/* ======================================================================
 *
 *  CCCCC          vr_input.dtrack.c
 * CC   CC         Author(s): Bill Sherman
 * CC              Created: November 4, 2010
 * CC   CC         Last Modified: February 24, 2024
 *  CCCCC
 *
 * Code file for FreeVR inputs from the DTrack input device.
 *
 * Copyright 2024, Bill Sherman, All rights reserved.
 * With the intent to provide an open-source license to be named later.
 * ====================================================================== */
/*************************************************************************
USAGE:
	The configuration of the inputs is similar to all input devices:

	Inputs are specified with the "input" option:
		input "<logical-name>" = "2switch(fs1[<number>]|fs2[<number>])";
		input "<logical-name>" = "valuator(fs2[<number>])"; -- NOTE: the old flysticks (fs1) had no valuators
		input "<logical-name>" = "6sensor(6body[<number>] {, 'id'|'r2e'|'xform ...'} {, 'ipos'})";
	 :-(	input "<logical-name>" = "6sensor(fs1[<number>] {, 'id'|'r2e'|'xform ...'} {, 'ipos'})";
		input "<logical-name>" = "6sensor(fs2[<number>] {, 'id'|'r2e'|'xform ...'} {, 'ipos'})";
	 :-[	input "<logical-name>" = "6sensor(3body[<number>] {, 'id'|'r2e'|'xform ...'} {, 'ipos'})";   NOTE: The parsing of 3d (3boday) data is done, but there's no option to configure it in FreeVR
	 :-(	input "<logical-name>" = "6sensor(mt[<number>] {, 'id'|'r2e'|'xform ...'} {, 'ipos'})";
	 :-(	input "<logical-name>" = "6sensor(gl[<number>] {, 'id'|'r2e'|'xform ...'} {, 'ipos'})";
	 :-(	input "<logical-name>" = "Nsensor(glove[<number>])"; -- There are DTrack glove devices, but I do not have access to one
	  XX	input "<logical-name>" = "Nswitch(switch[<number>])"; -- The DTrack has no concept of N-switch

	Controls are specified with the "control" option:
		control "<control option>" = "2switch(...)";
		control "<control option>" = "valuator(...)"; -- NOTE: no valuator oriented controls yet available for this device

	Here are the available control options for FreeVR:
		"print_help" -- print info on how to use the input device
		"system_pause_toggle" -- toggle the system pause flag
		"print_context" -- print the overall FreeVR context data structure (for debugging)
		"print_config" -- print the overall FreeVR config data structure (for debugging)
		"print_input" -- print the overall FreeVR input data structure (for debugging)
		"print_devinfo" -- print the internal DTrack data structure (for debugging)

	Here are the FreeVR configuration argument options for the DTrack:
		"port" - UDP socket port listening for data servers
			sending data.  (5000 is the default)
		"cmdHost" - machine running the ARTracking DTrack server.
			("localhost" is the default)
			NOTE: this is only necessary for the control features
			which currently aren't used.
		"cmdPort" -- port to send control commands (5001 is the default)
		"valScale" - scaling factor used to tune the valuator input range
		"trackScale"/"scale" - set the scaling factor for the location
			of the 6-sensor inputs.

	NOTE: presently the command to put the "DTrack" server into streaming
		mode via UDP is not implemented.  Thus the tracking stream must
		be manually started via the DTrack GUI interface.


	Using the "dtracktest" application (optional arguments are set by
	environment variables):
		% setenv DTRACK_HOST <value>
		% setenv DTRACK_PORT <value>
		% ./dtracktest


NOTES on the ARTracking "DTrack" protocol:

	The primary form of communication for the "DTrack" protocol is for
	the "DTrack" server to send tracking and other input information via
	UDP to a select set of hosts at port 5000.  Optionally, the server
	can be sent control commands via UDP to it's port 5001 (I suppose the
	server might send the data to itself, which would preclude the use
	of port 5000 for the commands).  Note that the ability to accept
	commands is a menu option in the DTrack server software, so be sure
	to enable this if you hope to have the end software (ie. FreeVR)
	make use of this feature.  The data can be sent either via an ASCII
	protocol or in binary form.  Thus far, only the ASCII format will be
	described and handled here.

	Commands can also be sent to port "50105" (at least for SMARTTRACK
	devices).

	Command protocol:
		There are only a handful of commands that are recognized by
		the "DTrack" server (at least that I'm aware of!):
			- "dtrack 10 0" -- deactivate (turn off) the cameras
			- "dtrack 10 1" -- activate (turn on) the cameras
			- "dtrack 10 3" -- activate the cameras and do reconstruction/calculation
			- "dtrack 31" -- start continuous update mode
			- "dtrack 32" -- stop continuous update mode
			- "dtrack 33 %d" -- send <N> data packets (i.e. poll mode)

		There is also a "new" protocol (DTrack2):
			- "dtrack2 get system version^@" -- respond with the DTrack version
			- "dtrack2 getmsg^@" -- get a status update?
			- "dtrack2 tracking start^@" -- start continuous update mode
			- "dtrack2 tracking stop^@" -- stop continuous update mode
			- "dtrack2 set <parameter>" -- ??

			NOTE: It's an error to send a "stop" when already
				stopped, and likewise for "start".  (These will
				produce an 'err -5 "function not allowed"' response.

			NOTE: I don't yet know what parameter settings are possible.

			NOTE: the NUL character (^@) is needed when using
				"telnet" to send commands, but not "ncat"/"nc".
				(Because "telnet" includes the newline in the
				sent message.)

			Return error codes:
			- "ok" -- no error
			- "err -1" -- receive timeout
			- "err -2" -- wrong system type
			- "err -3" -- command too long
			- "err -5" -- "function not allowed"
			- "err -9" -- broken tcp connection
			- "err -10" -- tcp connection invalid
			- "err -11" -- "command syntax error"
			NOTE: quotes indicate confirmed actual response

	Finding a SMARTRACK on a known subnet:
		tcsh% foreach i ( `loop 1 255` )
		foreach? echo $i
		foreach? echo -n "dtrack2 get system version" | nc -w 1 -v 192.168.5.$i 50105
		foreach? end

	ASCII Data protocol:
		Several data items (parcels?) are generally sent as a single
		packet, with a CR/NL newline pair ('\n') separating items:
			- "fr %d\n" -- frame counter
			- "ts %f\n" -- timestamp
			- "6dcal %d\n" -- number of calibrated 6-DOF (non-flystick)
				bodies -- see questions for ART below for additional info
			- "6d %d ..." -- standard 6-DOF bodies (ie. ones that
					only have tracking info)
				num-reports
				[id quality] [sx sy sz nu theta phi] [r0..r8 (3x3 rotation matrix)]
			- "6df %d ..." -- old flystick format
				num-reports
				[id quality button-mask][sx sy sz nu theta phi][r0..r8 (3x3 mat)]
			- "6df2 %d %d ..." -- new flystick format
				num-exist num-reports
				[id quality num-buttons num-controllers][sx sy sz][r0..r8 (3x3 mat)][bt0..btN ct0..ctN]
				NOTE: there are no Euler angles for the flystick2
				NOTE: there is one button integer for every 32
					buttons -- so normally just one.
		  :-(	- "6dmt %d ..." -- measurement tool
				num-reports
				[id quality unused][sx sy sz][r0..r8 (3x3 mat)]
		  :-(	- "gl %d ..." -- glove/finger format
				num-reports
				[id quality hand num-fingers][sx sy sz][r0..b8] {(for each fingertip) [sx sy sz][b0..b8][radius, phalange length, phalanx, phalanx angle, 2nd phalanx angle]}
		  :-(	- "glcal %d\n" -- number of defined hands
			- "3d %d ..." -- other 3d position (location-only) markers
				num-reports
				[id quality][sx sy sz]

		Questions for AR-Tracking:
			- are the unit identifiers unique across all types?
				A (from Andreas Werner): No (just as I expected)
			- why would there be a "6d 0\n" parcel in the sample data?
				A (from Andreas): not sure
			- what is the purpose of the "6dcal %d\n" parcel?
				(what does, does not get counted, and why
				 would this be used?)
				A (from Andreas): "6dcal gives the number of
				calibrated 6D objects (not Flysticks etc.) -
				some drivers wanted that information to
				allocate fields statically."
				FQ: why not flysticks?
			- how do I know how many button integers will be
				reported? (the manual explicitly says that
				"num_buttons" is number of buttons, ie. it
				doesn't suggest that it's the number of bytes)
				A (from Andreas): "There is one integer per
				32 buttons - and since none of our devices has
				that many buttons it will only be one. The
				format is just flexible so if we ever build a
				button with 33 buttons, there will be two."


		Where the rotation matrixii are defined as:
			    / b0  b3  b6 \
			R = | b1  b4  b7 |
			    \ b2  b5  b8 /

			NOTE: That the "6d" position trackers report both Euler angles and a 3x3
				rotation matrix, but the "6df2" flysticks only report the 3x3
				rotation matrix for the orientation.  For FreeVR, only the 3x3
				rotation matrix is used for creating the 4x4 position matrix
				regardless, so it doesn't really matter.

			NOTE: Applying the FreeVR vrMatrixCompositionFromMatrix() function to the
				3x3 matrix, and returns values that are the reverse of ROLL & ELEV,
				while the AZIM seems correct.

		Other notes:
			* it seems that with the flystick-1 units, when more
				than one button is pressed (including the
				china-hat), the button values (including the
				valuators) are all reported as 0.  So it's not
				even possible to push forward on the china-hat
				and push the trigger at the same time (and have
				them recognized) -- it will look as though no
				buttons are being pushed.  This is very bad.
				The good news is that the flystick-2 models do
				not have this limitation.
				NOTE: in my test setup, the data for both
				 models of flystick are reported as "6df2".
				A (from Andreas): Yes - the Fly-1 only supports
				one button at a time.

			* it seems as though the quality value is binary (1.0 or -1.0).
				A (from Andreas): No - that was intended for
				future use, currently it only tells you whether
				the Flystick is seen or not

			* it seems as though the 6d units (or at least the head
				tracker at the AVL cave) never report a quality
				other than 1.0, and when that unit is out of
				range, the position numbers freeze.  Whereas w/
				both of the flysticks (model 1 and model 2),
				... [what?]
				A (from Andreas): Yes: a 6d unit is not reported
				when it is not observed, but a Flystick might
				send button data when it is not observed.


HISTORY:
	04-08 November 2010 (Bill Sherman)
		Wrote initial version starting from vr_input.vruidd.c.  I
		  presently have the basic test program working -- it reads
		  the UDP data, stores it in the _DTrackPrivateInfo structure,
		  and then prints it.  Presently it only parses the two types
		  of position data I expect to get from most setups ("6d" and
		  "6df2"), but adding other types will be simple once the basic
		  FreeVR interface is completed.
		I have also defined the FreeVR interface, but not yet
		  implemented it.

	09 November 2010 (Bill Sherman)
		Continued writing the initial FreeVR interface.  Currently I'm
		  only parsing the "6d" and "6df2" position packets since I
		  don't have a lot of time right now, and I think those will
		  probably cover 98% of the use cases at the moment.  And then
		  I did some cleanup (got rid of the VruiDD stuff, etc.)

	11 November 2010 (Bill Sherman)
		Did first live tests of full FreeVR version on a CAVE with
		  AR-tracking.  From this made several bug fixes, and now the
		  interface works with "6d" and "6df2" input types.

		[corresponds to FreeVR Version 0.6b]

	5 December 2011 (Bill Sherman)
		Just some small adjustments to debug output styles.

		[corresponds to FreeVR Version 0.6c]

	2 September 2013 (Bill Sherman)
		Adopting a consistent usage of str<...>cmp() functions (using
		  logical negation).

	14 September 2013 (Bill Sherman)
		I am making use of the newly created SELFCTRL_DBGLVL for all
		  the self-control outputs.  I then added some reasonable
		  output to the "print_help" callback to print out what all
		  the device's inputs are doing.

	15 September 2013 (Bill Sherman)
		Changed the "0x%p" format to the improved "%#p" format.

		I added command line options:
			- "-list"  (synonym is "-nodata")

		[corresponds to FreeVR Version 0.6d]

	9 October 2020 (Bill Sherman)
		Some very minor edits to fix implicit function declarations
		  that had worked their way in.  Also the compiler is now
		  more picky.

		[corresponds to FreeVR Version 0.6g]

	20 October 2020 (Bill Sherman)
		I added comments to the top of all the functions that did
		  not yet have them.  (I also copied these over to the
		  "vr_input.skeleton.c" example code.)

		[corresponds to FreeVR Version 0.6g]

	22 January 2021 (Bill Sherman)
		Minor clean-up of the private information printing.  As part of
		  this, renamed "print_struct" callback to "print_devinfo", and
		  renamed _DTrackPrintStruct() to _DTrackPrintPrivateInfo().

	5 May 2021 (Bill Sherman)
		Minor clean-up to remove sprintf().

		[corresponds to FreeVR Version 0.7b]

	27 October 2021 (Bill Sherman)
		Added the 6-sensor initial position ("ipos") code to place
		  them at a reasonable default location prior to making a
		  network connection.
		Also did some general clean-up and change to more closely
		  match other position tracker input devices.

		[corresponds to FreeVR Version 0.7c]

	23 June 2022 (Bill Sherman) [At NIST]
		Began adding code for the "-screen" (renderscreen) option.
		  Copy-pasted code from "vruiddtest", and began changing
		  wording, but no functionality was completed.
		(Warning: "dtracktest" won't compile in this release!)

		[corresponds to FreeVR Version 0.7d]

	11 May 2023 (Bill Sherman) [At NIST]
		Worked more on the "-screen" option, and while not completing
			it, at least I have some values being reported.  (It
			does not yet do the moving "X" "Y", etc. values.)
		I came up with some ideas for improving the tracker implementation.

		There was a bug in my code where the EGT5 head tracker (2nd head
			tracker) is reported as a flystick.  At first it looked
			like the DTrack2 was sending that information, but I
			figured it was actually a parsing error.
			And indeed it was -- I figured it out -- I was going
			beyond the MAX_UNITS_PT value!
		I added some todo items to the list.
		(still have to do the actual graphical rendering, right now just
			printing numbers to particular locations on the screen.)

		NOTE: That the "6d" position trackers report both Euler angles
			and a 3x3 rotation matrix, but the "6df2" flysticks
			only report the 3x3 rotation matrix for the orientation.
			For FreeVR, only the 3x3 rotation matrix is used for
			creating the 4x4 position matrix regardless, so it
			doesn't really matter.

		[corresponds to FreeVR Version 0.7e]

	03 November 2023 (Bill Sherman)
		I improved the warning output for unknown parcels
			(e.g. "6dcal" & "3d").

		[corresponds to FreeVR Version 0.7e]

	12 February 2024 (Bill Sherman)
		Adding new top-comment sub-section with the "dtrack2" commands.
		I did a slight rearrange of the "-screen" render 6-dof tracker
		  values for the flystick to align it with the "regular" 6-dof
		  inputs.
		I fixed the terminal screen mode after a RENDERSCREEN display.

		[corresponds to FreeVR Version 0.7e]

	16-17 February 2024 (Bill Sherman)
		I added the moving-character feature to the "-screen" rendering
		  option.  Unlike past input device screen-rendering, I created
		  a new helper function that renders a complete column of data
		  for a single 6-dof input (including fly-sticks, of course).
		  (I'm probably going to expand that methodology to the other
		  input test programs that have a "-screen" option.

		[corresponds to FreeVR Version 0.7e]

	24 February 2024 (Bill Sherman)
		I wrote the CSV rendering output for "dtracktest".
		I added parsing for "3d" and "6dcal" input data.
		I added new command line arguments to "dtracktest":
		  - the "-once/-1" command line argument 
		  - the "-port/-p" command line argument 
		I'm also updating the man page.

		[corresponds to FreeVR Version 0.7e]

TODO:
	- DONE: Add the parsing of the "6dcal" input

	- DONE: Add the parsing of the "3d" input

	- DONE: Write a man page for "dtracktest"
		- NOTE: include information on how to find out what other
			program is bound to a UDP port:
			% netstat -nlup | grep 5000 (or whatever port)

	- include the man page material at the end of this source file

	- DONE: determine whether I'm parsing "6d" vs "6df2" packets correctly.
		(I think maybe I'm not, and so the 2nd "6d" unit is being
		reported as a "6df2" unit -- when printing the data from a
		"fr" (frame) report the data from DTrack2 looks correct.
		A: I was parsing them correctly, but I wasn't checking to see
		   whether new data was within the bounds of the MAX_UNITS_PT
		   value, so the 7th "6d" unit was being written into the
		   memory of the 2nd "6df2" unit!

	- Add a device field "anyframe" that reports the last frame number
		received from any of the devices -- that way we can know
		when a device is getting out of date relative to other
		devices.  (Though if "fr" frame packets are being received
		on a regular basis, that should be sufficient.)

	- Add a unit field "ever_reported" to allow us to remember that a
		device exists, and when it goes away we can "freeze" the
		reported values, and set the quality to "-1.0".
		- it looks like "fr" data is actually being reported, so this
			might not be necessary -- but could be interesting
			to print any negative deltas for devices

	- Add an output to the screen rendering (and I guess line rendering
		too, and why not, the CSV rendering when I get to it!) that
		indicates how many frames it's been since the last positive
		quality value -- or really what needs to be stored is the
		frame number of the last frame with a positive quality value
		(which may not be available, when only negative values have
		thus-far been received).

	- Add a timeout when trying to connect to a DTrack server that
		isn't there or isn't responding.

	- DONE: Add curses (aka "screen") rendering to "dtracktest" (the test program)

	- DONE: Add a CSV rendering style to "dtracktest" (the test program)

	- DONE: Add the ability to set the data-port in "dtracktest"

	- Complete/enhance the FreeVR driver:
		- See whether there's a clean way to only update inputs
			(especially buttons and valuators) when actual
			changes have occurred -- mostly that's important
			for buttons I would think.
	 -->	- Implement the other position types: (fs1, mt, gl, 3body)
		- Consider putting the time-stamp into the FreeVR data.
		- finish _DTrackPrintPrivateInfo()
		- fill in "version" and "op_params" fields
		- handle the "fs1" button case in _DTrackButtonInput()
		- add "quality" as a valuator value
		- add "quality" as a button value
		- add a feature to freeze flysticks (or any 6body)
			when the quality value is -1.0 -- right now the
			position data jumps to all zeros.

	- Add the ability to send control commands.  Currently, we assume
		that the user manually initiates the data stream from the
		"DTrack" server software GUI interface.

**************************************************************************/
#if !defined(FREEVR) && !defined(TEST_APP) && !defined(CAVE)
/* Choose how this code should be compiled (ie. for CAVE, standalone, etc). */
#  define	FREEVR
#  undef	TEST_APP
#  undef	CAVE
#endif

#undef COMM_DEBUG	/* define this to add a lot of output for debugging communications w/ the device */
#undef DEBUG_NOW	/* define this to add a little extra output for testing unimplemented parcels */


/* Header includes needed for the both the FreeVR library, and test utility */
#include <stdio.h>		/* needed for printf & all other standard I/O */
#include <ctype.h>		/* for isprint() */
#include <stdint.h>		/* needed for uint16_t & uint32_t types */
#include <stdlib.h>		/* needed for atoi() */ /* by why am I not using vrAtoI()? -- test on a DTrack at NIST and then hopefully delete me */
#include <fcntl.h>		/* for socket parameter settings */
#include <string.h>		/* for strchr(), strspn() */

#include "vr_socket.h"		/* Tracking data is communicated over sockets */
#include "vr_debug.h"		/* Standard FreeVR debugging features */
#include "vr_math.h"		/* for VR_{X,Y,Z,W,AZIM,ELEV,ROLL} */


/* Header includes needed for the FreeVR library, but not the test utility */
#if defined (FREEVR)
#  include "vr_input.h"
#  include "vr_input.opts.h"
#  include "vr_parse.h"
#  include "vr_utils.h"		/* for vrAtoI() */
#endif


/* Header includes needed just for the "dtracktest" utility */
/* NOTE: this includes some FreeVR *source code* */
#if defined(TEST_APP) /* { */
#  include <stdlib.h>		/* needed for getenv() & atoi() */
#  include <signal.h>		/* for catching interrupt signal */
#  include "vr_enums.h"
#  include "vr_socket.c"	/* for vrSocketCreateListenUDP & vrSocketClose*/
#  include "vr_utils.c"		/* needed for vrSleep() (temporary) */
#  define MATH_ONLY		/* exclude the FreeVR material in vr_math.c  */
#  include "vr_math.h"          /* for VR_{X,Y,Z,W,AZIM,ELEV,ROLL} */
#  include "vr_math.c"          /* needed for matrix to components conversion */
#  define vrShmemAlloc malloc
#  define vrShmemFree free

/*****************************************************************************/
/* vrInputDevice is a structure type used in the test application (TEST_APP) */
/*   to mimic the type from vr_input.h, and allowing the test app to make    */
/*   use of the FreeVR input callback functions.                             */
typedef struct vrID_st {
		char            *name;          /* CONFIG: name assigned to device in config file */
		int             operating;      /* whether the device is operating */
		void            *aux_data;      /* auxiliary data, specific to the type of input */
	} vrInputDevice;
#endif /* } TEST_APP */


/* Header includes for interfacing with the CaveLib */
/* NOTE: It's been over a decade since this has been tested */
#if defined(CAVE) /* { */
#  include <math.h>
#  include "cave.h"
#  include "cave.private.h"
#  include "cave.tracker.h"
static CAVE_SENSOR_ST initial_cave_sensor = {
		0,5,0,
		0,0,0,
		0,	/*timestamp*/
		FALSE,	/*calibrated*/
		CAVE_TRACKER_FRAME
	};

#  define	VR_X	0
#  define	VR_Y	1
#  define	VR_Z	2
#  define	VR_W	3
#  define	VR_AZIM	0
#  define	VR_ELEV	1
#  define	VR_ROLL	2
#endif /* } CAVE */


	/***************************************************/
	/*** definitions for interfacing with the device ***/
	/***                                             ***/

#define	BUFSIZE		1024

#define	DEFAULT_DATAPORT	5000
#define DEFAULT_CMDPORT		5001
#define	DEFAULT_CMDHOST		"127.0.0.1"


/* DTrack position tracker types */
#define DTRACK_TYPE_UNK		0
#define DTRACK_TYPE_6BODY	1
#define DTRACK_TYPE_3BODY	2
#define DTRACK_TYPE_FS1		3
#define DTRACK_TYPE_FS2		4
#define DTRACK_TYPE_MT		5
#define DTRACK_TYPE_GLOVE	6

/* DTrack sensitivity values */
#define TRANS_SENSITIVITY	0.001
#define ROT_SENSITIVITY		0.02
#define VALUATOR_SENSITIVITY	1.0



/****************************************************************/
/*** auxiliary structures of the current data from the device. ***/

	/* _DTrackUnit: contains data for a single unit of the ARTracking DTrack */
typedef struct {
		int		id;		/* the numeric identifier for this unit */
		int		type;		/* the type of incoming position data (e.g. standard 6-DOF body, flystick-2, 3-DOF body) */
		int		new;		/* a flag to indicate whether new data has arrived since this unit was last processed (NOTE: not currently setting the flag to 0 since that would prevent multiple values from the same tracker from registering) */
		int		frame;		/* the frame in which this data was received (0 if no data for this device) */
		float		time_stamp;	/* the time stamp for when this data was received */
		float		quality;	/* the quality of this data */
		int		num_buttons;	/* the number of buttons for this unit */
		int		num_valuators;	/* the number of valuators for this unit */
		float		location[3];	/* the location of this unit */
		float		angles[3];	/* the orientation of this unit as reported by Euler angles */
		float		rotation[9];	/* the orientation of this unit (elements of a 3x3 matrix) */
		uint32_t	buttons;	/* information of the buttons */
		float		valuators[32];	/* information of the valuators */
		int		active;		/* a flag to indicate whether position data is being received for this unit */
#ifdef CAVE
		/* CAVE specific fields here */

#elif defined(FREEVR)
		/* FREEVR specific fields here */
		vr6sensor	sensor6_input;	/* the assembled 6-dof data */

#endif /* end library-specific fields */
	} _DTrackUnit;

	/* _DTrackPrivateInfo: Overall data for the system */
typedef struct {
		/* these are for interfacing with the hardware */
		int		fd_socket;		/* communication file descriptor */
		char		*server_host;		/* name of daemon's host */
		int		server_port;		/* port on host listening for application commands */
		int		data_port;		/* port on this machine to which data will be sent from the server */
		int		open;			/* flag with DTrack successfully open */

		/* these are for internal data parsing */
		char		buf[BUFSIZE];
		int		eobuf_pos;
		char		version[256];		/* self-reported version of the device */
		char		op_params[256];		/* operating parameters of the device (according to it) */

		/* information about the current values */
		int		frame;			/* the number of the last frame of data received */
		float		time_stamp;		/* the last time stamp received */
		int		num_6dcal_bodies;	/* the number of 6-DOF calibrated (non-flystick) bodies known to the system */
#define MAX_UNITS_PT 15	/* the maximum number of units per position input type */
		_DTrackUnit	units_6body[MAX_UNITS_PT];	/* an array of all the 6body units */
		_DTrackUnit	units_fs2[MAX_UNITS_PT];	/* an array of all the flystick-2 units */

		/* information about how to filter the data for use with the VR library (FreeVR or CAVE) */
		float		scale_valuator;		/* scaling factor for valuators */
		float		scale_trans;		/* multiplier to scale from the DTrack location units (inches) */

		/* information about the inputs and mappings */
#ifdef CAVE
		/* CAVE specific fields here */

#elif defined(FREEVR)
		/* FREEVR specific fields here */
		int		num_buttons;		/* it is often wise to store the number of button inputs in addition to knowing the maximum possible */
		vr2switch	**button_inputs;	/* the maximum number of buttons is devinfo->num_2ways + devinfo->num_scontrols (all the self-controls are not necessarily buttons) */
		int		*button_map_bit;	/* An array of indices mapping from the server button id to the FreeVR logical id */
		int		*button_map_tracker;	/* An array of indices mapping from the FreeVR logical id to the tracker containing the data */
		int		*button_map_type;	/* An array of DTrack position tracker types for which the buttons are mapped from */

		int		num_switches;
		vrNswitch	**switch_inputs;

		int		num_valuators;
		vrValuator	**valuator_inputs;
		int		*valuator_map_valuator;	/* An array of indices mapping from the server valuator id to the FreeVR logical id */
		int		*valuator_map_tracker;	/* An array of indices mapping from the FreeVR logical id to the tracker containing the data */
		int		*valuator_map_type;	/* An array of DTrack position tracker types for which the valuators are mapped from */
		float		*valuator_map_sign;	/* An array indicating the sign of each valuator input (a multiplier so either "1.0" or "-1.0") */

		int		num_6sensors;
		vr6sensor	**sensor6_inputs;
		int		*tracker_map_tracker;	/* An array of indices mapping from the server receiver id to the FreeVR logical id */
		int		*tracker_map_type;	/* An array of DTrack position tracker types for which the trackers are mapped from */

		int		num_Nsensors;
		vrNsensor	**sensorN_inputs;

#endif /* end library-specific fields */

	} _DTrackPrivateInfo;



	/*********************************************/
	/*** General NON public interface routines ***/
	/*********************************************/

/*********************************************************/
/* _DTrackInputTypeName(): Convert from internal defined */
/*   enumerated values to ART DTrack string codes.       */
char *_DTrackInputTypeName(int type)
{
	switch (type) {
	case DTRACK_TYPE_UNK:	return "unknown";
	case DTRACK_TYPE_6BODY:	return "6body";
	case DTRACK_TYPE_3BODY:	return "3body";
	case DTRACK_TYPE_FS1:	return "fs1";
	case DTRACK_TYPE_FS2:	return "fs2";
	case DTRACK_TYPE_MT:	return "mt";
	case DTRACK_TYPE_GLOVE:	return "gl";
	}

	return "unknown";
}


/*****************************************************************/
/* _DTrackInputTypeValue(): Convert from ART DTrack string codes */
/*   to enumerated values defined in this source file for        */
/*   interfacing through FreeVR.                                 */
int _DTrackInputTypeValue(char *name)
{
	if      (!strcasecmp(name, "unknown"))	return DTRACK_TYPE_UNK;
	else if (!strcasecmp(name, "6body"))	return DTRACK_TYPE_6BODY;
	else if (!strcasecmp(name, "6d"))	return DTRACK_TYPE_6BODY;
	else if (!strcasecmp(name, "3body"))	return DTRACK_TYPE_3BODY;
	else if (!strcasecmp(name, "3d"))	return DTRACK_TYPE_3BODY;
	else if (!strcasecmp(name, "fs1"))	return DTRACK_TYPE_FS1;
	else if (!strcasecmp(name, "6df"))	return DTRACK_TYPE_FS1;
	else if (!strcasecmp(name, "fs2"))	return DTRACK_TYPE_FS2;
	else if (!strcasecmp(name, "6df2"))	return DTRACK_TYPE_FS2;
	else if (!strcasecmp(name, "mt"))	return DTRACK_TYPE_MT;
	else if (!strcasecmp(name, "gl"))	return DTRACK_TYPE_GLOVE;
	else if (!strcasecmp(name, "glove"))	return DTRACK_TYPE_GLOVE;
	else {
		vrErrPrintf("Unknown position receiver type '%s'\n", name);
		return DTRACK_TYPE_6BODY;
	}

	/* Can't get to this statement */
	return DTRACK_TYPE_6BODY;
}


/*************************************************************************/
/* _DTrackInitializeStruct(): sets initial values in the device specific */
/*   internal structure to prepare for making the initial connection.    */
/*   For some devices, memory will be allocated for arrays of values.    */
/*                                                                       */
/* NOTE: Some (few) versions of this routine take a second "typename"    */
/*   argument that can be used to specify a particular device among many */
/*   that share (more or less) the same protocol.  The typename is then  */
/*   used to determine what specific features are available with this    */
/*   particular type of device.                                          */
static void _DTrackInitializeStruct(_DTrackPrivateInfo *aux, char *typename)
{
	aux->version[0] = '\0';
	aux->op_params[0] = '\0';

	aux->buf[0] = '\0';
	aux->eobuf_pos = 0;

	aux->data_port = DEFAULT_DATAPORT;
	aux->server_port = DEFAULT_CMDPORT;
	aux->server_host = vrShmemStrDup(DEFAULT_CMDHOST);

	aux->scale_valuator = VALUATOR_SENSITIVITY;	/* set the default valuator scaling factor */
	aux->scale_trans = 1.0/304.8;			/* convert from mm to feet by default */

	aux->frame = -1;
	aux->time_stamp = -1;
	aux->num_6dcal_bodies = -1;

	/* everything else is zero'd by default */
}


/**************************************************************************/
/* _DTrackPrintPrivateInfo(): outputs the contents of the device specific */
/*   internal structure to help with configuration and system debugging.  */
/*   Output verbosity can be controlled by the "style" argument.          */
static void _DTrackPrintPrivateInfo(FILE *file, _DTrackPrivateInfo *aux, vrPrintStyle style)
{
	int	count;
	int	ptv_count;

	vrFprintf(file, "DTrack input-device internal structure (%p):\n", aux);
	vrFprintf(file, "\r\tversion -- '%s'\n", aux->version);
	vrFprintf(file, "\r\toperating parameters -- '%s'\n", aux->op_params);
	vrFprintf(file, "\r\tfd_socket = %d\n\tdata port = %d\n\tserver host = '%s'\n\tserver port = %d\n\topen = %d\n",
		aux->fd_socket,
		aux->data_port,
		aux->server_host,
		aux->server_port,
		aux->open);

	/* print the frame info */
	vrFprintf(file, "\r\tframe num = %d\n", aux->frame);
	vrFprintf(file, "\r\ttime stamp= %f\n", aux->time_stamp);
	vrFprintf(file, "\r\tnum 6dcal (not-fs) bodies= %d\n", aux->num_6dcal_bodies);

	/* print the filter values */
	vrFprintf(file, "\r\tscale_valuator = %f\n", aux->scale_valuator);
	vrFprintf(file, "\r\tscale_trans = %f\n", aux->scale_trans);

	/* print the raw values */
	/* TODO: print the raw values */

#ifdef FREEVR /* { */
	/* print the FreeVR input objects */
	vrFprintf(file, "\r\tbutton inputs:\n");
	for (count = 0; count < aux->num_buttons; count++)
		vrFprintf(file, "\r\t\tbutton_input[%d] = %p%s\n", count, aux->button_inputs[count], ((aux->button_inputs[count]->input_type == VRINPUT_CONTROL) ? " (*control*)" : ""));
	/* TODO: print position-tracker values */

	vrFprintf(file, "\r\tvaluator inputs:\n");
	for (count = 0; count < aux->num_valuators; count++)
		vrFprintf(file, "\r\t\tvaluator_input[%d] = %p%s\n", count, aux->valuator_inputs[count], ((aux->valuator_inputs[count]->input_type == VRINPUT_CONTROL) ? " (*control*)" : ""));

#if 0 /* NOT YET IMPLEMENTED */
	vrFprintf(file, "\r\t6sensor inputs (active = %d):\n", aux->active_sim6sensor);
	for (count = 0; count < aux->num_6sensors; count++)
		vrFprintf(file, "\r\t\t6sensor_inputs[%d] = %p\n", count, aux->sensor6_inputs[count]);
#endif

#  if 0 /* TODO: not sure if we want/need this */
	vrFprintf(file, "\r\tcontrol inputs:\n");
	for (count = 0; count < MAX_CONTROLS; count++)
		vrFprintf(file, "\r\t\tcontrol_inputs[%d] = %p\n", count, aux->control_inputs[count]);
#  endif
#endif /* } FREEVR */

	/* TODO: print some info about the current values */
}


/****************************************************************************/
/* _DTrackPrintHelp(): outputs the mappings from the physical device to the */
/*    virtual inputs and controls to help the user understand the interface.*/
static void _DTrackPrintHelp(FILE *file, _DTrackPrivateInfo *aux)
{
	int	count;		/* looping variable */

#if 0 || defined(TEST_APP)
	vrFprintf(file, BOLD_TEXT "Sorry, DTrack - print_help control not yet implemented.\n" NORM_TEXT);
#else
	vrFprintf(file, BOLD_TEXT "DTrack - inputs:" NORM_TEXT "\n");
	for (count = 0; count < aux->num_buttons; count++) {
		if (aux->button_inputs[count] != NULL) {
			vrFprintf(file, "\t%s -- %s%s\n",
				aux->button_inputs[count]->my_object->desc_str,
				(aux->button_inputs[count]->input_type == VRINPUT_CONTROL ? "control:" : ""),
				aux->button_inputs[count]->my_object->name);
		}
	}
	for (count = 0; count < aux->num_valuators; count++) {
		if (aux->valuator_inputs[count] != NULL) {
			vrFprintf(file, "\t%s -- %s%s\n",
				aux->valuator_inputs[count]->my_object->desc_str,
				(aux->valuator_inputs[count]->input_type == VRINPUT_CONTROL ? "control:" : ""),
				aux->valuator_inputs[count]->my_object->name);
		}
	}
	for (count = 0; count < aux->num_6sensors; count++) {
		if (aux->sensor6_inputs[count] != NULL) {
			vrFprintf(file, "\t%s -- %s%s\n",
				aux->sensor6_inputs[count]->my_object->desc_str,
				(aux->sensor6_inputs[count]->input_type == VRINPUT_CONTROL ? "control:" : ""),
				aux->sensor6_inputs[count]->my_object->name);
		}
	}
#endif
}


/***************************************************************************/
/* _DTrackReadInput(): function reads data from the socket, and then looks */
/*   for and parses packets from the daemon.  The data is then placed into */
/*   the generic portion of the "_DTrackPrivateInfo" structure.            */
/* NOTE: _DTrackReadInput() is placed above the _DTrackInitializeDevice() */
/*   function to allow the latter to call the former.                      */
static int _DTrackReadInput(_DTrackPrivateInfo *aux)
/* Returns number of packets read this time (-1 indicates reception of the STOPSTREAM_REPLY message) */
{
	//uint16_t	message;		/* VRUI VRDeviceDaemon messages are 2 byte sequences */ /* VruiDD-probably delete me */
	int		read_result;		/* return value of the call to "read()" */
	int		packets_decoded = 0;	/* return the number of packets that are decoded */
	int		num_bodies;		/* the number of position data received for a given parcel */
	int		count;			/* loop counter */
	int		body_count;		/* loop counter over number of bodies */
	int		shift;			/* number of bytes to shift after parsing */
	int		bytes;			/* number of bytes parsed in a particular operation */
	int		parsed_bytes;		/* number of bytes parsed thus far */
	int		return_count;		/* the number of items that were parsed via sscanf() */
	_DTrackUnit	tempunit;		/* a holding place for incoming unit data */

	do {
		read_result = (ssize_t)read(aux->fd_socket, &aux->buf[aux->eobuf_pos], sizeof(aux->buf) - aux->eobuf_pos);
		if (read_result > 0) {
			aux->eobuf_pos += read_result;
			aux->buf[aux->eobuf_pos] = '\0';
#if defined(COMM_DEBUG) || defined(DEBUG_NOW) || 0
			vrDbgPrintf("_DTrackReadInput(): %d bytes currently in buffer.\n", aux->eobuf_pos);
			vrPrintf("%3d bytes (%d): %s\n", (int)read_result, aux->eobuf_pos, aux->buf);
#endif
		}
	} while (read_result > 0);
#if defined(COMM_DEBUG) || 0
if (aux->eobuf_pos > 0)
	vrPrintf("_DTrackReadInput():  about to decode: eofbuf[pos-1] = '%02x', eobuf_pos = %d\n", aux->buf[aux->eobuf_pos-1], aux->eobuf_pos);
#endif

	/* The type of packet is determined by the first 2 to 5 characters, but even the 2-character */
	/*   messages have followon data, so there's nothing to parse until we get at least 5 chars. */
#if defined(DEBUG_NOW) && 0	/* delete me */
vrPrintf("HEYHEYHEY buf-pos = %d, size = %d\n", aux->eobuf_pos, sizeof(aux->buf) - aux->eobuf_pos);
vrPrintf("memchr = %p\n", memchr(aux->buf, '\n', aux->eobuf_pos));
vrPrintf("Done with memchr\n");
#endif
	while ((aux->eobuf_pos >= 5) && (memchr(aux->buf, '\n', aux->eobuf_pos) != NULL)) { /* TODO: make sure this isn't an off-by-one bug */
#if defined(DEBUG_NOW) && 0	/* delete me */
		vrPrintf("_DTrackReadInput(): memchr returning %p, aux->eobuf_pos = %d, size = %d\n", memchr(aux->buf, '\n', aux->eobuf_pos), aux->eobuf_pos, sizeof(aux->buf));
#endif
		/**********************************/
		/*** parcel type: frame -- "fr" ***/
		if (!strncmp(aux->buf, "fr ", 3)) {
#if defined(DEBUG_NOW) || 0	/* delete me */
printf("got fr parcel -- %s\n", vrPrintableString(aux->buf));
#endif
			/* the frame number */
			aux->frame = atoi(&aux->buf[3]);

			/* adjust the buffer */
			shift = (void *)index(aux->buf, '\n') - (void *)aux->buf + 1;
			memmove(aux->buf, &aux->buf[shift], aux->eobuf_pos - shift + 1);				/* shift the data by one packet */
			aux->eobuf_pos -= shift;
			packets_decoded++;

		/***************************************/
		/*** parcel type: time stamp -- "ts" ***/
		} else if (!strncmp(aux->buf, "ts ", 3)) {
#ifdef DEBUG_NOW	/* delete me */
printf("got ts parcel -- %s\n", vrPrintableString(aux->buf));
#endif
			/* a time stamp */
			aux->time_stamp = atof(&aux->buf[3]);

			/* adjust the buffer */
			shift = (void *)index(aux->buf, '\n') - (void *)aux->buf + 1;
			memmove(aux->buf, &aux->buf[shift], aux->eobuf_pos - shift + 1);				/* shift the data by one packet */
			aux->eobuf_pos -= shift;
			packets_decoded++;

		/***************************************/
		/*** parcel type: time stamp -- "6dcal" ***/
		} else if (!strncmp(aux->buf, "6dcal ", 6)) {
			/* a count of the number of calibrated 6-DOF (non-flystick) bodies known to DTrack */
			aux->num_6dcal_bodies = atof(&aux->buf[6]);

			/* adjust the buffer */
			shift = (void *)index(aux->buf, '\n') - (void *)aux->buf + 1;
			memmove(aux->buf, &aux->buf[shift], aux->eobuf_pos - shift + 1);				/* shift the data by one packet */
			aux->eobuf_pos -= shift;
			packets_decoded++;

		/*********************************************/
		/*** parcel type: standard bodies --  "6d" ***/
		} else if (!strncmp(aux->buf, "6d ", 3)) {
#ifdef DEBUG_NOW	/* delete me */
printf("got 6d parcel -- %s\n", vrPrintableString(aux->buf));
#endif
			/* a standard body */
			num_bodies = atof(&aux->buf[3]);
			parsed_bytes = 3; /* the first 3 bytes have been parsed */

			/* First assign all standard bodies to have inactive status -- and then we'll set the ones that provide data with active status */
			/* NOTE: the ways that active/inactive is handled for standard bodies is different that for flysticks -- just the nature of how AR-tracking defined the protocol. */
			for (body_count = 0; body_count < MAX_UNITS_PT; body_count++) {
				aux->units_6body[body_count].active = 0;
			}

			/* Next get the number of bodies */
			return_count = sscanf(&(aux->buf[parsed_bytes]), "%d%n", &num_bodies, &bytes);
			parsed_bytes += bytes;

			for (body_count = 0; body_count < num_bodies; body_count++) {
				memset(&tempunit, 0, sizeof(tempunit));
				tempunit.type = DTRACK_TYPE_6BODY;
				tempunit.frame = aux->frame;
				tempunit.time_stamp = aux->time_stamp;
				tempunit.new = 1;
				tempunit.active = 1;

				/* parse the first section (info) */
				return_count = sscanf(&(aux->buf[parsed_bytes]), " [%d %f]%n", &tempunit.id, &tempunit.quality, &bytes);
				tempunit.num_buttons = 0;
				tempunit.num_valuators = 0;
				parsed_bytes += bytes;

				/* parse the second section (location & angles) */
				return_count += sscanf(&(aux->buf[parsed_bytes]), "[%f %f %f %f %f %f]%n",
					&tempunit.location[VR_X], &tempunit.location[VR_Y], &tempunit.location[VR_Z],
					&tempunit.angles[0], &tempunit.angles[1], &tempunit.angles[2], &bytes);
				parsed_bytes += bytes;

				/* parse the third section (3x3 rotation) */
				return_count += sscanf(&(aux->buf[parsed_bytes]), "[%f %f %f %f %f %f %f %f %f]%n",
					&tempunit.rotation[0], &tempunit.rotation[1], &tempunit.rotation[2],
					&tempunit.rotation[3], &tempunit.rotation[4], &tempunit.rotation[5],
					&tempunit.rotation[6], &tempunit.rotation[7], &tempunit.rotation[8], &bytes);
				parsed_bytes += bytes;

				/* save the data into the generic space */
				if (aux->units_6body[tempunit.id].frame == tempunit.frame) {
					vrDbgPrintf("_DTrackReadInput(): Warning: 6d overwriting duplicate frame (%d) for id = %d.\n", tempunit.frame, tempunit.id);
				}
				aux->units_6body[tempunit.id] = tempunit;
			}

			/* adjust the buffer */
			shift = (void *)index(aux->buf, '\n') - (void *)aux->buf + 1;
			if (shift < parsed_bytes) {
				vrDbgPrintf("_DTrackReadInput(): Warning: 6d parsed more bytes (%d) than the shift (%d).\n", parsed_bytes, shift);
			}
			memmove(aux->buf, &aux->buf[shift], aux->eobuf_pos - shift + 1);				/* shift the data by one packet */
			aux->eobuf_pos -= shift;
			packets_decoded++;

		/******************************************/
		/*** parcel type: flystick-2 --  "6df2" ***/
		} else if (!strncmp(aux->buf, "6df2 ", 5)) {
			/* a flystick (new format) */
			/* num-??? num-bodies {[id qual num-buttons num-controllers][sx sy sz][r0..r8 (3x3 mat)][bt0..btN ct0..ctN]}* */
			parsed_bytes = 5; /* the first five bytes have been parsed */
#ifdef DEBUG_NOW	/* delete me */
printf("got 6df2 parcel -- %s\n", vrPrintableString(aux->buf));
#endif

			/* First assign all flystick-2's to have inactive status -- and then we'll set the ones that provide data with active status */
			for (body_count = 0; body_count < MAX_UNITS_PT; body_count++) {
				aux->units_fs2[body_count].active = 0;
			}

			/* Next get the number of bodies */
			/* TODO: I'm not sure what the first number means! (so it's being ignored for now) */
			return_count = sscanf(&(aux->buf[parsed_bytes]), "%*d %d%n", &num_bodies, &bytes);
			parsed_bytes += bytes;

			for (body_count = 0; body_count < num_bodies; body_count++) {
				memset(&tempunit, 0, sizeof(tempunit));
				tempunit.type = DTRACK_TYPE_FS2;
				tempunit.frame = aux->frame;
				tempunit.time_stamp = aux->time_stamp;
				tempunit.new = 1;

				/* parse the first section (info) */
				return_count += sscanf(&(aux->buf[parsed_bytes]), " [%d %f %d %d]%n",
					&tempunit.id, &tempunit.quality, &tempunit.num_buttons, &tempunit.num_valuators, &bytes);
				parsed_bytes += bytes;

				/* Now set the active status based on the quality -- NOTE: this is different than for standard bodies */
				if (tempunit.quality > 0.0)
					tempunit.active = 1;
				else	tempunit.active = 0;

				/* parse the second section (positions) */
				return_count += sscanf(&(aux->buf[parsed_bytes]), "[%f %f %f][%f %f %f %f %f %f %f %f %f]%n",
					&tempunit.location[VR_X], &tempunit.location[VR_Y], &tempunit.location[VR_Z],
					&tempunit.rotation[0], &tempunit.rotation[1], &tempunit.rotation[2],
					&tempunit.rotation[3], &tempunit.rotation[4], &tempunit.rotation[5],
					&tempunit.rotation[6], &tempunit.rotation[7], &tempunit.rotation[8], &bytes);
				parsed_bytes += bytes;

				/* parse the third section (buttons & valuators) */
				/* NOTE: it's important to scan that last closing square-bracket since we're in a loop and there may be more data to parse */
				return_count += sscanf(&(aux->buf[parsed_bytes]), "[%d %n", &tempunit.buttons, &bytes); /* NOTE: WARNING: if num_buttons>32, then more than one uint8 of buttons */
				parsed_bytes += bytes;
				for (count = 0; count < tempunit.num_valuators; count++) {
					return_count += sscanf(&(aux->buf[parsed_bytes]), "%f %n", &tempunit.valuators[count], &bytes);
					parsed_bytes += bytes;
				}
				return_count += sscanf(&(aux->buf[parsed_bytes]), "]%n", &bytes);
				parsed_bytes += bytes;

				/* save the data into the generic space */
				if (aux->units_fs2[tempunit.id].frame == tempunit.frame) {
					vrDbgPrintf("_DTrackReadInput(): Warning: 6df2 overwriting duplicate frame (%d) for id = %d.\n", tempunit.frame, tempunit.id);
				}
				aux->units_fs2[tempunit.id] = tempunit;
			}

#if 1 /* TODO: working on the bug at VisBox -- full buffers may cause a lack of CR character */	/* seemed fixed, delete me (ie. delete the #if, and the #else section) */
			/* adjust the buffer */
			shift = (void *)index(aux->buf, '\n') - (void *)aux->buf + 1;
			if (shift < parsed_bytes) {
				vrDbgPrintf("_DTrackReadInput(): Warning: 6df2 parsed more bytes (%d) than the shift (%d).\n", parsed_bytes, shift);
			}
#else
			shift = parsed_bytes;	/* this was my test code at Visbox -- with the "improved" while statement, this hack is not needed. */
#endif
			memmove(aux->buf, &aux->buf[shift], aux->eobuf_pos - shift + 1);				/* shift the data by one packet */
			aux->eobuf_pos -= shift;
			packets_decoded++;

		/*********************************************/
		/*** parcel type: standard bodies --  "3d" ***/
		} else if (!strncmp(aux->buf, "3d ", 3)) {
			/* a 3-DOF only body */
			num_bodies = atof(&aux->buf[3]);
			parsed_bytes = 3; /* the first 3 bytes have been parsed */

			/* First assign all standard bodies to have inactive status -- and then we'll set the ones that provide data with active status */
			/* NOTE: the ways that active/inactive is handled for standard bodies is different that for flysticks -- just the nature of how AR-tracking defined the protocol. */
			for (body_count = 0; body_count < MAX_UNITS_PT; body_count++) {
				aux->units_6body[body_count].active = 0;
			}

			/* Next get the number of bodies */
			return_count = sscanf(&(aux->buf[parsed_bytes]), "%d%n", &num_bodies, &bytes);
			parsed_bytes += bytes;

			for (body_count = 0; body_count < num_bodies; body_count++) {
				memset(&tempunit, 0, sizeof(tempunit));
				tempunit.type = DTRACK_TYPE_3BODY;
				tempunit.frame = aux->frame;
				tempunit.time_stamp = aux->time_stamp;
				tempunit.new = 1;
				tempunit.active = 1;

				/* parse the first section (info) */
				return_count = sscanf(&(aux->buf[parsed_bytes]), " [%d %f]%n",
					&tempunit.id, &tempunit.quality, &bytes);
				tempunit.num_buttons = 0;
				tempunit.num_valuators = 0;
				parsed_bytes += bytes;

				/* parse the second section (location) */
				return_count += sscanf(&(aux->buf[parsed_bytes]), "[%f %f %f]%n",
					&tempunit.location[VR_X], &tempunit.location[VR_Y], &tempunit.location[VR_Z], &bytes);
				parsed_bytes += bytes;

				/* save the data into the generic space */
				if (aux->units_6body[tempunit.id].frame == tempunit.frame) {
					vrDbgPrintf("_DTrackReadInput(): Warning: 6d overwriting duplicate frame (%d) for id = %d.\n", tempunit.frame, tempunit.id);
				}
				aux->units_6body[tempunit.id] = tempunit;
			}

			/* adjust the buffer */
			shift = (void *)index(aux->buf, '\n') - (void *)aux->buf + 1;
			if (shift < parsed_bytes) {
				vrDbgPrintf("_DTrackReadInput(): Warning: 3d parsed more bytes (%d) than the shift (%d).\n", parsed_bytes, shift);
			}
			memmove(aux->buf, &aux->buf[shift], aux->eobuf_pos - shift + 1);				/* shift the data by one packet */
			aux->eobuf_pos -= shift;
			packets_decoded++;

		} else {
			/* skip unknown/un-parsed data */
			shift = (void *)index(aux->buf, '\n') - (void *)aux->buf + 1;
#ifdef DEBUG_NOW /* delete me */
printf("\n_DTrackReadInput(): Unknown parcel, shifting %d bytes: '%s'\n", shift, vrPrintableString(aux->buf));
#endif
			vrDbgPrintf("\nYOYO_DTrackReadInput(): Warning: shifting %d bytes for unknown parcel ('", shift);
			for (count = 0; count < shift; count++) {
				vrDbgPrintf("%c", ((aux->buf[count] == '\n') ? 'N' : (aux->buf[count] == '\r') ? 'R' : aux->buf[count]));
			}
			vrDbgPrintf("' -- 0x");
			for (count = 0; count < shift; count++) {
				vrDbgPrintf("%02x ", (aux->buf[count]));
			}
			vrDbgPrintf(")\n");
			memmove(aux->buf, &aux->buf[shift], aux->eobuf_pos - shift + 1);				/* shift the data by one packet */
			aux->eobuf_pos -= shift;
			packets_decoded++;
		}
	}
#if defined(DEBUG_NOW) && 0 /* delete me */
	vrPrintf("_DTrackReadInput(): memchr returning %p, aux->eobuf_pos = %d, size = %d\n", memchr(aux->buf, '\n', aux->eobuf_pos), aux->eobuf_pos, sizeof(aux->buf));
#endif

#if defined(COMM_DEBUG) || defined(DEBUG_NOW) || 0 /* delete me */
if (packets_decoded > 0)
	vrPrintf("_DTrackReadInput(): %d packets decoded this call.\n", packets_decoded);
#endif
	return (packets_decoded);
}


/************************************************************/
/* _DTrackInitializeDevice() is called in the OPEN phase of */
/*   input interface -- after the types of inputs have been */
/*   determined (during the CREATE phase).                  */
/* TODO: In the future, this function could be used to      */
/*   initialize the DTrack server, such as putting it into  */
/*   streaming mode.                                        */
static int _DTrackInitializeDevice(_DTrackPrivateInfo *aux)
{
#if 0 /* From the VruiDD code, probably delete me */
	//uint16_t	message;	/* VRUI VRDeviceDaemon messages are 2 byte sequences */
	//uint32_t	value;		/* VRUI 4-byte value */
	//int		read_result;	/* return value of the call to "read()" */
	//int		bytes_read = 0;	/* number of bytes read from socket (the Vrui Daemon) thus far */
	//int		count;		/* loop counter */
#endif

	if (aux == NULL) {
		vrErrPrintf("_DTrackInitializeDevice(): " RED_TEXT "Warning, no auxiliary data for DTrack.\n" NORM_TEXT);
		return -1;
	}

	if (fcntl(aux->fd_socket, F_SETFL, O_RDWR | O_NONBLOCK | O_NOCTTY) < 0) {
		vrErrPrintf("_DTrackInitializeDevice(): An error occurred while trying to set the file controls.\n");
	}

	/***********************************************/
	/* TODO: When using server commands, this may be where they go */

	return 0;
}


/**********************************************************/
/* _DTrackCloseDevice() is called in the CLOSE phase      */
/*   of input interface -- just before the ports are      */
/*   closed, etc.                                         */
static int _DTrackCloseDevice(_DTrackPrivateInfo *aux)
{
#if 0 /* From the VruiDD code, probably delete me */
	uint16_t		message;	/* VRUI VRDeviceDaemon messages are 2 byte sequences */
#endif

	/* Don't close a device that's not open */
	if (aux->open == 0)
		return 0;

	/* TODO: determine what all needs to be done here */
	return -1;
}



	/*===========================================================*/
	/*===========================================================*/
	/*===========================================================*/
	/*===========================================================*/


#if defined(FREEVR) /* { */
/*****************************************************************/
/***  Functions for FreeVR access of DTrack devices for input  ***/
/*****************************************************************/


	/************************************/
	/***  FreeVR NON public routines  ***/
	/************************************/


/************************************************************************/
/* _DTrackParseArgs(): this function is used to read input device (or   */
/*    stream) configuration options from the FreeVR configuration file, */
/*    command-line or environment variable string.  The meaning and     */
/*    syntax for each option is described at the top of this source     */
/*    file (and will often be repeated in the man page documentation    */
/*    at the end.                                                       */
static void _DTrackParseArgs(_DTrackPrivateInfo *aux, char *args)
{
	float	scale_value = -1.0;		/* for reading one of the scaling factor values */

	/* In the rare case of no arguments, just return */
	if (args == NULL)
		return;

	/*****************************************/
	/** Argument format: "cmdHost" "=" file **/
	/*****************************************/
	vrArgParseString(args, "cmdhost", &(aux->server_host));

	/*******************************************/
	/** Argument format: "cmdPort" "=" number **/
	/*******************************************/
	vrArgParseInteger(args, "cmdport", &(aux->server_port));

	/*******************************************/
	/** Argument format: "port" "=" number **/
	/*******************************************/
	vrArgParseInteger(args, "port", &(aux->data_port));

#if 1 /* not sure whether we'll want this */
	/********************************************/
	/** Argument format: "valScale" "=" number **/
	/********************************************/
	if (vrArgParseFloat(args, "valscale", &(aux->scale_valuator))) {
		aux->scale_valuator = VALUATOR_SENSITIVITY;
	}
#endif

	/**********************************************/
	/** Argument format: "trackScale" "=" number **/
	/**********************************************/
	if (vrArgParseFloat(args, "trackscale", &scale_value)) {
		aux->scale_trans = scale_value;
	}

	/*****************************************/
	/** Argument format: "scale" "=" number **/ /* (redundant with above) */
	/*****************************************/
	if (vrArgParseFloat(args, "scale", &scale_value)) {
		aux->scale_trans = scale_value;
	}

	/** TODO: other arguments to parse go here **/
}


/**************************************************************************/
/* The _DTrackGetData(): function calls the _DTrackReadInput() function   */
/*   to get the latest data, and then puts that data into the FreeVR data */
/*   structures.  I.e. this is the FreeVR-specific portion of the data    */
/*   parsing.                                                             */
static void _DTrackGetData(vrInputDevice *devinfo)
{
	_DTrackPrivateInfo	*aux = (_DTrackPrivateInfo *)devinfo->aux_data;
	vrGenericInput		*input;			/* pointer to the input we're processing */
	int			num_inputs;		/* the number of input parcels parsed by _DTrackReadInput() */
	int			dtrack_num;		/* the particular button/valuator data to be extracted for the input */
	int			dtrack_tracker;		/* local storage of the mapping code from FreeVR input number to DTrack input number */
	int			dtrack_type;		/* the type of position input that the buttons/valuators are associated with */
	_DTrackUnit		*type_array;		/* pointer to a particular array of units */
	int			button_value;		/* holder for extracting the button flag from an integer */
	float			valuator_value;		/* holder for compositing a value together from components (e.g. sign & value) */
	float			scale_trans = aux->scale_trans;	/* a valuator scaling factor */
	int			count;			/* a loop counter */

	/*******************/
	/* gather the data */
	num_inputs = _DTrackReadInput(aux);			/* Make sure we have the latest data stored in the "_FastrakPrivateInfo" structure */

	/* There's no point in processing the inputs if we don't have anything new */
      if (num_inputs > 0) {

	/*************/
	/** buttons **/
	for (count = 0; count < aux->num_buttons; count++) {
		dtrack_num = aux->button_map_bit[count];
		dtrack_tracker = aux->button_map_tracker[count];
		dtrack_type = aux->button_map_type[count];
		switch(dtrack_type) {
		case DTRACK_TYPE_6BODY:	type_array = aux->units_6body;	break;
		case DTRACK_TYPE_3BODY:	type_array = NULL;		break;
		case DTRACK_TYPE_FS1:	type_array = NULL;		break;
		case DTRACK_TYPE_FS2:	type_array = aux->units_fs2;	break;
		case DTRACK_TYPE_MT:	type_array = NULL;		break;
		case DTRACK_TYPE_GLOVE:	type_array = NULL;		break;
		default:		type_array = NULL;		break;
		}
		if (dtrack_tracker >= MAX_UNITS_PT) {	/* NOTE: '>=' because dtrack_map is a zero-based number */
			vrErrPrintf(RED_TEXT "_DTrackGetData(): request for button (%d) not available with the DTrack interface (max %d).\n", dtrack_tracker, MAX_UNITS_PT-1 /* correct for zero-base */);
			/* TODO: I should look into creating a dummy input, and using that after the first warning is printed */
		} else if (type_array == NULL) {
			vrErrPrintf(RED_TEXT "_DTrackGetData(): unknown (or un-handled) tracker type (%d) when processing button (%d).\n", dtrack_type, dtrack_tracker);
		} else if (type_array[dtrack_tracker].new == 1) {
			/* handle button inputs as buttons or self-controls */
			input = (vrGenericInput *)(aux->button_inputs[count]);
			button_value = ((type_array[dtrack_tracker].buttons & (0x01 << dtrack_num)) != 0);
			if (input != NULL) {
				switch (input->input_type) {
				case VRINPUT_BINARY:
					vrAssign2switchValue((vr2switch *)(input), button_value);
					break;
				case VRINPUT_CONTROL:
					vrCallbackInvokeDynamic(((vrControl *)(input))->callback, 1, button_value);
					break;
				default:
					vrErrPrintf(RED_TEXT "_DTrackGetData(): Unable to handle button inputs that aren't Binary or Control inputs\n" NORM_TEXT);
					break;
				}
			}
		}
	} /* button loop */

	/***************/
	/** valuators **/
	for (count = 0; count < aux->num_valuators; count++) {
		dtrack_num = aux->valuator_map_valuator[count];
		dtrack_tracker = aux->valuator_map_tracker[count];
		dtrack_type = aux->valuator_map_type[count];
		switch(dtrack_type) {
		case DTRACK_TYPE_6BODY:	type_array = aux->units_6body;	break;
		case DTRACK_TYPE_3BODY:	type_array = NULL;		break;
		case DTRACK_TYPE_FS1:	type_array = NULL;		break;
		case DTRACK_TYPE_FS2:	type_array = aux->units_fs2;	break;
		case DTRACK_TYPE_MT:	type_array = NULL;		break;
		case DTRACK_TYPE_GLOVE:	type_array = NULL;		break;
		default:		type_array = NULL;		break;
		}
		if (dtrack_tracker >= MAX_UNITS_PT) {	/* NOTE: '>=' because dtrack_tracker is a zero-based number */
			vrErrPrintf(RED_TEXT "_DTrackGetData(): request for valuator (%d) not available with the DTrack interface (max %d).\n", dtrack_tracker, MAX_UNITS_PT-1 /* correct for zero-base */);
			/* TODO: I should look into creating a dummy input, and using that after the first warning is printed */
		} else if (type_array == NULL) {
			vrErrPrintf(RED_TEXT "_DTrackGetData(): unknown (or un-handled) tracker type (%d) when processing valuator (%d).\n", dtrack_type, dtrack_tracker);
		} else if (type_array[dtrack_tracker].new == 1) {
			/* handle valuator inputs as valuators or self-controls */
			input = (vrGenericInput *)(aux->valuator_inputs[count]);
			if (input != NULL) {
				valuator_value = type_array[dtrack_tracker].valuators[dtrack_num] * aux->scale_valuator * aux->valuator_map_sign[count];
				switch (input->input_type) {
				case VRINPUT_VALUATOR:
					vrAssignValuatorValue((vrValuator *)(input), valuator_value);
					break;
				case VRINPUT_CONTROL:
					vrCallbackInvokeDynamic(((vrControl *)(input))->callback, 1, &valuator_value);
					break;
				default:
					vrErrPrintf(RED_TEXT "_DTrackGetData(): Unable to handle valuator inputs that aren't Floating or Control inputs\n" NORM_TEXT);
					break;
				}
			}
		}
	} /* valuator loop */

	/***************/
	/** 6-sensors **/
	for (count = 0; count < aux->num_6sensors; count++) {
		vr6sensor	*current_6sensor;
		vrMatrix	new_position;

		dtrack_tracker = aux->tracker_map_tracker[count];
		dtrack_type = aux->tracker_map_type[count];
		switch(dtrack_type) {
		case DTRACK_TYPE_6BODY:	type_array = aux->units_6body;	break;
		case DTRACK_TYPE_3BODY:	type_array = NULL;		break;
		case DTRACK_TYPE_FS1:	type_array = NULL;		break;
		case DTRACK_TYPE_FS2:	type_array = aux->units_fs2;	break;
		case DTRACK_TYPE_MT:	type_array = NULL;		break;
		case DTRACK_TYPE_GLOVE:	type_array = NULL;		break;
		default:		type_array = NULL;		break;
		}
		if (dtrack_tracker >= MAX_UNITS_PT) {	/* NOTE: '>=' because dtrack_tracker is a zero-based number */
			vrErrPrintf(RED_TEXT "_DTrackGetData(): request for valuator (%d) not available with the DTrack interface (max %d).\n", dtrack_tracker, MAX_UNITS_PT-1 /* correct for zero-base */);
			/* TODO: I should look into creating a dummy input, and using that after the first warning is printed */
		} else if (type_array == NULL) {
			vrErrPrintf(RED_TEXT "_DTrackGetData(): unknown (or un-handled) tracker type (%d) when processing valuator (%d).\n", dtrack_type, dtrack_tracker);
		} else if (type_array[dtrack_tracker].new == 1) {
			/* handle 6-sensor inputs */
			current_6sensor = aux->sensor6_inputs[count];

			if (type_array[dtrack_tracker].active) {
				/* Assemble the new_position matrix */
				vrMatrixSetIdentity(&new_position);
				VRMAT_ROWCOL(&new_position, 0, 0) = type_array[dtrack_tracker].rotation[0];
				VRMAT_ROWCOL(&new_position, 1, 0) = type_array[dtrack_tracker].rotation[1];
				VRMAT_ROWCOL(&new_position, 2, 0) = type_array[dtrack_tracker].rotation[2];
				VRMAT_ROWCOL(&new_position, 0, 1) = type_array[dtrack_tracker].rotation[3];
				VRMAT_ROWCOL(&new_position, 1, 1) = type_array[dtrack_tracker].rotation[4];
				VRMAT_ROWCOL(&new_position, 2, 1) = type_array[dtrack_tracker].rotation[5];
				VRMAT_ROWCOL(&new_position, 0, 2) = type_array[dtrack_tracker].rotation[6];
				VRMAT_ROWCOL(&new_position, 1, 2) = type_array[dtrack_tracker].rotation[7];
				VRMAT_ROWCOL(&new_position, 2, 2) = type_array[dtrack_tracker].rotation[8];
				VRMAT_ROWCOL(&new_position, 0, 3) = type_array[dtrack_tracker].location[VR_X] * scale_trans;
				VRMAT_ROWCOL(&new_position, 1, 3) = type_array[dtrack_tracker].location[VR_Y] * scale_trans;
				VRMAT_ROWCOL(&new_position, 2, 3) = type_array[dtrack_tracker].location[VR_Z] * scale_trans;

				//VRMAT_ROWCOL(&new_position, 0, 3) = type_array[dtrack_tracker].frame;
				vrAssign6sensorValue(current_6sensor, &new_position, 0 /*, TODO: timestamp from daemon */);
			}
			vrAssign6sensorActiveValue(current_6sensor, type_array[dtrack_tracker].active);
		}
	} /* 6-sensor loop */

      } /* (num_inputs > 0) */
}


	/****************************************************************/
	/*    Function(s) for parsing DTrack "input" declarations.      */
	/*                                                              */
	/*  These _DTrack<type>Input() functions are called during the  */
	/*  CREATE phase of the input interface.                        */

/* ----------------------------------------------------------------------- */
/* The arguments for each of these "input" declaration parsers follow the  */
/*   same form, no matter what type of input or device:                    */
/*                                                                         */
/* devinfo -- a pointer to the overall structure that defines this device. */
/*            It is used to get access to the auxiliary data of the device.*/
/*                                                                         */
/* input -- a pointer to the vrGenericInput structure that is a specific   */
/*          FreeVR input, such as a single vr2switch, that is part of the  */
/*          specified input device.                                        */
/*          NOTE: the "input_type" can be determined by accessing the      */
/*          "input->input_type" field.  This can be used to have a single  */
/*          mapping function handle requests for multiple types of inputs. */
/*          See _XwindowsKeyboardInput() for an example of this.           */
/*                                                                         */
/* The remaining argument contains three strings that are from the portion */
/*   of the input config statement contained within the parentheses.  This */
/*   is pre-parsed as the "DTI" -- {<sub-Device>:}<Type>[<Instance>].  Any */
/*   of these subfields can be an empty string.  Usually:                  */
/*                                                                         */
/* device -- a string that may be used to indicate a sub-device of the     */
/*           provided hardware.  For example, and X11 input device may     */
/*           have "keyboard", "mouse" and "dial+buttons" subdevices.       */
/*           In most cases the "device" value is the empty string, and     */
/*           probably could always be included as part of the "type" value.*/
/*                                                                         */
/* type -- a string that indicates the class of input interface on the     */
/*         particular device/subdevice.  For example, an X11 input device  */
/*         may have "key", "mousebutton" and "pointerlocation" types.      */
/*         Or, a Magellan may have "button" and "6dof" types.              */
/*                                                                         */
/*         NOTE that the "type" value is also used in determining which of */
/*         these functions should be called (by matching the first element */
/*         of the vrInputFunctions list below), so it has already played a */
/*         roll in how this input is handled.  Therefore it's value is     */
/*         often ignored in the parsing function itself.  However, it is   */
/*         possible for more than one <type> string to point to the same   */
/*         device parsing function.  Thus in those cases it may be         */
/*         important to make use of this information within the parsing    */
/*         function.                                                       */
/*                                                                         */
/*         NOTE also that this string is different than the "input_type"   */
/*         data mentioned above as part of the "input" in that this is a   */
/*         string name that might describe the physical form of the input, */
/*         versus the "input_type" which describes a logical FreeVR input  */
/*         type such as a BINARY, VALUATOR or 6SENSOR input.               */
/*                                                                         */
/* instance -- a string usually used to indicate a specific instance of    */
/*             the type of data.  For example button "2" on a Magellan,    */
/*             unit "1" of a Flock of Birds, or key "Esc" of an X11        */
/*             keyboard.  Other data may also be included for altering how */
/*             this input is handled (See the VRPN device for an example). */
/*                                                                         */
/*    NOTE: since the "device" subfield is often optional, and the "type"  */
/*    subfield is used to determine which function to call, many of these  */
/*    functions will only make use of the "instance" subfield.             */
/*                                                                         */
/* The return values specify whether the initial match is confirmed, or    */
/*   rejected.  For some devices (such as the Xwindows-keyboard), there    */
/*   may be an overall match for the keyboard, but then the specific key   */
/*   instance may not match a known key for the keyboard.  In such cases   */
/*   a VRINPUT_MATCH_UNABLE is returned, indicating that the search failed,*/
/*   but that the search for a matching device should cease.  If the match */
/*   is successful, then a VRINPUT_MATCH_ABLE is returned.  If the match   */
/*   altogether fails, then a VRINPUT_NOMATCH is returned, and the search  */
/*   continues.                                                            */
/*                                                                         */
/* NOTE: these functions are called before making the physical connection  */
/*   to the device, so it is not possible to query features of the device  */
/*   within these functions.                                               */
/*                                                                         */
/* NOTE: the vrInputDTI parsing (done in vr_parse.c) is for the common     */
/*   form of how inputs are specified in the FreeVR configuration (rc)     */
/*   files.  Thus the following input specification:                       */
/*       input "2switch[left]" = "2switch(Joystick0:button[left])";        */
/*   would parse the devices input as:                                     */
/*       device (d) = "Joystick0"    ("sub-device" is probably more exact) */
/*       type (t) = "button"                                               */
/*       instance (i) = "left"                                             */
/* ----------------------------------------------------------------------- */


/***************************************************************************/
/* _DTrackButtonInput(): this function takes a "device,type,instance"      */
/*    (DTI) configuration triplet for a "button" class of input, parses    */
/*    the "DTI" triplet and maps it to a FreeVR input system virtual input */
/*    in the syntax of "button(<sub-device>:<type>[<instance>, <args>])"   */
/*    NOTE: for the DTrack device, there are no sub-device options.        */
/*                                                                         */
/* Here is a sample definition for a DTrack button:                        */
/*     input "button-2" = "2switch(fs2[0, 1])";                            */
/* And since the numbers are 0-based, this means the 2nd button of the     */
/*    1st flystick-2.                                                      */
static vrInputMatch _DTrackButtonInput(vrInputDevice *devinfo, vrGenericInput *input, vrInputDTI *dti)
{
	_DTrackPrivateInfo	*aux = (_DTrackPrivateInfo *)devinfo->aux_data;
	int			button_num;		/* the current count of FreeVR button inputs */
	int			button_choice;		/* the chosen number for this button */
	char			*button_name_instance;	/* pointer to the 2nd half of the instance string -- after the comma */
	int			unit_number;		/* the number of the given type of tracker (eg. the 2nd flystick-2) */
	int			tracker_type;		/* the type of DTrack tracker (e.g. a flystick-2) */

	/* The DTrack method of handling input arrays may be a combination of the way */ /* TODO: verify/reject this comment */
	/*   suggested by vr_input.skeleton.c as the FoB, MS & Fastrak input drivers. */

	/* TODO: determine whether type "fs1" or "fs2" */

	/*******************/
	/* select a button */
	button_num = aux->num_buttons;
	aux->num_buttons++;

	tracker_type = _DTrackInputTypeValue(dti->type);
	unit_number = vrAtoI(dti->instance);
	button_name_instance = strchr(dti->instance, ',') + 1;			/* jump past the comma */

	if (button_name_instance != NULL) {
		button_choice = vrAtoI(button_name_instance);
	} else {
		button_choice = -1;
	}

	/*****************************/
	/* check the selected button */
	if (button_choice < 0) {
		vrDbgPrintfN(CONFIG_WARN_DBGLVL, "_DTrackButtonInput(): Warning, button['%s'] did not match any known button\n", dti->instance);
		return VRINPUT_MATCH_UNABLE;	/* device match, but still unable to handle request */
	} else if (unit_number >= MAX_UNITS_PT) {
		vrDbgPrintfN(CONFIG_WARN_DBGLVL, "_DTrackButtonInput(): Warning, button['%s'] cannot be mapped due to insufficient position tracker units (%d)\n", dti->instance, MAX_UNITS_PT);
		return VRINPUT_MATCH_UNABLE;	/* device match, but still unable to handle request */
	} else if (aux->button_inputs[button_num] != NULL) {
		vrDbgPrintfN(CONFIG_WARN_DBGLVL, "_DTrackButtonInput(): " RED_TEXT "Warning, new input from %s[%s] overwriting old value.\n" NORM_TEXT,
			dti->type, dti->instance);
	}

	/********************/
	/* make the mapping */
	aux->button_inputs[button_num] = (vr2switch *)input;
	aux->button_map_bit[button_num] = button_choice;
	aux->button_map_tracker[button_num] = unit_number;
	aux->button_map_type[button_num] = tracker_type;

	vrDbgPrintfN(INPUT_DBGLVL, "_DTrackButtonInput(): assigned button event of value %d (DTrack button %d) to input pointer = %p)\n",
		button_num, button_choice, aux->button_inputs[button_num]);

	return VRINPUT_MATCH_ABLE;	/* input declaration match */
}


/***************************************************************************/
/* _DTrackValuatorInput(): this function takes a "device,type,instance"    */
/*    (DTI) configuration triplet for a "valuator" class of input, parses  */
/*    the "DTI" triplet and maps it to a FreeVR input system virtual input */
/*    in the syntax of "valuator(<sub-device>:<type>[<instance>, <args>])" */
/*    NOTE: for the DTrack device, there are no sub-device options.        */
/*                                                                         */
/* Here is a sample definition for a DTrack valuator:                      */
/*     input "joy-left-Y" = "valuator(fs2[0, -1])";                        */
/* And since the numbers are 0-based, this means the 2nd valuator of the   */
/*    1st flystick-2.  And do a sign-change on that valuator.              */
static vrInputMatch _DTrackValuatorInput(vrInputDevice *devinfo, vrGenericInput *input, vrInputDTI *dti)
{
static	char			*whitespace = " \t\r\b\n";
	_DTrackPrivateInfo	*aux = (_DTrackPrivateInfo *)devinfo->aux_data;
	int			valuator_num;		/* the current count of FreeVR valuator inputs */
	int			valuator_choice;	/* the chosen number for this valuator */
	char			*valuator_name_instance;/* pointer to the 2nd half of the instance string -- after the comma */
	int			unit_number;		/* the instance number of the given type of tracker (eg. the 2nd flystick-2) */
	int			tracker_type;		/* the type of DTrack tracker (e.g. a flystick-2) */

	/* The DTrack method of handling input arrays may be a combination of the way */
	/*   suggested by vr_input.skeleton.c as the FoB, MS & Fastrak input drivers. */

	/********************************/
	/* select a valuator (and sign) */
	valuator_num = aux->num_valuators;
	aux->num_valuators++;

	tracker_type = _DTrackInputTypeValue(dti->type);
	unit_number = vrAtoI(dti->instance);
	valuator_name_instance = strchr(dti->instance, ',') + 1;		/* jump past the comma */

	if (valuator_name_instance != NULL) {
		valuator_name_instance += strspn(valuator_name_instance, whitespace);/* skip white */
		if (valuator_name_instance[0] == '-') {
			aux->valuator_map_sign[valuator_num] = -1.0;
			valuator_choice = vrAtoI(&(valuator_name_instance[1]));	/* skip the negation sign */
		} else {
			aux->valuator_map_sign[valuator_num] =  1.0;
			valuator_choice = vrAtoI(valuator_name_instance);
		}
	} else {
		valuator_choice = -1;
	}

	/*******************************/
	/* check the selected valuator */
	if (valuator_choice == -1) {
		vrDbgPrintfN(CONFIG_WARN_DBGLVL, "_DTrackValuatorInput(): Warning, valuator['%s'] did not match any known valuator\n", dti->instance);
		return VRINPUT_MATCH_UNABLE;	/* device match, but still unable to handle request */
	} else if (unit_number >= MAX_UNITS_PT) {
		vrDbgPrintfN(CONFIG_WARN_DBGLVL, "_DTrackValuatorInput(): Warning, valuator['%s'] cannot be mapped due to insufficient position tracker units (%d)\n", dti->instance, MAX_UNITS_PT);
		return VRINPUT_MATCH_UNABLE;	/* device match, but still unable to handle request */
	} else if (aux->valuator_inputs[valuator_num] != NULL) {
		vrDbgPrintfN(CONFIG_WARN_DBGLVL, "_DTrackValuatorInput(): " RED_TEXT "Warning, new input from %s[%s] overwriting old value.\n" NORM_TEXT,
			dti->type, dti->instance);
	}

	/********************/
	/* make the mapping */
	aux->valuator_inputs[valuator_num] = (vrValuator *)input;
	aux->valuator_map_valuator[valuator_num] = valuator_choice;
	aux->valuator_map_tracker[valuator_num] = unit_number;
	aux->valuator_map_type[valuator_num] = tracker_type;

	vrDbgPrintfN(INPUT_DBGLVL, "_DTrackValuatorInput(): assigned valuator event of value %d (DTrack valuator %d of %s(%d)) to input pointer = %p)\n",
		valuator_num, valuator_choice, dti->type, tracker_type, aux->valuator_inputs[valuator_num]);

	return VRINPUT_MATCH_ABLE;	/* input declaration match */
}


/***************************************************************************/
/* _DTrack6sensorInput(): this function takes a "device,type,instance"     */
/*    (DTI) configuration triplet for a "6sensor" class of input, parses   */
/*    the "DTI" triplet and maps it to a FreeVR input system virtual input */
/*    in the syntax of "6sensor(<sub-device>:<type>[<instance>, <args>])"  */
/*    NOTE: for the DTrack device, there are no sub-device options.        */
/*                                                                         */
/* Here is a sample definition for a DTrack position sensor:               */
/*     input "tracker[1]" = "6sensor(fs2[0, r2e])";                        */
/* And since the numbers are 0-based, this means the 1st flystick-2 will   */
/*    be mapped to the name "tracker[1]", and it will have the current     */
/*    receiver-to-entity matrix applied to it.                             */
static vrInputMatch _DTrack6sensorInput(vrInputDevice *devinfo, vrGenericInput *input, vrInputDTI *dti)
{
	_DTrackPrivateInfo	*aux = (_DTrackPrivateInfo *)devinfo->aux_data;
	vrMatrix		mat_store;		/* temporary matrix storage */
	vr6sensor		*sensor6_input;		/* pointer to this new 6-sensor input */
	int			tracker_num;		/* the current count of FreeVR tracker inputs */
	int			tracker_choice;		/* the chosen number for this tracker */
	int			tracker_type;		/* the type of dtrack tracker (e.g. a flystick-2) */

	/* The DTrack method of handling input arrays may be a combination of the way */
	/*   suggested by vr_input.skeleton.c as the FoB, MS & Fastrak input drivers. */

	/********************/
	/* select a 6sensor */
	tracker_num = aux->num_6sensors;
	aux->num_6sensors++;

	tracker_type = _DTrackInputTypeValue(dti->type);
	tracker_choice = vrAtoI(dti->instance);

	/*************************************************/
	/* verify that the requested 'instance' is valid */
	if (tracker_choice == -1) {
		vrDbgPrintfN(CONFIG_WARN_DBGLVL, "_DTrack6sensorInput(): Warning, tracker['%s'] did not match any known tracker\n", dti->instance);
		return VRINPUT_MATCH_UNABLE;	/* device match, but still unable to handle request */
	} else if (tracker_choice >= MAX_UNITS_PT) {
		vrDbgPrintfN(CONFIG_WARN_DBGLVL, "_DTrack6sensorInput(): Warning, tracker['%s'] cannot be mapped due to insufficient position tracker units (%d)\n", dti->instance, MAX_UNITS_PT);
		return VRINPUT_MATCH_UNABLE;	/* device match, but still unable to handle request */
	} else if (aux->sensor6_inputs[tracker_num] != NULL) {
		vrDbgPrintfN(CONFIG_WARN_DBGLVL, "_DTrack6sensorInput(): " RED_TEXT "Warning, new input from %s[%s] overwriting old value.\n" NORM_TEXT,
			dti->type, dti->instance);
	}

	/********************/
	/* make the mapping */
	aux->sensor6_inputs[tracker_num] = (vr6sensor *)input;
	aux->tracker_map_tracker[tracker_num] = tracker_choice;
	aux->tracker_map_type[tracker_num] = tracker_type;
	sensor6_input = aux->sensor6_inputs[tracker_num];

	/* vrAssign6sensorR2Exform() parses the rest of the instance setting */
	vrAssign6sensorR2Exform(sensor6_input, strchr(dti->instance, ','));

	/* Set the initial values */
	vrMatrixSetIdentity(&mat_store);
	vrMatrixPostMult(&mat_store, sensor6_input->ipos_xform);
	vrAssign6sensorValue(sensor6_input, &mat_store, -1 /* , vrCurrentWallTime() */);
	vrAssign6sensorActiveValue(sensor6_input, 0);
	vrAssign6sensorErrorValue(sensor6_input, 0);

	vrDbgPrintfN(INPUT_DBGLVL, "_DTrack6sensorInput(): assigned tracker event of value %d (DTrack tracker %d of type %s(%d)) to input pointer = %p)\n",
		tracker_num, tracker_choice, dti->type, tracker_type, aux->sensor6_inputs[tracker_num]);


	return VRINPUT_MATCH_ABLE;	/* input declaration match */
}



	/************************************************************/
	/***************** FreeVR Callback routines *****************/
	/************************************************************/

	/************************************************************/
	/*    Callbacks for controlling the device's features.      */
	/*                                                          */

/*************************************************************************/
/* This collection of callbacks (for controlling the device features) is */
/*   designed to allow the FreeVR system configuration to map physical   */
/*   inputs of this device to control the operation of the VR system     */
/*   as opposed to the norm, where inputs are used to affect the         */
/*   simulated virtual world.                                            */
/*                                                                       */
/* Each command callback takes exactly two arguments:                    */
/*   * devinfo -- this is information about this particular instance of  */
/*                  connecting with an input device or system.  This     */
/*                  provides:                                            */
/*                  1) the ability to alter the settings of the device   */
/*                     itself using the "aux" data field (for example    */
/*                     the SpaceTec devices can do a "reset" with        */
/*                     _SpacetecSensorResetCallback(), or the X11 input  */
/*                     can loop through all the simulated sensors with   */
/*                     _XwindowsSensorNextCallback();                    */
/*               or 2) these callbacks can be used to affect the overall */
/*                     VR system, such as toggling the "paused" flag.    */
/*   * value -- an integer flag that can affect the callback.            */
/*       NOTE: this assumes that the callback is from a button input, or */
/*       from a valuator passing a threshold.  For buttons, the value    */
/*       will be 1 (one) when the button is depressed.                   */
/*************************************************************************/


#if 0
/***************************************************************************/
/* _DTrackSystemPauseToggleCallback(): this callback toggles the "paused"  */
/*   field of the overall FreeVR "context" structure.  When the system is  */
/*   "paused", rendering still functions, but the inputs are frozen.  This */
/*   can be used for debugging, or for still photography of a CAVE.        */
static void _DTrackSystemPauseToggleCallback(vrInputDevice *devinfo, int value)
{
	_DTrackPrivateInfo	*aux = (_DTrackPrivateInfo *)devinfo->aux_data;

	if (value == 0)
		return;

	devinfo->context->paused ^= 1;
	vrDbgPrintfN(SELFCTRL_DBGLVL, "DTrack Control: system_pause = %d.\n",
		devinfo->context->paused);
}

/*****************************************************************************/
/* _DTrackPrintContextStructCallback(): this callback generates output of    */
/*   the overall system "context" data.  It is generally used when debugging */
/*   the library or the configuration.                                       */
static void _DTrackPrintContextStructCallback(vrInputDevice *devinfo, int value)
{
	if (value == 0)
		return;

	vrFprintContext(stdout, devinfo->context, verbose);
}

/************************************************************************/
/* _DTrackPrintConfigStructCallback(): this callback outputs the entire */
/*    VR system configuration to the text console.                      */
static void _DTrackPrintConfigStructCallback(vrInputDevice *devinfo, int value)
{
	if (value == 0)
		return;

	vrFprintConfig(stdout, devinfo->context->config, verbose);
}

/************************************************************************/
/* _DTrackPrintInputStructCallback(): this callback outputs the current */
/*    state of the entire input structure of the running VR system.     */
static void _DTrackPrintInputStructCallback(vrInputDevice *devinfo, int value)
{
	if (value == 0)
		return;

	vrFprintInput(stdout, devinfo->context->input, verbose);
}
#endif

/***************************************************************************/
/* _DTrackPrintPrivateInfoCallback(): this callback outputs the contents   */
/*    of the device specific internal structure to help with configuration */
/*    and system debugging.                                                */
static void _DTrackPrintPrivateInfoCallback(vrInputDevice *devinfo, int value)
{
	_DTrackPrivateInfo	*aux = (_DTrackPrivateInfo *)devinfo->aux_data;

	if (value == 0)
		return;

	_DTrackPrintPrivateInfo(stdout, aux, verbose);
}

/***********************************************************************/
/* _DTrackPrintHelpCallback(): this callback outputs the full system's */
/*     help information string.                                        */
static void _DTrackPrintHelpCallback(vrInputDevice *devinfo, int value)
{
	_DTrackPrivateInfo	*aux = (_DTrackPrivateInfo *)devinfo->aux_data;

	if (value == 0)
		return;

	_DTrackPrintHelp(stdout, aux);
}


	/*************************************************/
	/*   Callbacks for interfacing with the device.  */
	/*                                               */

/**********************************************************************/
/* This collection of callbacks (for interfacing with the device) are */
/*   the standard ways in which the FreeVR system creates an instance */
/*   of access to a particular device.  It consists of the six (6)    */
/*   basic operational functions:                                     */
/*     1) Creating the device instance                                */
/*     2) Open a communications channel with the device               */
/*     3) Close the communications channel with the device            */
/*     4) Reset the device                                            */
/*     5) Poll input data from the device                             */
/* and 6) Print information about the state of the device             */
/**********************************************************************/


/***********************************************************************/
/* _DTrackCreateFunction(): this callback maps the configuration input */
/*    options and control options to the above callback routines that  */
/*    will then execute the code to handle each type of input/control. */
/*                                                                     */
/* There are N primary tasks of this function:                         */
/*     1) Allocate the the memory for the device instance data.        */
/*     2) Parse the options for operating this device instance.        */
/*     3) Allocate the memory for all the inputs and self-controls.    */
/*     4) Actually map the callbacks for all inputs and self-controls. */
/* and 5) Set the flag that the device is configured and ready to go.  */
static void _DTrackCreateFunction(vrInputDevice *devinfo)
{
	/*** List of possible inputs ***/
static	vrInputFunction	_DTrackInputs[] = {
				{ "fs1", VRINPUT_2WAY, _DTrackButtonInput },
				{ "fs2", VRINPUT_2WAY, _DTrackButtonInput },
				{ "fs2", VRINPUT_VALUATOR, _DTrackValuatorInput },	/* NOTE: the older flysticks (fs1) had no valuators */
				{ "6body", VRINPUT_6SENSOR, _DTrack6sensorInput },
				{ "fs2", VRINPUT_6SENSOR, _DTrack6sensorInput },
				{ NULL, VRINPUT_UNKNOWN, NULL } };
	/*** List of control functions ***/
static	vrControlFunc	_DTrackControlList[] = {
				/* overall system controls */
				{ "system_pause_toggle", vrInputSystemPauseToggleCallback },

				/* informational output controls */
				{ "print_context", vrInputPrintContextStructCallback },
				{ "print_config", vrInputPrintConfigStructCallback },
				{ "print_input", vrInputPrintInputStructCallback },
				{ "print_devinfo", _DTrackPrintPrivateInfoCallback },
				{ "print_help", _DTrackPrintHelpCallback },
		/** TODO: other callback control functions go here **/
				/* other controls */

				/* end of the list */
				{ NULL, NULL } };

	_DTrackPrivateInfo	*aux = NULL;

	/************************************************************/
	/* step 1: allocate and initialize auxiliary (private) data */
	devinfo->aux_data = (void *)vrShmemAlloc0(sizeof(*aux));
	aux = (_DTrackPrivateInfo *)devinfo->aux_data;
	_DTrackInitializeStruct(aux, devinfo->type);

	/**************************/
	/* step 2: handle options */
	_DTrackParseArgs(aux, devinfo->args);

	/***********************************************/
	/* step 3: create the inputs and self-controls */
	vrInputCountDataContainers(devinfo);

	/* Allocate internal/private memory for processing specific input types */
	/* NOTE: we add the number of self-controls to each type just to be   */
	/*   sure, because we don't yet know to which inputs the self-controls*/
	/*   will be assigned, so we play it safe.                            */
	aux->button_inputs = (vr2switch **)vrShmemAlloc((devinfo->num_2ways + devinfo->num_scontrols) * sizeof(vr2switch *));
	aux->button_map_bit = (int *)vrShmemAlloc0((devinfo->num_2ways + devinfo->num_scontrols) * sizeof(int));
	aux->button_map_tracker = (int *)vrShmemAlloc0((devinfo->num_2ways + devinfo->num_scontrols) * sizeof(int));
	aux->button_map_type = (int *)vrShmemAlloc0((devinfo->num_2ways + devinfo->num_scontrols) * sizeof(int));

	aux->valuator_inputs = (vrValuator **)vrShmemAlloc0((devinfo->num_valuators + devinfo->num_scontrols) * sizeof(vrValuator *));
	aux->valuator_map_valuator = (int *)vrShmemAlloc0((devinfo->num_valuators + devinfo->num_scontrols) * sizeof(int));
	aux->valuator_map_tracker = (int *)vrShmemAlloc0((devinfo->num_valuators + devinfo->num_scontrols) * sizeof(int));
	aux->valuator_map_type = (int *)vrShmemAlloc0((devinfo->num_valuators + devinfo->num_scontrols) * sizeof(int));
	aux->valuator_map_sign = (float *)vrShmemAlloc0((devinfo->num_valuators + devinfo->num_scontrols) * sizeof(float));

	aux->sensor6_inputs = (vr6sensor **)vrShmemAlloc((devinfo->num_6sensors) * sizeof(vr6sensor *));
	aux->tracker_map_tracker = (int *)vrShmemAlloc0((devinfo->num_6sensors) * sizeof(int));
	aux->tracker_map_type = (int *)vrShmemAlloc0((devinfo->num_6sensors) * sizeof(int));

	/**************************************************************************/
	/* step 4: These two functions do the real work of this function, by      */
	/*   creating the mappings as they are laid out in the configuration file.*/
	vrInputCreateDataContainers(devinfo, _DTrackInputs);
	vrInputCreateSelfControlContainers(devinfo, _DTrackInputs, _DTrackControlList);
	/* TODO: anything to do for NON self-controls?  Implement for Magellan first. */

	/***************************************************/
	/* step 5: set the configured-and-ready-to-go flag */
	vrDbgPrintf("_DTrackCreateFunction(): Done creating DTrack inputs for '%s'\n", devinfo->name);
	devinfo->created = 1;

	return;
}

#endif /* } FREEVR */

/***************************************************************************/
/* _DTrackOpenFunction(): this callback opens the communications channel   */
/*    with the physical device or data streaming utility.  This is         */
/*    generally unique for each type of device or stream.                  */
/*                                                                         */
/* There are no return values, but there are two input-device flags that   */
/*    are set to indicate the state of the device:                         */
/*      * open -- indicates when the communications channel is working     */
/*      * operating -- indicates the device is actually ready to send data */
static void _DTrackOpenFunction(vrInputDevice *devinfo)
{
	vrTrace("_DTrackOpenFunction", devinfo->name);

	_DTrackPrivateInfo	*aux = (_DTrackPrivateInfo *)devinfo->aux_data;

	/*******************/
	/* open the device */
	aux->fd_socket = vrSocketCreateListenUDP(&(aux->data_port), 0);
	if (aux->fd_socket < 0) {
		aux->open = 0;
		devinfo->operating = 0;
		vrErrPrintf("(%s::_DTrackOpenFunction()::%d) error: "
			RED_TEXT "couldn't open UDP socket %d for %s\n" NORM_TEXT,
			__FILE__, __LINE__, aux->data_port, devinfo->name);
		snprintf(aux->version, sizeof(aux->version), "- unconnected DTrack -");
	} else {
		aux->open = 1;
		if (_DTrackInitializeDevice(aux) < 0) {
			devinfo->operating = 0;
			vrErrPrintf("_DTrackOpenFunction(): "
				RED_TEXT "Warning, unable to initialize '%s' DTrack.\n" NORM_TEXT,
				devinfo->name);
		} else {
			devinfo->operating = 1;
			vrDbgPrintf("_DTrackOpenFunction(): Done opening DTrack input device '%s' (operating = %d).\n", devinfo->name, devinfo->operating);
		}
	}

	return;
}


/***************************************************************************/
/* _DTrackCloseFunction(): this callback closes the communications channel */
/*   with the ART DTrack device.                                           */
static void _DTrackCloseFunction(vrInputDevice *devinfo)
{
	_DTrackPrivateInfo	*aux = (_DTrackPrivateInfo *)devinfo->aux_data;

	vrDbgPrintf("_DTrackCloseFunction(): About to close DTrack input device '%s'.\n", devinfo->name);

	_DTrackCloseDevice(aux);

	if (aux != NULL) {
		if (aux->open)
			vrSocketClose(aux->fd_socket);
#ifdef FREEVR
		/* free the FreeVR specific fields of "aux" */
		/* TODO: ... */
#endif

		vrShmemFree(aux);			/* aka devinfo->aux_data */
	}

	return;
}


#if defined(FREEVR) /* { */
/*************************************************************************/
/* _DTrackResetFunction(): this callback provides the means by which the */
/*    device or input stream can be reset to a good state.  For the ART  */
/*    DTrack2 system, this is currently a no-op.                         */
static void _DTrackResetFunction(vrInputDevice *devinfo)
{
	_DTrackPrivateInfo	*aux = (_DTrackPrivateInfo *)devinfo->aux_data;

	/* TODO: is there a method for resetting the ART DTrack? */

	return;
}


/****************************************************************************/
/* _DTrackPollFunction(): this callback actually reads the data from the    */
/*    input device or stream using a polling (request) method of operation. */
/*    A check is done to confirm that the device is in "operating" state    */
/*    before attempting to read the data.                                   */
static void _DTrackPollFunction(vrInputDevice *devinfo)
{
	_DTrackPrivateInfo	*aux = (_DTrackPrivateInfo *)devinfo->aux_data;

	if (devinfo->operating) {
		_DTrackGetData(devinfo);
	} else {
		/* TODO: try to open the device again */
	}

	return;
}



	/******************************/
	/*** FreeVR public routines ***/
	/******************************/

/* NOTE: actually, there is only one public routine */


/****************************************************************************/
/* vrDTrackInitInfo(): this function serves as the "public" means by which  */
/*   a device (or input stream) is interfaced with from the FreeVR library. */
/*   This function creates the shared memory for the device version string, */
/*   and then instantiates the new instance with all the callbacks needed   */
/*   for FreeVR to operate the input.                                       */
void vrDTrackInitInfo(vrInputDevice *devinfo)
{
	devinfo->version = (char *)vrShmemStrDup("-get from DTrack device-");
	devinfo->Create = vrCallbackCreateNamed("DTrack:Create-Def", _DTrackCreateFunction, 1, devinfo);
	devinfo->Open = vrCallbackCreateNamed("DTrack:Open-Def", _DTrackOpenFunction, 1, devinfo);
	devinfo->Close = vrCallbackCreateNamed("DTrack:Close-Def", _DTrackCloseFunction, 1, devinfo);
	devinfo->Reset = vrCallbackCreateNamed("DTrack:Reset-Def", _DTrackResetFunction, 1, devinfo);
	devinfo->PollData = vrCallbackCreateNamed("DTrack:PollData-Def", _DTrackPollFunction, 1, devinfo);
	devinfo->PrintAux = vrCallbackCreateNamed("DTrack:PrintAux-Def", _DTrackPrintPrivateInfo, 0);

	vrDbgPrintfN(INPUT_DBGLVL, "vrDTrackInitInfo(): callbacks created.\n");
}


#endif /* } FREEVR */



	/*===========================================================*/
	/*===========================================================*/
	/*===========================================================*/
	/*===========================================================*/


/****************************************************************************/
/****************************************************************************/
/* The following section is the independent "test" utility that can be used */
/*   to test communications and output from the input device (or stream).   */
#if defined(TEST_APP) /* { */

/*********************************************************************/
/* Ugh, I hate globals, but I don't know a better way to get the     */
/*   "done" value into the interrupt signal function exit_testapp(). */
static	int	done = 0;

#define	RENDERLINES	0
#define	RENDERCSV	1	/* NOTE: not yet implemented here -- see vr_input.fastrak.c for example */
#define	RENDERSCREEN	2


/**************************************************************************/
/* exit_testapp(): this function is used as a signal callback, such that  */
/*    when called, it simply sets the "done" flag, so the main while loop */
/*    in main() while know to terminate at the end of the current pass.   */
void exit_testapp()
{
	done = 1;
}

/*******************************************************************/
/* _RenderValuesLine(): Render the Tracking values in a streaming  */
/*   line style format -- with a carriage return, so will overwrite*/
/*   previous lines (unless the screen is insufficiently wide.     */
void _RenderValuesLine(_DTrackPrivateInfo *aux)
{
	int		count;				/* for looping over device inputs */
	_DTrackUnit	unit_6body;			/* data for a 6body unit */
	_DTrackUnit	unit_fs2;			/* data for a flystick-2 unit */

	/* loop over the standard 6-dof bodies */
	for (count = 0; count < MAX_UNITS_PT; count++) {
		if (aux->units_6body[count].frame > 0) {
			unit_6body= aux->units_6body[count];
			printf("6d (%d): frame = %d, qual = %.2f -- %f %f %f\n",
				count,
				unit_6body.frame,
				unit_6body.quality,
				unit_6body.location[VR_X],
				unit_6body.location[VR_Y],
				unit_6body.location[VR_Z]);
		}
//else { printf("frame value is %d\n", unit_6body.frame); }
	}

	/* loop over the flystick-2 bodies */
	for (count = 0; count < MAX_UNITS_PT; count++) {
		if (aux->units_fs2[count].frame > 0) {
			unit_fs2 = aux->units_fs2[count];
			printf("6df2 (%d): frame = %d, qual = %.2f -- %f %f %f  but: %2d  val-1 %4f  val-2 %4f\n",
				count,
				unit_fs2.frame,
				unit_fs2.quality,
				unit_fs2.location[VR_X],
				unit_fs2.location[VR_Y],
				unit_fs2.location[VR_Z],
				unit_fs2.buttons,
				unit_fs2.valuators[0],
				unit_fs2.valuators[1]);
		}
	}

	/* TODO: loop over the other types of position inputs */

	printf("\n");

}

/********************************************************************/
/* _RenderValuesCSV(): Render the Tracking values in a streaming    */
/*   format with terse comma-separated-value information.  All new  */
/*   data reports are placed on their own line (record), though the */
/*   columns will vary depending on the type of data.               */
/* NOTE: unlike the line-rendering style, each line will have one   */
/*   input type, so there is no need to keep track of whether a     */
/*   particular device is already being reported.                   */
void _RenderValuesCSV(_DTrackPrivateInfo *aux)
{
	int		count;				/* for looping over device inputs */
	_DTrackUnit	unit_6body;			/* data for a 6body unit */
	_DTrackUnit	unit_fs2;			/* data for a flystick-2 unit */
	vrMatrix	rot_matrix;			/* A 4x4 version of the 3x3 matrix returned by DTrack */
	vrMatrixComposition	components;		/* Calculated components of a matrix */
	int		singularity;			/* Whether or not matrix was singular */

	/* loop over the standard 6-dof bodies */
	for (count = 0; count < MAX_UNITS_PT; count++) {
		if (aux->units_6body[count].frame > 0) {
			unit_6body= aux->units_6body[count];

			/* calculate the quaternion (and other components) */
			vrMatrixSetIdentity(&rot_matrix);
			VRMAT_COLROW(&rot_matrix, 0, 0) = unit_6body.rotation[0];
			VRMAT_COLROW(&rot_matrix, 0, 1) = unit_6body.rotation[1];
			VRMAT_COLROW(&rot_matrix, 0, 2) = unit_6body.rotation[2];
			VRMAT_COLROW(&rot_matrix, 1, 0) = unit_6body.rotation[3];
			VRMAT_COLROW(&rot_matrix, 1, 1) = unit_6body.rotation[4];
			VRMAT_COLROW(&rot_matrix, 1, 2) = unit_6body.rotation[5];
			VRMAT_COLROW(&rot_matrix, 2, 0) = unit_6body.rotation[6];
			VRMAT_COLROW(&rot_matrix, 2, 1) = unit_6body.rotation[7];
			VRMAT_COLROW(&rot_matrix, 2, 2) = unit_6body.rotation[8];
			singularity = vrMatrixCompositionFromMatrix(&components, &rot_matrix);

			printf("%d,6dof,%d,%.6f,%.2f,%.6f,%.6f,%.6f,%d,%.6f,%.6f,%.6f,%.6f\n",
				unit_6body.frame,
				count,
				unit_6body.quality,
				unit_6body.time_stamp,
				unit_6body.location[VR_X],
				unit_6body.location[VR_Y],
				unit_6body.location[VR_Z],
				singularity,
				components.quaternion[VR_X],
				components.quaternion[VR_Y],
				components.quaternion[VR_Z],
				components.quaternion[VR_W]);
		}
	}

	/* loop over the flystick-2 bodies */
	for (count = 0; count < MAX_UNITS_PT; count++) {
		if (aux->units_fs2[count].frame > 0) {
			unit_fs2 = aux->units_fs2[count];

			/* calculate the quaternion (and other components) */
			vrMatrixSetIdentity(&rot_matrix);
			VRMAT_COLROW(&rot_matrix, 0, 0) = unit_fs2.rotation[0];
			VRMAT_COLROW(&rot_matrix, 0, 1) = unit_fs2.rotation[1];
			VRMAT_COLROW(&rot_matrix, 0, 2) = unit_fs2.rotation[2];
			VRMAT_COLROW(&rot_matrix, 1, 0) = unit_fs2.rotation[3];
			VRMAT_COLROW(&rot_matrix, 1, 1) = unit_fs2.rotation[4];
			VRMAT_COLROW(&rot_matrix, 1, 2) = unit_fs2.rotation[5];
			VRMAT_COLROW(&rot_matrix, 2, 0) = unit_fs2.rotation[6];
			VRMAT_COLROW(&rot_matrix, 2, 1) = unit_fs2.rotation[7];
			VRMAT_COLROW(&rot_matrix, 2, 2) = unit_fs2.rotation[8];
			singularity = vrMatrixCompositionFromMatrix(&components, &rot_matrix);

			printf("%d,6df2,%d,%.6f,%.2f,%.6f,%.6f,%.6f,%d,%.6f,%.6f,%.6f,%.6f,%d,%d,%d,%.6f,%.6f\n",
				unit_fs2.frame,
				count,
				unit_fs2.quality,
				unit_fs2.time_stamp,
				unit_fs2.location[VR_X],
				unit_fs2.location[VR_Y],
				unit_fs2.location[VR_Z],
				singularity,
				components.quaternion[VR_X],
				components.quaternion[VR_Y],
				components.quaternion[VR_Z],
				components.quaternion[VR_W],
				unit_fs2.num_buttons,
				unit_fs2.num_valuators,
				unit_fs2.buttons,		/* NOTE: WARNING: if num_buttons>32, then more than one uint8 of buttons */
				unit_fs2.valuators[0],
				unit_fs2.valuators[1]);
		}
	}
}


/* _RenderValuesScreen6dofColumn(): Render a single column of 6-DOF */
/*   tracking data, including center and quartile markers.          */
/*                                                                  */
/* NOTE: this will serve as a template that I plan to move back to  */
/*   other input devices with screen renders (vrpn, vruidd,         */
/*   fastrak, fob, maybe more).                                     */
void _RenderValuesScreen6dofColumn(int column, int total_lines, int center_line, float max_trans, float *location, float *angles, vrQuat *quat)
{
	int	vertical[12];			/* the vertical line to render on -- used to prevent underflow/overflow */
const	char	*codes = "XYZAERijkW";
//const	char	*codes = "XYZAER";
	int	col;				/* loop counter for columns */
	int	total_columns = 0;		/* The number of columns assigned for display */

	float	trans_scale = total_lines * -0.5 / max_trans;
	float	rot_scale = total_lines * -0.5 / 180.0;	/* (180.0 = max_rot) */
	float	quat_scale = total_lines * -0.5 / 1.0;	/* (1.0 = max_quat) */
	int	plus_minus = max_trans*trans_scale;

	/* print a *zero* line */
	vrPrintf(ANSI_MOVETOROWCOL RED_TEXT ">----------<" NORM_TEXT, center_line, column);
	vrPrintf(ANSI_MOVETOROWCOL RED_TEXT "|==========|" NORM_TEXT, center_line + plus_minus + 1, column); /* top line */
	vrPrintf(ANSI_MOVETOROWCOL RED_TEXT "|==========|" NORM_TEXT, center_line - plus_minus, column); /* bottom line */
#if 0 /* 0 = less obscuration */
	vrPrintf(ANSI_MOVETOROWCOL RED_TEXT "|..........|" NORM_TEXT, center_line + (int)(plus_minus * 0.5), column); /* 1/4 down line */
#else
	vrPrintf(ANSI_MOVETOROWCOL RED_TEXT "|" NORM_TEXT, center_line + (int)(plus_minus * 0.5), column-1); /* 1/4 down line */
	vrPrintf(ANSI_MOVETOROWCOL RED_TEXT "|" NORM_TEXT, center_line + (int)(plus_minus * 0.5), column+12); /* 1/4 down line */
#endif
	vrPrintf(ANSI_MOVETOROWCOL RED_TEXT "|..........|" NORM_TEXT, center_line - (int)(plus_minus * 0.5), column); /* 3/4 down line */

#if 1 /* [optional] switch to reverse mode for moving values */
	vrPrintf(REVERSE_TEXT); /* switch to reverse */
#endif
	/* display the location values */
	vertical[0] = (location[VR_X]*trans_scale)+center_line;
	vertical[1] = (location[VR_Y]*trans_scale)+center_line;
	vertical[2] = (location[VR_Z]*trans_scale)+center_line;
//vrPrintf("\fvert[0,1,2] = %d,%d,%d, loc[X] = %f, scale = %f, cl = %d\n", vertical[0], vertical[1], vertical[2], location[VR_X], trans_scale,center_line); /* debugging code, eventually delete me */
	total_columns += 3;

	/* display the Euler Angle values */
	vertical[3] = (angles[VR_X]*rot_scale)+center_line;
	vertical[4] = (angles[VR_Y]*rot_scale)+center_line;
	vertical[5] = (angles[VR_Z]*rot_scale)+center_line;
	total_columns += 3;

	/* [optional] display the quaternion values */
	vertical[6] = (quat->v[VR_X]*quat_scale)+center_line;
	vertical[7] = (quat->v[VR_Y]*quat_scale)+center_line;
	vertical[8] = (quat->v[VR_Z]*quat_scale)+center_line;
	vertical[9] = (quat->v[VR_W]*quat_scale)+center_line;
	total_columns += 4;

	/* Loop over the columes using data provided above */
	for (col = 0; col < total_columns; col++) {
		if (vertical[col] < 0) {
			/* if we go off the top, stop and make it red */
			vertical[col] = 0;
			vrPrintf(RED_TEXT);
		}
		if (vertical[col] > center_line - plus_minus) {
			/* if we go off the bottom, stop and make it red */
			vertical[col] = center_line - plus_minus;
			vrPrintf(RED_TEXT);
		}
		vrPrintf(ANSI_MOVETOROWCOL "%c" ANSI_TEXTFORENORM, vertical[col], column + 1 + col, codes[col]);
	}

	vrPrintf(NORM_TEXT);    /* switch to normal mode (non-reverse) */
}

/*******************************************************************/
/* _RenderValuesScreen(): Render the Tracking values in a text-GUI */
/*   style format.                                                 */
/* NOTE: this routine relies on using the ANSI terminal character  */
/*   codes.  Since this is just a test application, I decided to   */
/*   forgo the requirement of the curses library, and instead      */
/*   limit the terminal on which this version would work.          */
/*   On Linux do "man console_codes" to get the sequence codes     */
/*   for other operations.                                         */
/*                                                                 */
/* The MAX_UNITS_PT value is the most of any given "part-type" that*/
/*   will be reported.  (I think) we aren't told the number of     */
/*   possible devices, we just get what we get.                    */
void _RenderValuesScreen(_DTrackPrivateInfo *aux)
{
	int		count;				/* for looping over device inputs */
	int		devnum;			/* for looping through devices */
	int		start_line = 2;		/* first line on the screen to place data */	/* TODO: for some reason things scroll up two lines -- need to investigate */
	int		line;			/* the line to print on */
	float		percent;		/* used for simple valuator visualization */

	int		num_lines = 40;		/* number of lines for showing chart */
	int		header_lines = 4;	/* used to shift the center and the min/max scaling */
	int		center = (num_lines-header_lines) / 2.0 + header_lines + 1;
	int		col_width = 22;		/* the number of characters per column */
	int		col1_offset = 5;	/* the offset before the first column */
	int		column_start;		/* starting column number for a particular unit */
	int		butval_offset;		/* the column to output the button & valuators */
#if 0 /* delete me */
	int		num_columns = 100;
#endif

	float		max_trans = 2000.0;	/* 2 meters (in mm) in one direction */
#if 0 /* These are now all handled in _RenderValuesScreen6dofColumn() -- delete me */
	float		max_rot = 180.0;	/* 180 degrees max in each direction */
	float		max_qual = 255.0;	/* 0-255 is the tracking quality range */	/* TODO: check the range on the quality values */

	float		trans_scale = (num_lines-header_lines) * -0.5 / max_trans;
	float		rot_scale = (num_lines-header_lines) * -0.5 / max_rot;
	float		quat_scale = (num_lines-header_lines) * -0.5 / 1.0; /* (1.0 = max_quat) */
#endif

	//_DTrackUnit	units_6body[MAX_UNITS_PT];	/* an array of all the 6body units */
	//_DTrackUnit	units_fs2[MAX_UNITS_PT];	/* an array of all the flystick-2 units */
	_DTrackUnit	unit_6body;			/* data for a 6body unit */
	_DTrackUnit	unit_fs2;			/* data for a flystick-2 unit */
	int		num_units = 0;			/* the number of position tracked units (to know what column to place data into) */

	/* for calculating the Euler angles of a position tracker input */
#if 0 /* delete me */
	vrPoint		location_input;		/* The location/translation portion */
	vrQuat		quat_input;		/* The quaternion portion */
#endif
	vrEuler		euler_angles;		/* The Euler+translation representation */
	vrMatrix	rot_matrix;		/* A 4x4 version of the 3x3 matrix returned by DTrack */
	vrMatrixComposition	components;	/* Calculated components of a matrix */
	int		singularity;


	/* move to upper left corner of screen */
	vrPrintf(ANSI_MOVETOUPPERLEFT);

	/* clear to end of screen */
	vrPrintf(ANSI_CLEARTOEOS);

	/* draw a data vs. debug separating line */
	vrPrintf(ANSI_MOVETOROWCOL, num_lines+1, 0);
	vrPrintf("=" TTY_REPEATCHAR, 80);
	vrPrintf(ANSI_MOVETOROWCOL, 1000, 0);

	/* save current cursor position */
	vrPrintf(ANSI_CURSORLOC_SAVE);

	/******************************************/
	/* Display header information */
	vrPrintf(ANSI_MOVETOROWCOL "DTrack Server reported devices:", start_line + 0, 2);
	vrPrintf(ANSI_MOVE "Frame : %d, Calibrated bodies: %d", start_line + 1, 10, aux->frame, aux->num_6dcal_bodies);
#if 0 /* Vrui version {{{ */ /* delete me */
	vrPrintf(ANSI_MOVE "Trackers : %d", start_line + 1, 10, aux->num_dd_trackers);
	vrPrintf(ANSI_MOVE "Buttons  : %d", start_line + 2, 10, aux->num_buttons);
	vrPrintf(ANSI_MOVE "Valuators: %d", start_line + 3, 10, aux->num_valuators);
#endif /* }}} Vrui version */

	/* Set normal text output to be within the given margin      */
	/*   (This pushes all debug output into the "margin" region, */
	/*   and thus doesn't obscure the screen rendering.)         */
	/*   NOTE: we exaggerate the bottom line to allow the screen */
	/*   to be expanded and still work properly.                 */
	vrPrintf(ANSI_SETMARGIN_TOPBOT, num_lines+2, num_lines+200);

	/******************************************/
	/* Display tracker (6-sensor) information */
	/* start counting at '1' to skip the simulator placeholder */
#if 0 /* Vrui version -- display tracker information {{{ */ /* delete me */
	for (devnum = 0; devnum < aux->num_dd_trackers; devnum++) {
		int	ptv_count;		/* for looping over the position-tracker values */

		/* NOTE: Ideally we'd have a "vruidd_dev" pointer, but for  */
		/*   some reason, the [devnum] index is on the outside, so  */
		/*   we can't really have a pointer to a single entity from */
		/*   which we can just get the X,Y,Z,etc. values.           */

		/* Store the given <input> in FreeVR types (inc. Eulers) */
		vrPrintf(ANSI_CURSORLOC_RETURN);        /* jump to output margin for possible debugging output */
		vrPointSet3d(&location_input, aux->incoming_trackers[VR_X][devnum], aux->incoming_trackers[VR_Y][devnum], aux->incoming_trackers[VR_Z][devnum]);
		vrQuatSet4d(&quat_input, aux->incoming_trackers[VR_QX][devnum], aux->incoming_trackers[VR_QY][devnum], aux->incoming_trackers[VR_QZ][devnum], aux->incoming_trackers[VR_QW][devnum]);
		vrEulerSetFromQuatPoint(&euler_input, &quat_input, &location_input);
		vrPrintf(ANSI_CURSORLOC_SAVE);          /* save cursor for next time */

		/* now print info headers */
		vrPrintf(ANSI_MOVE "Tracker %d", start_line + 4, devnum*col_width + col1_offset, devnum+1);
		vrPrintf(ANSI_MOVE "XYZAERijkW", start_line + 5, devnum*col_width + col1_offset + 1);
#  if 1 /* [optional] show the XYZ values under the header */
		vrPrintf(ANSI_MOVE "X=% 9.5f", start_line + 7, devnum*col_width + col1_offset, aux->incoming_trackers[VR_X][devnum]);
		vrPrintf(ANSI_MOVE "Y=% 9.5f", start_line + 8, devnum*col_width + col1_offset, aux->incoming_trackers[VR_Y][devnum]);
		vrPrintf(ANSI_MOVE "Z=% 9.5f", start_line + 9, devnum*col_width + col1_offset, aux->incoming_trackers[VR_Z][devnum]);
#  endif


		/* print a *zero* line */
		vrPrintf(ANSI_MOVE RED_TEXT ">----------<" NORM_TEXT, center, devnum*col_width + col1_offset);

#  if 1
		vrPrintf(ANSI_MOVE RED_TEXT "|==========|" NORM_TEXT, center + (int)(max_trans*trans_scale)+1, devnum*col_width + col1_offset);	/* top line -- presently overwrites the "XYZAERijkW" header */
#  endif
		vrPrintf(ANSI_MOVE RED_TEXT "|==========|" NORM_TEXT, center - (int)(max_trans*trans_scale), devnum*col_width + col1_offset);	/* bottom line */
		vrPrintf(ANSI_MOVE RED_TEXT "|..........|" NORM_TEXT, center + (int)(max_trans*trans_scale*0.5), devnum*col_width + col1_offset);	/* 1/4 down line */
		vrPrintf(ANSI_MOVE RED_TEXT "|..........|" NORM_TEXT, center - (int)(max_trans*trans_scale*0.5), devnum*col_width + col1_offset);	/* 3/4 down line */


#  if 1 /* [optional] switch to reverse mode for moving values */
		vrPrintf(REVERSE_TEXT);	/* switch to reverse */
#  endif

		/* print the X value */
		vertical = (int)(aux->incoming_trackers[VR_X][devnum]*trans_scale)+center;
		if (vertical < 0) {
			/* if we go off the top, stop and make it red */
			vertical = 0;
			vrPrintf(RED_TEXT);
		}
		vrPrintf(ANSI_MOVETOROWCOL "X" ANSI_TEXTFORENORM, vertical, devnum*col_width + col1_offset+1);

		/* print the Y value */
		vertical = (int)(aux->incoming_trackers[VR_Y][devnum]*trans_scale)+center;
		if (vertical < 0) {
			/* if we go off the top, stop and make it red */
			vertical = 0;
			vrPrintf(RED_TEXT);
		}
		vrPrintf(ANSI_MOVETOROWCOL "Y" ANSI_TEXTFORENORM, vertical, devnum*col_width + col1_offset+1);

		/* print the Z value */
		vertical = (int)(aux->incoming_trackers[VR_X][devnum]*trans_scale)+center;
		if (vertical < 0) {
			/* if we go off the top, stop and make it red */
			vertical = 0;
			vrPrintf(RED_TEXT);
		}
		vrPrintf(ANSI_MOVETOROWCOL "X" ANSI_TEXTFORENORM, vertical, devnum*col_width + col1_offset+1);

		/* print the A value (Azimuth) */
		vrPrintf(ANSI_MOVETOROWCOL "A", (int)(euler_input.r[VR_AZIM]*rot_scale)+center, devnum*col_width + col1_offset + 4);

		/* print the E value (Elevation) */
		vrPrintf(ANSI_MOVETOROWCOL "E", (int)(euler_input.r[VR_ELEV]*rot_scale)+center, devnum*col_width + col1_offset + 5);

		/* print the R value (Roll) */
		vrPrintf(ANSI_MOVETOROWCOL "R", (int)(euler_input.r[VR_ROLL]*rot_scale)+center, devnum*col_width + col1_offset + 6);

		/* Output the Quaternion values */
		vrPrintf(ANSI_MOVETOROWCOL "i", (int)(aux->incoming_trackers[VR_QX][devnum]*quat_scale)+center, devnum*col_width + col1_offset + 7);
		vrPrintf(ANSI_MOVETOROWCOL "j", (int)(aux->incoming_trackers[VR_QY][devnum]*quat_scale)+center, devnum*col_width + col1_offset + 8);
		vrPrintf(ANSI_MOVETOROWCOL "k", (int)(aux->incoming_trackers[VR_QZ][devnum]*quat_scale)+center, devnum*col_width + col1_offset + 9);
		vrPrintf(ANSI_MOVETOROWCOL "W", (int)(aux->incoming_trackers[VR_QW][devnum]*quat_scale)+center, devnum*col_width + col1_offset + 10);
		//vrQuatSet4d(&quat_input, aux->incoming_trackers[VR_QX][devnum], aux->incoming_trackers[VR_QY][devnum], aux->incoming_trackers[VR_QZ][devnum], aux->incoming_trackers[VR_QW][devnum]);

		vrPrintf(NORM_TEXT);	/* switch to normal mode (non-reverse) */

#  if 0
		/* print the joystick values */
		if (unitdata->has_button) {
			vrPrintf(ANSI_MOVE "x", (int)((unitdata->joystick[VR_X]+1)*qual_scale)+num_lines-1, unit*col_width + col1_offset+7);
			vrPrintf(ANSI_MOVE "y", (int)((unitdata->joystick[VR_Y]+1)*qual_scale)+num_lines-1, unit*col_width + col1_offset+8);
		}
#  endif
	}

	butval_offset = aux->num_dd_trackers * col_width + col1_offset;
#endif /* }}} Vrui version */

	/* loop over the standard 6-dof bodies */
	for (count = 0; count < MAX_UNITS_PT; count++) {
		column_start = num_units*col_width + col1_offset;

		/* Regular 6dof body */
		if (aux->units_6body[count].frame > 0) {
			num_units++;
			unit_6body = aux->units_6body[count];

#if 1
			/* calculate the quaternion (and other components) */
			vrMatrixSetIdentity(&rot_matrix);
			VRMAT_COLROW(&rot_matrix, 0, 0) = unit_6body.rotation[0];
			VRMAT_COLROW(&rot_matrix, 0, 1) = unit_6body.rotation[1];
			VRMAT_COLROW(&rot_matrix, 0, 2) = unit_6body.rotation[2];
			VRMAT_COLROW(&rot_matrix, 1, 0) = unit_6body.rotation[3];
			VRMAT_COLROW(&rot_matrix, 1, 1) = unit_6body.rotation[4];
			VRMAT_COLROW(&rot_matrix, 1, 2) = unit_6body.rotation[5];
			VRMAT_COLROW(&rot_matrix, 2, 0) = unit_6body.rotation[6];
			VRMAT_COLROW(&rot_matrix, 2, 1) = unit_6body.rotation[7];
			VRMAT_COLROW(&rot_matrix, 2, 2) = unit_6body.rotation[8];
			singularity = vrMatrixCompositionFromMatrix(&components, &rot_matrix);
#endif

			/* now print info headers */
			line = start_line + 2;
			vrPrintf(ANSI_MOVE "6D Tracker %d", line++, column_start, count+1);
			vrPrintf(ANSI_MOVE "XYZAERijkW", line++, column_start + 1);
#  if 1 /* [optional] show the XYZ values (and more) under the header */
			/* display many (all?) of the values in a column */
			vrPrintf(ANSI_MOVE "            id=%d",    line++, column_start, unit_6body.id);
			vrPrintf(ANSI_MOVE "active=%d",    line++, column_start, unit_6body.active);
			vrPrintf(ANSI_MOVE "frame=%9d",    line++, column_start, unit_6body.frame);
			vrPrintf(ANSI_MOVE "frames late=%5d",    line++, column_start, aux->frame-unit_6body.frame);
			vrPrintf(ANSI_MOVE "quality=% 3.1f",    line++, column_start, unit_6body.quality);
			vrPrintf(ANSI_MOVE "X=% 10.4f",    line++, column_start, unit_6body.location[VR_X]);
			vrPrintf(ANSI_MOVE "Y=% 10.4f",    line++, column_start, unit_6body.location[VR_Y]);
			vrPrintf(ANSI_MOVE "Z=% 10.4f",    line++, column_start, unit_6body.location[VR_Z]);
			vrPrintf(ANSI_MOVE "|% 5.2f % 5.2f % 5.2f |", line++, column_start, unit_6body.rotation[0], unit_6body.rotation[1], unit_6body.rotation[2]);
			vrPrintf(ANSI_MOVE "|% 5.2f % 5.2f % 5.2f |", line++, column_start, unit_6body.rotation[3], unit_6body.rotation[4], unit_6body.rotation[5]);
			vrPrintf(ANSI_MOVE "|% 5.2f % 5.2f % 5.2f |", line++, column_start, unit_6body.rotation[6], unit_6body.rotation[7], unit_6body.rotation[8]);
			vrPrintf(ANSI_MOVE "AZIM=% 9.4f", line++, column_start, unit_6body.angles[VR_AZIM]);
			vrPrintf(ANSI_MOVE "ELEV=% 9.4f", line++, column_start, unit_6body.angles[VR_ELEV]);
			vrPrintf(ANSI_MOVE "ROLL=% 9.4f", line++, column_start, unit_6body.angles[VR_ROLL]);
			vrPrintf(ANSI_MOVE "AZIM2=% 9.4f", line++, column_start, components.rotate[VR_AZIM]);
			vrPrintf(ANSI_MOVE "ELEV2=% 9.4f", line++, column_start, components.rotate[VR_ELEV]);
			vrPrintf(ANSI_MOVE "ROLL2=% 9.4f", line++, column_start, components.rotate[VR_ROLL]);
			vrPrintf(ANSI_MOVE "singularity=%d", line++, column_start, !singularity);
			//vrPrintf(ANSI_MOVE "Q=%2.2fi,%2.2fj,%2.2fk,%2.2f", line++, column_start, components.quaternion[0], components.quaternion[1], components.quaternion[2], components.quaternion[3]); /* debugging code, eventually delete me */
#  endif

			/* render the moving characters of fhe 6-dof position values */
			_RenderValuesScreen6dofColumn(column_start, num_lines-header_lines, center,
				max_trans, unit_6body.location,
				unit_6body.angles, (vrQuat*)components.quaternion);
		}
	}

	for (count = 0; count < MAX_UNITS_PT; count++) {
		/* Flystick 6dof body */
		if (aux->units_fs2[count].frame > 0) {
			num_units++;
			unit_fs2 = aux->units_fs2[count];

			/* calculate the quaternion (and other components) */
			vrMatrixSetIdentity(&rot_matrix);
			VRMAT_COLROW(&rot_matrix, 0, 0) = unit_fs2.rotation[0];
			VRMAT_COLROW(&rot_matrix, 0, 1) = unit_fs2.rotation[1];
			VRMAT_COLROW(&rot_matrix, 0, 2) = unit_fs2.rotation[2];
			VRMAT_COLROW(&rot_matrix, 1, 0) = unit_fs2.rotation[3];
			VRMAT_COLROW(&rot_matrix, 1, 1) = unit_fs2.rotation[4];
			VRMAT_COLROW(&rot_matrix, 1, 2) = unit_fs2.rotation[5];
			VRMAT_COLROW(&rot_matrix, 2, 0) = unit_fs2.rotation[6];
			VRMAT_COLROW(&rot_matrix, 2, 1) = unit_fs2.rotation[7];
			VRMAT_COLROW(&rot_matrix, 2, 2) = unit_fs2.rotation[8];
			singularity = vrMatrixCompositionFromMatrix(&components, &rot_matrix);
			euler_angles.r[VR_AZIM] = 1.0;//components.rotate[VR_AZIM];
			euler_angles.r[VR_ELEV] = 2.0;//components.rotate[VR_ELEV];
			euler_angles.r[VR_ROLL] = 3.0;//components.rotate[VR_ROLL];

			/* now print info headers */
			line = start_line + 2;
			vrPrintf(ANSI_MOVE "6D FS Tracker %d", line++, column_start, count+1);
			vrPrintf(ANSI_MOVE "XYZAERijkW", line++, column_start + 1);
#  if 1 /* [optional] show the XYZ values under the header */
			vrPrintf(ANSI_MOVE "            id=%d",    line++, column_start, unit_fs2.id);
			vrPrintf(ANSI_MOVE "active=%d",    line++, column_start, unit_fs2.active);
			vrPrintf(ANSI_MOVE "frame=%9d",    line++, column_start, unit_fs2.frame);
			vrPrintf(ANSI_MOVE "frames late=%5d",    line++, column_start, aux->frame-unit_fs2.frame);
			vrPrintf(ANSI_MOVE "quality=% 3.1f",    line++, column_start, unit_fs2.quality);
			vrPrintf(ANSI_MOVE "X=% 10.4f",    line++, column_start, unit_fs2.location[VR_X]);
			vrPrintf(ANSI_MOVE "Y=% 10.4f",    line++, column_start, unit_fs2.location[VR_Y]);
			vrPrintf(ANSI_MOVE "Z=% 10.4f",    line++, column_start, unit_fs2.location[VR_Z]);
			vrPrintf(ANSI_MOVE "|% 5.2f % 5.2f % 5.2f |", line++, column_start, unit_fs2.rotation[0], unit_fs2.rotation[1], unit_fs2.rotation[2]);
			vrPrintf(ANSI_MOVE "|% 5.2f % 5.2f % 5.2f |", line++, column_start, unit_fs2.rotation[3], unit_fs2.rotation[4], unit_fs2.rotation[5]);
			vrPrintf(ANSI_MOVE "|% 5.2f % 5.2f % 5.2f |", line++, column_start, unit_fs2.rotation[6], unit_fs2.rotation[7], unit_fs2.rotation[8]);
			vrPrintf(ANSI_MOVE "buttons=%d",    line++, column_start, unit_fs2.buttons); /* TODO: I could also add a button pressed visualization after the integer number */ /* NOTE: If num_buttons > 32, there will be additional uint8 button values */
			vrPrintf(ANSI_MOVE "valuators=% 4.2f,% 4.2f",    line++, column_start, unit_fs2.valuators[0], unit_fs2.valuators[1]);
			line++;		/* skip one line to keep the 3x3 matrices aligned / TODO: this line could be used for a pair of valuator visualization bars */
#    if 0 /* "6df2 devices don't report Euler angles */
			vrPrintf(ANSI_MOVE "AZIM=% 9.4f", line++, column_start, unit_fs2.angles[VR_AZIM]);
			vrPrintf(ANSI_MOVE "ELEV=% 9.4f", line++, column_start, unit_fs2.angles[VR_ELEV]);
			vrPrintf(ANSI_MOVE "ROLL=% 9.4f", line++, column_start, unit_fs2.angles[VR_ROLL]);
#    endif
			vrPrintf(ANSI_MOVE "AZIM2=% 9.4f", line++, column_start, components.rotate[VR_AZIM]);
			vrPrintf(ANSI_MOVE "ELEV2=% 9.4f", line++, column_start, components.rotate[VR_ELEV]);
			vrPrintf(ANSI_MOVE "ROLL2=% 9.4f", line++, column_start, components.rotate[VR_ROLL]);
			vrPrintf(ANSI_MOVE "singularity=%d", line++, column_start, !singularity);
			//vrPrintf(ANSI_MOVE "Q=%2.2fi,%2.2fj,%2.2fk,%2.2f", line++, column_start, components.quaternion[0], components.quaternion[1], components.quaternion[2], components.quaternion[3]); /* debugging code, eventually delete me */
#  endif

			_RenderValuesScreen6dofColumn(column_start, num_lines-header_lines, center,
				max_trans, unit_fs2.location,
				unit_fs2.angles, (vrQuat*)components.quaternion);
		}
	}

#if 0 /* Vrui version - Display button information {{{ */
	/*****************************************/
	/* Display button (2-switch) information */
	if (aux->num_buttons > 0) {
		vrPrintf(ANSI_MOVE "buts: ", start_line + 1, butval_offset);
		for (devnum = 0; devnum < aux->num_buttons; devnum++) {
			vrPrintf("%c", (aux->incoming_buttons[devnum] == 0 ? '.' : 'X'));
		}
	}
#endif /* }}} Vrui version */

#if 0 /* Vrui version -- Display valuator information {{{ */
	/********************************/
	/* Display valuator information */
	if (aux->num_valuators > 0) {
		float	valmin = -1.0;	/* temporary min/max values */
		float	valmax = +1.0;
		//printf(" vals: ");
		for (devnum = 0; devnum < aux->num_valuators; devnum++) {
			vrPrintf(ANSI_MOVE "val %d: %6.3f ", start_line + devnum + 3, butval_offset, devnum, aux->incoming_valuators[devnum]);
#  if 0 /* TODO: add min & max fields to the general vrValuator type */
			if ((vrpn_dev->analog_max[input]-vrpn_dev->analog_min[input]) < 0.0001) {
				/* When we don't have min/max values yet, then fill with dashes */
				vrPrintf(" %11s", "|---------|");
			} else {
			}
#  endif
				percent = (aux->incoming_valuators[devnum]-valmin) / (valmax-valmin);
				vrPrintf(" %11s",
					(percent < 0.11 ? "|-        |" :
					(percent < 0.22 ? "| -       |" :
					(percent < 0.33 ? "|  -      |" :
					(percent < 0.44 ? "|   -     |" :
					(percent < 0.56 ? "|    +    |" :
					(percent < 0.67 ? "|     -   |" :
					(percent < 0.78 ? "|      -  |" :
					(percent < 0.89 ? "|       - |" :
							  "|        -|")))))))));
		}
	}
#endif /* }}} Vrui version */

	/* return to the original cursor position */
	vrPrintf(ANSI_CURSORLOC_RETURN);
	fflush(stdout);
}


/*******************************************************************/
/* A test program to communicate with a DTrack device and print the results. */
int main(int argc, char *argv[])
{
static	char			*err_missing_arg = "%s: missing argument to '%s'.\n";
	_DTrackPrivateInfo	*aux;				/* pointer to device info struct */
	char			*progname;			/* name of the program executable */
	vrInputDevice		device;				/* using FreeVR input device struct */
	int			read_return;			/* return value from the read function */
	int			count;				/* loop counter */

	/* command line options */
	int			rendermode = RENDERLINES;	/* default: how to render the data */
	int			justonce = 0;			/* default: quit after just one output */
	int			nodata_flag = 0;		/* whether or not to actual show input data */

	done = 0;
	signal(SIGINT, exit_testapp);


	/******************************/
	/* setup the device structure */
	aux = (_DTrackPrivateInfo *)malloc(sizeof(*aux));
	memset(aux, 0, sizeof(*aux));
	_DTrackInitializeStruct(aux, "default");

	device.name = "DTrack test device";
	device.operating = 0;
	device.aux_data = aux;


	/*********************************************************/
	/* set default parameters based on environment variables */
	aux->server_host = getenv("DTRACK_HOST");
	if (aux->server_host == NULL)
		aux->server_host = DEFAULT_CMDHOST;		/* default, if no host given */
	aux->server_port = ((getenv("DTRACK_PORT")==NULL) ? DEFAULT_CMDPORT : atoi(getenv("DTRACK_PORT")));
	aux->data_port = ((getenv("DTRACK_DATAPORT")==NULL) ? DEFAULT_DATAPORT : atoi(getenv("DTRACK_DATAPORT")));


	/*****************************************************/
	/* adjust parameters based on command-line-arguments */
	progname = argv[0];
	while ((argc > 1) && (argv[1][0] == '-')) {

		/****************/
		/* Set the port */
		if (!strcmp(argv[1], "-p") || !strcmp(argv[1], "-port")) {
			argv++; argc--;
			if (argc > 1) {
				aux->data_port = atoi(argv[1]);
				argv++; argc--;
			} else {
				fprintf(stderr, err_missing_arg, progname, argv[0]);
				fprintf(stderr, RED_TEXT "Usage:" NORM_TEXT " %s [-screen|-csv] [-list|-nodata|-once|-1] [-p <port>(%d)] [<DTrack host> (default = '%s')]\n",
					progname, aux->data_port, aux->server_host);	/* NOTE: I'm reporting the default based on what might be changed by the environment variable */
				exit(1);
			}
			continue;               /* return to the while */
		}

		/*****************************************************/
		/* Report information about the device and then quit */
		/* NOTE: the "-nodata" option already basically does a "list" of what's coming from the DTrack server */
		if (!strcmp(argv[1], "-nodata") || !strcmp(argv[1], "-list")) {
			nodata_flag = 1;
			argv++; argc--;
		}

		/***************************/
		/* set the "justonce" flag */
		if (!strcmp(argv[1], "-1") || !strcmp(argv[1], "-once")) {
			argv++; argc--;
			justonce = 1;
			continue;               /* return to the while */
		}

		/************************************/
		/* set the "rendermode" flag to CSV */
		if (!strcmp(argv[1], "-csv")) {
			argv++; argc--;
			rendermode = RENDERCSV;
			continue;		/* return to the while */
		}

		/***************************************/
		/* set the "rendermode" flag to screen */
		if (!strcmp(argv[1], "-screen")) {
			argv++; argc--;
			rendermode = RENDERSCREEN;
			continue;		/* return to the while */
		}

		/******************/
		/* Unknown option */
		else {
			/* There are currently no other "-" options, so this is an error */
			fprintf(stderr, RED_TEXT "Usage:" NORM_TEXT " %s [-screen|-csv] [-list|-nodata|-once|-1] [-p <port>(%d)] [<DTrack host> (default = '%s')]\n",
				progname, aux->data_port, aux->server_host);	/* NOTE: I'm reporting the default based on what might be changed by the environment variable */
			exit(1);
		}
	}

	/* if there are any arguments left, use the first as the server host */
	if (argc > 1) {
		aux->server_host = strdup(argv[1]);
	}


	/******************************************************/
	/* open the UDP socket port and initialize the device */
#if 0 /* 0 -- redefine how to connect to the DTrack server / 1 -- use the FreeVR version of the code */	/* both ways work -- VRPN uses '0' */
	aux->fd_socket = vrSocketCreateListenUDP(&(aux->data_port), 1);
	if (aux->fd_socket < 0) {
		aux->open = 0;
		fprintf(stderr, RED_TEXT "couldn't open UDP socket '%d'\n" NORM_TEXT, aux->data_port);
		snprintf(aux->version, sizeof(aux->version), "- unconnected DTrack -");
	} else {
		aux->open = 1;
		if (_DTrackInitializeDevice(aux) < 0) {
			vrErrPrintf("main: " RED_TEXT "Warning, unable to initialize DTrack.\n" NORM_TEXT);
		}
	}
#else /* Use the FreeVR version of the open-function code */
	_DTrackOpenFunction(&device);
#endif

	/**********************/
	/* Display the Output */

	/* Output information about the device and connection */
	_DTrackPrintPrivateInfo(stdout, aux, verbose);

	/* Quit if flagged to print the device info but not data */
	if (nodata_flag)
		exit(0);

	/* For CSV-rendering, output a simple header */
	if (rendermode == RENDERCSV && aux->open) {
		vrPrintf("# ---------- DTrack Output in CSV format ----------\n");
		vrPrintf("#Meta,begin\n");
		vrPrintf("#Header, frame, quality, device, message-type, input-number, input-value(s)\n");
	}

	/* For screen-rendering, scroll everything up so it won't be erased */
	if (rendermode == RENDERSCREEN && !justonce) {
		for (count = 0; count < 60; count++)
			vrPrintf("\n");
vrPrintf("[%d;%dH" " Test message", 28, 55);
	}

	/****************************************/
	/* Read the data and display the output */
	while(aux->open && !done) {

		/* If any new data was received then display it */
		read_return = _DTrackReadInput(aux);
		if (read_return > 0) {
			switch(rendermode) {
			default:
			case RENDERLINES:
				_RenderValuesLine(aux);
				break;;
			case RENDERCSV:
				_RenderValuesCSV(aux);
				break;;
			case RENDERSCREEN:
				_RenderValuesScreen(aux);
				break;;
			}

			/* We're done already if just one printout */
			if (justonce)
				done = 1;
		} else if (read_return == -1) {
			/* There was a communication error during the read, so end */
			done = 1;
		}
	}


	/*****************/
	/* Close up shop */
#if 0 /* 0 -- redefine how to disconnect to the DTrack server / 1 -- use the FreeVR version of the code */	/* both ways work -- VRPN uses '0' */
	if (aux != NULL) {
		if (rendermode == RENDERCSV && aux->open) {
			vrPrintf("#Meta,end\n");
		}

		/* for screen-rendering, correct the terminal mode */
		if (rendermode == RENDERSCREEN && aux->open) {
			/* clean up the screen rendering */
			vrPrintf(ANSI_SETMARGIN_TOPBOT_OFF);
			vrPrintf(ANSI_CURSORLOC_RETURN);
		}

		if (aux->open) {
			_DTrackCloseDevice(aux);
			vrSocketClose(aux->fd_socket);
		}

		free(aux);
	}
#else /* Use the FreeVR version of the close-function code */
	if (aux != NULL) {
		if (rendermode == RENDERCSV && aux->open) {
			vrPrintf("#Meta,end\n");
		}

		/* For screen-rendering, correct the terminal mode */
		if (rendermode == RENDERSCREEN && aux->open) {
			/* clean up the screen rendering */
			vrPrintf(ANSI_SETMARGIN_TOPBOT_OFF);
			vrPrintf(ANSI_CURSORLOC_RETURN);
		}

		/* Shut down communication with the device */
		_DTrackCloseFunction(&device);
	}
#endif

	vrDbgPrintf(BOLD_TEXT "\nDTrack device closed\n" NORM_TEXT);
}

#endif /* } TEST_APP */



	/*===========================================================*/
	/*===========================================================*/
	/*===========================================================*/
	/*===========================================================*/



#if defined(CAVE) /* { */


	/* ... CAVE stuff here if to also work with CAVElib */


#endif /* } CAVE */

