/* ======================================================================
 *
 *  CCCCC          vr_math.c
 * CC   CC         Author(s): Ed Peters, Stuart Levy, Bill Sherman, John Stone
 * CC              Created: (probably) June 4, 1998
 * CC   CC         Last Modified: February 17, 2024 (minor post-release fix 03/04/24)
 *  CCCCC
 *
 * Source file for FreeVR math routines.
 *
 * Copyright 2024, Bill Sherman, All rights reserved.
 * With the intent to provide an open-source license to be named later.
 * ====================================================================== */
/*************************************************************************

FreeVR USAGE for types:
	NOTE: by default all the mathematical types use doubles to represent
		the component values, but float (f) versions also exist, mostly
		for interfacing with software that only provides 4-bytes float
		values.
	...
	- vrMatrix -- represents a 4x4 homogeneous matrix (though accessed as a 1-D 16 element array)
	- vrPoint -- 3-value vector, with 1.0 assumed as a virtual 4th element
	- vrPointf -- float version of "vrPoint"
	- vrVector -- 3-value vector, with 0.0 assumed as a virtual 4th element
	- vrVectorf -- float version of "vrVector"
	- vrVector4 -- only used for matrix decomposition
	- vrQuat -- quaternion stored in i(X), j(Y), k(Z), w(scalar) ordering
	- vrQuatf -- float version of vrQuatf
	- vrEuler -- 6-dof values using rotations and translations (generally avoided, but useful for billboarding & porting old code)
	- vrFrustum -- 6 element structure with left/right/top/bottom/near/far clipping values
	- vrMatrixComposition
	- vrVQSTransform -- presently unused

----------------------------------
A word on Matrix math, and OpenGL:

	Copying matrix data to/from OpenGL is actually done as a 16 value
	array.  The translational elements of that array are:
		12 -- X translation
		13 -- Y translation
		14 -- Z translation

	The overall array/memory ordering is:
		x.x x.y x.z 0  y.x y.y y.z 0  z.x z.y z.z 0  p.x p.y p.z 1

	Thus each basis vector has components in contiguous elements.
	(And in row-major memory order (ala C), rows are in contiguous memory.)

	When the 4x4 array is represented in row-major order, the matrix is:
		x.x x.y x.z 0 -- X basis vector
		y.x y.y y.z 0 -- Y basis vector
		z.x z.y z.z 0 -- Z basis vector
		p.x p.y p.z 1 -- translation vector

	However, the OpenGL documentation represents matrices as though they
	were in column-major order -- with the basis and translation vectors
	represented vertically (which also happens to be how I learned
	homogeneous coordinates) -- following the OpenGL documentation, the
	FreeVR Matrix output routines show the basis vectors as columns.
	Thus in a typical C-array storage, and how OpenGL actually stores
	the matrices (in a 16-element array) is as shown above.

	NOTE: in "OpenGL Programming Guide" Release 1, the matrix is clearly
	shown to be in column-major ordering on page 75 with the caution of
	C's row-major ordering on the top of 76.  Also, appendix G covers
	Homogeneous Matrix operations on pages 475-480, the rotations are
	specifically shown on page 479.

	Here's a StackOverflow (and other) URLs with some details:
		- https://stackoverflow.com/questions/17717600/confusion-between-c-and-opengl-matrix-order-row-major-vs-column-major
		- http://steve.hollasch.net/cgindex/math/matrix/column-vec.html
		- https://en.wikipedia.org/wiki/Row-_and_column-major_order

	As order of multiplication matters:
		* post-multiplying with column-major matrices produces the
		  same result as pre-multiplying with row-major matrices

	These representations can also be referred to as:
		* prefix == translation vector in rightmost column
		* postfix = translation vector in bottom row

	Within the FreeVR vrMatrix defined type, elements are stored as a
	single indexed array (one dimension) of 16 double precision elements.
	FreeVR provides two #define macros to access individual elements of
	the array as though it were a 4x4 matrix:
		#define VRMAT_MAXDIM 4
		#define VRMAT_ROWCOL(m,r,c) ((m)->v[(r)+VRMAT_MAXDIM*(c)]) -- row major
			0,0 -> 0
			1,0 -> 1
			2,0 -> 2
			3,0 -> 3
			0,1 -> 4
		#define VRMAT_COLROW(m,c,r) ((m)->v[(r)+VRMAT_MAXDIM*(c)]) -- this should just be COLUMN_MAJOR

	Thus, for row-major order:
	TODO: confirm this
		v[0]   v[4]   v[8]  v[12]     rc(x,x)  rc(x,y)  rc(x,z)  rc(x,w)
		v[1]   v[5]   v[9]  v[13]  =  rc(y,x)  rc(y,y)  rc(y,z)  rc(y,w)
		v[2]   v[6]  v[10]  v[14]     rc(z,x)  rc(z,y)  rc(z,z)  rc(z,w)
		v[3]   v[7]  v[11]  v[15]     rc(w,x)  rc(w,y)  rc(w,z)  rc(w,w)

	Coordinate Systems (Y-up vs. Z-up):
		OpenGL is right-handed Y-up:
			* +X to the right
			* +Y up
			* +Z into the screen (away from viewer)
			* Positive rotation is counterclockwise about the axis

		FreeVR is also right-handed Y-up (matches OpenGL).
		Performer is right-handed Z-up.
		OpenSceneGraph (OSG) is Y-up.
		Vrui is Z-up (defacto).
----------------------------------

HISTORY:
	5 June 1998 -- First version written by Ed Peters (probably).
		...
		[corresponds to FreeVR Version 0.1a]

	...

	25-26 January 2021 (Bill Sherman) -- 
		...
		- added (i,j,k) to quaternion print output
		- added this file top-comment section

		[corresponds to FreeVR Version 0.6g]

	8-9 April 2021 (Bill Sherman) -- 
		...
		- did some variable name cleanup in the matrix operation section

		I slightly reordered the output of the matrix decomposition
		data when printed -- moved the translation sooner and the
		shear later.

		[corresponds to FreeVR Version 0.6g]

	4-5 May 2021 (Bill Sherman)
		Deleted some old and deprecated code.
		Minor clean-up to make sizeof() safer.
		Renamed many arguments to make them more consistent.

		[corresponds to FreeVR Version 0.7b]

	30 March & 2 April 2022 (Bill Sherman)
		Enhanced some of the matrix print functions to duplicate any
		  (non-newline) whitespace contained at the beginning of the
		  "before" argument to help with aligned indentation.

		[corresponds to FreeVR Version 0.7d]

	17 February 2024 (Bill Sherman)
		Enhanced the top comments a bit (yesterday).
		Enhanced the vrMatrixCompositionFromMatrix() function to now
		  include the value for "singularity" as one of the components.
		  And now also return all zero values for the components when
		  the matrix is singular.  (Well, returning VR_QW as 1.0 for
		  the quaternion -- I'm undecided on which is better.)

		[corresponds to FreeVR Version 0.7e]

	4 March 2024 (Bill Sherman)
		Fixed a small bug in the Matrix decomposition code found while
		  doing the release testing.

		[corresponds to FreeVR Version 0.7e (post)]

TODO:
	* confirm (and probably fix) the Euler/Quaternion/Matrix conversion
		routines
		[04/02/2022: It's possible, perhaps likely, I did this already.]

	* DONE: move the time functions elsewhere [01/25/2021]
		[Fixed 05/05/2021 for FreeVR version 0.7b -- moved to vr_system.c]

	* DONE: move the "misc" functions to vr_utils.c [01/26/2021]
		[Fixed 05/05/2021 for FreeVR version 0.7b -- moved to vr_utils.c]
		

*************************************************************************/

#if !defined(FREEVR) && !defined(TEST_APP) && !defined(MATH_ONLY)
#  define	FREEVR
#  undef	TEST_APP
#  undef	MATH_ONLY
#endif

/* 01/25/2021: DONE: see which of these header inclusions can be removed */
/* 01/25/2021: DONE: move the time functions elsewhere (vr_system.c) */
/* 01/26/2021: DONE: move the "misc" functions to vr_utils.c */
/* Header includes needed for the both the FreeVR library, and test utility */
#include <math.h>	/* needed for sqrt(), cos(), sin(), asin(), atan2() */
#include <float.h>	/* needed for #define of FLT_EPSILON */
#include <string.h>	/* needed for memcpy() & memset() */
#if 0 /* test me */
#include <memory.h>	/* seems unneeded in Linux or MacOS -- do other OSes require it? */
#endif

#include "vr_math.h"

#if defined(FREEVR)
/* Header includes needed for the FreeVR library, but not the test utility */
#  include "vr_debug.h"
#  include "vr_shmem.h"
#endif

#if defined(TEST_APP) || defined(MATH_ONLY) /* { */
/* Includes/defines needed for the "gfxtest" and other non-FreeVR utilities */

#  include <stdio.h>	/* needed for FILE type and [f]printf() in TEST_APP */
#  include <stdlib.h>	/* needed for alloc(), free() */
#  include "vr_debug.h"	/* needed for the ANSI #defines (for "gfxtextwm"), so change to "vr_ansi.h" when created */

#if 0 /* this is not needed anymore (since the time functions where moved */ /* delete me */
/* Declare some alternate and dummy functions, macros and structures to  */
/*   allow the "mathtest" application to be compiled without complaints. */

typedef	double	vrTime;
struct temp_struct {
		vrTime	time_immemorial;
		vrTime	paused_time;
		vrTime	pause_wtime;
	} *vrContext;
#endif

void *vrShmemAlloc0(size_t size)
{
	void *buf = malloc(size);
	memset(buf, 0, size);
	return buf;
}
void vrShmemFree(void *buf) { free(buf); }

/* First undefine some macros defined in other headers that might */
/*   be overridden here.                                          */
#  undef	vrAtoI
#  undef	vrErrPrintf
#  undef	vrErr
/* Now #define faux versions of most of the debug helpers */
#  define	MATH_DBGLVL		260
#  define	ALMOSTNEVER_DBGLVL	1000
#  define	vrPrintf	printf
#  define	vrFprintf	fprintf
#  define	vrDbgPrintf	printf
#  ifndef vrDbgPrintfN	/* don't redefine the macro if already #define'd */
//#    define	AltvrDbgPrintfN(a,b,c,d,e,f,g,h,i,j,k)	/* print nothing (for different compilers, different forms) */;
#    define	vrDbgPrintfN(...)			/* print nothing (for different compilers, different forms) */;
#  endif
#  define	vrMsgPrintf	printf
#  define	vrErrPrintf	printf
#  define	vrErr(s)	{ printf(s); printf("\n"); }

#endif /* } TEST_APP || MATH_ONLY */

#if defined(TEST_APP) && !defined(MATH_ONLY) /* { */
/* NOTE: we won't need this once I create "vr_ansi.c" */
//#  define	ANSI_CSI	"\033["
//#  define	NORM_TEXT	ANSI_CSI "0m"
//#  define	BOLD_TEXT	ANSI_CSI "1m"
//#  define	RED_TEXT	ANSI_CSI "31m"
//#  define	SET_TEXT	ANSI_CSI "%dm"
#  undef	vrDbgDo
#  define	vrDbgDo(n)	1			/* Note, redefining this here can override expectations in other source files. */
#endif /* } TEST_APP && !MATH_ONLY */


#if 0 /* future delete me -- after some time to make sure things aren't broken */
/*****************************************************************/
/*                       TIME routines                           */
/*****************************************************************/

/*********************************************************************/
vrTime vrStartTime()
{
	if (vrContext == NULL)
		return 0.0;

	return vrContext->time_immemorial;
}


/*********************************************************************/
vrTime vrCurrentWallTime()
{
	vrTime		time_now;
#if 1
	struct timeval	now;

	gettimeofday(&now, NULL);
	time_now = ((vrTime)now.tv_sec + (vrTime)now.tv_usec / 1000000.0);
#else
	struct timespec	now;

	clock_gettime(CLOCK_SGI_CYCLE, &now);
	time_now = ((vrTime)now.tv_sec + (vrTime)now.tv_nsec / 1000000000.0);
#endif

	return (time_now);
}


/*********************************************************************/
vrTime vrSimTimeOf(vrTime given_wtime)
{
	if (vrContext->pause_wtime > 0.0)
		return (vrContext->pause_wtime - vrStartTime() - vrContext->paused_time);
	else	return (           given_wtime - vrStartTime() - vrContext->paused_time);
}


/*********************************************************************/
vrTime vrCurrentSimTime()
{
	return vrSimTimeOf(vrCurrentWallTime());
}


/*********************************************************************/
vrTime vrTimeSince(vrTime then)
{
	return vrCurrentWallTime() - then;
}
#endif



/* ==================================================== */
/* MATRIX OPERATIONS                                    */
/* ==================================================== */


/**************************************/
/****** Matrix Creation Routines ******/
/**************************************/

/*********************************************************************/
/* Create an empty (all zeros) 4x4 matrix                            */
vrMatrix *vrMatrixCreate(void)
{
	return (vrMatrix *)vrShmemAlloc0(sizeof(vrMatrix));
}


/*********************************************************************/
/* Delete the given matrix (freeing it's memory)                     */
void vrMatrixDelete(vrMatrix *mat)
{
	vrShmemFree(mat);
}


/**************************************************************************/
/* Compare the values of two matrices                                     */
/* NOTE: if either matrix is NULL, return a non-equivalency value (ie. 0) */
/*   We may want to consider a third result for this: -1.                 */
/*   Just added a new check, such that if both matrices are NULL, then    */
/*   they are considered to be the same.                                  */
int vrMatrixEqual(vrMatrix *mat_dst, vrMatrix *mat_r)
{
	int	count;

	if (mat_dst == NULL && mat_r == NULL)
		return 1;

	if (mat_dst == NULL)
		return 0;

	if (mat_r == NULL)
		return 0;

	for (count = 0; count < 16; count++) {
		if (mat_dst->v[count] != mat_r->v[count])
			return 0;
	}

	return 1;
}


/********************************************************************/
/* Copy the given matrix; ie, mat_dst = mat_r                         */
vrMatrix *vrMatrixCopy(vrMatrix *mat_dst, vrMatrix *mat_r)
{
	if (mat_dst == NULL) {
#if 0 /* for debugging on machines that don't crash when mat_dst == NULL */
#include <unistd.h>	/* needed by getpid() */
		vrErrPrintf("vrMatrixCopy(): asked to copy to NULL memory.\n"
			" -- pausing -- use 'gdb -p %d'\n", getpid());
		pause();
#endif
	}

	if (mat_r == NULL) {
#if 0 /* for more debugging on machines that don't crash when mat_r == NULL */
		vrErrPrintf("vrMatrixCopy(): asked to copy to NULL memory.\n"
			" -- pausing -- use 'gdb -p %d'\n", getpid());
		pause();
#endif
		/* when not debugging, just return the identity matrix */
		vrMatrixSetIdentity(mat_dst);
	} else {
		memcpy(mat_dst, mat_r, sizeof(*mat_dst));
	}

	return mat_dst;
}


/*********************************************************************/
/* Set an existing 4x4 matrix to the identity matrix               */
vrMatrix *vrMatrixSetIdentity(vrMatrix *mat_dst)
{
	return vrMatrixSetScale3d(mat_dst, 1.0f, 1.0f, 1.0f);
}


/*********************************************************************/
/* Create a 4x4 identity matrix                                      */
vrMatrix *vrMatrixCreateIdentity(void)
{
	return vrMatrixSetScale3d(vrMatrixCreate(), 1.0f, 1.0f, 1.0f);
}


/*********************************************************************/
/* Set an existing 4x4 matrix to the identity matrix in the rotation */
/*   portion only.                                                   */
vrMatrix *vrMatrixSetIdentityRotationOnly(vrMatrix *mat_dst)
{
	vrVector	trans_vec;	/* temp storage of the translation */

	/* m2o: move matrix to origin and store offset */
	vrMatrixGetResetTranslationAd(mat_dst, trans_vec.v);

	/* set the entire matrix to the identity */
	return vrMatrixSetScale3d(mat_dst, 1.0f, 1.0f, 1.0f);

	/* m2o-1: return to original translation */
	vrMatrixSetTranslationOnlyAd(mat_dst, trans_vec.v);

	return (mat_dst);
}


/******************************************************************************/
/* [01/12/21: NOTE: This code is actually untested as it's only "inclusion"   */
/*   is in a #if-d out section of the "mathtest" program.                     */
/*   Furthermore, I'm not 100% convinced these are actually in the order      */
/*   used by FreeVR.  This will need to be confirmed.                         */
/*                                                                            */
/* Set a 4x4 transform matrix from an array of 16 doubles                     */
/* Here is how the OpenGL-style matrix corresponds to                         */
/*   the matrix array elements:                                               */
/*        v[0]   v[4]   v[8]  v[12]                                           */
/*        v[1]   v[5]   v[9]  v[13]                                           */
/*        v[2]   v[6]  v[10]  v[14]                                           */
/*        v[3]   v[7]  v[11]  v[15]                                           */
/* The transpose of this looks more familiar:                                 */
/*        a  b  c  0                                                          */
/*        d  e  f  0                                                          */
/*        g  h  i  0                                                          */
/*        x  y  z  1                                                          */
/* where a-i is the 3x3 rotation matrix and [x,y,z] is the translation vector */
vrMatrix *vrMatrixSetAd(vrMatrix *mat_dst, double *array)
{
	mat_dst->v[ 0] = array[ 0];		/* a */
	mat_dst->v[ 1] = array[ 1];		/* b */
	mat_dst->v[ 2] = array[ 2];		/* c */
	mat_dst->v[ 3] = array[ 3];		/* 0 */
	mat_dst->v[ 4] = array[ 4];		/* d */
	mat_dst->v[ 5] = array[ 5];		/* e */
	mat_dst->v[ 6] = array[ 6];		/* f */
	mat_dst->v[ 7] = array[ 7];		/* 0 */
	mat_dst->v[ 8] = array[ 8];		/* g */
	mat_dst->v[ 9] = array[ 9];		/* h */
	mat_dst->v[10] = array[10];		/* i */
	mat_dst->v[11] = array[11];		/* 0 */
	mat_dst->v[12] = array[12];		/* x */
	mat_dst->v[13] = array[13];		/* y */
	mat_dst->v[14] = array[14];		/* z */
	mat_dst->v[15] = array[15];		/* 1 */

	return mat_dst;
}


/*********************************************************************/
/* Set matrix to be a 4x4 translation matrix from 3 doubles          */
vrMatrix *vrMatrixSetTranslation3d(vrMatrix *mat_dst, double x, double y, double z)
{
	vrMatrixSetIdentity(mat_dst);
	VRMAT_ROWCOL(mat_dst, VR_X, VR_W) = x;
	VRMAT_ROWCOL(mat_dst, VR_Y, VR_W) = y;
	VRMAT_ROWCOL(mat_dst, VR_Z, VR_W) = z;

	return mat_dst;
}


/*********************************************************************/
/* Set a 4x4 translation matrix from an array of 3 doubles           */
vrMatrix *vrMatrixSetTranslationAd(vrMatrix *mat_dst, double *array)
{
	vrMatrixSetIdentity(mat_dst);
	VRMAT_ROWCOL(mat_dst, VR_X, VR_W) = array[VR_X];
	VRMAT_ROWCOL(mat_dst, VR_Y, VR_W) = array[VR_Y];
	VRMAT_ROWCOL(mat_dst, VR_Z, VR_W) = array[VR_Z];

	return mat_dst;
}


/**************************************************************************/
/* Set only translation values of a 4x4 matrix from an array of 3 doubles */
vrMatrix *vrMatrixSetTranslationOnlyAd(vrMatrix *mat_dst, double *array)
{
	VRMAT_ROWCOL(mat_dst, VR_X, VR_W) = array[VR_X];
	VRMAT_ROWCOL(mat_dst, VR_Y, VR_W) = array[VR_Y];
	VRMAT_ROWCOL(mat_dst, VR_Z, VR_W) = array[VR_Z];

	return mat_dst;
}


/************************************************************************************/
/* Get the translation values into an array of 3 doubles from a 4x4 mat,            */
/*   and then reset the current matrix's translation values to zero.                */
/*   TODO: usage of this should go away once we fully adopt the vrVector operations */
/*   [1/16/02] Hmmm, actually, this can be used by the vector class by passing the  */
/*     "v" field.  And is an efficient way of translating to origin and back for    */
/*     rotations about the origin.                                                  */
vrMatrix *vrMatrixGetResetTranslationAd(vrMatrix *mat_dst, double *array)
{
	array[VR_X] = VRMAT_ROWCOL(mat_dst, VR_X, VR_W);
	array[VR_Y] = VRMAT_ROWCOL(mat_dst, VR_Y, VR_W);
	array[VR_Z] = VRMAT_ROWCOL(mat_dst, VR_Z, VR_W);

	VRMAT_ROWCOL(mat_dst, VR_X, VR_W) = 0.0;
	VRMAT_ROWCOL(mat_dst, VR_Y, VR_W) = 0.0;
	VRMAT_ROWCOL(mat_dst, VR_Z, VR_W) = 0.0;

	return mat_dst;
}


/*********************************************************************/
/* Set an existing 4x4 translation matrix from an array of 6 doubles */
/*   from X,Y,Z location and Azim, Elev, Roll orientation.           */
vrMatrix *vrMatrixSetXYZAERAd(vrMatrix *mat_dst, double *array)
{
	vrMatrixSetIdentity(mat_dst);

	VRMAT_ROWCOL(mat_dst, VR_X, VR_W) = array[VR_X];
	VRMAT_ROWCOL(mat_dst, VR_Y, VR_W) = array[VR_Y];
	VRMAT_ROWCOL(mat_dst, VR_Z, VR_W) = array[VR_Z];

	vrMatrixPostRotateId(mat_dst, VR_Y, array[VR_AZIM+3]);
	vrMatrixPostRotateId(mat_dst, VR_X, array[VR_ELEV+3]);
	vrMatrixPostRotateId(mat_dst, VR_Z, array[VR_ROLL+3]);

	return mat_dst;
}


/*********************************************************************/
/* Set an existing 4x4 translation matrix from an array of 6 floats  */
/*   from X,Y,Z location and Azim, Elev, Roll orientation.           */
vrMatrix *vrMatrixSetXYZAERAf(vrMatrix *mat_dst, float *array)
{
	vrMatrixSetIdentity(mat_dst);

	VRMAT_ROWCOL(mat_dst, VR_X, VR_W) = array[VR_X];
	VRMAT_ROWCOL(mat_dst, VR_Y, VR_W) = array[VR_Y];
	VRMAT_ROWCOL(mat_dst, VR_Z, VR_W) = array[VR_Z];

	vrMatrixPostRotateId(mat_dst, VR_Y, (double)(array[VR_AZIM+3]));
	vrMatrixPostRotateId(mat_dst, VR_X, (double)(array[VR_ELEV+3]));
	vrMatrixPostRotateId(mat_dst, VR_Z, (double)(array[VR_ROLL+3]));

	return mat_dst;
}


/***********************************************************************/
/* Set an existing 4x4 translation matrix from a vrEuler, with the     */
/*   axis about which the azimuth is derived specified.  The elevation */
/*   and roll axes are assumed to follow in logical fashion.           */
/* TODO: One day, and that day may never come, we may need an Euler conversion   */
/*   function that is more general than this, allowing any axis to be associated */
/*   with any rotation about a body, and in either direction.  However, until    */
/*   that days comes, or someone is sufficiently bored, that function won't be   */
/*   written.                                                                    */
/* TODO: rename this to be vrMatrixSetFromEulerAzimaxis().             */ /* delete me -- actually, do me */
vrMatrix *vrMatrixSetEulerAzimaxis(vrMatrix *mat_dst, vrEuler *euler, int axis)
{
	vrMatrixSetIdentity(mat_dst);

	/*** copy the translational values ***/
	VRMAT_ROWCOL(mat_dst, VR_X, VR_W) = euler->t[VR_X];
	VRMAT_ROWCOL(mat_dst, VR_Y, VR_W) = euler->t[VR_Y];
	VRMAT_ROWCOL(mat_dst, VR_Z, VR_W) = euler->t[VR_Z];

	/*** calculate the rotation from the Euler angles ***/
	switch (axis) {
	case VR_X:	/* ie. an X-up coordinate system (an unlikely choice) */
		vrMatrixPostRotateId(mat_dst, VR_X, euler->r[VR_AZIM]);
		vrMatrixPostRotateId(mat_dst, VR_Z, euler->r[VR_ELEV]);
		vrMatrixPostRotateId(mat_dst, VR_Y, euler->r[VR_ROLL]);
		break;
	case VR_Y:	/* ie. a Y-up coordinate system */
		vrMatrixPostRotateId(mat_dst, VR_Y, euler->r[VR_AZIM]);
		vrMatrixPostRotateId(mat_dst, VR_X, euler->r[VR_ELEV]);
		vrMatrixPostRotateId(mat_dst, VR_Z, euler->r[VR_ROLL]);
		break;
	default:
		vrMsgPrintf("Warning, incorrect azimuth axis specified in Euler conversion, using 'Z'.\n");
	case VR_Z:	/* ie. a Z-up coordinate system */
		vrMatrixPostRotateId(mat_dst, VR_Z, euler->r[VR_AZIM]);
		vrMatrixPostRotateId(mat_dst, VR_Y, euler->r[VR_ELEV]);
		vrMatrixPostRotateId(mat_dst, VR_X, euler->r[VR_ROLL]);
		break;
	}

	return mat_dst;
}


/*********************************************************************/
vrEuler *vrEulerSetFromMatrix(vrEuler *euler, vrMatrix *mat_src)
{
	double	azim;
	double	elev;
	double	roll;
	double	azim_sin;
	double	azim_cos;

	/*** first handle the translational values -- very easy ***/
	euler->t[VR_X] = VRMAT_ROWCOL(mat_src, VR_X, VR_W);
	euler->t[VR_Y] = VRMAT_ROWCOL(mat_src, VR_Y, VR_W);
	euler->t[VR_Z] = VRMAT_ROWCOL(mat_src, VR_Z, VR_W);

	/*** set the rotational values ***/
	/* This section of code calculates the proper Euler angles from */
	/*   the 4x4 matrix using techniques from Robotics, Fu et.al.   */
	/*   Actually, the matrix order seems backward, and I lucked    */
	/*   into getting them backward - figure it out later, it works!*/
	azim = atan2(
		VRMAT_ROWCOL(mat_src, VR_X, VR_Z),
		VRMAT_ROWCOL(mat_src, VR_Z, VR_Z));
	azim_sin = sin(azim);
	azim_cos = cos(azim);

	elev = atan2(
		VRMAT_ROWCOL(mat_src, VR_Y, VR_Z),
		azim_sin*VRMAT_ROWCOL(mat_src, VR_X, VR_Z) + azim_cos*VRMAT_ROWCOL(mat_src, VR_Z, VR_Z));
	roll = atan2(
		azim_sin*VRMAT_ROWCOL(mat_src, VR_Z, VR_Y) - azim_cos*VRMAT_ROWCOL(mat_src, VR_X, VR_Y),
		azim_cos*VRMAT_ROWCOL(mat_src, VR_X, VR_X) - azim_sin*VRMAT_ROWCOL(mat_src, VR_Z, VR_X));

#if 1 /* NOTE: 5/2/03: it seems as though the Azim and Elev values are swapped from the other lib */ /* 01/29/13 -- changed to "1" for testing */ /* 03/18/15 considered changing to "0", but issue was more with knowing that elevation is the X-rotation */
	euler->r[VR_AZIM] =  RTOD(azim);
	euler->r[VR_ELEV] = -RTOD(elev);
#else
	euler->r[VR_ELEV] =  RTOD(azim);
	euler->r[VR_AZIM] = -RTOD(elev);
#endif
	euler->r[VR_ROLL] =  RTOD(roll);

//vrPrintEuler(">>>>", euler);

	return euler;
}


/**********************************************************************/
/* vrEulerSetFromQuatPoint(): this function takes Quaternion (vrQuat) */
/*   and a 3D point (vrPoint), and returns the equivalent values in a */
/*   vrEuler structure, which contains six values, three for the Euler*/
/*   angles, and three for the translational values.                  */
vrEuler *vrEulerSetFromQuatPoint(vrEuler *euler_dst, vrQuat *quat_src, vrPoint *pnt_src)
{
	/* NOTE: these definitions aren't needed for "method 3" */
	double	qx = quat_src->v[VR_X];
	double	qy = quat_src->v[VR_Y];
	double	qz = quat_src->v[VR_Z];
	double	qw = quat_src->v[VR_W];

	double	qx_squared = qx * qx;
	double	qy_squared = qy * qy;
	double	qz_squared = qz * qz;

	double	partial0;
	double	partial1;
	double	partial2;
	double	partial3;
	double	partial4;

	/*** first handle the translational values -- very easy ***/
	euler_dst->t[VR_X] = pnt_src->v[VR_X];
	euler_dst->t[VR_Y] = pnt_src->v[VR_Y];
	euler_dst->t[VR_Z] = pnt_src->v[VR_Z];

	/*** set the rotational values ***/
#if 1 /* conversion method 1 -- I think I fixed this [02/11/21] { */ 
	/* Using code example from: http://www.euclideanspace.com/maths/geometry/rotations/conversions/quaternionToEuler/index.htm */
	/* NOTE: 02/11/21: There are some very odd things with this   */
	/*   method.  For one, it seems as though the ELEV & ROLL     */
	/*   values are swapped.  Assuming that, the ROLL value also  */
	/*   seems to be reflected when it hits +90, with a 180.0     */
	/*   value in the ELEV/attitude entry.  Also, the singularity */
	/*   test was much too low -- was |0.499| but that covers     */
	/*   several degrees and forces the other two rotations to    */
	/*   fixed values, so that's pretty horrible!  So I changed   */
	/*   that value to |0.499999|.                                */
	/* On the plus side, I think it does a better job with what   */
	/*   it calls "attitude" (ie. elevation) than method 3 -- not */
	/*   that method 3 is wrong, it just adds two 180.0 rotations */
	/*   about the X and Z axes that aren't necessary.            */
	/* Likewise, method 2 does the double-180 when the Azimuth/   */
	/*   heading is not in the range -90 to 90, which is odd for  */
	/*   the azimuth, that's really the one that should definitely*/
	/*   go the full 360.                                         */
	double	singularity_test;

	/* First test for singularities */
	//singularity_test = quat_src->v[VR_X] * quat_src->v[VR_Y] + quat_src->v[VR_Z] * quat_src->v[VR_W];
	singularity_test = qx * qy + qz * qw;
	if (singularity_test > 0.499999) {
		/* singularity at north pole */
		vrPrintf("vrEulerSetFromQuatPoint(): north singularity. ******************************\n");
		euler_dst->r[VR_AZIM] = RTOD( 2.0 * atan2(qx, qw));		/* aka heading */
		euler_dst->r[VR_ELEV] = RTOD( PI / 2.0);			/* aka attitude */
		euler_dst->r[VR_ROLL] = RTOD( 0.0);				/* aka bank */
	} else if (singularity_test < -0.499999) {
		/* singularity at south pole */
		vrPrintf("vrEulerSetFromQuatPoint(): south singularity. ----------------------------------\n");
		euler_dst->r[VR_AZIM] = RTOD(-2.0 * atan2(qx, qw));		/* aka heading */
		euler_dst->r[VR_ELEV/*was VR_ROLL*/] = RTOD( 0.0);		/* aka attitude */
		euler_dst->r[VR_ROLL/*was VR_ELEV*/] = RTOD(-PI / 2.0);	/* aka bank */
	} else {
		/* normal values */
//		vrPrintf("vrEulerSetFromQuatPoint(): no singularity.\n");

		partial0 = 2.0 * (qy * qw - qx * qz);
		partial1 = 1.0 - 2.0 * (qy_squared - qz_squared);
		euler_dst->r[VR_AZIM] = RTOD(atan2(partial0, partial1));	/* aka heading */
		//euler_dst->r[VR_AZIM] = RTOD(atan2(2.0 * (quat_src->v[VR_Y] * quat_src->v[VR_W]
		//				    - quat_src->v[VR_X] * quat_src->v[VR_Z]),
		//				1 - 2.0 * (qy_squared - qz_squared)));
		partial2 = 2.0 * (qx * qw - qy * qz);
		partial3 = 1.0 - 2.0 * (qx_squared + qz_squared);
		euler_dst->r[VR_ELEV/*was VR_ROLL*/] = RTOD(atan2(partial2, partial3));	/* aka attitude */
		//euler_dst->r[VR_ELEV/*was VR_ROLL*/] = RTOD(atan2(2.0 * (quat_src->v[VR_X] * quat_src->v[VR_W]
		//				    - quat_src->v[VR_Y] * quat_src->v[VR_Z]),
		//				1 - 2.0 * (qx_squared + qz_squared)));
		euler_dst->r[VR_ROLL/*was VR_ELEV*/] = RTOD(asin(2.0 * singularity_test));	/* aka bank */
	}
//	vrPrintEuler("Quat->Euler method 1:", euler_dst);
	return euler_dst;
#endif /* } "method 1" */
#if 0 /* conversion method 2 { */ 
	/* this code from: https://math.stackexchange.com/questions/2975109/how-to-convert-euler-angles-to-quaternions-and-get-the-same-euler-angles-back-fr (see answer) */
	/* And actually, it basically comes from the Wikipedia article on */
	/*   Euler to Quaternion (and back) conversions for "intrinsic    */
	/*   Tait–Bryan angles following the z-y′-x″ convention."  Though */
	/*   even there pitch and azimuth are exchanged, probably because */
	/*   it really should be z-x'-y".                                 */
	/* https://en.wikipedia.org/wiki/Rotation_formalisms_in_three_dimensions#Euler_angles_%E2%86%94_quaternion */

	partial2 = 2.0 * (qw * qy - qz * qx);
//vrPrintf("p2 = %lf -- ", partial2);
	if (partial2 > 2.3) //1.0)
		partial2 = 1.0;
	if (partial2 < -2.3) //-1.0)
		partial2 = -1.0;
	euler_dst->r[VR_AZIM] = RTOD(asin(partial2));

	partial0 = 2.0 * (qw * qx + qy * qz);
	partial1 = 1.0 - 2.0 * (qx_squared + qy_squared);
	euler_dst->r[VR_ELEV] = RTOD(atan2(partial0, partial1));

	partial3 = 2.0 * (qw * qz + qx * qy);
	partial4 = 1.0 - 2.0 * (qy_squared + qz_squared);
	euler_dst->r[VR_ROLL] = RTOD(atan2(partial3, partial4));

	vrPrintEuler("Quat->Euler method 2:", euler_dst);
	return euler_dst;
#endif /* } "method 2" */
#if 0 /* method 3 { */
	vrMatrix	temp_mat;
	vrMatrixSetFromQuat(&temp_mat, quat_src);
	vrMatrixSetTranslationOnlyAd(&temp_mat, pnt_src->v);
	vrEulerSetFromMatrix(euler_dst, &temp_mat);		// Also calls vrPrintEuler()
	vrPrintEuler("Quat->Euler method 3:", euler_dst);	// This will always echo the call from vrEulerSetFromMatrix()
#endif /* } "method 3" */
vrPrintf("\n");
	return euler_dst;
}


/*********************************************************************/
/* vrCoordsSetIdentity(): Create three 3-tuples of points that will  */
/*   produce an identity matrix when processed by the function       */
/*   vrMatrixSetFromCoords().                                        */
/* The "coords" argument is a pointer to a 9-tuple array of doubles  */
/*   that represent the three 3-tuple points (LL, LR, UL).           */
void vrCoordsSetIdentity(double *coords)
{
	/* LL: For an identity translation, the lower left point is at origin */
	coords[VR_X + 0] = coords[VR_Y + 0] = coords[VR_Z + 0] = 0.0;

	/* LR: For Y-up, +X is to the right */
	coords[VR_X + 3] = coords[VR_Y + 3] = coords[VR_Z + 3] = 0.0;
	coords[VR_X + 3] = 1.0;

	/* UL: For Y-up, +Y is up */
	coords[VR_X + 6] = coords[VR_Y + 6] = coords[VR_Z + 6] = 0.0;
	coords[VR_Y + 6] = 1.0;

	return;
}

/*********************************************************************/
/* Set an existing 4x4 translation matrix from a coordinate triplet. */
/*   This is primarily used to configure the windows based on three  */
/*   locations in space (lower-left point, lower-right point, upper- */
/*   left point.                                                     */
vrMatrix *vrMatrixSetFromCoords(vrMatrix *mat_dst, double *coords)
{
	vrPoint		P_ll, P_lr, P_ul;	/* three points that describe a rectangular window*/
	vrVector	x;			/* vector from P_ll to P_lr */
	vrVector	y_basis;		/* vector from P_ll to P_ul */
	vrVector	y;			/* y_basis vector with x vector component trimmed */
	vrVector	xhat, yhat, zhat;	/* normalized x, y & z vectors (orthogonal) */
	vrMatrix	Tw2rw;			/* Transform matrix from window to real-world */
	vrVector	tmpvec;

	/* step1: get the three coordinates from the configuration */
	vrPointSet3d(&P_ll, coords[0], coords[1], coords[2]);
	vrPointSet3d(&P_lr, coords[3], coords[4], coords[5]);
	vrPointSet3d(&P_ul, coords[6], coords[7], coords[8]);
	vrDbgPrintfN(MATH_DBGLVL, "vrMatrixSetFromCoords():1 P_ll = (%lf %lf %lf), P_lr = (%lf %lf %lf), P_ul = (%lf %lf %lf)\n",
		P_ll.v[0], P_ll.v[1], P_ll.v[2],
		P_lr.v[0], P_lr.v[1], P_lr.v[2],
		P_ul.v[0], P_ul.v[1], P_ul.v[2]);

	/* step2: calculate the x & y vectors */
	vrVectorFromTwoPoints(&x, &P_ll, &P_lr);
	vrVectorFromTwoPoints(&y_basis, &P_ll, &P_ul);
	/* from Lengyel pg 9, equation 1.19 -- perp-q(P), ie. the perpendicular-to-Q element of P */
	/* Y = Yb - X*(X dot Yb / X dot X)  -- ie. subtract the projection of Yb onto X from Yb resulting in a Y orthogonal to X */
	vrVectorSubtract(&y, &y_basis, vrVectorScale(vrVectorCopy(&tmpvec, &x), (vrVectorDotProduct(&x, &y_basis) / vrVectorDotProduct(&x, &x))));
	vrDbgPrintfN(MATH_DBGLVL, "vrMatrixSetFromCoords():2 x = (%lf %lf %lf), y_basis = (%lf %lf %lf), y = (%lf %lf %lf)\n",
		x.v[0], x.v[1], x.v[2],
		y_basis.v[0], y_basis.v[1], y_basis.v[2],
		y.v[0], y.v[1], y.v[2]);


	/* step3: calculate the unit X & Y vectors */
	vrVectorNormalize(&xhat, &x);
	vrVectorNormalize(&yhat, &y);

	/* step4: calculate the unit Z vector */
	vrVectorCrossProduct(&zhat, &xhat, &yhat);
	vrDbgPrintfN(MATH_DBGLVL, "vrMatrixSetFromCoords():3 x^ = (%lf %lf %lf), y^ = (%lf %lf %lf), z^ = (%lf %lf %lf)\n",
		xhat.v[0], xhat.v[1], xhat.v[2],
		yhat.v[0], yhat.v[1], yhat.v[2],
		zhat.v[0], zhat.v[1], zhat.v[2]);

	/* step5: Make the Tw2rw matrix */
	vrMatrixSetIdentity(&Tw2rw);
	/* remember to use coordinates based on OpenGL documentation */
	VRMAT_ROWCOL(&Tw2rw, VR_X, VR_X) = xhat.v[VR_X];
	VRMAT_ROWCOL(&Tw2rw, VR_Y, VR_X) = xhat.v[VR_Y];
	VRMAT_ROWCOL(&Tw2rw, VR_Z, VR_X) = xhat.v[VR_Z];
	VRMAT_ROWCOL(&Tw2rw, VR_X, VR_Y) = yhat.v[VR_X];
	VRMAT_ROWCOL(&Tw2rw, VR_Y, VR_Y) = yhat.v[VR_Y];
	VRMAT_ROWCOL(&Tw2rw, VR_Z, VR_Y) = yhat.v[VR_Z];
	VRMAT_ROWCOL(&Tw2rw, VR_X, VR_Z) = zhat.v[VR_X];
	VRMAT_ROWCOL(&Tw2rw, VR_Y, VR_Z) = zhat.v[VR_Y];
	VRMAT_ROWCOL(&Tw2rw, VR_Z, VR_Z) = zhat.v[VR_Z];
	/* the translation is by the lower-left corner */
	VRMAT_ROWCOL(&Tw2rw, VR_X, VR_W) = P_ll.v[VR_X];
	VRMAT_ROWCOL(&Tw2rw, VR_Y, VR_W) = P_ll.v[VR_Y];
	VRMAT_ROWCOL(&Tw2rw, VR_Z, VR_W) = P_ll.v[VR_Z];
	if (vrDbgDo(MATH_DBGLVL))
		vrPrintMatrix("vrMatrixSetFromCoords():4 Tw2rw -- matrix", &Tw2rw);

	/* step5: Calculate Trw2w from Tw2rw */
	vrMatrixInvertEuclidean(mat_dst, &Tw2rw);
	if (vrDbgDo(MATH_DBGLVL))
		vrPrintMatrix("vrMatrixSetFromCoords():5 Trw2w -- matrix", mat_dst);

//vrPrintMatrixComposition("vrMatrixSetFromCoords():6: decomp:", mat_dst); /* for debugging (eventually delete me) */
	return mat_dst;
}


/**********************************************************************/
/* Create a 4x4 matrix for rotation around one of the coordinate axes */
vrMatrix *vrMatrixSetRotationId(vrMatrix *mat_dst, int axis, double theta)
{
	double	cost = cos(vrDegreesToRadians(theta));
	double	sint = sin(vrDegreesToRadians(theta));

	vrMatrixSetIdentity(mat_dst);

	switch (axis) {					/* OpenGL doc ordering*/
	case VR_X:
		VRMAT_ROWCOL(mat_dst, VR_Y, VR_Y) = cost;	/* |  1  0  0  0  | */
		VRMAT_ROWCOL(mat_dst, VR_Y, VR_Z) = -sint;	/* |  0  c -s  0  | */
		VRMAT_ROWCOL(mat_dst, VR_Z, VR_Y) = sint;	/* |  0  s  c  0  | */
		VRMAT_ROWCOL(mat_dst, VR_Z, VR_Z) = cost;	/* |  0  0  0  1  | */
		break;
	case VR_Y:
		VRMAT_ROWCOL(mat_dst, VR_X, VR_X) = cost;	/* |  c  0  s  0  | */
		VRMAT_ROWCOL(mat_dst, VR_Z, VR_X) = -sint;	/* |  0  1  0  0  | */
		VRMAT_ROWCOL(mat_dst, VR_X, VR_Z) = sint;	/* | -s  0  c  0  | */
		VRMAT_ROWCOL(mat_dst, VR_Z, VR_Z) = cost;	/* |  0  0  0  1  | */
		break;
	case VR_Z:
		VRMAT_ROWCOL(mat_dst, VR_X, VR_X) = cost;	/* |  c -s  0  0  | */
		VRMAT_ROWCOL(mat_dst, VR_X, VR_Y) = -sint;	/* |  s  c  0  0  | */
		VRMAT_ROWCOL(mat_dst, VR_Y, VR_X) = sint;	/* |  0  0  1  0  | */
		VRMAT_ROWCOL(mat_dst, VR_Y, VR_Y) = cost;	/* |  0  0  0  1  | */
		break;
	default:
		vrErr("unknown axis specified");
		break;
	}
	return mat_dst;
}


/*********************************************************************/
/* Create a 4x4 matrix for rotation around an arbitrary axis         */
vrMatrix *vrMatrixSetRotation4d(vrMatrix *mat_dst, double x, double y, double z, double theta)
{
	vrQuat	quat_tmp;

	vrQuatSet4dThetaAxis(&quat_tmp, theta, x, y, z);
	mat_dst = vrMatrixSetFromQuat(mat_dst, &quat_tmp);

	return mat_dst;
}


/*********************************************************************/
/* Create a 4x4 matrix for rotation around an arbitrary axis         */
/* NOTE: order of the array is X, Y, Z, theta (in degrees)           */
vrMatrix *vrMatrixSetRotationAd(vrMatrix *mat_dst, double *array)
{
	vrQuat	quat_tmp;

	vrQuatSet4dThetaAxis(&quat_tmp, array[VR_W], array[VR_X], array[VR_Y], array[VR_Z]);
	mat_dst = vrMatrixSetFromQuat(mat_dst, &quat_tmp);
#if 0 /* for debugging */
	vrPrintMatrix("vrMatrixSetRotationAd() -- matrix", mat_dst);
#endif

	return mat_dst;
}


/*********************************************************************/
/* Set an existing 4x4 matrix to the scaling transformation          */
vrMatrix *vrMatrixSetScale3d(vrMatrix *mat_dst, double sx, double sy, double sz)
{
	memset(mat_dst, 0, sizeof(*mat_dst));
	VRMAT_ROWCOL(mat_dst, VR_X, VR_X) = sx;
	VRMAT_ROWCOL(mat_dst, VR_Y, VR_Y) = sy;
	VRMAT_ROWCOL(mat_dst, VR_Z, VR_Z) = sz;
	VRMAT_ROWCOL(mat_dst, VR_W, VR_W) = 1.0f;

	return mat_dst;
}


/****************************************************************/
/* Multiply two 4x4 matrices (mat_dst = mat_l * mat_r).         */
/*   When working in the OpenGL documentation convention, then  */
/*   consider the operation to be:                              */
/*       mat_dst * point = mat_l * mat_r * point                */
/*                                                              */
/*   When working in the C-style matrix convention, then:       */
/*       point * mat_dst = point * mat_r * mat_l                */
vrMatrix *vrMatrixProduct(vrMatrix *mat_dst, vrMatrix *mat_l, vrMatrix *mat_r)
{
	int		count;
	int		row;
	int		col;
	double		sum;

	for (row = 0; row < VRMAT_MAXDIM; row++) {
		for (col = 0; col < VRMAT_MAXDIM; col++) {
			sum = 0.0f;
			for (count = 0; count < VRMAT_MAXDIM; count++)
				sum += VRMAT_ROWCOL(mat_l, row, count) * VRMAT_ROWCOL(mat_r, count, col);
			VRMAT_ROWCOL(mat_dst, row, col) = sum;
		}
	}

	return mat_dst;
}


/*********************************************************************/
/* Transpose the given (src) matrix, putting the result in the (dst) */
/*   destination matrix.  A pointer to the resultant dst matrix is   */
/*   also returned as an argument.  It is safe to pass the same      */
/*   matrix-pointer as both the source and destination matrices, in  */
/*   which case that matrix is transposed in place.                  */
/*      Eg.  t = vrMatrixTranspose(m, m);                            */
vrMatrix *vrMatrixTranspose(vrMatrix *mat_dst, vrMatrix *mat_src)
{
	int	row;
	int	col;
	double	tmp;

	for (row = 0; row < VRMAT_MAXDIM-1; row++) {
		for (col = row + 1; col < VRMAT_MAXDIM; col++) {
			tmp = VRMAT_ROWCOL(mat_dst, row, col);
			VRMAT_ROWCOL(mat_dst, row, col) = VRMAT_ROWCOL(mat_src, col, row);
			VRMAT_ROWCOL(mat_src, col, row) = tmp;
		}
	}

	return mat_dst;
}


/************************************************************/
double vrMatrixGetScale(vrMatrix *mat_src)
{
	vrVector	*row1;		/* row-1 of the 3x3 rotation matrix */
	vrVector	*row2;		/* row-2 of the 3x3 rotation matrix */
	vrVector	*row3;		/* row-3 of the 3x3 rotation matrix */
	double		scale_sq1;	/* the scales of row-1 of the rotation portion */
	double		scale_sq2;	/* the scales of row-2 of the rotation portion */
	double		scale_sq3;	/* the scales of row-3 of the rotation portion */

	row1 = (vrVector *)&mat_src->v[0];	/* NOTE: these are actually columns of the OpenGL style */
	row2 = (vrVector *)&mat_src->v[4];	/*   matrix, which makes them rows of the transpose.    */
	row3 = (vrVector *)&mat_src->v[8];

	scale_sq1 = row1->v[0]*row1->v[0] + row1->v[1]*row1->v[1] + row1->v[2]*row1->v[2];
	scale_sq2 = row2->v[0]*row2->v[0] + row2->v[1]*row2->v[1] + row2->v[2]*row2->v[2];
	scale_sq3 = row3->v[0]*row3->v[0] + row3->v[1]*row3->v[1] + row3->v[2]*row3->v[2];

	/* check whether all the scale factors are the same (within a reasonable epsilon) */
	if ((fabs(scale_sq1 - scale_sq2) > 10000*VR_EPSILON) || (fabs(scale_sq1 - scale_sq3) > 10000*VR_EPSILON)) {
		/* if they are not the same, print an error and return the identity value */
		vrErrPrintf("vrMatrixGetScale(): " ANSI_TEXTRED "Warning: an asymmetrically scaled matrix! (%.9f %.9f %.9f)\n" NORM_TEXT, scale_sq1, scale_sq2, scale_sq3);
		return (1.0);		/* NOTE: 1.0 is the identity operation for scaling */
	}

	return (sqrt(scale_sq1));
}


/************************************************************/
vrMatrix *vrMatrixSetScale(vrMatrix *dst, vrMatrix *src, double scale)
{
	vrMatrix	tmpmat;		/* memory holder for src matrix to allow same value for dst and src */
	vrVector	*row1;		/* row-1 of the 3x3 rotation matrix */
	vrVector	*row2;		/* row-2 of the 3x3 rotation matrix */
	vrVector	*row3;		/* row-3 of the 3x3 rotation matrix */
	double		scale_sq;	/* the square of the given scale */

	if(src == dst) {	/* if modifying in-place: */
		tmpmat = *src;	/*   copy source data and */
		src = &tmpmat;	/*   change src arg to point to local copy */
	} else {
		*dst = *src;	/* start by copying the source matrix to the destination */
	}

	scale_sq = scale * scale;

	row1 = (vrVector *)&dst->v[0];	/* NOTE: these are actually columns of the OpenGL style */
	row2 = (vrVector *)&dst->v[4];	/*   matrix, which makes them rows of the transpose.    */
	row3 = (vrVector *)&dst->v[8];

	/* now rescale all the rows to have the average scale */
	vrVectorScale(row1, sqrt(scale_sq / vrVectorDotProduct(row1, row1)));
	vrVectorScale(row2, sqrt(scale_sq / vrVectorDotProduct(row2, row2)));
	vrVectorScale(row3, sqrt(scale_sq / vrVectorDotProduct(row3, row3)));

	return (dst);
}


/************************************************************/
/* Report whether the given matrix conforms to the rules of */
/*   a Euclidean matrix.  A Euclidean transformation matrix */
/*   must consist only of rotations, uniform scales, and    */
/*   translations.                                          */
int vrMatrixIsEuclidean(vrMatrix *mat_src)
{
	vrVector	*row1;		/* row-1 of the 3x3 rotation matrix */
	vrVector	*row2;		/* row-2 of the 3x3 rotation matrix */
	vrVector	*row3;		/* row-3 of the 3x3 rotation matrix */
	double		scale_sq1;	/* the scales of row-1 of the rotation portion */
	double		scale_sq2;	/* the scales of row-2 of the rotation portion */
	double		scale_sq3;	/* the scales of row-3 of the rotation portion */
	int		result = 1;

	row1 = (vrVector *)&mat_src->v[0];	/* NOTE: these are actually columns of the OpenGL style */
	row2 = (vrVector *)&mat_src->v[4];	/*   matrix, which makes them rows of the transpose.    */
	row3 = (vrVector *)&mat_src->v[8];

	scale_sq1 = row1->v[0]*row1->v[0] + row1->v[1]*row1->v[1] + row1->v[2]*row1->v[2];
	scale_sq2 = row2->v[0]*row2->v[0] + row2->v[1]*row2->v[1] + row2->v[2]*row2->v[2];
	scale_sq3 = row3->v[0]*row3->v[0] + row3->v[1]*row3->v[1] + row3->v[2]*row3->v[2];

	/* NOTE: these operations seem to be below the desired EPSILON value so need to soften the requirement */
	if ((fabs(scale_sq1 - scale_sq2) > 1000000*VR_EPSILON) || (fabs(scale_sq1 - scale_sq3) > 1000000*VR_EPSILON)) {
		vrErrPrintf("vrMatrixIsEuclidean(): " ANSI_TEXTRED "Warning: an asymmetrically scaled matrix is not Euclidean!  Scales: %.18lf, %.18lf, %.18lf\n" NORM_TEXT, scale_sq1, scale_sq2, scale_sq3);
		result = 0;
	}

	if (fabs(vrVectorDotProduct(row1, row2)) > 1000000*VR_EPSILON) {
		vrErrPrintf("vrMatrixIsEuclidean(): " ANSI_TEXTRED "Warning: Row 1 and 2 of matrix do not give a 0 dot product as expected from a Euclidean matrix! (%.18lf)\n" NORM_TEXT, vrVectorDotProduct(row1, row2));
		result = 0;
	}

	if (fabs(vrVectorDotProduct(row1, row3)) > 1000000*VR_EPSILON) {
		vrErrPrintf("vrMatrixIsEuclidean(): " ANSI_TEXTRED "Warning: Row 1 and 3 of matrix do not give a 0 dot product as expected from a Euclidean matrix! (%.18lf)\n" NORM_TEXT, vrVectorDotProduct(row1, row3));
		result = 0;
	}

	if (fabs(vrVectorDotProduct(row2, row3)) > 1000000*VR_EPSILON) {
		vrErrPrintf("vrMatrixIsEuclidean(): " ANSI_TEXTRED "Warning: Row 2 and 3 of matrix do not give a 0 dot product as expected from a Euclidean matrix! (%.18lf)\n" NORM_TEXT, vrVectorDotProduct(row2, row3));
		result = 0;
	}

	return result;
}


/**************************************************************/
/* Transform the given matrix into one that conforms to the   */
/*   rules of a Euclidean matrix.  A Euclidean transformation */
/*   matrix must consist only of rotations, uniform scales,   */
/*   and translations.                                        */
/* We will use the algorithm expressed by Stuart in an email. */
vrMatrix *vrMatrixMakeEuclidean(vrMatrix *dst, vrMatrix *src)
{
	vrMatrix	tmpmat;		/* memory holder for src matrix to allow same value for dst and src */
	vrVector	tmpvec;		/* memory holder for copying a vector */
	vrVector	*row1;		/* row-1 of the 3x3 rotation matrix */
	vrVector	*row2;		/* row-2 of the 3x3 rotation matrix */
	vrVector	*row3;		/* row-3 of the 3x3 rotation matrix */
	double		scale_sq1;	/* the scales of row-1 of the rotation portion */
	double		scale_sq2;	/* the scales of row-2 of the rotation portion */
	double		scale_sq3;	/* the scales of row-3 of the rotation portion */
	double		avg_scale_sq;	/* square of the average scaling factor */
	double		row1dot1;	/* dot product of row-1 & itself */
	double		row1dot2;	/* dot product of row-1 & row-2 */
	double		row1dot3;	/* dot product of row-1 & row-3 */
	double		row2dot2;	/* dot product of row-2 & row-2 (after first normalization) */
	double		row2dot3;	/* dot product of row-2 & row-3 (after first normalization) */

	if(src == dst) {	/* if modifying in-place: */
		tmpmat = *src;	/*   copy source data and */
		src = &tmpmat;	/*   change src arg to point to local copy */
	} else {
		*dst = *src;	/* start by copying the source matrix to the destination */
	}

	/* assign the three row vectors to their location in the 4x4 matrix */
	row1 = (vrVector *)&dst->v[0];	/* NOTE: these are actually columns of the OpenGL style */
	row2 = (vrVector *)&dst->v[4];	/*   matrix, which makes them rows of the transpose.    */
	row3 = (vrVector *)&dst->v[8];

	/* calculate the squared scale values of each row and their average */
	scale_sq1 = row1->v[0]*row1->v[0] + row1->v[1]*row1->v[1] + row1->v[2]*row1->v[2];
	scale_sq2 = row2->v[0]*row2->v[0] + row2->v[1]*row2->v[1] + row2->v[2]*row2->v[2];
	scale_sq3 = row3->v[0]*row3->v[0] + row3->v[1]*row3->v[1] + row3->v[2]*row3->v[2];
	avg_scale_sq = (scale_sq1 + scale_sq2 + scale_sq3) / 3.0;

	/* calculate the dot product of row-1 with each row */
	row1dot1 = vrVectorDotProduct(row1, row1);
	row1dot2 = vrVectorDotProduct(row1, row2);
	row1dot3 = vrVectorDotProduct(row1, row3);

	/* assume row-1 is pointing in the correct direction, */
	/*   and subtract its component from row-2 and row-3. */
	vrVectorSubtract(row2, row2, vrVectorScale(vrVectorCopy(&tmpvec, row1), (row1dot2 / row1dot1)));
	vrVectorSubtract(row3, row3, vrVectorScale(vrVectorCopy(&tmpvec, row1), (row1dot3 / row1dot1)));

	/* now calculate the dot products of row-2 with itself and row-3 */
	/*   and then normalize row-3 with respect to row-2.             */
	row2dot2 = vrVectorDotProduct(row2, row2);
	row2dot3 = vrVectorDotProduct(row2, row3);
	vrVectorSubtract(row3, row3, vrVectorScale(vrVectorCopy(&tmpvec, row2), (row2dot3 / row2dot2)));

	/* now rescale all the rows to have the average scale */
	vrVectorScale(row1, sqrt(avg_scale_sq / vrVectorDotProduct(row1, row1)));
	vrVectorScale(row2, sqrt(avg_scale_sq / vrVectorDotProduct(row2, row2)));
	vrVectorScale(row3, sqrt(avg_scale_sq / vrVectorDotProduct(row3, row3)));

	/* Just to be safe, we'll check for the Euclideanness */
	if (!vrMatrixIsEuclidean(dst)) {
		vrErrPrintf("vrMatrixMakeEuclidean(): " ANSI_TEXTRED "Warning: attempt to Euclideanize the matrix failed!\n" NORM_TEXT);
	}
	return dst;
}


/***********************************************************************/
/* Calculate the inverse of a matrix that is of the Euclidean variety. */
/* Stuart: so long as the 4x4 is a Euclidean transformation (ie.       */
/*   consisting only of a combination of rotations, uniform scales and */
/*   translations, this works.  Matrix operations that cannot be       */
/*   accommodated are non-uniform scales, shears and projections.      */
/*   "Euclidean" is slightly too broad a term, geometers might call    */
/*   the class of matrices this works on: "Euclidean Similarity" is    */
/*   probably a more precise term.                                     */
vrMatrix *vrMatrixInvertEuclidean(vrMatrix *dst, vrMatrix *src)
{
	vrMatrix	tmp;		/* memory holder for src matrix to allow same value for dst and src */
	int		col, row;
	vrVector	trans;		/* the translation part of the matrix */
	vrVector	old_trans;	/* the translation part of the matrix */
	double		inv_scale_factor;/* reciprocal of the scale part of the matrix */

	if(src == dst) {	/* if inverting in-place: */
		tmp = *src;	/*   copy source data and */
		src = &tmp;	/*   change src arg to point to local copy */
	}

	/* Measure square-of-scaling from the first row.  If this matrix is */
	/*   actually a uniform scaling (plus rotation plus translation, as */
	/*   described above), then any row or column should give the same  */
	/*   result.                                                        */
	inv_scale_factor = 1.0 / (src->v[0]*src->v[0] + src->v[1]*src->v[1] + src->v[2]*src->v[2]);

	/* transpose 3x3 rotation sub matrix into a 4x4 matrix */
	for (col = 0; col < 3; col++) {
		for (row = 0; row < 3; row++) {
			VRMAT_ROWCOL(dst, col, row) = VRMAT_ROWCOL(src, row, col) * inv_scale_factor;
		}
		VRMAT_ROWCOL(dst, VR_W, col) = 0.0;
	}

	/* now calculate the new translation portion of the inversion */
	vrVectorGetTransFromMatrix(&old_trans, src);
	vrVectorTransformByMatrix(&trans, &old_trans, dst);	/* NOTE: this operation does not use the 4th row of the 4x4 matrix (which is not presently set) */
	vrVectorScale(&trans, -1.0);
	vrMatrixSetTransFromVector(dst, &trans);

	VRMAT_ROWCOL(dst, VR_W, VR_W) = 1.0;			/* NOTE: we haven't yet set this element of the matrix */

	return dst;
}


/********************************************************************/
double vr2x2Determinant(double a, double b, double c, double d)
{
	return (a * d - b * c);
}


/**************************************************************************/
/* Calculate the determinant of a 3x3 matrix.                             */
/*     | a1,  b1,  c1 |                                                   */
/*     | a2,  b2,  c2 |                                                   */
/*     | a3,  b3,  c3 |                                                   */
double vr3x3Determinant(double a1, double a2, double a3, double b1, double b2, double b3, double c1, double c2, double c3)
{
	double	result;

	result = a1 * vr2x2Determinant(b2, b3, c2, c3)
		- b1 * vr2x2Determinant(a2, a3, c2, c3)
		+ c1 * vr2x2Determinant(a2, a3, b2, b3);

	return result;
}


/*****************************************************************************/
double vrMatrixDeterminant(vrMatrix *mat_src)
{
	double	exx, eyx, ezx, ewx;	/* a1 - a4 */
	double	exy, eyy, ezy, ewy;	/* b1 - b4 */
	double	exz, eyz, ezz, ewz;	/* c1 - c4 */
	double	exw, eyw, ezw, eww;	/* d1 - d4 */
	double	result;

	exx = VRMAT_ROWCOL(mat_src, VR_X, VR_X);	/* a1 */
	eyx = VRMAT_ROWCOL(mat_src, VR_Y, VR_X);	/* a2 */
	ezx = VRMAT_ROWCOL(mat_src, VR_Z, VR_X);	/* a3 */
	ewx = VRMAT_ROWCOL(mat_src, VR_W, VR_X);	/* a4 */

	exy = VRMAT_ROWCOL(mat_src, VR_X, VR_Y);	/* b1 */
	eyy = VRMAT_ROWCOL(mat_src, VR_Y, VR_Y);	/* b2 */
	ezy = VRMAT_ROWCOL(mat_src, VR_Z, VR_Y);	/* b3 */
	ewy = VRMAT_ROWCOL(mat_src, VR_W, VR_Y);	/* b4 */

	exz = VRMAT_ROWCOL(mat_src, VR_X, VR_Z);	/* c1 */
	eyz = VRMAT_ROWCOL(mat_src, VR_Y, VR_Z);	/* c2 */
	ezz = VRMAT_ROWCOL(mat_src, VR_Z, VR_Z);	/* c3 */
	ewz = VRMAT_ROWCOL(mat_src, VR_W, VR_Z);	/* c4 */

	exw = VRMAT_ROWCOL(mat_src, VR_X, VR_W);	/* d1 */
	eyw = VRMAT_ROWCOL(mat_src, VR_Y, VR_W);	/* d2 */
	ezw = VRMAT_ROWCOL(mat_src, VR_Z, VR_W);	/* d3 */
	eww = VRMAT_ROWCOL(mat_src, VR_W, VR_W);	/* d4 */

	result = exx * vr3x3Determinant(eyy, ezy, ewy, eyz, ezz, ewz, eyw, ezw, eww)
		- exy * vr3x3Determinant(eyx, ezx, ewx, eyz, ezz, ewz, eyw, ezw, eww)
		+ exz * vr3x3Determinant(eyx, ezx, ewx, eyy, ezy, ewy, eyw, ezw, eww)
		- exw * vr3x3Determinant(eyx, ezx, ewx, eyy, ezy, ewy, eyz, ezz, ewz);

	return result;
}



/**************************************/
/****** Matrix Pre-Mult Routines ******/
/**************************************/


/******************************************************/
/* Compute the matrix product, mat_l' = mat_r * mat_l */
/*   where "pre" is in C-style.                       */
/*   ie.  point * mat_l' = point * mat_r * mat_l      */
vrMatrix *vrMatrixPreMult(vrMatrix *mat_l, vrMatrix *mat_r)
{
	vrMatrix	tmpmat;		/* memory for the rotation matrix */
	vrMatrix	*product = vrMatrixProduct(&tmpmat, mat_r, mat_l);

	vrMatrixCopy(mat_l, product);

	return (mat_l);
}


/*************************************************************************/
/* Translate the given matrix by composing it with a translation matrix. */
/*   ie, mat_l = transmat * mat_l                                        */
vrMatrix *vrMatrixPreTranslate3d(vrMatrix *mat_l, double x, double y, double z)
{
	vrMatrix	tmpmat;		/* memory for the rotation matrix */
	vrMatrix	*tmat = vrMatrixSetTranslation3d(&tmpmat, x, y, z);

	vrMatrixPreMult(mat_l, tmat);

	return (mat_l);
}


/*************************************************************************/
/* Translate the given matrix by composing it with a translation matrix. */
/*   ie, mat_l = transmat * mat_l                                        */
vrMatrix *vrMatrixPreTranslateAd(vrMatrix *mat_l, double *array)
{
	vrMatrix	tmpmat;		/* memory for the rotation matrix */
	vrMatrix	*tmat = vrMatrixSetTranslationAd(&tmpmat, array);

	vrMatrixPreMult(mat_l, tmat);

	return (mat_l);
}


/*******************************************************************/
/* Rotate the given matrix by composing it with a rotation matrix. */
/*    ie, mat_l = rotmat * mat_l                                   */
vrMatrix *vrMatrixPreRotateId(vrMatrix *mat_l, int axis, double theta)
{
	vrMatrix	tmpmat;		/* memory for the rotation matrix */
	vrMatrix	*rotmat = vrMatrixSetRotationId(&tmpmat, axis, theta);

	vrMatrixPreMult(mat_l, rotmat);

	return (mat_l);
}


/**********************************************************************/
/* Rotate the given matrix by composing it with a rotation matrix.    */
/*    ie, mat_l = m2o * rotmat * mat_l * m2o-1                        */
/*     (or visa versa on the translation to and from the origin)      */
/* NOTE: This is now implemented more efficiently by simply doing the */
/*   m2o and m2o-1 transformations by simply clearing and restoring   */
/*   restoring the translation portion of the matrix, with functions: */
/*   vrMatrixGetResetTranslationAd()/vrMatrixSetTranslationOnlyAd().  */
vrMatrix *vrMatrixPreRotateAboutOriginId(vrMatrix *mat_l, int axis, double theta)
{
	vrMatrix	tmpmat;		/* memory for the rotation matrix */
	vrMatrix	*rotmat = vrMatrixSetRotationId(&tmpmat, axis, theta);
	vrVector	trans_vec;

	/* m2o: move matrix to origin and store offset */
	vrMatrixGetResetTranslationAd(mat_l, trans_vec.v);

	vrMatrixPreMult(mat_l, rotmat);

	/* m2o-1: return to original translation */
	vrMatrixSetTranslationOnlyAd(mat_l, trans_vec.v);

	return (mat_l);
}


/**********************************************************************/
/* Rotate the given matrix by composing it with a rotation matrix.    */
/*    ie, mat_l = m2p * rotmat * mat_l * m2p-1                        */
/*     (or visa versa on the translation to and from the origin)      */
/* NOTE: This is now implemented more efficiently by simply doing the */
/*   m2o and m2o-1 transformations by simply clearing and restoring   */
/*   restoring the translation portion of the matrix, with functions: */
/*   vrMatrixGetResetTranslationAd()/vrMatrixSetTranslationOnlyAd().  */
vrMatrix *vrMatrixPreRotateAboutPointId(vrMatrix *mat_l, vrPoint *point, int axis, double theta)
{
	vrMatrix	tmpmat;		/* memory for the rotation matrix */
	vrMatrix	*rotmat = vrMatrixSetRotationId(&tmpmat, axis, theta);
	vrVector	trans_vec;

	/* move matrix to origin and store offset */
	vrMatrixGetResetTranslationAd(mat_l, trans_vec.v);

	/* now jump to the point to rotate about */
	vrMatrixSetTranslationOnlyAd(mat_l, point->v);

	vrMatrixPreMult(mat_l, rotmat);

	/* return to original translation */
	vrMatrixSetTranslationOnlyAd(mat_l, trans_vec.v);

	return (mat_l);
}


/*******************************************************************/
/* Rotate the given matrix by composing it with a rotation matrix. */
/*    ie, mat_l = rotmat * mat_l                                   */
vrMatrix *vrMatrixPreRotate4d(vrMatrix *mat_l, double x, double y, double z, double theta)
{
	vrMatrix	tmpmat;
	vrMatrix	*rotmat = vrMatrixSetRotation4d(&tmpmat, x, y, z, theta);

	vrMatrixPreMult(mat_l, rotmat);

	return (mat_l);
}


/******************************************************************/
/* Scale the given matrix by composing it with a scaling matrix.  */
/*     ie, mat_l = scalemat * mat_l                               */
vrMatrix *vrMatrixPreScale3d(vrMatrix *mat_l, double sx, double sy, double sz)
{
	vrMatrix	tmpmat;
	vrMatrix	*smat = vrMatrixSetScale3d(&tmpmat, sx, sy, sz);

	vrMatrixPreMult(mat_l, smat);

	return (mat_l);
}

/***************************************/
/****** Matrix Post-Mult Routines ******/
/***************************************/


/********************************************************/
/* Compute the matrix product, mat_l' = mat_l * mat_r   */
/*   ie. point * mat_l' = point * mat_l * mat_r         */
vrMatrix *vrMatrixPostMult(vrMatrix *mat_l, vrMatrix *mat_r)
{
	vrMatrix	tmpmat;		/* memory for the rotation matrix */
	vrMatrix	*product = vrMatrixProduct(&tmpmat, mat_l, mat_r);

	vrMatrixCopy(mat_l, product);

	return (mat_l);
}


/*************************************************************************/
/* Translate the given matrix by composing it with a translation matrix. */
/*     ie, mat_l = mat_l * transmat                                      */
vrMatrix *vrMatrixPostTranslate3d(vrMatrix *mat_l, double x, double y, double z)
{
	vrMatrix	tmpmat;		/* memory for the rotation matrix */
	vrMatrix	*tmat = vrMatrixSetTranslation3d(&tmpmat, x, y, z);

	vrMatrixPostMult(mat_l, tmat);

	return (mat_l);
}


/*************************************************************************/
/* Translate the given matrix by composing it with a translation matrix. */
/*     ie, mat_l = mat_l * transmat                                      */
vrMatrix *vrMatrixPostTranslateAd(vrMatrix *mat_l, double *array)
{
	vrMatrix	tmpmat;		/* memory for the rotation matrix */
	vrMatrix	*transmat = vrMatrixSetTranslationAd(&tmpmat, array);

	vrMatrixPostMult(mat_l, transmat);

	return (mat_l);
}


/*******************************************************************/
/* Rotate the given matrix by composing it with a rotation matrix. */
/* ie, mat_l = mat_l * rotmat                                      */
vrMatrix *vrMatrixPostRotateId(vrMatrix *mat_l, int axis, double theta)
{
	vrMatrix	tmpmat;			/* memory for the rotation matrix */
	vrMatrix	*rotmat = vrMatrixSetRotationId(&tmpmat, axis, theta);

	vrMatrixPostMult(mat_l, rotmat);

	return (mat_l);
}


/*******************************************************************/
/* Rotate the given matrix by composing it with a rotation matrix. */
/*    ie, mat_l = m2o * mat_l * rotmat * m2o-1                     */
/*     (or visa versa on the translation to and from the origin)   */
vrMatrix *vrMatrixPostRotateAboutOriginId(vrMatrix *mat_l, int axis, double theta)
{
	vrMatrix	tmpmat;			/* memory for the rotation matrix */
	vrMatrix	*rotmat = vrMatrixSetRotationId(&tmpmat, axis, theta);
	vrVector	trans_vec;

	/* move matrix to origin and store offset */
	vrMatrixGetResetTranslationAd(mat_l, trans_vec.v);

	vrMatrixPostMult(mat_l, rotmat);

	/* return to original translation */
	vrMatrixSetTranslationOnlyAd(mat_l, trans_vec.v);

	return (mat_l);
}


/*******************************************************************/
/* Rotate the given matrix by composing it with a rotation matrix. */
/*    ie, mat_l = m2p * mat_l * rotmat * m2p-1                     */
/*     (or visa versa on the translation to and from the origin)   */
vrMatrix *vrMatrixPostRotateAboutPointId(vrMatrix *mat_l, vrPoint *point, int axis, double theta)
{
	vrMatrix	tmpmat;			/* memory for the rotation matrix */
	vrMatrix	*rotmat = vrMatrixSetRotationId(&tmpmat, axis, theta);
	vrVector	trans_vec;

	/* move matrix to origin and store offset */
	vrMatrixGetResetTranslationAd(mat_l, trans_vec.v);

	/* now jump to the point to rotate about */
	vrMatrixSetTranslationOnlyAd(mat_l, point->v);

	vrMatrixPostMult(mat_l, rotmat);

	/* return to original translation */
	vrMatrixSetTranslationOnlyAd(mat_l, trans_vec.v);

	return (mat_l);
}


/*******************************************************************/
/* Rotate the given matrix by composing it with a rotation matrix. */
/*     ie, mat_l = mat_l * rotmat                                  */
vrMatrix *vrMatrixPostRotate4d(vrMatrix *mat_l, double x, double y, double z, double theta)
{
	vrMatrix	tmpmat;
	vrMatrix	*rotmat = vrMatrixSetRotation4d(&tmpmat, x, y, z, theta);

	vrMatrixPostMult(mat_l, rotmat);

	return (mat_l);
}


/*****************************************************************/
/* Scale the given matrix by composing it with a scaling matrix. */
/*      ie, mat_l = mat_l * scalemat                             */
vrMatrix *vrMatrixPostScale3d(vrMatrix *mat_l, double sx, double sy, double sz)
{
	vrMatrix	tmpmat;
	vrMatrix	*smat = vrMatrixSetScale3d(&tmpmat, sx, sy, sz);

	vrMatrixPostMult(mat_l, smat);

	return (mat_l);
}



/* ==================================================== */
/* QUATERNION OPERATIONS                                */
/* ==================================================== */

/*********************************************************************/
/* Create an empty quaternion                                        */
vrQuat *vrQuatCreate(void)
{
	return (vrQuat *)vrShmemAlloc0(sizeof(vrQuat));
}


/*********************************************************************/
/* Create a quaternion representing the specified rotation           */
vrQuat *vrQuatSet4dThetaAxis(vrQuat *quat_dst, double theta, double x, double y, double z)
{
	double	theta_2 = theta / 2.0f;
	double	sint = sin(vrDegreesToRadians(theta_2));
	double	cost = cos(vrDegreesToRadians(theta_2));

	quat_dst->v[VR_X] = x * sint;
	quat_dst->v[VR_Y] = y * sint;
	quat_dst->v[VR_Z] = z * sint;
	quat_dst->v[VR_W] = cost;

	return quat_dst;
}


/*********************************************************************/
/* Create a standard quaternion from the four given values.          */
vrQuat *vrQuatSet4d(vrQuat *quat_dst, double x, double y, double z, double w)
{
	quat_dst->v[VR_X] = x;
	quat_dst->v[VR_Y] = y;
	quat_dst->v[VR_Z] = z;
	quat_dst->v[VR_W] = w;

	return quat_dst;
}


/***********************************************************************/
/* Create the product of the two given quaternions, quat_dst = ql * qr */
/*   to create a new product from quats ql & qr:                       */
/*      quat_dst = vrQuatProduct(vrQuatCreate(), ql, qr)               */
/* NOTE: quat_dst cannot point to the same quaternion as either ql or  */
/*   qr, as it will overwrite values during the computation.  Use the  */
/*   vrQuatProductEquals() function which uses a temporary holding     */
/*   variable to handle the computation.                               */
vrQuat *vrQuatProduct(vrQuat *quat_dst, vrQuat *ql, vrQuat *qr)
{
	quat_dst->v[VR_X] = (ql->v[VR_W]*qr->v[VR_X]) + (ql->v[VR_X]*qr->v[VR_W]) +
			    (ql->v[VR_Y]*qr->v[VR_Z]) - (ql->v[VR_Z]*qr->v[VR_Y]);
	quat_dst->v[VR_Y] = (ql->v[VR_W]*qr->v[VR_Y]) + (ql->v[VR_Y]*qr->v[VR_W]) +
			    (ql->v[VR_Z]*qr->v[VR_X]) - (ql->v[VR_X]*qr->v[VR_Z]);
	quat_dst->v[VR_Z] = (ql->v[VR_W]*qr->v[VR_Z]) + (ql->v[VR_Z]*qr->v[VR_W]) +
			    (ql->v[VR_X]*qr->v[VR_Y]) - (ql->v[VR_Y]*qr->v[VR_X]);
	quat_dst->v[VR_W] = (ql->v[VR_W]*qr->v[VR_W]) - (ql->v[VR_X]*qr->v[VR_X]) -
			    (ql->v[VR_Y]*qr->v[VR_Y]) - (ql->v[VR_Z]*qr->v[VR_Z]);

	return quat_dst;
}


/*********************************************************************/
/* Create the product of the two given quaternions, ql *= qr         */
vrQuat *vrQuatProductEquals(vrQuat *tmpquat, vrQuat *ql, vrQuat *qr)
{
	vrQuatCopy(ql, vrQuatProduct(tmpquat, ql, qr));

	return ql;
}


/**************************************************************************************/
/* Convert the given quaternion from Y-UP coordinate system to Z-UP: ie, ql = Y2Z(qr) */
vrQuat *vrQuatConvertYUPtoZUP(vrQuat *quat_dst, vrQuat *quat_src)
{
	vrQuat	swap;	/* used to make it safe to give the same value for both arguments */

	swap.v[VR_X] = quat_src->v[VR_X];
	swap.v[VR_Y] = quat_src->v[VR_Y];
	swap.v[VR_Z] = quat_src->v[VR_Z];
	swap.v[VR_W] = quat_src->v[VR_W];

	quat_dst->v[VR_X] = -swap.v[VR_X];
	quat_dst->v[VR_Y] =  swap.v[VR_Z];
	quat_dst->v[VR_Z] = -swap.v[VR_Y];
	quat_dst->v[VR_W] =  swap.v[VR_W];

	return (quat_dst);
}


/**************************************************************************************/
/* Convert the given quaternion from Y-UP coordinate system to Z-UP: ie, ql = Z2Y(qr) */
vrQuat *vrQuatConvertZUPtoYUP(vrQuat *quat_dst, vrQuat *quat_src)
{
	vrQuat	swap;	/* used to make it safe to give the same value for both arguments */

	swap.v[VR_X] = quat_src->v[VR_X];
	swap.v[VR_Y] = quat_src->v[VR_Y];
	swap.v[VR_Z] = quat_src->v[VR_Z];
	swap.v[VR_W] = quat_src->v[VR_W];

	quat_dst->v[VR_X] = -swap.v[VR_X];
	quat_dst->v[VR_Y] = -swap.v[VR_Z];
	quat_dst->v[VR_Z] =  swap.v[VR_Y];
	quat_dst->v[VR_W] =  swap.v[VR_W];

	return (quat_dst);
}


/*********************************************************************/
/* copy the given quaternion: ie, quat_dst = quat_src                */
vrQuat *vrQuatCopy(vrQuat *quat_dst, vrQuat *quat_src)
{
	return ((vrQuat *)memcpy(quat_dst, quat_src, sizeof(*quat_dst)));
}


/*********************************************************************/
/* delete the given quaternion                                       */
void vrQuatDelete(vrQuat *quat)
{
	vrShmemFree(quat);
}



/* ==================================================== */
/* VECTOR OPERATIONS                                    */
/* ==================================================== */

/*********************************************************************/
vrVector *vrVectorCopy(vrVector *vec_dst, vrVector *vec_src)
{
	return ((vrVector *)memcpy(vec_dst, vec_src, sizeof(*vec_dst)));
}


/*********************************************************************/
/* calculate the vector from point-1 to point-2                      */
vrVector *vrVectorFromTwoPoints(vrVector *vec_dst, vrPoint *pnt1, vrPoint *pnt2)
{
	vec_dst->v[VR_X] = pnt2->v[VR_X] - pnt1->v[VR_X];
	vec_dst->v[VR_Y] = pnt2->v[VR_Y] - pnt1->v[VR_Y];
	vec_dst->v[VR_Z] = pnt2->v[VR_Z] - pnt1->v[VR_Z];

	return vec_dst;
}


/*********************************************************************/
/* NOTE: vec_dst and vec_src can point to the same point.                    */
vrVector *vrVectorNormalize(vrVector *vec_dst, vrVector *vec_src)
{
	double	magnitude;

	vrDbgPrintfN(MATH_DBGLVL, "vrVectorNormalize(): from: (%.2f %.2f %.2f)\n", vec_src->v[VR_X], vec_src->v[VR_Y], vec_src->v[VR_Z]);
	magnitude = sqrt(vec_src->v[VR_X]*vec_src->v[VR_X] + vec_src->v[VR_Y]*vec_src->v[VR_Y] + vec_src->v[VR_Z]*vec_src->v[VR_Z]);

	vec_dst->v[VR_X] = vec_src->v[VR_X] / magnitude;
	vec_dst->v[VR_Y] = vec_src->v[VR_Y] / magnitude;
	vec_dst->v[VR_Z] = vec_src->v[VR_Z] / magnitude;
	vrDbgPrintfN(MATH_DBGLVL, "vrVectorNormalize(): to: Mag = %.2f (%.2f %.2f %.2f)\n", magnitude, vec_dst->v[VR_X], vec_dst->v[VR_Y], vec_dst->v[VR_Z]);

	return vec_dst;
}


/*********************************************************************/
double vrVectorLength(vrVector *vec)
{
	double	magnitude;

	magnitude = sqrt(vec->v[VR_X]*vec->v[VR_X] + vec->v[VR_Y]*vec->v[VR_Y] + vec->v[VR_Z]*vec->v[VR_Z]);

	return magnitude;
}


/*********************************************************************/
double vrVectorDotProduct(vrVector *vec1, vrVector *vec2)
{
	double	result;

	result  = vec1->v[VR_X] * vec2->v[VR_X];
	result += vec1->v[VR_Y] * vec2->v[VR_Y];
	result += vec1->v[VR_Z] * vec2->v[VR_Z];

	return result;
}


/********************************************************************/
/* NOTE: unlike many of the other vector operations, it is NOT SAFE */
/*   to use the same memory pointer for the result (vec_dst) and    */
/*   either of the operands.  We may want to consider fixing this.  */
vrVector *vrVectorCrossProduct(vrVector *vec_dst, vrVector *vec1, vrVector *vec2)
{
	vrVector	result;

	result.v[VR_X] = vec1->v[VR_Y]*vec2->v[VR_Z] - vec1->v[VR_Z]*vec2->v[VR_Y];
	result.v[VR_Y] = vec1->v[VR_Z]*vec2->v[VR_X] - vec1->v[VR_X]*vec2->v[VR_Z];
	result.v[VR_Z] = vec1->v[VR_X]*vec2->v[VR_Y] - vec1->v[VR_Y]*vec2->v[VR_X];

	*vec_dst = result;

	return vec_dst;
}


/*********************************************************************/
/* NOTE: this changes the vector passed as the (first) argument      */
vrVector *vrVectorScale(vrVector *vec, double scale)
{
	vec->v[VR_X] *= scale;
	vec->v[VR_Y] *= scale;
	vec->v[VR_Z] *= scale;

	return vec;
}


/*****************************************************************************************/
/* NOTE: this should be safe for vec_dst to be the same as either vec1 or vec2 (or both) */
vrVector *vrVectorSubtract(vrVector *vec_dst, vrVector *vec1, vrVector *vec2)
{
	vec_dst->v[VR_X] = vec1->v[VR_X] - vec2->v[VR_X];
	vec_dst->v[VR_Y] = vec1->v[VR_Y] - vec2->v[VR_Y];
	vec_dst->v[VR_Z] = vec1->v[VR_Z] - vec2->v[VR_Z];

	return vec_dst;
}


/**********************************************************************/
/* It is safe to pass the source and destination vectors as the same. */
/* NOTE: this is the equivalent of extending the 3-tuple vector by    */
/*   one element with a value of 0, and then multiplying the 4-tuple  */
/*   by the 4x4 matrix.                                               */
/* NOTE: this is a Vector operation, so translations have no effect!  */
vrVector *vrVectorTransformByMatrix(vrVector *vec_dst, vrVector *vec_src, vrMatrix *mat_src)
{
	int		row;
	vrVector	vec_tmp;		/* store here in case src and dst are the same */

	vec_tmp = *vec_src;
	for (row = VR_X; row <= VR_Z; row++) {
		vec_dst->v[row] = vec_tmp.v[VR_X] * VRMAT_ROWCOL(mat_src, row, VR_X) +
				  vec_tmp.v[VR_Y] * VRMAT_ROWCOL(mat_src, row, VR_Y) +
				  vec_tmp.v[VR_Z] * VRMAT_ROWCOL(mat_src, row, VR_Z);
	}

	return vec_dst;
}


/*************************************************************************************/
/* It is (currently) NOT safe to pass the source and dest as the same vector memory  */
/* NOTE: this function is (currently) only called by vrMatrixCompositionFromMatrix() */
/*   which also is using VRMAT_COLROW -- probably in an attempt to fix the latter.   */
vrVector4 *vrVector4TransformByMatrix(vrVector4 *vec_dst, vrVector4 *vec_src, vrMatrix *mat_src)
{
	vec_dst->v[VR_X] = (vec_src->v[VR_X] * VRMAT_COLROW(mat_src, VR_X, VR_X)) +
			   (vec_src->v[VR_Y] * VRMAT_COLROW(mat_src, VR_Y, VR_X)) +
			   (vec_src->v[VR_Z] * VRMAT_COLROW(mat_src, VR_Z, VR_X)) +
			   (vec_src->v[VR_W] * VRMAT_COLROW(mat_src, VR_W, VR_X));

	vec_dst->v[VR_Y] = (vec_src->v[VR_X] * VRMAT_COLROW(mat_src, VR_X, VR_Y)) +
			   (vec_src->v[VR_Y] * VRMAT_COLROW(mat_src, VR_Y, VR_Y)) +
			   (vec_src->v[VR_Z] * VRMAT_COLROW(mat_src, VR_Z, VR_Y)) +
			   (vec_src->v[VR_W] * VRMAT_COLROW(mat_src, VR_W, VR_Y));

	vec_dst->v[VR_Z] = (vec_src->v[VR_X] * VRMAT_COLROW(mat_src, VR_X, VR_Z)) +
			   (vec_src->v[VR_Y] * VRMAT_COLROW(mat_src, VR_Y, VR_Z)) +
			   (vec_src->v[VR_Z] * VRMAT_COLROW(mat_src, VR_Z, VR_Z)) +
			   (vec_src->v[VR_W] * VRMAT_COLROW(mat_src, VR_W, VR_Z));

	vec_dst->v[VR_W] = (vec_src->v[VR_X] * VRMAT_COLROW(mat_src, VR_X, VR_W)) +
			   (vec_src->v[VR_Y] * VRMAT_COLROW(mat_src, VR_Y, VR_W)) +
			   (vec_src->v[VR_Z] * VRMAT_COLROW(mat_src, VR_Z, VR_W)) +
			   (vec_src->v[VR_W] * VRMAT_COLROW(mat_src, VR_W, VR_W));


	return vec_dst;
}


/*********************************************************************/
vrVector *vrVectorAddScaledVectors(vrVector *vec_dst, vrVector *vec1, vrVector *vec2, double scale1, double scale2)
{
	vrVector	local_vec;

	vrDbgPrintfN(MATH_DBGLVL, "vrVectorAddScaledVectors(): from: (%.2f %.2f %.2f) * %.2f + (%.2f %.2f %.2f) * %.2f\n", vec1->v[VR_X], vec1->v[VR_Y], vec1->v[VR_Z], scale1, vec2->v[VR_X], vec2->v[VR_Y], vec2->v[VR_Z], scale2);
	local_vec.v[VR_X] = vec1->v[VR_X] * scale1 + vec2->v[VR_X] * scale2;
	local_vec.v[VR_Y] = vec1->v[VR_Y] * scale1 + vec2->v[VR_Y] * scale2;
	local_vec.v[VR_Z] = vec1->v[VR_Z] * scale1 + vec2->v[VR_Z] * scale2;

	*vec_dst = local_vec;
	vrDbgPrintfN(MATH_DBGLVL, "vrVectorAddScaledVectors(): to become: (%.2f %.2f %.2f)\n", vec_dst->v[VR_X], vec_dst->v[VR_Y], vec_dst->v[VR_Z]);

	return vec_dst;
}



/* ==================================================== */
/* POINT OPERATIONS                                     */
/* ==================================================== */

/*********************************************************************/
vrPoint *vrPointCreate(void)
{
	return (vrPoint *)vrShmemAlloc0(sizeof(vrPoint));
}


/*********************************************************************/
vrPoint *vrPointCreate3d(double x, double y, double z)
{
	vrPoint	*pt = vrPointCreate();

	vrPointSet3d(pt, x, y, z);

	return pt;
}


/*********************************************************************/
vrPoint *vrPointCopy(vrPoint *pnt_dst, vrPoint *pnt_src)
{
	return ((vrPoint *)memcpy(pnt_dst, pnt_src, sizeof(*pnt_dst)));
}


/*********************************************************************/
void vrPointDelete(vrPoint *pnt)
{
	vrShmemFree(pnt);
}


/*********************************************************************/
vrPoint	*vrPointAddScaledVector(vrPoint *pnt_dst, vrPoint *pnt_src, vrVector *vec_src, double scale)
{
	pnt_dst->v[VR_X] = pnt_src->v[VR_X] + vec_src->v[VR_X] * scale;
	pnt_dst->v[VR_Y] = pnt_src->v[VR_Y] + vec_src->v[VR_Y] * scale;
	pnt_dst->v[VR_Z] = pnt_src->v[VR_Z] + vec_src->v[VR_Z] * scale;

	return pnt_dst;
}


/* ==================================================== */
/* TRANSFORMATIONS AND CONVERSIONS                      */
/* ==================================================== */


#if 1 /* it's time to get rid of this mistake [03/18/2015 -- actually it's time to make this the real version!] { */
/* same as above, but using the COLROW method */
/* 07/01/2014: I suspect that I should get rid of this function.  It seems to be */ /* future delete me -- these comments */
/*   used as a means to convert data from Z-up to Y-up coordinate systems,       */
/*   though without any real understanding that that's what was going on.        */
/* 03/18/2015: And today I suspect it's the other version that we can get rid of */
/*   so I've renamed this one to have the normal name, and the other one to be   */
/*   oddly named, and now destined for deprecation.                              */
vrMatrix *vrMatrixSetRotationFromQuat(vrMatrix *mat_dst, vrQuat *quat_src)
{
	double		xs,ys,zs,wx,wy,wz,xx,xy,xz,yy,yz,zz;

#  if 0
	vrPrintQuat("vrMatrixSetRotationFromQuat() -- quat_src", quat_src);
#  endif

	xs = quat_src->v[VR_X] * 2.0;
	ys = quat_src->v[VR_Y] * 2.0;
	zs = quat_src->v[VR_Z] * 2.0;

	wx = quat_src->v[VR_W] * xs;
	wy = quat_src->v[VR_W] * ys;
	wz = quat_src->v[VR_W] * zs;
	xx = quat_src->v[VR_X] * xs;
	xy = quat_src->v[VR_X] * ys;
	xz = quat_src->v[VR_X] * zs;
	yy = quat_src->v[VR_Y] * ys;
	yz = quat_src->v[VR_Y] * zs;
	zz = quat_src->v[VR_Z] * zs;

#  if 1
	/* NOTE: we're using COLROW here instead of our usual ROWCOL, because */
	/*   the formula we're using to convert from a Quaternion to a matrix */
	/*   assumes the normal matrix ordering, which is to say, not the     */
	/*   OpenGL ordering, which is what FreeVR uses.                      */
/* 02/10/21: probably use _ROWCOL & swap the arguments */
	VRMAT_COLROW(mat_dst, VR_X, VR_X) = 1.0 - yy - zz;
	VRMAT_COLROW(mat_dst, VR_X, VR_Y) = xy + wz;
	VRMAT_COLROW(mat_dst, VR_X, VR_Z) = xz - wy;

	VRMAT_COLROW(mat_dst, VR_Y, VR_X) = xy - wz;
	VRMAT_COLROW(mat_dst, VR_Y, VR_Y) = 1.0 - xx - zz;
	VRMAT_COLROW(mat_dst, VR_Y, VR_Z) = yz + wx;

	VRMAT_COLROW(mat_dst, VR_Z, VR_X) = xz + wy;
	VRMAT_COLROW(mat_dst, VR_Z, VR_Y) = yz - wx;
	VRMAT_COLROW(mat_dst, VR_Z, VR_Z) = 1.0 - xx - yy;
#  else /* delete me -- and as mentioned, change to ROWCOL above, but flip the arguements */
	/* 05/26/2006: NOTE further (or instead), when I use the COLROW, my */
	/*   configuration is backward! (ie. I get t2r_xform & r2e_xform    */
	/*   values that don't work with the configuration I've already     */
	/*   worked out.)  So, what is the solution???                      */
	VRMAT_ROWCOL(mat_dst, VR_X, VR_X) = 1.0 - yy - zz;
	VRMAT_ROWCOL(mat_dst, VR_X, VR_Y) = xy + wz;
	VRMAT_ROWCOL(mat_dst, VR_X, VR_Z) = xz - wy;

	VRMAT_ROWCOL(mat_dst, VR_Y, VR_X) = xy - wz;
	VRMAT_ROWCOL(mat_dst, VR_Y, VR_Y) = 1.0 - xx - zz;
	VRMAT_ROWCOL(mat_dst, VR_Y, VR_Z) = yz + wx;

	VRMAT_ROWCOL(mat_dst, VR_Z, VR_X) = xz + wy;
	VRMAT_ROWCOL(mat_dst, VR_Z, VR_Y) = yz - wx;
	VRMAT_ROWCOL(mat_dst, VR_Z, VR_Z) = 1.0 - xx - yy;
#  endif

#  if 0 /* Option: extra debug output */
	vrPrintMatrix("vrMatrixSetRotationFromQuat() -- matrix", mat_dst);
#  endif
	return mat_dst;
}
#endif /* } */


/*********************************************************************/
/* 01/14/2021: By default, we use the Column-Row ordering.           */
vrMatrix *vrMatrixSetFromQuat(vrMatrix *mat_dst, vrQuat *quat_src)
{
#if 0 /* future delete me [ */
	vrMatrixSetFromQuatCR(mat_dst, quat_src);
	return mat_dst;
}

/*********************************************************************/
vrMatrix *vrMatrixSetFromQuatCR(vrMatrix *mat_dst, vrQuat *quat_src)
{
#endif /* ] */

	vrMatrixSetRotationFromQuat(mat_dst, quat_src);

/* 02/10/21: probably use _ROWCOL & swap the arguments (do this then delete me, the comment) */
	/* reset translation to none */
	VRMAT_COLROW(mat_dst, VR_W, VR_X) = VRMAT_COLROW(mat_dst, VR_W, VR_Y) = VRMAT_COLROW(mat_dst, VR_W, VR_Z) = 0.0f;

	/* reset the 4th column to normal */
	VRMAT_COLROW(mat_dst, VR_X, VR_W) = 0.0f;
	VRMAT_COLROW(mat_dst, VR_Y, VR_W) = 0.0f;
	VRMAT_COLROW(mat_dst, VR_Z, VR_W) = 0.0f;
	VRMAT_COLROW(mat_dst, VR_W, VR_W) = 1.0f;

	return mat_dst;
}


/*********************************************************************/
/* 01/13/2021: NOTE that based on my comparisons today of matrix     */ /* future delete me -- these comments */
/*   and quaternion operations in Octave it seems that there is a    */
/*   sign-error in the output quaternion from this calculation.      */
/*   So I need to recheck the algorithm.                             */
/* 02/10/2021: checked, and yes, the row/col indices were flipped.   */
vrQuat *vrQuatSetFromMatrix(vrQuat *quat_dst, vrMatrix *mat_src)
{
#if 0 /* this is Ed's method (set to '1'), for new method, set to '0' { */	/* future delete me */
	double	w_2, x_2, y_2;
	double	scale;
	double	len;

#  if 0
	vrPrintMatrix("Ed's vrQuatSetFromMatrix() -- matrix", mat_src);
#  endif

	w_2 = (1.0 + VRMAT_ROWCOL(mat_src, VR_X, VR_X) + VRMAT_ROWCOL(mat_src, VR_Y, VR_Y) + VRMAT_ROWCOL(mat_src, VR_Z, VR_Z)) * 0.25;

	if (w_2 > FLT_EPSILON) {
		quat_dst->v[VR_W] = sqrt(w_2);
		scale = 4.0 * quat_dst->v[VR_W];
		quat_dst->v[VR_X] = (VRMAT_ROWCOL(mat_src, VR_Y, VR_Z) - VRMAT_ROWCOL(mat_src, VR_Z, VR_Y)) / scale;
		quat_dst->v[VR_Y] = (VRMAT_ROWCOL(mat_src, VR_Z, VR_X) - VRMAT_ROWCOL(mat_src, VR_X, VR_Z)) / scale;
		quat_dst->v[VR_Z] = (VRMAT_ROWCOL(mat_src, VR_X, VR_Y) - VRMAT_ROWCOL(mat_src, VR_Y, VR_X)) / scale;
	} else {
		quat_dst->v[VR_W] = 0.0;
		x_2 = (VRMAT_ROWCOL(mat_src, VR_Y, VR_Y) + VRMAT_ROWCOL(mat_src, VR_Z, VR_Z)) * -0.5;

		if (x_2 > FLT_EPSILON) {
			quat_dst->v[VR_X] = sqrt(x_2);
			scale = 2.0 * quat_dst->v[VR_X];
			quat_dst->v[VR_Y] = VRMAT_ROWCOL(mat_src, VR_X, VR_Y) / scale;
			quat_dst->v[VR_Z] = VRMAT_ROWCOL(mat_src, VR_X, VR_Z) / scale;
		} else {
			quat_dst->v[VR_X] = 0.0;
			y_2 = (1.0 - VRMAT_ROWCOL(mat_src, VR_Z, VR_Z)) * 0.5;

			if (y_2 > FLT_EPSILON) {
				quat_dst->v[VR_Y] = sqrt(y_2);
				scale = 2.0 * quat_dst->v[VR_Y];
				quat_dst->v[VR_Z] = VRMAT_ROWCOL(mat_src, VR_Y, VR_Z) / scale;
			} else {
				quat_dst->v[VR_Y] = 0.0;
				quat_dst->v[VR_Z] = 1.0;
			}
		}
	}

	/* Normalize the quaternion */
	len = sqrt(vrQuatLengthSq(quat_dst));

	quat_dst->v[VR_X] = quat_dst->v[VR_X] / len;
	quat_dst->v[VR_Y] = quat_dst->v[VR_Y] / len;
	quat_dst->v[VR_Z] = quat_dst->v[VR_Z] / len;
	quat_dst->v[VR_W] = quat_dst->v[VR_W] / len;

#else /* Ed's method } { new method */

	/* This version of the calculation comes from the webpage:  */
	/*   http://skal.planet-d.net/demo/matrixfaq.htm#Q55        */
	/* [02/10/21: now: https://web.archive.org/web/20051229080102/http://skal.planet-d.net/demo/matrixfaq.htm */
	/*   Or: http://www.j3d.org/matrix_faq/matrfaq_latest.html */
	double	trace;
	double	scale;

	//vrPrintf("vrQuatSetFromMatrix() -- Warning, likely sign error.\n");

	trace = 1.0 + VRMAT_ROWCOL(mat_src, VR_X, VR_X) + VRMAT_ROWCOL(mat_src, VR_Y, VR_Y) + VRMAT_ROWCOL(mat_src, VR_Z, VR_Z);

	/* test for epsilon above zero to avoid distortions that can arise when near zero */
	if (trace > FLT_EPSILON) {
//vrPrintf("vrQuatSetFromMatrix() -- calculating using positive trace.\n");
		scale = sqrt(trace) * 2.0;
		quat_dst->v[VR_W] = scale * 0.25;
		quat_dst->v[VR_X] = (VRMAT_ROWCOL(mat_src, VR_Z, VR_Y) - VRMAT_ROWCOL(mat_src, VR_Y, VR_Z)) / scale;
		quat_dst->v[VR_Y] = (VRMAT_ROWCOL(mat_src, VR_X, VR_Z) - VRMAT_ROWCOL(mat_src, VR_Z, VR_X)) / scale;
		quat_dst->v[VR_Z] = (VRMAT_ROWCOL(mat_src, VR_Y, VR_X) - VRMAT_ROWCOL(mat_src, VR_X, VR_Y)) / scale;
	} else {
		/* If the trace of the matrix is equal to zero (or nearly) then identify */
		/*   which major diagonal element has the greatest value.                */
		/* Depending on this, calculate the following:                           */

		if (VRMAT_ROWCOL(mat_src, VR_X, VR_X) > VRMAT_ROWCOL(mat_src, VR_Y, VR_Y) &&
			VRMAT_ROWCOL(mat_src, VR_X, VR_X) > VRMAT_ROWCOL(mat_src, VR_Z, VR_Z))  {

			/* X,X is greatest */
			scale  = sqrt(1.0 + VRMAT_ROWCOL(mat_src, VR_X, VR_X) - VRMAT_ROWCOL(mat_src, VR_Y, VR_Y) - VRMAT_ROWCOL(mat_src, VR_Z, VR_Z)) * 2;
			quat_dst->v[VR_W] = (VRMAT_ROWCOL(mat_src, VR_Z, VR_Y) - VRMAT_ROWCOL(mat_src, VR_Y, VR_Z)) / scale;
			quat_dst->v[VR_X] = 0.25 * scale;
			quat_dst->v[VR_Y] = (VRMAT_ROWCOL(mat_src, VR_Y, VR_X) + VRMAT_ROWCOL(mat_src, VR_X, VR_Y)) / scale;
			quat_dst->v[VR_Z] = (VRMAT_ROWCOL(mat_src, VR_X, VR_Z) + VRMAT_ROWCOL(mat_src, VR_Z, VR_X)) / scale;
		} else if (VRMAT_ROWCOL(mat_src, VR_Y, VR_Y) > VRMAT_ROWCOL(mat_src, VR_Z, VR_Z)) {
			/* Y,Y is greatest */
			scale  = sqrt(1.0 + VRMAT_ROWCOL(mat_src, VR_Y, VR_Y) - VRMAT_ROWCOL(mat_src, VR_X, VR_X) - VRMAT_ROWCOL(mat_src, VR_Z, VR_Z)) * 2;
			quat_dst->v[VR_W] = (VRMAT_ROWCOL(mat_src, VR_X, VR_Z) - VRMAT_ROWCOL(mat_src, VR_Z, VR_X)) / scale;
			quat_dst->v[VR_X] = (VRMAT_ROWCOL(mat_src, VR_Y, VR_X) + VRMAT_ROWCOL(mat_src, VR_X, VR_Y)) / scale;
			quat_dst->v[VR_Y] = 0.25 * scale;
			quat_dst->v[VR_Z] = (VRMAT_ROWCOL(mat_src, VR_Z, VR_Y) + VRMAT_ROWCOL(mat_src, VR_Y, VR_Z)) / scale;
		} else {
			/* Z,Z is greatest  */
			scale  = sqrt(1.0 + VRMAT_ROWCOL(mat_src, VR_Z, VR_Z) - VRMAT_ROWCOL(mat_src, VR_X, VR_X) - VRMAT_ROWCOL(mat_src, VR_Y, VR_Y)) * 2;
			quat_dst->v[VR_W] = (VRMAT_ROWCOL(mat_src, VR_Y, VR_X) - VRMAT_ROWCOL(mat_src, VR_X, VR_Y)) / scale;
			quat_dst->v[VR_X] = (VRMAT_ROWCOL(mat_src, VR_X, VR_Z) + VRMAT_ROWCOL(mat_src, VR_X, VR_Z)) / scale;
			quat_dst->v[VR_Y] = (VRMAT_ROWCOL(mat_src, VR_Z, VR_Y) + VRMAT_ROWCOL(mat_src, VR_Y, VR_Z)) / scale;
			quat_dst->v[VR_Z] = 0.25 * scale;
		}
	}
#endif /* } new method */

#if 0
	vrPrintf("vrQuatSetFromMatrix() -- Warning, likely sign error.\n\t");
	vrPrintQuat("vrQuatSetFromMatrix() -- quat_dst", quat_dst);
#endif

	return quat_dst;
}


/*********************************************************************/
/* p = q p q^(-1)                                                    */
vrPoint *vrPointTransformByQuat(vrPoint *pnt_dst, vrPoint *pnt_src, vrQuat *quat_src)
{
	vrErr("vrPointTransformByQuat() -- not implemented yet");
	return pnt_src;
}


/*********************************************************************/
/* dp = m * sp                                                       */
/* It is safe to pass the source and destination points as the same. */
/* NOTE: this is the equivalent of extending the 3-tuple vector by   */
/*   one element with a value of 1, and then multiplying the 4-tuple */
/*   by the 4x4 matrix.                                              */
vrPoint *vrPointTransformByMatrix(vrPoint *pnt_dst, vrPoint *pnt_src, vrMatrix *mat_src)
{
	double	tmp[4];
	int	row;
	int	col;

	for (row = VR_X; row <= VR_W; row++) {
		tmp[row] = 0.0f;
		for (col = VR_X; col <= VR_Z; col++)
			tmp[row] += VRMAT_ROWCOL(mat_src, row, col) * pnt_src->v[col];
		tmp[row] += VRMAT_ROWCOL(mat_src, row, VR_W);
	}
	for (row = VR_X; row <= VR_Z; row++)
		pnt_dst->v[row] = tmp[row] / tmp[VR_W];

#ifndef MATH_ONLY
	if (vrDbgDo(ALMOSTNEVER_DBGLVL/2))
		vrPrintMatrix("vrMatrixXformPoint():", mat_src);
#endif

	return pnt_dst;
}


/*********************************************************************/
/* Get the translation values into a vrPoint from a 4x4 vrMatrix.    */
/*   TODO: usage of this should go away once we fully adopt the vrVector operations */
/*      9/7/02 -- why?  It appears vrVectorGetTransFromMatrix() is even being used  */
/*        in vr_entity.c where this function is more appropriate.                   */
/*      9/25/02 -- now vr_entity.c is using this function.                          */
vrPoint *vrPointGetTransFromMatrix(vrPoint *pnt_dst, vrMatrix *mat_src)
{
	pnt_dst->v[VR_X] = VRMAT_ROWCOL(mat_src, VR_X, VR_W);
	pnt_dst->v[VR_Y] = VRMAT_ROWCOL(mat_src, VR_Y, VR_W);
	pnt_dst->v[VR_Z] = VRMAT_ROWCOL(mat_src, VR_Z, VR_W);

	return pnt_dst;
}


/*********************************************************************/
vrVector *vrVectorGetTransFromMatrix(vrVector *vec_dst, vrMatrix *mat_src)
{
	vec_dst->v[VR_X] = VRMAT_ROWCOL(mat_src, VR_X, VR_W);
	vec_dst->v[VR_Y] = VRMAT_ROWCOL(mat_src, VR_Y, VR_W);
	vec_dst->v[VR_Z] = VRMAT_ROWCOL(mat_src, VR_Z, VR_W);

	return vec_dst;
}


/*********************************************************************/
vrMatrix *vrMatrixSetTransFromVector(vrMatrix *mat_dst, vrVector *vec_src)
{
	VRMAT_ROWCOL(mat_dst, VR_X, VR_W) = vec_src->v[VR_X];
	VRMAT_ROWCOL(mat_dst, VR_Y, VR_W) = vec_src->v[VR_Y];
	VRMAT_ROWCOL(mat_dst, VR_Z, VR_W) = vec_src->v[VR_Z];

	return mat_dst;
}


/*********************************************************************/
double vrDegreesToRadians(double deg)
{
	return (deg * M_PI) / 180.0f;
}


/*********************************************************************/
double vrRadiansToDegrees(double rad)
{
	return (rad * 180.0) / M_PI;
}



/* ==================================================== */
/* DEBUGGING PRINTERS                                   */
/* ==================================================== */


/*********************************************************************/
/* vrFprintMatrix(): prints the matrix in OpenGL-style -- ie. with   */
/*   the basis vectors in columns with the translation values down   */
/*   the right column.                                               */
/* Here is how the OpenGL-style matrix corresponds to the matrix     */
/*   array elements (i.e. column-major ordering):                    */
/*        v[0]   v[4]   v[8]  v[12]                                  */
/*        v[1]   v[5]   v[9]  v[13]                                  */
/*        v[2]   v[6]  v[10]  v[14]                                  */
/*        v[3]   v[7]  v[11]  v[15]                                  */
/*                                                                   */
/* Arguments:                                                        */
/*   * file -- file descriptor to send output to                     */
/*   * before -- a descriptive string that is printed before the mat */
/*   * mat_src -- pointer to the matrix to be printed                */
/* NOTE: to maintain indentation, any non-newline whitespace at the  */
/*   beginning of "before" will be prepended to each line.           */
void vrFprintMatrix(FILE *file, const char *before, vrMatrix *mat_src)
{
static	char	*whitespace = " \t\r\b-";
static	char	before_whitespace[32];
	int	row;

	if (before) {
		vrFprintf(file, "%s\n", before);

		/* copy just the whitespace */
		strncpy(before_whitespace, before, sizeof(before_whitespace));
		before_whitespace[strspn(before_whitespace, whitespace)] = '\0';
	}

	if (mat_src == NULL) {
		vrFprintf(file, "\t%s<nil>\n", before_whitespace);

		return;
	}

	for (row = 0; row < VRMAT_MAXDIM; row++) {
		vrFprintf(file, "\t%s%6.3f  %6.3f  %6.3f  %6.3f\n", before_whitespace,
			VRMAT_ROWCOL(mat_src, row, VR_X), VRMAT_ROWCOL(mat_src, row, VR_Y), VRMAT_ROWCOL(mat_src, row, VR_Z), VRMAT_ROWCOL(mat_src, row, VR_W));
	}

#if 1 /* Option: Add output of the data in single dimension array form (in memory order): */
	int	element;
	vrFprintf(file, "\t%s", before_whitespace);
	for (element = 0; element < 16; element++) {
		vrFprintf(file, "%6.3f ", mat_src->v[element]);
		if (element % 4 == 3)
			vrFprintf(file, "| ");
	}
	vrFprintf(file, "\n");
#endif
}


/*********************************************************************/
/* vrPrintMatrix(): same as vrFprintMatrix(), except prints directly */
/*   to standard-out (stdout).                                       */
void vrPrintMatrix(const char *before, vrMatrix *mat_src)
{
	vrFprintMatrix(stdout, before, mat_src);
}


/*********************************************************************/
/* vrFprintMatrixT(): print the transpose of the matrix.             */
/* Arguments are the same as for vrFprintMatrix().                   */
/*                                                                   */
/* NOTE: this is often done to print a 4x4 matrix with the           */
/*   translation values in the bottom row (C-style).                 */
void vrFprintMatrixT(FILE *file, const char *before, vrMatrix *mat_src)
{
static	char	*whitespace = " \t\r\b-";
static	char	before_whitespace[32];
	int	row;

	if (before) {
		vrFprintf(file, "%s\n", before);

		/* copy just the whitespace */
		strncpy(before_whitespace, before, sizeof(before_whitespace));
		before_whitespace[strspn(before_whitespace, whitespace)] = '\0';
	}

	if (mat_src == NULL) {
		vrFprintf(file, "\t%s<nil>\n", before_whitespace);

		return;
	}

	for (row = 0; row < VRMAT_MAXDIM; row++) {
		vrFprintf(file, "\t%s%6.3f  %6.3f  %6.3f  %6.3f\n", before_whitespace,
			VRMAT_ROWCOL(mat_src, VR_X, row), VRMAT_ROWCOL(mat_src, VR_Y, row), VRMAT_ROWCOL(mat_src, VR_Z, row), VRMAT_ROWCOL(mat_src, VR_W, row));
	}
}


/***********************************************************************/
/* vrPrintMatrixT(): same as vrFprintMatrixT(), except prints directly */
/*   to standard-out (stdout).                                         */
void vrPrintMatrixT(const char *before, vrMatrix *mat_src)
{
	vrFprintMatrixT(stdout, before, mat_src);
}


/*********************************************************************/
/* vrPrintMatrixComposition(): print the extraced components of a    */
/*   4x4 homogeneous matrix.                                         */
/*                                                                   */
/* Arguments:                                                        */
/*   * file -- file descriptor to send output to                     */
/*   * before -- a descriptive string that is printed before the mat */
/*   * mat_src -- pointer to the matrix to be printed                */
/*                                                                   */
/* NOTE: to maintain indentation, any non-newline whitespace at the  */
/*   beginning of "before" will be prepended to each line.           */
/*                                                                   */
/* NOTE: Euler angles have no one answer -- which is what quaternions*/
/*   are generally preferred -- but can help with human reasoning.   */
void vrPrintMatrixComposition(const char *before, vrMatrix *mat_src)
{
static	char			*whitespace = " \t\r\b-";
static	char			before_whitespace[32];
	vrMatrixComposition	components;
	int			no_singularity;

	no_singularity = vrMatrixCompositionFromMatrix(&components, mat_src);
	if (before) {
		vrPrintf("%s\n", before);

		/* copy just the whitespace */
		strncpy(before_whitespace, before, sizeof(before_whitespace));
		before_whitespace[strspn(before_whitespace, whitespace)] = '\0';
	}

	if (!no_singularity)
		vrPrintf("\t%s*singularity*\n", before_whitespace);

	vrPrintf("\t%sscale: (%2.2f, %2.2f, %2.2f)\n"
		"\t%stranslate: (%2.2f, %2.2f, %2.2f)\n"
		"\t%srotate: (%2.2f, %2.2f, %2.2f)\n"
		"\t%squaternion: (%2.3fi, %2.3fj, %2.3fk, %2.3f)\n"
		"\t%sshear: (%2.2f, %2.2f, %2.2f)\n"
		"\t%sperspective: (%2.2f, %2.2f, %2.2f, %2.2f)\n",
		before_whitespace, components.scale[VR_X], components.scale[VR_Y], components.scale[VR_Z],
		before_whitespace, components.translate[VR_X], components.translate[VR_Y], components.translate[VR_Z],
		before_whitespace, components.rotate[VR_X], components.rotate[VR_Y], components.rotate[VR_Z],
		before_whitespace, components.quaternion[VR_X], components.quaternion[VR_Y], components.quaternion[VR_Z], components.quaternion[VR_W],
		before_whitespace, components.shear[VR_X], components.shear[VR_Y], components.shear[VR_Z],
		before_whitespace, components.perspective[VR_X], components.perspective[VR_Y], components.perspective[VR_Z], components.perspective[VR_W]);
}


/***********************************************************************/
/* vrFprintMatrixCompositionOL(): The "One-Line" printed version of    */
/*   vrPrintMatrixComposition() -- also prints to FILE pointer instead */
/*   of directl to stdout.                                             */
/*                                                                     */
/* Often used when outputting a 6-sensor input in "brief" or "verbose" */
/*   styles.                                                           */
void vrFprintMatrixCompositionOL(FILE *fp, const char *before, vrMatrix *mat_src)
{
	vrMatrixComposition	components;
	int			no_singularity;

	no_singularity = vrMatrixCompositionFromMatrix(&components, mat_src);
	if (before)
		vrFprintf(fp, "%s", before);

	if (!no_singularity)
		vrFprintf(fp, "\t*singularity*\n");

	vrFprintf(fp,
		"sc (%2.1f, %2.1f, %2.1f)"
		" t (%2.2f, %2.2f, %2.2f)"
		" r (%2.1f, %2.1f, %2.1f)"
		" q (%2.2fi, %2.2fj, %2.2fk, %2.2f)"
		" sh (%2.1f, %2.1f, %2.1f)"
		" p (%2.1f, %2.1f, %2.1f, %2.1f)\n",
		components.scale[VR_X], components.scale[VR_Y], components.scale[VR_Z],
		components.translate[VR_X], components.translate[VR_Y], components.translate[VR_Z],
		components.rotate[VR_X], components.rotate[VR_Y], components.rotate[VR_Z],
		components.quaternion[VR_X], components.quaternion[VR_Y], components.quaternion[VR_Z], components.quaternion[VR_W],
		components.shear[VR_X], components.shear[VR_Y], components.shear[VR_Z],
		components.perspective[VR_X], components.perspective[VR_Y], components.perspective[VR_Z], components.perspective[VR_W]);
}


/*********************************************************************/
/* NOTE: (i,j,k) are shown in order to make it clear that FreeVR */
/*   quaternions are ordered scalar-last.                        */
void vrPrintQuat(const char *before, vrQuat *quat_src)
{
	if (before)
		vrPrintf("%s [(%2.3fi, %2.3fj, %2.3fk), %2.3f]\n",
			before, quat_src->v[VR_X], quat_src->v[VR_Y], quat_src->v[VR_Z], quat_src->v[VR_W]);
	else
		vrPrintf(   "[(%2.3fi, %2.3fj, %2.3fk), %2.3f]\n",
			        quat_src->v[VR_X], quat_src->v[VR_Y], quat_src->v[VR_Z], quat_src->v[VR_W]);
}


/*********************************************************************/
void vrPrintPoint(const char *before, vrPoint *pnt)
{
	if (before)
		vrPrintf("%s (%2.2f, %2.2f, %2.2f)\n", before, pnt->v[VR_X], pnt->v[VR_Y], pnt->v[VR_Z]);
	else	vrPrintf(   "(%2.2f, %2.2f, %2.2f)\n",         pnt->v[VR_X], pnt->v[VR_Y], pnt->v[VR_Z]);
}


/*********************************************************************/
void vrPrintVector(const char *before, vrVector *vec)
{
	if (before)
		vrPrintf("%s (%2.2f, %2.2f, %2.2f)\n", before, vec->v[VR_X], vec->v[VR_Y], vec->v[VR_Z]);
	else	vrPrintf(   "(%2.2f, %2.2f, %2.2f)\n",         vec->v[VR_X], vec->v[VR_Y], vec->v[VR_Z]);
}


/*********************************************************************/
void vrPrintVector4(const char *before, vrVector4 *vec)
{
	if (before)
		vrPrintf("%s (%2.2f, %2.2f, %2.2f, %2.2f)\n", before, vec->v[VR_X], vec->v[VR_Y], vec->v[VR_Z], vec->v[VR_W]);
	else	vrPrintf(   "(%2.2f, %2.2f, %2.2f, %2.2f)\n",         vec->v[VR_X], vec->v[VR_Y], vec->v[VR_Z], vec->v[VR_W]);
}


/*********************************************************************/
/* NOTE: It's important to remember that AZIM is a rotation about the up-axis, */
/*    and ELEV is a rotation about what is normally the X-axis.                */
void vrPrintEuler(const char *before, vrEuler *euler)
{
	if (before)
		vrPrintf("%s (%2.2f, %2.2f, %2.2f), (%2.2f, %2.2f, %2.2f)\n", before,
			euler->t[VR_X], euler->t[VR_Y], euler->t[VR_Z],
			euler->r[VR_AZIM], euler->r[VR_ELEV], euler->r[VR_ROLL]);
	else	vrPrintf(   "(%2.2f, %2.2f, %2.2f), (%2.2f, %2.2f, %2.2f)\n",
			euler->t[VR_X], euler->t[VR_Y], euler->t[VR_Z],
			euler->r[VR_AZIM], euler->r[VR_ELEV], euler->r[VR_ROLL]);
}


/*
 * ===================================================
 *
 * DEFINITIONS FOR MATRIX INVERSION
 * I'm doing matrix inversion with Gauss-Jordan reduction,
 * so I need to define some auxiliary data structures
 * ===================================================
 */
/* 04/06/2021: vrAugMat can't be a normal vrMatrix because it's actually a */
/*   4x8 matrix of double values -- but I wonder if there isn't a better   */
/*   way to represent this.                                                */
#define VRAUG_ROWCOL(mat, row, col)	((mat)[(8*(row) + (col))])
typedef double *vrAugMat;	/* TODO (perhaps): make this a structure like vrMatrix */


/*********************************************************************/
vrAugMat vrAugMatCreate(vrMatrix *m)
{
	int		row;
	int		col;
	vrAugMat	augmat = vrShmemAlloc0(8*4*sizeof(augmat[0]));

	for (row = 0; row < 4; row++) {
		for (col = 0; col < 4; col++) {
			VRAUG_ROWCOL(augmat, row, col) = VRMAT_ROWCOL(m, row, col);
		}
		VRAUG_ROWCOL(augmat, row, 4+row) = 1.0f;
	}
	return augmat;
}


/*********************************************************************/
void vrAugMatDelete(vrAugMat augmat) {
	vrShmemFree(augmat);
}


/*********************************************************************/
void vrAugMatPrint(const char *before, vrAugMat augmat)
{
	int	row;
	int	col;

	if (before) vrPrintf("%s\n", before);
	for (row = 0; row < 4; row++) {
		vrPrintf("\t");
		for (col = 0; col < 8; col++)
			vrPrintf("%2.2f  ", VRAUG_ROWCOL(augmat, row, col));
		vrPrintf("\n");
	}
}


/*********************************************************************/
/* elementary row operations on an augmented matrix                  */
void swap_rows(vrAugMat augmat, int row1, int row2)
{
	int	col;
	double	tmp;

	for (col = 0; col < 8; col++) {
		tmp = VRAUG_ROWCOL(augmat, row1, col);
		VRAUG_ROWCOL(augmat, row1, col) = VRAUG_ROWCOL(augmat, row2, col);
		VRAUG_ROWCOL(augmat, row2, col) = tmp;
	}
}


/*********************************************************************/
void scale_row(vrAugMat augmat, int row, double s)
{
	int	col;

	for (col = 0; col < 8; col++)
		VRAUG_ROWCOL(augmat, row, col) *= s;
}


/*********************************************************************/
void combine_rows(vrAugMat augmat, int row1, int row2, double s)
{
	int	col;

	for (col = 0; col < 8; col++)
		VRAUG_ROWCOL(augmat, row1, col) += VRAUG_ROWCOL(augmat, row2, col) * s;
}


/*********************************************************************/
/*  helper routines for reduction                                    */
#include <limits.h>
int fewest_zeros(vrAugMat augmat, int start)
{
	int	row;
	int	best = INT_MAX;
	int	best_idx = -1;

	for (row = start; row < 4; row++) {
		int zero_cnt = 0;
		int col = 0;
		while (VRAUG_ROWCOL(augmat, row, col++) == 0)
			zero_cnt++;
		if (zero_cnt < best) {
			best = zero_cnt;
			best_idx = row;
		}
	}
	return best_idx;
}


/*********************************************************************/
/*  do the reduction                                                 */
void vrAugMatReduce(vrAugMat augmat)
{
	int	row;
	double	factor;

	for (row = 0; row < 4; row++) {
		int	row2;
		int	best;

		/* adjust rows if necessary */
		best = fewest_zeros(augmat, row);

		if (best > row) {
			swap_rows(augmat, row, best);
		}

		/* scale the row if necessary */
		if (VRAUG_ROWCOL(augmat, row, row) != 1.0f &&
			VRAUG_ROWCOL(augmat, row, row) != 0.0f) {
			factor = 1.0f/VRAUG_ROWCOL(augmat, row, row);
			scale_row(augmat, row, factor);
		}

		/* Remove nonzero entries in the following rows.  */
		/* We're going to cheat a little here and use     */
		/* the fact that we're doing this on a 4x4 matrix.*/
		for (row2=0; row2<4; row2++) {
			factor = - VRAUG_ROWCOL(augmat, row2, row);
			if (factor != 0.0f && row2 != row) {
				combine_rows(augmat, row2, row, factor);
			}
		}
	}
}


/*********************************************************************/
void vrGetInverseFromAug(vrAugMat augmat, vrMatrix *inv)
{
	int	row;
	int	col;

	for (row = 0; row < 4; row++)
		for (col = 0; col < 4; col++)
			VRMAT_ROWCOL(inv, row, col) = VRAUG_ROWCOL(augmat, row, col+4);
}


/*********************************************************************/
vrMatrix *vrMatrixInvert(vrMatrix *dst, vrMatrix *src)
{
	vrAugMat	aug = vrAugMatCreate(src);

	vrAugMatReduce(aug);
	vrGetInverseFromAug(aug, dst);
	vrAugMatDelete(aug);

	return dst;
}


/*********************************************************************/
vrMatrix *vrMatrixInvertInPlace(vrMatrix *mat_dst)
{
	vrAugMat	aug = vrAugMatCreate(mat_dst);

	vrAugMatReduce(aug);
	vrGetInverseFromAug(aug, mat_dst);
	vrAugMatDelete(aug);

	return (mat_dst);
}

/**************************************************************************/
/* NOTE: this function doesn't work -- the intent is to decompose a       */
/*   4x4 matrix into the component parts (scale, shear, rotation,         */
/*   translation, and perspective).                                       */
/* 02/10/2021: I think this code maybe DOES work now.                     */  /* delete me -- these preface comments that are out of date */
/*                                                                        */
/* The function returns '1' when it can successfully decompose the        */
/*   matrix, 0 when it cannot.                                            */
/* [01/11/2021: My code here comes from the book Graphics Gems II.        */
/*   unmatrix - Decompose a non-degenerate 4x4 transformation matrix      */
/* 	into the sequence of transformations that produced it.            */
/* [Sx][Sy][Sz][Shearx/y][Sx/z][Sz/y][Rx][Ry][Rz][Tx][Ty][Tz][P(x,y,z,w)] */
/*                                                                        */
/*   Returns 1 upon success, 0 if the matrix is singular.                 */
/*                                                                        */
/*   The original "unmatrix.c" code by Spencer W. Thomas is at (https):   */
/*   - www.realtimerendering.com/resources/GraphicsGems/gemsii/unmatrix.c */
/*                                                                        */
/*   See also the web pages:                                              */
/*   - https://caff.de/posts/4X4-matrix-decomposition                     */
/*   - https://caff.de/posts/4X4-matrix-decomposition/decomposition.pdf   */
/*   - https://stackoverflow.com/questions/15022630/how-to-calculate-the-angle-from-rotation-matrix (has both solution for independent Eulers, or for angle around a single axis) */
/*   - https://gamedev.net/forums/topic/384075-extracting-scale-vector-from-a-transformation-matrix/3537443 */
/*   - https://www.gregslabaugh.net/publications/euler.pdf                */
/*   - https://www.geometrictools.com/Documentation/EulerAngles.pdf       */
int vrMatrixCompositionFromMatrix(vrMatrixComposition *comp, vrMatrix *mat_src)
{
	int		count, count2;
	vrMatrix	local_mat;	/* "locmat" in unmatrix.c */
	vrMatrix	pmat;
	vrMatrix	invpmat;
	vrVector4	prhs;
	vrVector	row[3];		/* Three rows of the upper-left 3x3 */
	vrVector	tmpvec;		/* "pdum3" in unmatrix.c */

	vrQuat		local_quat;
	double		trace;		/* rotation matrix trace (for quaternion) */
	double		halfinvsqrtrace;/* 0.5 / sqrt(trace) */

	/* copy the matrix to local space */
	local_mat = *mat_src;

	comp->singularity = 0;		/* initially assume not singular */

	/*** Normalize the matrix ***/
	if (VRMAT_COLROW(&local_mat, VR_W, VR_W) != 0) {
		for (count = VR_X; count <= VR_W; count++)
			for (count2 = VR_X; count2 <= VR_W; count2++)
				VRMAT_COLROW(&local_mat, count, count2) /= VRMAT_COLROW(&local_mat, VR_W, VR_W);
	} else {
		/* degenerate matrix -- can't divide by 0 */
		comp->singularity = 1;	/* singular matrix */
	}

	/*** Initialize pmat and check for 3x3 singularity ***/
	pmat = local_mat;
	for (count = VR_X; count <= VR_Z; count++)
		VRMAT_COLROW(&pmat, count, VR_W) = 0.0;
	VRMAT_COLROW(&pmat, VR_W, VR_W) = 1.0;

	if (vrMatrixDeterminant(&pmat) == 0.0)
		comp->singularity = 1;	/* singular matrix */
	vrDbgPrintfN(MATH_DBGLVL, "vrMatrixCompositionFromMatrix(): 4x4 matrix determinant = %.3f\n", vrMatrixDeterminant(&pmat));

	/*** In cases of singularity, send back all zeros ***/
	if (comp->singularity == 1) {
		comp->scale[VR_X] = comp->scale[VR_Y] = comp->scale[VR_Z] = 0.0;
		comp->shear[0] = comp->shear[1] = comp->shear[2] = 0.0;
		comp->rotate[VR_X] = comp->rotate[VR_Y] = comp->rotate[VR_Z] = 0.0;
		comp->quaternion[VR_X] = comp->quaternion[VR_Y] = comp->quaternion[VR_Z] = 0.0;
		comp->quaternion[VR_W] = 1.0;	/* or send back 0.0 since it's singular */
		comp->translate[VR_X] = comp->translate[VR_Y] = comp->translate[VR_Z] = 0.0;
		comp->perspective[VR_X] = comp->perspective[VR_Y] = comp->perspective[VR_Z] = 0.0;
		comp->perspective[VR_W] = 1.0; /* Not sure 0.0 or 1.0? */

		return 0;	/* matrix is singular (not valid) */
	}

	/*** Find the perspective values (this is the messiest) ***/
	if (VRMAT_COLROW(&local_mat, VR_X, VR_W) != 0 ||
		VRMAT_COLROW(&local_mat, VR_Y, VR_W) != 0 ||
		VRMAT_COLROW(&local_mat, VR_Z, VR_W) != 0) {

		/* prhs is the right hand side of the equation */
		prhs.v[VR_X] = VRMAT_COLROW(&local_mat, VR_X, VR_W);
		prhs.v[VR_Y] = VRMAT_COLROW(&local_mat, VR_Y, VR_W);
		prhs.v[VR_Z] = VRMAT_COLROW(&local_mat, VR_Z, VR_W);
		prhs.v[VR_W] = VRMAT_COLROW(&local_mat, VR_W, VR_W);

		/* solve the equation by inverting pmat and multiplying prhs by the */
		/*   inverse.  (This is the easiest way, not necessarily the best.) */
		vrMatrixInvert(&invpmat, &pmat);
		/* [01/12/21: NOTE: in the original unmatrix.c code, there's a transposition here, but I may have embedded that into the following line -- should check] */
		vrVector4TransformByMatrix((vrVector4 *)comp->perspective, &prhs, &invpmat);

		/* Clear the perspective partition */
		VRMAT_COLROW(&local_mat, VR_X, VR_W) = 0.0;
		VRMAT_COLROW(&local_mat, VR_Y, VR_W) = 0.0;
		VRMAT_COLROW(&local_mat, VR_Z, VR_W) = 0.0;
		VRMAT_COLROW(&local_mat, VR_W, VR_W) = 1.0;

	} else {
		vrDbgPrintfN(MATH_DBGLVL, "vrMatrixCompositionFromMatrix(): no perspective\n");
		/* There isn't any perspective */
		comp->perspective[VR_X] = 0.0;
		comp->perspective[VR_Y] = 0.0;
		comp->perspective[VR_Z] = 0.0;
		comp->perspective[VR_W] = 1.0;	/* NOTE: in the original, this is also set to 0.0, but that seems incorrect to me. */
	}

	/*** Take care of translation (easy) ***/
	for (count = VR_X; count <= VR_Z; count++) {
		comp->translate[count] = VRMAT_COLROW(&local_mat, VR_W, count);
		VRMAT_COLROW(&local_mat, VR_W, count) = 0.0;	/* erase from our local copy */
	}

	/*** Get the scale and shear ***/
	for (count = VR_X; count <= VR_Z; count++) {
#if 0 /* 01/12/21: these were COLROW, but seems wrong */ /* delete me */
		row[count].v[VR_X] = VRMAT_ROWCOL(&local_mat, count, VR_X);	/* 01/12/21: these were COLROW, but seems wrong */
		row[count].v[VR_Y] = VRMAT_ROWCOL(&local_mat, count, VR_Y);
		row[count].v[VR_Z] = VRMAT_ROWCOL(&local_mat, count, VR_Z);
#else /* 01/13/21: I compared the results with Octave's quaternion code, and this gives the correct answer. */
		row[count].v[VR_X] = VRMAT_COLROW(&local_mat, count, VR_X);
		row[count].v[VR_Y] = VRMAT_COLROW(&local_mat, count, VR_Y);
		row[count].v[VR_Z] = VRMAT_COLROW(&local_mat, count, VR_Z);
#endif
	}

	/* Compute X scale factor and normalize first row */
	comp->scale[VR_X] = vrVectorLength(&row[VR_X]);
	vrVectorNormalize(&row[VR_X], &row[VR_X]);

	/* Compute XY shear factor and make 2nd row orthogonal to 1st */
	comp->shear[0] = vrVectorDotProduct(&row[VR_X], &row[VR_Y]);	/* Shear X/Y */
	vrDbgPrintfN(MATH_DBGLVL, "vrMatrixCompositionFromMatrix(): Shear X/Y = %.2f\n", comp->shear[0]);
	vrVectorAddScaledVectors(&row[VR_Y], &row[VR_Y], &row[VR_X], 1.0, -comp->shear[0]);

	/* Compute Y scale and normalize 2nd row */
	comp->scale[VR_Y] = vrVectorLength(&row[VR_Y]);
	vrVectorNormalize(&row[VR_Y], &row[VR_Y]);

	comp->shear[0] /= comp->scale[VR_Y];				/* Adjust Shear X/Y */
	vrDbgPrintfN(MATH_DBGLVL, "vrMatrixCompositionFromMatrix(): Now Shear X/Y = %.2f\n", comp->shear[0]);

	/* Compute XZ and YZ shears, orthogonalize 3rd row */
	comp->shear[1] = vrVectorDotProduct(&row[VR_X], &row[VR_Z]);	/* Shear X/Z */
	vrDbgPrintfN(MATH_DBGLVL, "vrMatrixCompositionFromMatrix(): Shear X/Z = %.2f\n", comp->shear[0]);
	vrVectorAddScaledVectors(&row[VR_Z], &row[VR_Z], &row[VR_X], 1.0, -comp->shear[1]);
	comp->shear[2] = vrVectorDotProduct(&row[VR_Y], &row[VR_Z]);	/* Shear Y/Z */
	vrDbgPrintfN(MATH_DBGLVL, "vrMatrixCompositionFromMatrix(): Shear Y/Z = %.2f\n", comp->shear[0]);
	vrVectorAddScaledVectors(&row[VR_Z], &row[VR_Z], &row[VR_Y], 1.0, -comp->shear[2]);

	/* Compute Z scale and normalize 3rd row */
	comp->scale[VR_Z] = vrVectorLength(&row[VR_Z]);
	vrVectorNormalize(&row[VR_Z], &row[VR_Z]);
	comp->shear[1] /= comp->scale[VR_Z];				/* Adjust Shear X/Z */
	vrDbgPrintfN(MATH_DBGLVL, "vrMatrixCompositionFromMatrix(): Now Shear X/Z = %.2f\n", comp->shear[0]);
	comp->shear[2] /= comp->scale[VR_Z];				/* Adjust Shear Y/Z */
	vrDbgPrintfN(MATH_DBGLVL, "vrMatrixCompositionFromMatrix(): Now Shear Y/Z = %.2f\n", comp->shear[0]);

	/* At this point, the matrix (in rows[]) is orthonormal.  Check */
	/*   for a coordinate system flip.  If the determinant is -1,   */
	/*   then negate the matrix and the scaling factors.            */
	if (vrVectorDotProduct(&row[VR_X], vrVectorCrossProduct(&tmpvec, &row[VR_Y], &row[VR_Z])) < 0) {
		for (count = VR_X; count <= VR_Z; count++) {
			comp->scale[count] *= -1.0;
			row[count].v[VR_X] *= -1.0;
			row[count].v[VR_Y] *= -1.0;
			row[count].v[VR_Z] *= -1.0;
		}
	}

	/*** Now extract the rotation as Euler angles ***/
	/* TODO: 01/12/2021: These numbers are the negative of what I expect -- sign error here or elsewhere? */
	comp->rotate[VR_Y] = RTOD(asin(-row[VR_X].v[VR_Z]));
	if (cos(comp->rotate[VR_Y]) != 0.0) {
		comp->rotate[VR_X] = RTOD(atan2(row[VR_Y].v[VR_Z], row[VR_Z].v[VR_Z]));
		comp->rotate[VR_Z] = RTOD(atan2(row[VR_X].v[VR_Y], row[VR_X].v[VR_X]));
	} else {
		comp->rotate[VR_X] = RTOD(atan2(-row[VR_Z].v[VR_X], row[VR_Y].v[VR_Y]));
		comp->rotate[VR_Z] = 0.0;
	}

	/*** Now extract the rotation as a Quaternion ***/
	if (row[VR_Z].v[VR_Z] < 0.0) {
		/* |(x,y)| is bigger than |(z,w)| */

		if (row[VR_X].v[VR_X] > row[VR_Y].v[VR_Y]) {
			/* |x| is bigger than |y|, so use x-form */
			trace = 1.0 + row[VR_X].v[VR_X] - row[VR_Y].v[VR_Y] - row[VR_Z].v[VR_Z];
			//vrPrintf("x-form, trace = %.2f\n", trace);
			vrQuatSet4d(&local_quat,
				trace,
				row[VR_X].v[VR_Y] + row[VR_Y].v[VR_X],
				row[VR_Z].v[VR_X] + row[VR_X].v[VR_Z],
				row[VR_Y].v[VR_Z] - row[VR_Z].v[VR_Y]);
		} else {
			/* use y-form */
			trace = 1.0 - row[VR_X].v[VR_X] + row[VR_Y].v[VR_Y] - row[VR_Z].v[VR_Z];
			//vrPrintf("y-form, trace = %.2f\n", trace);
			vrQuatSet4d(&local_quat,
				row[VR_X].v[VR_Y] + row[VR_Y].v[VR_X],
				trace,
				row[VR_Y].v[VR_Z] + row[VR_Z].v[VR_Y],
				row[VR_Z].v[VR_X] - row[VR_X].v[VR_Z]);
		}

	} else {
		if (row[VR_X].v[VR_X] < -row[VR_Y].v[VR_Y]) {
			/* |z| is bigger than |y|, so use z-form */
			trace = 1.0 - row[VR_X].v[VR_X] - row[VR_Y].v[VR_Y] + row[VR_Z].v[VR_Z];
			//vrPrintf("z-form, trace = %.2f\n", trace);
			vrQuatSet4d(&local_quat,
				row[VR_Z].v[VR_X] + row[VR_X].v[VR_Z],
				row[VR_Y].v[VR_Z] + row[VR_Z].v[VR_Y],
				trace,
				row[VR_X].v[VR_Y] - row[VR_Y].v[VR_X]);
		} else {
			/* use w-form */
			trace = 1.0 + row[VR_X].v[VR_X] + row[VR_Y].v[VR_Y] + row[VR_Z].v[VR_Z];
			//vrPrintf("w-form, trace = %.2f\n", trace);
			vrQuatSet4d(&local_quat,
				row[VR_Y].v[VR_Z] - row[VR_Z].v[VR_Y],
				row[VR_Z].v[VR_X] - row[VR_X].v[VR_Z],
				row[VR_X].v[VR_Y] - row[VR_Y].v[VR_X],
				trace);
		}
	}

	/* Quat-scalar multiply */
	/* NOTE: FreeVR doesn't presently have a function for this, so doing it inline */
	halfinvsqrtrace = 0.5 / sqrt(trace);
	comp->quaternion[VR_X] = local_quat.v[VR_X] * halfinvsqrtrace;
	comp->quaternion[VR_Y] = local_quat.v[VR_Y] * halfinvsqrtrace;
	comp->quaternion[VR_Z] = local_quat.v[VR_Z] * halfinvsqrtrace;
	comp->quaternion[VR_W] = local_quat.v[VR_W] * halfinvsqrtrace;

#if 0
	/* use the original FreeVR math code for converting from matrix to quat (and override above calculation) */
	vrQuatSetFromMatrix(&local_quat, mat_src);
#endif

	return 1;	/* no singularity (i.e. is-valid) */
}


/***********************************************************************/
/* This section create a program that tests some of the math routines, */
/*   printing the result of various operations.  Tests are added on an */
/*   as-needed basis (ie. I think there might be a problem with either */
/*   one routine, or a section of routines.                            */
/*                                                                     */
/* The included Makefile will compile this program with the target:    */
/*   % make mathtest                                                   */
/***********************************************************************/
#if defined(TEST_APP) && !defined(MATH_ONLY) /* { */
int main(int argc, char **argv)
{
	vrPrintf("NOTE: cos(30) = %2.3f, sin(30) = %2.3f\n", cos(vrDegreesToRadians(30.0)), sin(vrDegreesToRadians(30.0)));
	vrPrintf("NOTE: cos(45) = %2.3f, sin(45) = %2.3f\n", cos(vrDegreesToRadians(45.0)), sin(vrDegreesToRadians(45.0)));
	vrPrintf("NOTE: cos(90) = %2.3f, sin(90) = %2.3f\n", cos(vrDegreesToRadians(90.0)), sin(vrDegreesToRadians(90.0)));
	vrPrintf("NOTE: sqrt(2) = %2.3f, sq(2)/2 = %2.3f, 1/sq(2) = %2.3f\n", sqrt(2.0), sqrt(2.0)/2.0, 1.0/sqrt(2.0));
	vrPrintf("NOTE: sqrt(3) = %2.3f, sq(3)/2 = %2.3f, 1/sq(3) = %2.3f\n", sqrt(3.0), sqrt(3.0)/2.0, 1.0/sqrt(3.0));
	vrPrintf("NOTE: pi      = %2.3f, pi/2    = %2.3f, pi/4    = %2.3f\n", PI, (PI/2.0), (PI/4.0));
#if 0 /* Option: { */
	vrPrintf(BOLD_TEXT "\nSome basic matrix operation tests:\n\n" NORM_TEXT);

	vrMatrix	*m = vrMatrixSetRotationId(vrMatrixCreate(), VR_X, 30.0f);
	vrMatrix	*i = vrMatrixInvert(vrMatrixCreate(), m);
	vrMatrix	*p = vrMatrixProduct(vrMatrixCreate(), i, m);

	vrPrintMatrix("30d rot about X = ", m);
	vrPrintMatrix("inverse of above = ", i);
	vrPrintMatrix("product of these two (s/b id) = ", p);


  {
	/* tests of the Euclidean inversion stuff with travel matrices from Kalev */
	vrMatrix	cl;
	vrMatrix	clinv;
	double		clvals[16] = { -0.964685, 0.000000, -0.165759, 0.000000,
					0.000000, 1.000000, 0.000000, 0.000000,
					0.165759, 0.000000, -0.964685, 0.000000,
					5.373073, -23.577421, -20.114393, 1.000000 };

	vrPrintf(BOLD_TEXT "\ntests of the Euclidean inversion stuff:\n\n" NORM_TEXT);
	vrMatrixSetAd(&cl, clvals);
	vrMatrixInvertEuclidean(&clinv, &cl);
	vrPrintMatrix("Supposed Euclidean matrix", &cl);
	vrPrintMatrixComposition("decomp: Euclidean matrix", &cl);
	vrPrintMatrix("Inverted Euclidean matrix", &clinv);
	vrPrintMatrixComposition("decomp: Inv-Euclidean matrix", &clinv);

	if (!vrMatrixIsEuclidean(&cl)) {
		vrPrintf("Actually it wasn't Euclidean\n");
		vrMatrixMakeEuclidean(&cl, &cl);
		vrMatrixInvertEuclidean(&clinv, &cl);
		vrPrintMatrix("Converted Euclidean matrix", &cl);
		vrPrintMatrixComposition("decomp: Conv-Euclidean matrix", &cl);
		vrPrintMatrix("Inverted Euclidean matrix", &clinv);
		vrPrintMatrixComposition("decomp: Inv-Conv-Euclidean matrix", &clinv);
	}

  }
  {
	/* further tests of the Euclidean inversion stuff with travel matrices from Kalev */
	vrMatrix	tmpmat;
	vrMatrix	cl;
	vrMatrix	cleuc;
	double		clvals[16] = { 0.040264,   0.000000,  -0.978011, 0.000000,	/* row-major */
					 0.000000,   1.000000,   0.000000, 0.000000,
					 0.978011,   0.000000,   0.040264, 0.000000,
					20.632961, -14.370701, -36.058109, 1.000000 };
	double		scale;

	vrPrintf(BOLD_TEXT "\nFurther tests of the Euclidean inversion stuff:\n\n" NORM_TEXT);
	vrMatrixSetAd(&cl, clvals);
	vrPrintMatrix("initial matrix", &cl);

	vrMatrixMakeEuclidean(&cleuc, &cl);
	vrPrintMatrix("first conversion", &cleuc);

	vrMatrixMakeEuclidean(&cl, &cleuc);
	vrPrintMatrix("second conversion", &cl);

	vrMatrixMakeEuclidean(&cleuc, &cl);
	vrPrintMatrix("third conversion", &cleuc);

	scale = vrMatrixGetScale(&cleuc);
	vrPrintf("The scale of the latest matrix is %f\n", scale);
	vrMatrixSetScale3d(&tmpmat, 1.0f/scale, 1.0f/scale, 1.0f/scale);
	vrMatrixPostMult(&cleuc, &tmpmat);
	vrPrintMatrix("rescaled matrix is", &cleuc);

	vrMatrixSetScale(&tmpmat, &cleuc, 1.0);
	vrPrintMatrix("rescaled via vrMatrixSetScale() matrix is", &cleuc);

	scale = vrMatrixGetScale(&cleuc);
	vrPrintf("The scale of the rescaled matrix is %f\n", scale);

  }
  {
	/* tests of the Quaternion functions */
	vrMatrix	*mid = vrMatrixCreateIdentity();
	vrMatrix	*m90x = vrMatrixSetRotationId(vrMatrixCreate(), VR_X, 90.0f);
	vrMatrix	*m30x = vrMatrixSetRotationId(vrMatrixCreate(), VR_X, 30.0f);
	vrQuat		quat;
	vrQuat		*q30x = vrQuatSet4dThetaAxis(vrQuatCreate(), 30.0, 1.0, 0.0, 0.0);

	vrPrintf(BOLD_TEXT "\n\nSome Quaternion function tests:\n\n" NORM_TEXT);
	vrPrintMatrix("identity matrix is", mid);
	vrQuatSetFromMatrix(&quat, mid);
	vrPrintQuat("vrQuatSetFromMatrix() -- identity", &quat);
	vrMatrixSetRotationFromQuat(mid, &quat);
	vrPrintMatrix("and converted back to matrix again is", mid);

	vrPrintMatrix("\n90 about X matrix is", m90x);
	vrQuatSetFromMatrix(&quat, m90x);
	vrPrintQuat("vrQuatSetFromMatrix() -- m90x", &quat);
	vrMatrixSetRotationFromQuat(m90x, &quat);
	vrPrintMatrix("and converted back to matrix again is", m90x);

	vrPrintMatrix("\n30 about X matrix is", m30x);
	vrQuatSetFromMatrix(&quat, m30x);
	vrPrintQuat("vrQuatSetFromMatrix() -- m30x", &quat);
	vrMatrixSetRotationFromQuat(m30x, &quat);
	vrPrintMatrix("and converted back to matrix again is", m30x);

	vrPrintQuat("\nvrQuatSet4dThetaAxis() -- q30x", q30x);
	vrMatrixSetRotationFromQuat(m30x, q30x);
	vrPrintMatrix("and converted to matrix is", m30x);
	vrPrintMatrix("with the transpose of:", vrMatrixTranspose(m30x, m30x));

  }
#elif 1 /* } Option: { */
  {
	vrPrintf(BOLD_TEXT "\n\nMore Quaternion function tests (Z-up vs. Y-up):\n" NORM_TEXT);
	/* tests of the Quaternion Y-UP and Z-UP conversion routines */
	vrPoint		*point1 = vrPointCreate3d(10.0, 20.0, 30.0);
	vrQuat		*p90x = vrQuatSet4dThetaAxis(vrQuatCreate(),  90.0, 1.0, 0.0, 0.0);
	vrQuat		*n90x = vrQuatSet4dThetaAxis(vrQuatCreate(), -90.0, 1.0, 0.0, 0.0);
	vrQuat		quat1;
	vrQuat		quat2;
	vrMatrix	mat;

	vrQuatSet4dThetaAxis(&quat1, 120.0, 0.0, 0.0, 1.0);
	vrPrintf("------------\n");
	vrPrintQuat("Original quaternion", &quat1);
	vrPrintQuat("Converted from Z-up to Y-up:", vrQuatConvertZUPtoYUP(&quat1, &quat1));
	vrPrintQuat("Converted back to Z-up from Y-up:", vrQuatConvertYUPtoZUP(&quat1, &quat1));

	/* now do the same with multiplies */
	vrQuatSet4dThetaAxis(&quat1, 45, 0.0, 1.0, 0.0);
	vrQuatSet4dThetaAxis(&quat1, 120.0, 0.0, 0.0, 1.0);
	vrQuatSet4dThetaAxis(&quat1, -30.0, 1.0, 0.0, 0.0);
	vrPrintf("------------\n");
	vrPrintQuat("Original quaternion", &quat1);
	vrQuatProduct(&quat2, n90x, &quat1);
	vrPrintQuat("Converted from Z-up to Y-up:", &quat2);
	vrQuatProduct(&quat1, p90x, &quat2);
	vrPrintQuat("Converted back to Z-up from Y-up:", &quat1);
	vrQuatProduct(&quat2, n90x, &quat1);
	vrPrintQuat("Converted again from Z-up to Y-up:", &quat2);
	vrQuatProduct(&quat1, p90x, &quat2);
	vrPrintQuat("Converted again back to Z-up from Y-up:", &quat1);
	vrPrintf("------------\n");
	vrQuatSet4dThetaAxis(&quat1, 90.0, 1.0, 0.0, 0.0);
	vrPrintQuat("+90 x-rot as Quat:", &quat1);
	//vrMatrixSetIdentity(&mat);
	//vrMatrixSetRotation
	vrMatrixSetTranslation3d(&mat, 30.0, 20.0, 10.0);
	vrMatrixSetRotationFromQuat(&mat, &quat1);
	vrPrintMatrix("+90 x-rot as Matrix", &mat);
	vrPrintMatrixComposition("decomp: +90 x-rot as Matrix", &mat);
	vrEuler	euler;
	vrEulerSetFromMatrix(&euler, &mat);
	vrPrintEuler("+90 x-rot as Euler (a,e,r)", &euler);
	vrEulerSetFromQuatPoint(&euler, &quat1, point1);
	vrPrintEuler("From Quat: +90 x-rot as Euler (a,e,r)", &euler);

	vrPrintf("------------\n");
	vrQuatSet4dThetaAxis(&quat1, 90.0, 0.0, 1.0, 0.0);
	vrPrintQuat("+90 Y-rot as Quat:", &quat1);
	vrMatrixSetRotationFromQuat(&mat, &quat1);
	vrPrintMatrix("+90 y-rot as Matrix", &mat);
	vrEulerSetFromMatrix(&euler, &mat);
	vrPrintEuler("+90 y-rot as Euler (a,e,r)", &euler);

	/* now the same with matrix multiplies */
	vrPrintf("------------\n");
	vrQuatSet4dThetaAxis(&quat1, 90.0, 0.0, 0.0, 1.0);
	vrPrintQuat("+90 Z-rot as Quat:", &quat1);
	vrMatrixSetRotationFromQuat(&mat, &quat1);
	vrPrintMatrix("+90 Z-rot as Matrix", &mat);
	vrEulerSetFromMatrix(&euler, &mat);
	vrPrintEuler("+90 Z-rot as Euler (a,e,r)", &euler);

	/* now the same with matrix multiplies */
	vrQuatSet4dThetaAxis(&quat1, 90.0, 0.0, 1.0, 0.0);
	vrPrintf("------------\n");
	vrPrintQuat("Original quaternion", &quat1);
	vrPrintPoint("test point is: ", point1);
	vrPointTransformByQuat(point1, point1, &quat1);
	vrPrintPoint("test point by quat: ", point1);

	/* Basic tests of vrEulerSetFromQuatPoint() */
	/* point1 will remain (10, 20, 30) */
	/* I need to test different angles because of the singularity issue */
	vrQuatSet4dThetaAxis(&quat1, 45.0, 0.0, 1.0, 0.0);
	vrPrintQuat("Quaternion for 45 about Y:", &quat1);
	vrPrintPoint("test point is: ", point1);
	vrEulerSetFromQuatPoint(&euler, &quat1, point1);
	vrPrintEuler("From Quat: +45 y-rot as Euler (a,e,r)", &euler);

	vrQuatSet4dThetaAxis(&quat1, 45.0, 0.0, 0.0, 1.0);
	vrPrintQuat("Quaternion for 45 about Z:", &quat1);
	vrPrintPoint("test point is: ", point1);
	vrEulerSetFromQuatPoint(&euler, &quat1, point1);
	vrPrintEuler("From Quat: +45 z-rot as Euler (a,e,r)", &euler);


/* Other quaternion stuff to test:

DONE: vrQuat	*vrQuatProduct(vrQuat *prod, vrQuat *ql, vrQuat *qr);
vrQuat		*vrQuatProductEquals(vrQuat *prod, vrQuat *ql, vrQuat *qr);
DONE: vrMatrix	*vrMatrixSetFromQuat(vrMatrix *mat, vrQuat *quat);
vrPoint		*vrPointTransformByQuat(vrPoint *pnt_dst, vrPoint *pnt_src, vrQuat *quat);
	[and when I add it: vrQuatSetFromVector()]
	[and when I add it: vrVectorTransformByQuat() -- verify that this would be different from vrPointTransformByquat -- or should this be vrVectorRotateByQuat()?]
*/

	/* Quick test to confirm that I'm making an identity matrix from coords */
	double	coords[9];
	vrPrintf("\n");
	vrCoordsSetIdentity(coords);
	vrMatrixSetFromCoords(&mat, coords);
	vrPrintMatrix("Matrix from Coords", &mat);
	vrPrintMatrixComposition("decomp: Matrix from Coords", &mat);

  }
#else /* } Option: { */
	vrMatrix	*m = vrMatrixSetRotationId(vrMatrixCreate(), VR_X, -10.0f);
	vrPrintMatrix("-10d rot about X = ", m);
	vrPrintMatrixComposition("decomp: -10d rot about X = ", m);
#endif /* } */
}
#endif /* } TEST_APP */



#if 0 /* { */
/*******************************************************************/
/*******************************************************************/
/*******************************************************************/
/*******************************************************************/
/* Stuff from cave.boom.c (written by Bill Sherman) ****************/
/*   This old code has some routines used to convert a 4x4 matrix  */
/*   into 6 Euler values.  At the moment we don't need this code,  */
/*   but just in case it's needed in the future, I figure the math */
/*   source file is the best place to store it.  I (Bill) wrote    */
/*   this to integrate a Fakespace BOOM-2 into the CAVE library -- */
/*   which uses Eulers for dealing with all 6-sensors.             */


/******************************/
/*** CAVEGetBoomTracking(): ***/
/******************************/
int CAVEGetBoomTracking(CAVE_ST *cave,CAVE_SENSOR_ST *sensor)
{
static	CAVE_SENSOR_ST	head = { 0.0, 0.0, 0.0,   0.0, 0.0, 0.0 };
	float		mat[4][4];
	float		*lmat;
	float		azi, elev, roll;
	float		azi_s, azi_c;

	if (boom_fd < 0) {
		fprintf(stderr,"CAVE tracking: no boom communications port.\n");
		return 0;
	}

	lmat = BOOM_read(boom_fd);
	CAVEGetTimestamp(&head.timestamp);
	memcpy(mat, lmat, 16*sizeof(*mat));

	head.x = mat[3][0]*3.0;
	head.y = mat[3][1]*3.0 + 5.0;
	head.z = mat[3][2]*3.0;

	/*** This section of code calculates the proper Euler angles from ***/
	/***   the 4x4 matrix using techniques from Robotics, Fu et.al.   ***/
	/***   Actually, the matrix order seems backward, and I lucked    ***/
	/***   into getting them backward - figure it out later, it works!***/
	azi  = atan2(mat[2][0], mat[2][2]);
	azi_s = sin(azi);
	azi_c = cos(azi);
	elev = atan2(mat[2][1], azi_s*mat[2][0] + azi_c*mat[2][2]);
	roll = atan2(azi_s*mat[1][2] - azi_c*mat[1][0], azi_c*mat[0][0] - azi_s*mat[0][2]);
	head.azim =  RTOD(azi);
	head.elev = -RTOD(elev);
	head.roll =  RTOD(roll);

	/*** put information into return structs ***/
	sensor[0] = head;
	sensor[1] = head;	/* The BOOM only has one tracked object */

	return 1;
}


/*****************************/
/*** BOOM_angles2matrix(): ***/
/*****************************/
static void BOOM_angles2matrix(float angles[6],float matrix[4][4])
{
	int	i, j;	/* loop counters              */

	/*** initialize matrix to 4x4 identity matrix ***/
	for (i = 0; i < 4; i++)
		for (j = 0; j < 4; j++)
			matrix[i][j] = (i == j) ? 1 : 0;

	matrix_rotate(matrix, angles[0] * (-1.0), 'y');
	matrix_rotate(matrix, angles[1],          'x');
	matrix_translate(matrix, 0.0, 1.0, 0.0);
	matrix_rotate(matrix, angles[2] - M_PI_2, 'x');
	matrix_translate(matrix, 0.0, 0.0, 1.0);
	matrix_rotate(matrix, angles[3] - M_PI_2, 'x');
	matrix_translate(matrix, 0.0, -0.5, 0.0);
	matrix_rotate(matrix, angles[4],          'y');
	matrix_rotate(matrix, angles[5],          'x');

	matrix[3][1] -= 0.5;	/* display is at y=0	      */
}


/*****************************************************************/
/*** BOOM_matrix2euler(): convert a 4x4 matrix to Euler angles ***/
/***   plus a Cartesian 3-tuple.                               ***/
/***                                                           ***/
/*** NOTE: this code isn't necessarily correct, I don't use it ***/
/***   It has the essence of solving for the Euler angles, but ***/
/***   to see how it is really done, refer to "Robotics" by Fu ***/
/***   Gonzales & Lee.                                         ***/
/*****************************************************************/
static void BOOM_matrix2euler(float matrix[4][4], float eulers[6])
{
static	float	epsilon = (1e-10);
	float	sinPitch, cosPitch;
	float	sinRoll, cosRoll;
	float	sinYaw, cosYaw;

	/* Cartesian location is just copied from matrix */
	eulers[3] = matrix[3][0];
	eulers[4] = matrix[3][1];
	eulers[5] = matrix[3][2];

	/* determine Euler angles from matrix - a little harder */
	sinPitch = -matrix[2][0];
	cosPitch = sqrt(1 - sinPitch*sinPitch);
	if (matrix[0][2] < 0) cosPitch = -cosPitch;

	if (fabs(cosPitch) > epsilon) {
		sinRoll = matrix[2][1] / cosPitch;
		cosRoll = matrix[2][2] / cosPitch;
		sinYaw = matrix[1][0] / cosPitch;
		cosYaw = matrix[0][0] / cosPitch;
	} else {
		sinRoll = -matrix[1][2];
		cosRoll = matrix[1][1];
		sinYaw = 0;
		cosYaw = 1;
	}

	eulers[0] = atan2(sinYaw, cosYaw);
	eulers[1] = atan2(sinPitch, cosPitch);
	eulers[2] = atan2(sinRoll, cosRoll);
}


/************************/
/*** matrix_rotate(): ***/
/************************/
static void matrix_rotate(float mat[4][4],float angle,char axis)
{
	float	sin_angle, cos_angle, temp;
	int	i, j, k;

	switch (axis) {
	case 'x':
		i = 1;
		j = 2;
		break;
	case 'y':
		i = 2;
		j = 0;
		break;
	case 'z':
		i = 0;
		j = 1;
		break;
	}

	sin_angle = sinf(angle);
	cos_angle = cosf(angle);

	for (k = 0; k < 4; k++) {
		temp = cos_angle * mat[j][k] - sin_angle * mat[i][k];
		mat[i][k] = sin_angle * mat[j][k] + cos_angle * mat[i][k];
		mat[j][k] = temp;
	}
}


/***************************/
/*** matrix_translate(): ***/
/***************************/
static void matrix_translate(float mat[4][4], float x, float y, float z)
{
	int	count;

	for (count = 0; count < 3; count++)
		mat[3][count] += mat[0][count] * x + mat[1][count] * y + mat[2][count] * z;
}

#endif /* } */

