/* ======================================================================
 *
 *  CCCCC          vr_output.stream.c
 * CC   CC         Author(s): Bill Sherman
 * CC              Created: November 23, 2021
 * CC   CC         Last Modified: July 27, 2023 (minor fix 02/29/2024)
 *  CCCCC
 *
 * Code file for FreeVR outputs of select data format streams.
 *
 * Copyright 2024, Bill Sherman, All rights reserved.
 * With the intent to provide an open-source license to be named later.
 * ====================================================================== */
/*************************************************************************

USAGE:
	The "stream" FreeVR output device is a simple half-duplex exporter
	of input data from FreeVR, which includes tracking, valuators and
	buttons.  By not requiring any feedback from the other side, we can
	use this device to write to files, or stream to a socket with the
	receiving application unable to send requests back to FreeVR.

	[NOTE: there's a chance that the Monado connection is more full-duplex
	than I originally thought because the "monado-gui" tool seems to know
	the last values sent by the FreeVR stream, which implies that a request
	is being made -- this can be explored with "socketspy".]


	Ideally, this output type would have the ability to minutely
	control which outputs are streamed out and in what particular
	order, but for now there are just a handful of hardcoded steaming
	options.

	1) CSV: simply takes all the inputs available (except for the
		first button, which is the 'ESC' key), and writes them
		as comma-separated ASCII values either to a file (or
		standard-out/stdout), or to a socket listening on the
		given port.

	2) Monado "remote device" style 1 (based on Google Daydream):
		outputs the first three sensors as HMD, left controller,
		and right controller.  Each with a "pose" (quat & location),
		a "hand curl", a flag for hand tracking, and 4 buttons.

	3) Monado "remote device" style 2 (based on the Valve Index):
		also outputs three sensors as an HMD, and two controllers
		(left & right), which also have embedded "button" and
		valuator inputs.  (Many of the "button" (boolean/2switch)
		inputs are from touch sensors.)


	Controls are specified in the freevrrc file:
		control "<control option>" = "2switch(...)";

	Here are the available control options for FreeVR:
	  :-(	"print_help" -- print info on how to use the output device
	  :-(	"system_pause_toggle" -- toggle the system pause flag
	  :-(	"print_context" -- print the overall FreeVR context data structure (for debugging)
	  :-(	"print_config" -- print the overall FreeVR config data structure (for debugging)
	  :-(	"print_input" -- print the overall FreeVR input data structure (for debugging)
	  :-(	"print_devinfo" -- print the internal Shared Memory data structure (for debugging)
	  :-(	"print_values" -- print just the input values of the memory


	Here are the FreeVR configuration options for the Stream output Device:
		"output" - filename to write data to (or at "@" address will
			activate the socket writing code)
			Socket format: "@[<hostname>][:<port>]"

			Example: "output = @hmdmachine:3883;" -- will send to
				a host called "hmdmachine" on port 3883.
			Example: "output = @;" -- will send to "localhost:4242".
			Example: "output = @:3883" -- will send to
				"localhost:3883".

		"port" - socket port listening for applications requesting data.
			(port 4242 is the default).
			NOTE: the port can also be set as part of the "output"
			when sending to a socket, and any port specified as
			part of the "output" argument, will take precedence.

			NOTE: if no connection can be made to the specifed port
			then the data will stream to standard out (stdout).

NOTES:
	As mentioned, this is a half-duplex only streaming output feature
	for FreeVR.  As it happens, the Monado "protocol" does actually have
	a slight full-duplex aspect as the Monado OpenXR runtime will send
	the current values for the input data back through the socket (twice)
	after a new connection is made.  But because this is half-duplex, that
	data is ignored.

	CSV data can be sent to a file, "stdout", or a socket.  For use with a
	socket, you can test be listening on port 4242 (the Monado default):
		% nc -l 4242

	The CSV stream has no metadata or header, so one needs to know the
	configuration in effect when the output was streamed.


MONADO "remote" protocol-1 (~20220413): based on Google Daydream
	Perhaps refer to as "Monado-simple" or "Monado1".

	From: src/xrt/drivers/remote/r_interface.h
	From: src/xrt/include/xrt/xrt_defines.h [contains the xrt_pose/xrt_vecN/etc struct typedef's]

	[200 bytes total]
	header (8bytes) = "mndrmt1\0"
	// HMD
	"xrt_pose": ("hmd.pose")
		[28 bytes, start at byte 9]
		"xrt_quat" (4 floats: x,y,z,w) ("orientation")
		"xrt_vec3" (3 floats: x,y,z) ("position")
		(no padding here)
	// Left Controller
	"r_remote_controller": ("left")
		[80 bytes, start at byte 37]
		xrt_pose: (7 floats: qx,qy,qz,qw,x,y,z) ("pose")
		xrt_vec3: (3 floats: x,y,z) ("linear_velocity")
		xrt_vec3: (3 floats: x,y,z) ("angular_velocity")
		float[5]: (5 floats:...) ("hand_curl")
		bool: (1 ??) ("hand_tracking_active")
		bool: (1 ??) ("active")
		bool: (1 ??) ("select")
		bool: (1 ??) ("menu")
		bool: (1 ??) ("_pad")
		(3 bytes of padding to get to an 8-byte boundary)
	// Right Controller
	//   [same as Left Controller]
		[80 bytes, start at byte 117]
	// Padding
		[4 bytes of padding needed to get to an 8-byte boundary]

MONADO "remote" protocol-2 (~20220614): based on Valve Index
	Perhaps refer to as "Monado-Index" or "Monado2".
	R_HEADER_VALUE = "mndrmt2\0"

	From: src/xrt/drivers/remote/r_interface.h
	From: src/xrt/include/xrt/xrt_defines.h [contains the xrt_pose/xrt_vecN/etc struct typedef's]

	[280 bytes total]
	header (8bytes) = "mndrmt2\0"
	// HMD
	"xrt_pose": ("hmd.pose") [same as "protocol-1"]
		[28 bytes, start at byte 9]
		"xrt_quat" (4 floats: x,y,z,w) ("orientation")
		"xrt_vec3" (3 floats: x,y,z) ("position")
		(no padding here)
	// Left Controller
	"r_remote_controller": ("left")
		[120 bytes, start at byte 37]
		xrt_pose: (7 floats: qx,qy,qz,qw,x,y,z) ("pose")
		xrt_vec3: (3 floats: x,y,z) ("linear_velocity")
		xrt_vec3: (3 floats: x,y,z) ("angular_velocity")

		float[5]: (5 floats:...) ("hand_curl")

			[byte: 72+36 = 108]
	(new)	xrt_vec1: (1 float: x)    ("trigger_value")
	(new)	xrt_vec1: (1 float: x)    ("squeeze_value")
	(new)	xrt_vec1: (1 float: x)    ("squeeze_force")
	(new)	xrt_vec2: (2 floats: x,y) ("thumbstick")
	(new)	xrt_vec1: (1 float: x)    ("trackpack_force")
	(new)	xrt_vec2: (2 floats: x,y) ("trackpack")

			[byte: 72+36+32 = 140]
		bool: (1 ??) ("hand_tracking_active")
		bool: (1 ??) ("active")

		bool: (1 ??) ("system_click")
		bool: (1 ??) ("system_touch")
		bool: (1 ??) ("a_click")
		bool: (1 ??) ("a_touch")
		bool: (1 ??) ("b_click")
		bool: (1 ??) ("b_touch")
	   ----
		bool: (1 ??) ("trigger_click")
		bool: (1 ??) ("trigger_touch")
		bool: (1 ??) ("thumbstick_click")
		bool: (1 ??) ("thumbstick_touch")
		bool: (1 ??) ("trackpad_touch")
		bool: (1 ??) ("_pad0")
		bool: (1 ??) ("_pad1")
		bool: (1 ??) ("_pad2")
		(no bytes of padding needed to get to an 8-byte boundary)
	// Right Controller
	//   [same as Left Controller]
		[120 bytes, start at byte 156]
	// Padding
		[4 bytes of padding to get to an 8-byte boundary]

MONADO "remote" protocol-3 (~20230718): based on ?? HMD?
	(Still in effect on 02/29/2024)
	Perhaps refer to as "Monado3".
	R_HEADER_VALUE = "mndrmt3\0"

	Change from protocol-2:
		* replacing:
			struct { struct xrt_pose pose } hmd
		  with:
			struct r_head_data head;
				(which now has a lot more data!)


	From: src/xrt/drivers/remote/r_interface.h
	From: src/xrt/include/xrt/xrt_defines.h [contains the xrt_pose/xrt_vecN/etc struct typedef's]

	[376 bytes total]
	header [8 bytes] = "mndrmt3\0"
	// HMD
	"head_data": ("hmd.pose") [changed from "protocol-1" & "protocol-2"]
	[128 bytes total, start at byte 9]
	 (new)	"views"[2] [48 bytes each = 96 bytes total]
			"fov" ("struct xrt_fov"):
		 (new)		"angle_left" (1 float)
		 (new)		"angle_right" (1 float)
		 (new)		"angle_up" (1 float)
		 (new)		"angle_down" (1 float)
			"pose" ("xrt_pose"):
				[28 bytes, start at byte 41]
				"xrt_quat" (4 floats: x,y,z,w) ("orientation")
				"xrt_vec3" (3 floats: x,y,z) ("position")
		 (new)		[4 bytes of padding (uint32_t) to get to an 8-byte boundary]
		"center" ("xrt_pose"): [starts at byte 105]
			"xrt_quat" (4 floats: x,y,z,w) ("orientation")
			"xrt_vec3" (3 floats: x,y,z) ("position")
		"per_view_data_valid" ("bool" - 1 byte) -- "Is the per view data valid and should be used?"
		[3 bytes of padding (bool * 3)]
	[the rest is unchanged from protocol-2, but plus 100 bytes even (except
	4 bytes of padding got moved to the "head_data" sub-structure)]
	// Left Controller
	"r_remote_controller": ("left")
		[120 bytes, start at byte 137]
		xrt_pose: (7 floats: qx,qy,qz,qw,x,y,z) ("pose")
		xrt_vec3: (3 floats: x,y,z) ("linear_velocity")
		xrt_vec3: (3 floats: x,y,z) ("angular_velocity")

		float[5]: (5 floats:...) ("hand_curl")

			[byte: 72+36 = 208]
	(v2)	xrt_vec1: (1 float: x)    ("trigger_value")
	(v2)	xrt_vec1: (1 float: x)    ("squeeze_value")
	(v2)	xrt_vec1: (1 float: x)    ("squeeze_force")
	(v2)	xrt_vec2: (2 floats: x,y) ("thumbstick")
	(v2)	xrt_vec1: (1 float: x)    ("trackpack_force")
	(v2)	xrt_vec2: (2 floats: x,y) ("trackpack")

			[byte: 72+36+32 = 240]
		bool: (1 ??) ("hand_tracking_active")
		bool: (1 ??) ("active")

		bool: (1 ??) ("system_click")
		bool: (1 ??) ("system_touch")
		bool: (1 ??) ("a_click")
		bool: (1 ??) ("a_touch")
		bool: (1 ??) ("b_click")
		bool: (1 ??) ("b_touch")
	   ----
		bool: (1 ??) ("trigger_click")
		bool: (1 ??) ("trigger_touch")
		bool: (1 ??) ("thumbstick_click")
		bool: (1 ??) ("thumbstick_touch")
		bool: (1 ??) ("trackpad_touch")
		bool: (1 ??) ("_pad0")
		bool: (1 ??) ("_pad1")
		bool: (1 ??) ("_pad2")
		(no bytes of padding needed to get to an 8-byte boundary)
	// Right Controller
	//   [same as Left Controller]
		[120 bytes, start at byte 256]
	// Padding
		[4 bytes of padding to get to an 8-byte boundary]


HISTORY:
	23 November, 2021 (Bill Sherman) -- began writing the initial version.
		Code was started as a copy of the "vr_output.vruidd.c" routines.
		  And at the time was called "vr_output.csvdata.c", but then
		  sat undeveloped for 5 months.

	21-23 April, 2022 (Bill Sherman)
		I wrote the very basic code for outputting CSV data to either
		  a file (or stdout), or to a socket.  This included excluding
		  some of the VruiDD code, (sometimes by removing it, and other
		  times just #if'ing it out).
		Using source code from Monado, I documented the expected data
		  protocol.  Which I was (through trial and error) able to
		  confirm its correctness (after adding padding).
		I renamed this file a couple of times: from "csvdata" to
		  "generic" then to "stream".
		Both CSV and Monado output work, though for now selecting
		  between them requires setting a #if at compile-time.

	14 June, 2022 (Bill Sherman)
		I upgraded the Monado output to match the new, extended "remote"
		  input controls for Monado that Jakob Bornecratz informed me of
		  in the Monado Discord back on 05/28/22.

		[corresponds to FreeVR Version 0.7d]

	15 June, 2022 (Bill Sherman)
		Added the "protocol" configuration argument that allows the
		  format of the output to be selected from "csv", "monado1"
		  (based on Google Daydream) or "monado2" (based on Valve Index).
		Removing references to having an ASCII flag, since the choice
		  of protocol dictates that.
		I enhanced the "output" argument to take a formal address
		  setting regex rather than just the word "socket" that would
		  require other arguments to fill in the details.  Now an
		  address beginning with the '@' character indicates a socket
		  should be used.

		NOTE: testing of this feature is when I discovered a significant
		  shortcoming (bug) of the way my vrArgParse<type>() routines
		  work -- they just look for the first time an argument name
		  appears, which could even be on the right side of a setting
		  for some other value.  E.g. "output = port; port = 4242;"
		  will ignore the "port = 4242;" part because that's the 2nd
		  match of the string "port" in the arguments.

		[corresponds to FreeVR Version 0.7d]

	16 June, 2022 (Bill Sherman)
		Cleaned up most of the VruiDD code (left some as an example for
		  the possible future partial-full-duplex of reading the initial
		  output from a Monado connection).
		Corrected the _StreamPrintPrivateInfo() function to be for the
		  Stream-output rather than VruiDD.

		[corresponds to FreeVR Version 0.7d]

	17 July, 2023 (Bill Sherman)
		Added the new(ish) Monado protocol-3 (mndrmt3") stream output.

		[corresponds to FreeVR Version 0.7e]

	29 February, 2024 (Bill Sherman)
		Fixed a small bug where _StreamPrintPrivateInfo() was trying
		  to output the values of it's inputs, but Stream-output doesn't
		  have input values!  (It just sends the input values from the
		  other (actual) input devices!)

		[corresponds to FreeVR Version 0.7e]

TODO:
	* Consider adding an option to send time data with the CSV data

	* While one *feature* of this output device is that it's half-duplex,
		and only sends data.  I might consider at least reading the
		stream, perhaps only at the beginning, to evaluate whether
		there is any meta data that can be obtained.  And I think
		because Monado sends two copies of the current data back
		through a socket, I could look for two copies of the header
		string ("mndrmt1") and calculate how many bytes are in a
		single packet (or "mndrmt2" or "mndrmt3" for those protocols).

	* Consider using the Delta version of the vrGet<input>() functions
		so as not to mess with any "Delta()" forms.  Though this
		would be somewhat pedantic, as we really shouldn't be doing
		output stuff when running a real application -- but I like
		to be pedantic, it can save from future bugs.

	* Consider adding the ability to alter the endianness of binary data

	* Perhaps: break _StreamPutData() into sub-functions for each
		protocol (format) type.

	* DONE: Clean out the old VruiDD code
		[02/29/2024: it seems like there still is some VruiDD code here]

	* DONE: Add a configuration argument to control which protocol to output

	* DONE: Determine whether we need the "ascii" flag.
		Decision: remove all references to this flag.

	* DONE: Consider making "/dev/null" the default file

	* DONE: Change the syntax for setting socket usage from using the
		keyword "socket" to using the format "@<hostname>[:<port>]"
		(I'll keep the separate "port" option too.)

	* DONE: Basic CSV output works (both to file and socket)

	* DONE: Monado head and controller tracking
		[well the controller tracking still needs testing]

	* DONE: Determine whether there are full-duplex commands available
		(using socketspy).  There are no full-duplex commands that I
		found, but the Monado OpenXR runtime does send it's currently
		held input values out (twice) through a newly established
		connection.

	* DONE: Add Monado button inputs

	* DONE: Find out why Monado remote device doesn't have valuators
		Answer: because at the time it was created the developers
		were only testing simple input devices.

	* DONE: Update to new protocol-3 ("mndrmt3") Monado remote control.

BUGS: (or potential bugs)
	* The configuration arguments can interfere with one another if
		the name of one argument (e.g. "port") appears as the value
		of an earlier argument.  This is an overall FreeVR issue.

*************************************************************************/
#include <stdio.h>		/* needed for snprintf() */
#include <string.h>		/* needed for strchr() */
#include <stdint.h>		/* needed for uint16_t & uint32_t types */
#include <sys/time.h>		/* needed for struct timeval */
#include <fcntl.h>		/* for open(), O_RDWR, etc. */
#if 1 /* for Linger */
#include <netinet/tcp.h>	/* needed for TCP_NODELAY and TCP_CORK */
#include <netinet/in.h>		/* needed for IPPROTO_TCP */
#endif

#include "vr_input.h"
#include "vr_input.opts.h"
#include "vr_socket.h"
#include "vr_debug.h"
#include "vr_config.h"		/* for vrFprintConfig() */
#include "vr_parse.h"		/* for vrArgParseInteger() */
#include "vr_utils.h"		/* for vrAtoI() */

	/***************************************************/
	/*** definitions for interfacing with the device ***/
	/***                                             ***/

#define	BUFSIZE		2048
#define	DEFAULT_PORT	4242

enum OutputFormat {
		OUTPUT_CSV,		/* Output ASCII CSV data */
		OUTPUT_MONADO1,		/* Output binary Monado-1 "remote device" data (original protocol) */
		OUTPUT_MONADO2,		/* Output binary Monado-2 "remote device" data (new protocol) */
		OUTPUT_MONADO3		/* Output binary Monado-3 "remote device" data (new protocol) */
	};


/*****************************************************************/
/*** auxiliary structure of the current data about the device. ***/
typedef struct {
		/* this is for printing extra info in _StreamPrintValues() */
		vrInputDevice	*devinfo;

		/* these are for interfacing with the client (usually the application) */
		int		fd_output;		/* the file descriptor where the output should be sent */
		char		*output;		/* "output" filename (if "socket" then use socket code) */
#ifdef VRUIDD_STUFF /* probably delete me */
		int		fd_receive_socket;	/* socket to communicate with client */
#endif
		int		socket_port;		/* port on server listening for application connections */
		int		open;			/* flag that Stream successfully open */
		int		use_socket;		/* flag that indicates whether to use socket or stdio */

		int	 	format;			/* The format (aka protocol) of the data to send: from OutputFormat enumeration (e.g. CSV, Monado, etc) */

		/* these are for internal data parsing/constructing */
		int		packet_size;		/* the number of bytes in each packet reported by the Daemon */
		unsigned char	buf[BUFSIZE];		/* buffer for sending data to the Vrui Client */
		int		eobuf_pos;		/* position of the last active byte in the buffer */

		/* TODO: most of these fields were part of vr_input.shmemd.c, I suspect most can be deleted */

		/* these are pointers to the FreeVR inputs */
		int		num_buttons;
#if 0 /* saving these for a time when I might do output-mapping */
		vr2switch	**button_inputs;
		int		*button_values;		/* array of current values */
		int		*button_offsets;	/* array of offsets into the Vrui memory structure */
		int		*button_dummy;		/* array of flags whether button data coming from dummy values */
#endif

		int		num_valuators;
#if 1 /* saving these for a time when I might do output-mapping */
		vrValuator	**valuator_inputs;
		float		*valuator_values;	/* array of current values */
		int		*valuator_offsets;	/* array of offsets into the Vrui memory structure */
		int		*valuator_dummy;	/* array of flags whether valuator data coming from dummy values */
#endif

		int		num_6sensors;
#if 1 /* saving these for a time when I might do output-mapping */
		vr6sensor	**sensor6_inputs;
		vrMatrix	*sensor6_values;	/* array of current values */
		int		*sensor6_offsets;	/* array of offsets into the Vrui memory structure */
		int		*sensor6_dummy;		/* array of flags whether 6sensor data coming from dummy values */
#endif

#if 0 /* this part certainly is not required for an "output" device (delete me) */
		/* these are for storing dummy values when too many inputs are requested */
		int		dummy_button_value;
		int		dummy_switch_value;
		float		dummy_valuator_value;
		vrMatrix	dummy_6sensor_value;
#endif
	} _StreamPrivateInfo;



	/*********************************************/
	/*** General NON public interface routines ***/
	/*********************************************/

/**************************************************************************/
static void _StreamInitializeStruct(_StreamPrivateInfo *aux)
{
	aux->fd_output = STDOUT_FILENO;		/* default output is standard-out (stdout) -- only used when given bad arguments */
	aux->output = "/dev/null";		/* Send the data nowhere by default */
	aux->socket_port = DEFAULT_PORT;

	aux->format = OUTPUT_CSV;		/* default output format (other option is OUTPUT_MONADO1|2) */
	//aux->format = OUTPUT_MONADO2;		/* default output format (other option is OUTPUT_CSV) */

#ifdef VRUIDD_STUFF /* probably delete me */
	aux->fd_receive_socket = -1;		/* i.e. indicate that it's not yet initialized */
#endif
}


/**************************************************************************/
static void _StreamPrintPrivateInfo(FILE *file, _StreamPrivateInfo *aux, vrPrintStyle style)
{
	int	count;

	switch (style) {
		case brief:

		vrFprintf(file, "stream-output: destination='%s', format=%d, port=%d, open=%d, fd_output=%d\n",
			aux->output, aux->format, aux->socket_port, aux->open, aux->fd_output);
#if 0
	char msg[1024];
	snprintf(mesg, ,sizeof(msg), "stream-output: destination='%s', format=%d, open=%d, fd_output=%d\n",
		aux->output, aux->format, aux->open, aux->fd_output);
	vrSystemSetStatusDescription(msg);
#endif
		break;

		case verbose:
		default:
		vrFprintf(file, "Stream output-device internal structure (%#p):\n", aux);
		vrFprintf(file, "\r\tStream open = %d\n", aux->open);
		vrFprintf(file, "\r\tStream output format = %d\n", aux->format);
		vrFprintf(file, "\r\tStream output file descriptor = %d\n", aux->fd_output);
		vrFprintf(file, "\r\tStream output destination = %s\n", aux->output);
		vrFprintf(file, "\r\tStream output socket port = %d\n", aux->socket_port);

		vrFprintf(file, "\r\t%d buttons:\n", aux->num_buttons);
#if 0 /* Note that no inputs actually exist for the stream output I/O device */ /* So could delete me */
		vrFprintf(file, "\r\tbutton_inputs = %#p, button_values = %#p, button_offsets = %#p\n",
			aux->button_inputs, aux->button_values, aux->button_offsets);
		for (count = 0; count < aux->num_buttons; count++) {
			vrFprintf(file, "\r\t\tbutton %d: value = %d, offset = %d, input = %#p (type = %d)\n",
				count,
				aux->button_values[count],
				aux->button_offsets[count],
				aux->button_inputs[count],
				aux->button_inputs[count]->input_type);
		}

		vrFprintf(file, "\r\t%d valuators:\n", aux->num_valuators);
		vrFprintf(file, "\r\tvaluator_inputs = %#p, valuator_values = %#p, valuator_offsets = %#p\n",
			aux->valuator_inputs, aux->valuator_values, aux->valuator_offsets);
		for (count = 0; count < aux->num_valuators; count++) {
			vrFprintf(file, "\r\t\tvaluator %d: value = %.2f, offset = %d, input = %#p (type = %d)\n",
				count,
				aux->valuator_values[count],
				aux->valuator_offsets[count],
				aux->valuator_inputs[count],
				aux->valuator_inputs[count]->input_type);
		}

		vrFprintf(file, "\r\t%d 6-sensors:\n", aux->num_6sensors);
		vrFprintf(file, "\r\t6sensor_inputs = %#p, 6sensor_values = %#p, 6sensor_offsets = %#p\n",
			aux->sensor6_inputs, aux->sensor6_values, aux->sensor6_offsets);
		for (count = 0; count < aux->num_6sensors; count++) {
			vrFprintf(file, "\r\t\t6-sensor %d: value = [... (TODO!)], offset = %d, input = %#p (type = %d)\n",
				count,
			/* TODO: put matrix values here  -- perhaps convert to quat/trans */
				aux->sensor6_offsets[count],
				aux->sensor6_inputs[count],
				aux->sensor6_inputs[count]->input_type);
		}
#endif
	}
}


/**************************************************************************/
static void _StreamPrintHelp(FILE *file, _StreamPrivateInfo *aux)
{
	vrFprintf(file, BOLD_TEXT "Sorry, Stream -- print_help control not yet implemented.\n" NORM_TEXT);
}


/**************************************************************************/
static void _StreamPrintValues(FILE *file, _StreamPrivateInfo *aux)
{
	int	count;

	vrFprintf(file, "Stream output device inputs to send:\n");

	vrFprintf(file, "\r\t%d buttons\n", aux->num_buttons);
#if 0 /* 02/29/24: WRS: This is an "output" device, so it doesn't actually have it's own buttons */
	for (count = 0; count < aux->num_buttons; count++) {
		if (aux->button_inputs[count]->input_type == VRINPUT_BINARY) {
			vrFprintf(file, "\r\t\tbutton %d ('%s') [offset = %d]: = %d\n",
				count,
				aux->devinfo->switch2[count].my_object->name,
				aux->button_offsets[count],
				aux->button_values[count]);
		} else {
			vrFprintf(file, "\r\t\tbutton %d ('%s') [offset = %d]: not reported as VRINPUT_BINARY (%d instead)\n",
				count,
				aux->devinfo->switch2[count].my_object->name,
				aux->button_offsets[count],
				aux->button_inputs[count]->input_type);
		}
	}
#endif

	vrFprintf(file, "\r\t%d valuators\n", aux->num_valuators);
#if 0 /* 02/29/24: WRS: This is an "output" device, so it doesn't actually have it's own valuators */
	for (count = 0; count < aux->num_valuators; count++) {
		if (aux->valuator_inputs[count]->input_type == VRINPUT_VALUATOR) {
			vrFprintf(file, "\r\t\tvaluator %d ('%s') [offset = %d]: = %.2f\n",
				count,
				aux->devinfo->valuator[count].my_object->name,
				aux->valuator_offsets[count],
				aux->valuator_values[count]);
		} else {
			vrFprintf(file, "\r\t\tvaluator %d ('%s') [offset = %d]: not reported as VRINPUT_VALUATOR (%d instead)\n",
				count,
				aux->devinfo->switch2[count].my_object->name,
				aux->valuator_offsets[count],
				aux->valuator_inputs[count]->input_type);
		}
	}
#endif

	vrFprintf(file, "\r\t%d 6-sensors\n", aux->num_6sensors);
#if 0 /* 02/29/24: WRS: This is an "output" device, so it doesn't actually have it's own 6-sensors */
	for (count = 0; count < aux->num_6sensors; count++) {
		if (aux->sensor6_inputs[count]->input_type == VRINPUT_6SENSOR) {
			/* TODO: print the matrix */
			vrFprintf(file, "\r\t\t6sensor %d ('%s') [offset = %d]: = ...\n",
				count,
				aux->devinfo->sensor6[count].my_object->name,
				aux->sensor6_offsets[count]);
		} else {
			vrFprintf(file, "\r\t\t6sensor %d ('%s') [offset = %d]: not reported as VRINPUT_6SENSOR (%d instead)\n",
				count,
				aux->devinfo->switch2[count].my_object->name,
				aux->sensor6_offsets[count],
				aux->sensor6_inputs[count]->input_type);
		}
	}
#endif
}


	/************************************/
	/***  FreeVR NON public routines  ***/
	/************************************/


/***************************************************************/
static void _StreamParseArgs(_StreamPrivateInfo *aux, char *args)
{
static	char	*protocol_choices[] = { "csv", "monado1", "monado2", "monado3", NULL };
static	int	protocol_values[] = { OUTPUT_CSV, OUTPUT_MONADO1, OUTPUT_MONADO2, OUTPUT_MONADO3 };

	/* In the rare case of no arguments, just return */
	if (args == NULL)
		return;

	/****************************************/
	/** Argument format: "port" "=" number **/
	/****************************************/
	vrArgParseInteger(args, "port", &(aux->socket_port));

	/******************************************/
	/** Argument format: "output" "=" <file> **/
	/******************************************/
	vrArgParseString(args, "output", &(aux->output));

	/***********************************************************************/
	/** Argument format: "protocol" "=" { "csv" | "monado1" | "monado2" | "monado3" } **/
	/***********************************************************************/
	vrArgParseChoiceInteger(args, "protocol", &(aux->format), protocol_choices, protocol_values);


	/* it would be reasonable to add some arguments for setting the dummy values here */
}


/***************************************************************/
/* _StreamPutData(): ... */
/* TODO: Consider having a sub-function for each of the formats (protocols). */
/* TODO: probably should use a "Direct()" form of the vrGet<input>()  */
/*   functions so as not to mess with any "Delta()" forms.  Though    */
/*   this would be somewhat pedantic, as we really shouldn't be doing */
/*   output stuff when running a real application -- but I like to be */
/*   pedantic, it can save from future bugs.                          */
static void _StreamPutData(vrInputDevice *devinfo)
{
static	int			initialized_flag = 0;
	_StreamPrivateInfo	*aux = (_StreamPrivateInfo *)devinfo->aux_data;
	int			count;
	int			offset;
	int			return_val;
	char			buffer[BUFSIZE];
	int			buf_pos = 0;

	/* The first time set the values of the outputs -- TODO: this should probably be determined some other way */
	if (initialized_flag == 0 && devinfo->context->input->all_devices_initialized) {
		aux->num_buttons = devinfo->context->input->num_2ways - 1;	/* NOTE: 1 is subtracted because we won't send the Esc key */
		aux->num_valuators = devinfo->context->input->num_valuators;
		aux->num_6sensors = devinfo->context->input->num_6sensors;
		initialized_flag = 1;
	}

	bzero(buffer, BUFSIZE);	/* clear the buffer */

    /* I'm only half indenting the switch/case syntax because I expect */
    /*   to split the stream types into separate sub-functions.        */
    /*******************************************************************/
    switch (aux->format) {
    case OUTPUT_CSV: {
    default:
	/*************/
	/** buttons **/

	/* NOTE: we skip button 0, which is the escape key.         */
	/* Note further: that "aux->num_buttons" already is reduced */
	/*   by 1 to account for the lack of ESC key.               */

	for (count = 0; count < aux->num_buttons; count++) {
		return_val = snprintf(&buffer[buf_pos], (BUFSIZE-buf_pos), "%1d,", vrGet2switchValue(count+1));
		buf_pos += return_val;
	}

	/***************/
	/** valuators **/
	for (count = 0; count < aux->num_valuators; count++) {
		return_val = snprintf(&buffer[buf_pos], (BUFSIZE-buf_pos), "%f,", vrGetValuatorValue(count+1));
		buf_pos += return_val;
	}

	for (count = 0; count < aux->num_6sensors; count++) {
		vrMatrix	sensor_mat;		/* a transitional place to get the sensor data */
		vrQuat		sensor_quat;		/* a place to put the quaternion version of the rotation */

		/* extract the actual matrix for this sensor */
		vrMatrixGet6sensorValues(&sensor_mat, count);

		/* modify the matrix based on configuration parameters     */
		/* NOTE: typically this would be the identity operation,   */
		/*   but for some special cases (namely ParaView), it's    */
		/*   beneficial to be able to move the coordinate system.  */
		/* NOTE: this operation uses the "t2rw_xform" field,       */
		/*   but when specified in a configuration file, the       */
		/*   "frame-of-reference" synonym will generally be used   */
		/*   instead (e.g. "for_translate").                       */
		vrMatrixPreMult(&sensor_mat, devinfo->t2rw_xform);

		/* extract the rotation from the new matrix as a quaternion */
		vrQuatSetFromMatrix(&sensor_quat, &sensor_mat);

		return_val = snprintf(&buffer[buf_pos], (BUFSIZE-buf_pos), "%f,%f,%f,",
			VRMAT_ROWCOL(&sensor_mat, VR_X, VR_W),
			VRMAT_ROWCOL(&sensor_mat, VR_Y, VR_W),
			VRMAT_ROWCOL(&sensor_mat, VR_Z, VR_W));
		buf_pos += return_val;
	}

	/* terminate the output */
	return_val = snprintf(&buffer[buf_pos], (BUFSIZE-buf_pos), "\n");
	buf_pos += return_val;
    }
    break;

    /*********************************************************************/
    case OUTPUT_MONADO1: {
	/* Monado binary format #1 (designed around the Google Daydream) */

	vrMatrix	sensor_mat;		/* a transitional place to get the sensor data */
	vrQuat		sensor_quat;		/* a place to put the quaternion version of the rotation */
	vrQuatf		monado_quat;		/* Monado works in floats */
	vrVectorf	monado_loc;		/* Monado works in floats */
	/************/
	/** header **/
	// MONADO -- 8 bytes: 'mndrmt1\0'
	return_val = snprintf(&buffer[buf_pos], (BUFSIZE-buf_pos), "mndrmt1\0");
	buf_pos += return_val;
	buf_pos++;	/* One extra for the '\0' character */

	/***************/
	/** 6-sensors **/

	vrMatrixGet6sensorValues(&sensor_mat, 0/* the head */);
	vrMatrixPreMult(&sensor_mat, devinfo->t2rw_xform);
	vrQuatSetFromMatrix(&sensor_quat, &sensor_mat);

	monado_quat.v[VR_X] = sensor_quat.v[VR_X];
	monado_quat.v[VR_Y] = sensor_quat.v[VR_Y];
	monado_quat.v[VR_Z] = sensor_quat.v[VR_Z];
	monado_quat.v[VR_W] = sensor_quat.v[VR_W];

	buf_pos = 8;		/* HMD pose starts at byte 9 */
	*((float *)&(buffer[buf_pos +  0])) =  monado_quat.v[VR_X];
	*((float *)&(buffer[buf_pos +  4])) =  monado_quat.v[VR_Y];
	*((float *)&(buffer[buf_pos +  8])) =  monado_quat.v[VR_Z];
	*((float *)&(buffer[buf_pos + 12])) =  monado_quat.v[VR_W];
	buf_pos += 4 * sizeof(float);

	monado_loc.v[VR_X] = VRMAT_ROWCOL(&sensor_mat, VR_X, VR_W);
	monado_loc.v[VR_Y] = VRMAT_ROWCOL(&sensor_mat, VR_Y, VR_W);
	monado_loc.v[VR_Z] = VRMAT_ROWCOL(&sensor_mat, VR_Z, VR_W);
	*((float *)&(buffer[buf_pos +  0])) =  monado_loc.v[VR_X];
	*((float *)&(buffer[buf_pos +  4])) =  monado_loc.v[VR_Y];
	*((float *)&(buffer[buf_pos +  8])) =  monado_loc.v[VR_Z];
	buf_pos += 3 * sizeof(float);

	/*** Set Quaternions & Position for the controllers ***/

	/* Skip to the left controller */
	buf_pos = 36;		/* Left Controller starts at byte 37 */
	if (aux->num_6sensors > 2) {
		vrMatrixGet6sensorValues(&sensor_mat, 2/* the left controller */);
		vrMatrixPreMult(&sensor_mat, devinfo->t2rw_xform);
		vrQuatSetFromMatrix(&sensor_quat, &sensor_mat);
		monado_quat.v[VR_X] = sensor_quat.v[VR_X];
		monado_quat.v[VR_Y] = sensor_quat.v[VR_Y];
		monado_quat.v[VR_Z] = sensor_quat.v[VR_Z];
		monado_quat.v[VR_W] = sensor_quat.v[VR_W];
		monado_loc.v[VR_X] = VRMAT_ROWCOL(&sensor_mat, VR_X, VR_W);
		monado_loc.v[VR_Y] = VRMAT_ROWCOL(&sensor_mat, VR_Y, VR_W);
		monado_loc.v[VR_Z] = VRMAT_ROWCOL(&sensor_mat, VR_Z, VR_W);
	} else {
		/* if there is no sensor output generic values */
		monado_quat.v[VR_X] = 0.0;
		monado_quat.v[VR_Y] = 0.0;
		monado_quat.v[VR_Z] = 0.0;
		monado_quat.v[VR_W] = 1.0;
		monado_loc.v[VR_X] = 0.0;
		monado_loc.v[VR_Y] = 0.0;
		monado_loc.v[VR_Z] = 0.0;
	}
	*((float *)&(buffer[buf_pos +  0])) =  monado_quat.v[VR_X];
	*((float *)&(buffer[buf_pos +  4])) =  monado_quat.v[VR_Y];
	*((float *)&(buffer[buf_pos +  8])) =  monado_quat.v[VR_Z];
	*((float *)&(buffer[buf_pos + 12])) =  monado_quat.v[VR_W];
	buf_pos += 4 * sizeof(float);
	*((float *)&(buffer[buf_pos +  0])) =  monado_loc.v[VR_X];
	*((float *)&(buffer[buf_pos +  4])) =  monado_loc.v[VR_Y];
	*((float *)&(buffer[buf_pos +  8])) =  monado_loc.v[VR_Z];

	/* Skip to the right controller */
	buf_pos = 116;		/* Right Controller starts at byte 117 */
	if (aux->num_6sensors > 1) {
		vrMatrixGet6sensorValues(&sensor_mat, 1/* the right controller */);
		vrMatrixPreMult(&sensor_mat, devinfo->t2rw_xform);
		vrQuatSetFromMatrix(&sensor_quat, &sensor_mat);
		monado_quat.v[VR_X] = sensor_quat.v[VR_X];
		monado_quat.v[VR_Y] = sensor_quat.v[VR_Y];
		monado_quat.v[VR_Z] = sensor_quat.v[VR_Z];
		monado_quat.v[VR_W] = sensor_quat.v[VR_W];
		monado_loc.v[VR_X] = VRMAT_ROWCOL(&sensor_mat, VR_X, VR_W);
		monado_loc.v[VR_Y] = VRMAT_ROWCOL(&sensor_mat, VR_Y, VR_W);
		monado_loc.v[VR_Z] = VRMAT_ROWCOL(&sensor_mat, VR_Z, VR_W);
	} else {
		/* if there is no sensor output generic values */
		monado_quat.v[VR_X] = 0.0;
		monado_quat.v[VR_Y] = 0.0;
		monado_quat.v[VR_Z] = 0.0;
		monado_quat.v[VR_W] = 1.0;
		monado_loc.v[VR_X] = 0.0;
		monado_loc.v[VR_Y] = 0.0;
		monado_loc.v[VR_Z] = 0.0;
	}
	*((float *)&(buffer[buf_pos +  0])) =  monado_quat.v[VR_X];
	*((float *)&(buffer[buf_pos +  4])) =  monado_quat.v[VR_Y];
	*((float *)&(buffer[buf_pos +  8])) =  monado_quat.v[VR_Z];
	*((float *)&(buffer[buf_pos + 12])) =  monado_quat.v[VR_W];
	buf_pos += 4 * sizeof(float);
	*((float *)&(buffer[buf_pos +  0])) =  monado_loc.v[VR_X];
	*((float *)&(buffer[buf_pos +  4])) =  monado_loc.v[VR_Y];
	*((float *)&(buffer[buf_pos +  8])) =  monado_loc.v[VR_Z];
	buf_pos += 3 * sizeof(float);
	buf_pos += 52;


	/***************/
	/** valuators **/

	/* there are no valuators in this format/protocol */


	/*************/
	/** buttons **/

	/* Set some fixed booleans */
	buffer[108] = 0;	/* Left controller "hand_tracking_active" */
	buffer[109] = 1;	/* Left controller "active" (ie. there's a left controller) */
	buffer[188] = 0;	/* Right controller "hand_tracking_active" */
	buffer[189] = 1;	/* Right controller "active" (ie. there's a right controller) */

	/* Set some button booleans */
	buffer[110] = vrGet2switchValue(1);	/* Left controller "select" */
	buffer[111] = vrGet2switchValue(2);	/* Left controller "menu" */
	buffer[190] = vrGet2switchValue(3);	/* Right controller "select" */
	buffer[191] = vrGet2switchValue(4);	/* Right controller "menu" */

	buf_pos += 4;	/* padding to 8-byte boundary */
    }
    break;

    /*********************************************************************/
    case OUTPUT_MONADO2: {
	/* Monado binary format #2 (designed around the Valve Index) */

	vrMatrix	sensor_mat;		/* a transitional place to get the sensor data */
	vrQuat		sensor_quat;		/* a place to put the quaternion version of the rotation */
	vrQuatf		monado_quat;		/* Monado works in floats */
	vrVectorf	monado_loc;		/* Monado works in floats */

	/************/
	/** header **/
	// MONADO -- 8 bytes: 'mndrmt2\0'
	return_val = snprintf(&buffer[buf_pos], (BUFSIZE-buf_pos), "mndrmt2\0");
	buf_pos += return_val;
	buf_pos++;	/* One extra for the '\0' character */

	/***************/
	/** 6-sensors **/

	vrMatrixGet6sensorValues(&sensor_mat, 0/* the head */);
	vrMatrixPreMult(&sensor_mat, devinfo->t2rw_xform);
	vrQuatSetFromMatrix(&sensor_quat, &sensor_mat);

	monado_quat.v[VR_X] = sensor_quat.v[VR_X];
	monado_quat.v[VR_Y] = sensor_quat.v[VR_Y];
	monado_quat.v[VR_Z] = sensor_quat.v[VR_Z];
	monado_quat.v[VR_W] = sensor_quat.v[VR_W];

	buf_pos = 8;		/* HMD pose starts at byte 9 */
	*((float *)&(buffer[buf_pos +  0])) =  monado_quat.v[VR_X];
	*((float *)&(buffer[buf_pos +  4])) =  monado_quat.v[VR_Y];
	*((float *)&(buffer[buf_pos +  8])) =  monado_quat.v[VR_Z];
	*((float *)&(buffer[buf_pos + 12])) =  monado_quat.v[VR_W];
	buf_pos += 4 * sizeof(float);

	monado_loc.v[VR_X] = VRMAT_ROWCOL(&sensor_mat, VR_X, VR_W);
	monado_loc.v[VR_Y] = VRMAT_ROWCOL(&sensor_mat, VR_Y, VR_W);
	monado_loc.v[VR_Z] = VRMAT_ROWCOL(&sensor_mat, VR_Z, VR_W);
	*((float *)&(buffer[buf_pos +  0])) =  monado_loc.v[VR_X];
	*((float *)&(buffer[buf_pos +  4])) =  monado_loc.v[VR_Y];
	*((float *)&(buffer[buf_pos +  8])) =  monado_loc.v[VR_Z];
	buf_pos += 3 * sizeof(float);

	/*** Set Quaternions & Position for the controllers ***/

	/* Skip to the left controller */
	buf_pos = 36;		/* Left Controller starts at byte 37 */
	/*   determine the controller position tracking data */
	if (aux->num_6sensors > 2) {
		vrMatrixGet6sensorValues(&sensor_mat, 2/* the left controller */);
		vrMatrixPreMult(&sensor_mat, devinfo->t2rw_xform);
		vrQuatSetFromMatrix(&sensor_quat, &sensor_mat);
		monado_quat.v[VR_X] = sensor_quat.v[VR_X];
		monado_quat.v[VR_Y] = sensor_quat.v[VR_Y];
		monado_quat.v[VR_Z] = sensor_quat.v[VR_Z];
		monado_quat.v[VR_W] = sensor_quat.v[VR_W];
		monado_loc.v[VR_X] = VRMAT_ROWCOL(&sensor_mat, VR_X, VR_W);
		monado_loc.v[VR_Y] = VRMAT_ROWCOL(&sensor_mat, VR_Y, VR_W);
		monado_loc.v[VR_Z] = VRMAT_ROWCOL(&sensor_mat, VR_Z, VR_W);
	} else {
		/* if there is no sensor output generic values */
		monado_quat.v[VR_X] = 0.0;
		monado_quat.v[VR_Y] = 0.0;
		monado_quat.v[VR_Z] = 0.0;
		monado_quat.v[VR_W] = 1.0;
		monado_loc.v[VR_X] = 0.0;
		monado_loc.v[VR_Y] = 0.0;
		monado_loc.v[VR_Z] = 0.0;
	}
	/*   output the controller position tracking data */
	*((float *)&(buffer[buf_pos +  0])) =  monado_quat.v[VR_X];
	*((float *)&(buffer[buf_pos +  4])) =  monado_quat.v[VR_Y];
	*((float *)&(buffer[buf_pos +  8])) =  monado_quat.v[VR_Z];
	*((float *)&(buffer[buf_pos + 12])) =  monado_quat.v[VR_W];
	buf_pos += 4 * sizeof(float);
	*((float *)&(buffer[buf_pos +  0])) =  monado_loc.v[VR_X];
	*((float *)&(buffer[buf_pos +  4])) =  monado_loc.v[VR_Y];
	*((float *)&(buffer[buf_pos +  8])) =  monado_loc.v[VR_Z];

	/* Skip to the right controller */
	buf_pos = 156;		/* Right Controller starts at byte 157 */
	/*   determine the controller position tracking data */
	if (aux->num_6sensors > 1) {
		vrMatrixGet6sensorValues(&sensor_mat, 1/* the right controller */);
		vrMatrixPreMult(&sensor_mat, devinfo->t2rw_xform);
		vrQuatSetFromMatrix(&sensor_quat, &sensor_mat);
		monado_quat.v[VR_X] = sensor_quat.v[VR_X];
		monado_quat.v[VR_Y] = sensor_quat.v[VR_Y];
		monado_quat.v[VR_Z] = sensor_quat.v[VR_Z];
		monado_quat.v[VR_W] = sensor_quat.v[VR_W];
		monado_loc.v[VR_X] = VRMAT_ROWCOL(&sensor_mat, VR_X, VR_W);
		monado_loc.v[VR_Y] = VRMAT_ROWCOL(&sensor_mat, VR_Y, VR_W);
		monado_loc.v[VR_Z] = VRMAT_ROWCOL(&sensor_mat, VR_Z, VR_W);
	} else {
		/* if there is no sensor output generic values */
		monado_quat.v[VR_X] = 0.0;
		monado_quat.v[VR_Y] = 0.0;
		monado_quat.v[VR_Z] = 0.0;
		monado_quat.v[VR_W] = 1.0;
		monado_loc.v[VR_X] = 0.0;
		monado_loc.v[VR_Y] = 0.0;
		monado_loc.v[VR_Z] = 0.0;
	}

	/*   output the controller position tracking data */
	*((float *)&(buffer[buf_pos +  0])) =  monado_quat.v[VR_X];
	*((float *)&(buffer[buf_pos +  4])) =  monado_quat.v[VR_Y];
	*((float *)&(buffer[buf_pos +  8])) =  monado_quat.v[VR_Z];
	*((float *)&(buffer[buf_pos + 12])) =  monado_quat.v[VR_W];
	buf_pos += 4 * sizeof(float);
	*((float *)&(buffer[buf_pos +  0])) =  monado_loc.v[VR_X];
	*((float *)&(buffer[buf_pos +  4])) =  monado_loc.v[VR_Y];
	*((float *)&(buffer[buf_pos +  8])) =  monado_loc.v[VR_Z];
	buf_pos += 3 * sizeof(float);


	/*************/
	/** buttons **/

	/* Set some fixed booleans */
	buffer[140] = 0;	/* Left controller "hand_tracking_active" */
	buffer[141] = 1;	/* Left controller "active" (ie. there's a left controller) */
	buffer[260] = 0;	/* Right controller "hand_tracking_active" */
	buffer[261] = 1;	/* Right controller "active" (ie. there's a right controller) */

	/* Set some button booleans */
	buffer[142] = vrGet2switchValue(1);	/* Left controller "system_click" */
	buffer[143] = 0;//vrGet2switchValue(2);	/* Left controller "system_touch" */
	buffer[144] = vrGet2switchValue(2);	/* Left controller "a_click" */
	buffer[145] = 0;//vrGet2switchValue(2);	/* Left controller "a_touch" */
	buffer[146] = vrGet2switchValue(3);	/* Left controller "b_click" */
	buffer[147] = 0;//vrGet2switchValue(2);	/* Left controller "b_touch" */
	buffer[148] = vrGet2switchValue(4);	/* Left controller "trigger_click" */
	buffer[149] = 0;//vrGet2switchValue(2);	/* Left controller "trigger_touch" */
	buffer[150] = 0;//vrGet2switchValue(2);	/* Left controller "thumbstick_click" */
	buffer[151] = 0;//vrGet2switchValue(2);	/* Left controller "thumbstick_touch" */
	buffer[152] = 0;//vrGet2switchValue(2);	/* Left controller "trackpad_touch" */

	buffer[262] = 0;//vrGet2switchValue(1);	/* Right controller "system_click" */
	buffer[263] = 0;//vrGet2switchValue(2);	/* Right controller "system_touch" */
	buffer[264] = 0;//vrGet2switchValue(2);	/* Right controller "a_click" */
	buffer[265] = 0;//vrGet2switchValue(2);	/* Right controller "a_touch" */
	buffer[266] = 0;//vrGet2switchValue(2);	/* Right controller "b_click" */
	buffer[267] = 0;//vrGet2switchValue(2);	/* Right controller "b_touch" */
	buffer[268] = 0;//vrGet2switchValue(2);	/* Right controller "trigger_click" */
	buffer[269] = 0;//vrGet2switchValue(2);	/* Right controller "trigger_touch" */
	buffer[270] = 0;//vrGet2switchValue(2);	/* Right controller "thumbstick_click" */
	buffer[271] = 0;//vrGet2switchValue(2);	/* Right controller "thumbstick_touch" */
	buffer[272] = 0;//vrGet2switchValue(2);	/* Right controller "trackpad_touch" */

	/***************/
	/** valuators **/

	/* Set some valuators values */
	*((float *)&(buffer[108])) = 0.0;//vrGetValuatorValue(0);	/* Left cont: trigger */
	*((float *)&(buffer[112])) = 0.0;//vrGetValuatorValue(0);	/* Left cont: squeeze */
	*((float *)&(buffer[116])) = 0.0;//vrGetValuatorValue(0);	/* Left cont: squeeze-force */
	*((float *)&(buffer[120])) = vrGetValuatorValue(0);		/* Left cont: thumbstick-x */
	*((float *)&(buffer[124])) = vrGetValuatorValue(1);		/* Left cont: thumbstick-y */
	*((float *)&(buffer[128])) = 0.0;//vrGetValuatorValue(0);	/* Left cont: trackpack-force */
	*((float *)&(buffer[132])) = 0.0;//vrGetValuatorValue(0);	/* Left cont: trackpack-x */
	*((float *)&(buffer[136])) = 0.0;//vrGetValuatorValue(0);	/* Left cont: trackpack-y */

	*((float *)&(buffer[228])) = 0.0;//vrGetValuatorValue(0);	/* Right cont: trigger */
	*((float *)&(buffer[232])) = 0.0;//vrGetValuatorValue(0);	/* Right cont: squeeze */
	*((float *)&(buffer[236])) = 0.0;//vrGetValuatorValue(0);	/* Right cont: squeeze-force */
	*((float *)&(buffer[240])) = 0.0;//vrGetValuatorValue(0);	/* Right cont: thumbstick-x */
	*((float *)&(buffer[244])) = 0.0;//vrGetValuatorValue(0);	/* Right cont: thumbstick-y */
	*((float *)&(buffer[248])) = 0.0;//vrGetValuatorValue(0);	/* Right cont: trackpack-force */
	*((float *)&(buffer[252])) = 0.0;//vrGetValuatorValue(0);	/* Right cont: trackpack-x */
	*((float *)&(buffer[256])) = 0.0;//vrGetValuatorValue(0);	/* Right cont: trackpack-y */
	buf_pos = 280;				/* This is the Monado2 packet size (including 4 bytes of padding) */
    }
    /*********************************************************************/
    case OUTPUT_MONADO3: {
	/* Monado binary format #2 (designed around the Valve Index) */

	vrMatrix	sensor_mat;		/* a transitional place to get the sensor data */
	vrQuat		sensor_quat;		/* a place to put the quaternion version of the rotation */
	vrQuatf		monado_quat;		/* Monado works in floats */
	vrVectorf	monado_loc;		/* Monado works in floats */


	/**************/
	/** views[2] **/
	/* NOTE: the monado remote protocol-3 has an array of 2 "views", and */
	/*   a boolean to flag whether they are used instead of the "center" */
	/*   of the HMD value.  For now, I'm going to leave all these values */
	/*   as 0, and the bool value as 0 (false), which means the rest of  */
	/*   the protocol is basically the same other than the offsets.      */
	/*   (This adds 48*2 + 4 padding bytes = 100 bytes to the protocol.) */

	/************/
	/** header **/
	// MONADO -- 8 bytes: 'mndrmt3\0'
	return_val = snprintf(&buffer[buf_pos], (BUFSIZE-buf_pos), "mndrmt3\0");
	buf_pos += return_val;
	buf_pos++;	/* One extra for the '\0' character */

	/***************/
	/** 6-sensors **/

	vrMatrixGet6sensorValues(&sensor_mat, 0/* the head */);
	vrMatrixPreMult(&sensor_mat, devinfo->t2rw_xform);
	vrQuatSetFromMatrix(&sensor_quat, &sensor_mat);

	monado_quat.v[VR_X] = sensor_quat.v[VR_X];
	monado_quat.v[VR_Y] = sensor_quat.v[VR_Y];
	monado_quat.v[VR_Z] = sensor_quat.v[VR_Z];
	monado_quat.v[VR_W] = sensor_quat.v[VR_W];

	buf_pos = 104;		/* HMD "center" pose starts at byte 105 */
	*((float *)&(buffer[buf_pos +  0])) =  monado_quat.v[VR_X];
	*((float *)&(buffer[buf_pos +  4])) =  monado_quat.v[VR_Y];
	*((float *)&(buffer[buf_pos +  8])) =  monado_quat.v[VR_Z];
	*((float *)&(buffer[buf_pos + 12])) =  monado_quat.v[VR_W];
	buf_pos += 4 * sizeof(float);

	monado_loc.v[VR_X] = VRMAT_ROWCOL(&sensor_mat, VR_X, VR_W);
	monado_loc.v[VR_Y] = VRMAT_ROWCOL(&sensor_mat, VR_Y, VR_W);
	monado_loc.v[VR_Z] = VRMAT_ROWCOL(&sensor_mat, VR_Z, VR_W);
	*((float *)&(buffer[buf_pos +  0])) =  monado_loc.v[VR_X];
	*((float *)&(buffer[buf_pos +  4])) =  monado_loc.v[VR_Y];
	*((float *)&(buffer[buf_pos +  8])) =  monado_loc.v[VR_Z];
	buf_pos += 3 * sizeof(float);

	buffer[buf_pos] = 0;	/* per_view_data_valid: whether to use the "views[2]" or "center" HMD pose(s) */

#if 1 /* also set the view[].pose values for now */
	buf_pos = 24;		/* view[0].pose */
	*((float *)&(buffer[buf_pos +  0])) =  monado_quat.v[VR_X];
	*((float *)&(buffer[buf_pos +  4])) =  monado_quat.v[VR_Y];
	*((float *)&(buffer[buf_pos +  8])) =  monado_quat.v[VR_Z];
	*((float *)&(buffer[buf_pos + 12])) =  monado_quat.v[VR_W];

	buf_pos = 72;		/* view[1].pose */
	*((float *)&(buffer[buf_pos +  0])) =  monado_quat.v[VR_X];
	*((float *)&(buffer[buf_pos +  4])) =  monado_quat.v[VR_Y];
	*((float *)&(buffer[buf_pos +  8])) =  monado_quat.v[VR_Z];
	*((float *)&(buffer[buf_pos + 12])) =  monado_quat.v[VR_W];
#endif

	/*** Set Quaternions & Position for the controllers ***/

	/* Skip to the left controller */
	buf_pos = 136;		/* Left Controller starts at byte 137 */
	/*   determine the controller position tracking data */
	if (aux->num_6sensors > 2) {
		vrMatrixGet6sensorValues(&sensor_mat, 2/* the left controller */);
		vrMatrixPreMult(&sensor_mat, devinfo->t2rw_xform);
		vrQuatSetFromMatrix(&sensor_quat, &sensor_mat);
		monado_quat.v[VR_X] = sensor_quat.v[VR_X];
		monado_quat.v[VR_Y] = sensor_quat.v[VR_Y];
		monado_quat.v[VR_Z] = sensor_quat.v[VR_Z];
		monado_quat.v[VR_W] = sensor_quat.v[VR_W];
		monado_loc.v[VR_X] = VRMAT_ROWCOL(&sensor_mat, VR_X, VR_W);
		monado_loc.v[VR_Y] = VRMAT_ROWCOL(&sensor_mat, VR_Y, VR_W);
		monado_loc.v[VR_Z] = VRMAT_ROWCOL(&sensor_mat, VR_Z, VR_W);
	} else {
		/* if there is no sensor output generic values */
		monado_quat.v[VR_X] = 0.0;
		monado_quat.v[VR_Y] = 0.0;
		monado_quat.v[VR_Z] = 0.0;
		monado_quat.v[VR_W] = 1.0;
		monado_loc.v[VR_X] = 0.0;
		monado_loc.v[VR_Y] = 0.0;
		monado_loc.v[VR_Z] = 0.0;
	}
	/*   output the controller position tracking data */
	*((float *)&(buffer[buf_pos +  0])) =  monado_quat.v[VR_X];
	*((float *)&(buffer[buf_pos +  4])) =  monado_quat.v[VR_Y];
	*((float *)&(buffer[buf_pos +  8])) =  monado_quat.v[VR_Z];
	*((float *)&(buffer[buf_pos + 12])) =  monado_quat.v[VR_W];
	buf_pos += 4 * sizeof(float);
	*((float *)&(buffer[buf_pos +  0])) =  monado_loc.v[VR_X];
	*((float *)&(buffer[buf_pos +  4])) =  monado_loc.v[VR_Y];
	*((float *)&(buffer[buf_pos +  8])) =  monado_loc.v[VR_Z];

	/* Skip to the right controller */
	buf_pos = 256;		/* Right Controller starts at byte 257 */
	/*   determine the controller position tracking data */
	if (aux->num_6sensors > 1) {
		vrMatrixGet6sensorValues(&sensor_mat, 1/* the right controller */);
		vrMatrixPreMult(&sensor_mat, devinfo->t2rw_xform);
		vrQuatSetFromMatrix(&sensor_quat, &sensor_mat);
		monado_quat.v[VR_X] = sensor_quat.v[VR_X];
		monado_quat.v[VR_Y] = sensor_quat.v[VR_Y];
		monado_quat.v[VR_Z] = sensor_quat.v[VR_Z];
		monado_quat.v[VR_W] = sensor_quat.v[VR_W];
		monado_loc.v[VR_X] = VRMAT_ROWCOL(&sensor_mat, VR_X, VR_W);
		monado_loc.v[VR_Y] = VRMAT_ROWCOL(&sensor_mat, VR_Y, VR_W);
		monado_loc.v[VR_Z] = VRMAT_ROWCOL(&sensor_mat, VR_Z, VR_W);
	} else {
		/* if there is no sensor output generic values */
		monado_quat.v[VR_X] = 0.0;
		monado_quat.v[VR_Y] = 0.0;
		monado_quat.v[VR_Z] = 0.0;
		monado_quat.v[VR_W] = 1.0;
		monado_loc.v[VR_X] = 0.0;
		monado_loc.v[VR_Y] = 0.0;
		monado_loc.v[VR_Z] = 0.0;
	}

	/*   output the controller position tracking data */
	*((float *)&(buffer[buf_pos +  0])) =  monado_quat.v[VR_X];
	*((float *)&(buffer[buf_pos +  4])) =  monado_quat.v[VR_Y];
	*((float *)&(buffer[buf_pos +  8])) =  monado_quat.v[VR_Z];
	*((float *)&(buffer[buf_pos + 12])) =  monado_quat.v[VR_W];
	buf_pos += 4 * sizeof(float);
	*((float *)&(buffer[buf_pos +  0])) =  monado_loc.v[VR_X];
	*((float *)&(buffer[buf_pos +  4])) =  monado_loc.v[VR_Y];
	*((float *)&(buffer[buf_pos +  8])) =  monado_loc.v[VR_Z];
	buf_pos += 3 * sizeof(float);


	/*************/
	/** buttons **/

	/* Set some fixed booleans */
	buffer[240] = 0;	/* Left controller "hand_tracking_active" */
	buffer[241] = 1;	/* Left controller "active" (ie. there's a left controller) */
	buffer[360] = 0;	/* Right controller "hand_tracking_active" */
	buffer[361] = 1;	/* Right controller "active" (ie. there's a right controller) */

	/* Set some button booleans */
	buffer[242] = vrGet2switchValue(1);	/* Left controller "system_click" */
	buffer[243] = 0;//vrGet2switchValue(2);	/* Left controller "system_touch" */
	buffer[244] = vrGet2switchValue(2);	/* Left controller "a_click" */
	buffer[245] = 0;//vrGet2switchValue(2);	/* Left controller "a_touch" */
	buffer[246] = vrGet2switchValue(3);	/* Left controller "b_click" */
	buffer[247] = 0;//vrGet2switchValue(2);	/* Left controller "b_touch" */
	buffer[248] = vrGet2switchValue(4);	/* Left controller "trigger_click" */
	buffer[249] = 0;//vrGet2switchValue(2);	/* Left controller "trigger_touch" */
	buffer[250] = 0;//vrGet2switchValue(2);	/* Left controller "thumbstick_click" */
	buffer[251] = 0;//vrGet2switchValue(2);	/* Left controller "thumbstick_touch" */
	buffer[252] = 0;//vrGet2switchValue(2);	/* Left controller "trackpad_touch" */

	buffer[362] = 0;//vrGet2switchValue(1);	/* Right controller "system_click" */
	buffer[363] = 0;//vrGet2switchValue(2);	/* Right controller "system_touch" */
	buffer[364] = 0;//vrGet2switchValue(2);	/* Right controller "a_click" */
	buffer[365] = 0;//vrGet2switchValue(2);	/* Right controller "a_touch" */
	buffer[366] = 0;//vrGet2switchValue(2);	/* Right controller "b_click" */
	buffer[367] = 0;//vrGet2switchValue(2);	/* Right controller "b_touch" */
	buffer[368] = 0;//vrGet2switchValue(2);	/* Right controller "trigger_click" */
	buffer[369] = 0;//vrGet2switchValue(2);	/* Right controller "trigger_touch" */
	buffer[370] = 0;//vrGet2switchValue(2);	/* Right controller "thumbstick_click" */
	buffer[371] = 0;//vrGet2switchValue(2);	/* Right controller "thumbstick_touch" */
	buffer[372] = 0;//vrGet2switchValue(2);	/* Right controller "trackpad_touch" */

	/***************/
	/** valuators **/

	/* Set some valuators values */
	*((float *)&(buffer[208])) = 0.0;//vrGetValuatorValue(0);	/* Left cont: trigger */
	*((float *)&(buffer[212])) = 0.0;//vrGetValuatorValue(0);	/* Left cont: squeeze */
	*((float *)&(buffer[216])) = 0.0;//vrGetValuatorValue(0);	/* Left cont: squeeze-force */
	*((float *)&(buffer[220])) = vrGetValuatorValue(0);		/* Left cont: thumbstick-x */
	*((float *)&(buffer[224])) = vrGetValuatorValue(1);		/* Left cont: thumbstick-y */
	*((float *)&(buffer[228])) = 0.0;//vrGetValuatorValue(0);	/* Left cont: trackpack-force */
	*((float *)&(buffer[232])) = 0.0;//vrGetValuatorValue(0);	/* Left cont: trackpack-x */
	*((float *)&(buffer[236])) = 0.0;//vrGetValuatorValue(0);	/* Left cont: trackpack-y */

	*((float *)&(buffer[328])) = 0.0;//vrGetValuatorValue(0);	/* Right cont: trigger */
	*((float *)&(buffer[332])) = 0.0;//vrGetValuatorValue(0);	/* Right cont: squeeze */
	*((float *)&(buffer[336])) = 0.0;//vrGetValuatorValue(0);	/* Right cont: squeeze-force */
	*((float *)&(buffer[340])) = 0.0;//vrGetValuatorValue(0);	/* Right cont: thumbstick-x */
	*((float *)&(buffer[344])) = 0.0;//vrGetValuatorValue(0);	/* Right cont: thumbstick-y */
	*((float *)&(buffer[348])) = 0.0;//vrGetValuatorValue(0);	/* Right cont: trackpack-force */
	*((float *)&(buffer[352])) = 0.0;//vrGetValuatorValue(0);	/* Right cont: trackpack-x */
	*((float *)&(buffer[356])) = 0.0;//vrGetValuatorValue(0);	/* Right cont: trackpack-y */
	buf_pos = 376;				/* This is the Monado3 packet size (including 4 bytes of padding) */
    }
    break;
    }

	/*******************/
	/** Send the data **/
	vrTrace("_StreamPutData():", BOLD_TEXT "About to send the data to client." NORM_TEXT);
	//return_val = write(aux->fd_receive_socket, aux->buf, aux->packet_size);	/* TODO: warning, no effort yet being made to deal with endianness of the order */
	///* return_val = */vrFprintf(stdout, buffer);
	if (initialized_flag) {
		return_val = write(aux->fd_output, buffer, buf_pos);
		//printf("wrote %d bytes to output stream\n", return_val); -- should = buf_pos
		/* TODO: check for error conditions */
	}

	vrTrace("_StreamPutData():", BOLD_TEXT "Sent the data to client." NORM_TEXT);
}


#if 0 /* I don't think I'm going to want this stuff for the "output" function, but I'll leave the code in place for now [04/11/2007] {{{ */

	/*****************************************************************/
	/*   Function(s) for parsing Shared Memory "input" declarations. */
	/*                                                               */
	/*  These _Stream<type>Input() functions are called during the   */
	/*  CREATE phase of the input interface.                         */

/* ----------------------------------------------------------------------- */
/* ----------------------------------------------------------------------- */

/**************************************************************************/
/* type might be "char", "int", or "long" (and in the future "float" or "double") */
/*    parsing:                                                                    */
/*       input "<name>" = "2switch(<type>[<offset>])";                            */
/*    thus:                                                                       */
/*       the device string is unused.                                             */
/*       the type string is <type> (eg. "char", "int", etc)                       */
/*       the instance string is <offset> (eg. 0, 4, 6)                            */
/**************************************************************************/
static vrInputMatch _StreamButtonInput(vrInputDevice *devinfo, vrStreamInput *input, vrInputDTI *dti)
{
	_StreamPrivateInfo	*aux = (_StreamPrivateInfo *)devinfo->aux_data;
	int			button_num;
	int			offset;
	_MemoryType		memtype;

	/* determine current button in our private arrays of button data */
	button_num = aux->num_buttons;
	aux->num_buttons++;

	memtype = _MemoryValue(dti->type);
	offset = vrAtoI(dti->instance);

	vrDbgPrintfN(INPUT_DBGLVL, "_StreamButtonInput(): Assigned button event of button %d, to type '%s'(%d) with offset %d\n",
		button_num, _MemoryTypeName(memtype), (int)memtype, offset);

	if (memtype == MEM_UNKNOWN) {
		vrErrPrintf("_StreamValuatorInput: " RED_TEXT "Warning, type['%s'] did not match any known type\n" NORM_TEXT, dti->type);
		return VRINPUT_NOMATCH;	/* unacceptable input declaration match, perhaps keep searching */
	}

#if 0 /* NOTE: this check isn't possible here because we haven't connected with the memory yet to know the size */
	if (offset 'out of bounds') {
		...
		return VRINPUT_NOMATCH;	/* unacceptable input declaration match, perhaps keep searching */
	}
#endif

	aux->button_inputs[button_num] = (vr2switch *)input;
	aux->button_values[button_num] = 0;
	aux->button_dummy[button_num] = 0;
	aux->button_offsets[button_num] = offset;

	return VRINPUT_MATCH_ABLE;	/* input declaration match */
}


/**************************************************************************/
static vrInputMatch _StreamValuatorInput(vrInputDevice *devinfo, vrStreamInput *input, vrInputDTI *dti)
{
	_StreamPrivateInfo	*aux = (_StreamPrivateInfo *)devinfo->aux_data;
	int			valuator_num;
	int			offset;
	_MemoryType		memtype;

	/* determine current valuator in our private arrays of valuator data */
	valuator_num = aux->num_valuators;
	aux->num_valuators++;

	memtype = _MemoryValue(dti->type);
	offset = vrAtoI(dti->instance);

	vrDbgPrintfN(INPUT_DBGLVL, "_StreamValuatorInput(): Assigned button event of button %d, to type '%s'(%d) with offset %d\n",
		valuator_num, _MemoryTypeName(memtype), (int)memtype, offset);

	if (memtype == MEM_UNKNOWN) {
		vrErrPrintf("_StreamValuatorInput: " RED_TEXT "Warning, type['%s'] did not match any known type\n" NORM_TEXT, dti->type);
		return VRINPUT_NOMATCH;	/* unacceptable input declaration match, perhaps keep searching */
	}

#if 0 /* NOTE: this check isn't possible here because we haven't connected with the memory yet to know the size */
	if (offset 'out of bounds') {
		...
		return VRINPUT_NOMATCH;	/* unacceptable input declaration match, perhaps keep searching */
	}
#endif

	aux->valuator_inputs[valuator_num] = (vrValuator *)input;
	aux->valuator_values[valuator_num] = 0;
	aux->valuator_dummy[valuator_num] = 0.0;
	aux->valuator_offsets[valuator_num] = offset;

	return VRINPUT_MATCH_ABLE;		/* input declaration match */
}


/**************************************************************************/
static vrInputMatch _Stream6sensorInput(vrInputDevice *devinfo, vrStreamInput *input, vrInputDTI *dti)
{
	_StreamPrivateInfo	*aux = (_StreamPrivateInfo *)devinfo->aux_data;
	int			sensor6_num;
	int			offset;
	_MemoryType		memtype;

	/* determine current 6-sensor in our private arrays of 6-sensor data */
	sensor6_num = aux->num_6sensors;
	aux->num_6sensors++;

	memtype = _MemoryValue(dti->type);
	offset = vrAtoI(dti->instance);

	vrDbgPrintfN(INPUT_DBGLVL, "_Stream6sensorInput(): Assigned 6-sensor event of 6-sensor %d, to type '%s'(%d) with offset %d\n",
		sensor6_num, _MemoryTypeName(memtype), (int)memtype, offset);

	if (memtype == MEM_UNKNOWN) {
		vrErrPrintf("_StreamValuatorInput: " RED_TEXT "Warning, type['%s'] did not match any known type\n" NORM_TEXT, dti->type);
		return VRINPUT_NOMATCH;	/* unacceptable input declaration match, perhaps keep searching */
	}

#if 0 /* NOTE: this check isn't possible here because we haven't connected with the memory yet to know the size */
	if (offset 'out of bounds') {
		...
		return VRINPUT_NOMATCH;	/* unacceptable input declaration match, perhaps keep searching */
	}
#endif

	aux->sensor6_inputs[sensor6_num] = (vr6sensor *)input;
	vrMatrixSetIdentity(&aux->sensor6_values[sensor6_num]);
	aux->sensor6_dummy[sensor6_num] = 0;
	aux->sensor6_offsets[sensor6_num] = offset;

	vrAssign6sensorR2Exform(aux->sensor6_inputs[sensor6_num], strchr(dti->instance, ','));

	return VRINPUT_MATCH_ABLE;	/* input declaration match */
}

#endif /* }}} */


	/************************************************************/
	/***************** FreeVR Callback routines *****************/
	/************************************************************/

	/************************************************************/
	/*    Callbacks for controlling the device's features.      */
	/*                                                          */

/************************************************************/
static void _StreamSystemPauseToggleCallback(vrInputDevice *devinfo, int value)
{
	if (value == 0)
		return;

vrPrintf("Pausing Stream output!\n");
	devinfo->context->paused ^= 1;
	vrDbgPrintfN(INPUT_DBGLVL, "Stream Control: system_pause = %d.\n",
		devinfo->context->paused);
}

/************************************************************/
static void _StreamPrintContextStructCallback(vrInputDevice *devinfo, int value)
{
	if (value == 0)
		return;

	vrFprintContext(stdout, devinfo->context, verbose);
}

/************************************************************/
static void _StreamPrintConfigStructCallback(vrInputDevice *devinfo, int value)
{
	if (value == 0)
		return;

	vrFprintConfig(stdout, devinfo->context->config, verbose);
}

/************************************************************/
static void _StreamPrintInputStructCallback(vrInputDevice *devinfo, int value)
{
	if (value == 0)
		return;

	vrFprintInput(stdout, devinfo->context->input, verbose);
}

/************************************************************/
static void _StreamPrintPrivateInfoCallback(vrInputDevice *devinfo, int value)
{
	_StreamPrivateInfo  *aux = (_StreamPrivateInfo *)devinfo->aux_data;

	if (value == 0)
		return;

	_StreamPrintPrivateInfo(stdout, aux, verbose);
}

/************************************************************/
static void _StreamPrintHelpCallback(vrInputDevice *devinfo, int value)
{
	_StreamPrivateInfo  *aux = (_StreamPrivateInfo *)devinfo->aux_data;

	if (value == 0)
		return;

	_StreamPrintHelp(stdout, aux);
}

/************************************************************/
static void _StreamPrintValuesCallback(vrInputDevice *devinfo, int value)
{
	_StreamPrivateInfo  *aux = (_StreamPrivateInfo *)devinfo->aux_data;

	if (value == 0)
		return;

	_StreamPrintValues(stdout, aux);
}




	/*************************************************/
	/*   Callbacks for interfacing with the device.  */
	/*                                               */

/**********************************************************/
static void _StreamCreateFunction(vrInputDevice *devinfo)
{
#if 0 /* There's a chance I'll want to control the outputs similar to the inputs, so I'll leave this here for now {{{ */
	/*** List of possible inputs ***/
static	vrInputFunction	_StreamInputs[] = {
			/* TODO: ideally, we should be able to catch all 2-switches with: { NULL, VRINPUT_BINARY, _StreamButtonInput }, */
				{ "char", VRINPUT_BINARY, _StreamButtonInput },
				{ "int",  VRINPUT_BINARY, _StreamButtonInput },
				{ "long", VRINPUT_BINARY, _StreamButtonInput },
			/* TODO: a "float/double" binary is also conceivable, perhaps with an epsilon */
				{ "float",  VRINPUT_VALUATOR, _StreamValuatorInput },
				{ "double", VRINPUT_VALUATOR, _StreamValuatorInput },
			/* TODO: valuators based on integer values are also conceivable */
				{ "vreuler",       VRINPUT_6SENSOR, _Stream6sensorInput },
				{ "vrmatrix",      VRINPUT_6SENSOR, _Stream6sensorInput },
				{ "vreulerfloat",  VRINPUT_6SENSOR, _Stream6sensorInput },
				{ "vrmatrixdouble",VRINPUT_6SENSOR, _Stream6sensorInput },
				{ "vreulerdouble", VRINPUT_6SENSOR, _Stream6sensorInput },
				{ "vrmatrixfloat", VRINPUT_6SENSOR, _Stream6sensorInput },
			/* TODO: I'd rather do without these, but needed to get visbox working ASAP */
				{ "visboxeuler",   VRINPUT_6SENSOR, _Stream6sensorInput },
				{ "visboxtrans",   VRINPUT_6SENSOR, _Stream6sensorInput },
				{ NULL, VRINPUT_UNKNOWN, NULL } };
#endif /* }}} */
	/*** List of control functions ***/
static	vrControlFunc	_StreamControlList[] = {
				/* overall system controls */
				{ "system_pause_toggle", _StreamSystemPauseToggleCallback },

				/* informational output controls */
				{ "print_context", _StreamPrintContextStructCallback },
				{ "print_config",  _StreamPrintConfigStructCallback },
				{ "print_input",   _StreamPrintInputStructCallback },
				{ "print_devinfo",  _StreamPrintPrivateInfoCallback },
				{ "print_help",    _StreamPrintHelpCallback },
				{ "print_values",  _StreamPrintValuesCallback },

		/** TODO: other callback control functions go here **/
				/* simulated 6-sensor selection controls */
				/* simulated 6-sensor manipulation controls */
				/* other controls */

				/* end of the list */
				{ NULL, NULL } };

	_StreamPrivateInfo	*aux = NULL;

	/******************************************/
	/* allocate and initialize auxiliary data */
	devinfo->aux_data = (void *)vrShmemAlloc0(sizeof(*aux));
	aux = (_StreamPrivateInfo *)devinfo->aux_data;
	_StreamInitializeStruct(aux);
	aux->devinfo = devinfo;		/* [1/31/00] -- we don't normally do this, but are experimenting */

	/******************/
	/* handle options */
	_StreamParseArgs(aux, devinfo->args);


	vrDbgPrintf("_StreamCreateFunction(): Done creating Stream outputs for '%s'\n", devinfo->name);
	devinfo->created = 1;

	return;
}


/**********************************************************/
/* _StreamOpenFunction(): ...*/
/* If the "output" name begins with the at ("@") character, then a socket */
/*   port will be used for the output instead.  The format for where to   */
/*   send the data is: "@<hostname>[:<port>]".                            */
static void _StreamOpenFunction(vrInputDevice *devinfo)
{
struct	linger			linger_opt = { 1, 0 };	/* values for the SO_LINGER control options */
	char			host_machine[1024] = "localhost";
	int			hostname_len;
	int			host_port;
	_StreamPrivateInfo	*aux = (_StreamPrivateInfo *)devinfo->aux_data;
	int			return_value;	/* return value for other calls -- eg. fsync() */

	if (aux->output != NULL) {
		/* First check whether the request is for a socket */
		if (aux->output[0] == '@') {
			int	fd_socket;		/* socket connection file descriptor */
			hostname_len = strcspn(aux->output, " \t,;=:");
			if (hostname_len > 1) {
				strncpy(host_machine, &(aux->output[1]), sizeof(host_machine)-1);
				host_machine[strcspn(aux->output, " \t,;=:")-1] = '\0';
			}
			if (aux->output[hostname_len] == ':') {
				aux->socket_port = vrAtoI(&(aux->output[hostname_len+1]));
			}

			fd_socket = vrSocketCall(host_machine, aux->socket_port);
			if (fd_socket < 0) {
				/* Unable to open the socket */
				aux->open = 0;
				devinfo->operating = 0;
				vrErrPrintf("(%s::_StreamOpenFunction()::%d) error: "
					RED_TEXT "couldn't open socket '%s:%d' for %s\n" NORM_TEXT,
					__FILE__, __LINE__, host_machine, aux->socket_port, devinfo->name);
				vrMsgPrintf("Sending Stream Output data to *stdout* -- Unable to send Stream Output data to socket '%s:%d'.\n", host_machine, aux->socket_port);
			} else {
				/* Socket connection established */
				aux->fd_output = fd_socket;
				vrMsgPrintf("Sending Stream Output data to socket '%s:%d'.\n", host_machine, aux->socket_port);
				aux->open = 1;
				devinfo->operating = 1;
			}
		} else { /* Open a regular file: create it if necessary (with permissions 600) */
			return_value = open(aux->output, O_WRONLY | O_CREAT | O_NOCTTY, S_IRUSR | S_IWUSR);
			if (return_value < 0) {
				perror("_StreamOpenFunction(): Failed to open file for Stream Tracking output");
			} else {
				aux->fd_output = return_value;
				vrMsgPrintf("Sending Stream Output data to file '%s'.\n", aux->output);
				aux->open = 1;
				devinfo->operating = 1;
			}
		}
	}

	vrDbgPrintf("_StreamOpenFunction(): Done opening Stream for output device '%s'\n", devinfo->name);
	devinfo->operating = 1;


	/*******************************************/
	/* Print output at appropriate debug level */
	if (vrDbgDo(STREAM_DBGLVL)) {
		_StreamPrintPrivateInfo(stdout, aux, brief);
		_StreamPrintValues(stdout, aux);
	}

	return;
}


/**********************************************************/
static void _StreamCloseFunction(vrInputDevice *devinfo)
{
	_StreamPrivateInfo	*aux = (_StreamPrivateInfo *)devinfo->aux_data;

	vrSocketClose(aux->fd_output);

#if 0 /* VruiDD stuff -- update for Stream */ /* delete me */
	vrDbgPrintf("Closing Stream sockets %d:%d\n", aux->fd_listen_socket, aux->fd_receive_socket);
	if (aux->fd_listen_socket > 0)
		vrSocketClose(aux->fd_listen_socket);
	if (aux->fd_receive_socket > 0)
		vrSocketClose(aux->fd_receive_socket);
	return;
#endif
}


/**********************************************************/
static void _StreamResetFunction(vrInputDevice *devinfo)
{
	/* TODO: figure out what it means to do a reset for a Stream device */

	return;
}


/**********************************************************/
static void _StreamPollFunction(vrInputDevice *devinfo)
{
	_StreamPrivateInfo	*aux = (_StreamPrivateInfo *)devinfo->aux_data;

#if 0 /* VruiDD stuff -- saving this for possible re-use in a partial full-duplex mode with Monado {{{ */
	char			inbuf[512];	/* buffer for data received from the Vrui client */
	uint16_t		message;	/* VRUI VRDeviceDaemon messages are 2 byte sequences */
	uint32_t		value;		/* VRUI 4-byte value */
	int			bytes_read = 0;	/* number of bytes read from socket thus far */
	int			read_result;	/* return value of the call to "read()" */
	int			return_value;	/* return value for other calls -- eg. fsync() */
	struct timeval		no_block = { 0, 0 };	/* set the time values for the no-blocking option */
	fd_set			fds;

	vrTrace("_VruiDDPollFunction():", BOLD_TEXT "at top." NORM_TEXT);
	/* First check for an existing connection, and if not, check for a new connection */
	if (!aux->vrui_link) {
		/* not connected means listen for a connection */
		if (aux->fd_receive_socket < 0 && aux->fd_listen_socket >= 0) {
			/* fd_receive_socket will be -1 when no one is connected */
			aux->fd_receive_socket = vrSocketAnswer(aux->fd_listen_socket);
			if (aux->fd_receive_socket > 0) {
				vrDbgPrintfN(AALWAYS_DBGLVL, "FreeVR: VruiDD connection has been established with client.\n");
int flag = 1;
setsockopt(aux->fd_listen_socket, IPPROTO_TCP, TCP_NODELAY, &flag, sizeof(flag));
			}
		}
	}

	if (aux->fd_receive_socket > 0) {
		/* if we have a connection, see whether there is a message (or disconnect) from the client */
		FD_ZERO(&fds);
		FD_SET(aux->fd_receive_socket, &fds);
		if (select(aux->fd_receive_socket + 1, &fds, NULL, NULL, &no_block)) {

			/***************************************/
			/* Read data from the Vrui VRDD client */

			bytes_read = 0;
			while ((bytes_read < 2) && (aux->fd_receive_socket > 0)) {
				read_result = (ssize_t)read(aux->fd_receive_socket, inbuf, sizeof(inbuf));
				if (read_result == 0) {
					/* this occurs when the other side has disconnected */
					aux->vrui_link = 0;		/* indicate we're not communicating with a Vrui client */
					aux->streaming = 0;		/* set streaming to default off-state */
					aux->fd_receive_socket = -1;	/* indicate that there's no one on the other side of the socket */
					vrDbgPrintfN(AALWAYS_DBGLVL, "FreeVR: VruiDD client has disconnected.\n", bytes_read);
					vrTrace("_VruiDDPollFunction():", BOLD_TEXT "client disconnected." NORM_TEXT);
				} else if (read_result > 0) {
					bytes_read += read_result;
				}
#ifdef COMM_DEBUG
				vrDbgPrintf("_VruiDDPollFunction(): %d bytes read thus far.\n", bytes_read);
#endif
			}

			/**********************************************************/
			/* handle the VRUI VRDeviceDaemon acknowledgment protocol */

			while (bytes_read >= 2) {

				/* verify expected response received (assuming LITTLE_ENDIAN for now) */
				message = (inbuf[0] + (inbuf[1] << 8));
				memmove(inbuf, &inbuf[2], bytes_read-2);	bytes_read -= 2;	/* shift the data by one message packet */
				switch (message) {
				default:
					vrDbgPrintf("_VruiDDPollFunction(): Got an unexpected message %d.\n", message);
					break;

				case CONNECT_REQUEST:	/* Request to connect to server (0x00 0x00) */
					vrDbgPrintf("_VruiDDPollFunction(): Got a CONNECT_REQUEST packet, will now respond with a CONNECT_REPLY.\n");
#if 0
vrPrintf("CONNECT_REQUEST: default protocol = %d\n", aux->protocol);
					/**********************************************************************/
					/* attempt to determine whether the client is using Vrui protocol > 0 */
					if (aux->protocol == 0) {
						/* For protocol 0, the Vrui client should have only sent 2 bytes. */
						/*   If we got 6 bytes (4 extra), then it probably is a newer     */
						/*   protocol, which sent a protocol number.                      */
						if (bytes_read == 4) {
							uint32_t	client_protocol = 0;
							((unsigned char *)(&client_protocol))[0] = inbuf[0];
							((unsigned char *)(&client_protocol))[1] = inbuf[1];
							((unsigned char *)(&client_protocol))[2] = inbuf[2];
							((unsigned char *)(&client_protocol))[3] = inbuf[3];
							//memmove(inbuf, &inbuf[4], bytes_read-4);	bytes_read -= 4;	/* shift the data by one integer packet */
							vrDbgPrintfN(AALWAYS_DBGLVL, "_VruiDDPollFunction(): Vrui application seems to be using protocol %d, but expecting protocol %d -- shifting to new protocol.\n", client_protocol, aux->protocol);
							aux->protocol = client_protocol;
						}
					}
#endif

					/****************************/
					/* Handle protocol exchange */
 					/* (only exchange for protocol==0 when the client sends a protocol value) */
					if (aux->protocol > 0 || (aux->protocol == 0 && bytes_read >= 4)) {
						uint32_t	client_protocol = 0;

						/* TODO: properly handle endianness */
						((unsigned char *)(&client_protocol))[0] = inbuf[0];
						((unsigned char *)(&client_protocol))[1] = inbuf[1];
						((unsigned char *)(&client_protocol))[2] = inbuf[2];
						((unsigned char *)(&client_protocol))[3] = inbuf[3];
						memmove(inbuf, &inbuf[4], bytes_read-4);	bytes_read -= 4;	/* shift the data by one integer packet */

						if (aux->protocol != client_protocol) {
							vrDbgPrintfN(AALWAYS_DBGLVL, "_VruiDDPollFunction(): Vrui protocol mismatch with the application: serving %d, but client reporting %d -- shifting to client protocol.\n", aux->protocol, client_protocol);
							aux->protocol = client_protocol;
						}
					}

					/************************************************/
					/* send the opening reply message to the client */

					/* TODO: I'd like these numbers to be determined in the  */
					/*   _VruiDDCreateFunction() function, but that doesn't  */
					/*   seem to be working like I expected.  It's not a big */
					/*   time sink if it's done here though.                 */
					aux->num_buttons = devinfo->context->input->num_2ways - 1;	/* NOTE: 1 is subtracted because we won't send the Esc key */
					aux->num_valuators = devinfo->context->input->num_valuators;
					aux->num_6sensors = devinfo->context->input->num_6sensors;

					aux->packet_size = sizeof(uint16_t) +				/* message code */
							(sizeof(float)*13 * aux->num_6sensors) +	/* A Vrui TrackerState struct is thirteen floats */
							(sizeof(uint8_t) * aux->num_buttons) +		/* A Vrui button is one uint8_t (aka "bool" in C++) */
							(sizeof(float) * aux->num_valuators);		/* A Vrui valuator is one float */
					vrDbgPrintfN(STREAM_DBGLVL, "_VruiDDPollFunction(): Daemon packet size is %d bytes (num_buttons = %d, valuators = %d, position trackers = %d).\n", aux->packet_size, aux->num_buttons, aux->num_valuators, aux->num_6sensors);

					switch (aux->protocol) {

					/* VruiDD protocol "0" [WRS: my numbering] */
					case 0:
					default:

						message = CONNECT_REPLY;		/* Positive connect reply with server layout (0x01 0x00) */
						aux->buf[ 0] = 0x01;
						aux->buf[ 1] = 0x00;
						aux->buf[ 2] = aux->num_6sensors;	/* TODO: fill the correct values for all bytes (but this will work when there's less than 256 of any type. */
						aux->buf[ 3] = 0x00;
						aux->buf[ 4] = 0x00;
						aux->buf[ 5] = 0x00;
						aux->buf[ 6] = aux->num_buttons;
						aux->buf[ 7] = 0x00;
						aux->buf[ 8] = 0x00;
						aux->buf[ 9] = 0x00;
						aux->buf[10] = aux->num_valuators;
						aux->buf[11] = 0x00;
						aux->buf[12] = 0x00;
						aux->buf[13] = 0x00;
						vrDbgPrintfN(1/*STREAM_DBGLVL*/, "_VruiDDPollFunction(): Sending out the CONNECT_REPLY message for protocol %d (0)\n", aux->protocol);
						write(aux->fd_receive_socket, aux->buf, 14);	/* TODO: warning, no effort yet being made to deal with endianness of the order */
						break;

					/* VruiDD protocol 1 */
					case 1:
						message = CONNECT_REPLY;		/* Positive connect reply with server layout (0x01 0x00) */
						aux->buf[ 0] = 0x01;
						aux->buf[ 1] = 0x00;
						aux->buf[ 2] = aux->protocol;		/* TODO: fill the correct values for all bytes (but this will work when the protocol number is less than 256. */
						aux->buf[ 3] = 0x00;
						aux->buf[ 4] = 0x00;
						aux->buf[ 5] = 0x00;
						aux->buf[ 6] = aux->num_6sensors;
						aux->buf[ 7] = 0x00;
						aux->buf[ 8] = 0x00;
						aux->buf[ 9] = 0x00;
						aux->buf[10] = aux->num_buttons;
						aux->buf[11] = 0x00;
						aux->buf[12] = 0x00;
						aux->buf[13] = 0x00;
						aux->buf[14] = aux->num_valuators;	/* TODO: fill the correct values for all bytes (but this will work when there's less than 256 of any type. */
						aux->buf[15] = 0x00;
						aux->buf[16] = 0x00;
						aux->buf[17] = 0x00;
						vrDbgPrintfN(1/*STREAM_DBGLVL*/, "_VruiDDPollFunction(): Sending out the CONNECT_REPLY message for protocol %d\n", aux->protocol);
						write(aux->fd_receive_socket, aux->buf, 18);	/* TODO: warning, no effort yet being made to deal with endianness of the order */
						aux->buf[18] = 0x00;
						aux->buf[19] = 0x00;
						//write(aux->fd_receive_socket, aux->buf, 20);	/* TODO: warning, no effort yet being made to deal with endianness of the order */
						break;

					/* VruiDD protocol 2 */
					case 2:
						message = CONNECT_REPLY;		/* Positive connect reply with server layout (0x01 0x00) */
						aux->buf[ 0] = 0x01;
						aux->buf[ 1] = 0x00;
						aux->buf[ 2] = aux->protocol;		/* TODO: fill the correct values for all bytes (but this will work when the protocol number is less than 256. */
						aux->buf[ 3] = 0x00;
						aux->buf[ 4] = 0x00;
						aux->buf[ 5] = 0x00;
						aux->buf[ 6] = aux->num_6sensors;
						aux->buf[ 7] = 0x00;
						aux->buf[ 8] = 0x00;
						aux->buf[ 9] = 0x00;
						aux->buf[10] = aux->num_buttons;
						aux->buf[11] = 0x00;
						aux->buf[12] = 0x00;
						aux->buf[13] = 0x00;
						aux->buf[14] = aux->num_valuators;	/* TODO: fill the correct values for all bytes (but this will work when there's less than 256 of any type. */
						aux->buf[15] = 0x00;
						aux->buf[16] = 0x00;
						aux->buf[17] = 0x00;
						vrDbgPrintfN(1/*STREAM_DBGLVL*/, "_VruiDDPollFunction(): Sending out the CONNECT_REPLY message for protocol %d\n", aux->protocol);
						aux->buf[18] = 0x00;			/* TODO: this is for the virtual devices which we are currently hard-coding to 0 */
						aux->buf[19] = 0x00;
						aux->buf[20] = 0x00;
						aux->buf[21] = 0x00;
						write(aux->fd_receive_socket, aux->buf, 22);	/* TODO: warning, no effort yet being made to deal with endianness of the order */
						break;
					}

					return_value = fsync(aux->fd_receive_socket); //-- this should work, but maybe it doesn't
					vrDbgPrintfN(STREAM_DBGLVL, "_VruiDDPollFunction(): Sent out the CONNECT_REPLY message\n");

					aux->vrui_link = 1;		/* set the flag to indicate we're now linked with a client */
					aux->streaming = 0;		/* set streaming to default off-state */
					break;

				case DISCONNECT_REQUEST:	/* Polite request to disconnect from server (0x02 0x00) */
					vrDbgPrintfN(STREAM_DBGLVL, "_VruiDDPollFunction(): Got a DISCONNECT_REQUEST message\n");
					close(aux->fd_receive_socket);
					aux->vrui_link = 0;		/* indicate we're not communicating with a Vrui client */
					aux->streaming = 0;		/* set streaming to default off-state */
					aux->fd_receive_socket = -1;	/* indicate that there's no one on the other side of the socket */
					break;

				case ACTIVATE_REQUEST:		/* Request to activate server (prepare for sending packets) (0x03 0x00) */
					vrDbgPrintfN(STREAM_DBGLVL, "_VruiDDPollFunction(): Got an ACTIVATE_REQUEST message -- treating as a no-op.\n");
					break;

				case DEACTIVATE_REQUEST:	/* Request to deactivate server (no more packet requests) (0x04 0x00) */
					vrDbgPrintfN(STREAM_DBGLVL, "_VruiDDPollFunction(): Got a DEACTIVATE_REQUEST message -- treating as a no-op.\n");
					/* TODO: Presently treating as a no-op */
					break;

				case PACKET_REQUEST:		/* Requests a single packet with current device state (0x05 0x00) */
					vrDbgPrintfN(STREAM_DBGLVL, "_VruiDDPollFunction(): Got a PACKET_REQUEST message -- sending.\n");
					_VruiDDPutData(devinfo);	/* includes the PACKET_REPLY (0x06 0x00) */
					break;

				case STARTSTREAM_REQUEST:	/* Requests entering stream mode (server sends packets automatically) (0x07 0x00) */
					vrDbgPrintfN(STREAM_DBGLVL, "_VruiDDPollFunction(): Got a STARTSTREAM_REQUEST message -- streaming.\n");
					aux->streaming = 1;
					break;

				case STOPSTREAM_REQUEST:	/* Requests leaving stream mode (0x08 0x00) */
					vrDbgPrintfN(STREAM_DBGLVL, "_VruiDDPollFunction(): Got a STOPSTREAM_REQUEST message -- cease streaming.\n");
					/* Set the outgoing message type -- STOPSTREAM_REPLY -- Server's reply after last stream packet has been sent (0x09 0x00) */
					aux->buf[0] = 0x09;
					aux->buf[1] = 0x00;
					write(aux->fd_receive_socket, aux->buf, 2);	/* TODO: warning, no effort yet being made to deal with endianness of the order */
					fsync(aux->fd_receive_socket);

					aux->streaming = 0;
					break;

				/* TODO: case STOPSTREAM_REPLY        * Server's reply after last stream packet has been sent (0x09 0x00) */

				}
			}
		}
	}
#endif /* }}} */

	/* Now if there is a connection, and we're to be streaming the data -- send it out */
	if (devinfo->operating) {
		vrTrace("_StreamPollFunction():", BOLD_TEXT "about to send data." NORM_TEXT);
		_StreamPutData(devinfo);
		vrTrace("_StreamPollFunction():", BOLD_TEXT "sent data." NORM_TEXT);
	}

	return;
}



	/******************************/
	/*** FreeVR public routines ***/
	/******************************/


/**********************************************************/
void vrStreamOutInitInfo(vrInputDevice *devinfo)
{
	devinfo->version = (char *)vrShmemStrDup("Stream out-dev 0.1");
	devinfo->Create = vrCallbackCreateNamed("Stream:Create-Def", _StreamCreateFunction, 1, devinfo);
	devinfo->Open = vrCallbackCreateNamed("Stream:Open-Def", _StreamOpenFunction, 1, devinfo);
	devinfo->Close = vrCallbackCreateNamed("Stream:Close-Def", _StreamCloseFunction, 1, devinfo);
	devinfo->Reset = vrCallbackCreateNamed("Stream:Reset-Def", _StreamResetFunction, 1, devinfo);
	devinfo->PollData = vrCallbackCreateNamed("Stream:PollData-Def", _StreamPollFunction, 1, devinfo);
	devinfo->PrintAux = vrCallbackCreateNamed("Stream:PrintAux-Def", _StreamPrintPrivateInfo, 0);
}

