/* ======================================================================
 *
 *  CCCCC          vr_output.vrpn.c
 * CC   CC         Author(s): Bill Sherman
 * CC              Created: August 27, 2021
 * CC   CC         Last Modified: April 4, 2022 (minor fix 02/29/2024)
 *  CCCCC
 *
 * Code file for FreeVR outputs using the VRPN protocol.
 *
 * Copyright 2024, Bill Sherman, All rights reserved.
 * With the intent to provide an open-source license to be named later.
 * ====================================================================== */
/*************************************************************************

USAGE:
	There currently are no options for flexibly specifying the
	configuration options for the VRPN output server.  At
	present it simply takes all the inputs available (except for
	the first button, which is the 'ESC' key), and puts them, in
	order, into the VRPN protocol.

	Controls are specified in the freevrrc file:
		control "<control option>" = "2switch(...)";

	Here are the available control options for FreeVR:
	  :-(	"print_help" -- print info on how to use the output device
	  :-(	"system_pause_toggle" -- toggle the system pause flag
	  :-(	"print_context" -- print the overall FreeVR context data structure (for debugging)
	  :-(	"print_config" -- print the overall FreeVR config data structure (for debugging)
	  :-(	"print_input" -- print the overall FreeVR input data structure (for debugging)
	  :-(	"print_devinfo" -- print the internal Shared Memory data structure (for debugging)
	  :-(	"print_values" -- print just the input values of the memory


	Here are the FreeVR configuration options for the VRPN output Device:
		"port" - socket port listening for applications
			requesting data.  (8555 is the default)
		"protocol" -- version of the VRPN communications protocol to
			output data (default is 7.35)
	  :-?	"valscale" - scaling factor used to tune the valuator input range
	  :-?	"transscale"/"scale" - set the scaling factor for the location
			of the 6-sensor inputs.


	In order to *record* user inputs in a file using the VRPN protocol
		format, do this:
			% nc localhost 3883 > save_vrpn_data
		while enabling a "vrpnout" device in FreeVR.


	To test the generated VRPN output, either (or both) an official VRPN
	tool can be used ("vrpn_print_devices"), and/or FreeVR's "vrpntest".
		% vrpn_print_devices FreeVR@tcp://localhost

NOTES:
	- VRPN is generally coordinate system agnostic, so in general, it will
		output data in the same space as the FreeVR configuration.
		(Of course, this can be altered with the frame-of-reference
		setting -- e.g. "for_translate".)

	- FreeVR on the other hand uses a Y-up coordinate system, and defaults
		to units as feet.
		BTW: FreeVR has +X to the left and -Z into the screen.

	- The VRPN protocol for position trackers is ...
		floating point (4-byte) values.  The first 3 are location,
		the next four are orientation (as a quaternion), and the
		remaining 6 are for velocities or something, but in practice
		are unused.  Therefore, this driver will send "0.0" values
		for the trailing 6 values.

	- Is there a mechanism in the protocol for times when the server is
		shutdown, and might want to warn the clients.
		[11/26/21: perhaps a shutting-down server can send a text
		message about going down -- I notice that it does do something
		like that when taking the Razer Hydra device down.

HISTORY:
	27 August 2021 - 2 September (Bill Sherman)
		Initial version.

		This version now handles the opening handshake between the
		  server (FreeVR) and the VRPN client.  Plus, button presses,
		  valuators and sensor 6-DOF positions are all handled.
		  Presently, though are all continuously sent rather than
		  just for state changes.

		[corresponds to FreeVR Version 0.7c-alpha]

	28 October 2021 (Bill Sherman)
		Added the code to only send data when it changed since the last
		  transmission -- or when a new connection is made, or when
		  the number of an input type changes (rare).
		Also, began working on adding a full-duplex feature in order to
		  respond to "ping" messages sent by *real* PRON clients.  But
		  even debug session lasting until 3:00am couldn't fix it.

		[corresponds to FreeVR Version 0.7c]

	29 October 2021 (Bill Sherman)
		I made the non-duplex version send a "pong" message to all new
		  connections, and that seems to satisfy "vrpn_print_devices".
		Unfortunately, the VRPN_DUPLEX (full duplex) mode doesn't yet
		  work, and I don't have time to debug it before I do the
		  release of FreeVR version 0.7c, so I'm going to disable it
		  for now and keep it on the TODO/BUG lists.

		[corresponds to FreeVR Version 0.7c]

	26-27 November 2021 (Bill Sherman)
		I added two VRPN text messages to what gets sent when the
		  "send_all" flag is given to _VrpnPutData().  These give some
		  simple provenance meta-data to the VRPN output, which normally
		  won't matter, but if that output is recorded for future
		  playback or analysis, then that will be included.

		[corresponds to FreeVR Version 0.7d]

	28 March 2022 (Bill Sherman) [entered 04/02/22]
		I fixed a *major* bug that I only today realized was there.
		  The sensor number in the posquat tracking payload was
		  mis-formatted in that I thought it was an 8-byte unsigned int,
		  but it was really a 4-byte signed int followed by 4 0x00
		  bytes.  Since vr_input.vrpn.c (and therefore "vrpntest") is
		  compiled with the same bug, I didn't notice it when using it
		  with my own reading code!  (Not sure how I didn't miss it
		  before this though.)

		[corresponds to FreeVR Version 0.7d]

	4 April 2022 (Bill Sherman)
		Further cleaning of the sensor-number-is-only-4-bytes issue.
		  (In looking at the VRPN "vrpn_Tracker.C" code, it turns out
		  that in most cases, the extra 4 bytes sent to get to ensure
		  8-byte alignment is usually a repeat of the sensor number.

		[corresponds to FreeVR Version 0.7d]

	29 February 2024 (Bill Sherman)
		Fixed a small bug where _VRPNPrintValues() tries to print about
		  inputs that don't exist -- because this is an output device!

		[corresponds to FreeVR Version 0.7e]

TODO:
	- DONE: Fix the sensor-number bug in the posquat tracking payload.
		(Was sending 8-byte unsigned int (uint64_t) for the sensor
		number, but needed to be 4-bytes signed int (int32_t) type.)
		[Fixed 03/28/22, entered 04/02/22.]

	- Check other places where the "sensor" number refers to tracking
		data, and probably is also just 4 bytes.

	- After some debugging, compare _VrpnReadInput() with the version in
		"vr_input.vrpn.c".

	- Check to see whether text_message "level" and "severity" values
		are reversed or not.

	- Add some metadata of the VRPN configuration to what the "vrpnout"
		I/O (aka "inputdevice") outputs when initialized.  Ie. send
		that data as a VRPN text message.  It should have information
		such as how many buttons/valuators/6-sensors, and also
		information such as "for_translate" values.

	- Fix a bug that causes the "telnet" process to crash when printing
		the inputdevice after a VRPN client is reading the data.
		I.e. doing this:
			FreeVR>p inputdevice[1]
			Connection closed by foreign host.
		But after "vrpntest" is reading the data.  (It doesn't crash
		prior to that -- though it does report zero buttons/valuators,
		etc., which may be related to the crash.

	- Allow a configuration option that directly dumps VRPN data to a file
		rather than through a socket.

	- DONE: Output the "vrpn_Button States" message which reports all the buttons
		at once (I think).  I think this is done to send the initial
		values for all the buttons.  Send this message once during
		initialization.

	- DONE: Make the button & valuator inputs only be reported for state changes.

	- DONE: Make the 6-sensor inputs only reported for state changes.

	- FAUX: Make it work with "vrpn_print_devices" (ie. make it full-duplex)
		"vrpn_print_devices" expects a "pong" response to its "ping"
		messages.
		[presently, the single mode fakes a duplex connection by sending
		a pong whenver a new connection is made.]

	- Make it handle UDP connections

	- Make it handle multiple connections (This is probably going to be a
		bit complex, so it may be lower priority.)

	- Need to make it configurable.

BUGS: (or potential bugs)
	- FIXED: when communication is not established, the "operating" value
		should be false (0)

	- full duplex (experimental) mode causes crashes and hangs

VRPN PROTOCOL NOTES:
	[These notes are not exhaustive, but only enough for parsing and
	generating the basic needs of a typical VR system, and in this
	case, one that (presently) uses TCP exclusively.]

	OVER-ARCHING:
		* Data is sent in network order (ie. big-endian), so on
			little-endian machines, the byte order is swapped.
		* Data packets are always sent on 8-byte boundaries, and
			messages that don't initially meet that standard
			are padded out to the next 8-byte boundary.

	HEADER:
		* The header is 24 bytes (5 4-byte values plus 4 more for padding)
			- complete message size
			- time (seconds component)
			- time (micro-seconds component)
			- message-code (what category the message is)
			- message-type (what the message contains)
			- padding (as claimed by the protcol, however seems to
				actually contain a count of number of sent messages)

	SYSTEM DESCRIPTION MESSAGES
		* Information from the server of what information is available
		* (Can also be sent from the client but basically never is)
		* The data "payload" includes a redundant string size as the
			first 4-bytes of the "payload" followed by the string.
		* SENDER_DESCRIPTION -- names of the "devices"
		* TYPE_DESCRIPTION -- names of the types of data server may send

	INPUT DATA MESSAGES
		* "vrpn_Button Change" -- 8 byte payload
			- button number (int32_t)
			- button value (int32_t)
			[could these be uint32_t ??]

		* "vrpn_Button States" -- 4 + 4 * num_buttons bytes payload
			- button number (int32_t)
			- (button number) button values (int32_t each)
			[could these be uint32_t ??]

		* "vrpn_Analog Channel" -- 8 + 8 * num_valuators bytes payload
			- number of valuator values (double)
			- each valuator value (double)
			NOTE: Odd that this isn't "States" instead of "Channel"

		* "vrpn_Tracker Pos_Quat" -- 8 * num_6sensors bytes payload
			- number of of the 6-sensor (int32_t)
			- 4 bytes of 0x00 (or sometimes the number of the 6-sensor is repeated just to take up 4 bytes)
			- location (3 * double)
			- quaternion (4 * double)

		* "vrpn_Tracker Velocity" -- 9 * num_6sensors bytes payload
			- number of of the 6-sensor (int32_t)
			- 4 bytes of 0x00 (or sometimes the number of the 6-sensor is repeated just to take up 4 bytes)
			- velocity (3 * double)
			- rotational velocity (4 * double)
			- delta time interval (double)

		* "vrpn_Tracker Acceleration" -- 9 * num_6sensors bytes payload
			[the same as velocity values]

		* "vrpn_Base text_message"
			- type/severity of message (enum)
			- level (uint32)
			- 1024 text message buffer

		* "vrpn_Base ping_message"
			[no payload]
			- set a flag to send a pong message


	OPENING HANDSHAKE
		...

	OTHER INFORMATION
		The position tracker data provided by VRPN is reported to be
		in VQS order (vector/quaternion/scale), with the quaternions
		calculated by taking the rotations in the order:
			- Q_YAW (rotation about Z)
			- Q_PITCH (rotation about Y)
			- Q_ROLL (rotation about X)
		(It's not clear that there really is a "scale" value.)


*************************************************************************/
#include <stdio.h>		/* needed for snprintf() */
#include <string.h>		/* needed for strchr() */
#include <stdint.h>		/* needed for uint16_t & uint32_t types */
#include <sys/time.h>		/* needed for struct timeval */
#include <netinet/tcp.h>	/* needed for TCP_NODELAY and TCP_CORK */
#include <netinet/in.h>		/* needed for IPPROTO_TCP */
#include <errno.h>		/* needed for perror() & errno */
	#include <time.h>	/*temp-maybe: for ctime() */
	#include <stdlib.h>	/*temp-maybe: for drand48() */
double floor(double x);		/* instead of including all of "math.h" */

#include "vr_input.h"
#include "vr_input.opts.h"
#include "vr_socket.h"
#include "vr_utils.h"		/* needed for vrEndianSwap4() & vrEndianSwap8() */
#include "vr_debug.h"
#include "vr_config.h"		/* for vrFprintConfig() */
#include "vr_parse.h"		/* for vrArgParseInteger() */

	/***************************************************/
	/*** definitions for interfacing with the device ***/
	/***                                             ***/

#define	BUFSIZE		4096	/* 10/28/21: was 1024, but there are larger messages */
#define HEADER_SIZE	24
#define	DEFAULT_PORT	3883

#undef VRPN_DUPLEX		/* #undef to only send and not to receive */

/************************************************************************/
/* VRPNServiceFlags can be used by VRPN to specify different tolerances */
/*   for reliability (usually versus latency).                          */
/* NOTE: only the "RELIABLE" flag is presently used.                    */
/* NOTE Also: FreeVR doesn't use this at all.                           */
typedef enum {
		RELIABLE	 = (1<<0),
		FIXED_LATENCY	 = (1<<1),
		LOW_LATENCY	 = (1<<2),
		FIXED_THROUGHPUT = (1<<3),
		HIGH_THROUGHPUT	 = (1<<4)
	} VRPNServiceFlags;


/*************************************************************************/
/* VRPNMessageType provides an enumeration of the different types of     */
/*   messages that can be expected from the server.  Messages provided   */
/*   by specific devices are assigned numbers starting with 0 (zero) as  */
/*   they are chronologically encountered in the vrpn.cfg file.  Special */
/*   assignments are pre-defined for general messages and are given      */
/*   negative values.                                                    */
typedef enum {
		SENDER_DESCRIPTION	= (-1),	/* sending the "sender name" */
		TYPE_DESCRIPTION	= (-2),	/* sending the "type name" */
		UDP_DESCRIPTION		= (-3),	/* sending the host's IP address */
		LOG_DESCRIPTION		= (-4),	/* set a new Log description value */
		DISCONNECT_MESSAGE	= (-5),	/* never sent -- for logging only */
		INVALID_MESSAGE		= (-6)	/* for FreeVR communications error detection */
		/* ... -- the non-negative message-type values are mapped to  */
		/*        ASCII strings immediately after a client-server     */
		/*        connection is established, and then used to indicate*/
		/*        what type of data is arriving (like tracker data or */
		/*        button data.                                        */
	} VRPNMessageType;


/***********************************************************************/
/* VRPNLogFlags uses the lower two bits to establish which information */
/*   should be logged: incoming data, outgoing data, or both.          */
typedef enum {
		LOG_NOTHING		= (0),
		LOG_INCOMING		= (1<<0),
		LOG_OUTGOING		= (1<<1)
	} VRPNLogFlags;


/******************************************************************************/
/* VRPNHeader: data packet header for communication between client and server */
typedef struct {	/* VRPNHeader */
		uint32_t	message_size;			/* header-length (24) + unpadded message length (4 + string_length) */
		uint32_t	time_seconds;			/* seconds component of time */
		uint32_t	time_useconds;			/* micro-seconds component of time */
		uint32_t	code;				/* numeric encoding describing message source (source code :-) */
								/*   code is "sender" when (type == SENDER_DESCRIPTION)        */
								/*   code is "message-number" when (type == TYPE_DESCRIPTION)  */
								/*   code is the device-number when sending input values       */
		int32_t		message_type;			/* type of message (from SystemMessageType) */
		int32_t		padding;			/* need 4 more bytes to be on an 8-byte boundary */
	} VRPNHeader;

#define MAX_VALUATORS	16
#ifdef VRPN_DUPLEX /* This is not used by the output, but perhaps the "name" field is needed for receiving from the client */
/*****************************************************************************/
/* VRPNdevice: structure containing input data from a particular VRPN device */
#define MAX_BUTTONS	16
#define MAX_VALUATORS	16
#define MAX_6SENSORS	8
typedef struct {	/* VRPNdevice */
		char	name[256];				/* name of this device -- according to VRPN server */

		uint8_t	ever_reported_button_any;		/* flag indicating whether the VRPN server has ever sent any button for this device */
		uint8_t	ever_reported_button[MAX_BUTTONS];	/* flag indicating whether the VRPN server has ever sent input numbered such */
		uint8_t	newly_reported_button[MAX_BUTTONS];	/* flag indicating whether the VRPN server just sent input numbered_such */
		uint8_t	mapped_button[MAX_BUTTONS];		/* flag indicating whether the given button has been mapped to a FreeVR input */
		uint8_t	button[MAX_BUTTONS];			/* values of VRPN button inputs */
		double	button_time[MAX_BUTTONS];		/* time of the last button event */

		uint8_t	ever_reported_analog_any;		/* flag indicating whether the VRPN server has ever sent any analog for this device */
		uint8_t	ever_reported_analog[MAX_VALUATORS];	/* flag indicating whether the VRPN server has ever sent input numbered such */
		uint8_t	newly_reported_analog[MAX_VALUATORS];	/* flag indicating whether the VRPN server just sent input numbered_such */
		uint8_t	mapped_analog[MAX_VALUATORS];		/* flag indicating whether the given analog has been mapped to a FreeVR input */
		double	analog[MAX_VALUATORS];			/* values of VRPN valuator inputs */
		double	analog_max[MAX_VALUATORS];		/* maximum value of VRPN valuator inputs reported thus far */
		double	analog_min[MAX_VALUATORS];		/* minimum value of VRPN valuator inputs reported thus far */
		double	analog_time[MAX_VALUATORS];		/* time of the last valuator event */

		uint8_t	ever_reported_posquat_any;		/* flag indicating whether the VRPN server has ever sent any posquat for this device */
		uint8_t	ever_reported_posquat[MAX_6SENSORS];	/* flag indicating whether the VRPN server has ever sent input numbered such */
		uint8_t	newly_reported_posquat[MAX_6SENSORS];	/* flag indicating whether the VRPN server just sent input numbered_such */
		uint8_t	mapped_tracker[MAX_6SENSORS];		/* flag indicating whether the given tracker has been mapped to a FreeVR input */
		double	posquat[MAX_6SENSORS][7];		/* values of VRPN 6-sensor inputs */
		double	posquat_time[MAX_6SENSORS];		/* time of the last 6-sensor event */

		uint8_t	ever_reported_velquat_any;		/* flag indicating whether the VRPN server has ever sent any velquat for this device */
		uint8_t	ever_reported_velquat[MAX_6SENSORS];	/* flag indicating whether the VRPN server has ever sent input numbered such */
		uint8_t	newly_reported_velquat[MAX_6SENSORS];	/* flag indicating whether the VRPN server just sent input numbered_such */
		double	velquat[MAX_6SENSORS][7];		/* values of VRPN 6-sensor input velocity values (currently unused) */
		double	velquat_time[MAX_6SENSORS];		/* time of the last 6-sensor velocity event */
	} VRPNdevice;
#endif


/***********************************************************************************/
/*** _VrpnPrivateInfo: auxiliary structure of the current data about the device. ***/
typedef struct {	/* _VrpnPrivateInfo */
		/* this is for printing extra info in _VRPNPrintValues() */
		vrInputDevice	*devinfo;

		/* these are for interfacing with the client (usually the application) */
		int		fd_listen_socket;	/* socket to listen for connections requests (file_descriptor) */
		int		fd_receive_socket;	/* socket to communicate with client */
		int		dd_port;		/* port on server listening for application connections */
		int		open;			/* flag that VRPN server successfully open */
		int		vrpn_link;		/* flag to indicate when a VRPN client is connected */	/* TODO: for VRPN (as opposed to VruiDD) we need to handle multiple links, so maybe rename this to "vrpn_link_count" at that point. */
		int		streaming;		/* flag indicating whether we're to be streaming data now */	/* TODO: for VRPN, we always stream, so we can remove this VruiDD option -- delete me */
		int		protocol;		/* An integer indicating the VRPN protocol */		/* TODO: for VRPN, the protocol is set by the VRPN version, so we can remove this VruiDD option -- delete me */
		int		server_ver_major;	/* Major version code of this VRPN server */
		int		server_ver_minor;	/* Minor version code of this VRPN server */
		int		server_logging;		/* How much (if any) logging to do */
		char		version[32];		/* The VRPN version "cookie" string -- really needs 24 bytes, but making p-o-2 */

		/* these are for internal data parsing/constructing */ /* TODO: determine whether these are needed for VRPN -- perhaps delete me */
		int		packet_size;		/* the number of bytes in each packet reported by the Daemon */
		unsigned char	buf[BUFSIZE];		/* buffer for sending data to the VRPN Client */
		int		eobuf_pos;		/* position of the last active byte in the buffer */

#define MAX_DEVICES 16 /* there can be <n> VRPN devices */
		/* information about the current (raw/unprocessed) values (SENDER_DESCRIPTION mappings) */
		int		num_devices;			/* just a counter to know how many have been mentioned by the VRPN server */
		unsigned char	device_name[MAX_DEVICES][256];	/* array of possible VRPN input device names (NOTE: for VRPN-input this is the VRPNdevice struct */
#ifdef VRPN_DUPLEX
VRPNdevice	device[MAX_DEVICES];		/* array of possible VRPN input devices *//* from vr_input.vrpn.c */
#endif

		/* these are the TYPE_DESCRIPTION ASCII to integer mappings */
#define MAX_CODES 128 /* the upper bounds on number of TYPE_DESCRIPTION codes in the array */
		int		biggestOut_code;		/* highest code value recorded */
		char		messageOut_codes[MAX_CODES][256];	/* string names for each message code VRPN server (us) is sending */
		int		MessageOutSensorPos;		/* VRPN code for sensor location data */
		int		MessageOutSensorVel;		/* VRPN code for sensor velocity data */
		int		MessageOutSensorAcc;		/* VRPN code for sensor acceleration data */
		int		MessageOutButton;		/* VRPN code for sensor button data */
		int		MessageOutAllButtons;		/* VRPN code for all button data */
		int		MessageOutValuator;		/* VRPN code for sensor velocity data */
		int		MessageOutPong;			/* VRPN code for sending a pong response */
		int		MessageOutText;			/* VRPN code for text message */

#ifdef VRPN_DUPLEX
		int		biggestIn_code;		/* highest code value recorded */
		char		messageIn_codes[MAX_CODES][256];	/* string names for each message code VRPN client is reporting */
		int		MessageInSensorPos;		/* VRPN code for sensor location data */
		int		MessageInSensorVel;		/* VRPN code for sensor velocity data */
		int		MessageInSensorAcc;		/* VRPN code for sensor acceleration data */
		int		MessageInButton;		/* VRPN code for sensor button data */
		int		MessageInAllButtons;		/* VRPN code for all button data */
		int		MessageInValuator;		/* VRPN code for sensor velocity data */
		int		MessageInPing;			/* VRPN code for receiving a ping request */
		int		MessageInText;			/* VRPN code for text message */

		int		unanswered_ping;		/* If we get a ping, we need to response with a pong */
#endif


		/* TODO: most of these fields were part of vr_input.shmemd.c, I suspect most can be delete me */

		/* these are pointers to the FreeVR inputs */
		int		num_buttons;
#if 0 /* saving these for a time when I might do output-mapping */
		vr2switch	**button_inputs;
		int		*button_values;		/* array of current values */
		int		*button_offsets;	/* array of offsets into the VRPN memory structure */
		int		*button_dummy;		/* array of flags whether button data coming from dummy values */
#endif

		int		num_valuators;
#if 1 /* saving these for a time when I might do output-mapping */
		vrValuator	**valuator_inputs;
		float		*valuator_values;	/* array of current values */
		int		*valuator_offsets;	/* array of offsets into the VRPN memory structure */
		int		*valuator_dummy;	/* array of flags whether valuator data coming from dummy values */
#endif

		int		num_6sensors;
#if 1 /* saving these for a time when I might do output-mapping */
		vr6sensor	**sensor6_inputs;
		vrMatrix	*sensor6_values;	/* array of current values */
		int		*sensor6_offsets;	/* array of offsets into the VRPN memory structure */
		int		*sensor6_dummy;		/* array of flags whether 6sensor data coming from dummy values */
#endif
	} _VrpnPrivateInfo;

/* _VrpnTypeDescriptionList is the list of message "Types" that the */
/*    server may send.  In practice most of these are not used, and */
/*    so we may shorten this list in the future.                    */
static char *_VrpnTypeDescriptionList[] = {
		"VRPN_Connection_Got_First_Connection",
		"VRPN_Connection_Got_Connection",
		"VRPN_Connection_Dropped_Connection",
		"VRPN_Connection_Dropped_Last_Connection",
	  /*!*/	"vrpn_Analog Channel",
	  /*.*/	"vrpn_Base text_message",
	  /*.*/	"vrpn_Base ping_message",
	  /*!*/	"vrpn_Base pong_message",
	  /*!*/	"vrpn_Button Change",
	  /*!*/	"vrpn_Button States",
		"vrpn_Button Admin",
		"vrpn_Button Alert",
	  /*!*/	"vrpn_Tracker Pos_Quat",
		"vrpn_Tracker Velocity",
		"vrpn_Tracker Acceleration",
		"vrpn_Tracker To_Room",
		"vrpn_Tracker Unit_To_Sensor",
		"vrpn_Tracker Request_Tracker_To_Room",
		"vrpn_Tracker Request_Unit_To_Sensor",
		"vrpn_Tracker Workspace",
		"vrpn_Tracker Request_Tracker_Workspace",
		"vrpn_Tracker set_update_rate",
		"vrpn_Tracker Reset_Origin",
		"vrpn_Forwarder_Brain start_forwarding",
		"vrpn_Forwarder_Brain forward",
		NULL
	};


	/*********************************************/
	/*** General NON public interface routines ***/
	/*********************************************/

/**************************************************************************/
static void _VrpnInitializeStruct(_VrpnPrivateInfo *aux)
{
	int	count;

	aux->fd_listen_socket = -1;		/* i.e. indicate that it's not yet initialized */
	aux->fd_receive_socket = -1;		/* i.e. indicate that it's not yet initialized */
	aux->dd_port = DEFAULT_PORT;		/* Set the default listening port */
	aux->vrpn_link = 0;			/* No client is connected yet */	/* TODO: for VRPN (as opposed to VruiDD) we need to handle multiple links, so maybe rename this to "vrpn_link_count" at that point. */
	aux->streaming = 0;			/* Not presently streaming data */
	aux->protocol = 0;			/* VRPN protocol ... (NOTE: VRPN has a major and minor number) */
	aux->server_ver_major = 7;		/* VRPN Version, major component */
	aux->server_ver_minor = 34;		/* VRPN Version, minor component */
	aux->server_logging = 0;		/* VRPN Logging value (default to no logging) */
#ifdef VRPN_DUPLEX
	aux->unanswered_ping = 0;		/* No need to pong until we get a ping */
#endif

	/* NOTE: no safety checks are done for major.minor -- make sure above they are between 0 and 99 inclusive */
	snprintf(aux->version, 24, "vrpn: ver. %02d.%02d  %d%c%c%c%c%c",
		aux->server_ver_major, aux->server_ver_minor, aux->server_logging,
		'\0', '\0', '\0', '\0', '\0');	/* NULL padding */

	/* Fill in the SENDER_DESCRIPTION device names */
	/* NOTE: presently only one, and that may always be the case */
	aux->num_devices = 1;
	strncpy(aux->device_name[0], "FreeVR", sizeof(aux->device_name[0]));

	/* Fill in the TYPE_DESCRIPTION device names */
	for (count = 0; _VrpnTypeDescriptionList[count] != NULL; count++) {
		strncpy(aux->messageOut_codes[count], _VrpnTypeDescriptionList[count], sizeof(aux->messageOut_codes[count]));

		/* store an internal mapping to the key data messages */
		if (!strcmp(aux->messageOut_codes[count], "vrpn_Tracker Pos_Quat")) {
			aux->MessageOutSensorPos = count;
			vrDbgPrintfN(VRPN_DBGLVL, "_VrpnInitializeStruct(): assigned MessageOutSensorPos to %d\n", aux->MessageOutSensorPos);
		} else if (!strcmp(aux->messageOut_codes[count], "vrpn_Tracker Velocity")) {
			aux->MessageOutSensorVel = count;
			vrDbgPrintfN(VRPN_DBGLVL, "_VrpnInitializeStruct(): assigned MessageOutSensorVel to %d\n", aux->MessageOutSensorVel);
		} else if (!strcmp(aux->messageOut_codes[count], "vrpn_Tracker Acceleration")) {
			aux->MessageOutSensorAcc = count;
			vrDbgPrintfN(VRPN_DBGLVL, "_VrpnInitializeStruct(): assigned MessageOutSensorAcc to %d\n", aux->MessageOutSensorAcc);
		} else if (!strcmp(aux->messageOut_codes[count], "vrpn_Button Change")) {
			aux->MessageOutButton = count;
			vrDbgPrintfN(VRPN_DBGLVL, "_VrpnInitializeStruct(): assigned MessageOutButton to %d\n", aux->MessageOutButton);
		} else if (!strcmp(aux->messageOut_codes[count], "vrpn_Analog Channel")) {
			aux->MessageOutValuator = count;
			vrDbgPrintfN(VRPN_DBGLVL, "_VrpnInitializeStruct(): assigned MessageOutValuator to %d\n", aux->MessageOutValuator);
		} else if (!strcmp(aux->messageOut_codes[count], "vrpn_Base text_message")) {
			aux->MessageOutText = count;
			vrDbgPrintfN(VRPN_DBGLVL, "_VrpnInitializeStruct(): assigned MessageOutText to %d\n", aux->MessageOutText);
		} else if (!strcmp(aux->messageOut_codes[count], "vrpn_Base pong_message")) {
			aux->MessageOutPong = count;
			vrDbgPrintfN(VRPN_DBGLVL, "_VrpnInitializeStruct(): assigned MessageOutPong to %d\n", aux->MessageOutPong);
		}
	}
	aux->biggestOut_code = count - 1;
}


/**************************************************************************/
static void _VrpnPrintPrivateInfo(FILE *file, _VrpnPrivateInfo *aux, vrPrintStyle style)
{
	int	count;

	switch (style) {
		case brief:

		vrFprintf(file, "VRPN-output: connection=(%d:%d:%d), open=%d, link=%d, streaming=%d, protocol=%d\n",
			aux->fd_listen_socket, aux->fd_receive_socket, aux->dd_port,
			aux->open, aux->vrpn_link, aux->streaming, aux->protocol);
#if 0
	char msg[1024];
	snprintf(msg, sizeof(msg), "VRPN-output: connection=(%d:%d:%d), open=%d, link=%d, streaming=%d, protocol=%d\n",
		aux->fd_listen_socket, aux->fd_receive_socket, aux->dd_port,
		aux->open, aux->vrpn_link, aux->streaming, aux->protocol);
	vrSystemSetStatusDescription(msg);
#endif
		break;

		case verbose:
		default:
		vrFprintf(file, "VRPN output-device internal structure (%#p):\n", aux);
		vrFprintf(file, "\r\tvrpn version = %02d.%02d\n", aux->server_ver_major, aux->server_ver_minor);
		vrFprintf(file, "\r\tvrpn server logging = %d\n", aux->server_logging);
		vrFprintf(file, "\r\tvrpn cookie = '%s'\n", aux->version);
		vrFprintf(file, "\r\tvrpn open = %d\n", aux->open);
		vrFprintf(file, "\r\tvrpn listening socket file descriptor = %d\n", aux->fd_listen_socket);

		vrFprintf(file, "\tVRPN Server reporting devices:\n");		/* SENDER_DESCRIPTION values */
		for (count = 0; count <= aux->num_devices; count++) {
			if (aux->device_name[count][0] != '\0')
				vrFprintf(file, "\t\tDevice %d is '%s'\n", count, aux->device_name[count]);
		}
		vrFprintf(file, "\tVRPN Server reporting message types:\n");	/* TYPE_DESCRIPTION values */
		for (count = 0; count <= aux->biggestOut_code; count++) {
			vrFprintf(file, "\t\tOutput Message Code %2d is '%s'\n", count, aux->messageOut_codes[count]);
		}

#ifdef VRPN_DUPLEX
		vrFprintf(file, "\t\t%2d: unanswered pong from client\n", aux->unanswered_ping);
#endif

		vrFprintf(file, "\tOutput Messages (to client):\n");		/* TYPE_DESCRIPTION values */
		vrFprintf(file, "\t\t%2d: Tracker position\n", aux->MessageOutSensorPos);
		vrFprintf(file, "\t\t%2d: Tracker velocity\n", aux->MessageOutSensorVel);
		vrFprintf(file, "\t\t%2d: Tracker acceleration\n", aux->MessageOutSensorAcc);
		vrFprintf(file, "\t\t%2d: Button event\n", aux->MessageOutButton);
		vrFprintf(file, "\t\t%2d: Valuator event\n", aux->MessageOutValuator);
		vrFprintf(file, "\t\t%2d: Text message\n", aux->MessageOutText);

#ifdef VRPN_DUPLEX
		vrFprintf(file, "\tWatch Messages (from client):\n");		/* TYPE_DESCRIPTION values */
		vrFprintf(file, "\t\t%2d: Tracker position\n", aux->MessageInSensorPos);
		vrFprintf(file, "\t\t%2d: Tracker velocity\n", aux->MessageInSensorVel);
		vrFprintf(file, "\t\t%2d: Tracker acceleration\n", aux->MessageInSensorAcc);
		vrFprintf(file, "\t\t%2d: Button event\n", aux->MessageInButton);
		vrFprintf(file, "\t\t%2d: Valuator event\n", aux->MessageInValuator);
		vrFprintf(file, "\t\t%2d: Ping message\n", aux->MessageInPing);
		vrFprintf(file, "\t\t%2d: Text message\n", aux->MessageInText);
#endif

		vrFprintf(file, "\r\t%d buttons:\n", aux->num_buttons);
#if 0 /* Note that no inputs actually exist for the stream output I/O device */ /* So could delete me */
		vrFprintf(file, "\r\tbutton_inputs = %#p, button_values = %#p, button_offsets = %#p\n",
			aux->button_inputs, aux->button_values, aux->button_offsets);
		for (count = 0; count < aux->num_buttons; count++) {
			vrFprintf(file, "\r\t\tbutton %d: value = %d, offset = %d, input = %#p (type = %d)\n",
				count,
				aux->button_values[count],
				aux->button_offsets[count],
				aux->button_inputs[count],
				aux->button_inputs[count]->input_type);
		}

		vrFprintf(file, "\r\t%d valuators:\n", aux->num_valuators);
		vrFprintf(file, "\r\tvaluator_inputs = %#p, valuator_values = %#p, valuator_offsets = %#p\n",
			aux->valuator_inputs, aux->valuator_values, aux->valuator_offsets);
		for (count = 0; count < aux->num_valuators; count++) {
			vrFprintf(file, "\r\t\tvaluator %d: value = %.2f, offset = %d, input = %#p (type = %d)\n",
				count,
				aux->valuator_values[count],
				aux->valuator_offsets[count],
				aux->valuator_inputs[count],
				aux->valuator_inputs[count]->input_type);
		}

		vrFprintf(file, "\r\t%d 6-sensors:\n", aux->num_6sensors);
		vrFprintf(file, "\r\t6sensor_inputs = %#p, 6sensor_values = %#p, 6sensor_offsets = %#p\n",
			aux->sensor6_inputs, aux->sensor6_values, aux->sensor6_offsets);
		for (count = 0; count < aux->num_6sensors; count++) {
			vrFprintf(file, "\r\t\t6-sensor %d: value = [... (TODO!)], offset = %d, input = %#p (type = %d)\n",
				count,
			/* TODO: put matrix values here  -- perhaps convert to quat/trans */
				aux->sensor6_offsets[count],
				aux->sensor6_inputs[count],
				aux->sensor6_inputs[count]->input_type);
		}
#endif
	}
}


/**************************************************************************/
static void _VrpnPrintHelp(FILE *file, _VrpnPrivateInfo *aux)
{
	vrFprintf(file, BOLD_TEXT "Sorry, VRPN -- print_help control not yet implemented.\n" NORM_TEXT);
}


/**************************************************************************/
static void _VrpnPrintValues(FILE *file, _VrpnPrivateInfo *aux)
{
	int	count;

	vrFprintf(file, "VRPN output device inputs to send:\n");

	vrFprintf(file, "\r\t%d buttons\n", aux->num_buttons);
#if 0 /* 02/29/24: WRS: This is an "output" device, so it doesn't actually have it's own buttons */
	for (count = 0; count < aux->num_buttons; count++) {
		if (aux->button_inputs[count]->input_type == VRINPUT_BINARY) {
			vrFprintf(file, "\r\t\tbutton %d ('%s') [offset = %d]: = %d\n",
				count,
				aux->devinfo->switch2[count].my_object->name,
				aux->button_offsets[count],
				aux->button_values[count]);
		} else {
			vrFprintf(file, "\r\t\tbutton %d ('%s') [offset = %d]: not reported as VRINPUT_BINARY (%d instead)\n",
				count,
				aux->devinfo->switch2[count].my_object->name,
				aux->button_offsets[count],
				aux->button_inputs[count]->input_type);
		}
	}
#endif

	vrFprintf(file, "\r\t%d valuators\n", aux->num_valuators);
#if 0 /* 02/29/24: WRS: This is an "output" device, so it doesn't actually have it's own valuators */
	for (count = 0; count < aux->num_valuators; count++) {
		if (aux->valuator_inputs[count]->input_type == VRINPUT_VALUATOR) {
			vrFprintf(file, "\r\t\tvaluator %d ('%s') [offset = %d]: = %.2f\n",
				count,
				aux->devinfo->valuator[count].my_object->name,
				aux->valuator_offsets[count],
				aux->valuator_values[count]);
		} else {
			vrFprintf(file, "\r\t\tvaluator %d ('%s') [offset = %d]: not reported as VRINPUT_VALUATOR (%d instead)\n",
				count,
				aux->devinfo->switch2[count].my_object->name,
				aux->valuator_offsets[count],
				aux->valuator_inputs[count]->input_type);
		}
	}
#endif

	vrFprintf(file, "\r\t%d 6-sensors\n", aux->num_6sensors);
#if 0 /* 02/29/24: WRS: This is an "output" device, so it doesn't actually have it's own 6-sensors */
	for (count = 0; count < aux->num_6sensors; count++) {
		if (aux->sensor6_inputs[count]->input_type == VRINPUT_6SENSOR) {
			/* TODO: print the matrix */
			vrFprintf(file, "\r\t\t6sensor %d ('%s') [offset = %d]: = ...\n",
				count,
				aux->devinfo->sensor6[count].my_object->name,
				aux->sensor6_offsets[count]);
		} else {
			vrFprintf(file, "\r\t\t6sensor %d ('%s') [offset = %d]: not reported as VRINPUT_6SENSOR (%d instead)\n",
				count,
				aux->devinfo->switch2[count].my_object->name,
				aux->sensor6_offsets[count],
				aux->sensor6_inputs[count]->input_type);
		}
	}
#endif
}


	/************************************/
	/***  FreeVR NON public routines  ***/
	/************************************/


#ifdef VRPN_DUPLEX /* { */
/************************************************************************/
/* FROM vr_input.vrpn.c (copied 10/29/2021) */
/* The _VrpnReadInput() function acquires incoming data from the server */
/*   source, parsing the stream of data.  The data is then placed into  */
/*   the generic portion of the "_VrpnPrivateInfo" structure.  Data is  */
/*   then transitioned into the FreeVR (or whatever) system through the */
/*   "_VrpnGetData()" function.                                         */
static int _VrpnReadInput(_VrpnPrivateInfo *aux)
{
	/************************************************************/
	/* This section reads the VRPN data by parsing the raw data */
	/************************************************************/
	VRPNHeader	VRPNdata;		/* storage of the incoming VRPN header */
	int		read_result;		/* return value of the call to "read()" */
	int		packets_decoded = 0;	/* return the number of packets that are decoded */
	int		packet_size;		/* size of the currently being parsed data packet -- adjusted as new information is determined */

	uint32_t	message_length;		/* for string-type messages */
	char		*message;		/* for string-type messages */
	int		padded_message_size;	/* the complete length of the message portion of the packet (or does it include header too?) */

	double		vrpntime;		/* time stamp of the incoming VRPN data */
	VRPNdevice	*vrpn_dev;		/* pointer to a device in the VRPN device list */

static int hello = 0;
	/*******************************************************/
	/*** read available data into available buffer space ***/
	do {
vrPrintf("hello, _VrpnReadInput() reading data -- %d\n", hello++);
		read_result = (ssize_t)read(aux->fd_receive_socket, &aux->buf[aux->eobuf_pos], BUFSIZE - aux->eobuf_pos);
vrPrintf("   read done.\n");

		/* NOTE: if read_result == 0, the socket was closed from the other side.  This condition is handled at the end of the function */
		if (read_result > 0) {
			aux->eobuf_pos += read_result;
			/* debug statement here */
		}
#  if 1
		if (read_result < -1) {
			/* we don't care about a -1 return value */
			/* error ? */
			vrErrPrintf("_VrpnReadInput(): Unexpected error: read_result = %d, errno = %d\n", read_result, errno);
			//perror("_VrpnReadInput(): socket read:");
		}
#  endif
	} while (read_result > 0 && aux->eobuf_pos < BUFSIZE);

	/*******************************************/
	/*** decode data contained in the buffer ***/

	/* decode until no more complete messages are contained in the buffer */
	packet_size = HEADER_SIZE + 8;	/* There at least must be data for the first header */  /* NOTE: in reality, we'll always need the first number as well */
	while (aux->eobuf_pos >= packet_size) {
		/* point the VRPN interpretation structure (aka "header") to the beginning of the buffer */
		VRPNdata = *(VRPNHeader *)aux->buf;

		/* interpret the header */
#  if __BYTE_ORDER == __LITTLE_ENDIAN
		vrEndianSwap4(&VRPNdata.message_size);
		vrEndianSwap4(&VRPNdata.time_seconds);
		vrEndianSwap4(&VRPNdata.time_useconds);
		vrEndianSwap4(&VRPNdata.code);
		vrEndianSwap4(&VRPNdata.message_type);
#  endif

		/* now make sure the entire packet is in the buffer */
		padded_message_size = VRPNdata.message_size;
		padded_message_size = (padded_message_size+7)&(~7);	/* use bit-wise operation to raise size to next 8-byte number */
//vrPrintf("padded_message_size = %d; bufsize = %d\n", padded_message_size, sizeof(aux->buf));
		if (padded_message_size > sizeof(aux->buf)) {
			/* but first make sure the entire packet can fit in the buffer! */
			vrDbgPrintfN(AALWAYS_DBGLVL, "Not enough room in the buffer for the VRPN packet: (%d of %d needed bytes).\n", sizeof(aux->buf), padded_message_size);
			return (packets_decoded);	/* this is a hack, but I don't know what else to do -- the actual solution is to make the buffer bigger */
		}

		if (aux->eobuf_pos < padded_message_size) {
			/* debug statement here */
			vrDbgPrintfN(VRPN_DBGLVL, "Not enough data in the buffer (%d of %d needed bytes), come back for more\n", aux->eobuf_pos, padded_message_size);
			//vrDbgPrintfN(VRPN_DBGLVL, "************** YO: need to drop out of the while loop here! *********\n");
			//vrDbgPrintf("message_size = %d, padded size = %d, bytes in the buffer = %d\n", VRPNdata.message_size, padded_message_size, aux->eobuf_pos);
			packet_size = padded_message_size;	/* skip the complete packet */
			continue;	/* jump out of the loop if we don't have enough data */
		}

		/* If the number is more negative than any of our messages, */
		/*   force it to be exactly the INVALID_MESSAGE value, and  */
		/*   then process it in the switch.                         */
		if (VRPNdata.message_type <= INVALID_MESSAGE) {
			VRPNdata.message_type = INVALID_MESSAGE;
		}

		/* now interpret the data in the message container */
		switch (VRPNdata.message_type) {
		case SENDER_DESCRIPTION:
			vrDbgPrintfN(VRPN_DBGLVL, "SENDER_DESCRIPTION\n");

			/* extract the string message -- starting with the size */
			packet_size = HEADER_SIZE + sizeof(message_length);		/* NOTE: this isn't really necessary */
			memcpy(&message_length, &aux->buf[HEADER_SIZE], sizeof(message_length));
#  if __BYTE_ORDER == __LITTLE_ENDIAN
			vrEndianSwap4(&message_length);
#  endif
//vrPrintf("message_length = %d\n", message_length);	/* delete me */

			/* now calculate the overall packet size */
			packet_size = HEADER_SIZE + sizeof(message_length) + message_length;
			packet_size = (packet_size+7)&(~7);	/* use bit-wise operation to raise size to next 8-byte number */
//vrPrintf("packet_size = %d, aux->eobuf_pos = %d\n", packet_size, aux->eobuf_pos);	/* delete me */
			if (aux->eobuf_pos < packet_size) {
				/* debug statement here */
				/* 01/12/13 NOTE: we should now never get here since we did a check before the switch */
				vrErrPrintf("\nVRPN: Parsing SENDER_DESCRIPTION: Not enough data in the buffer (%d of %d needed bytes)\n\n\n", aux->eobuf_pos, packet_size);
				vrErrPrintf("VRPN: ************** SHOULDN'T GET HERE! need to drop out of the while loop here! *********\n");
				continue;	/* jump out of the loop if we don't have enough data */
			}

			/* store the name of the device */
			message = &aux->buf[HEADER_SIZE + sizeof(message_length)];
			/* NOTE: subtract 1 from MAX_DEVICES because device[0] is the simulator */
			if (VRPNdata.code < MAX_DEVICES-1) {
				/* NOTE: 1 is added to "code" because of the simulator as device[0] */
				strncpy(aux->device[VRPNdata.code+1].name, message, sizeof(aux->device[0].name));
				vrDbgPrintfN(VRPN_DBGLVL, "VRPN device number (%d) has the name '%s'.\n", (int)(VRPNdata.code+1), message);
				aux->num_devices++;
				/* TODO: do a check here to see whether we've already initialized the FreeVR inputs and if so print a warning */
			} else {
#  if !defined(TEST_APP) /* { */
				vrDbgPrintfN(AALWAYS_DBGLVL, "WARNING: VRPN device number (%d) exceeds maximum (%d) -- '%s'.\n", (int)(VRPNdata.code+1), MAX_DEVICES-1, message);
#  else
				vrDbgPrintf("WARNING: VRPN device number (%d) exceeds maximum (%d) -- '%s'.\n", (int)(VRPNdata.code+1), MAX_DEVICES-1, message);
#  endif /* } TEST_APP */
			}
			vrDbgPrintfN(VRPN_DBGLVL, "VRPN: Sender: bytes = %d, mes-size = %u, time = %u:%06u, code = %u, mes-type = %d, mes-len = %u, message = '%s'\n", aux->eobuf_pos, VRPNdata.message_size, VRPNdata.time_seconds, VRPNdata.time_useconds, VRPNdata.code, VRPNdata.message_type, message_length, message);
			break;

		case TYPE_DESCRIPTION:
			vrDbgPrintfN(VRPN_DBGLVL, "TYPE_DESCRIPTION\n");
			vrDbgPrintfN(0, "-----------------------------------TYPE_DESCRIPTION\n");

			/* extract the string message -- starting with the size */
			packet_size = HEADER_SIZE + sizeof(message_length);		/* NOTE: this isn't really necessary */
			memcpy(&message_length, &aux->buf[HEADER_SIZE], sizeof(message_length));
			//message = &aux->buf[aux->eobuf_pos];
			message = &aux->buf[HEADER_SIZE + sizeof(message_length)];
#  if __BYTE_ORDER == __LITTLE_ENDIAN
			vrEndianSwap4(&message_length);
#  endif
//vrPrintf("message_length = %d\n", message_length);	/* delete me */

			/* now calculate the overall packet size */
			packet_size = HEADER_SIZE + sizeof(message_length) + message_length;
			packet_size = (packet_size+7)&(~7);	/* use bit-wise operation to raise size to next 8-byte number */
			if (aux->eobuf_pos < packet_size) {
				/* debug statement here */
				/* 01/12/13 NOTE: we should now never get here since we did a check before the switch */
				vrErrPrintf("\nVRPN: Parsing TYPE_DESCRIPTION: Not enough data in the buffer (%d of %d needed bytes)\n\n\n", aux->eobuf_pos, packet_size);
				vrErrPrintf("VRPN: ************** SHOULDN'T GET HERE! need to drop out of the while loop here! *********\n");
				continue;	/* jump out of the loop if we don't have enough data */
			}

vrPrintf("got TYPE_DESCRIPTION string: '%s' -- code is %d\n", message, VRPNdata.code);	/* delete me */
			/* now store the TYPE_DESCRIPTION message strings    */
			/*   (and store a mapping of the important messages) */
			if (!strncmp(message, "vrpn_Tracker Pos_Quat", message_length)) {
				aux->MessageInSensorPos = VRPNdata.code;
				vrDbgPrintfN(VRPN_DBGLVL FORCE_DBG, "_VrpnReadInput(): assigned MessageInSensorPos to %d\n", aux->MessageInSensorPos);
			} else if (!strncmp(message, "vrpn_Tracker Velocity", message_length)) {
				aux->MessageInSensorVel = VRPNdata.code;
				vrDbgPrintfN(VRPN_DBGLVL FORCE_DBG, "_VrpnReadInput(): assigned MessageInSensorVel to %d\n", aux->MessageInSensorVel);
			} else if (!strncmp(message, "vrpn_Tracker Acceleration", message_length)) {
				aux->MessageInSensorAcc = VRPNdata.code;
				vrDbgPrintfN(VRPN_DBGLVL FORCE_DBG, "_VrpnReadInput(): assigned MessageInSensorAcc to %d\n", aux->MessageInSensorAcc);
			} else if (!strncmp(message, "vrpn_Button Change", message_length)) {
				aux->MessageInButton = VRPNdata.code;
				vrDbgPrintfN(VRPN_DBGLVL FORCE_DBG, "_VrpnReadInput(): assigned MessageInButton to %d\n", aux->MessageInButton);
			} else if (!strncmp(message, "vrpn_Button States", message_length)) {
				aux->MessageInAllButtons = VRPNdata.code;
				vrDbgPrintfN(VRPN_DBGLVL FORCE_DBG, "_VrpnReadInput(): assigned MessageInAllButtons to %d\n", aux->MessageInAllButtons);
			} else if (!strncmp(message, "vrpn_Analog Channel", message_length)) {
				aux->MessageInValuator = VRPNdata.code;
				vrDbgPrintfN(VRPN_DBGLVL FORCE_DBG, "_VrpnReadInput(): assigned MessageInValuator to %d\n", aux->MessageInValuator);
			} else if (!strncmp(message, "vrpn_Base ping_message", message_length)) {
				aux->MessageInPing = VRPNdata.code;
				vrDbgPrintfN(VRPN_DBGLVL FORCE_DBG, "_VrpnReadInput(): assigned MessageInPing to %d\n", aux->MessageInPing);
			} else if (!strncmp(message, "vrpn_Base text_message", message_length)) {
				aux->MessageInText = VRPNdata.code;
				vrDbgPrintfN(VRPN_DBGLVL FORCE_DBG, "_VrpnReadInput(): assigned MessageInText to %d\n", aux->MessageInText);
			}
//vrPrintf("current biggestIn_code = %d, VRPNdata.code = %d\n", aux->biggestIn_code, VRPNdata.code);	/* delete me */
			if ((int32_t)(VRPNdata.code) < MAX_CODES) {
				if ((int32_t)(VRPNdata.code) > aux->biggestIn_code)
					aux->biggestIn_code = (int32_t)(VRPNdata.code);
				else	vrDbgPrintf("Warning: duplicate message code value received: %u\n", VRPNdata.code);
//vrPrintf("about to store description: code = %d, message = '%s', size = %d\n", VRPNdata.code, message, sizeof(aux->messageIn_codes[0]));	/* delete me */
				strncpy(aux->messageIn_codes[VRPNdata.code], message, sizeof(aux->messageIn_codes[0]));
			}
			vrDbgPrintfN(VRPN_DBGLVL, "Type: bytes = %d, mes-size = %u, time = %u:%06u, code = %u, mes-type = %d, mes-len = %u, message = '%s'\n", aux->eobuf_pos, VRPNdata.message_size, VRPNdata.time_seconds, VRPNdata.time_useconds, VRPNdata.code, VRPNdata.message_type, message_length, message);
//vrPrintf("finished decoding TYPE_DESCRIPTION\n");	/* delete me */
			break;

		case UDP_DESCRIPTION:
			vrDbgPrintfN(VRPN_DBGLVL, RED_TEXT "UDP_DESCRIPTION -- not parsed! (code = %u)\n" NORM_TEXT, VRPNdata.code);
			packet_size = padded_message_size;	/* skip the complete packet */
			break;

		case LOG_DESCRIPTION:
			vrDbgPrintfN(VRPN_DBGLVL, RED_TEXT "LOG_DESCRIPTION -- not parsed! (code = %u)\n" NORM_TEXT, VRPNdata.code);
			packet_size = padded_message_size;	/* skip the complete packet */
			break;

		case DISCONNECT_MESSAGE:
			vrDbgPrintfN(VRPN_DBGLVL, RED_TEXT "DISCONNECT_MESSAGE -- not parsed! (code = %u)\n" NORM_TEXT, VRPNdata.code);
			packet_size = padded_message_size;	/* skip the complete packet */
			break;

		case INVALID_MESSAGE:
			vrDbgPrintfN(VRPN_DBGLVL, "\nVRPN: Invalid messages type: %d -- not parsed!.\n", VRPNdata.message_type);
			packet_size = padded_message_size;	/* skip the complete packet */
			break;

		default:
			/************************************************************************/
			/* The default message is that we have some actual input data to parse. */
			/************************************************************************/

			/******************************/
			/*** Tracker Position Data? ***/
			if (VRPNdata.message_type == aux->MessageInSensorPos) {
				/* tracking data = 64bit sensor number, 3*64 bit float location, 4*64bit float orientation (64 bytes) */
				int		count;
/* TODO: this probably needs fixing too */
				uint64_t	sensor;		/* the 8-byte (64bit) sensor value in the VRPN data packet */
				int32_t		int_sensor;	/* the unsigned 64 bit sensor value wreaks havoc with array access */

				/* redetermine the packet size now that we know the type of data */
				packet_size = HEADER_SIZE + sizeof(sensor) + 7 * sizeof(vrpn_dev->posquat[0][0]);
				packet_size = (packet_size+7)&(~7);	/* use bit-wise operation to raise size to next 8-byte number */

				/* now determine whether there is enough data in the buffer! */
				if (aux->eobuf_pos < packet_size) {
					/* debug statement here */
					/* 01/12/13 NOTE: we should now never get here since we did a check before the switch */
					vrErrPrintf("\nVRPN: Not enough data in the buffer (%d of %d needed bytes)\n\n\n", aux->eobuf_pos, packet_size);
					vrErrPrintf("VRPN: ************** SHOULDN'T GET HERE! need to drop out of the while loop here! *********\n");
					continue;	/* jump out of the loop if we don't have enough data */
				}


				/* now parse the data! */
				memmove(&sensor, &aux->buf[HEADER_SIZE], sizeof(sensor));
				vrpntime = VRPNdata.time_seconds + (VRPNdata.time_useconds * 0.000001);
				vrpn_dev = &(aux->device[VRPNdata.code+1]);
#  if __BYTE_ORDER == __LITTLE_ENDIAN
				vrEndianSwap8(&sensor);
#  endif
				int_sensor = (int32_t)sensor;		/* convert to an integer to avoid gcc array indexing bug */
				if (int_sensor < MAX_6SENSORS) {
					vrpn_dev->ever_reported_posquat_any = 1u;
					vrpn_dev->ever_reported_posquat[int_sensor] = 1u;
					vrpn_dev->newly_reported_posquat[int_sensor] = 1u;
					for (count = 0; count < 7; count++) {
						memmove(&(vrpn_dev->posquat[int_sensor][count]), &aux->buf[HEADER_SIZE + sizeof(sensor) + (count * sizeof(vrpn_dev->posquat[0][0]))], sizeof(vrpn_dev->posquat[0][0]));
#  if __BYTE_ORDER == __LITTLE_ENDIAN
						vrEndianSwap8(&(vrpn_dev->posquat[int_sensor][count]));
#  endif
					}
					vrDbgPrintfN(VRPN_DBGLVL, BOLD_TEXT "\rTime = %0.6f" NORM_TEXT " -- PosQuat data: device = %u sensor = %d, xyz = %.3f %.3f %.3f, quat = %.3f %.3f %.3f %.3f", vrpntime, VRPNdata.code, int_sensor, vrpn_dev->posquat[int_sensor][0], vrpn_dev->posquat[int_sensor][1], vrpn_dev->posquat[int_sensor][2], vrpn_dev->posquat[int_sensor][3], vrpn_dev->posquat[int_sensor][4], vrpn_dev->posquat[int_sensor][5], vrpn_dev->posquat[int_sensor][6]);
				} else {
					vrDbgPrintfN(COMMON_DBGLVL, "VRPN: Number of 6-sensor (%d) exceeds maximum (%d).\n", int_sensor+1, MAX_6SENSORS);
				}

			}
			/**********************/
			/*** Velocity Data? ***/
			else if (VRPNdata.message_type == aux->MessageInSensorVel) {
				/* tracking data = 64bit sensor number, 3*64 bit float location vel, 4*64bit float orientation vel + one more unknown double! (72 bytes) */
				int		count;
/* TODO: this probably needs fixing too */
				uint64_t	sensor;
				double		velquat[7];
				double		unknown_quantity;

				/* redetermine the packet size now that we know the type of data */
				packet_size = HEADER_SIZE + sizeof(sensor) + 7 * sizeof(velquat[0]) + sizeof(unknown_quantity);
				packet_size = (packet_size+7)&(~7);	/* use bit-wise operation to raise size to next 8-byte number */

				/* now determine whether there is enough data in the buffer! */
				if (aux->eobuf_pos < packet_size) {
					/* debug statement here */
					/* 01/12/13 NOTE: we should now never get here since we did a check before the switch */
					//vrErrPrintf("\nVRPN: Not enough data in the buffer\n\n\n", packet_size);
					vrErrPrintf("\nVRPN: Not enough data in the buffer (%d of %d needed bytes)\n\n\n", aux->eobuf_pos, packet_size);
					vrErrPrintf("VRPN: ************** SHOULDN'T GET HERE! need to drop out of the while loop here! *********\n");
					continue;	/* jump out of the loop if we don't have enough data */
				}


				/* now parse the data! */
				memmove(&sensor, &aux->buf[HEADER_SIZE], sizeof(sensor));
				memmove(&unknown_quantity, &aux->buf[packet_size - sizeof(unknown_quantity)], sizeof(unknown_quantity));
#  if __BYTE_ORDER == __LITTLE_ENDIAN
				vrEndianSwap8(&sensor);
				vrEndianSwap8(&unknown_quantity);
#  endif
				for (count = 0; count < 7; count++) {
					memmove(&velquat[count], &aux->buf[HEADER_SIZE + sizeof(sensor) + (count * sizeof(velquat[0]))], sizeof(velquat[0]));
#  if __BYTE_ORDER == __LITTLE_ENDIAN
					vrEndianSwap8(&velquat[count]);
#  endif
				}

				vrDbgPrintfN(VRPN_DBGLVL, BOLD_TEXT "\nVRPN: Time = %u:%06u" NORM_TEXT " -- VelQuat data: device = %u sensor = %d, xyz = %.3f %.3f %.3f, quat = %.3f %.3f %.3f %.3f, unknown = %.3f\n", VRPNdata.time_seconds, VRPNdata.time_useconds, VRPNdata.code, sensor, velquat[0], velquat[1], velquat[2], velquat[3], velquat[4], velquat[5], velquat[6], unknown_quantity);

			}

			/**************************/
			/*** Acceleration Data? ***/
			else if (VRPNdata.message_type == aux->MessageInSensorAcc) {
				packet_size = HEADER_SIZE + 72;
			}

			/**********************/
			/*** Valuator Data? ***/
			else if (VRPNdata.message_type == aux->MessageInValuator) {
				/* valuator data = 64bit number of channels, num-channels * 64bit values */
				int		count;
				double		num_valuators;		/* Oddly this is sent as a double precision float */
				double		valuators[MAX_VALUATORS];

				memmove(&num_valuators, &aux->buf[HEADER_SIZE], sizeof(num_valuators));
#  if __BYTE_ORDER == __LITTLE_ENDIAN
				vrEndianSwap8(&num_valuators);
#  endif
				packet_size = HEADER_SIZE + sizeof(num_valuators) + sizeof(valuators[0]) * num_valuators;
				/* now determine whether there is enough data in the buffer! */
				if (aux->eobuf_pos < packet_size) {
					/* debug statement here */
					/* 01/12/13 NOTE: we should now never get here since we did a check before the switch */
					//vrErrPrintf("\nVRPN: Not enough data in the buffer (only %d)\n\n\n", packet_size);
					vrErrPrintf("\nVRPN: Not enough data in the buffer (%d of %d needed bytes)\n\n\n", aux->eobuf_pos, packet_size);
					vrErrPrintf("VRPN: ************** SHOULDN'T GET HERE! need to drop out of the while loop here! *********\n");
					continue;	/* jump out of the loop if we don't have enough data */
				}

				vrpntime = VRPNdata.time_seconds + (VRPNdata.time_useconds * 0.000001);
				vrpn_dev = &(aux->device[VRPNdata.code+1]);
				if (num_valuators > MAX_VALUATORS) {
					vrDbgPrintfN(COMMON_DBGLVL, "VRPN: Number of valuators available (%d) exceeds maximum (%d).\n", ((int)num_valuators), MAX_VALUATORS);
					num_valuators = MAX_VALUATORS;		/* cap number to maximum allowable */
				}

				vrDbgPrintfN(VRPN_DBGLVL, BOLD_TEXT "\nVRPN: Time = %u:%06u" NORM_TEXT " -- (device = %u) Valuators (%d): ", VRPNdata.time_seconds, VRPNdata.time_useconds, VRPNdata.code, (int)num_valuators);
				for (count = 0; count < num_valuators && count < MAX_VALUATORS; count++) {
					memmove(&valuators[count], &aux->buf[HEADER_SIZE + sizeof(num_valuators) + (count * sizeof(valuators[0]))], sizeof(valuators[0]));
#  if __BYTE_ORDER == __LITTLE_ENDIAN
					vrEndianSwap8(&valuators[count]);
#  endif
					if (!vrpn_dev->ever_reported_analog[count]) {
						/* This first time this analog/valuator input is received, set its min and max values */
						vrpn_dev->analog_max[count] = valuators[count];
						vrpn_dev->analog_min[count] = valuators[count];
					} else {
						/* otherwise adjust the min/max values as needed */
						if (valuators[count] > vrpn_dev->analog_max[count])
							vrpn_dev->analog_max[count] = valuators[count];
						if (valuators[count] < vrpn_dev->analog_min[count])
							vrpn_dev->analog_min[count] = valuators[count];
					}
					vrpn_dev->analog[count] = valuators[count];
					vrpn_dev->ever_reported_analog_any = 1u;
					vrpn_dev->ever_reported_analog[count] = 1u;
					vrpn_dev->newly_reported_analog[count] = 1u;
					vrpn_dev->analog_time[count] = vrpntime;

					vrDbgPrintfN(VRPN_DBGLVL, "% lf ", valuators[count]);
				}
				vrDbgPrintfN(VRPN_DBGLVL, "\n");
			}

			/***************************/
			/*** Single Button Data? ***/
			else if (VRPNdata.message_type == aux->MessageInButton) {
				/* button change data = 32bit button number, 32bit button value (8 bytes) */
				uint32_t	button_number;
				uint32_t	button_value;

				packet_size = HEADER_SIZE + 8;
				if (aux->eobuf_pos < packet_size) {		/* NOTE: we won't need this once I assume there's at least 8 bytes after the header */
					/* debug statement here */
					/* 01/12/13 NOTE: we should now never get here since we did a check before the switch */
					//vrErrPrintf("\nVRPN: Not enough data in the buffer (only %d bytes)\n\n\n", packet_size);
					vrErrPrintf("\nVRPN: Not enough data in the buffer (%d of %d needed bytes)\n\n\n", aux->eobuf_pos, packet_size);
					vrErrPrintf("VRPN: ************** SHOULDN'T GET HERE! need to drop out of the while loop here! *********\n");
					continue;	/* jump out of the loop if we don't have enough data */
				}
				memmove(&button_number, &aux->buf[HEADER_SIZE], sizeof(button_number));
				memmove(&button_value, &aux->buf[HEADER_SIZE + sizeof(button_number)], sizeof(button_value));
#  if __BYTE_ORDER == __LITTLE_ENDIAN
				vrEndianSwap4(&button_number);
				vrEndianSwap4(&button_value);
#  endif
				vrpntime = VRPNdata.time_seconds + (VRPNdata.time_useconds * 0.000001);
				vrpn_dev = &(aux->device[VRPNdata.code+1]);

				if (button_number < MAX_BUTTONS) {
					vrpn_dev->button[button_number] = button_value;
					vrpn_dev->ever_reported_button_any = 1u;
					vrpn_dev->ever_reported_button[button_number] = 1u;
					vrpn_dev->newly_reported_button[button_number] = 1u;
					vrpn_dev->button_time[button_number] = vrpntime;
				} else {
					vrDbgPrintfN(COMMON_DBGLVL, "VRPN: Number of button (%d) exceeds maximum (%d).\n", button_number+1, MAX_BUTTONS);
				}

				vrDbgPrintfN(VRPN_DBGLVL, BOLD_TEXT "\nTime = %u:%06u" NORM_TEXT " -- (device = %u) Button %d is now %d\n", VRPNdata.time_seconds, VRPNdata.time_useconds, VRPNdata.code, button_number, button_value);
			}

			/************************/
			/*** All Button Data? ***/
			else if (VRPNdata.message_type == aux->MessageInAllButtons) {
				/* button change data = 64bit button count, <N> 64bit button values */
				int		count;
				uint32_t	num_buttons;
				uint32_t	buttons[MAX_BUTTONS];

				memcpy(&num_buttons, &aux->buf[HEADER_SIZE], sizeof(num_buttons));
#  if __BYTE_ORDER == __LITTLE_ENDIAN
				vrEndianSwap4(&num_buttons);
#  endif

				packet_size = HEADER_SIZE + sizeof(num_buttons) + sizeof(buttons[0]) * num_buttons;
				packet_size = (packet_size+7)&(~7);	/* use bit-wise operation to raise size to next 8-byte number */
//vrPrintf("adjusted packet_size = %d, num_buttons = %d\n", packet_size, num_buttons);	/* delete me */
				/* now determine whether there is enough data in the buffer! */
				if (aux->eobuf_pos < packet_size) {
					/* debug statement here */
					/* 01/12/13 NOTE: we should now never get here since we did a check before the switch */
					//vrErrPrintf("\nVRPN: Not enough data in the buffer (only %d bytes)\n\n\n", packet_size);
					vrErrPrintf("\nVRPN: Not enough data in the buffer (%d of %d needed bytes)\n\n\n", aux->eobuf_pos, packet_size);
					vrErrPrintf("VRPN: ************** SHOULDN'T GET HERE! need to drop out of the while loop here! *********\n");
					continue;	/* jump out of the loop if we don't have enough data */
				}

				vrpntime = VRPNdata.time_seconds + (VRPNdata.time_useconds * 0.000001);
				vrpn_dev = &(aux->device[VRPNdata.code+1]);
				if (num_buttons > MAX_BUTTONS) {
					vrDbgPrintfN(COMMON_DBGLVL, "VRPN: Number of buttons available (%d) exceeds maximum (%d).\n", ((int)num_buttons), MAX_BUTTONS);
					num_buttons = MAX_BUTTONS;		/* cap number to maximum allowable */
				}

				vrDbgPrintfN(VRPN_DBGLVL, BOLD_TEXT "\nVRPN: Time = %u:%06u" NORM_TEXT " -- (device = %u) Buttons (%d): ", VRPNdata.time_seconds, VRPNdata.time_useconds, VRPNdata.code, (int)num_buttons);
				for (count = 0; count < num_buttons && count < MAX_BUTTONS; count++) {
					memmove(&buttons[count], &aux->buf[HEADER_SIZE + sizeof(num_buttons) + (count * sizeof(buttons[0]))], sizeof(buttons[0]));
#  if __BYTE_ORDER == __LITTLE_ENDIAN
					vrEndianSwap8(&buttons[count]);
#  endif
					vrpn_dev->button[count] = buttons[count];
					vrpn_dev->ever_reported_button_any = 1u;
					vrpn_dev->ever_reported_button[count] = 1u;
					vrpn_dev->newly_reported_button[count] = 1u;
					vrpn_dev->button_time[count] = vrpntime;

					vrDbgPrintfN(VRPN_DBGLVL, "%d ", buttons[count]);
				}
				vrDbgPrintfN(VRPN_DBGLVL, "\n");
			}

			/********************/
			/*** Ping message ***/
			else if (VRPNdata.message_type == aux->MessageInPing) {
				/* NOTE: "packet_size" need to be set to the expected size for this message */
				packet_size = HEADER_SIZE;

				vrDbgPrintfN(AALWAYS_DBGLVL, "VRPN Client pinged us\n");
				aux->unanswered_ping = 1;
			}

			/******************/
			/*** Text Data? ***/
			else if (VRPNdata.message_type == aux->MessageInText) {
				int32_t		level;
				int32_t		severity;
				char		message[1024];	/* max VRPN text message size */

				/* NOTE: "packet_size" need to be set to the expected size for this message */
				packet_size = HEADER_SIZE + sizeof(level) + sizeof(severity) + sizeof(message);

				/* now parse the data! */
				memcpy(&level, &aux->buf[HEADER_SIZE], sizeof(level));
				memcpy(&severity, &aux->buf[HEADER_SIZE + sizeof(level)], sizeof(severity));
				memcpy(message, &aux->buf[HEADER_SIZE + sizeof(level) + sizeof(severity)], sizeof(message));
				vrpntime = VRPNdata.time_seconds + (VRPNdata.time_useconds * 0.000001);
#  if __BYTE_ORDER == __LITTLE_ENDIAN
				vrEndianSwap4(&level);
				vrEndianSwap4(&severity);
#  endif
				vrDbgPrintfN(AALWAYS_DBGLVL, "VRPN Client message (%d/%d): '%s'\n", level, severity, message);
			}

			/************************/
			/*** Unknown Datatype ***/
			else {
				memcpy(&message_length, &aux->buf[HEADER_SIZE], sizeof(message_length));
#  if __BYTE_ORDER == __LITTLE_ENDIAN
				vrEndianSwap4(&message_length);
#  endif
				message = &aux->buf[HEADER_SIZE + sizeof(message_length)];
				packet_size = padded_message_size;	/* skip the complete packet */

				//vrDbgPrintfN(VRPN_DBGLVL, "\nVRPN: Unknown data type -- %d (code = %u)\n\n", VRPNdata.message_type, VRPNdata.code);
				vrDbgPrintfN(VRPN_DBGLVL, RED_TEXT "\nVRPN: Unwatched data type -- '%s' %d (code = %u)\n" NORM_TEXT, aux->messageIn_codes[VRPNdata.message_type], VRPNdata.message_type, VRPNdata.code);
				vrDbgPrintfN(VRPN_DBGLVL, "VRPN: bytes = %d, mes-size = %u, time = %u:%06u, code = %u, mes-type = %d, string? = '%s'\n\n", aux->eobuf_pos, VRPNdata.message_size, VRPNdata.time_seconds, VRPNdata.time_useconds, VRPNdata.code, VRPNdata.message_type, message);

				/* print the information about the unknown message */
				int		count;
				vrDbgPrintfN(AALWAYS_DBGLVL, RED_TEXT "Unwatched data header:" NORM_TEXT " size = %d, time = %d:%d, code = %d ('%s'), type = %d ('%s')\nPayload Bytes:",
					VRPNdata.message_size,
					VRPNdata.time_seconds,
					VRPNdata.time_useconds,
					VRPNdata.code, aux->device[VRPNdata.code+1].name,
					VRPNdata.message_type, aux->messageIn_codes[VRPNdata.message_type]);
				for (count = HEADER_SIZE; count < VRPNdata.message_size; count++) {
					vrPrintf("%02x ", aux->buf[count]);
				}
				/* TODO: also output the payload as 4-byte integers, 8-byte integers, doubles, and ASCII */
				vrDbgPrintfN(AALWAYS_DBGLVL, "\n");
#if 0 /* Set this to "1" to exit which can make it easier to analyze the data ordering */
				exit(0);
#endif
			}
		}

		/* do a sanity check!  (the reported message size should match the actual message size) */
		if (padded_message_size != packet_size) {
			vrDbgPrintfN(AALWAYS_DBGLVL, RED_TEXT "VRPN: SANITY FAIL: padded_message_size = %d, packet_size = %d -- bytes = %d, mes-size = %u, time = %u:%06u, code = %u, mes-type = %d\n" NORM_TEXT, padded_message_size, packet_size, aux->eobuf_pos, VRPNdata.message_size, VRPNdata.time_seconds, VRPNdata.time_useconds, VRPNdata.code, VRPNdata.message_type);
		}

		/* now shift the buffer by the message size (padded to 8 byte boundaries) */
		memmove(aux->buf, &aux->buf[packet_size], (aux->eobuf_pos - packet_size));	/* shift the data */
		aux->eobuf_pos -= packet_size;
		packets_decoded++;

		/* reset the packet size to the minimal size required for parsing */
		packet_size = HEADER_SIZE;	/* There at least must be data for the first header */  /* NOTE: in reality, we'll always need the first number as well */
	}

	/* return -1 if the socket connection is closed */
	if (read_result == 0) {
		vrDbgPrintf("\n\n_VRPNReadInput(): socket connection closed -- %d packets decoded in this pass\n", packets_decoded);
		return (-1);
	}
	return (packets_decoded);
}
#endif /* } VRPN_DUPLEX */


/***************************************************************/
static void _VrpnParseArgs(_VrpnPrivateInfo *aux, char *args)
{
	/* In the rare case of no arguments, just return */
	if (args == NULL)
		return;

	/****************************************/
	/** Argument format: "port" "=" number **/
	/****************************************/
	vrArgParseInteger(args, "port", &(aux->dd_port));

	/********************************************/
	/** Argument format: "protocol" "=" number **/
	/********************************************/
	vrArgParseInteger(args, "protocol", &(aux->protocol));


	/* it would be reasonable to add some arguments for setting the dummy values here */
}


/***************************************************************************/
/* _VrpnServerSendDescriptionMessage(): sends a single message comprised   */
/*   of a SENDER_D... or TYPE_D... message, preceded by the message header.*/
static void _VrpnServerSendDescriptionMessage(vrInputDevice *devinfo, VRPNHeader *header, char *message)
{
static	char			padded_message[2048];
	int			padded_message_size;
	int			data_message_size;
	_VrpnPrivateInfo	*aux = (_VrpnPrivateInfo *)devinfo->aux_data;
	int			return_value;		/* return value for socket calls */
	uint32_t		message_length;		/* for string-type messages */
	VRPNHeader		local_header;		/* make a local copy of the header because we change it! */

	/* make a local copy of the header to protect the original data */
	local_header = *header;

	/* calculate the full length of the message */
	message_length = strlen(message) + 1;
	local_header.message_size = sizeof(VRPNHeader) + message_length + sizeof(message_length);
	padded_message_size = (local_header.message_size+7)&(~7);	/* use bit-wise operation to raise size to next 8-byte number */
	data_message_size = padded_message_size - sizeof(VRPNHeader) - sizeof(message_length);
	strncpy(padded_message, message, data_message_size);	/* NOTE: strncpy fills with '\0' as needed */

#if __BYTE_ORDER == __LITTLE_ENDIAN
	/* Handle the Header byte ordering */
	vrEndianSwap4(&local_header.message_size);
	vrEndianSwap4(&local_header.time_seconds);
	vrEndianSwap4(&local_header.time_useconds);
	vrEndianSwap4(&local_header.code);
	vrEndianSwap4(&local_header.message_type);

	/* Handle the payload byte ordering */
	vrEndianSwap4(&message_length);
#endif

	/* Send the header */
	return_value = write(aux->fd_receive_socket, &local_header, sizeof(VRPNHeader));
	if (return_value != sizeof(VRPNHeader)) {
		vrDbgPrintfN(VRPN_DBGLVL, "_VrpnServerSendDescriptionMessage(): " RED_TEXT "An error occurred while trying to send VRPN header - %d bytes sent instead of %d.\n" NORM_TEXT, return_value, sizeof(VRPNHeader));
		if (return_value < 0)
			perror("_VrpnOpenFunction(): write()");
	}
	/* Send the length of the data (message) + 1 */
	return_value = write(aux->fd_receive_socket, &message_length, sizeof(message_length));
	/* Send the actual data (message) */
	return_value = write(aux->fd_receive_socket, padded_message, data_message_size);
	if (return_value != data_message_size) {
		vrDbgPrintfN(VRPN_DBGLVL, "_VrpnServerSendDescriptionMessage(): " RED_TEXT "An error occurred while trying to send VRPN message - %d bytes sent instead of %d.\n" NORM_TEXT, return_value, data_message_size);
		if (return_value < 0)
			perror("_VrpnOpenFunction(): write()");
	}

	/* Flush the output data */
	return_value = fsync(aux->fd_receive_socket); //-- this should work, but maybe it doesn't
#if !defined(__APPLE__) && !defined(__CYGWIN__)	/* Darwin & Cygwin don't have the TCP_CORK option */
	int flag = 0;
	setsockopt(aux->fd_receive_socket, IPPROTO_TCP, TCP_CORK, &flag, sizeof(flag));
	flag = 1;
	setsockopt(aux->fd_receive_socket, IPPROTO_TCP, TCP_CORK, &flag, sizeof(flag));
#endif

	vrDbgPrintfN(VRPN_DBGLVL, "_VrpnServerSendDescriptionMessage(): Message sent: %d endian(%d) bytes sent.\n", sizeof(VRPNHeader) + data_message_size + sizeof(message_length), header->message_size);
}


/*********************************************************************/
/* _VrpnServerSendMessage(): sends a single message comprised of the */
/*   message header followed by the message.                         */
static void _VrpnServerSendMessage(vrInputDevice *devinfo, VRPNHeader *header, void *message, uint32_t message_length)
{
static	char			padded_message[2048];
	int			padded_message_size;
	int			data_message_size;
	_VrpnPrivateInfo	*aux = (_VrpnPrivateInfo *)devinfo->aux_data;
	int			return_value;		/* return value for socket calls */
	//uint32_t		message_length;		/* for string-type messages */
	VRPNHeader		local_header;		/* make a local copy of the header because we change it! */
	int			count;

//vrPrintf("111\n"); /* delete me */
	/* make a local copy of the header to protect the original data */
	local_header = *header;

	/* calculate the full length of the message */
	local_header.message_size = sizeof(VRPNHeader) + message_length;
	padded_message_size = (local_header.message_size+7)&(~7);	/* use bit-wise operation to raise size to next 8-byte number */
	data_message_size = padded_message_size - sizeof(VRPNHeader);
#ifdef VRPN_DUPLEX
	if (data_message_size > 0) {
		memcpy(padded_message, message, data_message_size);
	}
#else
	memcpy(padded_message, message, data_message_size);
#endif

//vrPrintf("222\n"); /* delete me */
#if __BYTE_ORDER == __LITTLE_ENDIAN
	/* Handle the Header byte ordering */
	vrEndianSwap4(&local_header.message_size);
	vrEndianSwap4(&local_header.time_seconds);
	vrEndianSwap4(&local_header.time_useconds);
	vrEndianSwap4(&local_header.code);
	vrEndianSwap4(&local_header.message_type);
#endif

//vrPrintf("333\n"); /* delete me */
	/* Send the header */
	return_value = write(aux->fd_receive_socket, &local_header, sizeof(VRPNHeader));
//vrPrintf("333.1\n"); /* delete me */
#if 0 /* new code 10/28/21 */
	if (return_value < 0) {
		perror("_VrpnServerSendMessage(): An error occurred while trying to send VRPN header -- write()");
	} else {
		if (return_value != sizeof(VRPNHeader)) {
//vrPrintf("333.2\n"); /* delete me */
			vrDbgPrintfN(VRPN_DBGLVL FORCE_DBG, "_VrpnServerSendMessage(): " RED_TEXT "An error occurred while trying to send VRPN header - %d bytes sent instead of %d.\n" NORM_TEXT, return_value, sizeof(VRPNHeader));
//vrPrintf("333.3\n"); /* delete me */
		}
	}
#else /* original code */
	if (return_value != sizeof(VRPNHeader)) {
		vrDbgPrintfN(VRPN_DBGLVL FORCE_DBG, "_VrpnServerSendMessage(): " RED_TEXT "An error occurred while trying to send VRPN header - %d bytes sent instead of %d.\n" NORM_TEXT, return_value, sizeof(VRPNHeader));
		if (return_value < 0)
			perror("_VrpnServerSendMessage(): An error occurred while trying to send VRPN header -- write()");
	}
#endif

//vrPrintf("444\n"); /* delete me */
	/* Send the actual data (message) */
#if 0 /* new code 10/28/21 */
	if (data_message_size == 0) {
		return_value = write(aux->fd_receive_socket, padded_message, data_message_size);
		if (return_value < 0) {
			perror("_VrpnServerSendMessage(): An error occurred while trying to send VRPN header -- write()");
		} else {
			if (return_value != data_message_size) {
				vrDbgPrintfN(VRPN_DBGLVL, "_VrpnServerSendMessage(): " RED_TEXT "An error occurred while trying to send VRPN message - %d bytes sent instead of %d.\n" NORM_TEXT, return_value, data_message_size);
			}
		}
	}
#else /* original code */
	return_value = write(aux->fd_receive_socket, padded_message, data_message_size);
	if (return_value != data_message_size) {
		vrDbgPrintfN(VRPN_DBGLVL, "_VrpnServerSendMessage(): " RED_TEXT "An error occurred while trying to send VRPN message - %d bytes sent instead of %d.\n" NORM_TEXT, return_value, data_message_size);
		if (return_value < 0)
			perror("_VrpnServerSendMessage(): An error occurred while trying to send VRPN header -- write()");
	}
#endif

//vrPrintf("555\n"); /* delete me */
	/* Flush the output data */
	return_value = fsync(aux->fd_receive_socket); //-- this should work, but maybe it doesn't
#if !defined(__APPLE__) && !defined(__CYGWIN__)	/* Darwin & Cygwin don't have the TCP_CORK option */
	int flag = 0;
	setsockopt(aux->fd_receive_socket, IPPROTO_TCP, TCP_CORK, &flag, sizeof(flag));
	flag = 1;
	setsockopt(aux->fd_receive_socket, IPPROTO_TCP, TCP_CORK, &flag, sizeof(flag));
#endif
}


/***********************************************************************/
/* _VrpnNewClientConnectionData(): sends all the information needed by */
/*   a client after a new connection is established.                   */
static void _VrpnNewClientConnectionData(vrInputDevice *devinfo)
{
	_VrpnPrivateInfo	*aux = (_VrpnPrivateInfo *)devinfo->aux_data;
	VRPNHeader		VRPNdata;
	vrTime			time_now;
	int			count;

	/* get some data for the header */
	time_now = vrCurrentWallTime();

	/* Assemble the message header */
	VRPNdata.message_size = 0;			/* header-length (24) + unpadded message length (4 + string_length) */
	VRPNdata.time_seconds = floor(time_now);	/* seconds component of time */
	VRPNdata.time_useconds = (time_now - floor(time_now)) * 1000000;	/* micro-seconds component of time */
	VRPNdata.code = 0;				/* numeric encoding describing message source */
								/*   code is "sender" when (type == SENDER_DESCRIPTION)        */
								/*   code is "message-number" when (type == TYPE_DESCRIPTION)  */
	VRPNdata.message_type = 0;			/* type of message (from SystemMessageType) */

	/* Send all the "SENDER_DESCRIPTION" (device) messages */
	VRPNdata.message_type = SENDER_DESCRIPTION;
	VRPNdata.code = 0;	/* the 0th device */

	_VrpnServerSendDescriptionMessage(devinfo, &VRPNdata, aux->device_name[0]);

	/* Send all the "TYPE_DESCRIPTION" messages */
	VRPNdata.message_type = TYPE_DESCRIPTION;
	for (count = 0; count <= aux->biggestOut_code; count++) {
		VRPNdata.code = count;
		_VrpnServerSendDescriptionMessage(devinfo, &VRPNdata, aux->messageOut_codes[count]);
	}
}


/***************************************************************/
/* the "send_all" flag indicates all data should be forced out */
/*   (usually this means we have a new connection, so the client */
/*   needs all the data.                                         */
static void _VrpnPutData(vrInputDevice *devinfo, int send_all)
{
	_VrpnPrivateInfo	*aux = (_VrpnPrivateInfo *)devinfo->aux_data;
	VRPNHeader		VRPNdata;
	int			count;
	int			offset;
	int			return_val;
	vrTime			time_now;

//send_all = 0;	// WRS: 03/28/22: removing this because at the moment my own VRPN reader can't handle the VRPN text message that is sent.

	/* For new connections (and maybe some other cases) send all our information */
	if (send_all) {
		vrDbgPrintfN(DEFAULT_DBGLVL, "_VrpnPutData():  Sending all data.\n");
	}

	/* get some data for the header */
	time_now = vrCurrentWallTime();

	/* Assemble the message header */
	VRPNdata.message_size = 0;	/* header-length (24) + unpadded message length (4 + string_length) */
	VRPNdata.time_seconds = (uint32_t)(floor(time_now));	/* seconds component of time */
	VRPNdata.time_useconds = (time_now - (vrTime)(VRPNdata.time_seconds)) * 1000000;	/* micro-seconds component of time */
	VRPNdata.code = 0;		/* code is the number of the device being reported -- for now we are just doing a single "FreeVR" device */

	/**********/
	/** pong **/
#ifdef VRPN_DUPLEX
	/* When the client pings, send a pong */
	if (aux->unanswered_ping)
#else
	/* Send a pong even with one-way communication, just in case we're pinged! */
	if (send_all)
#endif
	{
		vrDbgPrintfN(VRPN_DBGLVL FORCE_DBG, "_VrpnPutData(): About to send pong response to client.\n");
		VRPNdata.message_type = aux->MessageOutPong;	/* the description message type: pong */
		_VrpnServerSendMessage(devinfo, &VRPNdata, NULL, 0);
	}

	/******************/
	/** Text_Message **/
	/* NOTE: I've added this as part of my recording/playback of VRPN   */
	/*   data experiments -- this is sample meta-data of the recording. */
	if (send_all /* || (drand48() < 0.01)*/) {
		char		text_msg[128];
		time_t		time_nowL = time_now;	/* convert from double to long int */
		int32_t		level = 0;
		int32_t		severity = 0;

#  if __BYTE_ORDER == __LITTLE_ENDIAN
		vrEndianSwap4(&level);
		vrEndianSwap4(&severity);
#  endif
		memcpy(text_msg, &level, sizeof(level));
		memcpy(text_msg+4, &severity, sizeof(severity));
		vrPrintf("The time is now '%s'\n", ctime_r(&time_nowL, &(text_msg[8])));
		/* send a sample text message */
		vrDbgPrintfN(VRPN_DBGLVL FORCE_DBG, "_VrpnPutData(): About to send text message to client.\n");
		VRPNdata.message_type = aux->MessageOutText;	/* the description message type: text_message */
		_VrpnServerSendMessage(devinfo, &VRPNdata, text_msg, 1+8+strlen(&(text_msg[8])));	/* add 1 for the null terminator */

		/* send another text message with the application name */
		snprintf(&(text_msg[8]), sizeof(text_msg)-8, "FreeVR Application: %s", devinfo->context->name);
		_VrpnServerSendMessage(devinfo, &VRPNdata, text_msg, 1+8+strlen(&(text_msg[8])));
	}

	/*************/
	/** buttons **/
static	uint32_t	*button_data_last = NULL;	/* store previous data */
static	uint32_t	button_data_last_num = 0;	/* size of previous data array */
	int		button_data_allocated = 0;	/* flag for when the button array is allocated -- if so, send-all buttons */
	uint32_t	button_number;
	uint32_t	button_value;
	char		button_buffer[16] = "\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0";

	/* For the purposes of only sending changed values, we store old */
	/*   values in the "button_data_last" array, which has to be     */
	/*   allocated (or re-allocated) when the button count changes.  */
	if (button_data_last_num != aux->num_buttons) {
		button_data_last_num = aux->num_buttons;
		button_data_last = vrShmemAlloc0(button_data_last_num * sizeof(button_data_last[0]));
		button_data_allocated = 1;
	}

	for (count = 0; count < aux->num_buttons; count++) {
		button_number = count;
		button_value = vrGet2switchValueNoLastUpdate(count+1);

		/* individual buttons are sent if they are different from previous */
		if ((button_data_last[count] != button_value) || send_all || button_data_allocated) {
			button_data_last[count] = button_value;

#if __BYTE_ORDER == __LITTLE_ENDIAN
			vrEndianSwap4(&button_number);
			vrEndianSwap4(&button_value);
#endif

			/* prepare the buffer -- will be 8 bytes */
			memmove(&(button_buffer[0]), &button_number, sizeof(button_number));
			memmove(&(button_buffer[4]), &button_value, sizeof(button_value));

			/* now send the message to the client */
			vrDbgPrintfN(VRPN_DBGLVL, "_VrpnPutData(): About to send button %d\n", count);
			VRPNdata.message_type = aux->MessageOutButton;	/* the description message type: buttons */
			_VrpnServerSendMessage(devinfo, &VRPNdata, button_buffer, 8);
		}
	}

	/***************/
	/** valuators **/
	/* NOTE: all the valuators are sent as one complete bundle */
static	double		valuators_data_last[MAX_VALUATORS+1] = { 0.0 };
	double		valuators_data[MAX_VALUATORS+1];	/* adding 1 to put the count first */
	int		changes = 0;

	valuators_data[0] = aux->num_valuators;
#if __BYTE_ORDER == __LITTLE_ENDIAN
	vrEndianSwap8(&valuators_data[0]);
#endif
	for (count = 0; count < aux->num_valuators; count++) {
		valuators_data[count+1] = vrGetValuatorValueNoLastUpdate(count);
		if ((valuators_data[count+1] != valuators_data_last[count+1]) || send_all) {
			/* if any valuator has changed, that's enough to send the batch */
			changes = 1;
			valuators_data_last[count+1] = valuators_data[count+1];
		}
#if __BYTE_ORDER == __LITTLE_ENDIAN
		vrEndianSwap8(&valuators_data[count+1]);
#endif
	}

	/* now send the message to the client */
	if (changes) {
		vrDbgPrintfN(VRPN_DBGLVL, "_VrpnPutData(): About to send all valuators\n");
		VRPNdata.message_type = aux->MessageOutValuator;	/* the description message type: valuators */
		_VrpnServerSendMessage(devinfo, &VRPNdata, valuators_data, sizeof(double)*(count+1));
	}


	/***************/
	/** 6-sensors **/
static	vrMatrix	*sensor6_data_last = NULL;	/* store previous data */
static	uint32_t	sensor6_data_last_num = 0;	/* size of previous data array */
	int		sensor6_data_allocated = 0;	/* flag for when the 6-sensor array is allocated -- if so, send-all buttons */
	double		posquat_data[8];		/* the 7 values: X, Y, Z, QX, QY, QZ, QW (plus the sensor value in [0]) */
	int32_t		sensor;				/* the 4-byte (32bit) sensor value in the VRPN data packet */

	/* For the purposes of only sending changed values, we store old */
	/*   values in the "sensor6_data_last" array, which has to be    */
	/*   allocated (or re-allocated) when the 6-sensor count changes.*/
	if (sensor6_data_last_num != aux->num_6sensors) {
		sensor6_data_last_num = aux->num_6sensors;
		sensor6_data_last = vrShmemAlloc0(sensor6_data_last_num * sizeof(sensor6_data_last[0]));
		sensor6_data_allocated = 1;
	}

	for (count = 0; count < aux->num_6sensors; count++) {
		vrMatrix	sensor_mat;			/* a transitional place to get the sensor data */
		vrQuat		sensor_quat;			/* a place to put the quaternion version of the rotation */

		/* extract the actual matrix for this sensor */
		vrMatrixGet6sensorValuesNoLastUpdate(&sensor_mat, count);

		/* individual 6sensors are sent if they are different from previous */
		if ((!vrMatrixEqual(&(sensor6_data_last[count]), &sensor_mat)) || send_all || sensor6_data_allocated) {
			sensor6_data_last[count] = sensor_mat;

			/* modify the matrix based on configuration parameters     */
			/* NOTE: typically this would be the identity operation,   */
			/*   but for some special cases (namely ParaView), it's    */
			/*   beneficial to be able to move the coordinate system.  */
			/* NOTE: this operation uses the "t2rw_xform" field,       */
			/*   but when specified in a configuration file, the       */
			/*   "frame-of-reference" synonym will generally be used   */
			/*   instead (e.g. "for_translate").                       */
#if 0 /* set to "1" to hardcode (delete me when ready) */
			/* TODO: for the moment (while developing Immersive ParaView),*/
			/*    I'm going to just hard-code the transformation.         */
			VRMAT_ROWCOL(&sensor_mat, VR_Y, VR_W) -= 5.0;
#else
			vrMatrixPreMult(&sensor_mat, devinfo->t2rw_xform);
#endif

			/* extract the rotation from the new matrix as a quaternion */
			vrQuatSetFromMatrix(&sensor_quat, &sensor_mat);

			sensor = count;							/* convert to a int32_t */
			memcpy(&((int32_t*)posquat_data)[0], &sensor, sizeof(sensor));	/* put into the proper 4 bytes */
			memset(&((int32_t*)posquat_data)[1], '\0', 4);			/* put 4 0x00 bytes next */
			//memcpy(&posquat_data[0], &sensor, sizeof(sensor));		/* bad version (thought it was 8 bytes) (delete me) */
			posquat_data[VR_X+1] = VRMAT_ROWCOL(&sensor_mat, VR_X, VR_W);
			posquat_data[VR_Y+1] = VRMAT_ROWCOL(&sensor_mat, VR_Y, VR_W);
			posquat_data[VR_Z+1] = VRMAT_ROWCOL(&sensor_mat, VR_Z, VR_W);
			posquat_data[VR_QX+1] = sensor_quat.v[VR_X];
			posquat_data[VR_QY+1] = sensor_quat.v[VR_Y];
			posquat_data[VR_QZ+1] = sensor_quat.v[VR_Z];
			posquat_data[VR_QW+1] = sensor_quat.v[VR_W];
#if __BYTE_ORDER == __LITTLE_ENDIAN
			vrEndianSwap4(posquat_data);
			vrEndianSwap8(&posquat_data[VR_X+1]);
			vrEndianSwap8(&posquat_data[VR_Y+1]);
			vrEndianSwap8(&posquat_data[VR_Z+1]);
			vrEndianSwap8(&posquat_data[VR_QX+1]);
			vrEndianSwap8(&posquat_data[VR_QY+1]);
			vrEndianSwap8(&posquat_data[VR_QZ+1]);
			vrEndianSwap8(&posquat_data[VR_QW+1]);
#endif

			/* now send the message to the client */
			vrDbgPrintfN(VRPN_DBGLVL, "_VrpnPutData(): About to send sensor %d\n", count);
			VRPNdata.message_type = aux->MessageOutSensorPos;	/* the description message type: 6-sensors */
			VRPNdata.code = 0;				/* the number of the 6-sensor's device (always 0 for now) */
			_VrpnServerSendMessage(devinfo, &VRPNdata, posquat_data, sizeof(double)*8);
		}
	}

	vrTrace("_VrpnPutData():", BOLD_TEXT "Sent the data to client." NORM_TEXT);
}


#if 0 /* I don't think I'm going to want this stuff for the "output" function, but I'll leave the code in place for now (VruiDD) [04/11/2007] { */
	/* [09/25/13: actually, it might now come in handy for specifying the virtual devices */
	/*   of protocol-2.  Not sure yet, we'll see when the time comes to implement them.]  */

	/*****************************************************************/
	/*   Function(s) for parsing Shared Memory "input" declarations. */
	/*                                                               */
	/*  These _Vrpn<type>Input() functions are called during the   */
	/*  CREATE phase of the input interface.                         */

/* ----------------------------------------------------------------------- */
/* ----------------------------------------------------------------------- */

/**************************************************************************/
/* type might be "char", "int", or "long" (and in the future "float" or "double") */
/*    parsing:                                                                    */
/*       input "<name>" = "2switch(<type>[<offset>])";                            */
/*    thus:                                                                       */
/*       the device string is unused.                                             */
/*       the type string is <type> (eg. "char", "int", etc)                       */
/*       the instance string is <offset> (eg. 0, 4, 6)                            */
/**************************************************************************/
static vrInputMatch _VrpnButtonInput(vrInputDevice *devinfo, vrGenericInput *input, vrInputDTI *dti)
{
	_VrpnPrivateInfo	*aux = (_VrpnPrivateInfo *)devinfo->aux_data;
	int			button_num;
	int			offset;
	_MemoryType		memtype;

	/* determine current button in our private arrays of button data */
	button_num = aux->num_buttons;
	aux->num_buttons++;

	memtype = _MemoryValue(dti->type);
	offset = vrAtoI(dti->instance);

	vrDbgPrintfN(INPUT_DBGLVL, "_VrpnButtonInput(): Assigned button event of button %d, to type '%s'(%d) with offset %d\n",
		button_num, _MemoryTypeName(memtype), (int)memtype, offset);

	if (memtype == MEM_UNKNOWN) {
		vrErrPrintf("_VrpnValuatorInput: " RED_TEXT "Warning, type['%s'] did not match any known type\n" NORM_TEXT, dti->type);
		return VRINPUT_NOMATCH;	/* unacceptable input declaration match, perhaps keep searching */
	}

#if 0 /* NOTE: this check isn't possible here because we haven't connected with the memory yet to know the size */
	if (offset 'out of bounds') {
		...
		return VRINPUT_NOMATCH;	/* unacceptable input declaration match, perhaps keep searching */
	}
#endif

	aux->button_inputs[button_num] = (vr2switch *)input;
	aux->button_values[button_num] = 0;
	aux->button_dummy[button_num] = 0;
	aux->button_offsets[button_num] = offset;

	return VRINPUT_MATCH_ABLE;	/* input declaration match */
}


/**************************************************************************/
static vrInputMatch _VrpnValuatorInput(vrInputDevice *devinfo, vrGenericInput *input, vrInputDTI *dti)
{
	_VrpnPrivateInfo	*aux = (_VrpnPrivateInfo *)devinfo->aux_data;
	int			valuator_num;
	int			offset;
	_MemoryType		memtype;

	/* determine current valuator in our private arrays of valuator data */
	valuator_num = aux->num_valuators;
	aux->num_valuators++;

	memtype = _MemoryValue(dti->type);
	offset = vrAtoI(dti->instance);

	vrDbgPrintfN(INPUT_DBGLVL, "_VrpnValuatorInput(): Assigned button event of button %d, to type '%s'(%d) with offset %d\n",
		valuator_num, _MemoryTypeName(memtype), (int)memtype, offset);

	if (memtype == MEM_UNKNOWN) {
		vrErrPrintf("_VrpnValuatorInput: " RED_TEXT "Warning, type['%s'] did not match any known type\n" NORM_TEXT, dti->type);
		return VRINPUT_NOMATCH;	/* unacceptable input declaration match, perhaps keep searching */
	}

#if 0 /* NOTE: this check isn't possible here because we haven't connected with the memory yet to know the size */
	if (offset 'out of bounds') {
		...
		return VRINPUT_NOMATCH;	/* unacceptable input declaration match, perhaps keep searching */
	}
#endif

	aux->valuator_inputs[valuator_num] = (vrValuator *)input;
	aux->valuator_values[valuator_num] = 0;
	aux->valuator_dummy[valuator_num] = 0.0;
	aux->valuator_offsets[valuator_num] = offset;

	return VRINPUT_MATCH_ABLE;		/* input declaration match */
}


/**************************************************************************/
static vrInputMatch _Vrpn6sensorInput(vrInputDevice *devinfo, vrGenericInput *input, vrInputDTI *dti)
{
	_VrpnPrivateInfo	*aux = (_VrpnPrivateInfo *)devinfo->aux_data;
	int			sensor6_num;
	int			offset;
	_MemoryType		memtype;

	/* determine current 6-sensor in our private arrays of 6-sensor data */
	sensor6_num = aux->num_6sensors;
	aux->num_6sensors++;

	memtype = _MemoryValue(dti->type);
	offset = vrAtoI(dti->instance);

	vrDbgPrintfN(INPUT_DBGLVL, "_Vrpn6sensorInput(): Assigned 6-sensor event of 6-sensor %d, to type '%s'(%d) with offset %d\n",
		sensor6_num, _MemoryTypeName(memtype), (int)memtype, offset);

	if (memtype == MEM_UNKNOWN) {
		vrErrPrintf("_VrpnValuatorInput: " RED_TEXT "Warning, type['%s'] did not match any known type\n" NORM_TEXT, dti->type);
		return VRINPUT_NOMATCH;	/* unacceptable input declaration match, perhaps keep searching */
	}

#if 0 /* NOTE: this check isn't possible here because we haven't connected with the memory yet to know the size */
	if (offset 'out of bounds') {
		...
		return VRINPUT_NOMATCH;	/* unacceptable input declaration match, perhaps keep searching */
	}
#endif

	aux->sensor6_inputs[sensor6_num] = (vr6sensor *)input;
	vrMatrixSetIdentity(&aux->sensor6_values[sensor6_num]);
	aux->sensor6_dummy[sensor6_num] = 0;
	aux->sensor6_offsets[sensor6_num] = offset;

	vrAssign6sensorR2Exform(aux->sensor6_inputs[sensor6_num], strchr(dti->instance, ','));

	return VRINPUT_MATCH_ABLE;	/* input declaration match */
}

#endif /* } */


	/************************************************************/
	/***************** FreeVR Callback routines *****************/
	/************************************************************/

	/************************************************************/
	/*    Callbacks for controlling the device's features.      */
	/*                                                          */

/************************************************************/
static void _VrpnSystemPauseToggleCallback(vrInputDevice *devinfo, int value)
{
	if (value == 0)
		return;

vrPrintf("Pausing VRPN output!\n");
	devinfo->context->paused ^= 1;
	vrDbgPrintfN(INPUT_DBGLVL, "VRPN Control: system_pause = %d.\n",
		devinfo->context->paused);
}

/************************************************************/
static void _VrpnPrintContextStructCallback(vrInputDevice *devinfo, int value)
{
	if (value == 0)
		return;

	vrFprintContext(stdout, devinfo->context, verbose);
}

/************************************************************/
static void _VrpnPrintConfigStructCallback(vrInputDevice *devinfo, int value)
{
	if (value == 0)
		return;

	vrFprintConfig(stdout, devinfo->context->config, verbose);
}

/************************************************************/
static void _VrpnPrintInputStructCallback(vrInputDevice *devinfo, int value)
{
	if (value == 0)
		return;

	vrFprintInput(stdout, devinfo->context->input, verbose);
}

/************************************************************/
static void _VrpnPrintPrivateInfoCallback(vrInputDevice *devinfo, int value)
{
	_VrpnPrivateInfo  *aux = (_VrpnPrivateInfo *)devinfo->aux_data;

	if (value == 0)
		return;

	_VrpnPrintPrivateInfo(stdout, aux, verbose);
}

/************************************************************/
static void _VrpnPrintHelpCallback(vrInputDevice *devinfo, int value)
{
	_VrpnPrivateInfo  *aux = (_VrpnPrivateInfo *)devinfo->aux_data;

	if (value == 0)
		return;

	_VrpnPrintHelp(stdout, aux);
}

/************************************************************/
static void _VrpnPrintValuesCallback(vrInputDevice *devinfo, int value)
{
	_VrpnPrivateInfo  *aux = (_VrpnPrivateInfo *)devinfo->aux_data;

	if (value == 0)
		return;

	_VrpnPrintValues(stdout, aux);
}




	/*************************************************/
	/*   Callbacks for interfacing with the device.  */
	/*                                               */

/**********************************************************/
static void _VrpnCreateFunction(vrInputDevice *devinfo)
{
#if 0 /* There's a chance I'll want to control the outputs similar to the inputs, so I'll leave this here for now { */
	/*** List of possible inputs ***/
static	vrInputFunction	_VrpnInputs[] = {
			/* TODO: ideally, we should be able to catch all 2-switches with: { NULL, VRINPUT_BINARY, _VrpnButtonInput }, */
				{ "char", VRINPUT_BINARY, _VrpnButtonInput },
				{ "int",  VRINPUT_BINARY, _VrpnButtonInput },
				{ "long", VRINPUT_BINARY, _VrpnButtonInput },
			/* TODO: a "float/double" binary is also conceivable, perhaps with an epsilon */
				{ "float",  VRINPUT_VALUATOR, _VrpnValuatorInput },
				{ "double", VRINPUT_VALUATOR, _VrpnValuatorInput },
			/* TODO: valuators based on integer values are also conceivable */
				{ "vreuler",       VRINPUT_6SENSOR, _Vrpn6sensorInput },
				{ "vrmatrix",      VRINPUT_6SENSOR, _Vrpn6sensorInput },
				{ "vreulerfloat",  VRINPUT_6SENSOR, _Vrpn6sensorInput },
				{ "vrmatrixdouble",VRINPUT_6SENSOR, _Vrpn6sensorInput },
				{ "vreulerdouble", VRINPUT_6SENSOR, _Vrpn6sensorInput },
				{ "vrmatrixfloat", VRINPUT_6SENSOR, _Vrpn6sensorInput },
			/* TODO: I'd rather do without these, but needed to get visbox working ASAP */
				{ "visboxeuler",   VRINPUT_6SENSOR, _Vrpn6sensorInput },
				{ "visboxtrans",   VRINPUT_6SENSOR, _Vrpn6sensorInput },
				{ NULL, VRINPUT_UNKNOWN, NULL } };
#endif /* } */
	/*** List of control functions ***/
static	vrControlFunc	_VrpnControlList[] = {
				/* overall system controls */
				{ "system_pause_toggle", _VrpnSystemPauseToggleCallback },

				/* informational output controls */
				{ "print_context", _VrpnPrintContextStructCallback },
				{ "print_config",  _VrpnPrintConfigStructCallback },
				{ "print_input",   _VrpnPrintInputStructCallback },
				{ "print_devinfo",  _VrpnPrintPrivateInfoCallback },
				{ "print_help",    _VrpnPrintHelpCallback },
				{ "print_values",  _VrpnPrintValuesCallback },

		/** TODO: other callback control functions go here **/
				/* simulated 6-sensor selection controls */
				/* simulated 6-sensor manipulation controls */
				/* other controls */

				/* end of the list */
				{ NULL, NULL } };

	_VrpnPrivateInfo	*aux = NULL;

	/******************************************/
	/* allocate and initialize auxiliary data */
	devinfo->aux_data = (void *)vrShmemAlloc0(sizeof(*aux));
	aux = (_VrpnPrivateInfo *)devinfo->aux_data;
	_VrpnInitializeStruct(aux);
//_VrpnPrintPrivateInfo(stdout, aux, verbose);
	aux->devinfo = devinfo;		/* [1/31/00] -- we don't normally do this, but are experimenting */

	/******************/
	/* handle options */
	_VrpnParseArgs(aux, devinfo->args);


	/* ?? Here is some hard-coding of what would typically be done in most setups. */
	/*   I.e. output all the buttons (except button 0), all the */
	/*   valuators, and all the 6-sensors.                      */
	/* TODO: these values don't seem to get filled in at this point -- I was */
	/*   thinking that since this "input" device would be last in the list,  */
	/*   that the values would be known by the time it is initialized, but I */
	/*   guess that's not the case, so I'm going to set the value at the top */
	/*   of the processing routine _VrpnPutData().                         */
	/*   08/31/21: TODO: so does this mean I can get rid of these lines?  Probably num_buttons is 1 too high anyway. */ /* maybe delete me */
	aux->num_buttons = devinfo->context->input->num_2ways;
	aux->num_valuators = devinfo->context->input->num_valuators;
	aux->num_6sensors = devinfo->context->input->num_6sensors;
	vrDbgPrintfN(VRPN_DBGLVL, "_VrpnCreateFunction(): presently set values for: num_buttons = %d, num_valuators = %d, num_6sensors = %d.\n", aux->num_buttons, aux->num_valuators, aux->num_6sensors);

	vrDbgPrintf("_VrpnCreateFunction(): Done creating Vrpn outputs for '%s'\n", devinfo->name);
	devinfo->created = 1;

	return;
}


/***************************************************************************/
/* _VrpnOpenFunction(): this callback opens a listening port for clients   */
/*    to initiate a connection for receiving data.                         */
/*                                                                         */
/* There are no return values, but there are two input-device flags that   */
/*    are set to indicate the state of the device:                         */
/*      * open -- indicates when the communications channel is listening   */
/*      * operating -- indicates the server is actually ready to send data */
static void _VrpnOpenFunction(vrInputDevice *devinfo)
{
struct	linger			linger_opt = { 1, 0 };	/* values for the SO_LINGER control options */
	char			host_machine[1024];
	int			host_port;
	_VrpnPrivateInfo	*aux = (_VrpnPrivateInfo *)devinfo->aux_data;
	int			return_value;	/* return value for other calls -- eg. fsync() */

	/****************************************/
	/* open the device (a listening socket) */
	gethostname(host_machine, 1024);
	host_port = aux->dd_port;
	vrTrace("_VrpnOpenFunction():", BOLD_TEXT "Creating the listen socket." NORM_TEXT);
	aux->fd_listen_socket = vrSocketCreateListen(&host_port, 0);

	if (aux->fd_listen_socket < 0) {
		vrDbgPrintfN(AALWAYS_DBGLVL, "_VrpnOpenFunction(): " RED_TEXT "FAILURE: Unable to open socket %d.\n" NORM_TEXT, aux->dd_port);
		snprintf(devinfo->version, strlen(devinfo->version), "- unconnected vrpn -");
		aux->open = 0;
		devinfo->operating = 0;
		vrTrace("_VrpnOpenFunction():", BOLD_TEXT "Unable to create listen socket." NORM_TEXT);
	} else {
		vrMsgPrintf("VRPN Output: Connect to VRPN port %s@tcp://%s:%d\n", "FreeVR", host_machine, host_port);
		aux->open = 1;
		devinfo->operating = 1;
		vrTrace("_VrpnOpenFunction():", BOLD_TEXT "Listen socket created." NORM_TEXT);

		return_value = setsockopt(aux->fd_listen_socket, SOL_SOCKET, SO_LINGER, (void *)&linger_opt, sizeof(linger_opt));
	}

	vrDbgPrintf("_VrpnOpenFunction(): Done opening VRPN for output device '%s'\n", devinfo->name);


	/*******************************************/
	/* Print output at appropriate debug level */
	if (vrDbgDo(VRPN_DBGLVL)) {
		_VrpnPrintPrivateInfo(stdout, aux, brief);
		_VrpnPrintValues(stdout, aux);
	}

	return;
}


/***************************************************************************/
/* _VrpnCloseFunction(): this callback closes the communications channel   */
/*   suspending current communications and no longer listening for new     */
/*   connections.                                                          */
static void _VrpnCloseFunction(vrInputDevice *devinfo)
{
	_VrpnPrivateInfo	*aux = (_VrpnPrivateInfo *)devinfo->aux_data;

	vrDbgPrintf("Closing VRPN sockets %d:%d\n", aux->fd_listen_socket, aux->fd_receive_socket);
	if (aux->fd_listen_socket > 0)
		vrSocketClose(aux->fd_listen_socket);
	if (aux->fd_receive_socket > 0)
		vrSocketClose(aux->fd_receive_socket);
	return;
}


/*************************************************************************/
/* _VrpnResetFunction(): this callback provides the means by which the   */
/*    device or input stream can be reset to a good state.               */
static void _VrpnResetFunction(vrInputDevice *devinfo)
{
	/* TODO: a VRPN reset should verify that the VRPN server is sending */
	/*   the same amount of data, and if not, update the vrInputDevice. */

	return;
}


/************************************************************************/
/* _VrpnPollFunction(): this callback is called inside a FreeVR "input" */
/*   process for each I/O "frame".  For VRPN it checks for any new      */
/*   connection attempts, and for open connections is sends out the     */
/*   requested data.                                                    */
static void _VrpnPollFunction(vrInputDevice *devinfo)
{
	_VrpnPrivateInfo	*aux = (_VrpnPrivateInfo *)devinfo->aux_data;
	int			new_connection = 0;	/* flag to indicate all data should be sent */
	char			inbuf[512];	/* buffer for data received from the VRPN client */
	int			bytes_read = 0;	/* number of bytes read from socket thus far */
	int			read_result;	/* return value of the call to "read()" */
	int			return_value;	/* return value for other calls -- eg. fsync() */
	struct timeval		no_block = { 0, 0 };	/* set the time values for the no-blocking option */
	fd_set			fds;

	vrTrace("_VrpnPollFunction():", BOLD_TEXT "at top." NORM_TEXT);
	/* TOOD: test for devinfo->operating */
#if 0	/* this is just for very low-level (initial) testing */
vrDbgPrintfN(VRPN_DBGLVL, "VRPN polling: operating = %d, vrpn_link = %d\n", devinfo->operating, aux->vrpn_link);
#endif

#if 0
_VrpnPrintPrivateInfo(stdout, aux, brief);
#endif
	/* First check for an existing connection, and if not, check for a new connection */
	/* 08/31/21: TODO: actually, for VRPN we need to always check for possible new connections -- we can have more than one */
	if (!aux->vrpn_link) {
		/* not connected means listen for a connection */
		if (aux->fd_receive_socket < 0 && aux->fd_listen_socket >= 0) {
			/* fd_receive_socket will be -1 when no one is connected */
			aux->fd_receive_socket = vrSocketAnswer(aux->fd_listen_socket);
			if (aux->fd_receive_socket > 0) {
				vrDbgPrintfN(AALWAYS_DBGLVL, "FreeVR: VRPN connection has been established with client.\n");
#if 0
/* 08/31/21: what was this testing in the VruiDD code?  TODO: test the necessity of this in VRPN (then maybe delete me) */
int flag = 1;
setsockopt(aux->fd_listen_socket, IPPROTO_TCP, TCP_NODELAY, &flag, sizeof(flag));
#endif

				/* New connection established, so send the version "cookie" */
				vrTrace("_VrpnPollFunction():", BOLD_TEXT "new connection." NORM_TEXT);
				return_value = write(aux->fd_receive_socket, aux->version, 24);    /* NOTE: not using vrSocketSendMsg() because the length is fixed, including several NULL terminator characters */
				//return_value = fsync(aux->fd_receive_socket); //-- this should work, but maybe it doesn't [10/28/21: trying this here.]
				if (return_value != 24) {
					vrDbgPrintfN(AALWAYS_DBGLVL, "_VrpnOpenFunction(): " RED_TEXT "An error occurred while trying to send VRPN hail message - %d bytes sent instead of %d.\n" NORM_TEXT, return_value, 24);
					if (return_value < 0)
						perror("_VrpnOpenFunction(): write()");
				}

				/* Now wait for the reply */
				vrSocketAwaitData(aux->fd_receive_socket);		/* NOTE: currently hard-coded to wait .15 second */
#ifdef VRPN_DUPLEX /* 10/28/21: test code to use the _VrpnPollFunction() function from vr_input.vrpn.c */
				//read_result = (ssize_t)read(aux->fd_receive_socket, inbuf, 24);
				read_result = _VrpnReadInput(aux);	/* read from the client */
				vrDbgPrintfN(VRPN_DBGLVL FORCE_DBG, "VRPN: from _VrpnReadInput(): got (%d bytes) '%s'\n", read_result, inbuf);
				vrTrace("_VrpnPollFunction():", BOLD_TEXT "first responses read from client." NORM_TEXT);
#else /* 10/28/21: original code */
				read_result = (ssize_t)read(aux->fd_receive_socket, inbuf, sizeof(inbuf));
				if (read_result <= 0) {
					return;				/* Didn't get a quick response -- return w/o setting the open flag */
				}

				/* Parse the version number */
			/* TODO: actually parse the data, don't just print it -- but for now we assume we're good */
				vrDbgPrintfN(VRPN_DBGLVL FORCE_DBG, "VRPN: got (%d bytes) '%s'\n", read_result, inbuf);
#endif
				aux->vrpn_link = 1;			/* Indicate we're now communicating with a VRPN client */

				/* Send information about what *this* VRPN server handles */
				_VrpnNewClientConnectionData(devinfo);

				aux->streaming = 1;			/* Indicate that we are now streaming data */

				/* The number of buttons/valuators/trackers/etc    */
				/*   now needs to be re-assigned since the initial */
				/*   setting occurs during a race for the input    */
				/*   device creation.                              */
				aux->num_buttons = devinfo->context->input->num_2ways - 1;	/* NOTE: 1 is subtracted because we won't send the Esc key */
				aux->num_valuators = devinfo->context->input->num_valuators;
				aux->num_6sensors = devinfo->context->input->num_6sensors;

				new_connection = 1;
			}
		}
	}

	if (aux->fd_receive_socket > 0) {
		vrTrace("_VrpnPollFunction():", BOLD_TEXT "reading from client." NORM_TEXT);
		/* if we have a connection, see whether there is a message (or disconnect) from the client */
		FD_ZERO(&fds);
		FD_SET(aux->fd_receive_socket, &fds);
		if (select(aux->fd_receive_socket + 1, &fds, NULL, NULL, &no_block)) {

			/**********************************/
			/* Read data from the VRPN client */

			/* TODO: For VRPN I'm not sure the client actually sends data back to the server */
			/* But we do at least need to determine whether */
			/*   the connection is still active.            */

			bytes_read = 0;
			while ((bytes_read < 2) && (aux->fd_receive_socket > 0)) {
#ifdef VRPN_DUPLEX
				read_result = _VrpnReadInput(aux);
#else
				read_result = (ssize_t)read(aux->fd_receive_socket, inbuf, sizeof(inbuf));
#endif
				if (read_result == 0) {
					/* this occurs when the other side has disconnected */
					aux->vrpn_link = 0;		/* indicate we're not communicating with a VRPN client */
					aux->streaming = 0;		/* set streaming to default off-state */
					aux->fd_receive_socket = -1;	/* indicate that there's no one on the other side of the socket */
					vrDbgPrintfN(AALWAYS_DBGLVL, "FreeVR: VRPN client has disconnected.\n", bytes_read);
					vrTrace("_VrpnPollFunction():", BOLD_TEXT "client disconnected." NORM_TEXT);
				} else if (read_result > 0) {
					bytes_read += read_result;
				}
if (bytes_read > 0)
				vrDbgPrintfN(VRPN_DBGLVL FORCE_DBG, "_VrpnPollFunction(): %d bytes read thus far.\n", bytes_read);
#ifdef COMM_DEBUG
				vrDbgPrintf("_VrpnPollFunction(): %d bytes read thus far.\n", bytes_read);
#endif
			}
		}
	}

	/* Now if there is a connection, and we're to be streaming the data -- send it out */
	if (aux->vrpn_link && aux->streaming) {
		vrTrace("_VrpnPollFunction():", BOLD_TEXT "about to send data." NORM_TEXT);
		_VrpnPutData(devinfo, new_connection);
		vrTrace("_VrpnPollFunction():", BOLD_TEXT "sent data." NORM_TEXT);
	}

	return;
}



	/******************************/
	/*** FreeVR public routines ***/
	/******************************/


/**********************************************************/
void vrVrpnOutInitInfo(vrInputDevice *devinfo)
{
	devinfo->version = (char *)vrShmemStrDup("VRPN out-dev 0.1");
	devinfo->Create = vrCallbackCreateNamed("VRPN:Create-Def", _VrpnCreateFunction, 1, devinfo);
	devinfo->Open = vrCallbackCreateNamed("VRPN:Open-Def", _VrpnOpenFunction, 1, devinfo);
	devinfo->Close = vrCallbackCreateNamed("VRPN:Close-Def", _VrpnCloseFunction, 1, devinfo);
	devinfo->Reset = vrCallbackCreateNamed("VRPN:Reset-Def", _VrpnResetFunction, 1, devinfo);
	devinfo->PollData = vrCallbackCreateNamed("VRPN:PollData-Def", _VrpnPollFunction, 1, devinfo);
	devinfo->PrintAux = vrCallbackCreateNamed("VRPN:PrintAux-Def", _VrpnPrintPrivateInfo, 0);
}

