/* ======================================================================
 *
 *  CCCCC          vr_parse.c
 * CC   CC         Author(s): Bill Sherman
 * CC              Created: October 12, 1998
 * CC   CC         Last Modified: February 23, 2024 (comment additions 02/29/24)
 *  CCCCC
 *
 * Code file for various parsing routines.  There are two styles of parsing
 *   implemented herein.  First, there are routines for parsing the
 *   configuration file; and second, there are also a different collection
 *   of functions for parsing argument options, as well as commands sent
 *   via sockets.
 *
 * Copyright 2024, Bill Sherman, All rights reserved.
 * With the intent to provide an open-source license to be named later.
 * ====================================================================== */
#include <stdio.h>		/* needed for printf & all other standard I/O */
#include <stdlib.h>		/* for atof() */
#include <string.h>		/* for strlen/strdup/strspn/strcspn/strstr/strcasecmp */

#include "vr_parse.h"
#include "vr_debug.h"
#include "vr_input.h"
#include "vr_utils.h"		/* for vrAtoI() */

static vrTokenInfo _ParseJustNextToken(vrParseInfo *parse);

/** Set this to see every token as it's parsed (also requires PARSE_DBGLVL) **/
#undef PRINT_ALL_PARSED_TOKENS

/** Set this to get warnings during configuration when a function without trailing parens **/
#define WARN_ABOUT_CONFIG_FUNCTIONS_WO_PARENS

/** Set this to enable concatenation of string tokens in config-file parsing **/ /* Probably should just make this permanent -- so perhaps delete me and also it's usage */
#define PARSE_CONCAT_STRINGS


/***************************************************************/
/* _TokenIsTerminal() takes a token code and returns a boolean */
/*   that indicates whether a token is a statement terminator. */
static int _TokenIsTerminal(int tokentype)
{
	switch (tokentype) {
	case VRTOKEN_SEMICOLON:
	case VRTOKEN_CLOSECURL:
	case VRTOKEN_EOF:
	case VRTOKEN_EOS:	/* this is special for when skipping to an EOS*/
		return 1;
	default:
		return 0;
	}
}


/*****************************************************************/
/* _TokenIsValue() takes a token code and returns a boolean that */
/*   indicates whether a token is a value specifier.             */
static int _TokenIsValue(int tokentype)
{
	switch (tokentype) {
	case VRTOKEN_NUMBER:
	case VRTOKEN_STRING:
	case VRTOKEN_NIL:
		return 1;
	default:
		return 0;
	}
}


/****************************************************************************/
/* _TokenIsAssignment() takes a token code and returns a boolean that       */
/*   indicates whether a token is an assignment operator ("=", "+=", "*="). */
static int _TokenIsAssignment(int tokentype)
{
	switch (tokentype) {
	case VRTOKEN_SETEQUALS:
	case VRTOKEN_PLUSEQUALS:
	case VRTOKEN_TIMESEQUALS:
		return 1;
	default:
		return 0;
	}
}


/*********************************************************************/
/* _TokenIsBiMathOperator() takes a token code and returns a boolean */
/*   that indicates whether a token is a bi-operand math operator.   */
static int _TokenIsBiMathOperator(int tokentype)
{
	if ((tokentype > VRTOKEN_BIOP_BEGIN) && (tokentype < VRTOKEN_BIOP_END))
		return 1;
	else	return 0;
}


/**********************************************************************/
/* _TokenIsUniMathOperator() takes a token code and returns a boolean */
/*   that indicates whether a token is a uni-operand math operator.   */
/* NOTE: currently we don't handle any such operators, so always      */
/*   returns false.                                                   */
static int _TokenIsUniMathOperator(int tokentype)
{
	return 0;
}


/*********************************************************************/
/* _ParseStringAsNumber(): convert certain string inputs to numbers. */
/*   Either returns a token for the list of special strings, or      */
/*   returns the special "unknown" token.                            */
static int _ParseStringAsNumber(char *str)
{
	if (!strcasecmp(str, "default"))	return VRTOKEN_DEFAULT;
	else if (!strcasecmp(str, "no"))	return VRTOKEN_NO;
	else if (!strcasecmp(str, "yes"))	return VRTOKEN_YES;
	else if (!strcasecmp(str, "false"))	return VRTOKEN_FALSE;
	else if (!strcasecmp(str, "true"))	return VRTOKEN_TRUE;
	else if (!strcasecmp(str, "off"))	return VRTOKEN_OFF;
	else if (!strcasecmp(str, "on"))	return VRTOKEN_ON;

	else
		return VRTOKEN_UNKNOWN;
}


/**************************************************************/
/* vrTokenStringRepresentation(): convert certain token types */
/*   into string representations that can be output to error  */
/*   and informational messages.                              */
/* NOTE: thus far this only includes tokens known to be output */
/*   in error and info messages.  It can be expanded as needed.*/
char *vrTokenStringRepresentation(vrToken token_type)
{
	switch (token_type) {
	case VRTOKEN_DEFAULT:
		return "default";

	case VRTOKEN_NO:	/* NOTE: same as VRTOKEN_FALSE & VRTOKEN_OFF */
		return "no";

	case VRTOKEN_YES:	/* NOTE: same as VRTOKEN_TRUE & VRTOKEN_ON */
		return "yes";

	default:
		return "<token>";
	}
}


/************************************************************************/
/* _NumericStringLength(): Determine the number of numeric characters   */
/*   that begin the given string.  A result of 0 means that the string  */
/*   does not begin with a number, so this function can be used to tell */
/*   whether a string begins with a number, and if so, the length (not  */
/*   counting a closing '\0') that can then be used to copy the numeric */
/*   string to another string.                                          */
static int _NumericStringLength(char *string)
{
static	char	*numchars = "-.0123456789";
static	char	*hexnumchars = "-.0123456789ABCDEFabcdef";
static	char	*numradix = "boxd";
	int	numlen;

	if (strspn(string, numchars) > 0) {
		if ((string[0] == '0') && (strspn(&string[1], numradix) > 0)) {
			/* this number has a radix */
			/* NOTE: we know there must be at least two array elements */
			/*   left in string because the first one isn't '\0', and  */
			/*   there must be one character that is '\0'.             */
			if (string[1] == 'x')
				numlen = strspn(string+2, hexnumchars) + 2;
			else	numlen = strspn(string+2, numchars) + 2;
		} else if ((string[0] == '-') && (string[1] == '0') && (strspn(&string[2], numradix) > 0)) {
			if (string[2] == 'x')
				numlen = strspn(string+3, hexnumchars) + 3;
			else	numlen = strspn(string+3, numchars) + 3;
		} else {
			numlen = strspn(string, numchars);
		}

		return (numlen);
	} else {
		return (0);
	}
}


/*******************************************************************/
/* _TokenNumber(): The next input is known to be a numeric string, */
/*   so parse that into a token -- which is returned.              */
static vrTokenInfo _TokenNumber(char *tokstring, vrParseInfo *parse)
{
	vrTokenInfo	tok = { VRTOKEN_UNKNOWN, "<unk>" };

	tok.token = VRTOKEN_NUMBER;
	tok.string = strdup(tokstring);
	tok.linenum = parse->linenum;

	return (tok);
}


/*****************************************************************/
/* _TokenString(): The next input is known to be an alphanumeric */
/*   string, so parse that into a token -- which is returned.    */
static vrTokenInfo _TokenString(char *tokstring, vrParseInfo *parse)
{
	vrTokenInfo	tok = { VRTOKEN_UNKNOWN, "<unk>" };
	int		count;

	tok.token = VRTOKEN_STRING;
	tok.string = strdup(tokstring);
	tok.linenum = parse->linenum;

	for (count = 0; count < parse->nummaps; count++) {
		if (!strcasecmp(parse->tsmap_list[count].string, tokstring)) {
			tok.token = parse->tsmap_list[count].token;
			tok.string = strdup(tokstring);

			return (tok);
		}
	}

	return (tok);
}


/************************************************************/
/* _TokenPunctuation(): The next input (character) is known */
/*   to be a punctuation  character, so figure out what the */
/*   token is -- and return that value.                     */
static vrTokenInfo _TokenPunctuation(char tokchar, vrParseInfo *parse)
{
	vrTokenInfo	tok = { VRTOKEN_UNKNOWN, "<unk>" };

//vrPrintf("TEMP: _TokenPunctuation(): char '%c', line = %d\n", tokchar, parse->linenum);
	tok.linenum = parse->linenum;

	switch (tokchar) {
	case '{':
		tok.token = VRTOKEN_OPENCURL;
		tok.string = "{";
		break;
	case '}':
		tok.token = VRTOKEN_CLOSECURL;
		tok.string = "}";
		break;
	case ',':
		tok.token = VRTOKEN_COMMA;
		tok.string = ",";
		break;
	case ';':
		tok.token = VRTOKEN_SEMICOLON;
		tok.string = ";";
		break;
	case '"':
		tok.token = VRTOKEN_QUOTE;
		tok.string = "\"";
		break;
	case '(':
		tok.token = VRTOKEN_OPENPAREN;
		tok.string = "(";
		break;
	case ')':
		tok.token = VRTOKEN_CLOSEPAREN;
		tok.string = ")";
		break;
	}

//vrPrintf("TEMP: _TokenPunctuation(): returning token '%s'(%d) @ %d\n", tok.string, tok.token, tok.linenum);
	return (tok);
}


/********************************************************************/
/* _TokenEvaluateVariable(): The next input (character) is known to */
/*    be a variable.  This returns either a numeric or string token */
/*    depending on the value of the variable.                       */
/* NOTE: this currently assumes all variables are environment variables */
/*   or special built-in variables such as the architecture.            */
static vrTokenInfo _TokenEvaluateVariable(vrTokenInfo var_token)
{
static	char		*whitespace = " \t\r\b";
	vrTokenInfo	value_token = { VRTOKEN_NIL, "<nil>" };	/* return the <nil> token if there's a problem */
	char		*variable_name = var_token.string;	/* the name of the variable */
	char		*var_string;				/* the string of the variable */
	int		toklen;

	/***************************************************/
	/* first set the line number of the returned token */
	value_token.linenum = var_token.linenum;		/* set to line with the variable */

	/*******************************************/
	/* next check the validity of the variable */
	if (variable_name == NULL) {
		/* NOTE: this really shouldn't happen */
		vrDbgPrintfN(ALWAYS_DBGLVL, "Configuration Error (line %d): empty variable name string.\n", var_token.linenum);
		return value_token;
	}

	if (variable_name[0] == '\0') {
		vrDbgPrintfN(ALWAYS_DBGLVL, "Configuration Error (line %d): empty variable name.\n", var_token.linenum);
		return value_token;
	}

	/**************************************/
	/* now do a simple parse on the value */
	/*   Currently we only check for numeric or string values. */

	/* point to the string value of the variable */
	var_string = vrEvaluateVariable(vrContext, variable_name);/* NOTE: using global vrContext */
	if (var_string == NULL) {
		vrDbgPrintfN(CONFIG_ERROR_DBGLVL, "Configuration Error (line %d): " RED_TEXT "no such variable '$%s'\n" NORM_TEXT,
			var_token.linenum, var_token.string);
		return value_token;
	}

	var_string += strspn(var_string, whitespace);		/* skip whitespace */

	/* check whether the variable value is a number */
	if ((toklen = _NumericStringLength(var_string)) > 0) {
		value_token.token = VRTOKEN_NUMBER;
		value_token.string = vrShmemAlloc(toklen+1);
		strncpy(value_token.string, var_string, toklen);
		value_token.string[toklen] = '\0';
	} else {
		/* NOTE: all non-numbers are strings, so we know token is a string */

		value_token.token = VRTOKEN_STRING;
		toklen = strlen(var_string);			/* copy the entire string to be the string value */
		value_token.string = vrShmemAlloc(toklen+1);
		strncpy(value_token.string, var_string, toklen);
		value_token.string[toklen] = '\0';
	}

	vrDbgPrintfN(PARSE_DBGLVL, "_TokenEvaluateVariable(), line %d: "
		BOLD_TEXT "variable $%s value is '%s', token type = %d\n" NORM_TEXT,
		value_token.linenum, variable_name, value_token.string, value_token.token);

	return (value_token);
}


/*********************************************************************/
/* _FunctionOrString(): The next input is known to be an alphanumeric  */
/*   string, so determine whether the token is a function, and if    */
/*   so, evaluate it.  This returns either a numeric or string token */
/*   depending on the value returned by the function.                */

static vrTokenInfo _FunctionOrString(char *tokstring, vrParseInfo *parse)
{
static	char		*whitespace = " \t\r\b";
	vrTokenInfo	tok = { VRTOKEN_NIL, "<nil>" };
	int		toklen;
	char		*(*function)(char *) = NULL;
	char		*string = NULL;
	char		*result = NULL;
	char		*tofree;
	vrTokenInfo	openparen_token;
	vrTokenInfo	arg_token;
	int		count;

	vrDbgPrintfN(TRACE_DBGLVL, "(%s::%d) %s -> " RED_TEXT "'%s' may be a function -- entering" NORM_TEXT "\n",
		__FILE__, __LINE__, "_FunctionOrString()", tokstring);

	/***************************************/
	/* Initialize the default token result */

#if 0 /* including these would work, except on the receiving end, it could be   */
	/* interpretted that the string token was a function, and just happened */
	/* to return a string that was the function's name.                     */
	tok.token = VRTOKEN_STRING;
	tok.string = strdup(tokstring);
#endif
	tok.linenum = parse->linenum;

	/**********************************/
	/* search for a matching function */
	for (count = 0; count < parse->numfuncs; count++) {
		if (!strcasecmp(parse->funcmap_list[count].string, tokstring)) {
			function = parse->funcmap_list[count].func;
		}
	}

	/***************************************************************/
	/* If no matching function is found, then return the NIL token */
	if (function == NULL) {
		vrDbgPrintfN(TRACE_DBGLVL, "(%s::%d) %s -> " RED_TEXT "'%s' is not a function -- exiting" NORM_TEXT "\n",
			__FILE__, __LINE__, "_FunctionOrString()", tokstring);
		return (tok);
	}

	/****************************************************************************/
	/* Otherwise, determine the string between parentheses and pass to function */
	openparen_token = _ParseJustNextToken(parse);

	if (openparen_token.token != VRTOKEN_OPENPAREN) {
#ifdef WARN_ABOUT_CONFIG_FUNCTIONS_WO_PARENS  /* Choice: one option (defined/"1") will */
	/* treat a function name not followed by an OP as though it is a function, and */
	/* the arguments start right away.  Another option (undefined/"0") will report */
	/* a matched function name that isn't followed by an open-paren as simply not  */
	/* a function, but just that string.                                           */

		/* Without a following open-paren, treat the next token as the function's argument */
		vrErrPrintf("FreeVR config: Error (line %d): "
			RED_TEXT "Function name ('%s') not followed by open paren.\n" NORM_TEXT,
			openparen_token.linenum, tokstring);
		//string = vrShmemStrDup(openparen_token.string);	/* no longer used */
		return (openparen_token);
#else
		/* Without a following open-paren, treat the function name as a string. */
		vrParseReturnPreviousToken(openparen_token, parse);
		vrDbgPrintfN(TRACE_DBGLVL, "(%s::%d) %s -> " RED_TEXT "'%s' is apparently not a function -- exiting" NORM_TEXT "\n",
			__FILE__, __LINE__, "_FunctionOrString()", tokstring);
		return (tok);
#endif
	}

	/* Concatenate subsequent tokens into a string with the arguments */
	arg_token = vrParseNextToken(parse);
	while (arg_token.token != VRTOKEN_CLOSEPAREN  && !_TokenIsTerminal(arg_token.token)) {

		if (string == NULL) {
			string = vrShmemStrDup(arg_token.string);
		} else {
			tofree = string;
#if 0
			string = vrShmemStrCat(string, " ");		/* add whitespace */
#endif
			string = vrShmemStrCat(string, arg_token.string);
			vrShmemFree(tofree);
		}

		arg_token = vrParseNextToken(parse);
	}

	/********************************************/
	/* Evaluate the function with the arguments */
	result = (*function)(string);		/* TODO: we may want to pass "context" here too -- for now using global when necessary [09/17/21] */

	/* Determine whether the result is a number */
	result += strspn(result, whitespace);			/* skip whitespace */
	if ((toklen = _NumericStringLength(result)) > 0) {
		tok.token = VRTOKEN_NUMBER;
	} else {
		/* NOTE: all non-numbers are strings, so we know token is a string */

		tok.token = VRTOKEN_STRING;
		toklen = strlen(result);			/* copy the entire string to be the string value */
	}
	tok.string = vrShmemAlloc(toklen+1);
	strncpy(tok.string, result, toklen);
	tok.string[toklen] = '\0';

	vrDbgPrintfN(TRACE_DBGLVL, "(%s::%d) %s -> " RED_TEXT "function '%s' was evaluated to '%s' -- exiting" NORM_TEXT "\n",
		__FILE__, __LINE__, "_FunctionOrString()", tokstring, tok.string);
	return (tok);
}


/******************************************************************/
/* vrParseNextToken(): just do a regular vrParseNextTokenParser(),*/
/*   and when compiled with PRINT_ALL_PARSED_TOKENS flag set, then*/
/*   print the result at the PARSE_DBGLVL debug level, and also   */
/*   return the token as normal.                                  */
/* NOTE: formerly neamed vrParseNextTokenPrint(), but renamed to  */
/*   just vrParseNextToken() to allow the decision to print all   */
/*   parsed tokens at this point.                                 */
vrTokenInfo vrParseNextToken(vrParseInfo *parse)
{
	vrTokenInfo	token;

	token = vrParseNextTokenParser(parse);

#ifdef PRINT_ALL_PARSED_TOKENS
	vrDbgPrintfN(PARSE_DBGLVL, "vrParseNextTokenPrint(): Parsed (line %d): token (%3d) #### '%s' ####\n",
		token.linenum, token.token, token.string);
#endif

	return token;
}


/*********************************************************************/
/* vrParseNextToken(): as the name implies, it read the next portion */
/*   of input and parses it.  However, if there's already been some  */
/*   pre-parsed data, then use that instead -- sometimes a parsed    */
/*   token will be deemed to have been read too soon and placed back */
/*   in the parsing chain.                                           */
/* When parsing new material, the possible expression types are      */
/*   handled in the order:                                           */
/*   * read a chunk with whitespace separation                       */
/*   * skip full-line #-comments                                     */
/*   * look for in-line #-comments                                   */
/*   * handle punctuation                                            */
/*   * handle math operators                                         */
/*   * handle numbers                                                */
/*   * handle keywords (from token list)                             */
/*   * handle variables                                              */
/*   * handle unknown characters                                     */
/*   * skip "in-line reevaluations"                                  */
/*   * handle math expressions                                       */
/*   * handle function evaluations                                   */
/*   * do string concatenation                                       */
/* NOTE: renamed to vrParseNextTokenParser() in order to wrap it with*/
/*   the actual vrParseNextToken() where printing vs non-printing can*/
/*   be chosen.                                                      */
vrTokenInfo vrParseNextTokenParser(vrParseInfo *parse)
{
static	char		*whitespace = " \t\r\b";
static	char		*punctuation = "(){},;\"";	/* NOTE: parens are here because they are parsed as single character tokens. */
static	char		*mathchars = "+*/=<>!|&";	/** TODO: work on parsing of "-" **/
static	char		*strchars = "_:abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ";
static	char		*strnumchars = "_:abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ.0123456789";	/* TODO: should '-' be in this list? */
	int		toklen;
	int		first_nonws;
	vrTokenInfo	eoftok = { VRTOKEN_EOF, "<eof>", -1 };
	vrTokenInfo	return_tok = { VRTOKEN_UNKNOWN, "<unk-' '>", -1 };
	vrTokenInfo	tok = { VRTOKEN_UNKNOWN, "<unk>", -1 };


	/***************************************************************/
	/** first check whether a token was already parsed and stored **/
	if (parse->preparsed != NULL) {
		vrTokenInfo	*freeme;

		/* if a token was already parsed and stored, use it */
		/* NOTE: this should only happen when a parsing process decided it */
		/*   shouldn't have read the next token, so it stores it here.     */
		tok = *(parse->preparsed);
		freeme = parse->preparsed;

		/* pop that token off the stack */
		parse->preparsed = parse->preparsed->next;
		vrShmemFree(freeme);
		vrDbgPrintfN(PARSE_DETAIL_DBGLVL, "vrParseNextToken: (line %d) using preparsed token **** (%3d) '%s' ****\n",
			tok.linenum, tok.token, tok.string);

		return_tok = tok;

#ifdef PARSE_CONCAT_STRINGS /* NOTE: not necessary when the string concatenation stuff has been disabled */
		/* NOTE: 5/12/03: I'm not sure that this is ever necessary, because the only */
		/*   way for this to get on the pre-parsed list is that some non-string      */
		/*   token was encountered -- I think!                                       */
		/* For now, I'll set it as a PARSE_DBGLVL message.                           */
		if (return_tok.token == VRTOKEN_STRING)
			vrDbgPrintfN(PARSE_DBGLVL, RED_TEXT "Hmmm, vrParseNextToken() pre-parsed token '%s' is a string!  We may need to check for concatenation.\n" NORM_TEXT, return_tok.string);
#endif
		return (return_tok);
	}

	/**********************************/
	/** otherwise, do the dirty work **/

	/* find the next non whitespace character */
	/*   NOTE: '\n' is treated differently than other whitespace in order to: */
	/*     1: advance the line number counter, and                            */
	/*     2: terminate comments.                                             */
	/*     (both of which are likely to occur when parsing strings).          */
	first_nonws = strspn(parse->next_input, whitespace);
	while ((parse->next_input[first_nonws] == '\n') ||
		(first_nonws == strlen(parse->next_input)) ||
		(parse->next_input[first_nonws] == '#')) {
	  if (parse->next_input[first_nonws] == '\n') {
		parse->linenum++;
		parse->next_input = &(parse->next_input[first_nonws+1]);
		first_nonws = strspn(parse->next_input, whitespace);
	  }

	/* advance the next_input pointer past comments (terminated by '\n') */
	  if (parse->next_input[first_nonws] == '#') {
		parse->next_input += first_nonws;
		parse->next_input += strcspn(parse->next_input, "\n");
		parse->next_input += strspn(parse->next_input, "\n");
		first_nonws = strspn(parse->next_input, whitespace);
	  }

	/* If the first non white space of the next stuff to parse is the end of */
	/*   the current input, then read more input -- if reading from a file.  */
	  if (first_nonws == strlen(parse->next_input)) {
		if (parse->fp != NULL) {
			/* Get the next non-comment, non-empty line.  If none left */
			/*   then return with EOF token.                           */
			do {
				parse->linenum++;
				eoftok.linenum = parse->linenum;
				if (fgets(parse->input, MAX_PARSE_INPUT, parse->fp) == NULL) {
					vrDbgPrintfN(PARSE_DETAIL_DBGLVL, "vrParseNextToken: (line %d) hit EOF token **** (%3d) <eof> ****\n",
						//parse->linenum, eoftok.token);
						eoftok.linenum, eoftok.token);
					return eoftok;
				}
				first_nonws = strspn(parse->input, whitespace);
			} while (parse->input[first_nonws] == '#' || (first_nonws == strlen(parse->input)));

			/* print an error if the line was longer than we can handle */
			if (parse->input[MAX_PARSE_INPUT-1] != '\0') {
				vrFprintf(stderr, RED_TEXT "%s Parse: input line greater than maximum (%d) characters on line %d of file '%s'.\n" NORM_TEXT, parse->parse_name, MAX_PARSE_INPUT, parse->linenum, parse->filename);
				parse->input[MAX_PARSE_INPUT-1] = '\0';

			}

			/* not really nece, but makes it easier to print */
			parse->input[strlen(parse->input)-1] = '\0';

			/* start parsing at the beginning of the line */
			parse->next_input = parse->input;
		} else {
			eoftok.linenum = parse->linenum;
			vrDbgPrintfN(PARSE_DETAIL_DBGLVL, "vrParseNextToken: (line %d) hit EOF token **** (%3d) <eof> ****\n",
				eoftok.linenum, eoftok.token);
			return eoftok;
		}
	  }
	}

	tok.linenum = parse->linenum;
	eoftok.linenum = parse->linenum;

	parse->next_input += strspn(parse->next_input, whitespace);	/* skip whitespace */

	/***************************/
	/** parse the next string **/
	/***************************/
	if (parse->next_input[0] == '\0') {
		vrErrPrintf("%s Parse Error (line %d): "
			RED_TEXT "hmmm, should I really get a null char as part of a token?\n" NORM_TEXT,
			parse->parse_name, tok.linenum);
		return (eoftok);

	/*********************/
	/** in-line comment **/
	} else if (parse->next_input[0] == '#') {
		tok.string = parse->next_input;
		parse->next_input += strlen(parse->next_input);
		tok.token = VRTOKEN_COMMENT;
		tok.string = "# in-line comment";
		vrDbgPrintfN(PARSE_DETAIL_DBGLVL, "vrParseNextToken: (line %d) in-line comment **** (%3d) '%s' ****\n",
			tok.linenum, tok.token, tok.string);

		/* recursively get the next token -- comments are white-space */
//vrPrintf("TEMP: getting next token for in-line comment\n");
		return_tok = vrParseNextToken(parse);

	/*****************/
	/** punctuation **/
	} else if (strspn(parse->next_input, punctuation) > 0) {
		tok = _TokenPunctuation(parse->next_input[0], parse);
		parse->next_input++;

		/* if we get a quote mark, then parse all the characters */
		/*   until the next quote as a string.                   */
		if (tok.token == VRTOKEN_QUOTE) {
			toklen = strcspn(parse->next_input, "\"");
			if (toklen == strlen(parse->next_input)) {
				/* no more quote marks in input is an error */
				tok.token = VRTOKEN_PARSEERROR;
				tok.string = "unmatched quote";
				return_tok = tok;
			} else {
				tok.token = VRTOKEN_STRING;
				tok.string = vrShmemAlloc(toklen+1);
				strncpy(tok.string, parse->next_input, toklen);
				tok.string[toklen] = '\0';
				parse->next_input += toklen + 1;

				return_tok = tok;
			}
			vrDbgPrintfN(PARSE_DETAIL_DBGLVL, "vrParseNextToken: (line %d) string token **** (%3d) '%s' ****\n",
				return_tok.linenum, return_tok.token, return_tok.string);
		} else {
			return_tok = tok;
			vrDbgPrintfN(PARSE_DETAIL_DBGLVL, "vrParseNextToken: (line %d) punctuation token **** (%3d) '%s' ****\n",
				return_tok.linenum, return_tok.token, return_tok.string);
		}

	/*******************/
	/** math operator **/
	} else if (strspn(parse->next_input, mathchars) > 0) {
		toklen = strspn(parse->next_input, mathchars);
		tok.string = vrShmemAlloc(toklen+1);
		strncpy(tok.string, parse->next_input, toklen);
		tok.string[toklen] = '\0';
		parse->next_input += toklen;

		/* NOTE: the math operators are also part of the list of token strings, */
		/*   but it is important to parse them separately, so we don't end up   */
		/*   with long strings containing both alphanumeric and math characters.*/
		return_tok = _TokenString(tok.string, parse);

		vrDbgPrintfN(PARSE_DETAIL_DBGLVL, "vrParseNextToken: (line %d) math operator token **** (%3d) '%s' ****\n",
			return_tok.linenum, return_tok.token, return_tok.string);

	/** TODO: somehow we need to handle the special case of "-" which can **/
	/**   be part of a numeric expression or part of a math expression.   **/
	/**   For now, it will be limited to being part of a number.          **/
	/** NOTE: I was thinking that perhaps it might be possible to have a  **/
	/**   MINUS operator that coexists with the "-" as part of the number.**/
	/**   So, if we're expecting a math expression, then the "-" can be   **/
	/**   assumed to be that, or if we're expecting a number then it's a  **/
	/**   number -- hmmm, maybe the code isn't quite formulated that way. **/

	/************************/
	/** numeric expression **/
	} else if ((toklen = _NumericStringLength(parse->next_input)) > 0) {
		tok.string = vrShmemAlloc(toklen+1);
		strncpy(tok.string, parse->next_input, toklen);
		tok.string[toklen] = '\0';
		parse->next_input += toklen;

		return_tok = _TokenNumber(tok.string, parse);
		vrDbgPrintfN(PARSE_DETAIL_DBGLVL, "vrParseNextToken: (line %d) numeric expression token **** (%3d) '%s' ****\n",
			return_tok.linenum, return_tok.token, return_tok.string);

	/***********************************/
	/** keyword (alphanumeric string) **/
	} else if (strspn(parse->next_input, strchars) > 0) {
		/* NOTE: the full string can have numbers in it, */
		/*   they just can't be first.                   */
		/* NOTE however that we wouldn't get here if the first char was a number */
		toklen = strspn(parse->next_input, strnumchars);
		tok.string = vrShmemAlloc(toklen+1);
		strncpy(tok.string, parse->next_input, toklen);
		tok.string[toklen] = '\0';
		parse->next_input += toklen;

		return_tok = _TokenString(tok.string, parse);
		vrDbgPrintfN(PARSE_DETAIL_DBGLVL, "vrParseNextToken: (line %d) keyword token **** (%3d) '%s' ****\n",
			return_tok.linenum, return_tok.token, return_tok.string);

	/*******************/
	/** variable name **/
	} else if (parse->next_input[0] == '$') {
		parse->next_input++;			/* skip past the '$' */
		toklen = 0;

		if (parse->next_input[0] == '?')
			toklen = 1 + strspn(&parse->next_input[1], strnumchars);	/* include and skip past an initial '?' */
		else	toklen = strspn(parse->next_input, strnumchars);	/* NOTE: initial whitespace will give empty variable */

		tok.string = vrShmemAlloc(toklen+1);
		tok.token = VRTOKEN_VARIABLE;
		strncpy(tok.string, parse->next_input, toklen);
		tok.string[toklen] = '\0';
		parse->next_input += toklen;
		tok.linenum = parse->linenum;


		/* return not the value of the variable token, but rather the value it contains (ie. evaluates to) */
		return_tok = _TokenEvaluateVariable(tok);
		vrDbgPrintfN(VARIABLE_DBGLVL, "vrParseNextToken(): Parsed variable (line %d): token (%3d) #### '%s' ####, len = %d\n", parse->linenum, return_tok.token, return_tok.string, toklen);
		vrDbgPrintfN(PARSE_DETAIL_DBGLVL, "vrParseNextToken(): Parsed variable (line %d): token (%3d) #### '%s' ####, len = %d\n", parse->linenum, return_tok.token, return_tok.string, toklen);
	}

	/***********************/
	/** unknown character **/

	/* If no token has been assigned, then the config file has */
	/*   a character that we don't recognize, so return it as  */
	/*   an unknown token, and advance the parsing counter.    */
	if (return_tok.token == VRTOKEN_UNKNOWN) {
		return_tok.linenum = parse->linenum;
#ifndef __linux /* TODO: for some unknown reason, Linux segfaults on this */
		return_tok.string[6] = parse->next_input[0];
#endif
		vrDbgPrintfN(PARSE_DETAIL_DBGLVL, "vrParseNextToken: (line %d): "
			"Parsing '%s' as an Unknown token\n",
			return_tok.linenum, return_tok.string);

		parse->next_input++;
		return (return_tok);
	}


	/******************************************/
	/** skip inline re-evaluation of tokens? **/
	if (parse->suppress_inline)
		return (return_tok);


	/********************************/
	/** Math expression evaluation **/

	/* This is where math expressions can be evaluated as they're parsed. */
	/*   If the current token is a number, a uni-operand math operator,   */
	/*   or an open parenthesis, then call vrParseMathExpression()        */
	/*   otherwise consider this to not be a math expression, and move on.*/

	if (_TokenIsUniMathOperator(return_tok.token) ||
			(return_tok.token == VRTOKEN_OPENPAREN) ||
			(return_tok.token == VRTOKEN_NUMBER)
		) {
		vrTokenInfo	mathval_token;

		vrDbgPrintfN(PARSE_DETAIL_DBGLVL, "vrParseNextToken: (line %d): "
			"vrParseNextToken(): Parsed '%s' as a Math token\n",
			return_tok.linenum, return_tok.string);

		/* need to put initial token back so it will be seen by vrParseMathExpression() */
		vrParseReturnPreviousToken(return_tok, parse);

		mathval_token = vrParseMathExpression(parse);

		return_tok = mathval_token;
	}

	/*************************/
	/** Function evaluation **/

	/* If the token is a string, then we can compare it with the array of     */
	/*   available function names, and if we find a match, call the function. */
	if (return_tok.token == VRTOKEN_STRING) {
		vrTokenInfo	func_token = _FunctionOrString(return_tok.string, parse);

		/* If the function was found and evaluated, then return */
		/*   the value without concatenation.                   */
		/* TODO: determine whether this is the behavior we want.*/
		/*   NOTE that the may still be concatenated to previous*/
		/*   strings.                                           */
		if (func_token.token != VRTOKEN_NIL) {
			vrDbgPrintfN(PARSE_DETAIL_DBGLVL, "vrParseNextToken: (line %d): "
				"Parsed '%s' as a FunctionString token\n",
				return_tok.linenum, return_tok.string);

			return (func_token);
		}
	}

#ifdef PARSE_CONCAT_STRINGS /* NOTE: this doesn't work when separate strings are desired (eg. "setenv" implementation without '=') */
	/**************************/
	/** String concatenation **/

	/* If the token is a string, then check whether the next token is also a   */
	/*   string, and if so concatenate them together -- and do so recursively. */
	if (return_tok.token == VRTOKEN_STRING) {
		tok = vrParseNextToken(parse);

		vrDbgPrintfN(PARSE_DETAIL_DBGLVL FORCE_DBG, "vrParseNextToken: (line %d): "
			"'%s' is a string token, so looking ahead to token '%s' for possible concatenation...",
			return_tok.linenum, return_tok.string, tok.string);

		if (tok.token == VRTOKEN_STRING || tok.token == VRTOKEN_NUMBER) {
			/* Concatenate onto the return token */
			return_tok.string = vrShmemStrCat(return_tok.string, tok.string);
			vrDbgPrintfN(PARSE_DETAIL_DBGLVL FORCE_DBG, "Concentenated string to: '%s'.\n", return_tok.string);
		} else {
			/* Store the next token as a pre-parsed token */
			vrDbgPrintfN(PARSE_DETAIL_DBGLVL FORCE_DBG, "Did not concatentate onto string: '%s'\n", tok.string);
			vrParseReturnPreviousToken(tok, parse);
		}
	}
#endif

	return (return_tok);
}


/************************************************************************/
/* _ParseJustNextToken(): This function is just a wrapper function for  */
/*   vrParseNextToken() that simply disables the inline post-processing */
/*   of tokens to determine whether they can be further parsed.         */
/* Possible uses of this function is to get exactly the next string, or */
/*   get an open-paren token without doing the math simplification --   */
/*   which is desirable when trying to determine whether a string is    */
/*   a function name.                                                   */
static vrTokenInfo _ParseJustNextToken(vrParseInfo *parse)
{
	vrTokenInfo	tok;

	parse->suppress_inline = 1;
	tok = vrParseNextToken(parse);
	parse->suppress_inline = 0;

	return (tok);
}


/**************************************************************************/
/* vrParseReturnPreviousToken(): put a previously parsed token onto a     */
/*   token stack such that it has parsing priority, and will be the next  */
/*   item handled in the parsing sequence.  This is for cases where we    */
/*   need to "look-ahead" in the parsing to see what action should be     */
/*   taken, and sometimes we won't use the upcoming token, so it will     */
/*   need to be addressed next.                                           */
/* NOTE: care must be taken to avoid having the possibility of attempting */
/*   to store more than one pre-parsed token.  Currently, there are three */
/*   instances where this may be called, and they are (or are believed to */
/*   be) mutually-exclusive:                                              */
/*   1) a math operation that may just be a numeric value, or may be      */
/*      followed by an operator.                                          */
/*   2) a string token that may be followed by another string, with the   */
/*      two being concatenated together.                                  */
/*   3) an open or close curly bracket is encountered as part of a search */
/*      for an end-of-statement indicator -- which would typically come   */
/*      about when skipping over an encountered error.  We'll still need  */
/*      to initiate or close the block associated with the bracket.       */
void vrParseReturnPreviousToken(vrTokenInfo token, vrParseInfo *parse)
{
	vrTokenInfo	*store_token = vrShmemAlloc0(sizeof(*store_token));

	*store_token = token;

	/* If there is already one (or more) pre-parsed tokens, */
	/*   add this one to the beginning of the list.         */
	if (parse->preparsed != NULL) {
		store_token->next = parse->preparsed;

	} else {
		store_token->next = NULL;
	}

	parse->preparsed = store_token;
	vrDbgPrintfN(PARSE_DETAIL_DBGLVL, "vrParseReturnPreviousToken: returning token **** (%3d) '%s' ****\n",
		token.token, token.string);
#ifdef PRINT_ALL_PARSED_TOKENS
	vrPrintf("vrParseReturnPreviousToken: returning token **** (%3d) '%s' ****\n",
		token.token, token.string);
#endif
}


/************************************************************************/
/* vrParseMathExpression(): Parse a mathematical expression, which are  */
/*   defined as these options (or not-yet-implimented options -- :-( ). */
/*                                                                      */
/* Possible math expressions are:                                       */
/*      - <number>                                                      */
/* :-(  - <uni-operator> <number>                                       */
/*      - <number> <bi-operator> <number>                               */
/* :-(  - <number> <bi-operator> <expression>                           */
/* :-(  - "(" <number> <bi-operator> <expression> ")"                   */
/*	- <string> <equality|non-equality bi-operator> <string>         */
/*                                                                      */
/* Expressions currently handled by this function are:                  */
/*	- unary value                                                   */
/*	- equality comparison                                           */
/*	- inequality comparison                                         */
/*	- greater-than comparison                                       */
/*	- greater-than-or-equal comparison                              */
/*	- less-than comparison                                          */
/*	- less-than-or-equal comparison                                 */
/*	- addition                                                      */
/*	- multiplication                                                */
/*	- division                                                      */
/*	- bitwise-or                                                    */
/*	- bitwise-and                                                   */
/*	- logical-or                                                    */
/*	- logical-and                                                   */
/*                                                                      */
/* TODO: implement the <uni-operator> form.                             */
/* TODO: if an open-paren is encountered, then make a recursive call.   */
/* NOTE: we need to not just parse the expression, but evaluate it too! */
/* NOTE: one problem we have is that if we get a number, and the next   */
/*   token isn't a bi-operator, then we need to put that token back!    */
/* TODO: there's a lot of redundancy in the comparison code -- probably */
/*   should write it with much less redundancy.                         */
/* NOTE: the logical comparison operations can operate on string or     */
/*   numeric values.  However, there are only limited places where this */
/*   function might be called with string values -- eg. currently the   */
/*   "if" operations.  Otherwise, the in-line numeric simplification    */
/*   parsing will not call this function for strings.                   */
/************************************************************************/
vrTokenInfo vrParseMathExpression(vrParseInfo *parse)
{
static	char		string[256];
	vrTokenInfo	value1_token;
	vrTokenInfo	value2_token;
	vrTokenInfo	operator_token;
	vrTokenInfo	close_token;
	vrTokenInfo	result_token = { VRTOKEN_NIL, "<nil>", -1 };
	int		openparen_linenum;

	vrTrace("vrParseMathExpression()", RED_TEXT "entering" NORM_TEXT);

	/*************************************/
	/* get first token of the expression */
//vrPrintf("TEMP: vrParseMathExpression(): about to call vrParseNextToken()\n");
	value1_token = vrParseNextToken(parse);
//vrPrintf("TEMP: vrParseMathExpression(): returned from vrParseNextToken() -- '%s'\n", value1_token.string);

	/* if value begins with an open-paren, then evaluate the sub-expression */
	if (value1_token.token == VRTOKEN_OPENPAREN) {
		vrDbgPrintfN(PARSE_DBGLVL, "vrParseMathExpression(): " RED_TEXT "recursively calling vrParseMathExpression() for 1st argument\n" NORM_TEXT);
		openparen_linenum = value1_token.linenum;
		value1_token = vrParseMathExpression(parse);
		close_token = vrParseNextToken(parse);
		if (close_token.token != VRTOKEN_CLOSEPAREN) {
			vrDbgPrintfN(ALWAYS_DBGLVL, "Configuration Error (line %d): open-paren is missing matching closing paren.\n", openparen_linenum);
			vrParseReturnPreviousToken(close_token, parse);
		}
		vrDbgPrintfN(PARSE_DBGLVL, "vrParseMathExpression(): " RED_TEXT "recursive call for 1st argument returned token string '%s'\n" NORM_TEXT, value1_token.string);
	}

	/******************************************/
	/* get possible operator, and 2nd operand */
	operator_token = vrParseNextToken(parse);

	/* if the operator is a bi-operator, then get the 2nd operand */
	if (_TokenIsBiMathOperator(operator_token.token)) {
		value2_token = vrParseNextToken(parse);

		/* if value begins with an open-paren, then evaluate the sub-expression */
		if (value2_token.token == VRTOKEN_OPENPAREN) {
			vrDbgPrintfN(PARSE_DBGLVL, "vrParseMathExpression(): recursively calling vrParseMathExpression() for 2nd argument\n");
			openparen_linenum = value2_token.linenum;
			value2_token = vrParseMathExpression(parse);
			close_token = vrParseNextToken(parse);
			if (close_token.token != VRTOKEN_CLOSEPAREN) {
				vrDbgPrintfN(ALWAYS_DBGLVL, "Configuration Error (line %d): open paren is missing matching closing paren.\n", openparen_linenum);
				vrParseReturnPreviousToken(close_token, parse);
			}
			vrDbgPrintfN(PARSE_DBGLVL, "vrParseMathExpression(): " RED_TEXT "recursive call for 2nd argument returned token string '%s'\n" NORM_TEXT, value2_token.string);
		}
	} else {
		/* if the operator token wasn't a known math operator,   */
		/*   then put it back into the parser as the next token, */
		/*   and return the first value as our result (ie. a     */
		/*   single value 'expression').                         */
		vrParseReturnPreviousToken(operator_token, parse);
		vrDbgPrintfN(PARSE_DETAIL_DBGLVL, "vrParseMathExpression(<number>): 'operator' (line %d): returning token %d with value '%s'\n",
			operator_token.linenum, operator_token.token, operator_token.string);
		return (value1_token);
	}

	/* NOTE: if this were a single-value 'expression' then we should have */
	/*   simply returned to the calling parse-routine, and should proceed */
	/*   no further.                                                      */

	/************************/
	/* now do the operation */
	vrDbgPrintfN(PARSE_DBGLVL, "vrParseMathExpression(): " RED_TEXT "performing operation: (%s %s %s)\n" NORM_TEXT,
		value1_token.string, operator_token.string, value2_token.string);

	switch (operator_token.token) {

	/********************/
	case VRTOKEN_EQUALITY:
		if (value1_token.token != VRTOKEN_NUMBER) {
			/* Do a string comparison */
			result_token.token = VRTOKEN_NUMBER;
			result_token.linenum = operator_token.linenum;

			/* TODO: consider whether we should ignore case */
			if (!strcmp(value1_token.string, value2_token.string))
				result_token.string = strdup("1");
			else	result_token.string = strdup("0");
		} else {
			/* Do a numeric comparison */
			result_token.token = VRTOKEN_NUMBER;
			result_token.linenum = operator_token.linenum;
			if (atof(value1_token.string) == atof(value2_token.string))
				result_token.string = strdup("1");
			else	result_token.string = strdup("0");
		}
		break;

	/**********************/
	case VRTOKEN_INEQUALITY:
		if (value1_token.token != VRTOKEN_NUMBER) {
			/* Do a string comparison */
			result_token.token = VRTOKEN_NUMBER;
			result_token.linenum = operator_token.linenum;

			/* TODO: consider whether we should ignore case */
			if (strcmp(value1_token.string, value2_token.string) != 0)
				result_token.string = strdup("1");
			else	result_token.string = strdup("0");
		} else {
			/* Do a numeric comparison */
			result_token.token = VRTOKEN_NUMBER;
			result_token.linenum = operator_token.linenum;
			if (atof(value1_token.string) != atof(value2_token.string))
				result_token.string = strdup("1");
			else	result_token.string = strdup("0");
		}
		break;

	/*******************/
	case VRTOKEN_GREATER:
		if (value1_token.token != VRTOKEN_NUMBER) {
			/* Do a string comparison */
			result_token.token = VRTOKEN_NUMBER;
			result_token.linenum = operator_token.linenum;

			/* TODO: consider whether we should ignore case */
			if (strcmp(value1_token.string, value2_token.string) > 0)
				result_token.string = strdup("1");
			else	result_token.string = strdup("0");
		} else {
			/* Do a numeric comparison */
			result_token.token = VRTOKEN_NUMBER;
			result_token.linenum = operator_token.linenum;
			if (atof(value1_token.string) > atof(value2_token.string))
				result_token.string = strdup("1");
			else	result_token.string = strdup("0");
		}
		break;

	/***********************/
	case VRTOKEN_GREATEROREQ:
		if (value1_token.token != VRTOKEN_NUMBER) {
			/* Do a string comparison */
			result_token.token = VRTOKEN_NUMBER;
			result_token.linenum = operator_token.linenum;

			/* TODO: consider whether we should ignore case */
			if (strcmp(value1_token.string, value2_token.string) >= 0)
				result_token.string = strdup("1");
			else	result_token.string = strdup("0");
		} else {
			/* Do a numeric comparison */
			result_token.token = VRTOKEN_NUMBER;
			result_token.linenum = operator_token.linenum;
			if (atof(value1_token.string) >= atof(value2_token.string))
				result_token.string = strdup("1");
			else	result_token.string = strdup("0");
		}
		break;

	/****************/
	case VRTOKEN_LESS:
		if (value1_token.token != VRTOKEN_NUMBER) {
			/* Do a string comparison */
			result_token.token = VRTOKEN_NUMBER;
			result_token.linenum = operator_token.linenum;

			/* TODO: consider whether we should ignore case */
			if (strcmp(value1_token.string, value2_token.string) < 0)
				result_token.string = strdup("1");
			else	result_token.string = strdup("0");
		} else {
			/* Do a numeric comparison */
			result_token.token = VRTOKEN_NUMBER;
			result_token.linenum = operator_token.linenum;
			if (atof(value1_token.string) < atof(value2_token.string))
				result_token.string = strdup("1");
			else	result_token.string = strdup("0");
		}
		break;

	/********************/
	case VRTOKEN_LESSOREQ:
		if (value1_token.token != VRTOKEN_NUMBER) {
			/* Do a string comparison */
			result_token.token = VRTOKEN_NUMBER;
			result_token.linenum = operator_token.linenum;

			/* TODO: consider whether we should ignore case */
			if (strcmp(value1_token.string, value2_token.string) <= 0)
				result_token.string = strdup("1");
			else	result_token.string = strdup("0");
		} else {
			/* Do a numeric comparison */
			result_token.token = VRTOKEN_NUMBER;
			result_token.linenum = operator_token.linenum;
			if (atof(value1_token.string) <= atof(value2_token.string))
				result_token.string = strdup("1");
			else	result_token.string = strdup("0");
		}
		break;

	/****************/
	case VRTOKEN_PLUS:
		/* NOTE: it might be reasonable to "add" two strings, returning  */
		/*   a longer string, but we're not going to implement that now. */
		if (value1_token.token != VRTOKEN_NUMBER) {
			vrDbgPrintfN(ALWAYS_DBGLVL, "Configuration Error (line %d): The addition operator (currently) expects the first operand to be a number -- got '%s'\n",
				value1_token.linenum, value1_token.string);
		}
		if (value2_token.token != VRTOKEN_NUMBER) {
			vrDbgPrintfN(ALWAYS_DBGLVL, "Configuration Error (line %d): The addition operator (currently) expects the second operand to be a number -- got '%s'\n",
				value2_token.linenum, value2_token.string);
		}
		if ((value1_token.token == VRTOKEN_NUMBER) && (value2_token.token == VRTOKEN_NUMBER)) {
			snprintf(string, sizeof(string), "%f", (atof(value1_token.string) + atof(value2_token.string)));
			result_token.token = VRTOKEN_NUMBER;
			result_token.string = strdup(string);
			result_token.linenum = operator_token.linenum;
		}
		break;

	/*****************/
	case VRTOKEN_TIMES:
		if (value1_token.token != VRTOKEN_NUMBER) {
			vrDbgPrintfN(ALWAYS_DBGLVL, "Configuration Error (line %d): The multiplication operator (currently) expects the first operand to be a number -- got '%s'\n",
				value1_token.linenum, value1_token.string);
		}
		if (value2_token.token != VRTOKEN_NUMBER) {
			vrDbgPrintfN(ALWAYS_DBGLVL, "Configuration Error (line %d): The multiplication operator (currently) expects the second operand to be a number -- got '%s'\n",
				value2_token.linenum, value2_token.string);
		}
		if ((value1_token.token == VRTOKEN_NUMBER) && (value2_token.token == VRTOKEN_NUMBER)) {
			snprintf(string, sizeof(string), "%f", (atof(value1_token.string) * atof(value2_token.string)));
			result_token.token = VRTOKEN_NUMBER;
			result_token.string = strdup(string);
			result_token.linenum = operator_token.linenum;
		}
		break;

	/******************/
	case VRTOKEN_DIVIDE:
		if (value1_token.token != VRTOKEN_NUMBER) {
			vrDbgPrintfN(ALWAYS_DBGLVL, "Configuration Error (line %d): The division operator (currently) expects the first operand to be a number -- got '%s'\n",
				value1_token.linenum, value1_token.string);
		}
		if (value2_token.token != VRTOKEN_NUMBER) {
			vrDbgPrintfN(ALWAYS_DBGLVL, "Configuration Error (line %d): The division operator (currently) expects the second operand to be a number -- got '%s'\n",
				value2_token.linenum, value2_token.string);
		}
		if ((value1_token.token == VRTOKEN_NUMBER) && (value2_token.token == VRTOKEN_NUMBER)) {
			snprintf(string, sizeof(string), "%f", (atof(value1_token.string) / atof(value2_token.string)));
			result_token.token = VRTOKEN_NUMBER;
			result_token.string = strdup(string);
			result_token.linenum = operator_token.linenum;
		}
		break;

	/*******************/
	case VRTOKEN_BIT_AND:
		if (value1_token.token != VRTOKEN_NUMBER) {
			vrDbgPrintfN(ALWAYS_DBGLVL, "Configuration Error (line %d): The bitwise-and operator (currently) expects the first operand to be a number -- got '%s'\n",
				value1_token.linenum, value1_token.string);
		}
		if (value2_token.token != VRTOKEN_NUMBER) {
			vrDbgPrintfN(ALWAYS_DBGLVL, "Configuration Error (line %d): The bitwise-and operator (currently) expects the second operand to be a number -- got '%s'\n",
				value2_token.linenum, value2_token.string);
		}
		if ((value1_token.token == VRTOKEN_NUMBER) && (value2_token.token == VRTOKEN_NUMBER)) {
			snprintf(string, sizeof(string), "%ld", (vrAtoI(value1_token.string) & vrAtoI(value2_token.string)));
			result_token.token = VRTOKEN_NUMBER;
			result_token.string = strdup(string);
			result_token.linenum = operator_token.linenum;
		}
		break;

	/******************/
	case VRTOKEN_BIT_OR:
		if (value1_token.token != VRTOKEN_NUMBER) {
			vrDbgPrintfN(ALWAYS_DBGLVL, "Configuration Error (line %d): The bitwise-or operator (currently) expects the first operand to be a number -- got '%s'\n",
				value1_token.linenum, value1_token.string);
		}
		if (value2_token.token != VRTOKEN_NUMBER) {
			vrDbgPrintfN(ALWAYS_DBGLVL, "Configuration Error (line %d): The bitwise-or operator (currently) expects the second operand to be a number -- got '%s'\n",
				value2_token.linenum, value2_token.string);
		}
		if ((value1_token.token == VRTOKEN_NUMBER) && (value2_token.token == VRTOKEN_NUMBER)) {
			snprintf(string, sizeof(string), "%ld", (vrAtoI(value1_token.string) | vrAtoI(value2_token.string)));
			result_token.token = VRTOKEN_NUMBER;
			result_token.string = strdup(string);
			result_token.linenum = operator_token.linenum;
		}
		break;

	/***********************/
	case VRTOKEN_LOGICAL_AND:
		if (value1_token.token != VRTOKEN_NUMBER) {
			vrDbgPrintfN(ALWAYS_DBGLVL, "Configuration Error (line %d): The logical-and operator (currently) expects the first operand to be a number -- got '%s'\n",
				value1_token.linenum, value1_token.string);
		}
		if (value2_token.token != VRTOKEN_NUMBER) {
			vrDbgPrintfN(ALWAYS_DBGLVL, "Configuration Error (line %d): The logical-and operator (currently) expects the second operand to be a number -- got '%s'\n",
				value2_token.linenum, value2_token.string);
		}
		if ((value1_token.token == VRTOKEN_NUMBER) && (value2_token.token == VRTOKEN_NUMBER)) {
			snprintf(string, sizeof(string), "%d", (vrAtoI(value1_token.string) && vrAtoI(value2_token.string)));
			result_token.token = VRTOKEN_NUMBER;
			result_token.string = strdup(string);
			result_token.linenum = operator_token.linenum;
		}
		break;

	/**********************/
	case VRTOKEN_LOGICAL_OR:
		if (value1_token.token != VRTOKEN_NUMBER) {
			vrDbgPrintfN(ALWAYS_DBGLVL, "Configuration Error (line %d): The logical-or operator (currently) expects the first operand to be a number -- got '%s'\n",
				value1_token.linenum, value1_token.string);
		}
		if (value2_token.token != VRTOKEN_NUMBER) {
			vrDbgPrintfN(ALWAYS_DBGLVL, "Configuration Error (line %d): The logical-or operator (currently) expects the second operand to be a number -- got '%s'\n",
				value2_token.linenum, value2_token.string);
		}
		if ((value1_token.token == VRTOKEN_NUMBER) && (value2_token.token == VRTOKEN_NUMBER)) {
			snprintf(string, sizeof(string), "%d", (vrAtoI(value1_token.string) || vrAtoI(value2_token.string)));
			result_token.token = VRTOKEN_NUMBER;
			result_token.string = strdup(string);
			result_token.linenum = operator_token.linenum;
		}
		break;

	/******/
	default:
		/* We shouldn't get here, because non-bi-operators should have */
		/*   been returned before the switch statement.                */
		vrDbgPrintfN(AALWAYS_DBGLVL, "vrParseMathExpression(<number>): (line %d): "
			RED_TEXT "Library ERROR: Unknown bi-operator\n" NORM_TEXT,
			operator_token.linenum, operator_token.token, operator_token.string);
		return (value1_token);
	}

	vrDbgPrintfN(PARSE_DBGLVL, "vrParseMathExpression(<num> <op> <num>): (line %d): returning token %d with value '%s'\n",
		result_token.linenum, result_token.token, result_token.string);

	return (result_token);
}


		/**********************************/
		/*** Generic expression parsing ***/
		/**********************************/

/*********************************************************************/
/* This is a second section of "vr_parse.c".  It is still crucial to */
/*   parsing the configuration file, but these functions are called  */
/*   specifically from within vr_config.c to handle the particular   */
/*   syntax expected by particular keywords.                         */
/*********************************************************************/



/************************************************************/
/* vrParseSingleStringExpr(): parses an expression to get a */
/*   single string, and returns that string.                */
/* Format: "<setting>" "=" string[;]                        */
/*                                                          */
/* Arguments:                                                               */
/*   * retvalue -- a pointer to a char-array where the result is placed.    */
/*   * setting -- the name of what's being set (only used for debug output).*/
/*   * parse -- parsing system info -- (used to extract upcoming tokens)    */
vrTokenInfo vrParseSingleStringExpr(char **retvalue, const char *setting, vrParseInfo *parse)
{
	vrTokenInfo	assignment_token;	/* type of assignment used ('=' '+=' or '*=') */
	vrTokenInfo	value_token;
	vrTokenInfo	terminal_token;
#ifdef UNPICKY_COMPILER
	char		*value;
#endif

	vrTrace("vrParseSingleStringExpr()", RED_TEXT "entering" NORM_TEXT);

	/*****************************/
	/** operator ( "=" | "+=" ) **/
	assignment_token = vrParseNextToken(parse);
	if (!_TokenIsAssignment(assignment_token.token)) {
		vrErrPrintf("%s Parse Error (line %d): "
			RED_TEXT "Missing assignment operator (got \"%s\") for %s.\n" NORM_TEXT,
			parse->parse_name, assignment_token.linenum, assignment_token.string, setting);

		/* unless the invalid token was a terminator, skip the rest of the statement */
		if (!_TokenIsTerminal(assignment_token.token)) {
			terminal_token = vrParseToEOS(parse);
			vrDbgPrintfN(ALWAYS_DBGLVL, "Configuration Error (line %d): " RED_TEXT "skipping '%s'\n" NORM_TEXT,
				terminal_token.linenum, terminal_token.string);
		} else {
			terminal_token = value_token;
		}

		/* return an obvious error string */
		*retvalue = vrShmemStrDup("<config error>");
		vrTrace("vrParseSingleStringExpr()", RED_TEXT "exiting from assignment error" NORM_TEXT);
		return terminal_token;
	}

	/***********************/
	/** do the assignment **/
	switch(assignment_token.token) {
	case VRTOKEN_SETEQUALS:
		*retvalue = NULL;
		break;

	case VRTOKEN_PLUSEQUALS:
		/* leave retvalue alone */
		break;

	case VRTOKEN_NOP:	/* TODO: if we ever create a VRTOKEN_NOP, change this to it */ /* [09/29/21: done! (used to be VRTOKEN_NIL)] */
		break;

	case VRTOKEN_TIMESEQUALS:
	default:
		/* NOTE: 09/30/21: Yesterday, I added the "(or missing)", but I'm */
		/*   now handling that case with a check for a proper assignment  */
		/*   token above, but I'm leaving the "(or missing)" for now.     */
		vrErrPrintf("%s Parse Error (line %d): "
			RED_TEXT "Invalid (or missing) assignment operator (got \"%s\") for %s.\n" NORM_TEXT,
			parse->parse_name, assignment_token.linenum, assignment_token.string, setting);

		terminal_token = vrParseToEOS(parse);
		vrDbgPrintfN(ALWAYS_DBGLVL, "Configuration Error (line %d): " RED_TEXT "skipping '%s'\n" NORM_TEXT,
			terminal_token.linenum, terminal_token.string);

		vrTrace("vrParseSingleStringExpr()", RED_TEXT "exiting from assignment error" NORM_TEXT);

		/* return an obvious error string */
		*retvalue = vrShmemStrDup("<config error>");

		vrTrace("vrParseSingleStringExpr()", RED_TEXT "exiting from assignment error" NORM_TEXT);
		return terminal_token;
	}

	/***********/
	/** value **/
	value_token = vrParseNextToken(parse);
	if (value_token.token == VRTOKEN_NIL) {
		/* probably have a non-existent variable, so don't do the assignment */
		assignment_token.token = VRTOKEN_NIL;
		assignment_token.string = "<nil>";
	}

	/*************************/
	/** assign return value **/
	if (value_token.token == VRTOKEN_STRING || value_token.token == VRTOKEN_NUMBER) {
		if (*retvalue == NULL)
			*retvalue = vrShmemStrDup(value_token.string);
		else	*retvalue = vrShmemStrCat(*retvalue, value_token.string);
	} else {
		vrErrPrintf("%s Parse Error (line %d): "
			RED_TEXT "Invalid %s given -- value (%s) is not a string or number.\n" NORM_TEXT,
			parse->parse_name, value_token.linenum, setting, value_token.string);
		*retvalue = vrShmemStrDup("<config error>");
	}

	if (assignment_token.token == VRTOKEN_NIL) {
		vrDbgPrintfN(PARSE_DBGLVL, "vrParseSingleStringExpr(): "
			BOLD_TEXT "leaving '%s' as value '%s'," NORM_TEXT
			" assignment token = '%s' (line %d), value token = '%s' (line %d)\n",
			setting, *retvalue,
			assignment_token.string, assignment_token.linenum,
			value_token.string, value_token.linenum);
	} else {
		vrDbgPrintfN(PARSE_DBGLVL, "vrParseSingleStringExpr(): "
			BOLD_TEXT "setting '%s' to value '%s'," NORM_TEXT
			" assignment token = '%s' (line %d), value token = '%s' (line %d)\n",
			setting, *retvalue,
			assignment_token.string, assignment_token.linenum,
			value_token.string, value_token.linenum);
	}

	/***************************/			/* '{' to match next line */
	/** ( ";" | "}" | <eof> ) **/
	terminal_token = vrParseNextToken(parse);
	if (!_TokenIsTerminal(terminal_token.token)) {
		vrErrPrintf("%s Parse Error (line %d): "
			RED_TEXT "%s takes only one argument.  (I.e. expected a terminal token, but got '%s'(%d/%s).\n" NORM_TEXT,
			parse->parse_name, terminal_token.linenum, setting,
			terminal_token.string, terminal_token.token,
			parse->tsmap_list[terminal_token.token+5].string	/* TODO: WARNING, the "+5" is a hack, what needs to happen is use a function that reverse maps VRTOKEN_... to an index of parse->tsmap_list (only works for first N toeksn (until strings duplicate) */ /* fix or delete me (this argument) */
			);
	}

	vrTrace("vrParseSingleStringExpr()", RED_TEXT "exiting" NORM_TEXT);
	return terminal_token;
}


/*********************************************************/
/* vrParseStringListExpr(): returns an array of strings. */
/* NOTE: unlike the numeric list-expression parsers, the string-list    */
/*   expression parser will set an unknown number of strings in the     */
/*   list based on the number given in the config file.  Also, unlike   */
/*   all the single expression parsers, this function will go ahead     */
/*   and set values even when there are unavailable variables.  In      */
/*   these cases there the parser skips over the non-existent variables */
/*   as though they weren't in the list.                                */
vrTokenInfo vrParseStringListExpr(char ***array, int *array_len, const char *setting, vrParseInfo *parse)
{
	vrTokenInfo	assignment_token;	/* type of assignment used ('=' '+=' NOT '*=') */
	vrTokenInfo	value_token;
	vrTokenInfo	terminal_token;
	int		count;
	int		num_strings = 0;
	char		**newarray = NULL;
	char		**oldarray;

	vrTrace("vrParseStringListExpr()", RED_TEXT "entering" NORM_TEXT);

	/*****************************/
	/** operator ( "=" | "+=" ) **/
	assignment_token = vrParseNextToken(parse);
	if (assignment_token.token != VRTOKEN_SETEQUALS && assignment_token.token != VRTOKEN_PLUSEQUALS) {
		vrErrPrintf("%s Parse Error (line %d): "
			RED_TEXT "Expecting an assignment operator for System Procs.\n" NORM_TEXT,
			parse->parse_name, assignment_token.linenum);

		/* unless the invalid token was a terminator, skip the rest of the statement */
		if (!_TokenIsTerminal(assignment_token.token)) {
			terminal_token = vrParseToEOS(parse);
			vrDbgPrintfN(ALWAYS_DBGLVL, "Configuration Error (line %d): " RED_TEXT "skipping '%s'\n" NORM_TEXT,
				terminal_token.linenum, terminal_token.string);
		} else {
			terminal_token = value_token;
		}

		/* for errors return the same values that were provided: */
		/*    (array, array_len)                                 */

		vrTrace("vrParseStringListExpr()", RED_TEXT "exiting from assignment error" NORM_TEXT);
		return terminal_token;
	}

	if (assignment_token.token == VRTOKEN_PLUSEQUALS) {
		/* This line causes the list to be appended to an existing list.*/
		/* Otherwise, newarray remains null, and a new list is begun.   */
		newarray = *array;
		num_strings = *array_len;
	}

	/*******************************/
	/** get all the string tokens **/
	value_token = vrParseNextToken(parse);
	while (_TokenIsValue(value_token.token)) {
		switch (value_token.token) {
		case VRTOKEN_STRING:	/* process all string tokens */
		case VRTOKEN_NUMBER:	/* process all numeric tokens (basically as though a string) */
			num_strings++;

			/**************************************/
			/** do the assignment for this token **/

			/* For the moment, the newarray is extended using the */
			/*   brute-force method of just creating another array*/
			/*   that is one element longer, and copying all the  */
			/*   old value into it, and adding the new value.     */
			oldarray = newarray;
			newarray = (char **)vrShmemAlloc0(num_strings * sizeof(*newarray));

			for (count = 0; count < num_strings-1; count++) {
				newarray[count] = oldarray[count];
				vrDbgPrintfN(PARSE_DETAIL_DBGLVL, "vrParseStringListExpr(): copied '%s' (%d)\n", newarray[count], count);
			}
			newarray[num_strings-1] = vrShmemStrDup(value_token.string);
			vrDbgPrintfN(PARSE_DETAIL_DBGLVL, "vrParseStringListExpr(): added '%s' (%d)\n", newarray[num_strings-1], num_strings-1);
			vrShmemFree(oldarray);
			break;

		case VRTOKEN_NIL:	/* just skip variables that weren't set, then get next token */
			break;

		default:		/* obviously this shouldn't happen -- see while-statement condition */
			break;
		}


		/************************/
		/** get the next token **/
		terminal_token = vrParseNextToken(parse);
		switch (terminal_token.token) {
		default:
			vrErrPrintf("%s Parse Error (line %d): "
				RED_TEXT "Expected a comma or semicolon, but got \"%s\" (%d/%s).\n" NORM_TEXT,
				parse->parse_name, terminal_token.linenum,
				terminal_token.string, terminal_token.token,
				parse->tsmap_list[terminal_token.token+5].string	/* TODO: WARNING, the "+5" is a hack, what needs to happen is use a function that reverse maps VRTOKEN_... to an index of parse->tsmap_list (only works for first N toeksn (until strings duplicate) */ /* fix or delete me (this argument) */
				);
		case VRTOKEN_SEMICOLON:
		case VRTOKEN_CLOSECURL:
		case VRTOKEN_EOF:
			value_token = terminal_token;	/* without this the while-loop won't terminate */
			break;
		case VRTOKEN_COMMA:
			value_token = vrParseNextToken(parse);
			break;
		}
	}
	*array_len = num_strings;
	*array = newarray;

	if (vrDbgDo(PARSE_DBGLVL)) {
		vrPrintf("vrParseStringListExpr(): " BOLD_TEXT "setting '%s' to list [", setting);
		for (count = 0; count < *array_len; count++) {
			vrPrintf( BOLD_TEXT "'%s' ", (*array)[count]);
		}
		vrPrintf(BOLD_TEXT "\b]\n" NORM_TEXT
			"\t-> assignment token = '%s' (line %d), last value token = '%s' (line %d)\n",
			assignment_token.string, assignment_token.linenum,
			value_token.string, value_token.linenum);
	}

	vrTrace("vrParseStringListExpr()", RED_TEXT "exiting" NORM_TEXT);
	return (terminal_token);
}


/**********************************************************/
/* vrParseSingleEnumerator(): parses an expression to get */
/*   an enumerated string.                                */
/* Format: "<setting>" "=" string[;]                      */
vrTokenInfo vrParseSingleEnumerator(int *retvalue, vrParseEnumeratorString stringparser, const char *setting, vrParseInfo *parse)
{
	vrTokenInfo	assignment_token;	/* type of assignment used ('=' '+=' or '*=') */
	vrTokenInfo	value_token;
	vrTokenInfo	terminal_token;
	int		value;

	vrTrace("vrParseSingleEnumeratorExpr()", RED_TEXT "entering" NORM_TEXT);

	/**********************/
	/** operator ( "=" ) **/
	assignment_token = vrParseNextToken(parse);
	if (!_TokenIsAssignment(assignment_token.token)) {
		vrErrPrintf("%s Parse Error (line %d): "
			RED_TEXT "Missing assignment operator (got \"%s\") for %s.\n" NORM_TEXT,
			parse->parse_name, assignment_token.linenum, assignment_token.string, setting);

		/* unless the invalid token was a terminator, skip the rest of the statement */
		if (!_TokenIsTerminal(assignment_token.token)) {
			terminal_token = vrParseToEOS(parse);
			vrDbgPrintfN(ALWAYS_DBGLVL, "Configuration Error (line %d): " RED_TEXT "skipping '%s'\n" NORM_TEXT,
				terminal_token.linenum, terminal_token.string);
		} else {
			terminal_token = value_token;
		}

		/* return an obvious error value */
		*retvalue = -1;

		vrTrace("vrParseSingleEnumerator()", RED_TEXT "exiting from assignment error" NORM_TEXT);
		return terminal_token;
	}

	/*********************************/
	/** enumeration value of string **/
	value_token = vrParseNextToken(parse);	/* first, get the value */
	switch(value_token.token) {
	case VRTOKEN_STRING:
		value = (*stringparser)(value_token.string);
		if (value == -2) {} /* TODO: print warning message  -- also make parsers return -2 */
		break;
	case VRTOKEN_NUMBER:
		value = vrAtoI(value_token.string);
		break;
	case VRTOKEN_NIL:
		/* probably have a non-existent variable, so don't do the assignment */
		assignment_token.token = VRTOKEN_NIL;
		assignment_token.string = "<nil>";
		break;
	default:
		vrErrPrintf("%s Parse Error (line %d): "
			RED_TEXT "Invalid %s given (\"%s\").\n" NORM_TEXT,
			parse->parse_name, value_token.linenum, setting, value_token.string);
	}

	/***********************/
	/** do the assignment **/
	switch(assignment_token.token) {
	case VRTOKEN_SETEQUALS:
		*retvalue = value;
		break;
	case VRTOKEN_NOP:	/* TODO: if we ever create a VRTOKEN_NOP, change this to it */ /* [09/29/21: done! (used to be VRTOKEN_NIL)] */
		break;
	case VRTOKEN_PLUSEQUALS:
	case VRTOKEN_TIMESEQUALS:
	default:
		vrErrPrintf("%s Parse Error (line %d): "
			RED_TEXT "Invalid (or missing) assignment operator (got \"%s\") for %s.\n" NORM_TEXT,
			parse->parse_name, assignment_token.linenum, assignment_token.string, setting);

		if (!_TokenIsTerminal(value_token.token)) {
			terminal_token = vrParseToEOS(parse);
			vrDbgPrintfN(ALWAYS_DBGLVL, "Configuration Error (line %d): " RED_TEXT "skipping '%s'\n" NORM_TEXT,
				terminal_token.linenum, terminal_token.string);
		} else {
			terminal_token = value_token;
		}
		vrTrace("vrParseSingleEnumerator()", RED_TEXT "exiting from assignment error" NORM_TEXT);
		return terminal_token;
	}

	if (assignment_token.token == VRTOKEN_NIL) {
		vrDbgPrintfN(PARSE_DETAIL_DBGLVL, "vrParseSingleEnumerator(): "
			BOLD_TEXT "leaving '%s' as value %d," NORM_TEXT
			" assignment token = '%s' (line %d), value token = '%s' (line %d)\n",
			setting, value,
			assignment_token.string, assignment_token.linenum,
			value_token.string, value_token.linenum);
	} else {
		vrDbgPrintfN(PARSE_DBGLVL, "vrParseSingleEnumerator(): "
			BOLD_TEXT "setting '%s' to value %d," NORM_TEXT
			" assignment token = '%s' (line %d), value token = '%s' (line %d)\n",
			setting, value,
			assignment_token.string, assignment_token.linenum,
			value_token.string, value_token.linenum);
	}

	/***************************/	/* "{" -- to balance the one in the line below (seems modern vim ignores ones in quotes) */
	/** ( ";" | "}" | <eof> ) **/
	terminal_token = vrParseNextToken(parse);
	if (!_TokenIsTerminal(terminal_token.token))
		vrErrPrintf("%s Parse Error (line %d): "
			RED_TEXT "%s takes only one argument.\n" NORM_TEXT,
			parse->parse_name, terminal_token.linenum, setting);


	vrTrace("vrParseSingleEnumeratorExpr()", RED_TEXT "exiting" NORM_TEXT);
	return terminal_token;
}


/***********************************************************/
/* vrParseSingleIntegerExpr(): parses an expression to get */
/*   a single integer, and returns that integer.           */
/* Format: "<setting>" "=" number[;]                       */
vrTokenInfo vrParseSingleIntegerExpr(int *retvalue, const char *setting, vrParseInfo *parse)
{
	vrTokenInfo	assignment_token;	/* type of assignment used ('=' '+=' or '*=') */
	vrTokenInfo	value_token;
	vrTokenInfo	terminal_token;		/* token following the value (expecting terminal token) */
	int		value;

	vrTrace("vrParseSingleIntegerExpr()", RED_TEXT "entering" NORM_TEXT);

	/************************************/
	/** operator ( "=" | "+=" | "*=" ) **/
	assignment_token = vrParseNextToken(parse);
	if (!_TokenIsAssignment(assignment_token.token)) {
		vrErrPrintf("%s Parse Error (line %d): "
			RED_TEXT "Missing assignment operator (got \"%s\") for %s.\n" NORM_TEXT,
			parse->parse_name, assignment_token.linenum, assignment_token.string, setting);

		/* unless the invalid token was a terminator, skip the rest of the statement */
		if (!_TokenIsTerminal(assignment_token.token)) {
			terminal_token = vrParseToEOS(parse);
			vrDbgPrintfN(ALWAYS_DBGLVL, "Configuration Error (line %d): " RED_TEXT "skipping '%s'\n" NORM_TEXT,
				terminal_token.linenum, terminal_token.string);
		} else {
			terminal_token = value_token;
		}

		/* NOTE: there is no good value to return when there's an error. */
		*retvalue = 0;

		vrTrace("vrParseSingleIntegerExpr()", RED_TEXT "exiting from assignment error" NORM_TEXT);
		return terminal_token;
	}

	/***********/
	/** value **/
	value_token = vrParseNextToken(parse);
	switch (value_token.token) {
	case VRTOKEN_NUMBER:
		/* if value_token is a number, then convert to the integer value */
		value = vrAtoI(value_token.string);
		break;
	case VRTOKEN_NIL:
		/* probably have a non-existent variable, so don't do the assignment */
		assignment_token.token = VRTOKEN_NIL;
		assignment_token.string = "<nil>";
		break;
	case VRTOKEN_STRING:
		/* if value_token is a string, then see whether it has a numeric value */
		value = _ParseStringAsNumber(value_token.string);
		if (value != VRTOKEN_UNKNOWN) {
			break;
		}
		/* otherwise, we fall through into the error message of the "default" case */
	default:
		vrErrPrintf("%s Parse Error (line %d): "
			RED_TEXT "Invalid %s given (\"%s\").\n" NORM_TEXT,
			parse->parse_name, value_token.linenum, setting, value_token.string);
		if (!_TokenIsTerminal(value_token.token)) {
			value_token = vrParseToEOS(parse);
			vrDbgPrintfN(ALWAYS_DBGLVL, "Configuration Error (line %d): " RED_TEXT "skipping '%s'\n" NORM_TEXT,
				value_token.linenum, value_token.string);
		}
	}

	/***********************/
	/** do the assignment **/
	switch(assignment_token.token) {
	case VRTOKEN_SETEQUALS:
		*retvalue = value;
		break;
	case VRTOKEN_PLUSEQUALS:
		*retvalue += value;
		break;
	case VRTOKEN_TIMESEQUALS:
		*retvalue *= value;
		break;
	case VRTOKEN_NOP:	/* TODO: if we ever create a VRTOKEN_NOP, change this to it */ /* [09/29/21: done! (used to be VRTOKEN_NIL)] */
		break;
	default:
		vrErrPrintf("%s Parse Error (line %d): "
			RED_TEXT "Invalid (or missing) assignment operator (got \"%s\") for %s.\n" NORM_TEXT,
			parse->parse_name, assignment_token.linenum, assignment_token.string, setting);

		if (!_TokenIsTerminal(value_token.token)) {
			terminal_token = vrParseToEOS(parse);
			vrDbgPrintfN(ALWAYS_DBGLVL, "Configuration Error (line %d): " RED_TEXT "skipping '%s'\n" NORM_TEXT,
				terminal_token.linenum, terminal_token.string);
		} else {
			terminal_token = value_token;
		}

		vrTrace("vrParseSingleIntegerExpr()", RED_TEXT "exiting from assignment error" NORM_TEXT);
		return terminal_token;
	}

	if (assignment_token.token == VRTOKEN_NIL) {
		vrDbgPrintfN(PARSE_DBGLVL, "vrParseSingleIntegerExpr(): "
			BOLD_TEXT "leaving '%s' as value %d," NORM_TEXT
			" assignment token = '%s' (line %d), value token = '%s' (line %d)\n",
			setting, *retvalue,
			assignment_token.string, assignment_token.linenum,
			value_token.string, value_token.linenum);
	} else {
		vrDbgPrintfN(PARSE_DBGLVL, "vrParseSingleIntegerExpr(): "
			BOLD_TEXT "setting '%s' to value %d," NORM_TEXT
			" assignment token = '%s' (line %d), value token = '%s' (line %d)\n",
			setting, *retvalue,
			assignment_token.string, assignment_token.linenum,
			value_token.string, value_token.linenum);
	}

	/***************************/	/* "{" -- to balance the one in the line below */
	/** ( ";" | "}" | <eof> ) **/
	terminal_token = vrParseNextToken(parse);
	if (!_TokenIsTerminal(terminal_token.token))
		vrErrPrintf("%s Parse Error (line %d): "
			RED_TEXT "%s takes only one argument.\n" NORM_TEXT,
			parse->parse_name, terminal_token.linenum, setting);

	vrTrace("vrParseSingleIntegerExpr()", RED_TEXT "exiting" NORM_TEXT);
	return terminal_token;
}


/*********************************************************************/
/* vrParseIntegerListExpr(): ... */
/* NOTE: the numeric list expression parsers generally require   */
/*   a specific number of values.  A warning message is printed  */
/*   when the number of given numbers doesn't match the expected */
/*   list length -- however this is not a hard error.  On the    */
/*   other hand, any string that is parsed will be sent to find  */
/*   the string value, and any missing variable will be replaced */
/*   with a zero value -- which doesn't follow the standard of   */
/*   the other expression parsers (except the other number lists.*/
vrTokenInfo vrParseIntegerListExpr(int *array, int maxlen, const char *setting, vrParseInfo *parse)
{
	vrTokenInfo	assignment_token;	/* type of assignment used ('=' '+=' or '*=') */
	vrTokenInfo	value_token;
	vrTokenInfo	terminal_token;		/* comma or standard terminating token */
	int		value;
	int		count;

	vrTrace("vrParseIntegerListExpr()", RED_TEXT "entering" NORM_TEXT);

	/***************************/
	/** ( "=" | "+=" | "*=" ) **/
	assignment_token = vrParseNextToken(parse);
	if (!_TokenIsAssignment(assignment_token.token)) {
		vrErrPrintf("%s Parse Error (line %d): "
			RED_TEXT "Expecting an assignment operator for System Procs.\n" NORM_TEXT,
			parse->parse_name, assignment_token.linenum);

		/* unless the invalid token was a terminator, skip the rest of the statement */
		if (!_TokenIsTerminal(assignment_token.token)) {
			terminal_token = vrParseToEOS(parse);
			vrDbgPrintfN(ALWAYS_DBGLVL, "Configuration Error (line %d): " RED_TEXT "skipping '%s'\n" NORM_TEXT,
				terminal_token.linenum, terminal_token.string);
		} else {
			terminal_token = value_token;
		}

		/* for errors return the same values that were provided: */
		/*   (array)                                             */

		vrTrace("vrParseIntegerListExpr()", RED_TEXT "exiting from assignment error" NORM_TEXT);
		return terminal_token;
	}

	/*******************/
	/** number values **/
	value_token = vrParseNextToken(parse);

	for (count = 0; count < maxlen && _TokenIsValue(value_token.token); count++) {
		switch (value_token.token) {
		case VRTOKEN_NUMBER:
			value = vrAtoI(value_token.string);
			break;
		case VRTOKEN_STRING:
			switch (_ParseStringAsNumber(value_token.string)) {
			case VRTOKEN_NO:
				value = 0;
				break;
			case VRTOKEN_YES:
				value = 1;
				break;
			default:
				vrErrPrintf("%s Parse Warning (line %d): "
					RED_TEXT "Non-numeric string '%s' given as number.\n" NORM_TEXT,
					parse->parse_name, value_token.linenum, value_token.string);
				value_token.token = VRTOKEN_NIL;	/* set to <nil> so no assignment will be done */
				break;
			}
			break;
		case VRTOKEN_NIL:
			break;
		}

		/***********************/
		/** do the assignment **/
		/* NOTE: assignments are not made when the value is the <nil> token, so old values remain */
		switch(assignment_token.token) {
		case VRTOKEN_SETEQUALS:
			if (value_token.token != VRTOKEN_NIL)
				array[count] = value;
			break;
		case VRTOKEN_PLUSEQUALS:
			if (value_token.token != VRTOKEN_NIL)
				array[count] += value;
			break;
		case VRTOKEN_TIMESEQUALS:
			if (value_token.token != VRTOKEN_NIL)
				array[count] *= value;
			break;
		default:
			vrErrPrintf("%s Parse Error (line %d): "
				RED_TEXT "Invalid (or missing) assignment operator (got \"%s\") for %s.\n" NORM_TEXT,
				parse->parse_name, assignment_token.linenum, assignment_token.string, setting);

#if 0 /* the condition of this if-statement can never be true (05/25/2006: "true"? or false?  Seems like the latter makes more sense. */
			if (!_TokenIsTerminal(value_token.token)) {
#endif
				terminal_token = vrParseToEOS(parse);
				vrDbgPrintfN(ALWAYS_DBGLVL, "Configuration Error (line %d): " RED_TEXT" skipping '%s'\n" NORM_TEXT,
					terminal_token.linenum, terminal_token.string);
#if 0 /* the condition of this if-statement can never be true */
			} else {
				terminal_token = value_token;
			}
#endif
			vrTrace("vrParseIntegerListExpr()", RED_TEXT "exiting from assignment error" NORM_TEXT);
			return terminal_token;
		}

		/*********************************/	/* "{" -- to balance the one in the line below */
		/** ( "," | ";" | "}" | <eof> ) **/
		terminal_token = vrParseNextToken(parse);
		switch (terminal_token.token) {
		case VRTOKEN_COMMA:
			value_token = vrParseNextToken(parse);
			break;
		case VRTOKEN_SEMICOLON:
		case VRTOKEN_CLOSECURL:
		case VRTOKEN_EOF:
			value_token = terminal_token;
			continue;
		default:
			vrErrPrintf("%s Parse Error (line %d): "
				RED_TEXT "%s: Expected a comma or statement terminator.\n" NORM_TEXT,
				parse->parse_name, terminal_token.linenum, setting);
			value_token = terminal_token;
		}
	}

	if (count < maxlen)
		vrErrPrintf("%s Parse Warning (line %d): "
			RED_TEXT "%s: Less than the full list of %d numbers given.\n" NORM_TEXT,
			parse->parse_name, value_token.linenum, setting, maxlen);

	if (vrDbgDo(PARSE_DBGLVL)) {
		vrPrintf("vrParseIntegerListExpr(): " BOLD_TEXT "setting '%s' to list [", setting);
		for (count = 0; count < maxlen; count++) {
			vrPrintf( BOLD_TEXT "%d ", array[count]);
		}
		vrPrintf(BOLD_TEXT "\b]" NORM_TEXT
			" assignment token = '%s' (line %d), last value token = '%s' (line %d)\n",
			assignment_token.string, assignment_token.linenum,
			value_token.string, value_token.linenum);
	}

	vrTrace("vrParseIntegerListExpr()", RED_TEXT "exiting" NORM_TEXT);
	return terminal_token;
}


/***********************************************************/
/* vrParseSingleFloatExpr(): parses an expression to get a */
/*   single float, and returns that float.                 */
/* Format: "<setting>" "=" number[;]                       */
vrTokenInfo vrParseSingleFloatExpr(float *retvalue, const char *setting, vrParseInfo *parse)
{
	vrTokenInfo	assignment_token;	/* type of assignment used ('=' '+=' or '*=') */
	vrTokenInfo	value_token;
	vrTokenInfo	terminal_token;		/* token following the value (expecting terminal token) */
	float		value;

	vrTrace("vrParseSingleFloatExpr()", RED_TEXT "entering" NORM_TEXT);

	/************************************/
	/** operator ( "=" | "+=" | "*=" ) **/
	assignment_token = vrParseNextToken(parse);
	if (!_TokenIsAssignment(assignment_token.token)) {
		vrErrPrintf("%s Parse Error (line %d): "
			RED_TEXT "Missing assignment operator (got \"%s\") for %s.\n" NORM_TEXT,
			parse->parse_name, assignment_token.linenum, assignment_token.string, setting);

		/* unless the invalid token was a terminator, skip the rest of the statement */
		if (!_TokenIsTerminal(assignment_token.token)) {
			terminal_token = vrParseToEOS(parse);
			vrDbgPrintfN(ALWAYS_DBGLVL, "Configuration Error (line %d): " RED_TEXT "skipping '%s'\n" NORM_TEXT,
				terminal_token.linenum, terminal_token.string);
		} else {
			terminal_token = value_token;
		}

		/* NOTE: there is no good value to return when there's an error. */
		*retvalue = 0.0;

		vrTrace("vrParseSingleFloatExpr()", RED_TEXT "exiting from assignment error" NORM_TEXT);
		return terminal_token;
	}

	/***********/
	/** value **/
	value_token = vrParseNextToken(parse);
	switch (value_token.token) {
	case VRTOKEN_NUMBER:
		/* if value_token is a number, then convert to the float value */
		value = atof(value_token.string);
		break;
	case VRTOKEN_NIL:
		/* probably have a non-existent variable, so don't do the assignment */
		assignment_token.token = VRTOKEN_NIL;
		assignment_token.string = "<nil>";
		break;
	case VRTOKEN_STRING:
		value = _ParseStringAsNumber(value_token.string);
		if (value != VRTOKEN_UNKNOWN) {
			break;
		}
		/* otherwise, we fall through into the error message of the "default" case */
	default:
		vrErrPrintf("%s Parse Error (line %d): "
			RED_TEXT "Invalid %s given (\"%s\").\n" NORM_TEXT,
			parse->parse_name, value_token.linenum, setting, value_token.string);

		/* At this point we're in a messed-up statement, so if we're */
		/*   not at a terminator, read read of statement.            */
		if (!_TokenIsTerminal(value_token.token)) {
			value_token = vrParseToEOS(parse);
			vrDbgPrintfN(ALWAYS_DBGLVL, "Configuration Error (line %d): " RED_TEXT "skipping '%s'\n" NORM_TEXT,
				value_token.linenum, value_token.string);
		}
	}

	/***********************/
	/** do the assignment **/
	switch(assignment_token.token) {
	case VRTOKEN_SETEQUALS:
		*retvalue = value;
		break;
	case VRTOKEN_PLUSEQUALS:
		*retvalue += value;
		break;
	case VRTOKEN_TIMESEQUALS:
		*retvalue *= value;
		break;
	case VRTOKEN_NOP:	/* TODO: if we ever create a VRTOKEN_NOP, change this to it */ /* [09/29/21: done! (used to be VRTOKEN_NIL)] */
		break;
	default:
		vrErrPrintf("%s Parse Error (line %d): "
			RED_TEXT "Invalid (or missing) assignment operator (got \"%s\") for %s.\n" NORM_TEXT,
			parse->parse_name, assignment_token.linenum, assignment_token.string, setting);

		if (!_TokenIsTerminal(value_token.token)) {
			terminal_token = vrParseToEOS(parse);
			vrDbgPrintfN(ALWAYS_DBGLVL, "Configuration Error (line %d): " RED_TEXT "skipping '%s'\n" NORM_TEXT,
				terminal_token.linenum, terminal_token.string);
		} else {
			terminal_token = value_token;
		}

		vrTrace("vrParseSingleFloatExpr()", RED_TEXT "exiting from assignment error" NORM_TEXT);
		return terminal_token;
	}

	if (assignment_token.token == VRTOKEN_NIL) {
		vrDbgPrintfN(PARSE_DBGLVL, "vrParseSingleFloatExpr(): "
			BOLD_TEXT "leaving '%s' as value %.4f," NORM_TEXT
			" assignment token = '%s' (line %d), value token = '%s' (line %d)\n",
			setting, *retvalue,
			assignment_token.string, assignment_token.linenum,
			value_token.string, value_token.linenum);
	} else {
		vrDbgPrintfN(PARSE_DBGLVL, "vrParseSingleFloatExpr(): "
			BOLD_TEXT "setting '%s' to value %.4f," NORM_TEXT
			" assignment token = '%s' (line %d), value token = '%s' (line %d)\n",
			setting, *retvalue,
			assignment_token.string, assignment_token.linenum,
			value_token.string, value_token.linenum);
	}

	/***************************/	/* "{" -- to balance the one in the line below */
	/** ( ";" | "}" | <eof> ) **/
	terminal_token = vrParseNextToken(parse);
	if (!_TokenIsTerminal(terminal_token.token))
		vrErrPrintf("%s Parse Error (line %d): "
			RED_TEXT "%s takes only one argument.\n" NORM_TEXT,
			parse->parse_name, terminal_token.linenum, setting);

	vrTrace("vrParseSingleFloatExpr()", RED_TEXT "exiting" NORM_TEXT);
	return terminal_token;
}


/*********************************************************************/
/* vrParseFloatListExpr(): ... */
vrTokenInfo vrParseFloatListExpr(float *array, int maxlen, const char *setting, vrParseInfo *parse)
{
	vrTokenInfo	assignment_token;	/* type of assignment used ('=' '+=' or '*=') */

#if 1 /* 12/17/02 -- test the necessity of the duplicate function by setting to 0 [ */
	/* TODO: [1/4/2] why do we do this?  It seems this isn't necessary. */
	/* 04/16/21: test this and if not needed -- delete me */
	/* 09/30/21: on the double-precision side, the "Assign" version is */
	/*   used quite frequently in order to get the "assignment_token"  */
	/*   returned back to the calling function.  So maybe we want to   */
	/*   keep this split here too, but then maybe we just don't ever   */
	/*   worry about floats when reading the configuration.            */

	return vrParseFloatListExprAssign(array, maxlen, &assignment_token, setting, parse);
}


/*********************************************************************/
/* NOTE: the numeric list expression parsers generally require a specific number of */
/*   values.  A warning message is printed when the number of given numbers doesn't */
/*   match the expected list length -- however this is not a hard error.  On the    */
/*   other hand, any string that is parsed will be sent to find the string value,   */
/*   and any missing variable will be replaced with a zero value -- which doesn't   */
/*   follow the standard of the other expression parsers (except the other number   */
/*   lists.                                                                         */
vrTokenInfo vrParseFloatListExprAssign(float *array, int maxlen, vrTokenInfo *assignment_token, const char *setting, vrParseInfo *parse)
{
# define PARSEFLOATLISTEXPR_DEREF
#endif /* ] */
	vrTokenInfo	value_token;
	vrTokenInfo	terminal_token;		/* comma or standard terminating token */
	float		value;
	int		count;

	vrTrace("vrParseFloatListExprAssign()", RED_TEXT "entering" NORM_TEXT);

	/************************************/
	/** operator ( "=" | "+=" | "*=" ) **/
#ifdef PARSEFLOATLISTEXPR_DEREF	/* NOTE: remove the dereference if combining this function with the above */
	*assignment_token = vrParseNextToken(parse);
	if (!_TokenIsAssignment(assignment_token->token))
#else
	assignment_token = vrParseNextToken(parse);
	if (!_TokenIsAssignment(assignment_token.token))
#endif
	{
		vrErrPrintf("%s Parse Error (line %d): "
			RED_TEXT "Missing assignment operator (got \"%s\") for %s.\n" NORM_TEXT,
#ifdef PARSEFLOATLISTEXPR_DEREF	/* NOTE: remove the dereference if combining this function with the above */
			parse->parse_name, assignment_token->linenum, assignment_token->string, setting);
#else
			parse->parse_name, assignment_token.linenum, assignment_token.string, setting);
#endif

		/* unless the invalid token was a terminator, skip the rest of the statement */
#ifdef PARSEFLOATLISTEXPR_DEREF	/* NOTE: remove the dereference if combining this function with the above */
		if (!_TokenIsTerminal(assignment_token->token))
#else
		if (!_TokenIsTerminal(assignment_token.token))
#endif
		{
			terminal_token = vrParseToEOS(parse);
			vrDbgPrintfN(ALWAYS_DBGLVL, "Configuration Error (line %d): " RED_TEXT "skipping '%s'\n" NORM_TEXT,
				terminal_token.linenum, terminal_token.string);
		} else {
			terminal_token = value_token;
		}

		/* NOTE: there is no good value to return when there's an error. */

		vrTrace("vrParseFloatListExprAssign()", RED_TEXT "exiting from assignment error" NORM_TEXT);
		return terminal_token;
	}

	/************/
	/** values **/
	value_token = vrParseNextToken(parse);

	for (count = 0; count < maxlen && _TokenIsValue(value_token.token); count++) {
		switch (value_token.token) {
		case VRTOKEN_NUMBER:
			value = atof(value_token.string);
			break;
		case VRTOKEN_STRING:
			switch (_ParseStringAsNumber(value_token.string)) {
			case VRTOKEN_NO:
				value = 0.0;
				break;
			case VRTOKEN_YES:
				value = 1.0;
				break;
			default:
				vrErrPrintf("%s Parse Warning (line %d): "
					RED_TEXT "Non-numeric string '%s' given as number.\n" NORM_TEXT,
					parse->parse_name, value_token.linenum, value_token.string);
				value_token.token = VRTOKEN_NIL;	/* set to <nil> so no assignment will be done */
				break;
			}
			break;
		case VRTOKEN_NIL:
			break;
		}

		/***********************/
		/** do the assignment **/
		/* NOTE: assignments are not made when the value is the <nil> token, so old values remain */
		switch(assignment_token->token) {
		case VRTOKEN_SETEQUALS:
			if (value_token.token != VRTOKEN_NIL)
				array[count] = value;
			break;
		case VRTOKEN_PLUSEQUALS:
			if (value_token.token != VRTOKEN_NIL)
				array[count] += value;
			break;
		case VRTOKEN_TIMESEQUALS:
			if (value_token.token != VRTOKEN_NIL)
				array[count] *= value;
			break;
		default:
			vrErrPrintf("%s Parse Error (line %d): "
				RED_TEXT "Invalid (or missing) assignment operator (got \"%s\") for %s.\n" NORM_TEXT,
				parse->parse_name, assignment_token->linenum, assignment_token->string, setting);

			terminal_token = vrParseToEOS(parse);
			vrDbgPrintfN(ALWAYS_DBGLVL, "Configuration Error (line %d): " RED_TEXT "skipping '%s'\n" NORM_TEXT,
				terminal_token.linenum, terminal_token.string);

			vrTrace("vrParseFloatListExprAssign()", RED_TEXT "exiting from assignment error" NORM_TEXT);
			return terminal_token;
		}

		/*********************************/	/* "{" -- to balance the one in the line below */
		/** ( "," | ";" | "}" | <eof> ) **/
		terminal_token = vrParseNextToken(parse);
		switch (terminal_token.token) {
		case VRTOKEN_COMMA:
			value_token = vrParseNextToken(parse);
			break;
		case VRTOKEN_SEMICOLON:
		case VRTOKEN_CLOSECURL:
		case VRTOKEN_EOF:
			value_token = terminal_token;
			continue;
		default:
			vrErrPrintf("%s Parse Error (line %d): "
				RED_TEXT "%s: Expected a comma or statement terminator.\n" NORM_TEXT,
				parse->parse_name, terminal_token.linenum, setting);
			value_token = terminal_token;
		}
	}

	if (count < maxlen)
		vrErrPrintf("%s Parse Warning (line %d): "
			RED_TEXT "%s: Less than the full list of %d numbers given.\n" NORM_TEXT,
			parse->parse_name, value_token.linenum, setting, maxlen);

	if (vrDbgDo(PARSE_DBGLVL)) {
		vrPrintf("vrParseFloatListExprAssign(): " BOLD_TEXT "setting '%s' to list [", setting);
		for (count = 0; count < maxlen; count++) {
			vrPrintf( BOLD_TEXT "%.4f ", array[count]);
		}
		vrPrintf(BOLD_TEXT "\b]" NORM_TEXT
			" assignment token = '%s' (line %d), last value token = '%s' (line %d)\n",
			assignment_token->string, assignment_token->linenum,
			value_token.string, value_token.linenum);
	}

	vrTrace("vrParseFloatListExprAssign()", RED_TEXT "exiting" NORM_TEXT);
	return terminal_token;
}


/******************************************************************************/
/* vrParseDoubleListExprAssign(): ... */
/* TODO: [1/4/2] why does this function take 'assignment_token' as an argument, */
/*   and vrParseFloatListExpr() doesn't???                                    */
/*   Answer: because this (these?) are called by the Matrix parsing functions below */
/* This function parses the assignment token, and all the (maxlen) numbers    */
/*   that go with a particular type of operation.  However, the operation     */
/*   itself is not performed here.  Thus, many different types of operations  */
/*   can use this function to get the basic inputs, and then perform the      */
/*   appropriate operation.                                                   */
/* [01/11/2021: TODO: I'm not sure about the "operation itself is not performed */
/*   here statement above.  Certainly there are += & *= operations done.]       */
/*                                                                            */
/* Arguments:                                                                 */
/*    * array -- the original values of the array[maxlen] & RETURN new values */
/*    * maxlen -- the maxiumum size of the array (number of numbers to parse) */
/*    * assignment_token -- RETURNS the type of assignment made (= / += / *=) */
/*    * parse -- parsing system info -- (used to extract upcoming tokens)     */
/* Returns:                                                                   */
/*    * via return -- the expression terminating token (',' ';' '<eof>' or curly */
/*    * array (via CBR) -- the modified array based on the given operation    */
/*    * assignment_token (via CBR) -- the type of assignment made (= / += / *=) */
vrTokenInfo vrParseDoubleListExprAssign(double *array, int maxlen, vrTokenInfo *assignment_token, const char *setting, vrParseInfo *parse)
{
	vrTokenInfo	value_token;
	vrTokenInfo	terminal_token;		/* comma or standard terminating token */
	double		value;
	int		count;

	vrTrace("vrParseDoubleListExprAssign()", RED_TEXT "entering" NORM_TEXT);

	/************************************/
	/** operator ( "=" | "+=" | "*=" ) **/
	*assignment_token = vrParseNextToken(parse);
	if (!_TokenIsAssignment(assignment_token->token)) {
		vrErrPrintf("%s Parse Error (line %d): "
			RED_TEXT "Missing assignment operator (got \"%s\") for %s.\n" NORM_TEXT,
			parse->parse_name, assignment_token->linenum, assignment_token->string, setting);

		/* unless the invalid token was a terminator, skip the rest of the statement */
		if (!_TokenIsTerminal(assignment_token->token)) {
			terminal_token = vrParseToEOS(parse);
			vrDbgPrintfN(ALWAYS_DBGLVL, "Configuration Error (line %d): " RED_TEXT "skipping '%s'\n" NORM_TEXT,
				terminal_token.linenum, terminal_token.string);
		} else {
			terminal_token = value_token;
		}

		/* NOTE: there is no good value to return when there's an error. */

		vrTrace("vrParseDoubleListExprAssign()", RED_TEXT "exiting from assignment error" NORM_TEXT);
		return terminal_token;
	}

	/************/
	/** values **/
	value_token = vrParseNextToken(parse);

	for (count = 0; count < maxlen && _TokenIsValue(value_token.token); count++) {
		switch (value_token.token) {
		case VRTOKEN_NUMBER:
			value = atof(value_token.string);
			break;
		case VRTOKEN_STRING:
			switch (_ParseStringAsNumber(value_token.string)) {
			case VRTOKEN_NO:
				value = 0.0;
				break;
			case VRTOKEN_YES:
				value = 1.0;
				break;
			default:
				vrErrPrintf("%s Parse Warning (line %d): "
					RED_TEXT "Non-numeric string '%s' given as number.\n" NORM_TEXT,
					parse->parse_name, value_token.linenum, value_token.string);
				value_token.token = VRTOKEN_NIL;	/* set to <nil> so no assignment will be done */
				break;
			}
			break;
		case VRTOKEN_NIL:
			break;
		}

		/***********************/
		/** do the assignment **/
		/* NOTE: assignments are not made when the value is the <nil> */
		/*   token, so old values remain.                             */
		/* TODO: consider setting the array[count] value to 0.0 for   */
		/*   VRTOKEN_NIL cases -- I think that might make more sense  */
		/*   here because those values will then be used to operate   */
		/*   on some other matrix or something.                       */
		switch(assignment_token->token) {
		case VRTOKEN_SETEQUALS:
			if (value_token.token != VRTOKEN_NIL)
				array[count] = value;
			break;
		case VRTOKEN_PLUSEQUALS:
			if (value_token.token != VRTOKEN_NIL)
				array[count] += value;
			break;
		case VRTOKEN_TIMESEQUALS:
			if (value_token.token != VRTOKEN_NIL)
				array[count] *= value;
			break;
		default:
			vrErrPrintf("%s Parse Error (line %d): "
				RED_TEXT "Invalid (or missing) assignment operator (got \"%s\") for %s.\n" NORM_TEXT,
				parse->parse_name, assignment_token->linenum, assignment_token->string, setting);

			terminal_token = vrParseToEOS(parse);
			vrDbgPrintfN(ALWAYS_DBGLVL, "Configuration Error (line %d): " RED_TEXT "skipping '%s'\n" NORM_TEXT,
				terminal_token.linenum, terminal_token.string);

			vrTrace("vrParseDoubleListExprAssign()", RED_TEXT "exiting from assignment error" NORM_TEXT);
			return terminal_token;
		}

		/*********************************/	/* "{" -- to balance the one in the line below */
		/** ( "," | ";" | "}" | <eof> ) **/
		terminal_token = vrParseNextToken(parse);
		switch (terminal_token.token) {
		case VRTOKEN_COMMA:
			value_token = vrParseNextToken(parse);
			break;
		case VRTOKEN_SEMICOLON:
		case VRTOKEN_CLOSECURL:
		case VRTOKEN_EOF:
			value_token = terminal_token;
			continue;
		default:
			vrErrPrintf("%s Parse Error (line %d): "
				RED_TEXT "%s: Unexpected end of list.\n" NORM_TEXT,
				parse->parse_name, terminal_token.linenum, setting);
			value_token = terminal_token;
		}
	}

	if (count < maxlen)
		vrErrPrintf("%s Parse Warning (line %d): "
			RED_TEXT "%s: Less than the full list of %d numbers given.\n" NORM_TEXT,
			parse->parse_name, value_token.linenum, setting, maxlen);

	if (vrDbgDo(PARSE_DBGLVL)) {
		vrPrintf("vrParseDoubleListExprAssign(): " BOLD_TEXT "setting '%s' to list [", setting);
		for (count = 0; count < maxlen; count++) {
			vrPrintf( BOLD_TEXT "%.4lf ", array[count]);
		}
		vrPrintf(BOLD_TEXT "\b]" NORM_TEXT
			" assignment token = '%s' (line %d), last value token = '%s' (line %d)\n",
			assignment_token->string, assignment_token->linenum,
			value_token.string, value_token.linenum);
	}

	vrTrace("vrParseDoubleListExprAssign()", RED_TEXT "exiting" NORM_TEXT);
	return terminal_token;
}


/************************************************************************/
/* vrParseMatrixFromCoords(): parses a sequence of numbers as three 3D  */
/*   coordinates that can be used to define an orientation of a         */
/*   coordinate system.  The created matrix is returned via             */
/*   call-by-reference in "xform", as well the 3 coordinates themselves */
/*   as call-by-reference in "coords_ll", "coords_lr", & "coords_ul".   */
vrTokenInfo vrParseMatrixFromCoords(vrMatrix **xform, double *coords_ll, double *coords_lr, double *coords_ul, const char *setting, vrParseInfo *parse)
{
	vrTokenInfo	assignment_token;	/* type of assignment used ('=' '+=' or '*=') */
	vrTokenInfo	terminal_token;		/* the terminal token returned by vrParseDoubleListExprAssign() */
	double		coords[9];		/* temporary space for the list of numbers to operate with */
	vrMatrix	tmp_matrix;		/* temporary matrix for doing the operation */

	/********************/
	/** get the values **/
	terminal_token = vrParseDoubleListExprAssign(coords, 9, &assignment_token, setting, parse);
	vrMatrixSetFromCoords(&tmp_matrix, coords);

	/******************************************************************/
	/** redundantly assign the values to call-by-reference arguments **/
	coords_ll[VR_X] = coords[0];
	coords_ll[VR_Y] = coords[1];
	coords_ll[VR_Z] = coords[2];
	coords_lr[VR_X] = coords[3];
	coords_lr[VR_Y] = coords[4];
	coords_lr[VR_Z] = coords[5];
	coords_ul[VR_X] = coords[6];
	coords_ul[VR_Y] = coords[7];
	coords_ul[VR_Z] = coords[8];

	/***********************/
	/** do the assignment **/
	switch (assignment_token.token) {
	case VRTOKEN_SETEQUALS:
		/* TODO: get rid of leak when vrObjectCopy fixed  [2/2/00 -- not so sure anymore] */
		*xform = vrMatrixCreate();
		vrMatrixCopy(*xform, &tmp_matrix);
		break;
	case VRTOKEN_TIMESEQUALS:
		/* TODO: get rid of leak when vrObjectCopy fixed  [2/2/00 -- not so sure anymore] */
		*xform = vrMatrixCopy(vrMatrixCreate(), *xform);
		vrMatrixPreMult(*xform, &tmp_matrix);
		break;
	case VRTOKEN_PLUSEQUALS:
	default:
		vrErrPrintf("%s Parse Error (line %d): "
			RED_TEXT "Invalid (or missing) assignment operator (got \"%s\") for %s.\n" NORM_TEXT,
			parse->parse_name, assignment_token.linenum, assignment_token.string, setting);

		if (!_TokenIsTerminal(terminal_token.token)) {
			terminal_token = vrParseToEOS(parse);
			vrDbgPrintfN(ALWAYS_DBGLVL, "Configuration Error (line %d): " RED_TEXT "skipping '%s'\n" NORM_TEXT,
				terminal_token.linenum, terminal_token.string);
		}
		vrTrace("vrParseMatrixFromCoords()", RED_TEXT "exiting from assignment error" NORM_TEXT);
		return terminal_token;
	}

	if (vrDbgDo(PARSE_DBGLVL)) {
		vrPrintf("vrParseMatrixFromCoords(): " BOLD_TEXT "setting '%s' with assignment '%s' (line %d):\n",
			setting, assignment_token.string, assignment_token.linenum);
		vrPrintMatrix(setting, *xform);
		vrPrintMatrixComposition(setting, *xform);

	}


	return terminal_token;
}


/*********************************************************************/
/* vrParseMatrixTransform(): ... */
vrTokenInfo vrParseMatrixTransform(vrMatrix **xform, const char *setting, vrParseInfo *parse)
{
	vrTokenInfo	assignment_token;	/* type of assignment used ('=' '+=' or '*=') */
	vrTokenInfo	terminal_token;		/* the terminal token returned by vrParseDoubleListExprAssign() */
	vrMatrix	tmp_matrix;		/* temporary matrix for doing the operation */

	/********************/
	/** get the values **/

	/* tmp_matrix values must be set 1.0 for the possible use of the *= assignment */
	tmp_matrix.v[ 0] = tmp_matrix.v[ 1] = tmp_matrix.v[ 2] = tmp_matrix.v[ 3] = 1.0;
	tmp_matrix.v[ 4] = tmp_matrix.v[ 5] = tmp_matrix.v[ 6] = tmp_matrix.v[ 7] = 1.0;
	tmp_matrix.v[ 8] = tmp_matrix.v[ 9] = tmp_matrix.v[10] = tmp_matrix.v[11] = 1.0;
	tmp_matrix.v[12] = tmp_matrix.v[13] = tmp_matrix.v[14] = tmp_matrix.v[15] = 1.0;

	terminal_token = vrParseDoubleListExprAssign(tmp_matrix.v, 16, &assignment_token, setting, parse);

	/***********************/
	/** do the assignment **/
	switch (assignment_token.token) {
	case VRTOKEN_SETEQUALS:
		*xform = vrMatrixCreate();	/* TODO: get rid of leak when vrObjectCopy fixed  [2/2/00 -- not so sure anymore] */
		vrMatrixCopy(*xform, &tmp_matrix);
		break;
	case VRTOKEN_TIMESEQUALS:
		/* TODO: get rid of leak when vrObjectCopy fixed  [2/2/00 -- not so sure anymore] */
		*xform = vrMatrixCopy(vrMatrixCreate(), *xform);
		vrMatrixPreMult(*xform, &tmp_matrix);
		break;
	case VRTOKEN_PLUSEQUALS:
	default:
		vrErrPrintf("%s Parse Error (line %d): "
			RED_TEXT "Invalid (or missing) assignment operator (got \"%s\") for %s.\n" NORM_TEXT,
			parse->parse_name, assignment_token.linenum, assignment_token.string, setting);

		if (!_TokenIsTerminal(terminal_token.token)) {
			terminal_token = vrParseToEOS(parse);
			vrDbgPrintfN(ALWAYS_DBGLVL, "Configuration Error (line %d): " RED_TEXT "skipping '%s'\n" NORM_TEXT,
				terminal_token.linenum, terminal_token.string);
		}
		vrTrace("vrParseMatrixTransform()", RED_TEXT "exiting from assignment error" NORM_TEXT);
		return terminal_token;
	}

	if (vrDbgDo(PARSE_DBGLVL)) {
		vrPrintf("vrParseMatrixTransform(): " BOLD_TEXT "setting '%s' with assignment '%s' (line %d):\n",
			setting, assignment_token.string, assignment_token.linenum);
		vrPrintMatrix(setting, *xform);
		vrPrintMatrixComposition(setting, *xform);
	}

	return terminal_token;
}


#ifdef EXPERIMENTAL
/*********************************************************************/
/* vrParseMatrixTransform(): ... */
/* NOTE: I hate that I basically have to duplicate the above function to  */
/*  add a coordinates option, but I guess that's the one limitation of C, */
/*  (without delving into varargs, or trying to calculate an inverse      */
/*  matrix after calling vrParseMatrixTranslate().                        */
vrTokenInfo vrParseMatrixTransformWithCoords(vrMatrix **xform, const char *setting, vrParseInfo *parse, vrPoint *coords_ll, vrPoint *coords_lr, vrPoint *coords_ul)
{
	vrTokenInfo	assignment_token;	/* type of assignment used ('=' '+=' or '*=') */
	vrTokenInfo	terminal_token;		/* the terminal token returned by vrParseDoubleListExprAssign() */
	vrMatrix	tmp_matrix;		/* temporary matrix for doing the operation */

	/********************/
	/** get the values **/

	/* tmp_matrix values must be set 1.0 for the possible use of the *= assignment */
	tmp_matrix.v[ 0] = tmp_matrix.v[ 1] = tmp_matrix.v[ 2] = tmp_matrix.v[ 3] = 1.0;
	tmp_matrix.v[ 4] = tmp_matrix.v[ 5] = tmp_matrix.v[ 6] = tmp_matrix.v[ 7] = 1.0;
	tmp_matrix.v[ 8] = tmp_matrix.v[ 9] = tmp_matrix.v[10] = tmp_matrix.v[11] = 1.0;
	tmp_matrix.v[12] = tmp_matrix.v[13] = tmp_matrix.v[14] = tmp_matrix.v[15] = 1.0;

	terminal_token = vrParseDoubleListExprAssign(tmp_matrix.v, 16, &assignment_token, setting, parse);

	/***********************/
	/** do the assignment **/
	switch (assignment_token.token) {
	case VRTOKEN_SETEQUALS:
		*xform = vrMatrixCreate();	/* TODO: get rid of leak when vrObjectCopy fixed  [2/2/00 -- not so sure anymore] */
		vrMatrixCopy(*xform, &tmp_matrix);
		break;
	case VRTOKEN_TIMESEQUALS:
		/* TODO: get rid of leak when vrObjectCopy fixed  [2/2/00 -- not so sure anymore] */
		*xform = vrMatrixCopy(vrMatrixCreate(), *xform);
		vrMatrixPreMult(*xform, &tmp_matrix);
		break;
	case VRTOKEN_PLUSEQUALS:
	default:
		vrErrPrintf("%s Parse Error (line %d): "
			RED_TEXT "Invalid (or missing) assignment operator (got \"%s\") for %s.\n" NORM_TEXT,
			parse->parse_name, assignment_token.linenum, assignment_token.string, setting);

		if (!_TokenIsTerminal(terminal_token.token)) {
			terminal_token = vrParseToEOS(parse);
			vrDbgPrintfN(ALWAYS_DBGLVL, "Configuration Error (line %d): " RED_TEXT "skipping '%s'\n" NORM_TEXT,
				terminal_token.linenum, terminal_token.string);
		}
		vrTrace("vrParseMatrixTransform()", RED_TEXT "exiting from assignment error" NORM_TEXT);
		return terminal_token;
	}

	if (vrDbgDo(PARSE_DBGLVL)) {
		vrPrintf("vrParseMatrixTransform(): " BOLD_TEXT "setting '%s' with assignment '%s' (line %d):\n",
			setting, assignment_token.string, assignment_token.linenum);
		vrPrintMatrix(setting, *xform);
		vrPrintMatrixComposition(setting, *xform);
	}

	return terminal_token;
}
#endif


/*********************************************************************/
/* vrParseMatrixTranslate(): ... */
vrTokenInfo vrParseMatrixTranslate(vrMatrix **xform, const char *setting, vrParseInfo *parse)
{
	vrTokenInfo	assignment_token;	/* type of assignment used ('=' '+=' or '*=') */
	vrTokenInfo	terminal_token;		/* the terminal token returned by vrParseDoubleListExprAssign() */
	vrVector	tmp_vector;		/* temporary 3-tuple for doing the operation */

	/********************/
	/** get the values **/

	/* tmp_vector values must be set 1.0 for the possible use of the *= assignment */
	tmp_vector.v[0] = tmp_vector.v[1] = tmp_vector.v[2] = 1.0;

	terminal_token = vrParseDoubleListExprAssign(tmp_vector.v, 3, &assignment_token, setting, parse);

	/***********************/
	/** do the assignment **/
	switch (assignment_token.token) {
	case VRTOKEN_SETEQUALS:
		*xform = vrMatrixSetTranslationAd(vrMatrixCreate(), tmp_vector.v);
		break;
	case VRTOKEN_TIMESEQUALS:
		/* TODO: get rid of leak when vrObjectCopy fixed  [2/2/00 -- not so sure anymore] */
		*xform = vrMatrixCopy(vrMatrixCreate(), *xform);
		vrMatrixPreMult(*xform, vrMatrixSetTranslationAd(vrMatrixCreate(), tmp_vector.v));
		break;
	case VRTOKEN_PLUSEQUALS:
		vrErrPrintf("%s Parse Error (line %d): "
			RED_TEXT "Invalid (or missing) assignment operator (got \"%s\") for %s.\n" NORM_TEXT,
			parse->parse_name, assignment_token.linenum, assignment_token.string, setting);

		if (!_TokenIsTerminal(terminal_token.token)) {
			terminal_token = vrParseToEOS(parse);
			vrDbgPrintfN(ALWAYS_DBGLVL, "Configuration Error (line %d): " RED_TEXT "skipping '%s'\n" NORM_TEXT,
				terminal_token.linenum, terminal_token.string);
		}
		vrTrace("vrParseMatrixTranslate()", RED_TEXT "exiting from assignment error" NORM_TEXT);
		return terminal_token;

	default:
		/* TODO: report grievous error  -- this should never happen */
		break;
	}

	if (vrDbgDo(PARSE_DBGLVL)) {
		vrPrintf("vrParseMatrixTranslate(): " BOLD_TEXT "setting '%s' with assignment '%s' (line %d):\n",
			setting, assignment_token.string, assignment_token.linenum);
		vrPrintMatrix(setting, *xform);
		vrPrintMatrixComposition(setting, *xform);
	}

	return terminal_token;
}


#ifdef EXPERIMENTAL
/*********************************************************************/
/* vrParseMatrixTranslate(): ... */
/* NOTE: I hate that I basically have to duplicate the above function to  */
/*  add a coordinates option, but I guess that's the one limitation of C, */
/*  (without delving into varargs, or trying to calculate an inverse      */
/*  matrix after calling vrParseMatrixTranslate().                        */
vrTokenInfo vrParseMatrixTranslateWithCoords(vrMatrix **xform, const char *setting, vrParseInfo *parse, vrPoint *coords_ll, vrPoint *coords_lr, vrPoint *coords_ul)
{
	vrTokenInfo	assignment_token;	/* type of assignment used ('=' '+=' or '*=') */
	vrTokenInfo	terminal_token;		/* the terminal token returned by vrParseDoubleListExprAssign() */
	vrVector	tmp_vector;		/* temporary 3-tuple for doing the operation */

	/********************/
	/** get the values **/

	/* tmp_vector values must be set 1.0 for the possible use of the *= assignment */
	tmp_vector.v[0] = tmp_vector.v[1] = tmp_vector.v[2] = 1.0;

	terminal_token = vrParseDoubleListExprAssign(tmp_vector.v, 3, &assignment_token, setting, parse);

	/***********************/
	/** do the assignment **/
	switch (assignment_token.token) {
	case VRTOKEN_SETEQUALS:
		*xform = vrMatrixSetTranslationAd(vrMatrixCreate(), tmp_vector.v);
		/* For the Set operation, make an identity window */
		vrPointSetAd(coords_ll, tmp_vector.v);
		vrPointSetAd(coords_lr, tmp_vector.v);
		vrPointSetAd(coords_ul, tmp_vector.v);
		coords_lr[VR_X] += 1.0;
		coords_ul[VR_Y] += 1.0;
		break;
	case VRTOKEN_TIMESEQUALS:
		/* TODO: get rid of leak when vrObjectCopy fixed  [2/2/00 -- not so sure anymore] */
		*xform = vrMatrixCopy(vrMatrixCreate(), *xform);
		vrMatrixPreMult(*xform, vrMatrixSetTranslationAd(vrMatrixCreate(), tmp_vector.v));
		break;
	case VRTOKEN_PLUSEQUALS:
		vrErrPrintf("%s Parse Error (line %d): "
			RED_TEXT "Invalid (or missing) assignment operator (got \"%s\") for %s.\n" NORM_TEXT,
			parse->parse_name, assignment_token.linenum, assignment_token.string, setting);

		if (!_TokenIsTerminal(terminal_token.token)) {
			terminal_token = vrParseToEOS(parse);
			vrDbgPrintfN(ALWAYS_DBGLVL, "Configuration Error (line %d): " RED_TEXT "skipping '%s'\n" NORM_TEXT,
				terminal_token.linenum, terminal_token.string);
		}
		vrTrace("vrParseMatrixTranslate()", RED_TEXT "exiting from assignment error" NORM_TEXT);
		return terminal_token;

	default:
		/* TODO: report grievous error  -- this should never happen */
		break;
	}

	if (vrDbgDo(PARSE_DBGLVL)) {
		vrPrintf("vrParseMatrixTranslate(): " BOLD_TEXT "setting '%s' with assignment '%s' (line %d):\n",
			setting, assignment_token.string, assignment_token.linenum);
		vrPrintMatrix(setting, *xform);
		vrPrintMatrixComposition(setting, *xform);
	}

	return terminal_token;
}
#endif


/*********************************************************************/
/* vrParseMatrixRotate(): ... */
vrTokenInfo vrParseMatrixRotate(vrMatrix **xform, const char *setting, vrParseInfo *parse)
{
	vrTokenInfo	assignment_token;	/* type of assignment used ('=' '+=' or '*=') */
	vrTokenInfo	terminal_token;		/* the terminal token returned by vrParseDoubleListExprAssign() */
	double		rot_values[4];		/* temporary space for the list of numbers to operate with */

	/********************/
	/** get the values **/

	/* rot_values values must be set 1.0 for the possible use of the *= assignment */
	rot_values[0] = rot_values[1] = rot_values[2] = rot_values[3] = 1.0;

	terminal_token = vrParseDoubleListExprAssign(rot_values, 4, &assignment_token, setting, parse);

	/***********************/
	/** do the assignment **/
	switch (assignment_token.token) {
	case VRTOKEN_SETEQUALS:
		*xform = vrMatrixSetRotationAd(vrMatrixCreate(), rot_values);
		break;
	case VRTOKEN_TIMESEQUALS:
		/* TODO: get rid of leak when vrObjectCopy fixed  [2/2/00 -- not so sure anymore] */
		*xform = vrMatrixCopy(vrMatrixCreate(), *xform);
		vrMatrixPostMult(*xform, vrMatrixSetRotationAd(vrMatrixCreate(), rot_values));
		break;
	case VRTOKEN_PLUSEQUALS:
		vrErrPrintf("%s Parse Error (line %d): "
			RED_TEXT "Invalid (or missing) assignment operator (got \"%s\") for %s.\n" NORM_TEXT,
			parse->parse_name, assignment_token.linenum, assignment_token.string, setting);

		if (!_TokenIsTerminal(terminal_token.token)) {
			terminal_token = vrParseToEOS(parse);
			vrDbgPrintfN(ALWAYS_DBGLVL, "Configuration Error (line %d): " RED_TEXT "skipping '%s'\n" NORM_TEXT,
				terminal_token.linenum, terminal_token.string);
		}
		vrTrace("vrParseMatrixRotate()", RED_TEXT "exiting from assignment error" NORM_TEXT);
		return terminal_token;

	default:
		/* TODO: report grievous error  -- this should never happen */
		break;
	}

	if (vrDbgDo(PARSE_DBGLVL)) {
		vrPrintf("vrParseMatrixRotate(): " BOLD_TEXT "setting '%s' with assignment '%s' (line %d):\n",
			setting, assignment_token.string, assignment_token.linenum);
		vrPrintMatrix(setting, *xform);
		vrPrintMatrixComposition(setting, *xform);
	}

	return terminal_token;
}


/*******************************************************************/
/* vrParseToEOS(): parses to the "end-of-statement".  Thus it will */
/*   return all tokens through the next statement termination token*/
/*   the token type will be that of the termination token itself,  */
/*   but the string value will contain all the intervening tokens. */
/* NOTE: in the case of encountering an open or close curly-brace  */
/*   while skipping the rest of the statement, we will still need  */
/*   to properly parse brace tokens as they do more than merely    */
/*   terminate statements, it is also used to group objects, so it */
/*   itself can't just be skipped.                                 */
vrTokenInfo vrParseToEOS(vrParseInfo *parse)
{
	vrTokenInfo	token;
	char		*string = NULL;		/* accumulate a string of all that is skipped */
	char		*tofree;		/* used for freeing old copy of "string" */
	int		paren_level = 0;	/* how deep in {} block statement are we */
	int		check_else = 0;		/* used like a counter of if-conditional levels, but shouldn't ever be more than "1", so could be a boolean */

	vrTrace("vrParseToEOS()", RED_TEXT "entering" NORM_TEXT);

	string = vrShmemStrDup("");

	do {
		token = vrParseNextToken(parse);

		/* when skipping a statement, make sure to consider {} statement blocks */
		if (token.token == VRTOKEN_OPENCURL)
			paren_level++;
		if (token.token == VRTOKEN_CLOSECURL)
			paren_level--;

		/* when skipping a conditional, need to take into account the optional "else" clause */
		if (check_else > 0) {
//vrPrintf("TEMP: checking for else token (%d) got: (%s)\n", check_else, token.string);
			/* if there's a possibility for an "else" clause, and we get an "else", then skip that statement too */
			if (token.token == VRTOKEN_ELSE) {
				/* append token to "string" */
				tofree = string;
				string = vrShmemStrCat(string, " ");		/* add whitespace */
				string = vrShmemStrCat(string, token.string);
				vrShmemFree(tofree);				/* free the previous copy */

				token = vrParseToEOS(parse);
//vrPrintf("TEMP: found else EOS (%d) token is: (%s)\n", check_else, token.string);
			} else {
				/* return any non "else" tokens to the parser */
				vrParseReturnPreviousToken(token, parse);

				/* now indicate that we are indeed at EOS */
				token.token = VRTOKEN_EOS;
				token.string = "<non-else>";
			}
			check_else--;		/* once "else" is checked for, we don't need to check again */
		}
		if (token.token == VRTOKEN_IF || token.token == VRTOKEN_IFEXEC) {
			/* append token to "string" */
			tofree = string;
			string = vrShmemStrCat(string, " ");		/* add whitespace */
			string = vrShmemStrCat(string, token.string);
			vrShmemFree(tofree);				/* free the previous copy */

			token = vrParseToEOS(parse);
			check_else++;
//vrPrintf("TEMP: check_else = %d; _TokenIsTerminal(%s) = %d\n", check_else, token.string, _TokenIsTerminal(token.token)); // delete me
		}

		/* make a string concatenating all the parsed tokens */
		tofree = string;
		string = vrShmemStrCat(string, " ");		/* add whitespace */
		string = vrShmemStrCat(string, token.string);
		vrShmemFree(tofree);				/* free the previous copy */
//
//vrPrintf("TEMP: paren_level = %d; _TokenIsTerminal(%s) = %d\n", paren_level, token.string, _TokenIsTerminal(token.token)); // delete me
	} while ((paren_level > 0) || (check_else > 0) || (!_TokenIsTerminal(token.token)));

	token.string = string;

//vrPrintf("vrParseToEOS(): returning '%s'\n", token.string); // delete me
	vrTrace("vrParseToEOS()", RED_TEXT "exiting" NORM_TEXT);
//vrPrintf("TEMP: returning: paren_level = %d; _TokenIsTerminal(%s) = %d\n", paren_level, token.string, _TokenIsTerminal(token.token)); // delete me
	return token;
}



/* =================================================================== */
/* =================================================================== */
/* =================================================================== */
/*                     Argument Parsing Functions                      */
/* This section has a different style of parsing that is used for      */
/*   handling arguments passed to input devices, processes, windows,   */
/*   etc. in the config file.  These are typically much simpler forms  */
/*   that just include an argument, some form of assignment, and a     */
/*   choice of assignees.  Thus these functions are generally called   */
/*   from code in the "vr_input.<type>.c" and "vr_visren.<type>.c"     */
/*   source files -- often inside a function called _<type>ParseArgs().*/
/*   (Though it would seem that "vr_visren.glx.c" does NOT use it --   */
/*   probably it was written before these functions existed.)          */
/* There are also some limited cases where these functions are used    */
/*   to augment the parsing routines in "vr_telnet.c".                 */
/* NOTE: individual arguments inside the "args" string are separated   */
/*   by semicolons (';'), which allows the value strings to contain    */
/*   commas (',').                                                     */
/* =================================================================== */

/* TODO: [6/3/03] specifically for vrArgParseString, trailing whitespace should be trimmed */
/* TODO: for all the vrArgParse...() functions, need to do the following: */
/*	- ignore case when searching for the arg option                   */
/*	- make sure the found arg immediately follows a separator (or     */
/*		the beginning of the line.)                               */
/*	- print a warning or use the last occurrence if more than one     */
/*		instance of an argument exists.                           */

/* NOTE: at one point both commas and semicolons were considered    */
/*   acceptable separators.  But commas were seldom used in the     */
/*   config file, yet were sometimes erroneously used as separating,*/
/*   within-argument tokens.  So, I decided to adopt the latter.    */
#undef COMMA_SEPARATOR	/* TODO: delete all the comma can be separator stuff */  /* delete me (and all associated code) */



/*********************************************************************/
/* vrArgParseString(): takes a string argument (args), along with a  */
/*   function-argument for the configuration-argument that is to be  */
/*   changed (argname), and then a call-by-reference string-pointer  */
/*   (arglvalue) into which the sought value is returned.  Shared    */
/*   Memory is allocated for the returned string.                    */
/* Returns boolean of whether the argument was set or not (otherwise)*/
/*   arglvalue will be untouched.                                    */
/* Format: "<argname>" "=" <argvalue> [";"]                          */
/* TODO: BUG-fix-needed: only the first occurrence of "argname" is   */
/*   sought in the "args" string, even when it might be part of a    */
/*   different setting!                                              */
/* TODO: BUG-fix-needed: should probably ignore case on the strcmp().*/
int vrArgParseString(char *args, char *argname, char **arglvalue)
{
	char	*arg_str;	/* pointer to the argument string */
#ifdef UNPICKY_COMPILER
	char	*end_str;	/* pointer to the end of the args string */
#endif
	char	*begin_str;	/* pointer to the beginning of the argument value */
	char	*equals_str;	/* pointer to the "next" '=' after the argument */
	char	*sep_str;	/* pointer to the "next" separator after the argument */
	char	sep_char;	/* temp storage for the separator character */
	int	arg_set = 0;	/* boolean of whether the argument was set */

	/* Handle the case of a NULL argument string */
	if (args == NULL)
		return (0);

#ifdef UNPICKY_COMPILER
	/* search for the end of the argument string */
	end_str = strchr(args, '\0');
#endif

	/*******************************************************/
	/** Argument format: "<argname>" "=" <argvalue> [";"] **/
	/*******************************************************/
	if ((arg_str = strstr(args, argname)) != NULL) {	/* TODO: need to make this ignore case */
		/* find the next '=' character */
		equals_str = strchr(arg_str, '=');

		/* find the beginning and ending (separator) boundaries */
		begin_str = equals_str + strspn(equals_str, "= \t");
#ifdef COMMA_SEPARATOR /* also removed the space which didn't seem correct */ /* deprecated */
		sep_str = begin_str + strcspn(begin_str, " \t,;");
#else
		sep_str = begin_str + strcspn(begin_str, "\t;");
#endif

		/* Error if no '=', or no '=' prior to the next separator */
		if ((begin_str == NULL) || (begin_str > sep_str)) {
			vrErrPrintf("vrArgParseString(): "
				RED_TEXT "Missing assignment character after '%s' argument.\n" NORM_TEXT, argname);
		} else {
			/* temporarily terminate the string at the next separator */
			sep_char = sep_str[0];
			sep_str[0] = '\0';

			if (strlen(begin_str) == 0) {
				vrErrPrintf("vrArgParseString(): "
					RED_TEXT "No string following '%s' argument.\n" NORM_TEXT, argname);
			} else {
				*arglvalue = vrShmemStrDup(begin_str);
				arg_set = 1;
				vrDbgPrintfN(PARSE_DBGLVL, "vrArgParseString(): parsed %s = \"%s\"\n", argname, *arglvalue);
			}

			/* restore the string's separator for other arguments */
			sep_str[0] = sep_char;
		}
	}

	return (arg_set);
}


/*********************************************************************/
/* vrArgParseFloat(): takes a string argument (args), along with a   */
/*   function-argument for the configuration-argument that is to be  */
/*   changed (argname), and then a call-by-reference float-pointer   */
/*   (arglvalue) into which the sought value is returned.            */
/* Returns boolean of whether the argument was set or not (otherwise)*/
/*   arglvalue will be untouched.                                    */
/* Format: "<argname>" "=" <argvalue> [";"]                          */
/* TODO: BUG-fix-needed: only the first occurrence of "argname" is   */
/*   sought in the "args" string, even when it might be part of a    */
/*   different setting!                                              */
/* TODO: BUG-fix-needed: should probably ignore case on the strcmp().*/
int vrArgParseFloat(char *args, char *argname, float *arglvalue)
{
	char	*arg_str;	/* pointer to the argument string */
	char	*end_str;	/* pointer to the end of the args string */
	char	*equals_str;	/* pointer to the "next" '=' after the argument */
#ifdef COMMA_SEPARATOR /* deprecated */
	char	*comma_str;	/* pointer to the "next" ',' after the argument */
	char	*semi_str;	/* pointer to the "next" ';' after the argument */
#endif
	char	*sep_str;	/* pointer to the "next" separator after the argument */
	int	arg_set = 0;	/* boolean of whether the argument was set */

	/* Handle the case of a NULL argument string */
	if (args == NULL)
		return (0);

	/* search for the end of the argument string */
	end_str = strchr(args, '\0');

	/*******************************************************/
	/** Argument format: "<argname>" "=" <argvalue> [";"] **/
	/*******************************************************/
	if ((arg_str = strstr(args, argname)) != NULL) {	/* TODO: need to make this ignore case */
		/* find the next '=' character */
		equals_str = strchr(arg_str, '=');

		/* find the next separator */
#ifdef COMMA_SEPARATOR /* deprecated */
		comma_str = strchr(arg_str, ',');
		semi_str = strchr(arg_str, ';');
		if (semi_str == NULL) semi_str = end_str;
		if (comma_str == NULL) comma_str = end_str;
		sep_str = (comma_str < semi_str ? comma_str : semi_str);
#else
		sep_str = strchr(arg_str, ';');
		if (sep_str == NULL) sep_str = end_str;
#endif

		/* Error if no '=', or no '=' prior to the next separator */
		if ((equals_str == NULL) || (equals_str > sep_str)) {
			vrErrPrintf("vrArgParseFloat(): "
				RED_TEXT "Missing assignment character after '%s' argument.\n" NORM_TEXT, argname);
		} else {
			*arglvalue = atof(equals_str+1);
			arg_set = 1;
			vrDbgPrintfN(PARSE_DBGLVL, "vrArgParseFloat(): parsed %s = %f\n", argname, *arglvalue);
		}
	}

	return (arg_set);
}


/**********************************************************************/
/* vrArgParseFloatList(): takes a string argument (args), along with  */
/*   a function-argument for the configuration-argument that is to be */
/*   changed (argname), and then a call-by-reference float-pointer    */
/*   (arglvalue) into which the sought values are returned.  And then */
/*   an integer (maxlen) indicates the amount floats for which memory */
/*   is allocated.                                                    */
/* Returns boolean of whether the argument was set or not (otherwise) */
/*   arglvalue will be untouched.                                     */
/* Format: "<argname>" "=" number [ "," number ]* [";"]               */
/* TODO: BUG-fix-needed: only the first occurrence of "argname" is    */
/*   sought in the "args" string, even when it might be part of a     */
/*   different setting!                                               */
/* TODO: BUG-fix-needed: should probably ignore case on the strcmp(). */
int vrArgParseFloatList(char *args, char *argname, float *arglvalue, int maxlen)
{
static	char	*whitespace = " \t\r\b\n";
static	char	*whiteandsep = " \t\r\b\n,;";
#ifndef COMMA_SEPARATOR /* deprecated */
static	char	*whiteandcomma = " \t\r\b\n,";
#endif
	char	*arg_str;	/* pointer to the argument string */
	char	*end_str;	/* pointer to the end of the args string */
	char	*equals_str;	/* pointer to the "next" '=' after the argument */
#ifdef COMMA_SEPARATOR /* deprecated */
	char	*comma_str;	/* pointer to the "next" ',' after the argument */
	char	*semi_str;	/* pointer to the "next" ';' after the argument */
#endif
	char	*sep_str;	/* pointer to the "next" separator after the argument */
	char	*num_str;	/* pointer to the "next" number in the argument */
	int	arg_set = 0;	/* boolean of whether the argument was set */
	int	count;		/* for putting the next number in the proper location */

	/* Handle the case of a NULL argument string */
	if (args == NULL)
		return (0);

	/* search for the end of the argument string */
	end_str = strchr(args, '\0');

	/*******************************************************************/
	/** Argument format: "<argname>" "=" number [ "," number ]* [";"] **/
	/*******************************************************************/
	if ((arg_str = strstr(args, argname)) != NULL) {	/* TODO: need to make this ignore case */
		/* find the next '=' character */
		equals_str = strchr(arg_str, '=');

		/* find the next separator */
#ifdef COMMA_SEPARATOR /* deprecated */
		comma_str = strchr(arg_str, ',');
		semi_str = strchr(arg_str, ';');
		if (semi_str == NULL) semi_str = end_str;
		if (comma_str == NULL) comma_str = end_str;
		sep_str = (comma_str < semi_str ? comma_str : semi_str);
#else
		sep_str = strchr(arg_str, ';');
		if (sep_str == NULL) sep_str = end_str;
#endif

		/* Error if no '=', or no '=' prior to the next separator */
		if ((equals_str == NULL) || (equals_str > sep_str)) {
			vrErrPrintf("vrArgParseFloat(): "
				RED_TEXT "Missing assignment character after '%s' argument.\n" NORM_TEXT, argname);
		} else {
			vrDbgPrintfN(PARSE_DBGLVL, "vrArgParseFloatList(): parsed %s = ", argname);
			num_str = equals_str+1;					/* skip the equals */
			num_str += strspn(num_str, whitespace);			/* skip whitespace */
			count = 0;
			do {
				arglvalue[count] = atof(num_str);

				/* these two lines basically skip the number and trailing whitespace */
				num_str += strcspn(num_str, whiteandsep);	/* skip non-whitespace (and separators) */
				num_str += strspn(num_str, whiteandcomma);	/* skip whitespace (and commas) */
				vrDbgPrintfN(PARSE_DBGLVL, "%f (next = '%s') ", arglvalue[count], num_str);

				count++;
			} while (count < maxlen && strcspn(num_str, ";") > 0);
			vrDbgPrintfN(PARSE_DBGLVL, "\n");
			arg_set = 1;
		}
	}

	return (arg_set);
}


/*********************************************************************/
/* vrArgParseInteger(): takes a string argument (args), along with a */
/*   function-argument for the configuration-argument that is to be  */
/*   changed (argname), and then a call-by-reference integer-pointer */
/*   (arglvalue) into which the sought value is returned.            */
/* Returns boolean of whether the argument was set or not (otherwise)*/
/*   arglvalue will be untouched.                                    */
/* Format: "<argname>" "=" <argvalue> [";"]                          */
/* TODO: BUG-fix-needed: only the first occurrence of "argname" is   */
/*   sought in the "args" string, even when it might be part of a    */
/*   different setting!                                              */
/* TODO: BUG-fix-needed: should probably ignore case on the strcmp().*/
int vrArgParseInteger(char *args, char *argname, int *arglvalue)
{
	char	*arg_str;	/* pointer to the argument string */
	char	*end_str;	/* pointer to the end of the args string */
	char	*equals_str;	/* pointer to the "next" '=' after the argument */
#ifdef COMMA_SEPARATOR /* deprecated */
	char	*comma_str;	/* pointer to the "next" ',' after the argument */
	char	*semi_str;	/* pointer to the "next" ';' after the argument */
#endif
	char	*sep_str;	/* pointer to the "next" separator after the argument */
	int	arg_set = 0;	/* boolean of whether the argument was set (return value) */

	/* Handle the case of a NULL argument string */
	if (args == NULL)
		return (0);

	/* search for the end of the argument string */
	end_str = strchr(args, '\0');

	/*******************************************************/
	/** Argument format: "<argname>" "=" <argvalue> [";"] **/
	/*******************************************************/
	if ((arg_str = strstr(args, argname)) != NULL) {	/* TODO: need to make this ignore case */
		/* find the next '=' character */
		equals_str = strchr(arg_str, '=');

		/* find the next separator */
#ifdef COMMA_SEPARATOR /* deprecated */
		comma_str = strchr(arg_str, ',');
		semi_str = strchr(arg_str, ';');
		if (semi_str == NULL) semi_str = end_str;
		if (comma_str == NULL) comma_str = end_str;
		sep_str = (comma_str < semi_str ? comma_str : semi_str);
#else
		sep_str = strchr(arg_str, ';');
		if (sep_str == NULL) sep_str = end_str;
#endif

		/* Error if no '=', or no '=' prior to the next separator */
		if ((equals_str == NULL) || (equals_str > sep_str)) {
			vrErrPrintf("vrArgParseInteger(): "
				RED_TEXT "Missing assignment character after '%s' argument.\n" NORM_TEXT, argname);
		} else {
			*arglvalue = vrAtoI(equals_str+1);
			arg_set = 1;
			vrDbgPrintfN(PARSE_DBGLVL, "vrArgParseInteger(): parsed %s = %d\n", argname, *arglvalue);
		}
	}

	return (arg_set);
}


/*********************************************************************/
/* vrArgParseBool(): takes a string argument (args), along with a    */
/*   function-argument for the configuration-argument that is to be  */
/*   changed (argname), and then a call-by-reference integer-pointer */
/*   (arglvalue) into which the sought value is returned.            */
/* NOTE: unlike the Integer parser, several strings (words) can be   */
/*   used to indicate the Boolean 0 or 1 value.                      */
/* Returns boolean of whether the argument was set or not (otherwise)*/
/*   arglvalue will be untouched.                                    */
/* Format: "<argname>" "=" { "on"|"off"|"yes"|"no"|"true"|"false"|number } [";"] */
/* TODO: BUG-fix-needed: only the first occurrence of "argname" is   */
/*   sought in the "args" string, even when it might be part of a    */
/*   different setting!                                              */
/* TODO: BUG-fix-needed: should probably ignore case on the strcmp().*/
int vrArgParseBool(char *args, char *argname, int *arglvalue)
{
	char	*arg_str;	/* pointer to the argument string */
	char	*begin_str;	/* pointer to the beginning of the argument value */
	char	*equals_str;	/* pointer to the "next" '=' after the argument */
	char	*sep_str;	/* pointer to the "next" separator after the argument */
	char	sep_char;	/* temp storage for the separator character */
	int	arg_set = 0;	/* boolean of whether the argument was set */

	/* Handle the case of a NULL argument string */
	if (args == NULL)
		return (0);

	/********************************************************************************************/
	/** Argument format: "<argname>" "=" { "on"|"off"|"yes"|"no"|"true"|"false"|number } [";"] **/
	/********************************************************************************************/
	if ((arg_str = strstr(args, argname)) != NULL) {	/* TODO: need to make this ignore case */
		/* find the next '=' character */
		equals_str = strchr(arg_str, '=');

		/* find the beginning and ending (separator) boundaries */
		begin_str = equals_str + strspn(equals_str, "= \t");
#ifdef COMMA_SEPARATOR /* deprecated */
		sep_str = begin_str + strcspn(begin_str, " \t,;");
#else
		sep_str = begin_str + strcspn(begin_str, " \t;");
#endif

		/* Error if no '=', or no '=' prior to the next separator */
		if ((begin_str == NULL) || (begin_str > sep_str)) {
			vrErrPrintf("vrArgParseBool(): "
				RED_TEXT "Missing assignment character after '%s' argument.\n" NORM_TEXT, argname);
		} else {
			/* temporarily terminate the string at the next separator */
			sep_char = sep_str[0];
			sep_str[0] = '\0';

			if (strlen(begin_str) == 0) {
				vrErrPrintf("vrArgParseBool(): "
					RED_TEXT "No value following '%s' argument.\n" NORM_TEXT, argname);
			} else {
				if (!strcasecmp(begin_str, "on")) {
					*arglvalue = 1;
				} else if (!strcasecmp(begin_str, "yes")) {
					*arglvalue = 1;
				} else if (!strcasecmp(begin_str, "true")) {
					*arglvalue = 1;
				} else if (!strcasecmp(begin_str, "off")) {
					*arglvalue= 0;
				} else if (!strcasecmp(begin_str, "no")) {
					*arglvalue= 0;
				} else if (!strcasecmp(begin_str, "false")) {
					*arglvalue= 0;
				} else {
					*arglvalue = !(!vrAtoI(begin_str));
				}
				arg_set = 1;
				vrDbgPrintfN(PARSE_DBGLVL, "vrArgParseBool(): parsed %s = %d\n", argname, *arglvalue);
			}

			/* restore the string's separator for other arguments */
			sep_str[0] = sep_char;
		}
	}

	return (arg_set);
}


/*********************************************************************/
/* vrArgParseChoiceInteger(): takes a string argument (args), and a  */
/*   function-argument for the configuration-argument that is to be  */
/*   changed (argname), and then a call-by-reference integer-pointer */
/*   (arglvalue) into which the sought value is returned.  Similar   */
/*   to the Boolean argument-parser, the Choice-Integer parser takes */
/*   string inputs and converts them to a numerical (integer) value. */
/*   Unlike the Boolean parser, the specific choices (choices), and  */
/*   the numbers they are mapped to (values) are passed to this      */
/*   function as arrays of strings (char*) and integers respectively.*/
/* NOTE: unlike the Integer parser, several strings (words) can be   */
/*   used to indicate the Boolean 0 or 1 value.                      */
/* Returns boolean of whether the argument was set or not (otherwise)*/
/*   arglvalue will be untouched.                                    */
/* Format: "<argname>" "=" { "<choice1>" | "<choice2>" | ... } [";"] */
/* TODO: BUG-fix-needed: only the first occurrence of "argname" is   */
/*   sought in the "args" string, even when it might be part of a    */
/*   different setting!                                              */
/* TODO: BUG-fix-needed: should probably ignore case on the strcmp().*/
int vrArgParseChoiceInteger(char *args, char *argname, int *arglvalue, char *choices[], int values[])
{
	char	*arg_str;	/* pointer to the argument string */
	char	*begin_str;	/* pointer to the beginning of the argument value */
	char	*equals_str;	/* pointer to the "next" '=' after the argument */
	char	*sep_str;	/* pointer to the "next" separator after the argument */
	char	sep_char;	/* temp storage for the separator character */
	int	arg_set = 0;	/* boolean of whether the argument was set */

	char	**choice;	/* current choice to test for */
	int	*value;		/* associated value of current choice */

	/* Handle the case of a NULL argument string */
	if (args == NULL)
		return (0);

	/********************************************************************************/
	/** Argument format: "<argname>" "=" { "<choice1>" | "<choice2>" | ... } [";"] **/
	/********************************************************************************/
	if ((arg_str = strstr(args, argname)) != NULL) {	/* TODO: need to make this ignore case */
		/* find the next '=' character */
		equals_str = strchr(arg_str, '=');

		/* find the beginning and ending (separator) boundaries */
		begin_str = equals_str + strspn(equals_str, "= \t");
#ifdef COMMA_SEPARATOR /* deprecated */
		sep_str = begin_str + strcspn(begin_str, " \t,;");
#else
		sep_str = begin_str + strcspn(begin_str, " \t;");
#endif

		/* Error if no '=', or no '=' prior to the next separator */
		if ((begin_str == NULL) || (begin_str > sep_str)) {
			vrErrPrintf("vrArgParseChoiceInteger(): "
				RED_TEXT "Missing assignment character after '%s' argument.\n" NORM_TEXT, argname);
		} else {
			/* temporarily terminate the string at the next separator */
			sep_char = sep_str[0];
			sep_str[0] = '\0';

			if (strlen(begin_str) == 0) {
				vrErrPrintf("vrArgParseChoiceInteger(): "
					RED_TEXT "No value following '%s' argument.\n" NORM_TEXT, argname);
			} else {
				for (choice = choices, value = values; *choice != NULL; choice++, value++) {
//vrPrintf("checking if '%s' matches with '%s' giving a value of %d\n", begin_str, *choice, *value);
					if (!strcasecmp(begin_str, *choice)) {
						*arglvalue = *value;
						arg_set = 1;
						vrDbgPrintfN(PARSE_DBGLVL, "ArgParseChoiceInteger(): parsed %s = %d ('%s')\n", argname, *arglvalue, *choice);
						break;
					}
				}
			}

			/* restore the string's separator for other arguments */
			sep_str[0] = sep_char;
		}
	}

	return (arg_set);
}



/* ==================================================================== */
/* ==================================================================== */
/* ==================================================================== */
/* This section has parsing routines that are specialized to particular */
/*   tasks that may or may not be common to several places in the code. */
/* ==================================================================== */


/************************************************************/
/* vrParseInputDescription(): ... */
/*  For parsing things of the form:                           */
/*       2switch(0) -- example of a static 2switch            */
/*       valuator(0.5) -- example of a static valuator        */
/*       2switch(button[1]) -- for a Magellan or similar unit */
vrInputDesc *vrParseInputDescription(char *description)
{
	vrInputDesc	*indesc;
	char		*desc;
	char		*open_paren;
	char		*close_paren;


	indesc = (vrInputDesc *)vrShmemAlloc(sizeof(*indesc));
	indesc->type = VRINPUT_UNKNOWN;
	indesc->args = vrShmemStrDup("");

	if (description == NULL) {
		vrErrPrintf(RED_TEXT "vrParseInputDescription(): NULL description\n" NORM_TEXT);
		return indesc;
	}
	vrDbgPrintfN(INPUT_DBGLVL, "vrParseInputDescription(): parsing input description:'%s'\n", description);

	/* copy and parse the description */
	desc = vrShmemStrDup(description);

	/* find the first '(' character */
	open_paren = strchr(desc, '(');
	if (open_paren == NULL) {
		vrErrPrintf(RED_TEXT "vrParseInputDescription(): Invalid description format '%s'\n" NORM_TEXT, desc);
		return indesc;
	}

	*open_paren = '\0';
	if (!strcasecmp(desc, "2switch")) {
		indesc->type = VRINPUT_BINARY;
	} else if (!strcasecmp(desc, "switch2")) {
		indesc->type = VRINPUT_BINARY;
	} else if (!strcasecmp(desc, "Nswitch")) {
		indesc->type = VRINPUT_NWAY;
	} else if (!strcasecmp(desc, "switchN")) {
		indesc->type = VRINPUT_NWAY;
	} else if (!strcasecmp(desc, "valuator")) {
		indesc->type = VRINPUT_VALUATOR;
	} else if (!strcasecmp(desc, "6sensor")) {
		indesc->type = VRINPUT_6SENSOR;
	} else if (!strcasecmp(desc, "sensor6")) {
		indesc->type = VRINPUT_6SENSOR;
	} else if (!strcasecmp(desc, "Nsensor")) {
		indesc->type = VRINPUT_NSENSOR;
	} else if (!strcasecmp(desc, "sensorN")) {
		indesc->type = VRINPUT_NSENSOR;
	} else if (!strcasecmp(desc, "control")) {
		indesc->type = VRINPUT_CONTROL;
	} else {
		vrErrPrintf(RED_TEXT "vrParseInputDescription: unknown input type '%s'\n", desc);
		indesc->type = VRINPUT_UNKNOWN;
	}
	vrDbgPrintfN(INPUT_DBGLVL, RED_TEXT "vrParseInputDescription: got input type '%s'\n", desc);

	/* now find the characters up to the next ')' character */
	close_paren = strchr(open_paren+1, ')');
	*close_paren = '\0';
	indesc->args = vrShmemStrDup(open_paren+1);

	return indesc;
}


/**************************************************************/
/* vrParseInputDTI(): is used to parse things of the          */
/*   form: <device>:type[instance]                            */
/*   In particular this was designed for Xevents, but may     */
/*   work well in other circumstances.                        */
/* Arguments:                                                 */
/*   input -- the input associated with the DTI string.       */
/*     [This is provided solely for reporting parsing errors. */
/*   string -- the char-string with the DTI description.      */
/* Returns:                                                   */
/*   Pointer to a DTI structure with the three elements in    */
/*   their own char-string fields.                            */
/*                                                            */
/* NOTE: [1/31/00] We have a choice of whether a string with  */
/*   neither a ':' or '[' & ']' characters is interpreted as  */
/*   either a type, with empty device & instance values, or   */
/*   as an instance with empty device & type values -- or     */
/*   even as device without the other two, but since many     */
/*   input declarations don't even use the device field,      */
/*   that doesn't make a lot of sense.  Previously, it had    */
/*   been the type that was the default, but today I'm going  */
/*   to try using the instance field, allowing "" to be a     */
/*   possible type.                                           */
vrInputDTI *vrParseInputDTI(vrInput *input, char *string)
{
	vrInputDTI	*dti;		/* pointer to a new DTI structure */
	char		*tmp_string;	/* duplicate copy of string used for manipulation */
	char		*colon;		/* locate of the first colon (':') in the string */
	char		*open_square;	/* locate of the first open-square ('[') in the string */
	char		*close_square;	/* locate of the first close-square (']') in the string */

	dti = (vrInputDTI *)vrShmemAlloc(sizeof(*dti));
	dti->device = vrShmemStrDup("");
	dti->type = vrShmemStrDup("");
	dti->instance = vrShmemStrDup("");
	tmp_string = vrShmemStrDup(string);


	/*******************/
	/*** dti->device ***/

	/* find the first ':' character */
	colon = strchr(tmp_string, ':');
	if (colon != NULL) {
		/* for some input parsing, the device field is not used */
		*colon = '\0';
		dti->device = vrShmemStrDup(tmp_string);
		*colon = ':';
	} else {
		colon = tmp_string-1;
	}

	/*****************/
	/*** dti->type ***/

	/* find the '[' character */
	open_square = strchr(tmp_string, '[');
	if (open_square == NULL ) {
#if 0 /* See 1/31/00 NOTE above */
		/* for some input parsing, the instance field is not used (when there's no '[' */
		dti->type = vrShmemStrDup(colon+1);
#else
		/* for some input parsing, the type field is not used (when there's no '[') */
		dti->instance = vrShmemStrDup(colon+1);
#endif
		return dti;
	}

	/* If the open-square happens before the colon, then the colon is really */
	/*   part of the instance name -- so make adjustments (added 08/19/13).  */
	if (open_square < colon) {
#if 0
		vrErrPrintf(RED_TEXT "vrParseInputDTI(): Invalid description format '%s'\n" NORM_TEXT, string);
		return dti;
#else
		/* the colon therefore is part of the instance name */
		dti->device = vrShmemStrDup("");	/* which means there is no sub-device */
#endif
		*open_square = '\0';
		dti->type = vrShmemStrDup(tmp_string);
		*open_square = '[';
	} else {
		*open_square = '\0';
		dti->type = vrShmemStrDup(colon+1);
		*open_square = '[';
	}

	/*********************/
	/*** dti->instance ***/

	/* find the ']' character */
	close_square = strchr(tmp_string, ']');
	if (close_square == NULL || close_square < open_square) {
		vrErrPrintf(RED_TEXT "vrParseInputDTI(): Invalid description format '%s'\n" NORM_TEXT, string);
		return dti;
	}

	if (strlen(close_square) > 1) {
		/* if there's more than just ']' left, then that's a configuration error */
		vrDbgPrintfN(CONFIG_ERROR_DBGLVL, "Configuration Error (line %d): " RED_TEXT "Ill-formed input description: '%s' for input '%s' (nothing should follow the brackets).\n" NORM_TEXT,
			input->line_created, string, input->name);
	}

	*close_square = '\0';		/* erase the ']' */
	dti->instance = vrShmemStrDup(open_square+1);	/* the 'instance' is between the square-brackets */
	*close_square = '[';

	return dti;
}


/* =================================================================== */
/* =================================================================== */
/* =================================================================== */


#ifdef TEST_APP

/* TODO: NOTE: the parse testing application presently (04/16/21) doesn't compile. */

/*********************************************************************/
static int vrParseFile(char *path, char *file)
{
	FILE		*fp_configfile;
	char		*fullfilename;
	vrTokenInfo	token;
	vrParseInfo	config_parse;

	fullfilename = (char *)vrShmemAlloc(strlen(path)+strlen(file)+3);
	sprintf(fullfilename, "%s/%s", path, file);

	fp_configfile = fopen(fullfilename, "r");
	if (fp_configfile == NULL) {
		fprintf(stderr, "vrReadConfigFile(): " RED_TEXT "could not open file '%s'\n" NORM_TEXT, fullfilename);
		return 0;
	}
	printf("FreeVR: reading configuration file '%s'.\n", fullfilename);
	config_parse.linenum = 0;
	config_parse.parse_name = vrShmemStrDup("Config");
	config_parse.filename = vrShmemStrDup(fullfilename);
	config_parse.fp = fp_configfile;
	memset(config_parse.input, '\0', 256);
	config_parse.next_input = "";

/* NOTE: _ConfigTSMap is local to vr_config.c -- perhaps #include <vr_config.c> will work here. */
	config_parse.tsmap_list = _ConfigTSMap;
	config_parse.nummaps = (sizeof (_ConfigTSMap) / sizeof (_ConfigTSMap[0]));
	do {
		token = vrParseNextToken(config_parse);
		printf("%s Parse: (line %d) got token **** (%3d) '%s' ****\n", config_parse->parse_name, config_parse.linenum, token.token, token.string);
	} while (token.token != VRTOKEN_EOF);

	return 1;
}


/*********************************************************************/
int main (int argc, char *argv[])
{
	vrParseFile(".", "freevr.rc");
}
#endif

