/* ======================================================================
 *
 *  CCCCC          vr_input.c
 * CC   CC         Author(s): Ed Peters, Bill Sherman, John Stone
 * CC              Created: June 3, 1998
 * CC   CC         Last Modified: February 9, 2024 (minor update 02/12/2024)
 *  CCCCC
 *
 * Code file for FreeVR input processes.  See vr_input.h for a
 * description of the relevant data structures.
 *
 * Copyright 2024, Bill Sherman, All rights reserved.
 * With the intent to provide an open-source license to be named later.
 * ====================================================================== */
/*************************************************************************
Handling FreeVR input processes
 
Usage:
	The primary functions of vr_input.c is to initialize, run, and then
	  terminate a FreeVR input-process.  These functions are called by
	  _RunChildProcess() (in vr_procs.c):

	void vrInputInitProc(vrProcessInfo *myproc_info)
		- initializes the data for a new input process
	void vrInputTermProc(vrProcessInfo *myproc_info)
		- cleans/clears the data for an ending input process
	void vrInputOneFrame(vrProcessInfo *myproc_info)
		- handles "one frame" of input, polling each input device
	void vrInputMainLoop(vrProcessInfo *myproc_info)
		- does process synchronization and then loops over the frames
		NOTE: Deprecated, as the frame looping is now handled in _RunChildProcess()

Structures to be aware of:
	vrGenericInput -- the top-level polymorphic type for all the
		input objects (containers?)
		Sub-types: vr2switch, vrNswitch, vrValuator, vr6sensor, vrNsensor, vrControl
	vrInput -- an input object
		- stores an input container(?)
		- child of an input device
		- part of a list of configured inputs of the system
	vrInputDevice -- a FreeVR object for an input device
		- points to lists of input objects for each type
		- contains input-device-specific information
	vrInputInfo -- a top-level FreeVR structure
		- points to all input devices
			(including controls, but not self-controls)
			NOTE: presently there are no non-self-controls
		- has the input object mapping
			NOTE: input-mapping has arrays of pointers to all the
			  active input objects (for each

The Input-process initialization process:
	... (see "inputdesc")

**************************************************************************/

#define	COREDUMP_IS_GOOD	/* core dumps are good when debugging */


#include <string.h>	/* needed for string functions */
#include <stdlib.h>	/* needed for atof(), and malloc, etc */
#include "vr_input.h"
#include "vr_callback.h"
#include "vr_config.h"
#include "vr_debug.h"
#include "vr_parse.h"
#include "vr_utils.h"
#include <signal.h>	/* for signal type SIGUSER2 */

#if !defined(__hpux)
#  include <dlfcn.h>	/* for DSO access */
#endif

#define VRINPUT_ONLY	/* needed by vr_input.opts.h */
#include "vr_input.opts.h"


/* static (file-scope) globals for the (TODO: each?) input process */
static	vrProcessInfo	*proc_info = NULL;	/* NOTE: currently only used by the signal handler */
static	int		num_devices = 0;	/* TODO: probably should be part of "proc_info" rather than global */
static	vrInputDevice	**devices = NULL;	/* TODO: probably should be part of "proc_info" rather than global */


/************************************************************************/
/* function declarations of local-only functions called before defined. */
vrInputDevice	*_MakeDummyInputDevice();
vrInput		*_MakeDummyInputObject(char *function, int line, char *description);



/*****************************************************************/
/* vrInputSignalHandler(): on a SIGINT signal, this process will */
/*   close down all devices it's responsible for and exit.       */
static void vrInputSignalHandler(int which)
{
	switch (which) {
	case SIGUSR2:
		vrMsgPrintf("vrInputSignalHandler(): Hey, input process got a USR2 signal\n");
		vrCallbackInvoke(vrContext->callbacks->HandleUSR2);
		break;

	default:
		/* for all other signals, assume death */
		vrMsgPrintf("vrInputSignalHandler(): (input proc %d) " RED_TEXT "dying with %s (%d).\n" NORM_TEXT, getpid(), vrSigName(which), which);

		vrMsgPrintf("vrInputSignalHandler(): Time of death = %f, number of frames = %d\n",
			proc_info->frame_wtime, proc_info->frame_count);

		vrMsgPrintf("vrInputSignalHandler(): Shared memory usage:%7ld (%ld freed)\n",
			vrShmemUsage(), vrShmemFreed());

		proc_info->req_proc_term = 1;
		vrFprintProcessInfo(stdout, proc_info, verbose);
		vrContext->status = VRSTATUS_PARTIAL_RUNNING;

#ifdef COREDUMP_IS_GOOD
		vrFprintf(stderr, "vrInputSignalHandler(): YO aborting in vr_input.c\n");
		abort();
#endif

#if 1
		vrInputTermProc(proc_info);	/* TODO: not sure we want to do this */
#endif

		/* PAUSE to allow debugger to attach to the process. */
		vrMsgPrintf("vrInputSignalHandler(): Process pausing, use 'gdb -p %d' to debug.\n", getpid());
		pause();
	}
}


/*************************************************************/
/* vrInputObjectClear(): clears all the values from a FreeVR */
/*   input container object.                                 */
void vrInputObjectClear(vrInput *object)
{
	object->object_type = VROBJECT_INPUT;

	object->map_refcount = 0;
	object->desc_str = NULL;
	object->desc = NULL;	
#if 1
	object->r2e_xform = NULL;
	object->ipos_xform = NULL;
#endif

	object->desc_ui[0] = '\0';
	object->container.generic = NULL;
}


/**********************************************************************/
/* vrInputObjectCopy(): copy the input information of a FreeVR        */
/*   input container object.  Note that it leaves untouched the       */
/*   general FreeVR object information, thus creating a "distinct"    */
/*   object that just happens to have the same values as the original.*/
void vrInputObjectCopy(vrInput *dest_object, vrInput *src_object)
{
	void	*dest_mem;
	void	*src_mem;
	int	memlen;

	/* copy only the memory after the generic vrObjectInfo stuff */
	dest_mem = (char *)dest_object + sizeof(vrObjectInfo);
	src_mem = (char *)src_object + sizeof(vrObjectInfo);
	memlen = sizeof(vrInput) - sizeof(vrObjectInfo);
	memcpy(dest_mem, src_mem, memlen);

	/* make independent copy of some fields */
	dest_object->r2e_xform = vrShmemMemDup(src_object->r2e_xform, sizeof(*(dest_object->r2e_xform)));
	dest_object->ipos_xform = vrShmemMemDup(src_object->ipos_xform, sizeof(*(dest_object->ipos_xform)));
}


/*******************************************************************/
/* vrInputTypeName(): returns a string representation of the given */
/*   vrInputType.                                                  */
char *vrInputTypeName(vrInputType type)
{
	switch (type) {
	case VRINPUT_BINARY:
		return "2-switch";
	case VRINPUT_NWAY:
		return "N-switch";
	case VRINPUT_VALUATOR:
		return "valuator";
	case VRINPUT_KEYSTROKE:
		return "keystroke";
	case VRINPUT_TEXT:
		return "text";
	case VRINPUT_6SENSOR:
		return "6-sensor";
	case VRINPUT_NSENSOR:
		return "N-sensor";
	case VRINPUT_POSITION:
		return "position";
	case VRINPUT_PLANE:
		return "plane";
	case VRINPUT_CONTROL:
		return "control";
	case VRINPUT_UNKNOWN:
	default:
		return "unknown";
	}
}


/************************************************************************/
/* vrInputGetFromMapname(): this function takes the "standardized" form */
/*   of an input name and returns a pointer to that newly created input.*/
/* The "standardized" form of an input is the device-type-instance (DTI)*/
/*   triple -- where the "instance" can be parsed further.  Thus:       */
/*       <device>:<type>[<instance>]                                    */
/*   The <device> portion is excluded in this case, and the "instance"  */
/*   is only used to provide the index-number, which is all that is     */
/*   needed to get a pointer to the particular input.                   */
/*                                                                      */
/* NOTE: as it happens, this function is only called from within the    */
/*   vr_telnet.c system, for getting and setting values of inputs via   */
/*   the telnet interface.                                              */
vrGenericInput *vrInputGetFromMapname(vrContextInfo *context, const char *mapname)
{
	int		index;
	vrGenericInput	*return_value;

	/* TODO: should skip beginning whitespace */

	if (!strncmp(mapname, "2-way[", 6)) {
		index = vrAtoI(&mapname[6]);
		if (index < 0 || index >= context->input->num_2ways)
			return NULL;
		return_value = (vrGenericInput *)context->input->switch2[index];
	} else
	if (!strncmp(mapname, "2-switch[", 8)) {
		index = vrAtoI(&mapname[8]);
		if (index < 0 || index >= context->input->num_2ways)
			return NULL;
		return_value = (vrGenericInput *)context->input->switch2[index];
	} else
	if (!strncmp(mapname, "2switch[", 8)) {
		index = vrAtoI(&mapname[8]);
		if (index < 0 || index >= context->input->num_2ways)
			return NULL;
		return_value = (vrGenericInput *)context->input->switch2[index];
	} else
	if (!strncmp(mapname, "switch2[", 8)) {	/* deprecated */
		index = vrAtoI(&mapname[8]);
		if (index < 0 || index >= context->input->num_2ways)
			return NULL;
		return_value = (vrGenericInput *)context->input->switch2[index];
	} else
	if (!strncmp(mapname, "s2[", 3)) {
		index = vrAtoI(&mapname[3]);
		if (index < 0 || index >= context->input->num_2ways)
			return NULL;
		return_value = (vrGenericInput *)context->input->switch2[index];
	} else

	if (!strncmp(mapname, "N-way[", 6)) {
		index = vrAtoI(&mapname[6]);
		if (index < 0 || index >= context->input->num_Nways)
			return NULL;
		return_value = (vrGenericInput *)context->input->switchN[index];
	} else
	if (!strncmp(mapname, "Nswitch[", 8)) {
		index = vrAtoI(&mapname[8]);
		if (index < 0 || index >= context->input->num_Nways)
			return NULL;
		return_value = (vrGenericInput *)context->input->switchN[index];
	} else
	if (!strncmp(mapname, "N-switch[", 8)) {
		index = vrAtoI(&mapname[8]);
		if (index < 0 || index >= context->input->num_Nways)
			return NULL;
		return_value = (vrGenericInput *)context->input->switchN[index];
	} else
	if (!strncmp(mapname, "switchN[", 8)) {	/* deprecated */
		index = vrAtoI(&mapname[8]);
		if (index < 0 || index >= context->input->num_Nways)
			return NULL;
		return_value = (vrGenericInput *)context->input->switchN[index];
	} else

	if (!strncmp(mapname, "valuator[", 9)) {
		index = vrAtoI(&mapname[9]);
		if (index < 0 || index >= context->input->num_valuators)
			return NULL;
		return_value = (vrGenericInput *)context->input->valuator[index];
	} else
	if (!strncmp(mapname, "v[", 2)) {
		index = vrAtoI(&mapname[2]);
		if (index < 0 || index >= context->input->num_valuators)
			return NULL;
		return_value = (vrGenericInput *)context->input->valuator[index];
	} else

	if (!strncmp(mapname, "6-sensor[", 9)) {
		index = vrAtoI(&mapname[9]);
		if (index < 0 || index >= context->input->num_6sensors)
			return NULL;
		return_value = (vrGenericInput *)context->input->sensor6[index];
	} else
	if (!strncmp(mapname, "6sensor[", 8)) {
		index = vrAtoI(&mapname[8]);
		if (index < 0 || index >= context->input->num_6sensors)
			return NULL;
		return_value = (vrGenericInput *)context->input->sensor6[index];
	} else
	if (!strncmp(mapname, "sensor6[", 8)) {	/* deprecated */
		index = vrAtoI(&mapname[8]);
		if (index < 0 || index >= context->input->num_6sensors)
			return NULL;
		return_value = (vrGenericInput *)context->input->sensor6[index];
	} else

	if (!strncmp(mapname, "N-sensor[", 9)) {
		index = vrAtoI(&mapname[9]);
		if (index < 0 || index >= context->input->num_Nsensors)
			return NULL;
		return_value = (vrGenericInput *)context->input->sensorN[index];
	} else
	if (!strncmp(mapname, "Nsensor[", 8)) {
		index = vrAtoI(&mapname[8]);
		if (index < 0 || index >= context->input->num_Nsensors)
			return NULL;
		return_value = (vrGenericInput *)context->input->sensorN[index];
	} else
	if (!strncmp(mapname, "sensorN[", 8)) {	/* deprecated */
		index = vrAtoI(&mapname[8]);
		if (index < 0 || index >= context->input->num_Nsensors)
			return NULL;
		return_value = (vrGenericInput *)context->input->sensorN[index];
	} else

	if (!strncmp(mapname, "control[", 8)) {
		index = vrAtoI(&mapname[8]);
		if (index < 0 || index >= context->input->num_controls)
			return NULL;
		return_value = (vrGenericInput *)context->input->control[index];
	} else

	/* default: */ {
		return_value = NULL;
	}

	vrDbgPrintfN(PARSE_DBGLVL, "vrInputGetFromMapname(%s) returning %#p.\n", mapname, return_value);

	return return_value;
}


/**************************************************************************/
/* vrInputGetFromTypeIndex(): this function takes an input type and index */
/*   number (and the system context), and returns a pointer to the FreeVR */
/*   input that is the N'th input of the given type.                      */
vrGenericInput *vrInputGetFromTypeIndex(vrContextInfo *context, vrInputType type, int input_num)
{
	vrGenericInput	*input = NULL;

	switch (type) {
	case VRINPUT_BINARY:
		if (input_num >= context->input->num_2ways) {
			if (vrDbgDo(CONFIG_WARN_DBGLVL)) {
				if (context->input->num_2ways == 0)
					vrErrPrintf("Invalid 2switch requested (%d) -- none available.\n", input_num);
				else	vrErrPrintf("Invalid 2switch requested (%d) -- valid range is [0..%d].\n",
					input_num, context->input->num_2ways-1);

				/* Now Create a new dummy input to handle the request */
				vrDbgPrintfN(AALWAYS_DBGLVL, "FreeVR (%.4f): " RED_TEXT "Creating dummy 2switch %d.\n" NORM_TEXT,
					vrCurrentSimTime(), input_num);
				input = vrInputAddDummyToInputMap(context->input, VRINPUT_BINARY, input_num);
			}
		} else {
			input = (vrGenericInput *)context->input->switch2[input_num];

			/* Ideally this should never happen, but until the vrInputWaitForAllInputsToBeCreated() */
			/*   routine creates dummy inputs for devices that don't initiate, this could happen.   */
			if (input == NULL) {
				vrDbgPrintfN(AALWAYS_DBGLVL, "FreeVR (%.4f): " RED_TEXT "NULL input!  Creating dummy 2-switch %d.\n" NORM_TEXT,
					vrCurrentSimTime(), input_num);
				input = vrInputAddDummyToInputMap(context->input, VRINPUT_BINARY, input_num);
				context->input->switch2[input_num] = (vr2switch *)input;
			}
		}
		break;

	case	VRINPUT_NWAY:
		if (input_num >= context->input->num_Nways) {
			if (vrDbgDo(CONFIG_WARN_DBGLVL)) {
				if (context->input->num_Nways == 0)
					vrErrPrintf("Invalid Nswitch requested (%d) -- none available.\n", input_num);
				else	vrErrPrintf("Invalid Nswitch requested (%d) -- valid range is [0..%d].\n",
					input_num, context->input->num_Nways-1);

				/* Now Create a new dummy input to handle the request */
				vrDbgPrintfN(AALWAYS_DBGLVL, "FreeVR (%.4f): " RED_TEXT "Creating dummy N-switch %d.\n" NORM_TEXT,
					vrCurrentSimTime(), input_num);
				input = vrInputAddDummyToInputMap(context->input, VRINPUT_NWAY, input_num);
			}
		} else {
			input = (vrGenericInput *)context->input->switchN[input_num];

			/* Ideally this should never happen, but until the vrInputWaitForAllInputsToBeCreated() */
			/*   routine creates dummy inputs for devices that don't initiate, this could happen.   */
			if (input == NULL) {
				vrDbgPrintfN(AALWAYS_DBGLVL, "FreeVR (%.4f): " RED_TEXT "NULL input!  Creating dummy N-switch %d.\n" NORM_TEXT,
					vrCurrentSimTime(), input_num);
				input = vrInputAddDummyToInputMap(context->input, VRINPUT_NWAY, input_num);
				context->input->switchN[input_num] = (vrNswitch *)input;
			}
		}
		break;

	case	VRINPUT_VALUATOR:
		if (input_num >= context->input->num_valuators) {
			if (vrDbgDo(CONFIG_WARN_DBGLVL)) {
				if (context->input->num_valuators == 0)
					vrErrPrintf("Invalid Valuator requested (%d) -- none available.\n", input_num);
				else	vrErrPrintf("Invalid Valuator requested (%d) -- valid range is [0..%d].\n",
					input_num, context->input->num_valuators-1);

				/* Now Create a new dummy input to handle the request */
				vrDbgPrintfN(AALWAYS_DBGLVL, "FreeVR (%.4f): " RED_TEXT "Creating dummy Valuator %d.\n" NORM_TEXT,
					vrCurrentSimTime(), input_num);
				input = vrInputAddDummyToInputMap(context->input, VRINPUT_VALUATOR, input_num);
			}
		} else {
			input = (vrGenericInput *)context->input->valuator[input_num];

			/* Ideally this should never happen, but until the vrInputWaitForAllInputsToBeCreated() */
			/*   routine creates dummy inputs for devices that don't initiate, this could happen.   */
			if (input == NULL) {
				vrDbgPrintfN(AALWAYS_DBGLVL, "FreeVR (%.4f): " RED_TEXT "NULL input!  Creating dummy Valuator %d.\n" NORM_TEXT,
					vrCurrentSimTime(), input_num);
				input = vrInputAddDummyToInputMap(context->input, VRINPUT_VALUATOR, input_num);
				context->input->valuator[input_num] = (vrValuator *)input;
			}
		}
		break;

	case	VRINPUT_6SENSOR:
		if (input_num >= context->input->num_6sensors) {
			if (vrDbgDo(CONFIG_WARN_DBGLVL)) {
				if (context->input->num_6sensors == 0)
					vrErrPrintf("Invalid Position Sensor requested (%d) -- none available.\n", input_num);
				else	vrErrPrintf("Invalid Position Sensor requested (%d) -- valid range is [0..%d].\n",
					input_num, context->input->num_6sensors-1);

				/* Now Create a new dummy input to handle the request */
				vrDbgPrintfN(AALWAYS_DBGLVL, "FreeVR (%.4f): " RED_TEXT "Creating dummy Position Sensor %d.\n" NORM_TEXT,
					vrCurrentSimTime(), input_num);
				input = vrInputAddDummyToInputMap(context->input, VRINPUT_6SENSOR, input_num);
			}
		} else {
			input = (vrGenericInput *)context->input->sensor6[input_num];

			/* Ideally this should never happen, but until the vrInputWaitForAllInputsToBeCreated() */
			/*   routine creates dummy inputs for devices that don't initiate, this could happen.   */
			if (input == NULL) {
				vrDbgPrintfN(AALWAYS_DBGLVL, "FreeVR (%.4f): " RED_TEXT "NULL input!  Creating dummy Position Sensor %d.\n" NORM_TEXT,
					vrCurrentSimTime(), input_num);
				input = vrInputAddDummyToInputMap(context->input, VRINPUT_6SENSOR, input_num);
				context->input->sensor6[input_num] = (vr6sensor *)input;
			}
		}
		break;

	case	VRINPUT_NSENSOR:
		if (input_num >= context->input->num_Nsensors) {
			if (vrDbgDo(CONFIG_WARN_DBGLVL)) {
				if (context->input->num_Nsensors == 0)
					vrErrPrintf("Invalid N-Sensor requested (%d) -- none available.\n", input_num);
				else	vrErrPrintf("Invalid N-Sensor requested (%d) -- valid range is [0..%d].\n",
					input_num, context->input->num_Nsensors-1);

				/* Now Create a new dummy input to handle the request */
				vrDbgPrintfN(AALWAYS_DBGLVL, "FreeVR (%.4f): " RED_TEXT "Creating dummy N-Sensor %d.\n" NORM_TEXT,
					vrCurrentSimTime(), input_num);
				input = vrInputAddDummyToInputMap(context->input, VRINPUT_NSENSOR, input_num);
			}
		} else {
			input = (vrGenericInput *)context->input->sensorN[input_num];

			/* Ideally this should never happen, but until the vrInputWaitForAllInputsToBeCreated() */
			/*   routine creates dummy inputs for devices that don't initiate, this could happen.   */
			if (input == NULL) {
				vrDbgPrintfN(AALWAYS_DBGLVL, "FreeVR (%.4f): " RED_TEXT "NULL input!  Creating dummy N-Sensor %d.\n" NORM_TEXT,
					vrCurrentSimTime(), input_num);
				input = vrInputAddDummyToInputMap(context->input, VRINPUT_NSENSOR, input_num);
				context->input->sensorN[input_num] = (vrNsensor *)input;
			}
		}
		break;

	case	VRINPUT_KEYSTROKE:
	case	VRINPUT_TEXT:
	case	VRINPUT_POSITION:
	case	VRINPUT_PLANE:
	case	VRINPUT_CONTROL:
		vrErrPrintf("vrInputGetFromTypeIndex(): Unimplemented input type requested.\n");
		break;

	default:
		vrErrPrintf("vrInputGetFromTypeIndex(): Unknown input type requested.\n");
		break;
	}

	return input;
}


/****************************************************************************/
/* vrFprintInputObject(): Output an InputObject (which might be any of the  */
/*   input types (2switch, valuator, 6sensor, ...) -- the "input container".*/
/*   This is a higher level structure that contains information about the   */
/*   input including what input device it belongs to, what line it was      */
/*   configured on, and what parameters were used to create it.  This then  */
/*   points to the specific input-value type (vr2switch, vrValuator, etc.). */
void vrFprintInputObject(FILE *file, vrInput *inobj, vrPrintStyle style)
{

	switch (style) {

	case brief:
	case one_line:
		/* Take advantage of the fact that we can at least print the value in one line */
		/* TODO: consider prepending the object name or something */
#if 0
		vrFprintf(file, "Object_type = %s (%d)\n\tid = %d\n\tname = \"%s\"\n",
			vrObjectTypeName(inobj->object_type),
			inobj->object_type,
			inobj->id,
			inobj->name);
#endif
		vrFprintInputValue(file, inobj->container.generic, style);
		break;

	case machine:
		/* TODO: not yet implemented -- output easily machine parsed */
		vrFprintf(file, "TODO: put special version of input object info here\n");
		break;

	default:
	case verbose:
		vrFprintf(file, "{\n");
		vrFprintf(file, "\r"
			"\tObject_type = %s (%d)\n\tid = %d\n\tname = \"%s\"\n"
			"\tmalleable = %d\n\tnext = %#p\n"
			"\tCreated at %s, line %d\n"
			"\tLast modified at %s, line %d\n",
			vrObjectTypeName(inobj->object_type),
			inobj->object_type,
			inobj->id,
			inobj->name,
			inobj->malleable,
			inobj->next,
			inobj->file_created,
			inobj->line_created,
			inobj->file_lastmod,
			inobj->line_lastmod);
		vrFprintf(file,
			"\r\tmap_refcount = %d\n"
			"\tdesc_str (%#p) = \"%s\"\n"
			"\tdesc_ui = \"%s\"\n"
			"\tvalue container = %#p\n",
			inobj->map_refcount,
			inobj->desc,
			inobj->desc_str,
			inobj->desc_ui,
			inobj->container.generic);
#if 1	/* 09/23/21: TODO: determine whether we actually need the r2e_xform and ipos_xform stored at the input object/input container level -- possibly delete me */
		if (inobj->r2e_xform == NULL)
			vrFprintf(file, "\r\tr2e.xform: (--)\n");
		else	vrFprintf(file, "\r\tr2e.xform: (%3.1lf %3.1lf %3.1lf %3.1f/%3.1lf %3.1lf %3.1lf %3.1f/%3.1lf %3.1lf %3.1lf %3.1f/%3.1lf %3.1lf %3.1lf %3.1f)\n",
				inobj->r2e_xform->v[ 0], inobj->r2e_xform->v[ 4], inobj->r2e_xform->v[ 8], inobj->r2e_xform->v[12],
				inobj->r2e_xform->v[ 1], inobj->r2e_xform->v[ 5], inobj->r2e_xform->v[ 9], inobj->r2e_xform->v[13],
				inobj->r2e_xform->v[ 2], inobj->r2e_xform->v[ 6], inobj->r2e_xform->v[10], inobj->r2e_xform->v[14],
				inobj->r2e_xform->v[ 3], inobj->r2e_xform->v[ 7], inobj->r2e_xform->v[11], inobj->r2e_xform->v[15]);

		if (inobj->ipos_xform == NULL)
			vrFprintf(file, "\r\tipos.xform: (--)\n");
		else	vrFprintf(file, "\r\tipos.xform: (%3.1lf %3.1lf %3.1lf %3.1f/%3.1lf %3.1lf %3.1lf %3.1f/%3.1lf %3.1lf %3.1lf %3.1f/%3.1lf %3.1lf %3.1lf %3.1f)\n",
				inobj->ipos_xform->v[ 0], inobj->ipos_xform->v[ 4], inobj->ipos_xform->v[ 8], inobj->ipos_xform->v[12],
				inobj->ipos_xform->v[ 1], inobj->ipos_xform->v[ 5], inobj->ipos_xform->v[ 9], inobj->ipos_xform->v[13],
				inobj->ipos_xform->v[ 2], inobj->ipos_xform->v[ 6], inobj->ipos_xform->v[10], inobj->ipos_xform->v[14],
				inobj->ipos_xform->v[ 3], inobj->ipos_xform->v[ 7], inobj->ipos_xform->v[11], inobj->ipos_xform->v[15]);
#endif

		vrFprintInputValue(file, inobj->container.generic, style);
		vrFprintf(file, "}\n");
		break;

	case file_format:
		vrFprintf(file, "#inputobject \"%s\" = {\n", inobj->name);

		if (!inobj->malleable)
			vrFprintf(file, "#\tmalleable = %d;\n", inobj->malleable);

		/* TODO: */

		vrFprintf(file, "}#\n\n");
		break;
	}
}


/***************************************************************/
/* vrInputDeviceClear(): * clears all the values from a FreeVR */
/*   input device object.                                      */
void vrInputDeviceClear(vrInputDevice *object)
{
	object->type = NULL;
	object->dso_file = NULL;
	object->dso_func = NULL;
	object->dso_handle = NULL;
	object->version = NULL;
	object->args = NULL;
	object->operating = 0;
	object->inputs = NULL;
	object->self_controls = NULL;
	object->counted = 0;

	/* When clearing the matrices just reset them if already allocated */
	if (object->t2rw_xform == NULL)
		object->t2rw_xform = vrMatrixCreateIdentity();
	else	vrMatrixSetIdentity(object->t2rw_xform);
	if (object->r2e_xform == NULL)
		object->r2e_xform = vrMatrixCreateIdentity();
	else	vrMatrixSetIdentity(object->r2e_xform);
	if (object->ipos_xform == NULL)
		object->ipos_xform = vrMatrixCreateIdentity();
	else	vrMatrixSetIdentity(object->ipos_xform);
}


/*********************************************************************/
/* vrInputDeviceCopy(): copy the input information of a FreeVR       */
/*   input device object.  Note that it leaves untouched the general */
/*   FreeVR object information, thus creating a "distinct" object    */
/*   that just happens to have the same values as the original.      */
void vrInputDeviceCopy(vrContextInfo *context, vrInputDevice *dest_object, vrInputDevice *src_object)
{
	void	*dest_mem;
	void	*src_mem;
	int	memlen;

	/* copy only the memory after the generic vrObjectInfo stuff */
	dest_mem = (void *)dest_object + sizeof(vrObjectInfo);		/* 03/08/2006: "(char *)" changed to "(void *)" -- see whether this breaks anything (it seems cleaner) */
	src_mem = (void *)src_object + sizeof(vrObjectInfo);		/* 03/08/2006: ditto */
	memlen = sizeof(vrInputDevice) - sizeof(vrObjectInfo);
	memcpy(dest_mem, src_mem, memlen);

	/* make independent copy of some fields */
	dest_object->t2rw_xform = vrShmemMemDup(src_object->t2rw_xform, sizeof(*(dest_object->t2rw_xform)));
	dest_object->r2e_xform = vrShmemMemDup(src_object->r2e_xform, sizeof(*(dest_object->r2e_xform)));
	dest_object->ipos_xform = vrShmemMemDup(src_object->ipos_xform, sizeof(*(dest_object->ipos_xform)));

	dest_object->inputs = vrObjectCopyLinkedList(context, src_object->inputs);
	dest_object->self_controls = vrObjectCopyLinkedList(context, src_object->self_controls);
}


/***************************************************************************/
/* _MakeDummyInputDevice(): creates a non-existent input device to handle  */
/*   occasions when the application expects (and configuration doesn't     */
/*   provide) an input, and rather than continuously streaming out error   */
/*   messages, a "dummy" device is created once.  (And usually the calling */
/*   function will output a message that a dummy device had to be created. */
vrInputDevice *_MakeDummyInputDevice()
{
	vrInputDevice	*new_indev = vrShmemAlloc0(sizeof(*new_indev));

	vrDbgPrintfN(ALWAYS_DBGLVL, "_MakeDummyInputDevice(): " ANSI_TEXTRED "A runtime need required the creation of a dummy input device!\n" ANSI_TEXTNORM);
	vrInputDeviceClear(new_indev);
	new_indev->name = vrShmemStrDup("Dummy Input Device");	/* under normal (non-dummy) circumstances, this would be created in vrObjectNew during the configuration parsing process */

	return new_indev;
}


/*************************************************************************/
/* _MakeDummyInputObject(): creates an unconfigured input (container)    */
/*   object that is then usually provided an input value object that     */
/*   needed to be created to address a shortcoming in the configuration. */
vrInput *_MakeDummyInputObject(char *function, int line, char *description)
{
static	vrInput 	*dummy_list = NULL;		/* the linked list of all the created input objects */
	vrInput		*input_object;

	/* create a new input object definition (adding it to the dummy inputdevice's private list of input objects) */
//	vrPrintf("BEFORE: (vrObjectInfo **)(&dummy_list) = %p; dummy_list = %p\n", (vrObjectInfo **)(&dummy_list), dummy_list);
	input_object = (vrInput *)vrObjectListNew(vrContext, VROBJECT_INPUT, (vrObjectInfo **)(&dummy_list), "unassigned input");
//	vrPrintf("AFTER: (vrObjectInfo **)(&dummy_list) = %p; dummy_list = %p\n", (vrObjectInfo **)(&dummy_list), dummy_list);
	input_object->line_created = line;
	snprintf(input_object->file_created, sizeof(input_object->file_created), "'on the fly' in %s()", function);

	input_object->desc_str = vrShmemStrDup(description);
	input_object->desc_ui[0] = '\0';
	vrDbgPrintfN(CONFIG_DBGLVL FORCE_DBG, "_MakeDummyInputObject(): Created new %s (at %p): '%s' (%d) -- %s\n", "input object", input_object, input_object->name, input_object->id, description);	/* This is to debug possible memory over-writes -- see 11/12/21 & 11/13/21 diary entries */
	//vrFprintInputObject(stdout, input_object, verbose);	/* ditto */

	return (input_object);
}


/****************************************************************************/
/* vrFprintInputDevice(): prints information about a complete input device. */
/*   For example the xwindows-input, or the VRPN input.  Depending on the   */
/*   given print-style, it may also print information about the actual      */
/*   inputs (input containers) configured for that input device.            */
void vrFprintInputDevice(FILE *file, vrInputDevice *device, vrPrintStyle style)
{
	vrInput		*input;		/* pointer for looping through the device's inputs */
	vrMatrix	*last_r2e;	/* used for the file_format output of inputs */
	vrMatrix	*last_ipos;	/* used for the file_format output of inputs */
	vrMatrix	*tmpmat;	/* pointer to matrix to be printed */

	switch (style) {
	case one_line:
	case brief:
		vrFprintf(file, "id = %d, %s'%s' " ANSI_TEXTNORM ", type = '%s', operating = %d\n",
			device->id,
			(device->operating ? "" : ANSI_TEXTRED), device->name,
			device->type,
			device->operating);

		/* For "brief" style, add a two lines with the t2rw matrix & decomposition. */
		if (style == brief) {
			tmpmat = device->t2rw_xform;
			vrFprintf(file, "\tt2rw = (%3.1f %3.1f %3.1f %3.1f/%3.1f %3.1f %3.1f %3.1f/%3.1f %3.1f %3.1f %3.1f/%3.1f %3.1f %3.1f %3.1f)\n",
				tmpmat->v[ 0], tmpmat->v[ 4], tmpmat->v[ 8], tmpmat->v[12],
				tmpmat->v[ 1], tmpmat->v[ 5], tmpmat->v[ 9], tmpmat->v[13],
				tmpmat->v[ 2], tmpmat->v[ 6], tmpmat->v[10], tmpmat->v[14],
				tmpmat->v[ 3], tmpmat->v[ 7], tmpmat->v[11], tmpmat->v[15]);
			vrFprintMatrixCompositionOL(file, "\tt2rw: ", tmpmat);
		}
		break;

	default:
	case verbose:
		vrFprintf(file, "{\n"
			"\tself = %#p\n"
			"\tObject_type = %s (%d)\n\tid = %d\n\t%sname = \"%s\"\n" ANSI_TEXTNORM
			"\ttype = '%s'\n"
			"\tmalleable = %d\n\tnext = %#p\n"
			"\tCreated at %s, line %d\n"
			"\tLast modified at %s, line %d\n",
			device,
			vrObjectTypeName(device->object_type),
			device->object_type,
			device->id,
			(device->operating ? "" : ANSI_TEXTRED), device->name,
			device->type,
			device->malleable,
			device->next,
			device->file_created,
			device->line_created,
			device->file_lastmod,
			device->line_lastmod);
		vrFprintf(file, "\r"
			"\tproc = %#p (pid %d)\n"
			"\tdso file = '%s'\n\tdso function = '%s'\n"
			"\tdso handle = %#p\n\tversion = '%s'\n\targs = '%s'\n",
			device->proc,
			(device->proc == NULL) ? -1 : device->proc->pid,
			(device->dso_file == NULL) ? "" : device->dso_file,
			(device->dso_func == NULL) ? "" : device->dso_func,
			device->dso_handle,
			device->version,
			device->args);
		if (device->t2rw_xform == NULL)
			vrFprintf(file, "\r\tt2rw_xform: (--)\n");
		else	vrFprintf(file, "\r"
				"\tt2rw_xform: (%3.1lf %3.1lf %3.1lf %3.1lf/%3.1lf %3.1lf %3.1lf %3.1lf/%3.1lf %3.1lf %3.1lf %3.1lf/%3.1lf %3.1lf %3.1lf %3.1lf)\n",
				device->t2rw_xform->v[ 0], device->t2rw_xform->v[ 4], device->t2rw_xform->v[ 8], device->t2rw_xform->v[12],
				device->t2rw_xform->v[ 1], device->t2rw_xform->v[ 5], device->t2rw_xform->v[ 9], device->t2rw_xform->v[13],
				device->t2rw_xform->v[ 2], device->t2rw_xform->v[ 6], device->t2rw_xform->v[10], device->t2rw_xform->v[14],
				device->t2rw_xform->v[ 3], device->t2rw_xform->v[ 7], device->t2rw_xform->v[11], device->t2rw_xform->v[15]);
		if (device->r2e_xform == NULL)
			vrFprintf(file, "\r\tr2e.xform: (--)\n");
		else	vrFprintf(file, "\r"
				"\tr2e.xform: (%3.1lf %3.1lf %3.1lf %3.1lf/%3.1lf %3.1lf %3.1lf %3.1lf/%3.1lf %3.1lf %3.1lf %3.1lf/%3.1lf %3.1lf %3.1lf %3.1lf)\n",
				device->r2e_xform->v[ 0], device->r2e_xform->v[ 4], device->r2e_xform->v[ 8], device->r2e_xform->v[12],
				device->r2e_xform->v[ 1], device->r2e_xform->v[ 5], device->r2e_xform->v[ 9], device->r2e_xform->v[13],
				device->r2e_xform->v[ 2], device->r2e_xform->v[ 6], device->r2e_xform->v[10], device->r2e_xform->v[14],
				device->r2e_xform->v[ 3], device->r2e_xform->v[ 7], device->r2e_xform->v[11], device->r2e_xform->v[15]);

		if (device->ipos_xform == NULL)
			vrFprintf(file, "\r\tipos.xform: (--)\n");
		else	vrFprintf(file, "\r"
				"\tipos_xform: (%3.1lf %3.1lf %3.1lf %3.1lf/%3.1lf %3.1lf %3.1lf %3.1lf/%3.1lf %3.1lf %3.1lf %3.1lf/%3.1lf %3.1lf %3.1lf %3.1lf)\n",
				device->ipos_xform->v[ 0], device->ipos_xform->v[ 4], device->ipos_xform->v[ 8], device->ipos_xform->v[12],
				device->ipos_xform->v[ 1], device->ipos_xform->v[ 5], device->ipos_xform->v[ 9], device->ipos_xform->v[13],
				device->ipos_xform->v[ 2], device->ipos_xform->v[ 6], device->ipos_xform->v[10], device->ipos_xform->v[14],
				device->ipos_xform->v[ 3], device->ipos_xform->v[ 7], device->ipos_xform->v[11], device->ipos_xform->v[15]);

		vrFprintf(file, "\r"
			"\t%soperating = %d\n" ANSI_TEXTNORM
			"\tinputs counted? = %d\n"
			"\tinput objects (%#p) [",
			(device->operating ? "" : ANSI_TEXTRED), device->operating,
			device->counted,
			device->inputs);
		for (input = device->inputs; input != NULL; input = (vrInput *)input->next)
			vrFprintf(file, " \"%s\"", input->name);

#if 0 /* Print all the input objects associated with this device */
		vrFprintf(file, "\nInput objects details:\n");
		for (input = device->inputs; input != NULL; input = (vrInput *)input->next)
			vrFprintInputObject(file, input, style);
		vrFprintf(file, "\n\n");
#endif

		vrFprintf(file, " ]\n\r"
			"\tself control input objects (%#p) [",
			device->self_controls);
		if (device->self_controls != NULL) {
			for (input = device->self_controls; input != NULL; input = (vrInput *)input->next) {
				vrFprintf(file, " \"%s\"", input->name);
#if 0 /* Print info about each control -- for debugging (bug solved, so at some point delete me) */
				vrFprintf(file,"line:%d\n", input->line_created);
				vrFprintf(file,"line-last:%d of %s\n", input->line_lastmod, input->file_lastmod);
				vrFprintf(file,"desc_string:%s\n", input->desc_str);
#endif
			}
		}
		vrFprintf(file, " ]\n\r"
			"\tnum_6sensors = %d\n\t6sensor (data ptr) = %#p\n"
			"\tnum_Nsensors = %d\n\tNsensor (data ptr) = %#p\n"
			"\tnum_2ways = %d\n\tswitch2 (data ptr) = %#p\n"
			"\tnum_Nways = %d\n\tswitchN (data ptr) = %#p\n"
			"\tnum_valuators = %d\n\tvaluator (data ptr) = %#p\n"
			"\tnum_texts = %d\n\ttext (data ptr) = %#p\n"
			"\tnum_controls = %d\n\tcontrol (data ptr) = %#p\n",
			device->num_6sensors,	device->sensor6,
			device->num_Nsensors,	device->sensorN,
			device->num_2ways,	device->switch2,
			device->num_Nways,	device->switchN,
			device->num_valuators,	device->valuator,
			device->num_texts,	device->text,
			device->num_controls,	device->control);

		/* If device->Create exists, the rest should exist */
		if (device->Create != NULL) {
			vrFprintf(file, "\r"
				"\tCreate = %#p (%s)\n\tOpen = %#p (%s)\n\tClose = %#p (%s)\n"
				"\tReset = %#p (%s)\n\tPollData = %#p (%s)\n\tPrintAux = %#p (%s)\n",
				device->Create,		(device->Create->name != NULL ? device->Create->name : "-"),
				device->Open,		(device->Open->name != NULL ? device->Open->name : "-"),
				device->Close,		(device->Close->name != NULL ? device->Close->name : "-"),
				device->Reset,		(device->Reset->name != NULL ? device->Reset->name : "-"),
				device->PollData,	(device->PollData->name != NULL ? device->PollData->name : "-"),
				device->PrintAux,	(device->PrintAux->name != NULL ? device->PrintAux->name : "-")
				);
		}
		vrFprintf(file, "\r"
			"\taux_data = %#p\n"
			"}\n",
			device->aux_data);

		/* Now print the auxiliary data, if available */
		if (device->PrintAux != NULL)
			vrCallbackInvokeDynamic(device->PrintAux, 3, file, device->aux_data, style);
		break;

	case file_format:
		vrFprintf(file, "inputdevice \"%s\" = {\n", device->name);

		if (!device->malleable)
			vrFprintf(file, "\tmalleable = %d;\n", device->malleable);
		vrFprintf(file, "\ttype = \"%s\";\n", device->type);
		if (device->args != NULL)
			vrFprintf(file, "\targs = \"%s\";\n", device->args);
		if (device->version != NULL)
			vrFprintf(file, "\t#version = \"%s\";\n", device->version);
		if (device->dso_file != NULL)
			vrFprintf(file, "\tdsofile = \"%s\";\n", device->dso_file);
		if (device->dso_func != NULL)
			vrFprintf(file, "\tdsofunc = \"%s\";\n", device->dso_file);
		vrFprintf(file, "\n");

		vrFprintf(file, "\t# WARNING: the order of the t2rw and r2e matrices is untested, and may very well be incorrect.\n");
		if (device->t2rw_xform == NULL)
			vrFprintf(file, "\t# No t2rw.transform defined.\n");
		else	vrFprintf(file, "\tt2rw.transform =%6.3lf,%6.3lf,%6.3lf,%6.3lf,  %6.3lf,%6.3lf,%6.3lf,%6.3lf,  %6.3lf,%6.3lf,%6.3lf,%6.3lf,  %6.3lf,%6.3lf,%6.3lf,%6.3lf;\n",
				device->t2rw_xform->v[ 0], device->t2rw_xform->v[ 4], device->t2rw_xform->v[ 8], device->t2rw_xform->v[12],
				device->t2rw_xform->v[ 1], device->t2rw_xform->v[ 5], device->t2rw_xform->v[ 9], device->t2rw_xform->v[13],
				device->t2rw_xform->v[ 2], device->t2rw_xform->v[ 6], device->t2rw_xform->v[10], device->t2rw_xform->v[14],
				device->t2rw_xform->v[ 3], device->t2rw_xform->v[ 7], device->t2rw_xform->v[11], device->t2rw_xform->v[15]);
		/* NOTE: there is almost no point in printing the r2e matrix here */
		/*   because the "device" copy will almost always be the one used */
		/*   with the last "input" copy, so if there are more than one    */
		/*   r2e matrices, the first input to use it will always override */
		/*   this one, and then eventually this one will be printed again.*/
		/*   Of course, if there is only one, then it looks nicer here!   */
		last_r2e = device->r2e_xform;
		if (device->r2e_xform == NULL)
			vrFprintf(file, "\t# No r2e.transform defined.\n");
		else	vrFprintf(file, "\tr2e.transform =%6.3lf,%6.3lf,%6.3lf,%6.3lf,  %6.3lf,%6.3lf,%6.3lf,%6.3lf,  %6.3lf,%6.3lf,%6.3lf,%6.3lf,  %6.3lf,%6.3lf,%6.3lf,%6.3lf;\n",
				device->r2e_xform->v[ 0], device->r2e_xform->v[ 4], device->r2e_xform->v[ 8], device->r2e_xform->v[12],
				device->r2e_xform->v[ 1], device->r2e_xform->v[ 5], device->r2e_xform->v[ 9], device->r2e_xform->v[13],
				device->r2e_xform->v[ 2], device->r2e_xform->v[ 6], device->r2e_xform->v[10], device->r2e_xform->v[14],
				device->r2e_xform->v[ 3], device->r2e_xform->v[ 7], device->r2e_xform->v[11], device->r2e_xform->v[15]);
		vrFprintf(file, "\n");

		last_ipos = device->ipos_xform;
		if (device->ipos_xform == NULL)
			vrFprintf(file, "\t# No ipos.transform defined.\n");
		else	vrFprintf(file, "\tipos.transform =%6.3lf,%6.3lf,%6.3lf,%6.3lf,  %6.3lf,%6.3lf,%6.3lf,%6.3lf,  %6.3lf,%6.3lf,%6.3lf,%6.3lf,  %6.3lf,%6.3lf,%6.3lf,%6.3lf;\n",
				device->ipos_xform->v[ 0], device->ipos_xform->v[ 4], device->ipos_xform->v[ 8], device->ipos_xform->v[12],
				device->ipos_xform->v[ 1], device->ipos_xform->v[ 5], device->ipos_xform->v[ 9], device->ipos_xform->v[13],
				device->ipos_xform->v[ 2], device->ipos_xform->v[ 6], device->ipos_xform->v[10], device->ipos_xform->v[14],
				device->ipos_xform->v[ 3], device->ipos_xform->v[ 7], device->ipos_xform->v[11], device->ipos_xform->v[15]);
		vrFprintf(file, "\n");

		for (input = device->inputs; input != NULL; input = (vrInput *)input->next) {
			/* Print r2e_xform for each input that uses it, and no   */
			/*   more.  Or maybe each time it's changed from before. */
			/* By storing the last printed r2e matrix, we only need  */
			/*   to print it again when it changes for a new input.  */
			/* Doing the same for the ipos_xofrm.                    */
#if 0
			vrFprintf(file, "\t# r2e matrix is %p, last is %p, compare is %d\n",
				input->r2e_xform,
				last_r2e,
				vrMatrixEqual(last_r2e, input->r2e_xform));
#endif
			if (!vrMatrixEqual(last_r2e, input->r2e_xform) && input->r2e_xform != NULL) {
				vrFprintf(file, "\tr2e.transform =%6.3lf,%6.3lf,%6.3lf,%6.3lf,  %6.3lf,%6.3lf,%6.3lf,%6.3lf,  %6.3lf,%6.3lf,%6.3lf,%6.3lf,  %6.3lf,%6.3lf,%6.3lf,%6.3lf;\n",
					input->r2e_xform->v[ 0], input->r2e_xform->v[ 4], input->r2e_xform->v[ 8], input->r2e_xform->v[12],
					input->r2e_xform->v[ 1], input->r2e_xform->v[ 5], input->r2e_xform->v[ 9], input->r2e_xform->v[13],
					input->r2e_xform->v[ 2], input->r2e_xform->v[ 6], input->r2e_xform->v[10], input->r2e_xform->v[14],
					input->r2e_xform->v[ 3], input->r2e_xform->v[ 7], input->r2e_xform->v[11], input->r2e_xform->v[15]);
				last_r2e = input->r2e_xform;
			}

			if (!vrMatrixEqual(last_ipos, input->ipos_xform) && input->ipos_xform != NULL) {
				vrFprintf(file, "\tipos.transform =%6.3lf,%6.3lf,%6.3lf,%6.3lf,  %6.3lf,%6.3lf,%6.3lf,%6.3lf,  %6.3lf,%6.3lf,%6.3lf,%6.3lf,  %6.3lf,%6.3lf,%6.3lf,%6.3lf;\n",
					input->ipos_xform->v[ 0], input->ipos_xform->v[ 4], input->ipos_xform->v[ 8], input->ipos_xform->v[12],
					input->ipos_xform->v[ 1], input->ipos_xform->v[ 5], input->ipos_xform->v[ 9], input->ipos_xform->v[13],
					input->ipos_xform->v[ 2], input->ipos_xform->v[ 6], input->ipos_xform->v[10], input->ipos_xform->v[14],
					input->ipos_xform->v[ 3], input->ipos_xform->v[ 7], input->ipos_xform->v[11], input->ipos_xform->v[15]);
				last_ipos = input->ipos_xform;
			}
			vrFprintf(file, "\tinput \"%s\" = \"%s\";\n", input->name, input->desc_str);
		}
		vrFprintf(file, "\n");

		for (input = device->self_controls; input != NULL; input = (vrInput *)input->next) {
			vrFprintf(file, "\tcontrol \"%s\" = \"%s\";\n", input->name, input->desc_str);
		}

		vrFprintf(file, "}\n\n");
		break;
	}
}


	/************************************************************/
	/************************************************************/
	/********** Generic self-control callback functions *********/

/***************************************************************************/
/* vrInputSystemPauseToggleCallback(): this callback toggles the "paused"  */
/*   field of the overall FreeVR "context" structure.  When the system is  */
/*   "paused", rendering still functions, but the inputs are frozen.  This */
/*   can be used for debugging, or for still photography of a CAVE.        */
void vrInputSystemPauseToggleCallback(vrInputDevice *devinfo, int value)
{
	if (value == 0)
		return;

	devinfo->context->paused ^= 1;
	vrDbgPrintfN(SELFCTRL_DBGLVL FORCE_DBG, "Input Control (%s): system_pause = %d.\n",
		devinfo->name, devinfo->context->paused);
}


/*****************************************************************************/
/* vrInputPrintContextStructCallback(): this callback generates output of    */
/*   the overall system "context" data.  It is generally used when debugging */
/*   the library or the configuration.                                       */
void vrInputPrintContextStructCallback(vrInputDevice *devinfo, int value)
{
	if (value == 0)
		return;

	vrFprintContext(stdout, devinfo->context, verbose);
}


/************************************************************************/
/* vrInputPrintConfigStructCallback(): this callback outputs the entire */
/*    VR system configuration to the text console.                      */
void vrInputPrintConfigStructCallback(vrInputDevice *devinfo, int value)
{
	if (value == 0)
		return;

	vrFprintConfig(stdout, devinfo->context->config, verbose);
}


/************************************************************************/
/* vrInputPrintInputStructCallback(): this callback outputs the current */
/*    state of the entire input structure of the running VR system.     */
void vrInputPrintInputStructCallback(vrInputDevice *devinfo, int value)
{
	if (value == 0)
		return;

	vrFprintInput(stdout, devinfo->context->input, verbose);
}


	/************************************************************/
	/************************************************************/
	/****** Functions to initialize the internal structures *****/

/***************************************************************************/
/* vrCreateInputMap(): create the structures in vrInputs that applications */
/*   refer to for input values (ie. button1, button2, valuator1, etc.).    */
/*                                                                         */
/* NOTE: it is important to note that this routine is called in the main   */
/*   process, not the input process.  This is done after all the input     */
/*   devices have reported an operating-status.                            */
void vrCreateInputMap(vrInputInfo *vrInputs, const char *mapname)
{
	int	count;
	int	devcount;
	int	num_2ways_assigned;
	int	num_Nways_assigned;
	int	num_valuators_assigned;
	int	num_6sensors_assigned;
	int	num_Nsensors_assigned;
	int	num_controls_assigned;

	vrTrace("vrCreateInputMap():", BOLD_TEXT "entering." NORM_TEXT);
	if (vrInputs->num_input_devices == 0) {
		vrErrPrintf("FreeVR (%.4f): " RED_TEXT "**WARNING**: No input devices have been defined in process! (Perhaps configuration error.)\n" NORM_TEXT, vrCurrentSimTime());
		return;
	}

	if (mapname == NULL) {
		vrErrPrintf("vrCreateInputMap(): " RED_TEXT "**WARNING**: NULL InputMap string!\n" NORM_TEXT);
		return;
	}

	if (!strcasecmp(mapname, "olddefault")) {	/* perhaps delete me -- maybe test it first */

		/* TODO: this hardwired monstrosity obviously needs to go away. */
		/*   That'll probably have to wait until most of the rest of    */
		/*   the config file parsing stuff is done, though.             */
		/* [10/13/99] -- actually, go away may be a bit strong.  It     */
		/*   might be wise to keep a "default" input map, but it        */
		/*   probably should do something like loop over all the input  */
		/*   devices and map all their inputs (in order) to the list of */
		/*   inputs in the configuration.  Then, make sure there is     */
		/*   some appropriate minimal number of inputs for each type,   */
		/*   and add dummy inputs where necessary.                      */

		vrDbgPrintfN(SELDOM_DBGLVL, "vrInputs data structure is at %#p w/ %d devices\n",
			vrInputs, vrInputs->num_input_devices);

		vrInputs->num_2ways += vrInputs->input_devices[0]->num_2ways;
		vrDbgPrintfN(SELDOM_DBGLVL, "vrInputs->num_2ways = %d\n", vrInputs->num_2ways);
		if (vrInputs->switch2 != NULL)
			vrShmemFree(vrInputs->switch2);
		if (vrInputs->input_devices[0]->num_2ways >= 3) {
			vrInputs->switch2 = (vr2switch **)vrShmemAlloc0(vrInputs->num_2ways * sizeof(vr2switch *));
			vrInputs->switch2[0] = &(vrInputs->input_devices[0]->switch2[0]);
			vrInputs->switch2[1] = &(vrInputs->input_devices[0]->switch2[1]);
			vrInputs->switch2[2] = &(vrInputs->input_devices[0]->switch2[2]);
		}

		vrInputs->num_Nways += vrInputs->input_devices[0]->num_Nways;
		vrDbgPrintfN(SELDOM_DBGLVL, "vrInputs->num_Nways = %d\n", vrInputs->num_Nways);
		if (vrInputs->switchN != NULL)
			vrShmemFree(vrInputs->switchN);
		if (vrInputs->input_devices[0]->num_Nways >= 2) {
			vrInputs->switchN = (vrNswitch **)vrShmemAlloc0(vrInputs->num_Nways * sizeof(vrNswitch *));
			vrInputs->switchN[0] = &(vrInputs->input_devices[0]->switchN[0]);
			vrInputs->switchN[1] = &(vrInputs->input_devices[0]->switchN[1]);
			vrInputs->switchN[2] = &(vrInputs->input_devices[0]->switchN[2]);
		}

		vrInputs->num_valuators += vrInputs->input_devices[0]->num_valuators;
		vrDbgPrintfN(SELDOM_DBGLVL, "vrInputs->num_valuators = %d\n", vrInputs->num_valuators);
		if (vrInputs->valuator != NULL)
			vrShmemFree(vrInputs->valuator);
		if (vrInputs->input_devices[0]->num_valuators >= 2) {
			vrInputs->valuator = (vrValuator **)vrShmemAlloc0(vrInputs->num_valuators * sizeof(vrValuator *));
			vrInputs->valuator[0] = &(vrInputs->input_devices[0]->valuator[0]);
			vrInputs->valuator[1] = &(vrInputs->input_devices[0]->valuator[1]);
			/* magellan & spaceorb have a third valuator (unused for now) */
			if (vrInputs->num_valuators >= 3)
				vrInputs->valuator[2] = &(vrInputs->input_devices[0]->valuator[2]);
		}

		vrInputs->num_6sensors += vrInputs->input_devices[0]->num_6sensors;
		vrDbgPrintfN(SELDOM_DBGLVL, "vrInputs->num_6sensors = %d\n", vrInputs->num_6sensors);
		if (vrInputs->sensor6 != NULL)
			vrShmemFree(vrInputs->sensor6);
		if (vrInputs->input_devices[0]->num_6sensors >= 2) {
			vrInputs->sensor6 = (vr6sensor **)vrShmemAlloc0(vrInputs->num_6sensors * sizeof(vr6sensor *));
			vrInputs->sensor6[0] = &(vrInputs->input_devices[0]->sensor6[0]);
			vrInputs->sensor6[1] = &(vrInputs->input_devices[0]->sensor6[1]);
			vrDbgPrintfN(SELDOM_DBGLVL, "vrInputs->sensor6[0] is at %#p -> %#p\n",
				&vrInputs->sensor6[0], vrInputs->sensor6[0]);
			vrDbgPrintfN(SELDOM_DBGLVL, "vrInputs->sensor6[1] is at %#p -> %#p\n",
				&vrInputs->sensor6[1], vrInputs->sensor6[1]);
			/* magellan has a third sensor6 (unused for now) */
			if (vrInputs->num_6sensors >= 3) {
				vrInputs->sensor6[2] = &(vrInputs->input_devices[0]->sensor6[2]);
				vrDbgPrintfN(SELDOM_DBGLVL, "vrInputs->sensor6[2] is at %#p -> %#p\n",
					&vrInputs->sensor6[2], vrInputs->sensor6[2]);
			}
		}

		/* assign user zero's head to 6-sensor 0 */
		vrInputs->users[0]->head = vrInputs->sensor6[0];


	} else if (!strcasecmp(mapname, "default")) {

		/* This version of default, simply puts all the available inputs */
		/*   into the map, in a first found, first added ordering.       */

		vrDbgPrintfN(SELDOM_DBGLVL, "vrInputs data structure is at %#p w/ %d devices\n",
			vrInputs, vrInputs->num_input_devices);

		/*************************************************************/
		/* first, we need to count all the different types of inputs */
		/*   from all the devices.                                   */
		vrInputs->num_2ways = 0;
		vrInputs->num_Nways = 0;
		vrInputs->num_valuators = 0;
		vrInputs->num_6sensors = 0;
		vrInputs->num_Nsensors = 0;
		vrInputs->num_controls = 0;
		for (devcount = 0; devcount < vrInputs->num_input_devices; devcount++) {
			vrInputs->num_2ways += vrInputs->input_devices[devcount]->num_2ways;
			vrInputs->num_Nways += vrInputs->input_devices[devcount]->num_Nways;
			vrInputs->num_valuators += vrInputs->input_devices[devcount]->num_valuators;
			vrInputs->num_6sensors += vrInputs->input_devices[devcount]->num_6sensors;
			vrInputs->num_Nsensors += vrInputs->input_devices[devcount]->num_Nsensors;
			vrInputs->num_controls += vrInputs->input_devices[devcount]->num_controls;
		}

		/***************************************************************/
		/* Now allocate the memory for the array of each type of input */

		/*** VRINPUT_BINARY ***/
		if (vrInputs->switch2 != NULL)
			vrShmemFree(vrInputs->switch2);
		vrInputs->switch2 = (vr2switch **)vrShmemAlloc0(vrInputs->num_2ways * sizeof(vr2switch *));
		num_2ways_assigned = 0;

		/*** VRINPUT_NWAY ***/
		if (vrInputs->switchN != NULL)
			vrShmemFree(vrInputs->switchN);
		vrInputs->switchN = (vrNswitch **)vrShmemAlloc0(vrInputs->num_Nways * sizeof(vrNswitch *));
		num_Nways_assigned = 0;

		/*** VRINPUT_VALUATOR ***/
		if (vrInputs->valuator != NULL)
			vrShmemFree(vrInputs->valuator);
		vrInputs->valuator = (vrValuator **)vrShmemAlloc0(vrInputs->num_valuators * sizeof(vrValuator *));
		num_valuators_assigned = 0;

		/*** VRINPUT_6SENSOR ***/
		if (vrInputs->sensor6 != NULL)
			vrShmemFree(vrInputs->sensor6);
		vrInputs->sensor6 = (vr6sensor **)vrShmemAlloc0(vrInputs->num_6sensors * sizeof(vr6sensor *));
		num_6sensors_assigned = 0;

		/*** VRINPUT_NSENSOR ***/
		if (vrInputs->sensorN != NULL)
			vrShmemFree(vrInputs->sensorN);
		vrInputs->sensorN = (vrNsensor **)vrShmemAlloc0(vrInputs->num_Nsensors * sizeof(vrNsensor *));
		num_Nsensors_assigned = 0;

		/*** VRINPUT_CONTROL ***/
		if (vrInputs->control != NULL)
			vrShmemFree(vrInputs->control);
		vrInputs->control = (vrControl **)vrShmemAlloc0(vrInputs->num_controls * sizeof(vrControl *));
		num_controls_assigned = 0;


		/***************************************************************************/
		/* Now, go through all the devices and add their inputs to the global list */
		for (devcount = 0; devcount < vrInputs->num_input_devices; devcount++) {

			/*** VRINPUT_BINARY ***/
			for (count = 0; count < vrInputs->input_devices[devcount]->num_2ways; count++) {
				vrInputs->switch2[num_2ways_assigned] = &(vrInputs->input_devices[devcount]->switch2[count]);
				vrInputs->switch2[num_2ways_assigned]->my_object->map_refcount++;
				num_2ways_assigned++;
			}

			/*** VRINPUT_NWAY ***/
			for (count = 0; count < vrInputs->input_devices[devcount]->num_Nways; count++) {
				vrInputs->switchN[num_Nways_assigned] = &(vrInputs->input_devices[devcount]->switchN[count]);
				vrInputs->switchN[num_Nways_assigned]->my_object->map_refcount++;
				num_Nways_assigned++;
			}

			/*** VRINPUT_VALUATOR ***/
			for (count = 0; count < vrInputs->input_devices[devcount]->num_valuators; count++) {
				vrInputs->valuator[num_valuators_assigned] = &(vrInputs->input_devices[devcount]->valuator[count]);
				vrInputs->valuator[num_valuators_assigned]->my_object->map_refcount++;
				num_valuators_assigned++;
			}

			/*** VRINPUT_6SENSOR ***/
			for (count = 0; count < vrInputs->input_devices[devcount]->num_6sensors; count++) {
				vrInputs->sensor6[num_6sensors_assigned] = &(vrInputs->input_devices[devcount]->sensor6[count]);
				vrInputs->sensor6[num_6sensors_assigned]->my_object->map_refcount++;
				num_6sensors_assigned++;
			}

			/*** VRINPUT_NSENSOR ***/
			for (count = 0; count < vrInputs->input_devices[devcount]->num_Nsensors; count++) {
				vrInputs->sensorN[num_Nsensors_assigned] = &(vrInputs->input_devices[devcount]->sensorN[count]);
				vrInputs->sensorN[num_Nsensors_assigned]->my_object->map_refcount++;
				num_Nsensors_assigned++;
			}

			/*** VRINPUT_CONTROL ***/
			for (count = 0; count < vrInputs->input_devices[devcount]->num_controls; count++) {
				vrInputs->control[num_controls_assigned] = &(vrInputs->input_devices[devcount]->control[count]);
				vrInputs->control[num_controls_assigned]->my_object->map_refcount++;
				num_controls_assigned++;
			}
		}


	} else if (!strcasecmp(mapname, "xtest")) {

		/* TODO: okay, this hardwired monstrosity obviously does need to go away. */ /* 11/12/21: explore whether to delete me */

		vrDbgPrintfN(SELDOM_DBGLVL, "vrInputs data structure is at %#p w/ %d devices\n",
			vrInputs, vrInputs->num_input_devices);

		vrInputs->num_2ways += vrInputs->input_devices[0]->num_2ways;
		vrDbgPrintfN(SELDOM_DBGLVL, "vrInputs->num_2ways = %d\n", vrInputs->num_2ways);
		if (vrInputs->switch2 != NULL)
			vrShmemFree(vrInputs->switch2);
		if (vrInputs->input_devices[0]->num_2ways >= 3) {
			vrInputs->switch2 = (vr2switch **)vrShmemAlloc0(vrInputs->num_2ways+1 * sizeof(vr2switch *));
			vrInputs->switch2[1] = &(vrInputs->input_devices[0]->switch2[0]);
			vrInputs->switch2[2] = &(vrInputs->input_devices[0]->switch2[1]);
			vrInputs->switch2[3] = &(vrInputs->input_devices[0]->switch2[2]);
		}

		/***** NEW stuff for testing X windows *****/
		/* we assume that the X device is the second input device */
		if (vrInputs->input_devices[1]->num_2ways > 0) {
			/* Notice I allocated an extra slot above */
			/* Also, I made this one the first one, just so it would */
			/*   be easier to see in the on-screen text output.      */
			vrInputs->switch2[0] = &(vrInputs->input_devices[1]->switch2[0]);
vrPrintf("Okay, xwin device is button[0]\n");
		} else {
			vrErr("YO: for some reason there wasn't a device\n");
			vrInputs->switch2[0] = &(vrInputs->input_devices[0]->switch2[0]);
		}

		vrInputs->num_Nways += vrInputs->input_devices[0]->num_Nways;
		vrDbgPrintfN(SELDOM_DBGLVL, "vrInputs->num_Nways = %d\n", vrInputs->num_Nways);
		if (vrInputs->switchN != NULL)
			vrShmemFree(vrInputs->switchN);
		if (vrInputs->input_devices[0]->num_Nways >= 2) {
			vrInputs->switchN = (vrNswitch **)vrShmemAlloc0(vrInputs->num_Nways * sizeof(vrNswitch *));
			vrInputs->switchN[0] = &(vrInputs->input_devices[0]->switchN[0]);
			vrInputs->switchN[1] = &(vrInputs->input_devices[0]->switchN[1]);
			vrInputs->switchN[2] = &(vrInputs->input_devices[0]->switchN[2]);
		}

		vrInputs->num_valuators += vrInputs->input_devices[0]->num_valuators;
		vrDbgPrintfN(SELDOM_DBGLVL, "vrInputs->num_valuators = %d\n", vrInputs->num_valuators);
		if (vrInputs->valuator != NULL)
			vrShmemFree(vrInputs->valuator);
		if (vrInputs->input_devices[0]->num_valuators >= 2) {
			vrInputs->valuator = (vrValuator **)vrShmemAlloc0(vrInputs->num_valuators * sizeof(vrValuator *));
			vrInputs->valuator[0] = &(vrInputs->input_devices[0]->valuator[0]);
			vrInputs->valuator[1] = &(vrInputs->input_devices[0]->valuator[1]);
			/* magellan & spaceorb have a third valuator (unused for now) */
			if (vrInputs->num_valuators >= 3)
				vrInputs->valuator[2] = &(vrInputs->input_devices[0]->valuator[2]);
		}

		vrInputs->num_6sensors += vrInputs->input_devices[0]->num_6sensors;
		vrDbgPrintfN(SELDOM_DBGLVL, "vrInputs->num_6sensors = %d\n", vrInputs->num_6sensors);
		if (vrInputs->sensor6 != NULL)
			vrShmemFree(vrInputs->sensor6);
		if (vrInputs->input_devices[0]->num_6sensors >= 2) {
			vrInputs->sensor6 = (vr6sensor **)vrShmemAlloc0(vrInputs->num_6sensors * sizeof(vr6sensor *));
			vrInputs->sensor6[0] = &(vrInputs->input_devices[0]->sensor6[0]);
			vrInputs->sensor6[1] = &(vrInputs->input_devices[0]->sensor6[1]);
			vrDbgPrintfN(SELDOM_DBGLVL, "vrInputs->sensor6[0] is at %#p -> %#p\n",
				&vrInputs->sensor6[0], vrInputs->sensor6[0]);
			vrDbgPrintfN(SELDOM_DBGLVL, "vrInputs->sensor6[1] is at %#p -> %#p\n",
				&vrInputs->sensor6[1], vrInputs->sensor6[1]);
			/* magellan has a third sensor6 (unused for now) */
			if (vrInputs->num_6sensors >= 3) {
				vrInputs->sensor6[2] = &(vrInputs->input_devices[0]->sensor6[2]);
				vrDbgPrintfN(SELDOM_DBGLVL, "vrInputs->sensor6[2] is at %#p -> %#p\n",
					&vrInputs->sensor6[2], vrInputs->sensor6[2]);
			}
		}

		/* assign user zero's head to 6-sensor 0 */
		vrInputs->users[0]->head = vrInputs->sensor6[0];


	} else {
		vrErrPrintf("vrCreateInputMap(): " RED_TEXT "input map other than 'default' specified -- don't know what to do!\n" NORM_TEXT);
	}

	vrTrace("vrCreateInputMap():", BOLD_TEXT "exiting." NORM_TEXT);
}


/******************************************************************************/
/* vrInputAddDummyToInputMap(): The intent of this function is to allow the   */
/*   system the ability to add new inputs that the application program has    */
/*   requested, but that don't exist in the configuration.  This way at least */
/*   things can function in a limited fashion.                                */
/*                                                                            */
/* TODO: I kind of put this together in a hurry, so it probably could use     */
/*   some rethinking and redesign.  [05/24/2006]                              */
vrGenericInput *vrInputAddDummyToInputMap(vrInputInfo *inputs, vrInputType type, int input_num)
{
	vrGenericInput	*input = NULL;
	vrGenericInput	**old_list;
	int		count;
	vrInput		*input_object;		/* a dummy input object */
#if 0
static	vrInput 	*dummy_list;		/* the linked list of all the created input objects */
#endif
static	char		buffer[512];		/* a place to snprintf into */

	switch (type) {
	case VRINPUT_BINARY:
		if (input_num >= inputs->num_2ways) {
			/* If we get here, then we need to extend the number of 2-way inputs */
			/* NOTE: if this condition isn't true, then we've previously extended the */
			/*   number of inputs, leaving some holes to be filled.                   */
			old_list = (vrGenericInput **)inputs->switch2;
			inputs->switch2 = (vr2switch **)vrShmemAlloc0((input_num+1) * sizeof(vr2switch *));
			for (count = 0; count < inputs->num_2ways; count++) {
				inputs->switch2[count] = ((vr2switch **)old_list)[count];
			}
			inputs->num_2ways = input_num + 1;
		}

		if (inputs->switch2[input_num] != NULL) {
			/* If we get here, there is an error in the library! */
			vrDbgPrintfN(ALWAYS_DBGLVL, "vrInputAddDummyToInputMap(): " RED_TEXT "Library Error: Created a new input that is non-NULL!\n" NORM_TEXT);
		}

		/* make the dummy input data container */
		input = vrInputCreateDataContainerArrayOfType(type, 1, _MakeDummyInputDevice());
		inputs->switch2[input_num] = (vr2switch *)input;

		snprintf(buffer, sizeof(buffer), "switch2(dummy[%d])", input_num);

		break;

	case VRINPUT_NWAY:
		if (input_num >= inputs->num_Nways) {
			/* If we get here, then we need to extend the number of N-way inputs */
			/* NOTE: if this condition isn't true, then we've previously extended the */
			/*   number of inputs, leaving some holes to be filled.                   */
			old_list = (vrGenericInput **)inputs->switchN;
			inputs->switchN = (vrNswitch **)vrShmemAlloc0((input_num+1) * sizeof(vrNswitch *));
			for (count = 0; count < inputs->num_Nways; count++) {
				inputs->switchN[count] = ((vrNswitch **)old_list)[count];
			}
			inputs->num_Nways = input_num + 1;
		}

		if (inputs->switchN[input_num] != NULL) {
			/* If we get here, there is an error in the library! */
			vrDbgPrintfN(ALWAYS_DBGLVL, "vrInputAddDummyToInputMap(): " RED_TEXT "Library Error: Created a new input that is non-NULL!" NORM_TEXT);
		}

		/* make the dummy input data container */
		input = vrInputCreateDataContainerArrayOfType(type, 1, _MakeDummyInputDevice());
		inputs->switchN[input_num] = (vrNswitch *)input;

		snprintf(buffer, sizeof(buffer), "switchN(dummy[%d])", input_num);

		break;

	case VRINPUT_VALUATOR:
		if (input_num >= inputs->num_valuators) {
			/* If we get here, then we need to extend the number of valuator inputs */
			/* NOTE: if this condition isn't true, then we've previously extended the */
			/*   number of inputs, leaving some holes to be filled.                   */
			old_list = (vrGenericInput **)inputs->valuator;
			inputs->valuator = (vrValuator **)vrShmemAlloc0((input_num+1) * sizeof(vrValuator *));
			for (count = 0; count < inputs->num_valuators; count++) {
				inputs->valuator[count] = ((vrValuator **)old_list)[count];
			}
			inputs->num_valuators = input_num + 1;
		}

		if (inputs->valuator[input_num] != NULL) {
			/* If we get here, there is an error in the library! */
			vrDbgPrintfN(ALWAYS_DBGLVL, "vrInputAddDummyToInputMap(): " RED_TEXT "Library Error: Created a new input that is non-NULL!" NORM_TEXT);
		}

		/* make the dummy input data container */
		input = vrInputCreateDataContainerArrayOfType(type, 1, _MakeDummyInputDevice());
		inputs->valuator[input_num] = (vrValuator *)input;

		snprintf(buffer, sizeof(buffer), "valuator(dummy[%d])", input_num);

		break;

	case VRINPUT_6SENSOR:
		if (input_num >= inputs->num_6sensors) {
			/* If we get here, then we need to extend the number of 6-sensor inputs   */
			/* NOTE: if this condition isn't true, then we've previously extended the */
			/*   number of inputs, leaving some holes to be filled.                   */
			old_list = (vrGenericInput **)inputs->sensor6;
			inputs->sensor6 = (vr6sensor **)vrShmemAlloc0((input_num+1) * sizeof(vr6sensor *));
			for (count = 0; count < inputs->num_6sensors; count++) {
				inputs->sensor6[count] = ((vr6sensor **)old_list)[count];
			}
			inputs->num_6sensors = input_num + 1;
		}

		if (inputs->sensor6[input_num] != NULL) {
			/* If we get here, there is an error in the library! */
			vrDbgPrintfN(ALWAYS_DBGLVL, "vrInputAddDummyToInputMap(): " RED_TEXT "Library Error: Created a new input that is non-NULL!" NORM_TEXT);
		}

		/* make the dummy input data container */
		input = vrInputCreateDataContainerArrayOfType(type, 1, _MakeDummyInputDevice());
		inputs->sensor6[input_num] = (vr6sensor *)input;

		snprintf(buffer, sizeof(buffer), "sensor6(dummy[%d])", input_num);

		break;

	case VRINPUT_NSENSOR:
		if (input_num >= inputs->num_Nsensors) {
			/* If we get here, then we need to extend the number of N-sensor inputs   */
			/* NOTE: if this condition isn't true, then we've previously extended the */
			/*   number of inputs, leaving some holes to be filled.                   */
			old_list = (vrGenericInput **)inputs->sensorN;
			inputs->sensorN = (vrNsensor **)vrShmemAlloc0((input_num+1) * sizeof(vrNsensor *));
			for (count = 0; count < inputs->num_Nsensors; count++) {
				inputs->sensorN[count] = ((vrNsensor **)old_list)[count];
			}
			inputs->num_Nsensors = input_num + 1;
		}

		if (inputs->sensorN[input_num] != NULL) {
			/* If we get here, there is an error in the library! */
			vrDbgPrintfN(ALWAYS_DBGLVL, "vrInputAddDummyToInputMap(): " RED_TEXT "Library Error: Created a new input that is non-NULL!" NORM_TEXT);
		}

		/* make the dummy input data container */
		input = vrInputCreateDataContainerArrayOfType(type, 1, _MakeDummyInputDevice());
		inputs->sensorN[input_num] = (vrNsensor *)input;

		snprintf(buffer, sizeof(buffer), "sensorN(dummy[%d])", input_num);

		break;

	default:
		/* Not implemented yet */
		break;
	}

	input->my_object = _MakeDummyInputObject("vrInputAddDummyToInputMap", __LINE__, buffer);
	input->my_object->container.generic = input;
	input->dummy = 1;			/* flag this as being a dummy input */

	return input;
}


/************************************************************************/
/* vrInputInitialize(): initialize the vrContext->input data structure. */
/*   This function sets up the vrContext->input structure that the AP   */
/*   will typically use to access input data.                           */
/*                                                                      */
/* NOTE: it is important to note that this routine is called in the     */
/*   main process, not the input process.  This is done after all the   */
/*   input devices have reported an operating-status.                   */
void vrInputInitialize(vrContextInfo *context)
{
	vrInputInfo	*vrInputs = context->input;
	vrConfigInfo	*vrConfig = context->config;

	vrTrace("vrInputInitialize():", BOLD_TEXT "entering." NORM_TEXT);

	/* Copy user and prop data from the overall configuration structure */
	/*   to the vrInputs data structure.                                */
	/* [04/14/21: is there a compelling reason why we can't just access */
	/*   the user and prop data from the configuration directly?        */
	vrInputs->users = vrConfig->users;
	vrInputs->num_users = vrConfig->num_users;
	if (vrInputs->num_users > 0)
		vrDbgPrintfN(SELDOM_DBGLVL, "vrInputs->num_users = %d, vrInputs->user[0] at %#p\n", vrInputs->num_users, vrInputs->users[0]);
	else	vrDbgPrintfN(SELDOM_DBGLVL, "vrInputs->num_users = %d\n", vrInputs->num_users);

	vrInputs->props = vrConfig->props;
	vrInputs->num_props = vrConfig->num_props;

	vrInputs->input_devices = vrConfig->input_devices;
	vrInputs->num_input_devices = vrConfig->num_input_devices;

	/* Map all the physical inputs and fixed values to the arrays */
	/*   of inputs referred to in the applications here.          */
	vrInputs->input_map_name = vrConfig->input_map_name;
	vrCreateInputMap(vrInputs, vrInputs->input_map_name);

	vrConfig->inputs_init = 1;

	vrTrace("vrInputInitialize():", BOLD_TEXT "exiting." NORM_TEXT);
	return;
}


/**********************************************************************/
/* vrInputsInitialized() returns a boolean that indicates whether all */
/*   the inputs defined in the system have been initialized.          */
int vrInputsInitialized(vrContextInfo *context)
{
	return context->config->inputs_init;
}


/******************************************************************************/
/* vrInputCreateDataContainerArrayOfType(): makes the memory arrangements for */
/*   adding an array of input data containers of the given type to the global */
/*   structure, and returns a pointer to the memory where the first input of  */
/*   that particular type is stored.                                          */
/*     Eg. A magellan will request to add 3 2-way input containers.           */
/* Initial settings and assignments are made for each individual container.   */
/*                                                                            */
/* Returns: a pointer to the first element in the input array.                */
void *vrInputCreateDataContainerArrayOfType(vrInputType type, int num, vrInputDevice *indev)
{
static	char		trace_msg[256];			/* space to create a string for vrTrace */
	int		count;				/* loop counter */
	vrGenericInput	*newinput;
	char		*memptr;
	float		*float_array;			/* memory for allocating Nsensor memory */

	vrTrace("vrInputCreateDataContainerArrayOfType", "beginning");

	/* TODO: this entire section (or each case separately) MUST be    */
	/*   mutex locked so only one change (per case) is being modified */
	/*   at any instance.  (We're safe with just one input proc.)     */
	/* Hmmm, now (12/16/98) I'm not so sure that we need to lock this code. */
	switch (type) {

	/******************/
	case VRINPUT_BINARY:	/* equivalent to VRINPUT_2WAY */
		snprintf(trace_msg, sizeof(trace_msg), "  creating %d 2-switches", num);
		vrTrace("vrInputCreateDataContainerArrayOfType", trace_msg);
		memptr = (char *)vrShmemAlloc0(num * sizeof(vr2switch));
		for (count = 0; count < num; count++) {
			newinput = (vrGenericInput *)(memptr + sizeof(vr2switch) * count);
			newinput->object_type = VROBJECT_INPUTDATA;
			newinput->input_type = VRINPUT_BINARY;
			newinput->my_device = indev;
			newinput->queue_me = 0;
			newinput->dummy = 0;
			newinput->lock = vrLockCreateName(vrContext, "binary input");
			newinput->my_object = _MakeDummyInputObject("vrInputCreateDataContainerArrayOfType", __LINE__, "dummy 2-switch"); /* assign a dummy object for now -- the real object is assigned later, after vrInputCreateDataContainerArrayOfType() completes and returns back to vrInputCreateDataContainers(). */
			newinput->my_object->container.generic = newinput;

			/* allocate the historical measurement array and initialize */
			((vr2switch *)newinput)->num_measures = 50;
			((vr2switch *)newinput)->current_measure = 0;
			((vr2switch *)newinput)->measures = vrShmemAlloc0(((vr2switch *)newinput)->num_measures * sizeof(int));

			/* make the initial assignment, and set last_value the same */
			vrAssign2switchValue((vr2switch *)(newinput), 0 /* , vrTime time */);
			((vr2switch *)newinput)->last_value = ((vr2switch *)newinput)->value;

			newinput->checksum = vrInputMakeChecksum(newinput);

			if (vrDbgDo(INPUT_DBGLVL)) {
				vrPrintf("Created: ");
				vrFprintInputValue(stdout, newinput, verbose);
			}
		}
		break;

	/****************/
	case VRINPUT_NWAY:	/* equivalent to VRINPUT_NARY */
		snprintf(trace_msg, sizeof(trace_msg), "  creating %d N-switches", num);
		vrTrace("vrInputCreateDataContainerArrayOfType", trace_msg);
		memptr = (char *)vrShmemAlloc0(num * sizeof(vrNswitch));
		for (count = 0; count < num; count++) {
			newinput = (vrGenericInput *)(memptr + sizeof(vrNswitch) * count);
			newinput->object_type = VROBJECT_INPUTDATA;
			newinput->input_type = VRINPUT_NWAY;
			newinput->my_device = indev;
			newinput->queue_me = 0;
			newinput->dummy = 0;
			newinput->lock = vrLockCreateName(vrContext, "n-ary input");

			/* make the initial assignment, and set last_value the same */
			vrAssignNswitchValue((vrNswitch *)(newinput), 0 /* , vrTime time */);
			((vrNswitch *)newinput)->last_value = ((vrNswitch *)newinput)->value;

			newinput->checksum = vrInputMakeChecksum(newinput);

			if (vrDbgDo(INPUT_DBGLVL)) {
				vrPrintf("Created: ");
				vrFprintInputValue(stdout, newinput, verbose);
			}
		}
		break;

	/*******************/
	case VRINPUT_VALUATOR:
		snprintf(trace_msg, sizeof(trace_msg), "  creating %d Valuators", num);
		vrTrace("vrInputCreateDataContainerArrayOfType", trace_msg);
		memptr = (char *)vrShmemAlloc0(num * sizeof(vrValuator));
		for (count = 0; count < num; count++) {
			newinput = (vrGenericInput *)(memptr + sizeof(vrValuator) * count);
			newinput->object_type = VROBJECT_INPUTDATA;
			newinput->input_type = VRINPUT_VALUATOR;
			newinput->my_device = indev;
			newinput->queue_me = 0;
			newinput->dummy = 0;
			newinput->lock = vrLockCreateName(vrContext, "valuator input");
			newinput->my_object = _MakeDummyInputObject("vrInputCreateDataContainerArrayOfType", __LINE__, "dummy valuator"); /* assign a dummy object for now -- the real object is assigned later, after vrInputCreateDataContainerArrayOfType() completes and returns back to vrInputCreateDataContainers(). */
			newinput->my_object->container.generic = newinput;

			/* allocate the historical measurement array and initialize */
			((vrValuator *)newinput)->num_measures = 50;
			((vrValuator *)newinput)->current_measure = 0;
			((vrValuator *)newinput)->measures = vrShmemAlloc0(((vrValuator *)newinput)->num_measures * sizeof(float));

			/* make the initial assignment, and set last_value the same */
			vrAssignValuatorValue((vrValuator *)(newinput), 0.0 /* , vrTime time */);
			((vrValuator *)newinput)->last_value = ((vrValuator *)newinput)->value;

			newinput->checksum = vrInputMakeChecksum(newinput);

			if (vrDbgDo(INPUT_DBGLVL)) {
				vrPrintf("Created: ");
				vrFprintInputValue(stdout, newinput, verbose);
			}
		}
		break;

	/*******************/
	case VRINPUT_6SENSOR:
		snprintf(trace_msg, sizeof(trace_msg), "  creating %d 6-sensors", num);
		vrTrace("vrInputCreateDataContainerArrayOfType", trace_msg);
//vrPrintf("For dev '%s': %s\n", indev->name, trace_msg);	/* temporary: delete me */
		memptr = (char *)vrShmemAlloc0(num * sizeof(vr6sensor));
		for (count = 0; count < num; count++) {
			newinput = (vrGenericInput *)(memptr + sizeof(vr6sensor) * count);
			newinput->object_type = VROBJECT_INPUTDATA;
			newinput->input_type = VRINPUT_6SENSOR;
			newinput->my_device = indev;
			newinput->queue_me = 0;
			newinput->dummy = 0;
			newinput->lock = vrLockCreateName(vrContext, "6sensor input");
#if defined(LOCK_DBG) || 0
vrPrintf("Just created lock %p for sensor6 %p\n", newinput->lock, newinput);
#endif

			((vr6sensor *)newinput)->dof = 9;
			((vr6sensor *)newinput)->raw_data = vrMatrixCreateIdentity();
			((vr6sensor *)newinput)->r2e_xform = vrMatrixCreateIdentity();
			((vr6sensor *)newinput)->ipos_xform = vrMatrixCreateIdentity();
			((vr6sensor *)newinput)->position = vrMatrixCreateIdentity();
			((vr6sensor *)newinput)->last_position = vrMatrixCreateIdentity();
			((vr6sensor *)newinput)->visren_position = vrMatrixCreateIdentity();
			if (indev == NULL)
				((vr6sensor *)newinput)->t2rw_xform = vrMatrixCreateIdentity();
			else	((vr6sensor *)newinput)->t2rw_xform = indev->t2rw_xform;

			/* make the initial assignment, and set last_value the same */
			vrAssign6sensorValue((vr6sensor *)newinput, vrMatrixCreateIdentity(), 0 /* , vrTime time */);
			vrMatrixCopy(((vr6sensor *)newinput)->last_position, ((vr6sensor *)newinput)->position);

			newinput->checksum = vrInputMakeChecksum(newinput);

			if (vrDbgDo(INPUT_DBGLVL)) {
				vrPrintf("Created: ");
				vrFprintInputValue(stdout, newinput, verbose);
			}
		}
		break;

	/*******************/
	case VRINPUT_NSENSOR:
		snprintf(trace_msg, sizeof(trace_msg), "  creating %d N-sensors", num);
		vrTrace("vrInputCreateDataContainerArrayOfType", trace_msg);
		memptr = (char *)vrShmemAlloc0(num * sizeof(vrNsensor));
		for (count = 0; count < num; count++) {
			newinput = (vrGenericInput *)(memptr + sizeof(vrNsensor) * count);
			newinput->object_type = VROBJECT_INPUTDATA;
			newinput->input_type = VRINPUT_NSENSOR;
			newinput->my_device = indev;
			newinput->queue_me = 0;
			newinput->dummy = 0;
			newinput->lock = vrLockCreateName(vrContext, "n-sensor input");

			((vrNsensor *)newinput)->dof = MAX_NSENSOR_VALUES;	/* TODO: this needs to be set somewhere */
			float_array = vrShmemAlloc0(MAX_NSENSOR_VALUES * sizeof(float));

			/* make the initial assignment, and set last_value the same */
			vrAssignNsensorArray((vrNsensor *)newinput, float_array /* , vrTime time */);
			memcpy(((vrNsensor *)newinput)->values, ((vrNsensor *)newinput)->last_values, MAX_NSENSOR_VALUES * sizeof(((vrNsensor *)newinput)->values[0]));

			newinput->checksum = vrInputMakeChecksum(newinput);

			if (vrDbgDo(INPUT_DBGLVL)) {
				vrPrintf("Created: ");
				vrFprintInputValue(stdout, newinput, verbose);
			}
		}
		break;

	/*******************/
	case VRINPUT_CONTROL:
		snprintf(trace_msg, sizeof(trace_msg), "  creating %d Controls", num);
		vrTrace("vrInputCreateDataContainerArrayOfType", trace_msg);
//vrPrintf("For dev '%s': %s\n", indev->name, trace_msg);	/* temporary: delete me */
		memptr = (char *)vrShmemAlloc0(num * sizeof(vrControl));
		for (count = 0; count < num; count++) {
			newinput = (vrGenericInput *)(memptr + sizeof(vrControl) * count);
			newinput->object_type = VROBJECT_INPUTDATA;
			newinput->input_type = VRINPUT_CONTROL;
			newinput->my_device = indev;
			newinput->queue_me = 0;
			newinput->dummy = 0;
			newinput->lock = vrLockCreateName(vrContext, "control input");

			/* assign the Null-callback until changed by the input device */
			((vrControl *)newinput)->callback_assigned = 0;
			((vrControl *)newinput)->callback = vrCallbackCreateNamed("CreateType:Control-DN", vrDoNothing, 0);

			newinput->checksum = vrInputMakeChecksum(newinput);

			if (vrDbgDo(INPUT_DBGLVL)) {
				vrPrintf("Created: ");
				vrFprintInputValue(stdout, newinput, verbose);
			}
		}
		break;

	/*********************/
	case VRINPUT_KEYSTROKE:
		/* TODO: implement VRINPUT_KEYSTROKE section */
		break;

	/****************/
	case VRINPUT_TEXT:
		/* TODO: implement VRINPUT_TEXT section */
		break;

	/********************/
	case VRINPUT_POSITION:	/* equivalent to VRINPUT_POINTER */
		/* TODO: implement VRINPUT_POSITION section */
		break;

	/*****************/
	case VRINPUT_PLANE:
		/* TODO: implement VRINPUT_PLANE section */
		break;

	/******/
	default:
		vrDbgPrintf("Unknown input type\n");
	}

	vrTrace("vrInputCreateDataContainerArrayOfType", "returning");

	return (void *)memptr;
}


/***************************************************************************/
/* vrInputCountDataContainers(): uses the linked list of input-objects for */
/*   the given device to count the inputs.                                 */
/***************************************************************************/
void vrInputCountDataContainers(vrInputDevice *devinfo)
{
	vrInput		*input_object;

	for (input_object = devinfo->inputs; input_object != NULL; input_object = (vrInput *)(input_object->next)) {
		input_object->desc = vrParseInputDescription(input_object->desc_str);
		vrTrace("vrInputCountDataContainers", input_object->desc_str);
		switch (input_object->desc->type) {
		case VRINPUT_BINARY:
			devinfo->num_2ways++;
			break;
		case VRINPUT_NWAY:
			devinfo->num_Nways++;
			break;
		case VRINPUT_VALUATOR:
			devinfo->num_valuators++;
			break;
		case VRINPUT_6SENSOR:
			devinfo->num_6sensors++;
			break;
		case VRINPUT_NSENSOR:
			devinfo->num_Nsensors++;
			break;
		case VRINPUT_CONTROL:
			devinfo->num_controls++;
			break;

		case VRINPUT_KEYSTROKE:
		case VRINPUT_TEXT:
		case VRINPUT_POSITION:
		case VRINPUT_PLANE:
		default:
			vrTrace("vrInputCountDataContainers", "encountered unimplemented input type");
			vrErrPrintf(RED_TEXT "vrInputCountDataContainers(): encountered not-yet-implemented input type: %d('%s').\n" NORM_TEXT, input_object->desc->type, input_object->desc_str);
			break;

		}
	}

	for (input_object = devinfo->self_controls; input_object != NULL; input_object = (vrInput *)input_object->next) {
		devinfo->num_scontrols++;
	}

	devinfo->counted = 1;

	vrDbgPrintfN(INPUT_DBGLVL, "vrInputCountDataContainers(): " RED_TEXT "Info: just counted inputs for '%s' -- (2:%d, N:%d, V:%d, 6:%d, N:%d, C:%d, SC:%d)\n" NORM_TEXT,
		devinfo->name,
		devinfo->num_2ways,
		devinfo->num_Nways,
		devinfo->num_valuators,
		devinfo->num_6sensors,
		devinfo->num_Nsensors,
		devinfo->num_controls,
		devinfo->num_scontrols);

	vrTrace("vrInputCountDataContainers", "inputs counted");
}


/*****************************************************************************/
/* vrInputCreateDataContainers(): uses the linked list of input-objects for  */
/*   the given device to:                                                    */
/*     1) count the input types for this device;                             */
/*     2) create arrays for each input type;                                 */
/*     Loop over the linked list to:                                         */
/*       3) assign the input's to an ordered list of particular input types  */
/*          for that device;                                                 */
/*       4) parse the DTI fields of the input; and then                      */
/*       5) go through the list of possible input functions available for    */
/*          the given input-device finding the match (if there is one), and  */
/*          then call the associated function (infunc->func) to initialize   */
/*          that input.                                                      */
/*        The "list of possible input functions" comes from the input device */
/*          by way of the "input_functions" argument.                        */
/*                                                                           */
/* For Step 5: the list/array of vrInputFunction pointers is provided by each*/
/*   input device definition at the top of the _<DEVICE>CreateFunction()     */
/*   function as a "List of possible inputs".                                */
/*                                                                           */
/*   And the return value (for the "list of possible input function) is one  */
/*      of:                                                                  */
/*      VRINPUT_MATCH_ABLE -- a match was found and assigned;                */
/*      VRINPUT_MATCH_UNABLE -- a match for the sub-device was found, but no */
/*         matching input for that device (the search stops at this point);  */
/*      VRINPUT_NOMATCH -- the function was found, but had no matching       */
/*         sub-device, thus the next function for the given type can be      */
/*         tested.  (For some devices, there is only one possible function   */
/*         for a given type.)                                                */
void vrInputCreateDataContainers(vrInputDevice *devinfo, vrInputFunction *input_functions)
{
	vrInput		*input_object;
	vrGenericInput	*input;
	vrInputFunction	*infunc;
	vrInputDTI	*dti;			/* the device-type-instance information of an input description */
	int		mapped;			/* flag for indicating if a successful map found for a declared input */
	int		map_count = 0;		/* count the number of mappings (NOTE: includes unable-to-handle matches) */
	int		num_2ways = 0;
	int		num_Nways = 0;
	int		num_valuators = 0;
	int		num_6sensors = 0;
	int		num_Nsensors = 0;
	int		num_controls = 0;

#if 0 /* TODO: decide which is better -- or spend effort to concatenate strings */
	vrTrace("vrInputCreateDataContainers", "beginning");
#else
	vrTrace("vrInputCreateDataContainers", devinfo->name);
#endif

	/********************************************************/
	/* step 1: First count the input types for this device. */
	if (!devinfo->counted)
		vrInputCountDataContainers(devinfo);

	/*****************************************************/
	/* step 2: Now create the input arrays for each type */
	devinfo->switch2  = (vr2switch  *)vrInputCreateDataContainerArrayOfType(VRINPUT_BINARY, devinfo->num_2ways, devinfo);
	devinfo->switchN  = (vrNswitch  *)vrInputCreateDataContainerArrayOfType(VRINPUT_NWAY, devinfo->num_Nways, devinfo);
	devinfo->valuator = (vrValuator *)vrInputCreateDataContainerArrayOfType(VRINPUT_VALUATOR, devinfo->num_valuators, devinfo);
	devinfo->sensor6  = (vr6sensor  *)vrInputCreateDataContainerArrayOfType(VRINPUT_6SENSOR, devinfo->num_6sensors, devinfo);
	devinfo->sensorN  = (vrNsensor  *)vrInputCreateDataContainerArrayOfType(VRINPUT_NSENSOR, devinfo->num_Nsensors, devinfo);
	devinfo->control  = (vrControl  *)vrInputCreateDataContainerArrayOfType(VRINPUT_CONTROL, devinfo->num_controls, devinfo);

	vrDbgPrintfN(INPUT_DBGLVL FORCE_DBG, "vrInputCreateDataContainers(): " RED_TEXT "Info: just created inputs for '%s' -- (2:%d, N:%d, V:%d, 6:%d, N:%d, C:%d, SC:%d)\n" NORM_TEXT,
		devinfo->name,
		devinfo->num_2ways,
		devinfo->num_Nways,
		devinfo->num_valuators,
		devinfo->num_6sensors,
		devinfo->num_Nsensors,
		devinfo->num_controls,
		devinfo->num_scontrols);

	/***************************************************************/
	/* Now initialize the input types for this device:             */
	/*   * step 3: For each input config object in the list assign */
	/*     it to an input of the particular type.                  */
	/*   * step 4: parse the DTI string.                           */
	/*   * step 5: Find a matching input mapping function based on */
	/*     the input type, and type name in the list of functions  */
	/*     given by "input_functions".                             */
	for (input_object = devinfo->inputs; input_object != NULL; input_object = (vrInput *)(input_object->next)) {
		/*******************************************************/
		/* Step 3: assign each input to the array of it's type */
		switch (input_object->desc->type) {
		case VRINPUT_BINARY:
			input = (vrGenericInput *)&devinfo->switch2[num_2ways];
			num_2ways++;
			break;

		case VRINPUT_NWAY:
			input = (vrGenericInput *)&devinfo->switchN[num_Nways];
			num_Nways++;
			break;

		case VRINPUT_VALUATOR:
			input = (vrGenericInput *)&devinfo->valuator[num_valuators];
			num_valuators++;
			break;

		case VRINPUT_6SENSOR:
			input = (vrGenericInput *)&devinfo->sensor6[num_6sensors];
			num_6sensors++;
			break;

		case VRINPUT_NSENSOR:
			input = (vrGenericInput *)&devinfo->sensorN[num_Nsensors];
			num_Nsensors++;
			break;

		case VRINPUT_CONTROL:
			input = (vrGenericInput *)&devinfo->control[num_controls];
			num_controls++;
			break;


		case VRINPUT_KEYSTROKE:
		case VRINPUT_TEXT:
		case VRINPUT_POSITION:
		case VRINPUT_PLANE:
		default:
			//vrErrPrintf(RED_TEXT "vrInputCreateDataContainers(): encountered not-yet-implemented input type.\n" NORM_TEXT);
			vrErrPrintf(RED_TEXT "vrInputCreateDataContainers(): encountered not-yet-implemented input type: %d('%s').\n" NORM_TEXT, input_object->desc->type, input_object->desc_str);
			break;
		}

		input->my_object = input_object;
		input_object->desc_ui[0] = '\0';
		input_object->container.generic = input;

		/********************************************************************************/
		/* Step 4: parse the input arguments from the format: '<device>:type[instance]' */
vrDbgPrintfN(INPUT_DBGLVL, "vrInputCreateDataContainers(): top of step 4\n");
		dti = vrParseInputDTI(input_object, input_object->desc->args);

#if 1 /* I'm including this because sometimes dti is corrupted, and I don't know why! [09/05/13] */
		while (dti == NULL || dti->device == NULL || dti->type == NULL || dti->instance == NULL) {
			vrErrPrintf(RED_TEXT "vrInputCreateDataContainers(): Parsing of the device type returned a corrupted value!  Trying again.\n" NORM_TEXT);
			dti = vrParseInputDTI(input_object, input_object->desc->args);
		}
#endif
		vrDbgPrintfN(INPUT_DBGLVL, "vrInputCreateDataContainers(): parsed dti as device:'%s', type:'%s', instance:'%s'\n", dti->device, dti->type, dti->instance);

		/*****************************************************************************************/
		/* Step 5: find a matching input by looping through the list of possible input functions */
		mapped = 0;
		for (infunc = input_functions; infunc->name != NULL && !mapped; infunc++) {
			vrDbgPrintfN(INPUT_DBGLVL, "vrInputCreateDataContainers(): trying to match input '%s' with input func '%s' (%s)\n",
				input_object->desc->args, infunc->name, vrInputTypeName(infunc->type));
			if (infunc->type == input->input_type && !strcasecmp(dti->type, infunc->name)) {
				vrDbgPrintfN(INPUT_DBGLVL, "vrInputCreateDataContainers(): ... possible match -- type = '%s' (%s) ... \n\t(B)",
					dti->type, vrInputTypeName(input->input_type));
				switch ((infunc->func)(devinfo, input, dti)) {
				case VRINPUT_NOMATCH:
					vrDbgPrintfN(INPUT_DBGLVL, "vrInputCreateDataContainers(): not a good match, try again\n");
					break;

				case VRINPUT_MATCH_UNABLE:	/* matches, but unable to handle request */
					/* We'll just march on without this input then, and claim it's mapped. */
					vrDbgPrintfN(INPUT_DBGLVL, "vrInputCreateDataContainers(): can't handle this ... but pretending to be a -- "/* "successful match" */);

				case VRINPUT_MATCH_ABLE:	/* matches, and able to handle request */
					mapped = 1;
					map_count++;
					vrDbgPrintfN(INPUT_DBGLVL, "vrInputCreateDataContainers(): successful match\n");
					break;
				}
			} else {
				vrDbgPrintfN(INPUT_DBGLVL, "vrInputCreateDataContainers(): type mismatch '%s' != '%s'\n", dti->type, vrInputTypeName(input->input_type));
			}
		}

		if (!mapped)
			vrDbgPrintfN(AALWAYS_DBGLVL, "vrInputCreateDataContainers(): "
				RED_TEXT "Warning, no mapping found for device '%s' input '%s', a %s of type '%s'\n" NORM_TEXT,
				devinfo->name, input_object->name, vrInputTypeName(input->input_type), input_object->desc->args);
	}
	vrDbgPrintfN(INPUT_DBGLVL, "vrInputCreateDataContainers(): " RED_TEXT "Info: just mapped %d inputs for '%s'\n" NORM_TEXT,
		map_count, devinfo->name);

	vrTrace("vrInputCreateDataContainers", "ending");
}


/*****************************************************************************/
/* vrDeleteInputs(): free the memory of all the inputs in the FreeVR system. */
/*   Basically loop over all input types, freeing each one.                  */
/*                                                                           */
/* NOTE: currently this function is only referenced in unused code, so at    */
/*   some point (TODO) it needs to be tested.                                */
void vrDeleteInputs(vrInputDevice *devinfo)
{
	int		count;
	vrGenericInput	*input;

	/* Loop through each type of input for the device delete the lock, and then the input. */
	for (count = 0; count < devinfo->num_2ways; count++) {
		input = (vrGenericInput *)&devinfo->switch2[count];
		vrLockFree(input->lock);
		vrShmemFree(input);
	}
	for (count = 0; count < devinfo->num_Nways; count++) {
		input = (vrGenericInput *)&devinfo->switchN[count];
		vrLockFree(input->lock);
		vrShmemFree(input);
	}
	for (count = 0; count < devinfo->num_valuators; count++) {
		input = (vrGenericInput *)&devinfo->valuator[count];
		vrLockFree(input->lock);
		vrShmemFree(input);
	}
	for (count = 0; count < devinfo->num_6sensors; count++) {
		input = (vrGenericInput *)&devinfo->sensor6[count];
		vrLockFree(input->lock);
		vrShmemFree(input);
	}
	for (count = 0; count < devinfo->num_Nsensors; count++) {
		input = (vrGenericInput *)&devinfo->sensorN[count];
		vrLockFree(input->lock);
		vrShmemFree(input);
	}
	for (count = 0; count < devinfo->num_controls; count++) {
		input = (vrGenericInput *)&devinfo->control[count];
		vrLockFree(input->lock);
		vrShmemFree(input);
	}
}


/**********************************************************************/
/* vrInputDeviceInList() returns a boolean value of whether the given */
/*   input device (provided as a string name) exists in this compiled */
/*   instance of FreeVR.  For example checking whether "xwindows" is  */
/*   present as a valid input device.                                 */
/* This is primarily called by _ConfigFunctionExistsInputDeviceType() */
/*   which is a configuration function available in the "freevrrc"    */
/*   and related configuration files.                                 */
int vrInputDeviceInList(const char *device)
{
	vrInputOptsType	*option;	/* for looping through the list of possible input devices */

	/* Loop through the input device options specified in vr_input.opts.h     */
	/*   until a string match is made with the "device" string, then call the */
	/*   associated initialization function to fill in the info structure.    */
	option = vrInputOpts;
	while (option->option_name != NULL) {
		if (!strcasecmp(option->option_name, device)) {
			return (1);
		}
		option++;
	}

	return (0);
}


/**********************************************************************/
/* vrGetFuncFromControlList(): takes a "vrControlFunc", which is just */
/*   an array of pairwise mappings between string names and control   */
/*   callback functions (aka self-controls).  These self-controls are */
/*   defined by each input device and pertain just to that device,    */
/*   though most input devices have nearly the same list of self-     */
/*   controls, especially if they simulate 6-sensor inputs.           */
/* Thus this function takes a string (funcname) and an array of       */
/*   mappings and returns a pointer to the desired function.          */
void *vrGetFuncFromControlList(const char *funcname, vrControlFunc *funclist)
{
	vrControlFunc	*cfunc;

	for (cfunc = funclist; cfunc->name != NULL; cfunc++) {
		if (!strcasecmp(funcname, cfunc->name))
			return (void *) cfunc->func;
	}

	return NULL;
}


/******************************************************************************/
/* vrInputCreateSelfControlContainers(): uses the linked list of self-controls*/
/*   listed with the input object of the given device to create and setup     */
/*   each self-control, assign initial values, and then ... (fill this in)    */
/*   using the list of Input Functions for each                               */
/*                                                                            */
/*     Loop over the linked list of this device's self-controls to:           */
/*       1) parse the self-control's input configuration string;              */
/*       2) assign the self-control's callback function;                      */
/*       3) parse the DTI fields of the input; and                            */
/*       4) go through the list of possible input functions available for     */
/*          the given input-device finding the match (if there is one), and   */
/*          then call the associated function (infunc->func) to initialize    */
/*          that input.                                                       */
/*        The "list of possible input functions" comes from the input device  */
/*          by way of the "input_functions" argument.                         */
/*                                                                            */
/* For Step 4: the list/array of vrInputFunction and vrControlFunc pointers   */
/*   is provided by each input device definition at the top of the            */
/*   _<DEVICE>CreateFunction() function as a "List of possible inputs".       */
/*                                                                            */
/*   And the return value (for the "list of possible input function) is one   */
/*      of:                                                                   */
/*      VRINPUT_MATCH_ABLE -- a match was found and assigned;                 */
/*      VRINPUT_MATCH_UNABLE -- a match for the sub-device was found, but no  */
/*         matching input for that device (the search stops at this point);   */
/*      VRINPUT_NOMATCH -- the function was found, but had no matching        */
/*         sub-device, thus the next function for the given type can be       */
/*         tested.  (For some devices, there is only one possible function    */
/*         for a given type.)                                                 */
void vrInputCreateSelfControlContainers(vrInputDevice *devinfo, vrInputFunction *input_functions, vrControlFunc *funclist)
{
static	char		string[256];		/* temp space for creating strings */
	vrInput		*scontrol;		/* loop variable for stepping through devinfo->self_controls */
	vrInputFunction	*infunc;		/* loop variable for stepping through "input_functions" */
	vrInputDTI	*dti;
	char		*input_args;
	vrInputType	 input_type;
	int		 mapped;		/* flag for indicating if a successful map found for a declared input */
	int		 map_count = 0;		/* count the number of mappings (NOTE: includes unable-to-handle matches) */
	void		(*callback_function)();

	/* this part originally from vr_input.magellan.c */

	/* initialize all the input device (self) controls */
	for (scontrol = devinfo->self_controls; scontrol != NULL; scontrol = (vrInput *)scontrol->next) {
		/***************************************************************/
		/* Step 1: parse the self-control's input configuration string */
		scontrol->desc = vrParseInputDescription(scontrol->desc_str);
		input_type = scontrol->desc->type;
		input_args = scontrol->desc->args;
		vrDbgPrintfN(DEFAULT_DBGLVL, "Handling '%s' self control '%s' with '%s'\n",
			devinfo->name, scontrol->name, input_args);

		scontrol->container.control = (vrControl *)vrInputCreateDataContainerArrayOfType(VRINPUT_CONTROL, 1, devinfo);
		scontrol->container.control->my_object = scontrol;

		/***************************************/
		/* Step 2: assign the control callback */
		callback_function = (void (*)())vrGetFuncFromControlList(scontrol->name, funclist);
		if (callback_function == NULL) {
			/* No function matching the name of the self-control found */
			scontrol->container.control->callback_assigned = -1;
			vrDbgPrintfN(AALWAYS_DBGLVL, "vrInputCreateSelfControlContainers(): "
				RED_TEXT "Warning, no callback function ('%s') found for device '%s' input '%s'\n" NORM_TEXT,
				scontrol->name, devinfo->name, scontrol->desc->args);
			/* scontrol->container.control->callback was set to a Do-nothing callback when created */
		} else {
			/* Matching function found, so assign it as a callback */
			snprintf(string, sizeof(string), "SelfControl: '%s' of device '%s", scontrol->name, devinfo->name);
			scontrol->container.control->callback = vrCallbackCreateNamed(string, callback_function, 1, devinfo);
			scontrol->container.control->callback_assigned =  1;
		}

		/********************************************************************************/
		/* Step 3: parse the input arguments from the format: '<device>:type[instance]' */
		dti = vrParseInputDTI(scontrol, input_args);

		/* this code from vr_input.c::vrInputCreateDataContainers() */

		/*****************************************************************************************/
		/* Step 4: find a matching input by looping through the list of possible input functions */
		mapped = 0;
		for (infunc = input_functions; infunc->name != NULL && !mapped; infunc++) {
			vrDbgPrintfN(INPUT_DBGLVL, "vrInputCreateSelfControlContainers(): trying to match input '%s' with input func '%s' (%d)\n",
				scontrol->desc->args, infunc->name, infunc->type);
			if (infunc->type == input_type && !strcasecmp(dti->type, infunc->name)) {
				vrDbgPrintfN(INPUT_DBGLVL, "vrInputCreateSelfControlContainers(): ... possible match -- type = '%s' (%d) ... \n\t(A)",
					dti->type, input_type);
				switch ((infunc->func)(devinfo, scontrol->container.generic, dti)) {
				case VRINPUT_NOMATCH:
					vrDbgPrintfN(INPUT_DBGLVL, "vrInputCreateSelfControlContainers(): not a good match, try again\n");
					break;

				case VRINPUT_MATCH_UNABLE:	/* matches, but unable to handle request */
					/* We'll just march on without this input then, and claim it's mapped. */
					vrDbgPrintfN(INPUT_DBGLVL, "vrInputCreateSelfControlContainers(): can't handle this ... ");

				case VRINPUT_MATCH_ABLE:	/* matches, and able to handle request */
					mapped = 1;
					map_count++;
					vrDbgPrintfN(INPUT_DBGLVL, "vrInputCreateSelfControlContainers(): successful match\n");
					break;
				}
			}
		}
		if (!mapped)
			vrDbgPrintfN(AALWAYS_DBGLVL, "vrInputCreateSelfControlContainers(): "
				RED_TEXT "Warning, no mapping found for device '%s' input '%s'\n" NORM_TEXT,
				devinfo->name, scontrol->desc->args);
	}
	vrDbgPrintfN(SELFCTRL_DBGLVL, "vrInputCreateSelfControlContainers(): " RED_TEXT "Info: just mapped %d self-controls for '%s'\n" NORM_TEXT,
		map_count, devinfo->name);
}


/******************************************************************************/
/* vrFprintInputValue(): This prints the actual low-level input structure     */
/*   which can be as simple as a boolean value, to a 6-sensor with all the    */
/*   settings used in configuring it (vr2switch, vrValuator, vr6sensor, etc.).*/
void vrFprintInputValue(FILE *file, vrGenericInput *input, vrPrintStyle style)
{
	vrMatrix	*tmpmat;	/* pointer to matrix to be printed */

	if (input == NULL) {
		vrFprintf(file, "\tInput Value = { <nil> }\n");
		return;
	}

	switch (style) {
	case none:
		break;

	case brief:
	case one_line:
		vrFprintf(file, "Input at %#p, object_name = '%s', ",
			input,
			(input->my_object != NULL ? input->my_object->name : "-"));
		switch(input->input_type) {

		case VRINPUT_BINARY:	/* equivalent to VRINPUT_2WAY */
			vrFprintf(file, "value = %d", ((vr2switch *)input)->value);
			break;
		case VRINPUT_NWAY:	/* equivalent to VRINPUT_NARY */
			vrFprintf(file, " NYI");	/* TODO: implement this */
			break;
		case VRINPUT_VALUATOR:
			vrFprintf(file, "value = %.2f", ((vrValuator *)input)->value);
			break;
		case VRINPUT_6SENSOR:
			tmpmat = ((vr6sensor *)input)->position;
			vrFprintf(file, "active = %d, oob = %d, error = %d, position = (%3.1f %3.1f %3.1f %3.1f/%3.1f %3.1f %3.1f %3.1f/%3.1f %3.1f %3.1f %3.1f/%3.1f %3.1f %3.1f %3.1f)",
				((vr6sensor *)input)->active,
				((vr6sensor *)input)->oob,
				((vr6sensor *)input)->error,
				tmpmat->v[ 0], tmpmat->v[ 4], tmpmat->v[ 8], tmpmat->v[12],
				tmpmat->v[ 1], tmpmat->v[ 5], tmpmat->v[ 9], tmpmat->v[13],
				tmpmat->v[ 2], tmpmat->v[ 6], tmpmat->v[10], tmpmat->v[14],
				tmpmat->v[ 3], tmpmat->v[ 7], tmpmat->v[11], tmpmat->v[15]);

			/* For "brief" style, add a 2nd line with the matrix decomposition. */
			if (style == brief) {
				vrFprintMatrixCompositionOL(file, "\n\tpos: ", tmpmat);
				tmpmat = ((vr6sensor *)input)->r2e_xform;
				vrFprintf(file, "\tr2e = (%3.1lf %3.1lf %3.1lf %3.1lf/%3.1lf %3.1lf %3.1lf %3.1lf/%3.1lf %3.1lf %3.1lf %3.1lf/%3.1lf %3.1lf %3.1lf %3.1lf)",
					tmpmat->v[ 0], tmpmat->v[ 4], tmpmat->v[ 8], tmpmat->v[12],
					tmpmat->v[ 1], tmpmat->v[ 5], tmpmat->v[ 9], tmpmat->v[13],
					tmpmat->v[ 2], tmpmat->v[ 6], tmpmat->v[10], tmpmat->v[14],
					tmpmat->v[ 3], tmpmat->v[ 7], tmpmat->v[11], tmpmat->v[15]);
				vrFprintMatrixCompositionOL(file, "\n\tr2e: ", tmpmat);
				tmpmat = ((vr6sensor *)input)->ipos_xform;
				vrFprintf(file, "\tipos = (%3.1lf %3.1lf %3.1lf %3.1lf/%3.1lf %3.1lf %3.1lf %3.1lf/%3.1lf %3.1lf %3.1lf %3.1lf/%3.1lf %3.1lf %3.1lf %3.1lf)",
					tmpmat->v[ 0], tmpmat->v[ 4], tmpmat->v[ 8], tmpmat->v[12],
					tmpmat->v[ 1], tmpmat->v[ 5], tmpmat->v[ 9], tmpmat->v[13],
					tmpmat->v[ 2], tmpmat->v[ 6], tmpmat->v[10], tmpmat->v[14],
					tmpmat->v[ 3], tmpmat->v[ 7], tmpmat->v[11], tmpmat->v[15]);
				vrFprintMatrixCompositionOL(file, "\n\tipos: ", tmpmat);
			}
			break;
		case VRINPUT_NSENSOR:
			vrFprintf(file, " NYI");	/* TODO: implement this */
			break;
		case VRINPUT_CONTROL:
			vrFprintf(file, "callback_assigned = %d, callback = %#p",
				((vrControl *)input)->callback_assigned,
				((vrControl *)input)->callback);
			break;

		default:
			vrFprintf(file, " NYI");	/* TODO: implement this */
			break;
		}
		vrFprintf(file, "\n");

		break;

	case file_format:
		/* TODO: or at least determine if this means anything -- probably not */
	case machine:
		/* TODO: not yet implemented -- output easily machine parsed */
		vrFprintf(file, "TODO: put special version of input value info here\n");
		break;

	default:
	case verbose:
	case everything:
		vrFprintf(file, "Input at %#p (self) = {\n"
			"\r\tInput_type = %d\n\tmy_device = %#p ('%s')\n"
			"\tmy_object = %#p ('%s')\n"
			"\tlock = %#p\n\tqueue_me = %d\n\tdummy = %d\n\ttimestamp = %.2f\n",
			input,
			input->input_type,
			input->my_device,
			(input->my_device != NULL ? input->my_device->name : "-"),
			input->my_object,
			(input->my_object != NULL ? input->my_object->name : "-"),
			input->lock,
			input->queue_me,
			input->dummy,
			input->timestamp);
#if 1 /* checksums not yet implemented */
		vrFprintf(file, "\tchecksum = %d\n",
			input->checksum);
#endif

		switch(input->input_type) {
		case VRINPUT_BINARY:	/* equivalent to VRINPUT_2WAY */
			vrFprintf(file, "\r"
				"\r\tvalue = %d\n\tlast_value = %d\n\tvisren_value = %d\n",
				((vr2switch *)input)->value,
				((vr2switch *)input)->last_value,
				((vr2switch *)input)->visren_value);
			break;
		case VRINPUT_NWAY:	/* equivalent to VRINPUT_NARY */
			vrFprintf(file, "\r"
				"\r\tvalue = %d\n\tlast_value = %d\n\tvisren_value = %d\n",
				((vrNswitch *)input)->value,
				((vrNswitch *)input)->last_value,
				((vrNswitch *)input)->visren_value);
			break;
		case VRINPUT_VALUATOR:
			vrFprintf(file, "\r"
				"\r\tvalue = %f\n\tlast_value = %f\n\tvisren_value = %f\n",
				((vrValuator *)input)->value,
				((vrValuator *)input)->last_value,
				((vrValuator *)input)->visren_value);
			break;
		case VRINPUT_6SENSOR:
			tmpmat = ((vr6sensor *)input)->position;
			vrFprintf(file, "\r"
				"\r\tdof = %d\n\tactive = %d\n\toob = %d\n\terror = %d\n"
				"\tposition  = [%#p] (%3.1lf %3.1lf %3.1lf %3.1lf/%3.1lf %3.1lf %3.1lf %3.1lf/%3.1lf %3.1lf %3.1lf %3.1lf/%3.1lf %3.1lf %3.1lf %3.1lf)\n",
				((vr6sensor *)input)->dof,
				((vr6sensor *)input)->active,
				((vr6sensor *)input)->oob,
				((vr6sensor *)input)->error,
				tmpmat,
				tmpmat->v[ 0], tmpmat->v[ 4], tmpmat->v[ 8], tmpmat->v[12],
				tmpmat->v[ 1], tmpmat->v[ 5], tmpmat->v[ 9], tmpmat->v[13],
				tmpmat->v[ 2], tmpmat->v[ 6], tmpmat->v[10], tmpmat->v[14],
				tmpmat->v[ 3], tmpmat->v[ 7], tmpmat->v[11], tmpmat->v[15]);
			vrFprintMatrixCompositionOL(file, "\tpos-decomp: ", tmpmat);
			tmpmat = ((vr6sensor *)input)->r2e_xform;
			vrFprintf(file, "\r"
				"\tr2e.xform = [%#p] (%3.1lf %3.1lf %3.1lf %3.1lf/%3.1lf %3.1lf %3.1lf %3.1lf/%3.1lf %3.1lf %3.1lf %3.1lf/%3.1lf %3.1lf %3.1lf %3.1lf)\n",
				tmpmat,
				tmpmat->v[ 0], tmpmat->v[ 4], tmpmat->v[ 8], tmpmat->v[12],
				tmpmat->v[ 1], tmpmat->v[ 5], tmpmat->v[ 9], tmpmat->v[13],
				tmpmat->v[ 2], tmpmat->v[ 6], tmpmat->v[10], tmpmat->v[14],
				tmpmat->v[ 3], tmpmat->v[ 7], tmpmat->v[11], tmpmat->v[15]);
			vrFprintMatrixCompositionOL(file, "\tr2e-decomp: ", tmpmat);
			tmpmat = ((vr6sensor *)input)->ipos_xform;
			vrFprintf(file, "\r"
				"\tipos.xform = [%#p] (%3.1lf %3.1lf %3.1lf %3.1lf/%3.1lf %3.1lf %3.1lf %3.1lf/%3.1lf %3.1lf %3.1lf %3.1lf/%3.1lf %3.1lf %3.1lf %3.1lf)\n",
				tmpmat,
				tmpmat->v[ 0], tmpmat->v[ 4], tmpmat->v[ 8], tmpmat->v[12],
				tmpmat->v[ 1], tmpmat->v[ 5], tmpmat->v[ 9], tmpmat->v[13],
				tmpmat->v[ 2], tmpmat->v[ 6], tmpmat->v[10], tmpmat->v[14],
				tmpmat->v[ 3], tmpmat->v[ 7], tmpmat->v[11], tmpmat->v[15]);
			vrFprintMatrixCompositionOL(file, "\tipos-decomp: ", tmpmat);
			tmpmat = ((vr6sensor *)input)->visren_position;
			vrFprintf(file, "\r"
				"\tvisren_pos= [%#p] (%3.1f %3.1lf %3.1lf %3.1lf/%3.1lf %3.1lf %3.1lf %3.1lf/%3.1lf %3.1lf %3.1lf %3.1lf/%3.1lf %3.1lf %3.1lf %3.1lf)\n",
				tmpmat,
				tmpmat->v[ 0], tmpmat->v[ 4], tmpmat->v[ 8], tmpmat->v[12],
				tmpmat->v[ 1], tmpmat->v[ 5], tmpmat->v[ 9], tmpmat->v[13],
				tmpmat->v[ 2], tmpmat->v[ 6], tmpmat->v[10], tmpmat->v[14],
				tmpmat->v[ 3], tmpmat->v[ 7], tmpmat->v[11], tmpmat->v[15]);
			vrFprintMatrixCompositionOL(file, "\tvis-decomp: ", tmpmat);
			break;
		case VRINPUT_NSENSOR:
			vrFprintf(file, " NYI");	/* TODO: implement this */
			break;
		case VRINPUT_CONTROL:
			vrFprintf(file, "\r"
				"\r\tcallback_assigned = %d\n\tcallback = %#p\n",
				((vrControl *)input)->callback_assigned,
				((vrControl *)input)->callback);
			break;
		default:
			vrFprintf(file, "\r\tvalue = ... [value printing not yet implemented for this type]\n");
			break;
		}
		vrFprintf(file, "\r}\n");

		break;
	}
}


/******************************************************************************/
/* vrFprint6sensorOptions(): This prints the values of the current settings   */
/*   of the vr6sensorConv field of an input device.  Options such as swapping */
/*   the Y and Z directions.                                                  */
/* This function is often called via an input device callback.                */
void vrFprint6sensorOptions(FILE *file, vr6sensorConv *sensor6_options, vrPrintStyle style)
{
	switch (style) {
		case brief:
		vrFprintf(file, "\r"
			"\t6sensor_options = %d, %d, %d, %d, %d, %d, %d, %d, %.2f %.2f (%.2f--%.2f %.2f--%.2f %.2f--%.2f)\n",
			sensor6_options->azimuth_axis,
			sensor6_options->ignore_all,
			sensor6_options->ignore_trans,
			sensor6_options->relative_axis,
			sensor6_options->return_to_zero,
			sensor6_options->restrict_space,
			sensor6_options->swap_yz,
			sensor6_options->swap_transrot,
			sensor6_options->trans_scale,
			sensor6_options->rot_scale,
			sensor6_options->working_volume_min[0],
			sensor6_options->working_volume_max[0],
			sensor6_options->working_volume_min[1],
			sensor6_options->working_volume_max[1],
			sensor6_options->working_volume_min[2],
			sensor6_options->working_volume_max[2]);
		break;

		case verbose:
		default:
		vrFprintf(file, "\t6sensor_options = {\n");
		vrFprintf(file, "\r"
			"\t\tazimuth-axis = '%c'\n"
			"\t\tignore-all = %d\n"
			"\t\tignore-trans = %d\n"
			"\t\trelative-axis = %d\n"
			"\t\treturn-to-zero = %d\n"
			"\t\trestrict-space = %d\n"
			"\t\tswap-yz = %d\n"
			"\t\tswap-transrot = %d\n"
			"\t\ttrans-scale = %.2f\n"
			"\t\trot-scale = %.2f\n"
			"\t\tworking-volume = (%.2f--%.2f %.2f--%.2f %.2f--%.2f)\n",
			(sensor6_options->azimuth_axis == VR_X ? 'X' :
				(sensor6_options->azimuth_axis == VR_Y ? 'Y' :
				(sensor6_options->azimuth_axis == VR_Z ? 'Z' :
				'?'))),
			sensor6_options->ignore_all,
			sensor6_options->ignore_trans,
			sensor6_options->relative_axis,
			sensor6_options->return_to_zero,
			sensor6_options->restrict_space,
			sensor6_options->swap_yz,
			sensor6_options->swap_transrot,
			sensor6_options->trans_scale,
			sensor6_options->rot_scale,
			sensor6_options->working_volume_min[0],
			sensor6_options->working_volume_max[0],
			sensor6_options->working_volume_min[1],
			sensor6_options->working_volume_max[1],
			sensor6_options->working_volume_min[2],
			sensor6_options->working_volume_max[2]);

		vrFprintf(file, "\t}\n");
		break;
	}
}


/**************************************************************************/
/* The vrFprintInput() function prints information about the entire input */
/*   structure of FreeVR -- the vrInputInfo structure, which has pointers */
/*   to all aspects of the running FreeVR input (/output) system.         */
void vrFprintInput(FILE *file, vrInputInfo *vrInputs, vrPrintStyle style)
{
	int		count;

	if (vrInputs == NULL) {
		vrErrPrintf("vrFprintInput(): " RED_TEXT "memory argument is NULL!\n" NORM_TEXT);
		return;
	}

	if (vrInputs->object_type != VROBJECT_INPUTINFO) {
		vrErrPrintf("vrFprintInput(): " RED_TEXT "memory argument does not point to the vrInputInfo structure!\n" NORM_TEXT);
		return;
	}

	switch(style) {
	case none:
		break;

	case brief:
		vrFprintf(file, "==================================================\n");
		vrFprintf(file, "vrInputs at %#p\n", vrInputs);
#if 0 /* TODO: when Queue stuff is implemented */
		vrFprintf(file, "\tQueue at %#p\n", vrInputs->queue);
#endif
		vrFprintf(file, "%d input devices\n", vrInputs->num_input_devices);
		vrFprintf(file, "%d users\n", vrInputs->num_users);
		vrFprintf(file, "%d props\n", vrInputs->num_props);
		vrFprintf(file, "%d 2-way switches\n", vrInputs->num_2ways);
		vrFprintf(file, "%d N-way switches\n", vrInputs->num_Nways);
		vrFprintf(file, "%d valuators\n", vrInputs->num_valuators);
		vrFprintf(file, "%d 6-sensors\n", vrInputs->num_6sensors);
		vrFprintf(file, "%d N-sensors\n", vrInputs->num_Nsensors);
		vrFprintf(file, "%d controls\n", vrInputs->num_controls);
		vrFprintf(file, "==================================================\n");
		break;

	case everything:
		/* Here, this is the same as verbose -- it also means more info */
		/*   during the read phase of configuration.                    */
		/* TODO: decide if it should be more.                           */

	case verbose:
		vrFprintf(file, "==================================================\n");
		vrFprintf(file, "vrInputs at %#p\n", vrInputs);
#if 0 /* TODO: when Queue stuff is implemented */
		vrFprintf(file, "\tQueue at %#p\n", vrInputs->queue);
#endif
		vrFprintf(file, "input map name = \"%s\"\n", vrInputs->input_map_name);

		vrFprintf(file, "%d input devices:\n", vrInputs->num_input_devices);
		for (count = 0; count < vrInputs->num_input_devices; count++) {
			if (vrInputs->input_devices[count] == NULL) {
				vrFprintf(file, "\tdevice[%d] is NULL\n", count);
			} else {
				vrFprintf(file, "\tdevice[%d] at %#p id:%d named: \"%s\" (%s)\n",
					count,
					vrInputs->input_devices[count],
					vrInputs->input_devices[count]->id,
					vrInputs->input_devices[count]->name,
					(vrInputs->input_devices[count]->operating ? "operating" : RED_TEXT "NOT OPERATING" NORM_TEXT) );
			}
		}
		vrFprintf(file, "%d users:\n", vrInputs->num_users);
		for (count = 0; count < vrInputs->num_users; count++) {
			if (vrInputs->users[count] == NULL) {
				vrFprintf(file, "\tuser[%d] is NULL\n", count);
			} else {
				vrFprintf(file, "\tuser[%d] at %#p, named \"%s\", head sensor is \"%s\"\n",
					count,
					vrInputs->users[count],
					vrInputs->users[count]->name,
					(vrInputs->users[count]->head->my_object != NULL ?
						vrInputs->users[count]->head->my_object->name : "-no object-"));
			}
		}
		vrFprintf(file, "%d props:\n", vrInputs->num_props);
		for (count = 0; count < vrInputs->num_props; count++) {
			if (vrInputs->props[count] == NULL) {
				vrFprintf(file, "\tprop[%d] is NULL\n", count);
			} else {
				vrFprintf(file, "\tprop[%d] at %#p, named \"%s\"\n",
					count,
					vrInputs->props[count],
					vrInputs->props[count]->name);
			}
		}
		vrFprintf(file, "%d 2-way switches (at %#p):\n", vrInputs->num_2ways, vrInputs->switch2);
		for (count = 0; count < vrInputs->num_2ways; count++) {
			if (vrInputs->switch2[count] == NULL) {
				vrFprintf(file, "\t2-way[%d] is NULL\n", count);
			} else {
				vrFprintf(file, "\t2-way[%d] at %#p, from \"%s\":\"%s\" (%d:%d) -- %d\n",
					count,
					 vrInputs->switch2[count],
					(vrInputs->switch2[count]->my_device != NULL ? vrInputs->switch2[count]->my_device->name : "-"),
					(vrInputs->switch2[count]->my_object != NULL ? vrInputs->switch2[count]->my_object->name : "-"),
					(vrInputs->switch2[count]->my_device != NULL ? vrInputs->switch2[count]->my_device->id : -1),
					(vrInputs->switch2[count]->my_object != NULL ? vrInputs->switch2[count]->my_object->id : -1),
					 vrInputs->switch2[count]->value);
			}
		}
		vrFprintf(file, "%d N-way switches (at %#p):\n", vrInputs->num_Nways, vrInputs->switchN);
		for (count = 0; count < vrInputs->num_Nways; count++) {
			if (vrInputs->switchN[count] == NULL) {
				vrFprintf(file, "\tN-way[%d] is NULL\n", count);
			} else {
				vrFprintf(file, "\tN-way[%d] at %#p, from \"%s\":\"%s\" (%d:%d) -- %d\n",
					count,
					 vrInputs->switchN[count],
					(vrInputs->switchN[count]->my_device != NULL ? vrInputs->switchN[count]->my_device->name : "-"),
					(vrInputs->switchN[count]->my_object != NULL ? vrInputs->switchN[count]->my_object->name : "-"),
					(vrInputs->switchN[count]->my_device != NULL ? vrInputs->switchN[count]->my_device->id : -1),
					(vrInputs->switchN[count]->my_object != NULL ? vrInputs->switchN[count]->my_object->id : -1),
					 vrInputs->switchN[count]->value);
			}
		}
		vrFprintf(file, "%d valuators (at %#p):\n", vrInputs->num_valuators, vrInputs->valuator);
		for (count = 0; count < vrInputs->num_valuators; count++) {
			if (vrInputs->valuator[count] == NULL) {
				vrFprintf(file, "\tvaluator[%d] is NULL\n", count);
			} else {
				vrFprintf(file, "\tvaluator[%d] at %#p, from \"%s\":\"%s\" (%d:%d) -- %.2f\n",
					count,
					 vrInputs->valuator[count],
					(vrInputs->valuator[count]->my_device != NULL ? vrInputs->valuator[count]->my_device->name : "-"),
					(vrInputs->valuator[count]->my_object != NULL ? vrInputs->valuator[count]->my_object->name : "-"),
					(vrInputs->valuator[count]->my_device != NULL ? vrInputs->valuator[count]->my_device->id : -1),
					(vrInputs->valuator[count]->my_object != NULL ? vrInputs->valuator[count]->my_object->id : -1),
					 vrInputs->valuator[count]->value);
			}
		}
		vrFprintf(file, "%d 6-sensors (at %#p):\n", vrInputs->num_6sensors, vrInputs->sensor6);
		for (count = 0; count < vrInputs->num_6sensors; count++) {
			if (vrInputs->sensor6[count] == NULL) {
				vrFprintf(file, "\t6-sensor[%d] is NULL\n", count);
			} else {
				vrFprintf(file, "\t6-sensor[%d] at %#p, from \"%s\":\"%s\" (%d:%d) -- [tr: %.2f %.2f %.2f]\n",
					count,
					 vrInputs->sensor6[count],
					(vrInputs->sensor6[count]->my_device != NULL ? vrInputs->sensor6[count]->my_device->name : "-"),
					(vrInputs->sensor6[count]->my_object != NULL ? vrInputs->sensor6[count]->my_object->name : "-"),
					(vrInputs->sensor6[count]->my_device != NULL ? vrInputs->sensor6[count]->my_device->id : -1),
					(vrInputs->sensor6[count]->my_object != NULL ? vrInputs->sensor6[count]->my_object->id : -1),
					 VRMAT_ROWCOL(vrInputs->sensor6[count]->position, VR_X, VR_W),
					 VRMAT_ROWCOL(vrInputs->sensor6[count]->position, VR_Y, VR_W),
					 VRMAT_ROWCOL(vrInputs->sensor6[count]->position, VR_Z, VR_W));
			}
		}
		vrFprintf(file, "%d N-sensors (at %#p):\n", vrInputs->num_Nsensors, vrInputs->sensorN);
		for (count = 0; count < vrInputs->num_Nsensors; count++) {
			if (vrInputs->sensorN[count] == NULL) {
				vrFprintf(file, "\tN-sensor[%d] is NULL\n", count);
			} else {
				vrFprintf(file, "\tN-sensor[%d] at %#p, from \"%s\":\"%s\" (%d:%d)\n",
					count,
					 vrInputs->sensorN[count],
					(vrInputs->sensorN[count]->my_device != NULL ? vrInputs->sensorN[count]->my_device->name : "-"),
					(vrInputs->sensorN[count]->my_object != NULL ? vrInputs->sensorN[count]->my_object->name : "-"),
					(vrInputs->sensorN[count]->my_device != NULL ? vrInputs->sensorN[count]->my_device->id : -1),
					(vrInputs->sensorN[count]->my_object != NULL ? vrInputs->sensorN[count]->my_object->id : -1));
			}
		}
		vrFprintf(file, "%d controls (at %#p):\n", vrInputs->num_controls, vrInputs->control);
		for (count = 0; count < vrInputs->num_controls; count++) {
			if (vrInputs->control[count] == NULL) {
				vrFprintf(file, "\tcontrol[%d] is NULL\n", count);
			} else {
				vrFprintf(file, "\tcontrol[%d] at %#p, from \"%s\":\"%s\" (%d:%d)\n",
					count,
					 vrInputs->control[count],
					(vrInputs->control[count]->my_device != NULL ? vrInputs->control[count]->my_device->name : "-"),
					(vrInputs->control[count]->my_object != NULL ? vrInputs->control[count]->my_object->name : "-"),
					(vrInputs->control[count]->my_device != NULL ? vrInputs->control[count]->my_device->id : -1),
					(vrInputs->control[count]->my_object != NULL ? vrInputs->control[count]->my_object->id : -1));
			}
		}
		vrFprintf(file, "==================================================\n");
		break;

	case file_format:
		/* TODO: not yet implemented -- output that can be reread. */
	case one_line:
		/* TODO: not yet implemented -- very brief output */
	case machine:
		/* TODO: not yet implemented -- output easily machine parsed */
		vrFprintf(file, "TODO: put special version of input info here\n");
		break;
	}
}


/******************************************************************************/
/* The vrSprintInputUI() function prints into a given character string array  */
/*   information about the FreeVR input system, as further augmented by the   */
/*   application program through the assignment of user-interface description */
/*   strings to particular inputs.                                            */
/*                                                                            */
/* We are currently ignoring the "style" argument, but we may want to make    */
/*   use of this in the future.                                               */
/*                                                                            */
/* It returns a copy of the pointer to the string.                            */
/*                                                                            */
/* NOTE: a warning about the use of snprintf() -- the return value is the     */
/*   number of bytes that ** would have been written ** were there enough     */
/*   space in the string -- which forces us to check whether we've hit the    */
/*   limit or not!  I think this is a bad return value, but nothing I can     */
/*   do about that!                                                           */
char *vrSprintInputUI(char *str, vrInputInfo *vrInputs, int str_size, vrPrintStyle style)
{
	int	bytes_remaining = str_size;
	int	bytes_used = 0;
	int	bytes;			/* return result from snprintf() calls -- usually # of bytes written */
	int	first_of_type;		/* has the first input of the current type been output yet? */
	int	count;
	vrInput	*input_object;

	str[0] = '\0';

	/********************/
	/* Application name */
	/********************/
	if (vrInputs->context->name != NULL) {
		/* NOTE: add an extra '\n' if there is no 'authors' line (and the space is necessary because of the strtok() method of string parsing) */
		bytes = snprintf(str, bytes_remaining, "Program: %s%s\n", vrInputs->context->name, (vrInputs->context->authors == NULL ? "\n " : ""));
		bytes_used += bytes;
		bytes_remaining -= bytes;
		if ((bytes < 1) || (bytes_remaining < 1)) {
			str[str_size] = '\0';		/* for safety */
			return str;
		}
	}

	/***********************/
	/* Application authors */
	/***********************/
	if (vrInputs->context->authors != NULL) {
		bytes = snprintf(&str[bytes_used], bytes_remaining, "Author(s): %s\n", vrInputs->context->authors);
		bytes_used += bytes;
		bytes_remaining -= bytes;
		if ((bytes < 1) || (bytes_remaining < 1)) {
			str[str_size] = '\0';		/* for safety */
			return str;
		}
	}


	/******************/
	/* FreeVR version */
	/******************/
	if (vrInputs->context->authors != NULL) {
		bytes = snprintf(&str[bytes_used], bytes_remaining, "FreeVR version: %s\n \n", &(vrInputs->context->version[15]));	/* adding a colon in the middle */
		bytes_used += bytes;
		bytes_remaining -= bytes;
		if ((bytes < 1) || (bytes_remaining < 1)) {
			str[str_size] = '\0';		/* for safety */
			return str;
		}
	}


	/*****************************/
	/*** 2-way's (aka buttons) ***/
	first_of_type = 1;
	for (count = 0; count < vrInputs->num_2ways; count++) {
		if (vrInputs->switch2[count] != NULL) {
			input_object = vrInputs->switch2[count]->my_object;

			if ((input_object != NULL) && (input_object->desc_ui[0] != '\0')) {
				/* print the section header just before the first button */
				if (first_of_type) {
					bytes = snprintf(&str[bytes_used], bytes_remaining, "Buttons:\n");
					bytes_used += bytes;
					bytes_remaining -= bytes;
					if ((bytes < 1) || (bytes_remaining < 1)) {
						str[str_size] = '\0';		/* for safety */
						return str;
					}

					first_of_type = 0;
				}

				/* now print the button information */
				bytes = snprintf(&str[bytes_used], bytes_remaining, "  button[%d] -- \"%s\" performs: %s\n",
						count,
						input_object->name,
						input_object->desc_ui);

				bytes_used += bytes;
				bytes_remaining -= bytes;
				if ((bytes < 1) || (bytes_remaining < 1)) {
					str[str_size] = '\0';		/* for safety */
					return str;
				}
			}
		}
	}

	/***************/
	/*** N-way's ***/
	first_of_type = 1;
	for (count = 0; count < vrInputs->num_Nways; count++) {
		if (vrInputs->switchN[count] != NULL) {
			input_object = vrInputs->switchN[count]->my_object;

			if ((input_object != NULL) && (input_object->desc_ui[0] != '\0')) {
				/* print the section header just before the first button */
				if (first_of_type) {
					bytes = snprintf(&str[bytes_used], bytes_remaining, "N-way switches:\n");
					bytes_used += bytes;
					bytes_remaining -= bytes;
					if ((bytes < 1) || (bytes_remaining < 1)) {
						str[str_size] = '\0';		/* for safety */
						return str;
					}

					first_of_type = 0;
				}

				/* now print the N-way information */
				bytes = snprintf(&str[bytes_used], bytes_remaining, "  switch[%d] -- \"%s\" performs: %s\n",
						count,
						input_object->name,
						input_object->desc_ui);

				bytes_used += bytes;
				bytes_remaining -= bytes;
				if ((bytes < 1) || (bytes_remaining < 1)) {
					str[str_size] = '\0';		/* for safety */
					return str;
				}
			}
		}
	}

	/*****************/
	/*** Valuators ***/
	first_of_type = 1;
	for (count = 0; count < vrInputs->num_valuators; count++) {
		if (vrInputs->valuator[count] != NULL) {
			input_object = vrInputs->valuator[count]->my_object;

			if ((input_object != NULL) && (input_object->desc_ui[0] != '\0')) {
				/* print the section header just before the first button */
				if (first_of_type) {
					bytes = snprintf(&str[bytes_used], bytes_remaining, "Valuators:\n");
					bytes_used += bytes;
					bytes_remaining -= bytes;
					if ((bytes < 1) || (bytes_remaining < 1)) {
						str[str_size] = '\0';		/* for safety */
						return str;
					}

					first_of_type = 0;
				}

				/* now print the valuator information */
				bytes = snprintf(&str[bytes_used], bytes_remaining, "  valuator[%d] -- \"%s\" performs: %s\n",
						count,
						input_object->name,
						input_object->desc_ui);

				bytes_used += bytes;
				bytes_remaining -= bytes;
				if ((bytes < 1) || (bytes_remaining < 1)) {
					str[str_size] = '\0';		/* for safety */
					return str;
				}
			}
		}
	}

	/*****************/
	/*** 6-Sensors ***/
	first_of_type = 1;
	for (count = 0; count < vrInputs->num_6sensors; count++) {
		if (vrInputs->sensor6[count] != NULL) {
			input_object = vrInputs->sensor6[count]->my_object;

			if ((input_object != NULL) && (input_object->desc_ui[0] != '\0')) {
				/* print the section header just before the first button */
				if (first_of_type) {
					bytes = snprintf(&str[bytes_used], bytes_remaining, "6DOF sensors:\n");
					bytes_used += bytes;
					bytes_remaining -= bytes;
					if ((bytes < 1) || (bytes_remaining < 1)) {
						str[str_size] = '\0';		/* for safety */
						return str;
					}

					first_of_type = 0;
				}

				/* now print the 6-sensor information */
				bytes = snprintf(&str[bytes_used], bytes_remaining, "  6-sensor[%d] -- \"%s\" performs: %s\n",
						count,
						input_object->name,
						input_object->desc_ui);

				bytes_used += bytes;
				bytes_remaining -= bytes;
				if ((bytes < 1) || (bytes_remaining < 1)) {
					str[str_size] = '\0';		/* for safety */
					return str;
				}
			}
		}
	}

	/*****************/
	/*** N-Sensors ***/
	first_of_type = 1;
	for (count = 0; count < vrInputs->num_Nsensors; count++) {
		if (vrInputs->sensorN[count] != NULL) {
			input_object = vrInputs->sensorN[count]->my_object;

			if ((input_object != NULL) && (input_object->desc_ui[0] != '\0')) {
				/* print the section header just before the first button */
				if (first_of_type) {
					bytes = snprintf(&str[bytes_used], bytes_remaining, "N-sensors:\n");
					bytes_used += bytes;
					bytes_remaining -= bytes;
					if ((bytes < 1) || (bytes_remaining < 1)) {
						str[str_size] = '\0';		/* for safety */
						return str;
					}

					first_of_type = 0;
				}

				/* now print the N-sensor information */
				bytes = snprintf(&str[bytes_used], bytes_remaining, "  N-sensor[%d] -- \"%s\" performs: %s\n",
						count,
						input_object->name,
						input_object->desc_ui);

				bytes_used += bytes;
				bytes_remaining -= bytes;
				if ((bytes < 1) || (bytes_remaining < 1)) {
					str[str_size] = '\0';		/* for safety */
					return str;
				}
			}
		}
	}

	/*********************************/
	/* Application Extra-Information */
	/*********************************/
	if (vrInputs->context->extra_info != NULL) {
		bytes = snprintf(&str[bytes_used], bytes_remaining, " \nAdditional information:\n  %s\n", vrInputs->context->extra_info);
		bytes_used += bytes;
		bytes_remaining -= bytes;
		if ((bytes < 1) || (bytes_remaining < 1)) {
			str[str_size] = '\0';		/* for safety */
			return str;
		}
	}

	/**********************************/
	/* Application Status Information */
	/**********************************/
	if (vrInputs->context->status_info != NULL) {
		bytes = snprintf(&str[bytes_used], bytes_remaining, " \nApplication status:\n  %s\n", vrInputs->context->status_info);
		bytes_used += bytes;
		bytes_remaining -= bytes;
		if ((bytes < 1) || (bytes_remaining < 1)) {
			str[str_size] = '\0';		/* for safety */
			return str;
		}
	}

	return str;
}


	/**************************************************************/
	/**************************************************************/
	/************ Input process operational functions *************/

/*******************************************************************/
/* vrInputInitProc(): Initialize the input devices associated with */
/*   this process.                                                 */
void vrInputInitProc(vrProcessInfo *myproc_info)
{
	int		count;

	vrTrace("vrInputInitProc", "beginning");
	proc_info = myproc_info;

#ifdef COREDUMP_IS_GOOD
	vrSetSignalHandler(vrInputSignalHandler);
#endif

	if (vrDbgDo(DEFAULT_DBGLVL+1)) {
		vrMsgPrintf("Input proc is: ");
		vrFprintProcessInfo(stdout, myproc_info, verbose);
	}

	/* copy data into static variables to avoid latency from indirection */
	num_devices = myproc_info->num_things;
	devices = (vrInputDevice **)(myproc_info->things);

	/**************************************************************/
	/*** Create inputs for all input devices (for this process) ***/
	/**************************************************************/
	vrTrace("vrInputInitProc", "creating input devices");
	vrDbgPrintfN(INPUT_DBGLVL, "vrInputInitProc(): creating %d input devices\n", num_devices);
	for (count = 0; count < num_devices; count++) {
		if (devices[count] != NULL) {
			/* create the device's inputs */
			vrDbgPrintfN(INPUT_DBGLVL, "vrInputInitProc(): calling Create callback for input device %d (%s)\n", count, devices[count]->name);
			vrCallbackInvoke(devices[count]->Create);
		} else {
			vrErrPrintf("vrInputInitProc()-creating: " RED_TEXT "Device '%s' is null.  Something is wrong.\n" NORM_TEXT, myproc_info->thing_names[count]);
		}
	}


	/*************************************************/
	/*** Open all input devices (for this process) ***/
	/*************************************************/
	vrTrace("vrInputInitProc", "opening input devices");
	vrDbgPrintfN(INPUT_DBGLVL, "vrInputInitProc(): opening %d input devices\n", num_devices);
	for (count = 0; count < num_devices; count++) {
		if (devices[count] != NULL) {
			/* start the input device */
			vrTrace("vrInputInitProc", devices[count]->name);
			vrCallbackInvoke(devices[count]->Open);
		} else {
			vrErrPrintf("vrInputInitProc()-opening: " RED_TEXT "Device '%s' is null.  Something is wrong.\n" NORM_TEXT, myproc_info->thing_names[count]);
		}
	}

	/************************************/
	/*** Initialize timing statistics ***/
	/************************************/
	/* NOTE: here are the meanings of all the stats values:        */
	/*    <1 to num_devices-1> is time spent polling that device   */
	/*    <num_devices> is time spent in sleep.                    */
	/*    <num_devices+1> is time spent waiting for sync barrier   */
	/*    <num_devices+2> is time spent waiting for freeze barrier */
	if (myproc_info->stats_args) {
		vrTrace("vrInputInitProc", "initializing process statistics");
		myproc_info->stats = vrProcessStatsCreate(myproc_info->name, num_devices+3, myproc_info->stats_args);

		/* if no y-location was assigned then change the default */
		if (!strstr(myproc_info->stats_args, "yloc"))
			myproc_info->stats->yloc = 0.667;

		/* set the element labels */
		for (count = 0; count < num_devices; count++)
			myproc_info->stats->elem_labels[count] = vrShmemStrDup(devices[count]->name);
		myproc_info->stats->elem_labels[num_devices+0] = vrShmemStrDup("sleep");
		myproc_info->stats->elem_labels[num_devices+1] = vrShmemStrDup("sync");
		myproc_info->stats->elem_labels[num_devices+2] = vrShmemStrDup("freeze");
	}

	return;
}


/**************************************************************/
/* vrInputTermProc(): Close the input devices associated with */
/*   this process.                                            */
void vrInputTermProc(vrProcessInfo *myproc_info)
{
#if 1
	int		count;
#endif

	vrTrace("vrInputTermProc", "beginning");

#if 1 /* 7/3/01: this sometimes hangs during the close-down phase (during a shmem-free call) */
	/**************************************************************/
	/*** Create inputs for all input devices (for this process) ***/
	/**************************************************************/
	for (count = 0; count < num_devices; count++) {
		if (devices[count] != NULL) {
#if 0 /* 11/15/01: instead of doing nothing, we at least need to close the device -- I think, I know xwindows should be closed to reenable keyrepeating */
			/* Delete the device's inputs */
			if (devices[count]->operating) {
				vrDeleteInputs(devices[count]);		/* TODO: Need to try doing this, but testing input existence when calculating perspective matrices */

			/* Close the device */
			if (devices[count]->operating) {
				vrCallbackInvoke(devices[count]->Close);

				vrShmemFree(devices[count]);		/* TODO: can't do this until the vrDeleteInputs() stuff works */
			}
#else
			/* Close the device (if it's open!) */
			if (devices[count]->operating)
				vrCallbackInvoke(devices[count]->Close);
#endif
		} else {
			vrErrPrintf("vrInputTermProc(): " RED_TEXT "Null device.  Something is wrong.\n" NORM_TEXT);
		}
	}
#endif

	return;
}


/*****************************************************************************/
/* vrInputOneFrame(): handle all the work for a single input frame.  Which   */
/*   basically just means poll each input device.  Since num_devices and     */
/*   devices are file-scope globals, there are no arguments to this function.*/
/*   We may want to change this, and have these values passed as arguments.  */
void vrInputOneFrame(vrProcessInfo *myproc_info)
{
static	char	trace_msg[256];				/* space to create a string for vrTrace */
	int	count;					/* loop counter */

	/* calculate frame rates and set the process time values */
	myproc_info->frame_count++;
	vrProcessCalcFrameRate(myproc_info);

	/* measure: update time measurement array index */
	vrProcessStatsNextFrame(myproc_info->stats);

	/* do the input polling */
	for (count = 0; count < num_devices; count++) {
		if (devices[count]) {
			snprintf(trace_msg, sizeof(trace_msg), "about to poll inputs from device %s", devices[count]->name);
			vrTrace("vrInputOneFrame", trace_msg);
			vrCallbackInvoke(devices[count]->PollData);
		}
		vrProcessStatsMark(myproc_info->stats, count, 0);
	}
}


/**********************************************************************/
/* vrInputMainLoop(): this process will just loop, invoking a polling */
/*   function for each of the devices it's responsible for.           */
void vrInputMainLoop(vrProcessInfo *myproc_info)
{
	int	sync_order;				/* the order of hitting the barrier */
	vrTime	loop_wtime = vrCurrentWallTime();	/* time of the beginning of each loop */

	vrTrace("vrInputMainLoop", "beginning");

	/* At this point, the inputs for all the devices _of this process_ */
	/*   have been created.  Some of the devices may not be fully open */
	/*   yet, and for those, more attempts to reopen can be made while */
	/*   the device is polled.                                         */

	/******************************/
	/*** Poll all input devices ***/
	/******************************/
	/* enter a (seemingly infinite) loop, polling each device */

	while (!myproc_info->req_proc_term) {
		vrTrace("vrInputMainLoop", BOLD_TEXT "*** top of input loop ***" NORM_TEXT);

		/* do minimal frame delay to allow other processes to get CPU time */
		/* NOTE: Delays of 10,000us or less allow 50fps frame rate for the input process (at least on my Thinkpad 770Z running Linux) */
		vrSleep(myproc_info->usec_min - (long)((vrCurrentWallTime() - loop_wtime) * 1000000.0));
		vrProcessStatsMark(myproc_info->stats, num_devices, 0);	/* time spent in sleep */

		vrProcessSync(myproc_info, num_devices+1, num_devices+2);

		/* start measuring for minimal loop time */
		loop_wtime = vrCurrentWallTime();

		vrInputOneFrame(myproc_info);
	}

	vrTrace("vrInputMainLoop", "ending");
	return;
}


/***************************************************************************/
/* vrInputCheckIfAllInputDevicesAreOpen(): this function simply checks the */
/*   operational state of each of the input devices in the configuration   */
/*   and returns true (1) if they all meet that criterion, otherwise false.*/
/* NOTE: never called. */
int vrInputCheckIfAllInputDevicesAreOpen(vrContextInfo *context)
{
	vrConfigInfo	*vrConfig = context->config;
	int		num_devices = vrConfig->num_input_devices;
	int		devnum;

	for (devnum = 0; devnum < num_devices; devnum++) {
		if (vrConfig->input_devices[devnum]->operating == 0) {
			return 0;
		}
	}

	return 1;
}


/*******************************************************************/
/* vrInputWaitForAllInputDevicesToBeOpen():                        */
/*   waits at a barrier until all the input devices are open -- is */
/*   (currently at least) necessary for some functions that assume */
/*   the input data structures are all taken care of.              */
/*                                                                 */
/* NOTE: at the moment (12/10/2002) this function is not actually  */
/*   called by the library.  The travel.c test program calls it    */
/*   before beginning the world simulation in earnest, however     */
/*   most programs don't call it, and they seem to run fine.  The  */
/*   reason the library doesn't call it is to allow the world      */
/*   simulation to begin, even when the system isn't fully engaged.*/
void vrInputWaitForAllInputDevicesToBeOpen()
{
	vrConfigInfo	*vrConfig = vrContext->config;
	int		num_devices = vrConfig->num_input_devices;
	int		devnum;

	for (devnum = 0; devnum < num_devices; devnum++) {
		vrDbgPrintfN(ALWAYS_DBGLVL, "FreeVR (%.4f): Input device '%s' is %s\n", vrConfig->input_devices[devnum]->name,
			vrCurrentSimTime(), (vrConfig->input_devices[devnum]->operating ? "operating." : RED_TEXT "NOT YET operating." NORM_TEXT));
	}
	for (devnum = 0; devnum < num_devices; devnum++) {
		while(vrConfig->input_devices[devnum]->operating == 0) {
			vrDbgPrintfN(ALWAYS_DBGLVL, "FreeVR (%.4f): " BOLD_TEXT "Still waiting for '%s' to be operating.\n" NORM_TEXT,
				vrCurrentSimTime(), vrConfig->input_devices[devnum]->name);
			vrSleep(1000000);
		}
	}
	vrDbgPrintfN(ALWAYS_DBGLVL, "FreeVR (%.4f): " BOLD_TEXT ">>>>>>>>>>>>>> All %d input devices reporting operating. <<<<<<<<<<<<<<\n"
		NORM_TEXT, vrCurrentSimTime(), num_devices);
}


/*********************************************************************/
/* vrInputWaitForAllInputsToBeCreated():                             */
/*   Waits until the inputs for all the input devices to be created  */
/*   This is (currently at least) necessary for some functions that  */
/*   assume the input data structures are all taken care of.         */
/* 03/08/2006: added the concept of a maximum number of loops before */
/*   ending the waiting.  This may be bad in some circumstances, but */
/*   for optional input devices, at least we can continue getting    */
/*   FreeVR up and running.                                          */
void vrInputWaitForAllInputsToBeCreated(vrContextInfo *context)
{
	vrConfigInfo	*vrConfig = context->config;
	int		devnum;
	int		num_devices = vrConfig->num_input_devices;
	int		wait_count;
	int		max_wait = 5;			/* TODO: probably want this to be configurable -- perhaps on a device by device basis */
	int		gave_up_flag = 0;

	for (devnum = 0; devnum < num_devices; devnum++) {
		vrDbgPrintfN(ALWAYS_DBGLVL, "FreeVR (%.4f): Inputs for device '%s' are %s\n", vrCurrentSimTime(), vrConfig->input_devices[devnum]->name,
			(vrConfig->input_devices[devnum]->created ? "created." : RED_TEXT "NOT YET created." NORM_TEXT));
	}
	for (devnum = 0; devnum < num_devices; devnum++) {
		wait_count = 0;
		while (vrConfig->input_devices[devnum]->created == 0 && wait_count < max_wait) {
			vrDbgPrintfN(ALWAYS_DBGLVL, "FreeVR (%.4f): " BOLD_TEXT "Still waiting for '%s' inputs to be created -- skipping in %d.\n" NORM_TEXT,
				vrCurrentSimTime(), vrConfig->input_devices[devnum]->name, (max_wait-wait_count));
			vrSleep(1000000);
			wait_count++;
		}
		if (vrConfig->input_devices[devnum]->created == 0) {
			vrDbgPrintfN(ALWAYS_DBGLVL, "FreeVR (%.4f): " BOLD_TEXT "*** Giving up on the creation of input '%s' ***.\n" NORM_TEXT,
				vrCurrentSimTime(), vrConfig->input_devices[devnum]->name);
			gave_up_flag++;

			/* TODO: Create dummy inputs for all the inputs that were anticipated for this device! */
		}
	}
	if (gave_up_flag) {
		vrDbgPrintfN(ALWAYS_DBGLVL, "FreeVR (%.4f): " BOLD_TEXT ">>>>>> Inputs for all but %d of the %d devices have been created. <<<<<<\n" NORM_TEXT,
			vrCurrentSimTime(), gave_up_flag, num_devices);
	} else {
		vrDbgPrintfN(ALWAYS_DBGLVL, "FreeVR (%.4f): " BOLD_TEXT ">>>>>> Inputs for all %d devices created. <<<<<<\n" NORM_TEXT,
			vrCurrentSimTime(), num_devices);
	}

	/* set the flag that all the devices have been initialized */
	/* NOTE: this includes those that we've given up on.       */
	context->input->all_devices_initialized = 1;
}


/*********************************************************************/
/* vrGetDefaultInputDeviceInfo(): this function assigns some default */
/*   callbacks for the 5 major operations of a FreeVR input device   */
/*   (Create/Open/Close/Reset/PollData).  Each device will then set  */
/*   it's own actual callback functions when it is initialized.      */
void vrGetDefaultInputDeviceInfo(vrInputDevice *info)
{
	info->version = (char *)vrShmemStrDup("Default (null) input device");
	info->Create = vrCallbackCreateNamed("Default Indev:Create-DN", vrDoNothing, 2, info);
	info->Open = vrCallbackCreateNamed("Default Indev:Open-DN", vrDoNothing, 2, info);
	info->Close = vrCallbackCreateNamed("Default Indev:Close-DN", vrDoNothing, 2, info);
	info->Reset = vrCallbackCreateNamed("Default Indev:Reset-DN", vrDoNothing, 2, info);
	info->PollData = vrCallbackCreateNamed("Default Indev:PollData-DN", vrDoNothing, 2, info);

	info->operating = 1;

	return;
}


/***************************************************************************/
/* vrGetInputDeviceInfo(): Fill in the info structure callbacks with the   */
/*   functions associated with the "type" field string indicating the type */
/*   of the input devices.  First checks for DSO functions, then searches  */
/*   the list specified in the vr_input.opts.h header file.                */
void vrGetInputDeviceInfo(vrInputDevice *info)
{
	char		*type = info->type;	/* string name that indicates type of the input device */

	/******************************************************************/
	/*** First check DSO values for initializing the info structure ***/

	/* NOTE: relies on lazy boolean evaluation.  The strlen operation is here */
	/*   to allow a configuration file to cancel a dso setting by giving the  */
	/*   empty string as an argument.                                         */
	if (info->dso_file && strlen(info->dso_file) > 0) {
		char		*dso_func = info->dso_func;
		char		*dso_file = info->dso_file;
		void		*dso_handle = NULL;
		void		(*dso_info_func)(vrInputDevice *) = NULL;

#if defined(__hpux)
		vrErrPrintf("vrGetInputDeviceInfo: "
			RED_TEXT "This platform does not support DSO loading (yet)\n" NORM_TEXT);
#else
		/* If a DSO file has been named, then first attempt to use it. */

		/* give a default function name if none given */
		if (!dso_func)
			dso_func = "initializeDeviceInfo";

vrPrintf(RED_TEXT "LD_LIBRARY_PATH is '%s'\n", getenv("LD_LIBRARY_PATH"));
vrPrintf(RED_TEXT "LD_LIBRARYN32_PATH is '%s'\n", getenv("LD_LIBRARYN32_PATH"));
		dso_handle = dlopen(dso_file, RTLD_LAZY);
		if (dso_handle == NULL) {
			vrErrPrintf("vrGetInputDeviceInfo: "
				RED_TEXT "Unable to load DSO input device file ('%s').\n" NORM_TEXT,
				dso_file);
			vrPrintf("dlerror: " BOLD_TEXT "'%s'\n" NORM_TEXT, dlerror());
		} else {
			info->dso_handle = dso_handle;	/* store the handle for removing the dso when the device is closed */
			dso_info_func = (void (*)(vrInputDevice *))dlsym(dso_handle, dso_func);
			if (dso_info_func == NULL) {
				vrErrPrintf("vrGetInputDeviceInfo: "
					RED_TEXT "No \"%s\" function in DSO input device file ('%s').\n" NORM_TEXT,
					dso_func, dso_file);
				vrPrintf(BOLD_TEXT "dlerror: '%s'\n" NORM_TEXT, dlerror());
			} else {
				/* If the dso operations worked, then fill in the info structure */
				/*   by calling the initialization function.                     */
				(*dso_info_func)(info);
				return;
			}
		}
#endif
	}

	/**********************************************************************************/
	/*** Next search the list of possible input devices listed in vr_inputs.opts.h. ***/

	/* an implied "else" here, since we'll have returned if the DSO worked */
	/*   (though if it didn't work, it will report an error, but still     */
	/*   attempt to function using the built-in version of the input).     */
	if (!type) {
		/* If no "type" string is given (ie. it's NULL), then use */
		/*   a default input device for this device.              */
		vrGetDefaultInputDeviceInfo(info);
		return;

	} else {
		vrInputOptsType	*option;	/* for looping through the list of possible input devices */

		/* Loop through the input device options specified in vr_input.opts.h   */
		/*   until a string match is made with the "type" string, then call the */
		/*   associated initialization function to fill in the info structure.  */
		option = vrInputOpts;
		while (option->option_name != NULL) {
			if (!strcasecmp(option->option_name, type)) {
				(*option->info_func)(info);
				return;
			}
			option++;
		}
	}

	/***************************************************/
	/*** Last option is the default nop input device ***/

	/* If we get here, then neither the DSO nor the input-option list provided */
	/*   a function for initializing the info structure, so use the default    */
	/*   function (which basically setups up a bunch of nop callbacks).        */
	vrGetDefaultInputDeviceInfo(info);
	return;
}


	/************************************************************/
	/************************************************************/
	/************ Generic input type access function ************/


/******************************************************************************/
/* vrAssignGenericInput(): this function takes a pointer to a FreeVR "input"  */
/*   and assigns a value to it from the provided string (char *) "value"      */
/*   parameter.  The type of the "input" argument is used to determine the    */
/*   type of data in the parameter -- "Binary/2switch" and "Nswitch" inputs   */
/*   are presumed to have an integer argument; valuators are assumed to have  */
/*   a floating point argument, and 6sensors are assumed to have a specific   */
/*   command, from the options: "id", "loc", "move", and "nr".  The "id"      */
/*   command takes no additional arguments and sets the 6sensor to the        */
/*   identity matrix.  The "loc" command takes 1 to 3 additional floating     */
/*   point values, which are interpreted one by one, so a single value is     */
/*   used to set the VR_X setting only, and two values will set VR_X and VR_Y,*/
/*   etc.  The "move" command expects 6 additional floating point numbers     */
/*   which are treated as 6 valuator inputs that control the simulated 6sensor*/
/*   using the vrAssign6sensorValueFromValuators() function.  Finally, the    */
/*   "nra" command (aka "nudger2eA") takes 1-3 numbers and adjusts only the   */
/*   rotation of the "r2e_xform" field for that input.  And now there is also */
/*   the "nrt" (aka "nudger2eT") command that takes 1-3 values and nudges the */
/*   "r2e_xform" matrix translationally.  Oh, and both of these now have the  */
/*   "id" option to reset the rotation or translation back to the identity    */
/*   operation.                                                               */
/*                                                                            */
/* NOTE: this is presently (11/29/2020) only called from vrInfoSet() in the   */
/*   vr_telnet.c process file.                                                */
/*                                                                            */
/* TODO: for all the input assignment functions, we could have a third        */
/*   argument (vrTime) that allows the time to be specified by the calling    */
/*   party, or if a value of 0 time is given then determine the time here.    */
/*   I suggest this because for devices like the tracker daemon, a time       */
/*   is already given for when the value was read by the remote process.      */
/*                                                                            */
/* TODO: do we also want get-functions to return the timestamp value/delta?   */
/*                                                                            */
char *vrAssignGenericInput(vrGenericInput *input, const char *value /* , vrTime time */)
{
static	char	*whitespace = " \t\r\b\n";
static	char	*help_msg_6sensor = "\b\b\b\b\b\b\b"	/* These erase the "Error:" message */
			"Available queries:\n"
			"\t" "help -- print a list of setting options\n"
			"\t" "id -- set the 6-sensor to the identity matrix\n"
			"\t" "loc <x> <y> <z> -- move to the given location\n"
			"\t" "move <x> <y> <z> <x-rot> <y-rot> <z-rot>\n"
			"\t" "nudger2eA|nra { id | <x-rot> <y-rot> <z-rot> }\n"
			"\t" "nudger2eT|nrt { id | <x> <y> <z> }\n"
		"";
	char	*parse_value;		/* An internally modfiable copy of "value" */

	if (value == NULL) {
		vrDbgPrintfN(AALWAYS_DBGLVL, "vrAssignGenericInput(): Null assignment string given.\n");
		return "Null assignment string given.";
	}

	if (input == NULL) {
		vrDbgPrintfN(AALWAYS_DBGLVL, "vrAssignGenericInput(): Null input pointer given.\n");
		return "Null input pointer given.";
	}

	parse_value = (char *)value + strspn(value, whitespace);		/* skip white */

	switch (input->input_type) {
	case VRINPUT_BINARY:	/* equivalent to VRINPUT_2WAY */
		vrAssign2switchValue((vr2switch *)input, vrAtoI(parse_value) /* , vrTime time */);
		break;

	case VRINPUT_NWAY:	/* equivalent to VRINPUT_NARY */
		vrAssignNswitchValue((vrNswitch *)input, vrAtoI(parse_value) /* , vrTime time */);
		break;

	case VRINPUT_VALUATOR:
		vrAssignValuatorValue((vrValuator *)input, atof(parse_value) /* , vrTime time */);
		break;

	case VRINPUT_6SENSOR:
vrPrintf("vrAssignGenericInput(): 6-sensor assignment: '%s' : '%s' (%#p)\n", input->my_object->name, parse_value, parse_value);

		/***********************/
		/*** <6-sensor> help ***/
		if (!strncasecmp(parse_value, "help", 4)) {
			/* the "help" command sends options for affecting a 6-sensor */
			return help_msg_6sensor;

		/*********************/
		/*** <6-sensor> id ***/
		} else if (!strncasecmp(parse_value, "id", 2)) {
			vrMatrix	mat_store;		/* temporary matrix storage */
			/* the "id" command resets 6-sensor to the identity matrix */
			vrMatrixSetIdentity(&mat_store);
			vrAssign6sensorValue((vr6sensor *)input, &mat_store, 0 /* , vrTime time */);

		/**********************/
		/*** <6-sensor> loc ***/
		} else if (!strncasecmp(parse_value, "loc", 3)) {
			/* the "loc" command repositions the sensor at x,y,z */
			vrMatrix	mat_store;		/* temporary matrix storage */

			vrMatrixGet6sensorRawValuesDirect(&mat_store, (vr6sensor *)input);

			/* replace the X, Y & Z values that are given -- using a "*" will leave it as is */
			parse_value = strchr(&parse_value[3], ' ');				/* skip to a space */
vrPrintf("vrAssignGenericInput(): 6-sensor assignment: '%s' (%#p)\n", parse_value, parse_value);
			if (parse_value != NULL && strlen(parse_value) > 0) {		/* done if no more spaces */
				parse_value += strspn(parse_value, whitespace);     		/* skip white */
				if (parse_value[0] != '*') {
vrPrintf("Setting X to %f (%s)\n", atof(parse_value), parse_value);
					VRMAT_ROWCOL(&mat_store, VR_X, VR_W) = atof(parse_value);	/* set 'x' parse_value */
				}
				parse_value = strchr(parse_value, ' ');				/* skip to next space */
			}

			if (parse_value != NULL && strlen(parse_value) > 0) {		/* done if no more spaces */
				parse_value += strspn(parse_value, whitespace);     		/* skip white */
				if (parse_value[0] != '*') {
vrPrintf("Setting Y to %f (%s)\n", atof(parse_value), parse_value);
					VRMAT_ROWCOL(&mat_store, VR_Y, VR_W) = atof(parse_value);	/* set 'y' parse_value */
				}
				parse_value = strchr(parse_value, ' ');				/* skip to next space */
			}

			if (parse_value != NULL && strlen(parse_value) > 0) {		/* done if no more spaces */
				parse_value += strspn(parse_value, whitespace);     		/* skip white */
				if (parse_value[0] != '*') {
vrPrintf("Setting Z to %f (%s)\n", atof(parse_value), parse_value);
					VRMAT_ROWCOL(&mat_store, VR_Z, VR_W) = atof(parse_value);	/* set 'z' parse_value */
				}
			}

			vrAssign6sensorValue((vr6sensor *)input, &mat_store, 0 /* , vrTime time */);

		/***********************/
		/*** <6-sensor> move ***/
		} else if (!strncasecmp(parse_value, "move", 4)) {
			/* the "move" command does a movement based on the X,Y,Z,AZIM,ELEV,ROLL values given */
			vr6sensorConv	default_options = {
						0,		/* ignore_all */
						0,		/* ignore_trans */
						0,		/* restrict_space */
						0,		/* return_to_zero */
						0,		/* relative_axis */
						VR_Y,		/* azimuth_axis */
						0,		/* swap_yz */
						0,		/* swap_transrot */
						1.0, 1.0,	/* trans_scale & rot_scale */
						-5.0, 0.0, -5.0,/* working-volume minimums */
						5.0, 10.0, 5.0	/* working-volume maximums */
					};
			float		valuators[6] = { 0.0, 0.0, 0.0, 0.0, 0.0, 0.0 };

			parse_value = strchr(&parse_value[4], ' ');				/* skip past next space */
			if (parse_value != NULL && strlen(parse_value) != 0) {		/* done if no more spaces */
				parse_value += strspn(parse_value, whitespace);    		/* skip white */
				valuators[0] = atof(parse_value);
				parse_value = strchr(parse_value, ' ');			/* skip to next space */
			}

			if (parse_value != NULL && strlen(parse_value) != 0) {		/* done if no more spaces */
				parse_value += strspn(parse_value, whitespace);     	/* skip white */
				valuators[1] = atof(parse_value);
				parse_value = strchr(parse_value, ' ');			/* skip to next space */
			}

			if (parse_value != NULL && strlen(parse_value) != 0) {		/* done if no more spaces */
				parse_value += strspn(parse_value, whitespace);     	/* skip white */
				valuators[2] = atof(parse_value);
				parse_value = strchr(parse_value, ' ');			/* skip to next space */
			}

			if (parse_value != NULL && strlen(parse_value) != 0) {		/* done if no more spaces */
				parse_value += strspn(parse_value, whitespace);     	/* skip white */
				valuators[3] = atof(parse_value);
				parse_value = strchr(parse_value, ' ');			/* skip to next space */
			}

			if (parse_value != NULL && strlen(parse_value) != 0) {		/* done if no more spaces */
				parse_value += strspn(parse_value, whitespace);     	/* skip white */
				valuators[4] = atof(parse_value);
				parse_value = strchr(parse_value, ' ');			/* skip to next space */
			}

			if (parse_value != NULL && strlen(parse_value) != 0) {		/* done if no more spaces */
				parse_value += strspn(parse_value, whitespace);    		/* skip white */
				valuators[5] = atof(parse_value);
			}

			vrAssign6sensorValueFromValuators((vr6sensor *)input, valuators, &default_options, -1 /* , vrTime time */);

		/*************************************************************/
		/*** <6-sensor> "nra|nudger2ea" { id | x-rot y-rot z-rot } ***/
		} else if (!strncasecmp(parse_value, "nudger2ea", 9) || !strncasecmp(parse_value, "nra", 3)) {
			/* the "nudger2eA" command allows for adjusting the 6-sensors calibration */
			/* Form: "nudger2eA|nra { id | <x-rot> <y-rot> <z-rot> } "                */
			vrMatrix	mat_store;				/* temporary matrix storage */
			vrMatrix	*tmpmat;				/* local pointer to r2e_xform */
			double		xrot,yrot,zrot = 0.0;			/* rotation values */
			char		*id_check;				/* for parsing for "id" */

			parse_value = strchr(&parse_value[2], ' ');				/* skip to a space */
			if (parse_value == NULL) {
				return "Nudge r2e Angles missing arguments.";
			}
vrPrintf("vrAssignGenericInput(): 6-sensor r2e angle nudge: '%s' (%#p)\n", parse_value, parse_value);

			tmpmat = ((vr6sensor *)input)->r2e_xform;		/* point to the r2e matrix */

			/* check whether to reset to identity rotation */
			/* NOTE: I need to leave the whitespace for the parse_value checks below, */
			/*   so I'm using a second, temporary, variable for the "id" check. */
			id_check = parse_value + strspn(parse_value, whitespace);		/* skip white for "id" check */
			if (!strncasecmp(id_check, "id", 2)) {
vrPrintf("vrAssignGenericInput(): 6-sensor r2e angle nudge id: '%s' (%#p)\n", parse_value, parse_value);
				vrMatrixSetIdentityRotationOnly(tmpmat);
			} else {
				/* otherwise expect up to three rotation values */
				if (parse_value != NULL && strlen(parse_value) > 0) {	/* done if no more spaces */
					parse_value += strspn(parse_value, whitespace);     /* skip white */
					xrot = atof(parse_value);
					parse_value = strchr(parse_value, ' ');		/* skip to next space */
				}
				if (parse_value != NULL && strlen(parse_value) > 0) {	/* done if no more spaces */
					parse_value += strspn(parse_value, whitespace);    	/* skip white */
					yrot = atof(parse_value);
					parse_value = strchr(parse_value, ' ');		/* skip to next space */
				}
				if (parse_value != NULL && strlen(parse_value) > 0) {	/* done if no more spaces */
					parse_value += strspn(parse_value, whitespace);    	/* skip white */
					zrot = atof(parse_value);
					parse_value = strchr(parse_value, ' ');		/* skip to next space */
				}
vrPrintf("vrAssignGenericInput(): 6-sensor r2e angle nudge %lf %lf %lf: '%s' (%#p)\n", parse_value, parse_value, xrot, yrot, zrot);

				/* Now do the rotation in-place */
				vrMatrixPostMult(tmpmat, vrMatrixSetRotationId(vrMatrixCreate(), VR_X, xrot));	/* TODO: a small memory leak here -- the created matrices aren't freed */
				vrMatrixPostMult(tmpmat, vrMatrixSetRotationId(vrMatrixCreate(), VR_Y, yrot));
				vrMatrixPostMult(tmpmat, vrMatrixSetRotationId(vrMatrixCreate(), VR_Z, zrot));
vrPrintMatrix("new r2e: ", tmpmat);
			}

vrPrintMatrix("recalc1: ", tmpmat);
			/* Cause the 6-sensors position to be recalculated */
			vrMatrixGet6sensorRawValuesDirect(&mat_store, (vr6sensor *)input);
			vrAssign6sensorValue((vr6sensor *)input, &mat_store, 0 /* , vrTime time */);

		/*************************************************/
		/*** <6-sensor> "nrt|nudger2et" { id | x y z } ***/
		} else if (!strncasecmp(parse_value, "nudger2et", 9) || !strncasecmp(parse_value, "nrt", 3)) {
			/* the "nudger2eT" command allows for adjusting the 6-sensors calibration */
			/* Form: "nudger2eT|nrt { id | <x-move> <y-move> <z-move> } "             */
			vrMatrix	mat_store;				/* temporary matrix storage */
			vrMatrix	*tmpmat;				/* local pointer to r2e_xform */
			vrVector	 translate = { 0.0, 0.0, 0.0 };		/* translation values */
			char		*id_check;				/* for parsing for "id" */

			parse_value = strchr(&parse_value[2], ' ');				/* skip to a space */
			if (parse_value == NULL) {
				return "Nudge r2e Translate missing arguments.";
			}
vrPrintf("vrAssignGenericInput(): 6-sensor r2e trans nudge: '%s' (%#p)\n", parse_value, parse_value);

			tmpmat = ((vr6sensor *)input)->r2e_xform;		/* point to the r2e matrix */

			/* check whether to reset to identity rotation */
			/* NOTE: I need to leave the whitespace for the parse_value checks below, */
			/*   so I'm using a second, temporary, variable for the "id" check. */
			id_check = parse_value + strspn(parse_value, whitespace);		/* skip white for "id" check */
			if (!strncasecmp(id_check, "id", 2)) {
vrPrintf("vrAssignGenericInput(): 6-sensor r2e trans trans id: '%s' (%#p)\n", parse_value, parse_value);
				vrMatrixSetTranslationOnlyAd(tmpmat, translate.v);
			} else {
				/* otherwise expect up to three rotation values */
				if (parse_value != NULL && strlen(parse_value) > 0) {	/* done if no more spaces */
					parse_value += strspn(parse_value, whitespace);     /* skip white */
					translate.v[VR_X] = atof(parse_value);
					parse_value = strchr(parse_value, ' ');		/* skip to next space */
				}
				if (parse_value != NULL && strlen(parse_value) > 0) {	/* done if no more spaces */
					parse_value += strspn(parse_value, whitespace);    	/* skip white */
					translate.v[VR_Y] = atof(parse_value);
					parse_value = strchr(parse_value, ' ');		/* skip to next space */
				}
				if (parse_value != NULL && strlen(parse_value) > 0) {	/* done if no more spaces */
					parse_value += strspn(parse_value, whitespace);    	/* skip white */
					translate.v[VR_Z] = atof(parse_value);
					parse_value = strchr(parse_value, ' ');		/* skip to next space */
				}
vrPrintf("vrAssignGenericInput(): 6-sensor r2e trans nudge %lf %lf %lf: '%s' (%#p)\n", parse_value, parse_value, translate.v[VR_X], translate.v[VR_Y], translate.v[VR_Z]);

				/* Now do the rotation in-place */
				vrMatrixPostTranslateAd(tmpmat, translate.v);
vrPrintMatrix("new r2e: ", tmpmat);
			}

			/* Cause the 6-sensors position to be recalculated */
vrPrintMatrix("recalc2: ", tmpmat);
			vrMatrixGet6sensorRawValuesDirect(&mat_store, (vr6sensor *)input);
			vrAssign6sensorValue((vr6sensor *)input, &mat_store, 0 /* , vrTime time */);

		} else {
			vrDbgPrintfN(AALWAYS_DBGLVL, "vrAssignGenericInput(): Improper 6-sensor input setting.\n");
			return "Improper 6-sensor input setting.";
		}
		break;

	default:
		break;

	}

	return NULL;	/* NULL indicates no error. */
}


	/**************************************************************/
	/**************************************************************/
	/************ Switch 2 input type access functions ************/


/*****************************************************************************/
/* vrInputSet2switchDescription(): sets the human-readable description text  */
/*   for the given input.  This information is purely for the end-user to be */
/*   provided "help" information that conveys what the given "button" does.  */
/* NOTE: returns 1 when the description is successfully set.  It returns 0   */
/*   when the input_num "button" doesn't exist in current configuration.     */
int vrInputSet2switchDescription(int input_num, const char *input_desc)
{
	vrGenericInput	*input;

	input = vrInputGetFromTypeIndex(vrContext, VRINPUT_BINARY, input_num);
	if (input == NULL) {
		/* Create a new dummy input to handle the request */

		vrDbgPrintfN(AALWAYS_DBGLVL, RED_TEXT "No button (%d) available to '%s' (Creating a dummy)\n" NORM_TEXT, input_num, input_desc);
		input = vrInputAddDummyToInputMap(vrContext->input, VRINPUT_BINARY, input_num);

		strncpy(input->my_object->desc_ui, input_desc, INPUT_UIDESC_LEN);

		return 0;
	}

	strncpy(input->my_object->desc_ui, input_desc, INPUT_UIDESC_LEN);

	return 1;
}


/***********************************************************************/
/* vrAssign2switchValue(): assigns a binary value to the given 2switch */
/*   input value object.  This function is called by the input device  */
/*   implementation code whenever it receives (or simulates) an input  */
/*   from a "button" type input.  Locks are used to protect the        */
/*   validity of the data.                                             */
/* NOTE: The provided value might be any integer, but this routine     */
/*   will convert any non-zero entry to specifically be 1 (true).      */
void vrAssign2switchValue(vr2switch *switch2, int newvalue /* , vrTime time */)
{
	int	assign_value;

	if (switch2->object_type != VROBJECT_INPUTDATA) {
		vrDbgPrintf("Attempt to assign a binary value to a non-input object type %d at %#p\n", switch2->object_type, switch2);
		return;
	}

	if (switch2->input_type != VRINPUT_BINARY) {
		vrDbgPrintf("Attempt to assign a binary value to an incorrect input type %d at %#p\n", switch2->input_type, switch2);
		return;
	}

	/* force value to be binary */
	assign_value = (newvalue != 0);

	/* store the value if it's changed */
	if (switch2->value != assign_value) {
#if 0
vrPrintf("switch %p is getting new value %d\n", switch2, assign_value);
#endif
		/* value has changed */
		vrLockWriteSet(switch2->lock);
		switch2->timestamp = 0.0;		/* TODO: assign timestamp */
		switch2->value = assign_value;
		vrLockWriteRelease(switch2->lock);

		if (switch2->queue_me) {
			/* TODO: put me in the queue */
		}
	}

	/* store historical measurement value */
	/* TODO: determine whether historical value measurements should only be recorded */
	/*   when values change, or always.  For the moment, we will do the "always"     */
	/*   method -- for the "when-change" method this section of code should be moved */
	/*   to be within the above if-statement.                                        */
	if (switch2->num_measures > 0) {
		switch2->current_measure++;
		switch2->current_measure %= switch2->num_measures;

		switch2->measures[switch2->current_measure] = switch2->value;
#if 0
printf("storing historical data for '%s', measure num %d is %d\n", switch2->my_object->name, switch2->current_measure, switch2->measures[switch2->current_measure]);
#endif
	}
}


/*************************************************************************/
/* vrGet2switchValue(): obtains the current value of the given vr2switch */
/*   ("button") in the running system.  The previous value of the input  */
/*   is stored as the "last_value" -- this is used for calculating the   */
/*   delta between calls of this function.  Locks are used to protect    */
/*   the validity of the data.  If the input doesn't exist 0 is returned.*/
int vrGet2switchValue(int input_num)
{
	vr2switch	*input;
	int		value;

	input = (vr2switch *)vrInputGetFromTypeIndex(vrContext, VRINPUT_BINARY, input_num);
	if (input == NULL) {
		/* TODO: this is where we can create new dummy inputs to handle the request */
		return 0;
	}

#if 0
vrTrace("vrGet2switchValue", "pre-read");
#endif
	vrLockReadSet(input->lock);
	value = input->value;
#if 0
vrTrace("vrGet2switchValue", "pre-read2write");
#endif
	vrLockReadToWrite(input->lock);
	input->last_value = value;
#if 0
vrTrace("vrGet2switchValue", "pre-write-rel");
#endif
	vrLockWriteRelease(input->lock);
#if 0
vrTrace("vrGet2switchValue", "end");
#endif

	return value;
}


/**************************************************************************/
/* vrGet2switchValueNoLastUpdate(): similar to vrGet2switchValue() this   */
/*   function returns the current value of the given vr2switch ("button").*/
/*   Locks are used to protect the validity of the data.                  */
/* This version is intended for Rendering processes, so doesn't update    */
/*   "last_value" and therefore won't disrupt the Delta calculations.     */
int vrGet2switchValueNoLastUpdate(int input_num)
{
	vr2switch	*input;
	int		value;

	input = (vr2switch *)vrInputGetFromTypeIndex(vrContext, VRINPUT_BINARY, input_num);
	if (input == NULL) {
		/* TODO: this is where we can create new dummy inputs to handle the request */
		return 0;
	}

	vrLockReadSet(input->lock);
	value = input->value;
	vrLockReadRelease(input->lock);

	return value;
}


/**************************************************************************/
/* vrGet2switchDelta(): returns the delta between the current value of    */
/*   the given vr2switch ("button") and the previous time the value was   */
/*   provided either by this function or by the vrGet2switchValue()       */
/*   function.  This enables the application code to act on events such   */
/*   as button-pressed and button-released, versus whether the button is  */
/*   currently pressed or not.  As with vrGet2switchValue(), the previous */
/*   value of the input is stored as the "last_value".  Locks are used to */
/*   protect the validity of the data.  If the input doesn't exist 0 is   */
/*   returned.                                                            */
int vrGet2switchDelta(int input_num)
{
	vr2switch	*input;
	int		delta;

	input = (vr2switch *)vrInputGetFromTypeIndex(vrContext, VRINPUT_BINARY, input_num);
	if (input == NULL) {
		/* TODO: this is where we can create new dummy inputs to handle the request */
		return 0;
	}

	vrLockReadSet(input->lock);
	delta = input->value - input->last_value;
	vrLockReadToWrite(input->lock);
	input->last_value = input->value;
	vrLockWriteRelease(input->lock);

	return delta;
}


/****************************************************************************/
/* vrGet2switchValueDirect(): similar to vrGet2switchValue(), this function */
/*   returns the current value of the given vr2switch ("button").  In this  */
/*   case however, rather than providing the number of the button within    */
/*   the FreeVR system, a direct pointer to the input is provided as the    */
/*   argument.                                                              */
/* NOTE: as of 09/28/21 (and going back at least to 01/13/09), this         */
/*   function is not used -- perhaps it can be deprecated.                  */
int vrGet2switchValueDirect(vr2switch *input)
{
	int		value;

	if (input == NULL) {
		if (vrDbgDo(CONFIG_WARN_DBGLVL)) {
			vrErrPrintf("vrGet2switchValueDirect(): Invalid NULL Binary input.\n");
		}
		return 0;
	}
	vrLockReadSet(input->lock);
	value = input->value;
	vrLockReadToWrite(input->lock);
	input->last_value = value;
	vrLockWriteRelease(input->lock);

	return value;
}


/******************************************************************************/
/* vrGet2switchDeltaDirect(): similar to vrGet2switchDelta(), this function   */
/*   provides the delta between now and the previous time this vr2switch      */
/*   ("button") value was requested. And similar to vrGet2switchValueDirect() */
/*   this version of requesting the input takes as an argument a direct       */
/*   pointer to the input rather than a FreeVR reference number.              */
/* NOTE: as of 09/28/21 (and going back at least to 01/13/09), this function  */
/*   is not used -- perhaps it can be deprecated.                             */
int vrGet2switchDeltaDirect(vr2switch *input)
{
	int		delta;

	if (input == NULL) {
		if (vrDbgDo(CONFIG_WARN_DBGLVL)) {
			vrErrPrintf("vrGet2switchDeltaDirect(): Invalid NULL Binary input.\n");
		}
		return 0;
	}
	vrLockReadSet(input->lock);
	delta = input->value - input->last_value;
	vrLockReadToWrite(input->lock);
	input->last_value = input->value;
	vrLockWriteRelease(input->lock);

	return delta;
}


	/**************************************************************/
	/**************************************************************/
	/************ Switch N input type access functions ************/


/*****************************************************************************/
/* vrInputSetNswitchDescription(): sets the human-readable description text  */
/*   for the given input.  This information is purely for the end-user to be */
/*   provided "help" information that conveys what the given "button" does.  */
/* NOTE: returns 1 when the description is successfully set.  It returns 0   */
/*   when the input_num "button" doesn't exist in current configuration.     */
int vrInputSetNswitchDescription(int input_num, const char *input_desc)
{
	vrGenericInput	*input;

	input = vrInputGetFromTypeIndex(vrContext, VRINPUT_NWAY, input_num);
	if (input == NULL) {
		/* Create a new dummy input to handle the request */

		vrDbgPrintfN(AALWAYS_DBGLVL, RED_TEXT "No N-switch (%d) available to '%s' (Creating a dummy)\n" NORM_TEXT, input_num, input_desc);
		input = vrInputAddDummyToInputMap(vrContext->input, VRINPUT_NWAY, input_num);

		strncpy(input->my_object->desc_ui, input_desc, INPUT_UIDESC_LEN);

		return 0;
	}

	strncpy(input->my_object->desc_ui, input_desc, INPUT_UIDESC_LEN);

	return 1;
}


/*************************************************************************/
/* vrAssignNswitchValue(): assigns an integer value to the given Nswitch */
/*   input value object.  This function is called by the input device    */
/*   implementation code whenever it receives (or simulates) an input    */
/*   from a "knob" type input.  Locks are used to protect the validity   */
/*   of the data.                                                        */
void vrAssignNswitchValue(vrNswitch *switchN, int newvalue /* , vrTime time */)
{
	if (switchN->object_type != VROBJECT_INPUTDATA) {
		vrDbgPrintf("Attempt to assign a N-switch value to a non-input object type %d at %#p\n", switchN->object_type, switchN);
		return;
	}

	if (switchN->input_type != VRINPUT_NARY) {
		vrDbgPrintf("Attempt to assign a N-switch value to an incorrect input type %d at %#p\n", switchN->input_type, switchN);
		return;
	}

	/* store the value if it's changed */
	if (switchN->value != newvalue) {
		/* value has changed */
		vrLockWriteSet(switchN->lock);
		switchN->timestamp = 0.0;		/* TODO: assign timestamp */
		switchN->value = newvalue;
		vrLockWriteRelease(switchN->lock);

		if (switchN->queue_me) {
			/* TODO: put me in the queue */
		}
	}
}


/*************************************************************************/
/* vrGetNswitchValue(): obtains the current value of the given vrNswitch */
/*   ("knob") in the running system.  The previous value of the input    */
/*   is stored as the "last_value" -- this is used for calculating the   */
/*   delta between calls of this function.  Locks are used to protect    */
/*   the validity of the data.  If the input doesn't exist 0 is returned.*/
int vrGetNswitchValue(int input_num)
{
	vrNswitch	*input;
	int		value;

	input = (vrNswitch *)vrInputGetFromTypeIndex(vrContext, VRINPUT_NWAY, input_num);
	if (input == NULL) {
		/* TODO: this is where we can create new dummy inputs to handle the request */
		return 0;
	}

	vrLockReadSet(input->lock);
	value = input->value;
	vrLockReadToWrite(input->lock);
	input->last_value = value;
	vrLockWriteRelease(input->lock);

	return value;
}


/**************************************************************************/
/* vrGetNswitchValueNoLastUpdate(): similar to vrGetNswitchValue() this   */
/*   function returns the current value of the given vrNswitch ("knob").  */
/*   Locks are used to protect the validity of the data.                  */
/* This version is intended for Rendering processes, so doesn't update    */
/*   "last_value" and therefore won't disrupt the Delta calculations.     */
int vrGetNswitchValueNoLastUpdate(int input_num)
{
	vrNswitch	*input;
	int		value;

	input = (vrNswitch *)vrInputGetFromTypeIndex(vrContext, VRINPUT_NWAY, input_num);
	if (input == NULL) {
		/* TODO: this is where we can create new dummy inputs to handle the request */
		return 0;
	}

	vrLockReadSet(input->lock);
	value = input->value;
	vrLockReadRelease(input->lock);

	return value;
}


/*************************************************************************/
/* vrGetNswitchDelta(): returns the delta between the current value of   */
/*   the given vr2switch ("knob") and the previous time the value was    */
/*   provided either by this function or by the vrGetNswitchValue()      */
/*   function.  This enables the application code to act on events such  */
/*   as knob-rotated-right or knob-rotated-left, versus the actual knob  */
/*   setting.  As with vrGetNswitchValue(), the previous value of the    */
/*   input is stored as the "last_value".  Locks are used to protect the */
/*   validity of the data.  If the input doesn't exist 0 is returned.    */
int vrGetNswitchDelta(int input_num)
{
	vrNswitch	*input;
	int		delta;

	input = (vrNswitch *)vrInputGetFromTypeIndex(vrContext, VRINPUT_NWAY, input_num);
	if (input == NULL) {
		/* TODO: this is where we can create new dummy inputs to handle the request */
		return 0;
	}

	vrLockReadSet(input->lock);
	delta = input->value - input->last_value;
	vrLockReadToWrite(input->lock);
	input->last_value = input->value;
	vrLockWriteRelease(input->lock);

	return delta;
}

/* NOTE: 09/28/21 (and going back at least to 01/13/09), these two functions */
/*   are not used, so perhaps can be deprecated.  This is true of all the    */
/*   input types except vr6sensors for which the "ValuesDirect()" function   */
/*   is used.                                                                */
/**********************************************************************/
int vrGetNswitchValueDirect(vr2switch *input) { return -10; }
/**********************************************************************/
int vrGetNswitchDeltaDirect(vr2switch *input) { return -10; }


	/**************************************************************/
	/**************************************************************/
	/************ Valuator input type access functions ************/


/*****************************************************************************/
/* vrInputSetValuatorDescription(): sets the human-readable description text */
/*   for the given input.  This information is purely for the end-user to be */
/*   provided "help" information that conveys what the given "button" does.  */
/* NOTE: returns 1 when the description is successfully set.  It returns 0   */
/*   when the input_num "button" doesn't exist in current configuration.     */
int vrInputSetValuatorDescription(int input_num, const char *input_desc)
{
	vrGenericInput	*input;

	input = vrInputGetFromTypeIndex(vrContext, VRINPUT_VALUATOR, input_num);
	if (input == NULL) {
		/* Create a new dummy input to handle the request */

		vrDbgPrintfN(AALWAYS_DBGLVL, "FreeVR (%.4f): " RED_TEXT "No Valuator (%d) available to '%s' (Creating a dummy)\n" NORM_TEXT,
			vrCurrentSimTime(), input_num, input_desc);
		input = vrInputAddDummyToInputMap(vrContext->input, VRINPUT_VALUATOR, input_num);

		strncpy(input->my_object->desc_ui, input_desc, INPUT_UIDESC_LEN);

		return 0;
	}

	strncpy(input->my_object->desc_ui, input_desc, INPUT_UIDESC_LEN);

	return 1;
}


/************************************************************************/
/* vrAssignValuatorValue(): assigns a float value to the given valuator */
/*   input value object.  This function is called by the input device   */
/*   implementation code whenever it receives (or simulates) an input   */
/*   from a "joystick" type input.  Locks are used to protect the       */
/*   validity of the data.                                              */
void vrAssignValuatorValue(vrValuator *valuator, float newvalue /* , vrTime time */)
{
	if (valuator->object_type != VROBJECT_INPUTDATA) {
		vrDbgPrintf("Attempt to assign a valuator value to a non-input object type %d at %#p\n", valuator->object_type, valuator);
		return;
	}

	if (valuator->input_type != VRINPUT_VALUATOR) {
		vrDbgPrintf("Attempt to assign a valuator value to an incorrect input type %d at %#p\n", valuator->input_type, valuator);
		return;
	}

	/* store the value if it's changed */
	if (valuator->value != newvalue) {
		/* value has changed */
		vrLockWriteSet(valuator->lock);
		valuator->timestamp = 0.0;		/* TODO: assign timestamp */
		valuator->value = newvalue;
		vrLockWriteRelease(valuator->lock);

		if (valuator->queue_me) {
			/* TODO: put me in the queue */
		}
	}

	/* store historical measurement value */
	/* TODO: determine whether historical value measurements should only be recorded */
	/*   when values change, or always.  For the moment, we will do the "always"     */
	/*   method -- for the "when-change" method this section of code should be moved */
	/*   to be within the above if-statement.                                        */
	if (valuator->num_measures > 0) {
		valuator->current_measure++;
		valuator->current_measure %= valuator->num_measures;

		valuator->measures[valuator->current_measure] = valuator->value;
#if 0
printf("storing historical data for '%s', measure num %d is %d\n", valuator->my_object->name, valuator->current_measure, valuator->measures[valuator->current_measure]);
#endif
	}
}


/***************************************************************************/
/* vrGetValuatorValue(): obtains the current value of the given vrValuator */
/*   ("joystick") in the running system.  The previous value of the input  */
/*   is stored as the "last_value" -- this is used for calculating the     */
/*   delta between calls of this function.  Locks are used to protect      */
/*   the validity of the data.  If the input doesn't exist 0 is returned.  */
float vrGetValuatorValue(int input_num)
{
	vrValuator	*input;
	float		value;

	input = (vrValuator *)vrInputGetFromTypeIndex(vrContext, VRINPUT_VALUATOR, input_num);
	if (input == NULL) {
		/* TODO: this is where we can create new dummy inputs to handle the request */
		return 0;
	}

	vrLockReadSet(input->lock);
	value = input->value;
	vrLockReadToWrite(input->lock);
	input->last_value = value;
	vrLockWriteRelease(input->lock);

	return value;
}


/**************************************************************************/
/* vrGetValuatorValueNoLastUpdate(): similar to vrGetValuatorValue() this */
/*   function returns the current value of the given vrValuator           */
/*   ("joystick").  Locks are used to protect the validity of the data.   */
/* This version is intended for Rendering processes, so doesn't update    */
/*   "last_value" and therefore won't disrupt the Delta calculations.     */
float vrGetValuatorValueNoLastUpdate(int input_num)
{
	vrValuator	*input;
	float		value;

	input = (vrValuator *)vrInputGetFromTypeIndex(vrContext, VRINPUT_VALUATOR, input_num);
	if (input == NULL) {
		/* TODO: this is where we can create new dummy inputs to handle the request */
		return 0;
	}

	vrLockReadSet(input->lock);
	value = input->value;
	vrLockReadRelease(input->lock);

	return value;
}


/*************************************************************************/
/* vrGetValuatorDelta(): returns the delta between the current value of  */
/*   the given vrValuator ("joystick") and the previous time the value   */
/*   was provided either by this function or by the vrGetValautorValue() */
/*   function.  This enables the application code to act on events such  */
/*   as stick-pushed-forward or stick-pulled-back (and at what rate).    */
/*   As with vrGetValuatorValue(), the previous value of the input is    */
/*   stored as the "last_value".  Locks are used to protect the validity */
/*   of the data.  If the input doesn't exist 0 is returned.             */
float vrGetValuatorDelta(int input_num)
{
	vrValuator	*input;
	float		delta;

	input = (vrValuator *)vrInputGetFromTypeIndex(vrContext, VRINPUT_VALUATOR, input_num);
	if (input == NULL) {
		/* TODO: this is where we can create new dummy inputs to handle the request */
		return 0;
	}

	vrLockReadSet(input->lock);
	delta = input->value - input->last_value;
	vrLockReadToWrite(input->lock);
	input->last_value = input->value;
	vrLockWriteRelease(input->lock);

	return delta;
}


/*** These place-holder functions return an obviously incorrect value ***/
/**********************************************************************/
float vrGetValuatorValueDirect(vrValuator *input) { return -10.0; }
/**********************************************************************/
float vrGetValuatorDeltaDirect(vrValuator *input) { return -10.0; }


	/**************************************************************/
	/**************************************************************/
	/************ Sensor 6 input type access functions ************/


/*****************************************************************************/
/* vrInputSet6sensorDescription(): sets the human-readable description text  */
/*   for the given input.  This information is purely for the end-user to be */
/*   provided "help" information that conveys what the given "button" does.  */
/* NOTE: returns 1 when the description is successfully set.  It returns 0   */
/*   when the input_num "button" doesn't exist in current configuration.     */
int vrInputSet6sensorDescription(int input_num, const char *input_desc)
{
	vrGenericInput	*input;

	input = vrInputGetFromTypeIndex(vrContext, VRINPUT_6SENSOR, input_num);
	if (input == NULL) {
		/* Create a new dummy input to handle the request */

		vrDbgPrintfN(AALWAYS_DBGLVL, "FreeVR (%.4f): " RED_TEXT "No Position Sensor (%d) available to '%s' (Creating a dummy)\n" NORM_TEXT,
			vrCurrentSimTime(), input_num, input_desc);
		input = vrInputAddDummyToInputMap(vrContext->input, VRINPUT_6SENSOR, input_num);

		strncpy(input->my_object->desc_ui, input_desc, INPUT_UIDESC_LEN);

		return 0;
	}

	strncpy(input->my_object->desc_ui, input_desc, INPUT_UIDESC_LEN);

	return 1;
}


/**************************************************************************/
/* vrAssign6sensorR2Exform() takes a 6-sensor input and the post-comma    */
/*   portion of the instance (of the DTI) string, and sets the r2e matrix */
/*   to either the identity matrix, the current r2e value or (in the      */
/*   future) a matrix specified by 16 values (TODO: the latter not yet    */
/*   implemented).  For the latter two options (ie. the non-identity      */
/*   matrix ones), the new r2e matrix is used to reassign the current     */
/*   matrix raw values -- this causes the newly set R2E values to take    */
/*   immediate affect.                                                    */
/* NOTE: repeat this for ipos -- I think  -- rename to vrAssign6sensorInstanceOptions() */
void vrAssign6sensorR2Exform(vr6sensor *sensor6, const char *xform_info)
{
static	char		*whitespace = " \t\n\r\b";
	vrMatrix	mat_store;

	/* If no string provided, then just leave the identity matrix in place */
	if (xform_info == NULL)
		return;

	/* If string is zero-length, then just leave the identity matrix in place */
	if (strlen(xform_info) == 0)
		return;

	/* skip the first comma and then any whitespace at the beginning of the string */
	xform_info++;					/* skip comma */
	xform_info += strspn(xform_info, whitespace);	/* skip whitespace */

	/* If we hit a comma before any data, then skip ahead to the initial position code */
	if (xform_info[0] != ',') {
		/* If value is "id" then just leave the identity matrix in place */
		if (!strncasecmp(xform_info, "id", 2)) {
			/* no-op (but skip the "id") */;
			xform_info += 2;
		} else

		if (!strncasecmp(xform_info, "r2e", 3)) {
			/* copy the R2E value in the input device to the input container */
			xform_info += 3;
			vrMatrixCopy(sensor6->r2e_xform, sensor6->my_object->r2e_xform);
			vrAssign6sensorValue(sensor6, vrMatrixGet6sensorRawValuesDirect(&mat_store, sensor6), -1 /* , vrCurrentWallTime() */);
		} else

		if (!strncasecmp(xform_info, "xform:", 6)) {
			/* input a full 16-value matrix into the R2E data (NYI) */
			xform_info += 6;
			vrErrPrintf("vrAssign6sensorR2Exform(): " RED_TEXT "Sorry, 'xform:' option not yet implemented, using identity matrix.\n" NORM_TEXT);
			vrAssign6sensorValue(sensor6, vrMatrixGet6sensorRawValuesDirect(&mat_store, sensor6), -1 /* , vrCurrentWallTime() */);
		}
	}
	xform_info = strchr(xform_info, ',');		/* skip to next comma */

	/* If there was no comma, then xform_info will be NULL, and we're done.*/
	if (xform_info == NULL) {
		return;
	}

	/* If we hit a comma, then we're expecting an ipos argument */
	if (xform_info[0] == ',') {	/* this should ALWAYS be true */
		xform_info++;					/* skip the comma */
		xform_info += strspn(xform_info, whitespace);	/* skip whitespace */
		//vrPrintf("IPOS: xform_info is now '%s'\n", xform_info);	// delete me

		if (!strncasecmp(xform_info, "id", 2)) {
			/* no-op (but skip the "id") */;
			xform_info += 2;
		} else
		if (!strncasecmp(xform_info, "ipos", 4)) {
			/* copy the IPos value in the input device to the input container */
			xform_info += 4;
//vrPrintf("Copying the ipos %p\n", sensor6->my_device->ipos_xform);	/* delete me */
//vrPrintMatrix("ipos", sensor6->my_device->ipos_xform);	/* delete me */
			vrMatrixCopy(sensor6->ipos_xform, sensor6->my_object->ipos_xform);
//vrPrintMatrix("ipos-copy", sensor6->ipos_xform);	/* delete me */
//vrFprintInputObject(stdout, sensor6->my_object, verbose);	/* delete me */
//vrFprintInputValue(stdout, (vrGenericInput *)sensor6, verbose);	/* delete me */
			vrAssign6sensorValue(sensor6, vrMatrixGet6sensorRawValuesDirect(&mat_store, sensor6), -1 /* , vrCurrentWallTime() */);
		} else
		if (!strncasecmp(xform_info, "xform:", 6)) {
			/* input a full 16-value matrix into the IPos data (NYI) */
			xform_info += 6;
			vrErrPrintf("vrAssign6sensorR2Exform(): " RED_TEXT "Sorry, 'xform:' option not yet implemented, using identity matrix.\n" NORM_TEXT);
			vrAssign6sensorValue(sensor6, vrMatrixGet6sensorRawValuesDirect(&mat_store, sensor6), -1 /* , vrCurrentWallTime() */);
		}
	} else {
		/* As the message says, we should NEVER get here due to the above logic. */
		vrErrPrintf("vrAssign6sensorR2Exform(): " RED_TEXT "Strange error in parsing ipos -- this should not happen.\n" NORM_TEXT);
	}
}


/**************************************************************************/
/* vrAssign6sensorValue(): assigns a new (presumably) value to a 6-sensor */
/*   type of input.  The first argument (sensor6) is a pointer to the     */
/*   input which is to be changed.  The second argument (new_mat) is a    */
/*   vrMatrix type with the new position values of the 6sensor input.  A  */
/*   third argument (oob) is a boolean that indicates whether the 6sensor */
/*   is currently out-of-bounds -- often used to flag a (simulated)       */
/*   representation of that input in another color (e.g. red).            */
/*                                                                        */
/* NOTE: "new_mat" is in the input device's (tracker's) coordinate system.*/
/*   It is the responsibility of this function to map that matrix         */
/*   into the coordinate system of the VR system (ie. the "real world"    */
/*   coordinate system).                                                  */
/* NOTE: values of oob less than zero are treated as a nop (ie. do not    */
/*   change) flag.                                                        */
/* NOTE: in general this function is called with "oob" set to "0".  In    */
/*   fact, the only place I presently (11/30/20) find where it is set     */
/*   to 1 (one) is in vrAssign6sensorValueFromValuators() when it does    */
/*   the out-of-bounds calculation.  (I.e. it isn't clear that this       */
/*   argument is actually needed.)                                        */
void vrAssign6sensorValue(vr6sensor *sensor6, vrMatrix *new_mat, int oob /* , vrTime time */)
{
	if (sensor6->object_type != VROBJECT_INPUTDATA) {
		vrDbgPrintf("Attempt to assign a position sensor value to a non-input object type %d at %#p\n", sensor6->object_type, sensor6);
		return;
	}

	if (sensor6->input_type != VRINPUT_6SENSOR) {
		vrDbgPrintf("Attempt to assign a position sensor value to an incorrect input type %d at %#p\n",
			sensor6->input_type, sensor6);
		return;
	}

	if (new_mat == NULL) {
		vrErr("vrAssign6sensorValue: Attempted to assign a Null pointer.\n");
		vrPrintf("Pausing: Feel free to use 'gdb -p %d' now to find out where\n", getpid());
		pause();
	}

#if 0 /* checksums not yet implemented */
	/* TODO: checksum check here -- I fixed the bug I was looking for, */
	/*   so I didn't implement this, but we may want to in the future. */
#endif

	if (1 /* here's where we assume the value has changed*/) {
		/* TODO: If there is a quick memory-checksum function, we    */
		/*   could add a checksum field for quickly determining      */
		/*   whether data has changed.  If the checksums match, then */
		/*   a slightly slower, brute force check should be done.    */

		/* assume (for now) value has changed */
		vrLockWriteSet(sensor6->lock);
		if (oob >= 0)
			sensor6->oob = oob;
		sensor6->timestamp = 0.0;		/* TODO: assign timestamp */
		vrMatrixCopy(sensor6->raw_data, new_mat);

#if 0
		incoming_sensor6.calibrated = 0;
		/* now calibrate the data */
		/* TODO: the calibration stuff */

		/* we may want to use callbacks for the calibration function */
#endif

		/* now transform the data into world space */
//vrPrintMatrix("about to copy from raw_data: ", sensor6->raw_data);
		vrMatrixCopy(sensor6->position, sensor6->raw_data);
/* 01/14/21: I'm still not convinced that I have the input matrix applications*/	/* eventually delete me -- (just the wrong ways) */
/*   done properly, so I'm making a new grouping of how the transformations   */
/*   are applied to inputs -- there are really 8 ways this can be done.       */
#define T2_FIRST	/* #define is original */
#define T2_PRE		/* #define is original */
#undef R2_PRE		/* #undef is original */
#if 0 /* 04/08/21: debbuging code for the r2e matrix */ /* delete me */
vrPrintMatrix("INIT: assign 6-sensor: position:", sensor6->position);
vrEuler	euler;
vrEulerSetFromMatrix(&euler, sensor6->position);
vrPrintEuler("INIT: assign 6-sensor: position:", &euler);
vrPrintMatrixComposition("INIT: assign 6-sensor: position:", sensor6->position); 
#endif
/* delete me -- the stuff that isn't used, but keep the stuff that is used */
#ifdef T2_FIRST

#  ifdef T2_PRE
		vrMatrixPreMult(sensor6->position, sensor6->t2rw_xform); /* TODO: I prefer using sensor6->my_device->t2rw_xform   */ /* fix me */ /* default */
#  else
		vrMatrixPostMult(sensor6->position, sensor6->t2rw_xform); /* TODO: I prefer using sensor6->my_device->t2rw_xform  */ /* fix me */
#  endif
#  ifdef R2_PRE
		vrMatrixPreMult(sensor6->position, sensor6->r2e_xform); /* TODO: the correctness of this is in question -- always tested with ident mat */
#  else
		vrMatrixPostMult(sensor6->position, sensor6->r2e_xform); /* TODO: the correctness of this is in question -- always tested with ident mat */ /* default */
#  endif

#else /* T2_FIRST vs. second */

#  ifdef R2_PRE
		vrMatrixPreMult(sensor6->position, sensor6->r2e_xform); /* TODO: the correctness of this is in question -- always tested with ident mat */
#  else
		vrMatrixPostMult(sensor6->position, sensor6->r2e_xform); /* TODO: the correctness of this is in question -- always tested with ident mat */
#  endif
#  ifdef T2_PRE
		vrMatrixPreMult(sensor6->position, sensor6->t2rw_xform); /* TODO: I prefer using sensor6->my_device->t2rw_xform   */ /* fix me */
#  else
		vrMatrixPostMult(sensor6->position, sensor6->t2rw_xform); /* TODO: I prefer using sensor6->my_device->t2rw_xform  */ /* fix me */
#  endif
#endif

#if 0 /* 04/08/21: debugging code for the r2e matrix */ /* delete me */
vrPrintMatrix("assign 6-sensor: r2e_xform:", sensor6->r2e_xform);
vrEulerSetFromMatrix(&euler, sensor6->r2e_xform);
vrPrintEuler("assign 6-sensor: r2e_xform:", &euler);
vrPrintMatrixComposition("assign 6-sensor: r2e_xform:", sensor6->r2e_xform); 

vrPrintMatrix("assign 6-sensor: position:", sensor6->position);
vrEulerSetFromMatrix(&euler, sensor6->position);
vrPrintEuler("assign 6-sensor: position:", &euler);
vrPrintMatrixComposition("POST: assign 6-sensor: position:", sensor6->position); 
#endif


	/***********************************************************************/
#if 0 /* This is the pre-01/14/21 code */ /* eventually delete me */
#if 1 /* 2/2/01 test -- top version is original -- after testing, I'm fairly certain the original was correct */ /* 01/14/21: testing the bottom version after correcting quaternion rotation creation */
		vrMatrixPreMult(sensor6->position, sensor6->t2rw_xform); /* TODO: I prefer using sensor6->my_device->t2rw_xform   */
#else
		vrMatrixPostMult(sensor6->position, sensor6->t2rw_xform); /* TODO: I prefer using sensor6->my_device->t2rw_xform  */
#endif
									/*   because then we can change the device's xform w/o    */
									/*   needing to copy it to all the inputs.  However,      */
									/*   because the pre-config, dummy user creates a 6sensor */
									/*   without an inputdevice, my_device is NULL for this   */
									/* TODO: use the newly created (02/22/2008) _MakeDummyInputDevice() */
									/*   function to have a dummy input device for the dummy  */
									/*   6-sensor of the dummy user.                          */

#if 0 /* this first test is wrong -- 2/2/01 testing original again, still doesn't seem to work */ /* delete me */
		vrMatrixPreMult(sensor6->position, sensor6->r2e_xform); /* TODO: the correctness of this is in question -- always tested with ident mat */
#else /* this seems to work fine. */
	/* 2/11/00 -- actually, this doesn't work either. rotations about the object */
	/*   (specifically xwindows sim-6), rotation about the origin, not the sensor */
		vrMatrixPostMult(sensor6->position, sensor6->r2e_xform); /* TODO: the correctness of this is in question -- always tested with ident mat */
#endif
#endif /* end of pre-01/14/21 code */
#if 0
		incoming_sensor6.frame_of_reference = 0;  /* TODO: set this to world space */
#endif

		vrLockWriteRelease(sensor6->lock);

		if (sensor6->queue_me) {
			/* TODO: put me in the queue */
		}
	}
}


/**************************************************************************/
/* vrAssign6sensorValueFromValuators(): this function is used to effect   */
/*   the control of a simulated 6sensor (provided as the first argument). */
/*   The second argument ("valuators") is a 6-element array of float      */
/*   values, the first three of which are the translational movements,    */
/*   and the final three of which are rotational movements.  The third    */
/*   argument ("conv_options") is a special structure with information    */
/*   of how the simulated 6sensor is to be controlled -- how the inputs   */
/*   in "valuators" are to be interpreted/converted.  Finally (for now),  */
/*   the fourth argument ("oob") specifies whether to set the 6sesnors    */
/*   out-of-bounds flag (boolean, plus -1 for nop).  This is merely       */
/*   passed along to the call to vrAssign6sensorValue() at the end of     */
/*   the function.                                                        */
/*                                                                        */
/* NOTE: presently (11/30/20) all calls to this function set oob to "-1". */
/*   Which implies that this argument may not be necessary.               */
/* NOTE: the array of valuators is of type (float) in order to work in    */
/*   conjunction with the vrEuler type.  If vrEuler changes to use        */
/*   doubles, then this function should also be similarly changed.        */
/* NOTE: presently (09/21/21) this function doesn't make use of the R2E   */
/*   value.  TODO: determine whether it should.                           */
/*                                                                        */
/* Here are what all the vr6SensorConv (conversion options) do:           */
/*   ignore_all -- don't actually make changes to the simulated 6sensor   */
/*   ignore_trans -- don't make any changes to the 6sensor's translations */
/*   restrict_space -- restrain translations to within the working volume */
/*   return_to_zero -- don't accumulate movements (vs. do accumulate)     */
/*   relative_axis -- base movements on sensor's CS (vs. real-world CS)   */
/*   azimuth_axis -- cardinal axis around which general rotations occur   */
/*   swap_yz -- swap movements between Y(up) & Z(in) axes                 */
/*   swap_transrot -- whether to swap translation and rotation values     */
/*   trans_scale -- translational scaling factor                          */
/*   rot_scale -- rotational scaling factor                               */
/*   working-volume minimums -- minimum X,Y,Z values of working volume    */
/*   working-volume maximums -- maximum X,Y,Z values of working volume    */
void vrAssign6sensorValueFromValuators(vr6sensor *sensor6, float *valuators, vr6sensorConv *conv_options, int oob /* , vrTime time */)
{
static	vr6sensorConv	default_options = { 0, 0, 0, 0, 0, VR_Y, 0, 0, 1.0, 1.0, { -5.0, 0.0, -5.0 }, { 5.0, 10.0, 5.0 } };
	vrMatrix	mat_store;		/* temporary matrix storage */
	vrMatrix	*mat = &mat_store;	/* pointer to temp matrix storage */
	float		*valuators_trans;	/* pointer to the translational data */
	float		*valuators_rot;		/* pointer to the rotational data */
	float		valuators_tmp[6];	/* holder for valuator values to move them around a bit */
	float		swapper;		/* place to hold a value while swapping */
	float		t_scale;		/* local copy of translational scaling factor */
	float		r_scale;		/* local copy of rotational scaling factor */

	if (conv_options == NULL)
		conv_options = &default_options;

	if (conv_options->ignore_all)
		return;

	/* make local copies of trans & rot _scale, use 1.0 if scale factor is 0.0 */
	/* TODO: once all the input devices have been updated to use the trans & rot factors, we can get rid of the check for zero */
	/* TODO: the t_scale and r_scale factors should ideally take time into account */
	t_scale = conv_options->trans_scale;
	if (t_scale == 0.0)
		t_scale = 1.0;

	r_scale = conv_options->rot_scale;
	if (r_scale == 0.0)
		r_scale = 1.0;

	/**********************************************************************/
	/*** manipulate the "valuators" based on 6sensor conversion options ***/

	/* The 6 "valuators" values are copied into "valuators_tmp" based */
	/*   on the current "conv_options" (type vr6sensorConv) settings. */
	if (conv_options->swap_transrot) {
		/* swapped */

		/* this is a kludge swap that works well for the way I've configured a joydev device */
		/* TODO: implementing input maps would provide for a more generic way of accomplishing this */
		valuators_tmp[VR_X] = -valuators[VR_AZIM+3];
		valuators_tmp[VR_Z] =  valuators[VR_ELEV+3];
		valuators_tmp[VR_Y] = -valuators[VR_ROLL+3];
		valuators_tmp[VR_AZIM+3] = -valuators[VR_X];
		valuators_tmp[VR_ELEV+3] =  valuators[VR_Z];
		valuators_tmp[VR_ROLL+3] = -valuators[VR_Y];
	} else {
		/* not swapped */
		valuators_tmp[VR_X] = valuators[VR_X];
		valuators_tmp[VR_Y] = valuators[VR_Y];
		valuators_tmp[VR_Z] = valuators[VR_Z];
		valuators_tmp[VR_AZIM+3] = valuators[VR_AZIM+3];
		valuators_tmp[VR_ELEV+3] = valuators[VR_ELEV+3];
		valuators_tmp[VR_ROLL+3] = valuators[VR_ROLL+3];
	}

	if (conv_options->swap_yz) {
		swapper = valuators[VR_Z];
		valuators_tmp[VR_Z] = -valuators[VR_Y];
		valuators_tmp[VR_Y] = -swapper;

		swapper = valuators[VR_ELEV+3];
		valuators_tmp[VR_ELEV+3] = valuators[VR_ROLL+3];
		valuators_tmp[VR_ROLL+3] = swapper;
	}

	valuators_trans = &valuators_tmp[0];
	valuators_rot   = &valuators_tmp[3];

	/************************************************************/
	/*** Set the starting matrix based on return_to_zero flag ***/

	/* The non-raw matrix has the tracker xform incorporated */
	/*   so using that causes that xform to be added to the  */
	/*   position every time.  And that's bad.               */
	if (conv_options->return_to_zero)
		vrMatrixSetIdentity(mat);
	else	vrMatrixGet6sensorRawValuesDirect(mat, sensor6);


	/****************************************/
	/*** calculate the sensor translation ***/
	if (!conv_options->ignore_trans) {
		/* only map the translation to the sensor when not doing valuator_override */
		if (conv_options->relative_axis) {
			/* do translation using the sensor's axes */
			vrMatrixPostTranslate3d(mat, valuators_trans[VR_X]*t_scale , valuators_trans[VR_Y]*t_scale, valuators_trans[VR_Z]*t_scale);
		} else {
			/* do translation using the world's axes */
			vrMatrixPreTranslate3d(mat, valuators_trans[VR_X]*t_scale, valuators_trans[VR_Y]*t_scale, valuators_trans[VR_Z]*t_scale);
		}
		if (conv_options->restrict_space) {
			/* compare the location with the given working volume, and move inside if nece */
			if (VRMAT_ROWCOL(mat, VR_X, VR_W) > conv_options->working_volume_max[VR_X])
				VRMAT_ROWCOL(mat, VR_X, VR_W) = conv_options->working_volume_max[VR_X];
			if (VRMAT_ROWCOL(mat, VR_X, VR_W) < conv_options->working_volume_min[VR_X])
				VRMAT_ROWCOL(mat, VR_X, VR_W) = conv_options->working_volume_min[VR_X];
			if (VRMAT_ROWCOL(mat, VR_Y, VR_W) > conv_options->working_volume_max[VR_Y])
				VRMAT_ROWCOL(mat, VR_Y, VR_W) = conv_options->working_volume_max[VR_Y];
			if (VRMAT_ROWCOL(mat, VR_Y, VR_W) < conv_options->working_volume_min[VR_Y])
				VRMAT_ROWCOL(mat, VR_Y, VR_W) = conv_options->working_volume_min[VR_Y];
			if (VRMAT_ROWCOL(mat, VR_Z, VR_W) > conv_options->working_volume_max[VR_Z])
				VRMAT_ROWCOL(mat, VR_Z, VR_W) = conv_options->working_volume_max[VR_Z];
			if (VRMAT_ROWCOL(mat, VR_Z, VR_W) < conv_options->working_volume_min[VR_Z])
				VRMAT_ROWCOL(mat, VR_Z, VR_W) = conv_options->working_volume_min[VR_Z];

			sensor6->oob = 0; /* it must be in-bounds, since we forced it to be */
		} else {
			/* determine whether location is out-of-bounds */
			if ((VRMAT_ROWCOL(mat, VR_X, VR_W) > conv_options->working_volume_max[VR_X])
				|| (VRMAT_ROWCOL(mat, VR_X, VR_W) < conv_options->working_volume_min[VR_X])
				|| (VRMAT_ROWCOL(mat, VR_Y, VR_W) > conv_options->working_volume_max[VR_Y])
				|| (VRMAT_ROWCOL(mat, VR_Y, VR_W) < conv_options->working_volume_min[VR_Y])
				|| (VRMAT_ROWCOL(mat, VR_Z, VR_W) > conv_options->working_volume_max[VR_Z])
				|| (VRMAT_ROWCOL(mat, VR_Z, VR_W) < conv_options->working_volume_min[VR_Z])) {
				sensor6->oob = 1;
			} else {
				sensor6->oob = 0;
			}
		}
	}


	/*************************************/
	/*** calculate the sensor rotation ***/
	if (conv_options->relative_axis) {
		/* This seems to work well for the head sensor.   */
		/* TODO: However, it's not the greatest interface */
		/* for controlling wand rotation -- wand rotation */
		/* should probably be relative to the head_sensor */
		/* as with the wand translation.                  */

		/* do rotations around the sensor's axes                  */
		/*   NOTE: in this case, order of rotation doesn't matter */
		switch (conv_options->azimuth_axis) {
		case VR_X:
			vrMatrixPostRotateId(mat, VR_X, valuators_rot[VR_AZIM]*r_scale);
			vrMatrixPostRotateId(mat, VR_Z, valuators_rot[VR_ELEV]*r_scale);
			vrMatrixPostRotateId(mat, VR_Y, valuators_rot[VR_ROLL]*r_scale);
			break;
		case VR_Y:
			vrMatrixPostRotateId(mat, VR_Y, valuators_rot[VR_AZIM]*r_scale);
			vrMatrixPostRotateId(mat, VR_X, valuators_rot[VR_ELEV]*r_scale);
			vrMatrixPostRotateId(mat, VR_Z, valuators_rot[VR_ROLL]*r_scale);
			break;
		case VR_Z:
			vrMatrixPostRotateId(mat, VR_Z, valuators_rot[VR_AZIM]*r_scale);
			vrMatrixPostRotateId(mat, VR_Y, valuators_rot[VR_ELEV]*r_scale);
			vrMatrixPostRotateId(mat, VR_X, valuators_rot[VR_ROLL]*r_scale);
			break;
		}
	} else {
		vrVector	vector;
/* for calculating the Euler angles of a position tracker input */
vrPoint		location_input;		/* The location/translation portion */
vrQuat		quat_input;		/* The quaternion portion */
vrEuler		euler_input;		/* The Euler+translation representation */

		/* do rotations around the world's axes */
		/* NOTE: the vrMatrixGetResetTranslationAd() and vrMatrixSetTranslationOnlyAd()   */
		/*   functions are effectively used to translate the matrix to the origin, do the */
		/*   rotation, and then translate back again.  This way saves a few operations.   */
		/* The other option is to use the vrMatrixPreRotateAboutOriginId() call.          */
		vrMatrixGetResetTranslationAd(mat, vector.v);

		switch (conv_options->azimuth_axis) {
		case VR_X:
			vrMatrixPreRotateId(mat, VR_X, valuators_rot[VR_AZIM]*r_scale);
			vrMatrixPreRotateId(mat, VR_Z, valuators_rot[VR_ELEV]*r_scale);
			vrMatrixPreRotateId(mat, VR_Y, valuators_rot[VR_ROLL]*r_scale);
vrEulerSetFromMatrix(&euler_input, mat); /* I think after we confirm the mathematics are working, delete me */
			break;
		case VR_Y:
			vrMatrixPreRotateId(mat, VR_Y, valuators_rot[VR_AZIM]*r_scale);
			vrMatrixPreRotateId(mat, VR_X, valuators_rot[VR_ELEV]*r_scale);
			vrMatrixPreRotateId(mat, VR_Z, valuators_rot[VR_ROLL]*r_scale);
/* Store the given <input> in FreeVR types (inc. Eulers) */
vrEulerSetFromMatrix(&euler_input, mat); /* I think after we confirm the mathematics are working, delete me */
#if 1
vrPointGetTransFromMatrix(&location_input, mat);
vrQuatSetFromMatrix(&quat_input, mat);
vrEulerSetFromQuatPoint(&euler_input, &quat_input, &location_input);
#endif
			break;
		case VR_Z:
vrEulerSetFromMatrix(&euler_input, mat); /* I think after we confirm the mathematics are working, delete me */
			vrMatrixPreRotateId(mat, VR_Z, valuators_rot[VR_AZIM]*r_scale);
			vrMatrixPreRotateId(mat, VR_Y, valuators_rot[VR_ELEV]*r_scale);
			vrMatrixPreRotateId(mat, VR_X, valuators_rot[VR_ROLL]*r_scale);
			break;
		}
		vrMatrixSetTranslationOnlyAd(mat, vector.v);
	}

	vrAssign6sensorValue(sensor6, mat, oob /* , TODO: time */);
}


/***************************************************************************/
/* vrGet6sensor(): this function simply returns a pointer to the vr6sensor */
/*   input value structure for the given integer 6sensor reference number. */
/*   It is used by the following functions that access vr6sensor values,   */
/*   and also by two functions in vr_visren.c.                             */
vr6sensor *vrGet6sensor(int input_num)
{
	return ((vr6sensor *)vrInputGetFromTypeIndex(vrContext, VRINPUT_6SENSOR, input_num));
}


/*************************************************************************/
/* vrAssign6sensorOobValue(): assign an Out-of-bounds (oob) value to the */
/*   given vr6sensor input value object.                                 */
/* NOTE: Values of oob less than zero are treated as a nop (ie. do not   */
/*   change) flag.                                                       */
void vrAssign6sensorOobValue(vr6sensor *sensor6, int oob)
{
	if (sensor6->object_type != VROBJECT_INPUTDATA) {
		vrDbgPrintf("Attempt to assign a position sensor value to a non-input object type %d at %#p\n", sensor6->object_type, sensor6);
		return;
	}

	if (sensor6->input_type != VRINPUT_6SENSOR) {
		vrDbgPrintf("Attempt to assign a position sensor value to an incorrect input type %d at %#p\n",
			sensor6->input_type, sensor6);
		return;
	}

	if (oob >= 0) {
		vrLockWriteSet(sensor6->lock);
		sensor6->oob = oob;
		vrLockWriteRelease(sensor6->lock);
	}

	return;
}


/***********************************************************************/
/* vrGet6sensorOobValue(): returns the value of a 6sensor's "oob" (out */
/*   of bounds) flag.  Typically this will be a boolean value of 0 or  */
/*   1, but also -1 is returned if input_num is not a valid 6-sensor.  */
int vrGet6sensorOobValue(int input_num)
{
	vr6sensor	*input;
	int		oob;

	input = vrGet6sensor(input_num);
	if (input == NULL) {
		vrErrPrintf("vrGet6sensorOobValue(): Hmmm, encountered a bad Position Sensor (%d) -- this shouldn't happen.\n", input_num);
		return -1;
	}

	vrLockReadSet(input->lock);
	oob = input->oob;
	vrLockReadRelease(input->lock);

	return oob;
}


/***************************************************************************/
/* vrAssign6sensorActiveValue(): assigns an "active" state value to the    */
/*   given vr6sensor input value object.  The active state is generally    */
/*   used to indicate which, of multiple, simulated 6-sensors is currently */
/*   being manipulated so it can be visually marked as the active 6-sensor.*/
/* NOTE: values of active less than zero are treated as a nop (ie. do not  */
/*   change) flag.                                                         */
void vrAssign6sensorActiveValue(vr6sensor *sensor6, int active)
{
	if (sensor6->object_type != VROBJECT_INPUTDATA) {
		vrDbgPrintf("Attempt to assign a position sensor value to a non-input object type %d at %#p\n", sensor6->object_type, sensor6);
		return;
	}

	if (sensor6->input_type != VRINPUT_6SENSOR) {
		vrDbgPrintf("Attempt to assign a position sensor value to an incorrect input type %d at %#p\n",
			sensor6->input_type, sensor6);
		return;
	}

	if (active >= 0) {
		vrLockWriteSet(sensor6->lock);
		sensor6->active = active;
		vrLockWriteRelease(sensor6->lock);
	}

	return;
}


/**************************************************************************/
/* vrGet6sensorActiveValue(): returns the current "active" state value of */
/*   the given vr6sensor input value object.  The value -1 is returned if */
/*   input_num is not a valid 6-sensor.                                   */
int vrGet6sensorActiveValue(int input_num)
{
	vr6sensor	*input;
	int		active;

	input = vrGet6sensor(input_num);
	if (input == NULL) {
		vrErrPrintf("vrGet6sensorActiveValue(): Hmmm, encountered a bad Position Sensor (%d) -- this shouldn't happen.\n", input_num);
		return -1;
	}

	vrLockReadSet(input->lock);
	active = input->active;
	vrLockReadRelease(input->lock);

	return active;
}


/*************************************************************************/
/* vrAssign6sensorErrorValue(): assigns an "error" state value to the    */
/*   given vr6sensor input value object.  The error state determined by  */
/*   the position tracking input device, and in practice is infrequently */
/*   used.  Indeed as of 09/28/21, the only device driver that reports   */
/*   position tracking errors is the Fastrak driver (also used by the    */
/*   Innersense acoustic tracking system).                               */
/* NOTE: values of error less than zero are treated as a nop (ie. do not */
/*   change) flag.                                                       */
void vrAssign6sensorErrorValue(vr6sensor *sensor6, int error)
{
	if (sensor6->object_type != VROBJECT_INPUTDATA) {
		vrDbgPrintf("Attempt to assign a position sensor value to a non-input object type %d at %#p\n", sensor6->object_type, sensor6);
		return;
	}

	if (sensor6->input_type != VRINPUT_6SENSOR) {
		vrDbgPrintf("Attempt to assign a position sensor value to an incorrect input type %d at %#p\n",
			sensor6->input_type, sensor6);
		return;
	}

	if (error >= 0) {
		vrLockWriteSet(sensor6->lock);
		sensor6->error = error;
		vrLockWriteRelease(sensor6->lock);
	}

	return;
}


/************************************************************************/
/* vrGet6sensorErrorValue(): returns the current "error" state value of */
/*   the given vr6sensor input value object.  The value -1 is returned  */
/*   if input_num is not a valid 6-sensor.                              */
int vrGet6sensorErrorValue(int input_num)
{
	vr6sensor	*input;
	int		error;

	input = vrGet6sensor(input_num);
	if (input == NULL) {
		vrErrPrintf("vrGet6sensorErrorValue(): Hmmm, encountered a bad Position Sensor (%d) -- this shouldn't happen.\n", input_num);
		return -1;
	}

	vrLockReadSet(input->lock);
	error = input->error;
	vrLockReadRelease(input->lock);

	return error;
}


/***********************************************************************/
/* vrGet6sensorDummyValue(): returns the "dummy" state value of the    */
/*   given vr6sensor input value object.  This state indicates whether */
/*   the given 6-sensor was created as a dummy to accommodate a failed */
/*   request for a non-existent 6-sensor.  (This action exists to      */
/*   alleviate the condition when the renderer requests a 6-sensor to  */
/*   produce head-perspective rendering, and using a "dummy" input, we */
/*   avoid a continuous stream of error messages.  The value -1 is     */
/*   returned if input_num is not a valid 6-sensor.                    */
/* NOTE: there is no accompanying vrAssign6sensorDummyValue() as this  */
/*   is an immutable attribute.                                        */
int vrGet6sensorDummyValue(int input_num)
{
	vr6sensor	*input;
	int		dummy;

	input = vrGet6sensor(input_num);
	if (input == NULL) {
		vrErrPrintf("vrGet6sensorDummyValue(): Hmmm, encountered a bad Position Sensor (%d) -- this shouldn't happen.\n", input_num);
		return -1;
	}

	vrLockReadSet(input->lock);
	dummy = input->dummy;
	vrLockReadRelease(input->lock);

	return dummy;
}


/**************************************************************************/
/* vrMatrixGet6sensorValues(): obtains the current values of the given    */
/*   vr6sensor ("tracker") in the running system.  Unlike the other input */
/*   types, FreeVR does not store a "last_value" for vr6sensor inputs.    */
/*   And thus there is no function to obtain the delta between requests.  */
/*   Locks are used to protect the validity of the data.  If the input    */
/*   doesn't exist a dummy vr6sensor is created (with the identity matrix */
/*   value) and returned.                                                 */
vrMatrix *vrMatrixGet6sensorValues(vrMatrix *mat, int input_num)
{
	vr6sensor	*input;
	vrGenericInput	*gen_input;

	input = vrGet6sensor(input_num);

	if (input == NULL) {
		/* Create a new dummy input to handle the request */
		vrDbgPrintfN(AALWAYS_DBGLVL, "FreeVR (%.4f): " RED_TEXT "vrMatrixGet6sensorValues(): No Position Sensor (%d) available (Creating a dummy)\n" NORM_TEXT,
			vrCurrentSimTime(), input_num);
		gen_input = vrInputAddDummyToInputMap(vrContext->input, VRINPUT_6SENSOR, input_num);
		strncpy(gen_input->my_object->desc_ui, "dummy", INPUT_UIDESC_LEN);

		vrMatrixSetIdentity(mat);
		return mat;
	}


	vrLockReadSet(input->lock);
	vrMatrixCopy(mat, input->position);
	vrLockReadToWrite(input->lock);
	vrMatrixCopy(input->last_position, input->position);
	vrLockWriteRelease(input->lock);

	return mat;
}


/****************************************************************************/
/* vrMatrixGet6sensorValuesNoLastUpdate(): This version of getting the      */
/*   vr6sensor values is intended for Rendering processes, so it doesn't    */
/*   update "last_value" and therefore won't disrupt the Delta calculations.*/
/* NOTE: presently, there is no ValuesDelta() function for vr6sensor inputs,*/
/*   so calling this is basically unnecessary, but it's good to be safe.    */
vrMatrix *vrMatrixGet6sensorValuesNoLastUpdate(vrMatrix *mat, int input_num)
{
	vr6sensor	*input;

	input = vrGet6sensor(input_num);
	if (input == NULL) {
		vrErrPrintf("vrMatrixGet6sensorValues(): Hmmm, encountered a bad Position Sensor (%d) -- this shouldn't happen.\n", input_num);
		vrMatrixSetIdentity(mat);
		return mat;
	}

	vrLockReadSet(input->lock);
	vrMatrixCopy(mat, input->position);
	vrLockReadRelease(input->lock);

	return mat;
}


/*********************************************************************/
/* vrMatrixGet6sensorValuesDirect(): this is the "direct" version of */
/*   vrMatrixGet6sensorValues(), which means instead of passing a    */
/*   FreeVR index to a 6-sensor, a pointer to the actual vr6sensor   */
/*   input value object is given.                                    */
/* NOTE: also, this does not apply the R2E or T2RW transformations.  */
vrMatrix *vrMatrixGet6sensorValuesDirect(vrMatrix *mat, vr6sensor *input)
{

	if (input == NULL) {
		if (vrDbgDo(CONFIG_WARN_DBGLVL)) {
			vrErrPrintf("vrMatrixGet6sensorValuesDirect(): Invalid NULL Position Sensor.\n");
		}
		vrMatrixSetIdentity(mat);
		return mat;
	}

	vrLockReadSet(input->lock);
	vrMatrixCopy(mat, input->position);
	vrLockReadToWrite(input->lock);
	vrMatrixCopy(input->last_position, input->position);
	vrLockWriteRelease(input->lock);

	return mat;
}


/**********************************************************************/
/* vrMatrixGet6sensorValuesDirectNoLastUpdate(): this is the "direct" */
/*   version of vrMatrixGet6sensorValuesNoLastUpdate(), which means   */
/*   instead of passing a FreeVR index to a 6-sensor, a pointer to    */
/*   the actual vr6sensor input value object is given.                */
vrMatrix *vrMatrixGet6sensorValuesDirectNoLastUpdate(vrMatrix *mat, vr6sensor *input)
{

	if (input == NULL) {
		vrErrPrintf("vrMatrixGet6sensorValuesDirectNoLastUpdate(): Invalid NULL Position Sensor.\n");
vrErrPrintf("aborting until a better solution is determined\n"); abort();
		vrMatrixSetIdentity(mat);
		return mat;
	}

	vrLockReadSet(input->lock);
	vrMatrixCopy(mat, input->position);
	vrLockReadRelease(input->lock);

	return mat;
}


/**********************************************************************/
/* vrMatrixGet6sensorRawValuesDirect(): this function for obtaining   */
/*   the current, and unfiltered values of a given 6-sensor.  What    */
/*   "Raw" means is that those values haven't been altered by         */
/*   modifiers such as t2rw_xform.                                    */
/* This is mostly used internally in vr_input.c to help make 6-sensor */
/*   value assignments.                                               */
vrMatrix *vrMatrixGet6sensorRawValuesDirect(vrMatrix *mat, vr6sensor *input)
{

	if (input == NULL) {
		vrErrPrintf("vrMatrixGet6sensorRawValuesDirect(): Invalid NULL Position Sensor.\n");
		vrMatrixSetIdentity(mat);
		return mat;
	}

	vrLockReadSet(input->lock);
	/* TODO: this will be different for generic non-6DOF sensors */
	vrMatrixCopy(mat, input->raw_data);
	/* TODO: copy raw_data to last_value? */
	vrLockReadRelease(input->lock);

	return mat;
}


/*********************************************************************/
/* vrVectorGetRWFrom6sensorDir(): Get the requested vector emanating */
/*   from the given sensor as it is located in the real world.       */
/*   For example, to get the head's forward vector:                  */
/*	vrVectorGet6sensorRW(head_forward_vector, <head sensor>, VRDIR_FORE); */
/* Thus a 2-DOF vector (normalized-triple) is reported, essentially  */
/*   equivalent to the azimuth and elevation of the sensor.          */
/* As the RW in the function name implies, the returned value is     */
/*   relative to the real-world coordinate system of the VR system.  */
/* NOTE: used to be: vrVector *vrGetSensor6VectorRW(vrDirection direction, int sensor_num, vrVector *vector) */
vrVector *vrVectorGetRWFrom6sensorDir(vrVector *vector, int sensor_num, vrDirection direction)
{
static	vrMatrix	mat;				/* used as storage space */
static	vrVector	fore_vec = { {  0.0,  0.0, -1.0 } };
static	vrVector	back_vec = { {  0.0,  0.0,  1.0 } };
static	vrVector	up_vec =   { {  0.0,  1.0,  0.0 } };
static	vrVector	down_vec = { {  0.0, -1.0,  0.0 } };
static	vrVector	left_vec = { { -1.0,  0.0,  0.0 } };
static	vrVector	right_vec= { {  1.0,  0.0,  0.0 } };
	vrVector	*unit_vec;

	switch (direction) {
	default:
		vrErrPrintf("vrVectorGetRWFrom6sensorDir(): " RED_TEXT "Unknown vector direction, using FORE\n" NORM_TEXT);
	case VRDIR_FORE:
		unit_vec = &fore_vec;
		break;
	case VRDIR_BACK:
		unit_vec = &back_vec;
		break;
	case VRDIR_UP:
		unit_vec = &up_vec;
		break;
	case VRDIR_DOWN:
		unit_vec = &down_vec;
		break;
	case VRDIR_LEFT:
		unit_vec = &left_vec;
		break;
	case VRDIR_RIGHT:
		unit_vec = &right_vec;
		break;
	}

	vrVectorTransformByMatrix(vector, unit_vec, vrMatrixGet6sensorValues(&mat, sensor_num));

	return vector;
}


/**************************************************************************/
/* vrVectorfGetRWFrom6sensorDir(): this operates the same as the function */
/*   vrVectorGetRWFrom6sensorDir(), above, but takes a vector of floats   */
/*   (vrVectorf) rather than the default vector of doubles (vrVector) as  */
/*   the first argument, and likewise provided as the return value.       */
/* NOTE: to reiterate, this is the "float" vector version.                */
vrVectorf *vrVectorfGetRWFrom6sensorDir(vrVectorf *vector, int sensor_num, vrDirection direction)
{
	vrVector	double_vector;

	vrVectorGetRWFrom6sensorDir(&double_vector, sensor_num, direction);
	vector->v[VR_X] = double_vector.v[VR_X];
	vector->v[VR_Y] = double_vector.v[VR_Y];
	vector->v[VR_Z] = double_vector.v[VR_Z];

	return vector;
}


/*************************************************************************/
/* vrVectorGetVWFromUser6sensorDir(): Get the requested vector emanating */
/*   from the given sensor relative to the virtual world location of the */
/*   given user.  (Multiple users may have travelled to different        */
/*   coordinates within the virtual world.)                              */
/* Thus a 2-DOF vector (normalized-triple) is reported, essentially      */
/*   equivalent to the azimuth and elevation of the sensor.              */
/* As the VW in the function name implies, the returned value is         */
/*   relative to the virtual-world coordinate system of the VR system.   */
/* NOTE: used to be: vrVector *vrGetSensor6VectorVW(int user_num, vrDirection direction, int sensor_num, vrVector *vector) */
vrVector *vrVectorGetVWFromUser6sensorDir(vrVector *vector, int user_num, int sensor_num, vrDirection direction)
{
	vrVectorGetRWFrom6sensorDir(vector, sensor_num, direction);
	vrVectorUserVWFromRW(user_num, vector, vector);

	return vector;
}


/******************************************************************************/
/* vrVectorfGetVWFromUser6sensorDir(): as with vrVectorfGetRWFrom6sensorDir(),*/
/*   this is the "float" (vs. "double" version of the get-VW-From-User        */
/*   operation.                                                               */
vrVectorf *vrVectorfGetVWFromUser6sensorDir(vrVectorf *vector, int user_num, int sensor_num, vrDirection direction)
{
	vrVector	double_vector;

	vrVectorGetVWFromUser6sensorDir(&double_vector, user_num, sensor_num, direction);
	vector->v[VR_X] = double_vector.v[VR_X];
	vector->v[VR_Y] = double_vector.v[VR_Y];
	vector->v[VR_Z] = double_vector.v[VR_Z];

	return vector;
}


/**************************************************************************/
/* vrPointGetRWFrom6sensor(): this function is used to get the real-world */
/*   location of a given 6sensor. That is, instead of the full 6-dof      */
/*   position of the sensor, only the 3-dof location is returned in the   */
/*   form of a vrPoint type.                                              */
/* NOTE: used to be: vrPoint *vrGetSensor6LocationRW(int sensor_num, vrPoint *real_point) */
vrPoint *vrPointGetRWFrom6sensor(vrPoint *real_point, int sensor_num)
{
	vrMatrix	mat_store;		/* temporary matrix storage */
	vrMatrix	*mat = &mat_store;	/* pointer to temp matrix storage */

	return (vrPoint *)vrVectorGetTransFromMatrix((vrVector *)real_point, vrMatrixGet6sensorValues(mat, sensor_num));
}


/*****************************************************************/
/* vrPointfGetRWFrom6sensor(): the "float" version of the above  */
/*   vrPointGetRWFrom6sensor() function which handles "doubles". */
vrPointf *vrPointfGetRWFrom6sensor(vrPointf *real_point, int sensor_num)
{
	vrPoint	double_point;

	vrPointGetRWFrom6sensor(&double_point, sensor_num);
	real_point->v[VR_X] = double_point.v[VR_X];
	real_point->v[VR_Y] = double_point.v[VR_Y];
	real_point->v[VR_Z] = double_point.v[VR_Z];

	return real_point;
}


/************************************************************************/
/* vrPointGetVWFromUser6sensor(): this function is used to get the      */
/*   virtual-world location of a given 6sensor relative to a particular */
/*   user's coordinate system.  As with vrPointGetRWFrom6sensor(), it   */
/*   returns only the 3-dof location in the form of a vrPoint type.     */
/* NOTE: used to be: vrPoint *vrGetSensor6LocationVW(int usernum, int sensor_num, vrPoint *virtual_point) */
vrPoint *vrPointGetVWFromUser6sensor(vrPoint *virtual_point, int usernum, int sensor_num)
{
	vrMatrix	mat_store;		/* temporary matrix storage */
	vrMatrix	*mat = &mat_store;	/* pointer to temp matrix storage */
	vrPoint		point_data;
	vrPoint		*real_point = &point_data;

	vrVectorGetTransFromMatrix((vrVector *)real_point, vrMatrixGet6sensorValues(mat, sensor_num));

	return vrPointGetVWUserFromRWPoint(virtual_point, usernum, real_point);
}


/*********************************************************************/
/* vrPointfGetVWFromUser6sensor(): the "float" version of the above  */
/*   vrPointGetVWFromUser6sensor() function which handles "doubles". */
vrPointf *vrPointfGetVWFromUser6sensor(vrPointf *virtual_point, int usernum, int sensor_num)
{
	vrPoint	double_point;

	vrPointGetVWFromUser6sensor(&double_point, usernum, sensor_num);
	virtual_point->v[VR_X] = double_point.v[VR_X];
	virtual_point->v[VR_Y] = double_point.v[VR_Y];
	virtual_point->v[VR_Z] = double_point.v[VR_Z];

	return virtual_point;
}


/****************************************************************************/
/* vrEulerGetRWFrom6sensor(): this function returns the full 6-dof position */
/*   of the given sensor, but instead of returning a 4x4 matrix, it returns */
/*   three translational values (the location), and three rotational values */
/*   (the orientation) in Euler angles.  All in real-world coordinates.     */
vrEuler *vrEulerGetRWFrom6sensor(vrEuler *euler, int sensor_num)
{
	vrMatrix	mat_store;		/* temporary matrix storage */
	vrMatrix	*mat = &mat_store;	/* pointer to temp matrix storage */

	return vrEulerSetFromMatrix(euler, vrMatrixGet6sensorValues(mat, sensor_num));
}


/*************************************************************************/
/* vrEulerGetVWFromUser6sensor(): this function returns the full 6-dof   */
/*   position of the given sensor (relative to the given user's traveled */
/*   coordinates within the virtual world), but instead of returning a   */
/*   4x4 matrix, it returns three translational values (the location),   */
/*   and three rotational values (the orientation) in Euler angles.      */
vrEuler *vrEulerGetVWFromUser6sensor(vrEuler *euler, int usernum, int sensor_num)
{
	vrMatrix	mat_store;		/* temporary matrix storage */
	vrMatrix	*mat = &mat_store;	/* pointer to temp matrix storage */

	return vrEulerSetFromMatrix(euler, vrMatrixPreMult(vrMatrixGet6sensorValues(mat, sensor_num), vrContext->input->users[usernum]->vw2rw_xform));
}



	/**************************************************************/
	/**************************************************************/
	/************ Sensor N input type access functions ************/


/*****************************************************************************/
/* vrInputSetNsensorDescription(): sets the human-readable description text  */
/*   for the given input.  This information is purely for the end-user to be */
/*   provided "help" information that conveys what the given "button" does.  */
/* NOTE: returns 1 when the description is successfully set.  It returns 0   */
/*   when the input_num "button" doesn't exist in current configuration.     */
int vrInputSetNsensorDescription(int input_num, const char *input_desc)
{
	vrGenericInput	*input;

	input = vrInputGetFromTypeIndex(vrContext, VRINPUT_NSENSOR, input_num);
	if (input == NULL) {
		/* Create a new dummy input to handle the request */

		vrDbgPrintfN(AALWAYS_DBGLVL, "FreeVR (%.4f): " RED_TEXT "No N-Sensor (%d) available to '%s' (Creating a dummy)\n" NORM_TEXT,
			vrCurrentSimTime(), input_num, input_desc);
		input = vrInputAddDummyToInputMap(vrContext->input, VRINPUT_NSENSOR, input_num);

		strncpy(input->my_object->desc_ui, input_desc, INPUT_UIDESC_LEN);

		return 0;
	}

	strncpy(input->my_object->desc_ui, input_desc, INPUT_UIDESC_LEN);

	return 1;
}


/*********************************/
/* TODO: implement this function */
/*************************************************************************/
/* vrAssignNsensorArray(): assigns an array of float values to the given */
/*   Nsensor input value object.  This function is called by the input   */
/*   device implementation code whenever it receives (or simulates) an   */
/*   input from a "glove" type input (for example).  Locks are used to   */
/*   protect the validity of the data.                                   */
/* NOTE: Not yet implemented (NYI).                                      */
void vrAssignNsensorArray(vrNsensor *sensorN, float *new_data /* , vrTime time */)
{
	if (sensorN->object_type != VROBJECT_INPUTDATA) {
		vrDbgPrintf("Attempt to assign a N-sensor value to a non-input object type %d at %#p\n", sensorN->object_type, sensorN);
		return;
	}

	if (sensorN->input_type != VRINPUT_NSENSOR) {
		vrDbgPrintf("Attempt to assign an N-sensor value to an incorrect input type %d at %#p\n", sensorN->input_type, sensorN);
		return;
	}

	if (new_data == NULL) {
		vrErr("vrAssignNsensorArray: Attempted to assign a Null pointer.\n");
		vrPrintf("Pausing: Feel free to use 'gdb -p %d' now to find out where\n", getpid());
		pause();
	}

#if 0 /* checksums not yet implemented */
	/* TODO: checksum check here -- I fixed the bug I was looking for, */
	/*   so I didn't implement this, but we may want to in the future. */
#endif

	if (1) {
		/* TODO: If there is a quick memory-checksum function, we could */
		/*   add a checksum field for quickly determining if data has   */
		/*   changed.  If the checksums match, then a slightly slower,  */
		/*   brute force check should be done.                          */

		/* assume (for now) value has changed */
		vrLockWriteSet(sensorN->lock);
		sensorN->timestamp = 0.0;		/* TODO: assign timestamp */
		memcpy(sensorN->values, new_data, sensorN->dof * sizeof(sensorN->values[0]));

#if 0
		sensorN.calibrated = 0;
		/* now calibrate the data */
		/* TODO: the calibration stuff */

		/* we may want to use callbacks for the calibration function */
#endif

#if 0
		sensorN.frame_of_reference = 0;  /* ?? TODO: set this to world space */
#endif

		vrLockWriteRelease(sensorN->lock);

		if (sensorN->queue_me) {
			/* TODO: put me in the queue */
		}
	}
}


/***************************************************************************/
/* vrGetNsensorValue(): obtains the current values of the given vrNsensor  */
/*   (e.g. "glove") in the running system.  As with the vr6sensor (and     */
/*   unlike all the other input types, FreeVR does not store a "last_value"*/
/*   for vrNsensor inputs.   And thus there is no function to obtain the   */
/*   delta between requests.                                               */
/* Locks are used to protect the validity of the data.  Unlike vr6sensor   */
/*   inputs, if the input doesn't exist only an error message is output,   */
/*   and NO dummy vrNsensor is created.                                    */
float *vrGetNsensorArray(int input_num, float *array)
{
	vrNsensor	*input;
	int		array_size;

	input = (vrNsensor *)vrInputGetFromTypeIndex(vrContext, VRINPUT_NSENSOR, input_num);
	if (input == NULL) {
		/* TODO: this is where we can create new dummy inputs to handle the request */
		vrErrPrintf("vrGetNsensorArray(): Hmmm, encountered a bad N-Sensor -- this shouldn't happen.\n");
		return array;
	}

	vrLockReadSet(input->lock);
	array_size = input->dof * sizeof(array[0]);
	memcpy(array, input->values, array_size);
	vrLockReadToWrite(input->lock);
	memcpy(input->last_values, input->values, array_size);
	vrLockWriteRelease(input->lock);

	return array;
}


/********************************************************************/
/* vrGetNsensorArrayDirect(): this is the "direct" version of the   */
/*   vrGetNsensorArray() function, which means instead of passing a */
/*   FreeVR index to a N-sensor, a pointer to the actual vrNsensor  */
/*   input value object is given.                                   */
float *vrGetNsensorArrayDirect(vrNsensor *input, float *array)
{
	int		array_size;

	if (input == NULL) {
		vrErrPrintf("vrGetNsensorValueDirect(): Invalid NULL N Sensor.\n");
		return 0;
	}

	array_size = input->dof * sizeof(array[0]);
	vrLockReadSet(input->lock);
	memcpy(array, input->values, array_size);
	vrLockReadToWrite(input->lock);
	memcpy(input->last_values, input->values, array_size);
	vrLockWriteRelease(input->lock);

	return array;
}


/***************************************************************************/
/* vrGetNsensorValue(): returns one element of the current values of the   */
/*   given vrNsensor (e.g. thumb-joint of a "gloved hand") in the running  */
/*   system.  As with the vr6sensor (and unlike all the other input types, */
/*   FreeVR does not store a "last_value" for vrNsensor inputs.  And thus  */
/*   there is no function to obtain the delta between requests.            */
/* Locks are used to protect the validity of the data.  Unlike vr6sensor   */
/*   inputs, if the input doesn't exist only an error message is output,   */
/*   and NO dummy vrNsensor is created.                                    */
float vrGetNsensorValue(int input_num, int datum_num)
{
	vrNsensor	*input;
	float		value;

	if (datum_num >= MAX_NSENSOR_VALUES) {
		vrErrPrintf("vrGetNsensorValue(): Invalid N Sensor datum requested (%d) -- valid range for an N-Sensor datum is [0..%d].\n", datum_num, MAX_NSENSOR_VALUES-1);
		return 0;
	}

	input = (vrNsensor *)vrInputGetFromTypeIndex(vrContext, VRINPUT_NSENSOR, input_num);
	if (input == NULL) {
		/* TODO: this is where we can create new dummy inputs to handle the request */
		vrErrPrintf("vrGetNsensorValue(): Hmmm, encountered a bad N-Sensor -- this shouldn't happen.\n");
		return 0;
	}


	vrLockReadSet(input->lock);
	value = input->values[datum_num];
#if 0
	/* TODO: decide if we want to do this for the single datum */
	vrLockReadToWrite(input->lock);
	input->last_values[datam_num] = value;
	vrLockWriteRelease(input->lock);
#else
	vrLockReadRelease(input->lock);
#endif

	return value;
}


/****************************************************************************/
/* vrGetNsensorValueNoLastUpdate(): This version of getting the vrNsensor   */
/*   values is intended for Rendering processes, so it doesn't update       */
/*   "last_value" and therefore won't disrupt the Delta calculations.       */
/* NOTE: presently, there is no ValuesDelta() function for vrNsensor inputs,*/
/*   so calling this is basically unnecessary, but it's good to be safe.    */
float vrGetNsensorValueNoLastUpdate(int input_num, int datum_num)
{
	vrNsensor	*input;
	float		value;

	if (datum_num >= MAX_NSENSOR_VALUES) {
		vrErrPrintf("vrGetNsensorValue(): Invalid N Sensor datum requested (%d) -- valid range for an N-Sensor datum is [0..%d].\n", datum_num, MAX_NSENSOR_VALUES-1);
		return 0;
	}

	input = (vrNsensor *)vrInputGetFromTypeIndex(vrContext, VRINPUT_NSENSOR, input_num);
	if (input == NULL) {
		/* TODO: this is where we can create new dummy inputs to handle the request */
		vrErrPrintf("vrGetNsensorValue(): Hmmm, encountered a bad N-Sensor -- this shouldn't happen.\n");
		return 0;
	}


	vrLockReadSet(input->lock);
	value = input->values[datum_num];
	vrLockReadRelease(input->lock);

	return value;
}


/********************************************************************/
/* vrGetNsensorValueDirect(): this is the "direct" version of the   */
/*   vrGetNsensorValue() function, which means instead of passing a */
/*   FreeVR index to a N-sensor, a pointer to the actual vrNsensor  */
/*   input value object is given.                                   */
float vrGetNsensorValueDirect(vrNsensor *input, int datum_num)
{
	float		value;

	if (input == NULL) {
		vrErrPrintf("vrGetNsensorValueDirect(): Invalid NULL N Sensor.\n");
		return 0;
	}

	if (datum_num >= MAX_NSENSOR_VALUES) {
		vrErrPrintf("vrGetNsensorValueDirect(): Invalid N Sensor datum requested (%d) -- valid range for N-Sensor x0%p is [0..%d].\n", datum_num, input, MAX_NSENSOR_VALUES-1);
		return 0;
	}

	vrLockReadSet(input->lock);
	value = input->values[datum_num];
#if 0
	/* TODO: decide if we want to do this for the single datum */
	vrLockReadToWrite(input->lock);
	input->last_values[datam_num] = value;
	vrLockWriteRelease(input->lock);
#else
	vrLockReadRelease(input->lock);
#endif

	return value;
}


/*****************************************************************************/
/* Neither vrGetNsensorDelta() nor vrGetNsensorDeltaDirect() are implemented */
/*   in this version of FreeVR.  As it stands, N-sensor inputs have not been */
/*   fully implemented, primarily due to lack of physical inputs that would  */
/*   benefit from this input type.                                           */
float vrGetNsensorDelta(int input_num, int datum_num) { return -10.0; }
float vrGetNsensorDeltaDirect(vrNsensor *input, int datum_num) { return -10.0; }


	/**************************************************************/
	/**************************************************************/

#ifdef GFX_PERFORMER /* { */
/***********************************************************************************/
/* Utility functions that translate FreeVR coordinates in to Performer Z-up coords */
/***********************************************************************************/
/* TODO: in the future we probably will just have a conversion matrix stored */
/*   with the context, and we'll use that matrix in the original functions   */
/*   to convert to whatever coordinate system the graphics system uses.      */


/***************************************************/
/* Performer version of vrPointGetRWFrom6sensor(). */
vrPoint *vrPfPointGetRWFrom6sensor(vrPoint *virtual_point, int sensor_num)
{
static	vrMatrix	mat_pos90x = { 1.0,0.0,0.0,0.0,  0.0,0.0,1.0,0.0,  0.0,-1.0,0.0,0.0,  0.0,0.0,0.0,1.0 };
#if 0
static	vrMatrix	mat_neg90x = { 1.0,0.0,0.0,0.0,  0.0,0.0,-1.0,0.0,  0.0,1.0,0.0,0.0,  0.0,0.0,0.0,1.0 };
#endif

	vrPointGetRWFrom6sensor(virtual_point, sensor_num);
	vrPointTransformByMatrix(virtual_point, virtual_point, &mat_pos90x);

	return virtual_point;
}


/*******************************************************/
/* Performer version of vrPointGetVWFromUser6sensor(). */
vrPoint *vrPfPointGetVWFromUser6sensor(vrPoint *virtual_point, int usernum, int sensor_num)
{
#if 0 /* for some reason, I can't declare these as static, even though it works in PF examples */
vrMatrix	*mat_neg90x = vrMatrixSetRotationIdNew(vrMatrixCreate(), VR_X, -90.0);	/* convert from Pf to OpenGL space */
vrMatrix	*mat_pos90x = vrMatrixSetRotationIdNew(vrMatrixCreate(), VR_X,  90.0);	/* convert from OpenGL to Pf space  (ogl2pf) */

vrPrintMatrix("neg", mat_neg90x);
vrPrintMatrix("pos", mat_pos90x);
#else
static	vrMatrix	mat_pos90x = { 1.0,0.0,0.0,0.0,  0.0,0.0,1.0,0.0,  0.0,-1.0,0.0,0.0,  0.0,0.0,0.0,1.0 };
#  if 0
	vrMatrix	*yo = vrMatrixSetRotationIdNew(vrMatrixCreate(), VR_X,  90.0);	/* convert from OpenGL to Pf space  (ogl2pf) */
	vrPrintMatrix("pos", &mat_pos90x);
	vrPrintMatrix("yo ", yo);
#  endif
#endif

	vrPointGetVWFromUser6sensor(virtual_point, usernum, sensor_num);
	vrPointTransformByMatrix(virtual_point, virtual_point, &mat_pos90x);

	return virtual_point;
}


/*******************************************************/
/* Performer version of vrVectorGetRWFrom6sensorDir(). */
vrVector *vrPfVectorGetRWFrom6sensorDir(vrVector *vector, int sensor_num, vrDirection direction)
{
static	vrMatrix	mat_pos90x = { 1.0,0.0,0.0,0.0,  0.0,0.0,1.0,0.0,  0.0,-1.0,0.0,0.0,  0.0,0.0,0.0,1.0 };

	vrVectorGetRWFrom6sensorDir(vector, sensor_num, direction);
	vrVectorTransformByMatrix(vector, vector, &mat_pos90x);

	return vector;
}


/***********************************************************/
/* Performer version of vrVectorGetVWFromUser6sensorDir(). */
vrVector *vrPfVectorGetVWFromUser6sensorDir(vrVector *vector, int usernum, int sensor_num, vrDirection direction)
{
static	vrMatrix	mat_pos90x = { 1.0,0.0,0.0,0.0,  0.0,0.0,1.0,0.0,  0.0,-1.0,0.0,0.0,  0.0,0.0,0.0,1.0 };

	vrVectorGetVWFromUser6sensorDir(vector, usernum, sensor_num, direction);
	vrVectorTransformByMatrix(vector, vector, &mat_pos90x);

	return vector;
}


/****************************************************/
/* Performer version of vrMatrixGet6sensorValues(). */
vrMatrix *vrPfMatrixGet6sensorValues(vrMatrix *mat, int input_num)
{
static	vrMatrix	mat_pos90x = { 1.0,0.0,0.0,0.0,  0.0,0.0,1.0,0.0,  0.0,-1.0,0.0,0.0,  0.0,0.0,0.0,1.0 };

	vrMatrixGet6sensorValues(mat, input_num);
	vrMatrixPreMult(mat, &mat_pos90x);

	return mat;
}


/**********************************************************/
/* Performer version of vrMatrixGet6sensorValuesDirect(). */
vrMatrix *vrPfMatrixGet6sensorValuesDirect(vrMatrix *mat, vr6sensor *input)
{
static	vrMatrix	mat_pos90x = { 1.0,0.0,0.0,0.0,  0.0,0.0,1.0,0.0,  0.0,-1.0,0.0,0.0,  0.0,0.0,0.0,1.0 };

	vrMatrixGet6sensorValuesDirect(mat, input);
	vrMatrixPreMult(mat, &mat_pos90x);

	return mat;
}


/**********************************************************************/
/* Performer version of vrMatrixGet6sensorValuesDirectNoLastUpdate(). */
vrMatrix *vrPfMatrixGet6sensorValuesDirectNoLastUpdate(vrMatrix *mat, vr6sensor *input)
{
static	vrMatrix	mat_pos90x = { 1.0,0.0,0.0,0.0,  0.0,0.0,1.0,0.0,  0.0,-1.0,0.0,0.0,  0.0,0.0,0.0,1.0 };

	vrMatrixGet6sensorValuesDirectNoLastUpdate(mat, input);
	vrMatrixPreMult(mat, &mat_pos90x);

	return mat;
}


/*************************************************************/
/* Performer version of vrMatrixGet6sensorRawValuesDirect(). */
vrMatrix *vrPfMatrixGet6sensorRawValuesDirect(vrMatrix *mat, vr6sensor *input)
{
static	vrMatrix	mat_pos90x = { 1.0,0.0,0.0,0.0,  0.0,0.0,1.0,0.0,  0.0,-1.0,0.0,0.0,  0.0,0.0,0.0,1.0 };

	vrMatrixGet6sensorRawValuesDirect(mat, input);
	vrMatrixPreMult(mat, &mat_pos90x);

	return mat;
}


/********************************************************/
/* Performer version of vrVectorfGetRWFrom6sensorDir(). */
vrVectorf *vrPfVectorfGetRWFrom6sensorDir(vrVectorf *vector, int sensor_num, vrDirection direction)
{
static	vrMatrix	mat_pos90x = { 1.0,0.0,0.0,0.0,  0.0,0.0,1.0,0.0,  0.0,-1.0,0.0,0.0,  0.0,0.0,0.0,1.0 };
	vrVector	vectord;

	vrVectorfGetRWFrom6sensorDir(vector, sensor_num, direction);
	vectord.v[VR_X] = vector->v[VR_X];
	vectord.v[VR_Y] = vector->v[VR_Y];
	vectord.v[VR_Z] = vector->v[VR_Z];
	vrVectorTransformByMatrix(&vectord, &vectord, &mat_pos90x);
	vector->v[VR_X] = vectord.v[VR_X];
	vector->v[VR_Y] = vectord.v[VR_Y];
	vector->v[VR_Z] = vectord.v[VR_Z];

	return vector;
}


/**************************************************************/
/* Performer version of vrPfVectorfGetVWFromUser6sensorDir(). */
vrVectorf *vrPfVectorfGetVWFromUser6sensorDir(vrVectorf *vector, int user_num, int sensor_num, vrDirection direction)
{
static	vrMatrix	mat_pos90x = { 1.0,0.0,0.0,0.0,  0.0,0.0,1.0,0.0,  0.0,-1.0,0.0,0.0,  0.0,0.0,0.0,1.0 };
	vrVector	vectord;

	vrVectorfGetVWFromUser6sensorDir(vector, user_num, sensor_num, direction);
	vectord.v[VR_X] = vector->v[VR_X];
	vectord.v[VR_Y] = vector->v[VR_Y];
	vectord.v[VR_Z] = vector->v[VR_Z];
	vrVectorTransformByMatrix(&vectord, &vectord, &mat_pos90x);
	vector->v[VR_X] = vectord.v[VR_X];
	vector->v[VR_Y] = vectord.v[VR_Y];
	vector->v[VR_Z] = vectord.v[VR_Z];

	return vector;
}


/****************************************************/
/* Performer version of vrPointfGetRWFrom6sensor(). */
vrPointf *vrPfPointfGetRWFrom6sensor(vrPointf *real_point, int sensor_num)
{
static	vrMatrix	mat_pos90x = { 1.0,0.0,0.0,0.0,  0.0,0.0,1.0,0.0,  0.0,-1.0,0.0,0.0,  0.0,0.0,0.0,1.0 };
	vrPoint		pointd;

	vrPointfGetRWFrom6sensor(real_point, sensor_num);
	pointd.v[VR_X] = real_point->v[VR_X];
	pointd.v[VR_Y] = real_point->v[VR_Y];
	pointd.v[VR_Z] = real_point->v[VR_Z];
	vrPointTransformByMatrix(&pointd, &pointd, &mat_pos90x);
	real_point->v[VR_X] = pointd.v[VR_X];
	real_point->v[VR_Y] = pointd.v[VR_Y];
	real_point->v[VR_Z] = pointd.v[VR_Z];

	return real_point;
}


/********************************************************/
/* Performer version of vrPointfGetVWFromUser6sensor(). */
vrPointf *vrPfPointfGetVWFromUser6sensor(vrPointf *virtual_point, int usernum, int sensor_num)
{
static	vrMatrix	mat_pos90x = { 1.0,0.0,0.0,0.0,  0.0,0.0,1.0,0.0,  0.0,-1.0,0.0,0.0,  0.0,0.0,0.0,1.0 };
	vrPoint		pointd;

	vrPointfGetVWFromUser6sensor(virtual_point, usernum, sensor_num);
	pointd.v[VR_X] = virtual_point->v[VR_X];
	pointd.v[VR_Y] = virtual_point->v[VR_Y];
	pointd.v[VR_Z] = virtual_point->v[VR_Z];
	vrPointTransformByMatrix(&pointd, &pointd, &mat_pos90x);
	virtual_point->v[VR_X] = pointd.v[VR_X];
	virtual_point->v[VR_Y] = pointd.v[VR_Y];
	virtual_point->v[VR_Z] = pointd.v[VR_Z];

	return virtual_point;
}


/********************************************************************/
/* Performer version of vrEulerGetRWFrom6sensor().                  */
/* NOTE: unlike the other -Pf- versions of the input routines, this */
/*   one actually mimics the non-Pf version rather than calling the */
/*   non-Pf version and then converting the result.                 */
vrEuler *vrPfEulerGetRWFrom6sensor(vrEuler *euler, int sensor_num)
{
static	vrMatrix	mat_pos90x = { 1.0,0.0,0.0,0.0,  0.0,0.0,1.0,0.0,  0.0,-1.0,0.0,0.0,  0.0,0.0,0.0,1.0 };
	vrMatrix	mat_store;		/* temporary matrix storage */
	vrMatrix	*mat = &mat_store;	/* pointer to temp matrix storage */

	vrMatrixGet6sensorValues(mat, sensor_num);
	return vrEulerSetFromMatrix(euler, vrMatrixPreMult(mat, &mat_pos90x));
}


/*******************************************************/
/* Performer version of vrEulerGetVWFromUser6sensor(). */
vrEuler *vrPfEulerGetVWFromUser6sensor(vrEuler *euler, int usernum, int sensor_num)
{
static	vrMatrix	mat_pos90x = { 1.0,0.0,0.0,0.0,  0.0,0.0,1.0,0.0,  0.0,-1.0,0.0,0.0,  0.0,0.0,0.0,1.0 };
	vrMatrix	mat_store;		/* temporary matrix storage */
	vrMatrix	*mat = &mat_store;	/* pointer to temp matrix storage */

	vrMatrixPreMult(vrMatrixGet6sensorValues(mat, sensor_num), vrContext->input->users[usernum]->vw2rw_xform);
	return vrEulerSetFromMatrix(euler, vrMatrixPreMult(mat, &mat_pos90x));
}

#endif /* GFX_PERFORMER } */


	/**************************************************************/
	/**************************************************************/


/*****************************************************************************/
/* Copy all the current input values into a secondary storage location (also */
/*   part of each inputs data structure). ...                                */
/* TODO: it would be more efficient to use a local variable to store the     */
/*   pointer to each input, as it is referenced 4 times per loop, and        */
/*   requires 2 dereferences and an addition.                                */
void vrInputFreeze(vrContextInfo *context)
{
	vrInputInfo	*vrInputs = context->input;
	int		count;

	/*************************/
	/* freeze all the 2-ways */
	for (count = 0; count < vrInputs->num_2ways; count++) {
		if (vrInputs->switch2[count] != NULL) {
			vrLockWriteSet(vrInputs->switch2[count]->lock);
			vrInputs->switch2[count]->visren_value = vrInputs->switch2[count]->value;
			vrLockWriteRelease(vrInputs->switch2[count]->lock);
		}
	}

	/*************************/
	/* freeze all the N-ways */
	for (count = 0; count < vrInputs->num_Nways; count++) {
		if (vrInputs->switchN[count] != NULL) {
			vrLockWriteSet(vrInputs->switchN[count]->lock);
			vrInputs->switchN[count]->visren_value = vrInputs->switchN[count]->value;
			vrLockWriteRelease(vrInputs->switchN[count]->lock);
		}
	}

	/****************************/
	/* freeze all the valuators */
	for (count = 0; count < vrInputs->num_valuators; count++) {
		if (vrInputs->valuator[count] != NULL) {
			vrLockWriteSet(vrInputs->valuator[count]->lock);
			vrInputs->valuator[count]->visren_value = vrInputs->valuator[count]->value;
			vrLockWriteRelease(vrInputs->valuator[count]->lock);
		}
	}

	/****************************/
	/* freeze all the 6-sensors */
	for (count = 0; count < vrInputs->num_6sensors; count++) {
		if (vrInputs->sensor6[count] != NULL) {
			vrLockWriteSet(vrInputs->sensor6[count]->lock);
			*(vrInputs->sensor6[count]->visren_position) = *(vrInputs->sensor6[count]->position);
			vrLockWriteRelease(vrInputs->sensor6[count]->lock);
		}
	}

	/****************************/
	/* freeze all the N-sensors */
	for (count = 0; count < vrInputs->num_Nsensors; count++) {
		if (vrInputs->sensorN[count] != NULL) {
			vrLockWriteSet(vrInputs->sensorN[count]->lock);
			*(vrInputs->sensorN[count]->visren_values) = *(vrInputs->sensorN[count]->values);
			vrLockWriteRelease(vrInputs->sensorN[count]->lock);
		}
	}
}


	/**************************************************************/
	/**************************************************************/



/**************************************************************************/
/* UNIMPLIMENTED: vrInputMakeChecksum(): create a checksum value that can */
/*   be used to make sure the input structure has not been corrupted by a */
/*   stray memory write.                                                  */
/* As it happens, I found the bug I was tracking down before implementing */
/*   this as a means to figured out that bug, so this hasn't yet been     */
/*   implemented, and perhaps never will.                                 */
int vrInputMakeChecksum(vrGenericInput *input)
{
#ifdef UNPICKY_COMPILER
	int	bytes_to_check;
#endif

	/* TODO: I didn't get around to implementing this before finding the bug */
	/*   that made me want to have this -- not sure if we will implement it. */

	return -10;	/* return a dummy value for now */
}


/**************************************************************************/
/* UNIMPLIMENTED: vrInputCheckChecksum(): check a checksum value to make  */
/*   make sure the input structure has not been corrupted by a stray      */
/*   memory write.                                                        */
/* As it happens, I found the bug I was tracking down before implementing */
/*   this as a means to figured out that bug, so this hasn't yet been     */
/*   implemented, and perhaps never will.                                 */
int vrInputCheckChecksum(vrGenericInput *input)
{
#ifdef UNPICKY_COMPILER
	int	bytes_to_check;
#endif

	/* TODO: I didn't get around to implementing this before finding the bug */
	/*   that made me want to have this -- not sure if we will implement it. */

	return 0;
}


